:- module(cvc4interface, [init_cvc4interface/0,
                          reset_cvc4interface/0, % TODO: make more lightweight once cvc 1.5 is out
                          cvc4_interface_call/1]).

:- use_module('../../src/module_information.pl').
:- module_info(group,experimental).
:- module_info(description,'This is the interface between ProB and the CVC4 SMT solver').

/*
:- use_module(probsrc(bsyntaxtree), [map_over_typed_bexpr/2,
                            find_typed_identifier_uses/3,
                            get_texpr_type/2,
                            get_texpr_id/2]).
*/
foreign_resource('cvc4interface', [init,
                                   mk_var,
                                   mk_bounded_var,
                                   mk_int_const,
                                   mk_string_const,
                                   mk_bool_const,
                                   mk_empty_set,
                                   mk_record_const,
                                   mk_record_field,
                                   mk_set,
                                   mk_sort,
                                   mk_sort_with_cardinality,
                                   pop_frame, push_frame,
                                   setup_enumerated_set,
                                   mk_op,
                                   mk_op_arglist,
                                   query,
                                   get_string,
                                   mk_quantifier,
                                   conjoin_negated_state,
                                   add_interpreter_constraint,
                                   mk_couple]).

% will be called by solver_interface:
:- public mk_var/3, mk_bounded_var/3, mk_empty_set/2, mk_int_const/2, mk_bool_const/2, mk_string_const/2.
:- public mk_set/2, mk_sort/1, mk_sort_with_cardinality/2, mk_record_const/3, mk_record_field/4.
:- public setup_enumerated_set/1, mk_op/4, mk_quantifier/4, mk_op_arglist/3, query/3, get_string/2, conjoin_negated_state/3.
:- public pop_frame/0, push_frame/0, add_interpreter_constraint/4, mk_couple/4.

% function declarations
foreign(init, c, init).
foreign(reset, c, reset).
foreign(mk_var, c, mk_var(+term, +string, [-integer])). % creates a variable; first argument is an atom of the type,
                                                         % second argument is the identifier
foreign(mk_bounded_var, c, mk_bounded_var(+term, +string, [-integer])).
foreign(mk_empty_set, c, mk_empty_set(+term, [-integer])).
foreign(mk_int_const, c, mk_int_const(+integer, [-integer])). % creates an int constant
foreign(mk_bool_const, c, mk_bool_const(+string, [-integer])). % creates a bool constant
foreign(mk_string_const, c, mk_string_const(+atom, [-integer])).
foreign(mk_set, c, mk_set(+term, [-integer])).
foreign(mk_sort, c, mk_sort(+atom)).
foreign(mk_sort_with_cardinality, c, mk_sort_with_cardinality(+atom,+integer)).
foreign(mk_record_const, c, mk_record_const(+term, +term, [-integer])).
foreign(mk_record_field, c, mk_record_field(+integer, +integer, +atom, [-integer])).
foreign(setup_enumerated_set, c, setup_enumerated_set(+term)).
foreign(mk_op, c, mk_op(+string, +integer, +integer, [-integer])).
foreign(mk_quantifier, c, mk_quantifier(+string, +term, +integer, [-integer])).
foreign(mk_op_arglist, c, mk_op_arglist(+string, +term, [-integer])).
foreign(query, c, query(+integer,+integer,[-atom])).
foreign(get_string, c, get_string(+integer, [-atom])).
foreign(conjoin_negated_state, c, conjoin_negated_state(+term,+integer,[-integer])).
foreign(pop_frame, c, pop_frame).
foreign(push_frame, c, push_frame).
foreign(add_interpreter_constraint, c, add_interpreter_constraint(+integer,+atom,+integer,[-term])).
foreign(mk_couple, c, mk_couple(+term,+integer,+integer,[-integer])).


%:- load_foreign_resource(cvc4interface).
:- dynamic is_initialised/0.

init_cvc4interface :- is_initialised,!.
init_cvc4interface :-
    catch(load_foreign_resource(library(cvc4interface)),_,fail),
    init,
    assert(is_initialised).

reset_cvc4interface :- reset.

:- meta_predicate cvc4_interface_call(+).
cvc4_interface_call(Call) :-
    call(Call).
