//
//  fair_counterexample.c
//  
//
//  Created by Ivaylo Dobrikov on 02/11/15.
//
//

#include "fair_counterexample.h"

#ifdef LTLFAIRNESSDEBUG

void print_fairness_arg(SP_term_ref argument, char const *type) {
    printf("fairness argument (%s): ", type);
    SP_pred_ref pred;
    /* Look up pp_tp/2 */
    if (!(pred = SP_predicate("pp_tp2",2,"ltl_translate"))) {
        fprintf(stderr, "Could not find ltl_translate:pp_tp/2\n");
        exit(1);
    }
    SP_term_ref res, arg = SP_new_term_ref();
    SP_qid goal;
    SP_put_variable(res = SP_new_term_ref());
    char const *text=NULL;
    SP_put_term(arg,argument);
    if (!(goal = SP_open_query(pred,arg,res))) {
        fprintf(stderr, "Failed to open query.\n");
        exit(1);
    }
    while (SP_next_solution(goal)==SP_SUCCESS) {
        SP_get_string(res,&text);
        printf("%s", text);
    }
    SP_close_query(goal);
    printf("\n");
}

static SP_term_ref get_node_id_term(ltlatom *atom) {
    SP_term_ref node_id;
    ltlnode *node = atom->node;
    if (node->type == int_node) {
        node_id = SP_new_term_ref();
        SP_put_integer(node_id, node->id.id);
    } else {
        node_id = node->id.ref;
    }
    return node_id;
}

#endif

/* returns the bitfield that represents the goals encountered
 * in the atom of the corresponding item */
static bitfield* until_goals_of_fairloopitem(fairloopitem *item) {
    return (bitfield*) (((uintptr_t) item) + sizeof(fairloopitem));
}

/* returns the bitfield that represents the used untils (untils which are evaluated to true)
 * in the atom of the corresponding item */
static bitfield* used_untils_of_fairloopitem(ltl_formula *f, fairloopitem *item) {
    return (bitfield*) (((uintptr_t) item) + sizeof(fairloopitem) + get_bitfield_memsize(f->number_of_untils));
}

void print_item(fairloopitem* item) {
    ltlatom *atom = item->atom;
    int nodeid = (atom->node)->id.id;
    int atomid = atom->tarjan_number;
    unsigned int transition_id = (item->executed_transition).transition_id;
    printf("loopitem { nodeid: %d, atomid: % d, transition_id: %d, \n",nodeid,atomid,transition_id);
    printf("           wf: ");
    show_bitfield(item->wf_executed,item->wf_size);
    printf("\n           sf: ");
    show_bitfield(item->sf_executed,item->sf_size);
    printf(" }\n");
}

void print_stack_content(loop_data *stack) {
    int i;
    fairloopitem *item;
    printf("STACK CONTENT:\n");
    for (i=stack->stacksize-1; i>=0; i--) {
        item = (fairloopitem*) get_array_element(&(stack->stack),i);
        printf("stack item: ");
        print_item(item);
    }
}

static void compute_untils_for_fairloopitems(ltl_formula *f, ltlarray *nodes, unsigned int first, unsigned int size,
                                             bitfield *used_untils, bitfield *encountered_goals) {
    fairloopitem *item;
    unsigned int i;
    for (i=first; i<size; i++) {
        item = (fairloopitem*) get_array_element(nodes, i);
        if (used_untils != NULL) {
            add_set_bits(used_untils_of_fairloopitem(f,item), used_untils, f->number_of_untils);
        }
        if (encountered_goals != NULL) {
            add_set_bits(until_goals_of_fairloopitem(item), encountered_goals, f->number_of_untils);
        }
    }
}

static int check_if_untils_are_matched_in_fair_loop(ltl_formula *f, ltlarray *nodes, int cycle_start, int size) {
    int result;
    bitfield *used_untils = f->tmp_evaluation;
    bitfield *encountered_goals = f->tmp_evaluation2;
    clear_bitfield(used_untils, f->number_of_untils);
    clear_bitfield(encountered_goals, f->number_of_untils);
    
    compute_untils_for_fairloopitems(f, nodes, cycle_start, size, used_untils, encountered_goals);
    if (is_sub_bitfield(used_untils, encountered_goals, f->number_of_untils)) {
        IFFAIRNESSDEBUG(printf("  Bitfield of used untils: ");
                        show_bitfield(used_untils,f->number_of_untils);
                        printf("\n  Bitfield of goals      : ");
                        show_bitfield(encountered_goals,f->number_of_untils);
                        printf("\n") );
        compute_untils_for_fairloopitems(f, nodes, 0, size, used_untils, encountered_goals);
        result = is_sub_bitfield(encountered_goals, used_untils, f->number_of_untils);
    } else {
        result = 0;
    }
    return result;
}

/*
 * calculates which untils are true and which goals are encountered in the item
 * the result is stored in the loop item
 */
static void store_used_untils(ltl_formula *f, fairloopitem *item) {
    bitfield *goalbuffer, *usedbuffer, *evalbuffer;
    unsigned int i, until;
    goalbuffer = until_goals_of_fairloopitem(item);
    usedbuffer = used_untils_of_fairloopitem(f,item);
    clear_bitfield(goalbuffer, f->number_of_untils);
    clear_bitfield(usedbuffer, f->number_of_untils);
    evalbuffer = get_atom_evaluation(item->atom);
    for (i=0; i<f->number_of_untils; i++) {
        until = f->untils[i];
        if (eval(evalbuffer, &(f->closure[until].args.case_until.goal))) {
            setbit(goalbuffer, i, 1);
        }
        if (getbit(evalbuffer,until)) {
            setbit(usedbuffer, i, 1);
        }
    }
}

static modelitem create_model_item(ltlatom *atom, unsigned int transition_id) {
    modelitem item;
    item.atom = atom;
    // the transition id is sometimes needed to know by which node transition the
    // atom is reached. especially when a fair loop should be found one has to
    // make sure that the relevant fairness transition are executed along the loop
    item.transition_id = transition_id;
    return item;
}


static SP_term_ref get_node_id_term(ltlatom *atom) {
    SP_term_ref node_id;
    ltlnode *node = atom->node;
    if (node->type == int_node) {
        node_id = SP_new_term_ref();
        SP_put_integer(node_id, node->id.id);
    } else {
        node_id = node->id.ref;
    }
    return node_id;
}

static int match_operations(ltlatom *atom, SP_term_ref *event, SP_term_ref *op_id) {
    SP_pred_ref pred;
    /* Look up match_transition/3 */
    if (!(pred = SP_predicate("match_transition",3,"ltl_fairness"))) {
        fprintf(stderr, "Could not find ltl_fairness:match_transition/3\n");
        exit(1);
    }
    SP_term_ref node_id = get_node_id_term(atom);
    int res = SP_query(pred,node_id,*event,*op_id);
    
    switch (res) {
        case SP_SUCCESS:
            return 1;
        case SP_FAILURE:
            return 0;
        default: // SP_ERROR
            fprintf(stderr, "Unexpected result on calling ltl_fairness:match_transition/3 from match_operations\n");
            exit(1);
    }
}

int match_transition_fair(SP_term_ref event, ltlatom *atom) {
    SP_pred_ref pred;
    /* Look up get_transition_id/4 */
    if (!(pred = SP_predicate("get_transition_id",4,"ltl_fairness"))) {
        fprintf(stderr, "Could not find ltl_fairness:get_transition_id/4\n");
        exit(1);
    }
    SP_term_ref node_id, trans_id, dst_id;
    node_id = get_node_id_term(atom);
    
    // variables are likely not to be needed
    SP_put_variable(trans_id = SP_new_term_ref());
    SP_put_variable(dst_id = SP_new_term_ref());
    
    int res = SP_query(pred,event,node_id,trans_id,dst_id);
    
    switch (res) {
        case SP_SUCCESS:
            return 1;
        case SP_FAILURE:
            return 0;
        default: // SP_ERROR
            fprintf(stderr, "Unexpected result on calling ltl_fairness:get_transition_id/4 from match_transition\n");
            exit(1);
    }
}

static int match_transition_everywhere(SP_term_ref event, ltlarray *nodes, int cycle_start, int size) {
    
    if (size > 0 && cycle_start > -1) {
        unsigned int i;
        fairloopitem *item;
        ltlatom *atom;
        for (i=cycle_start; i < size; i++) {
            item = (fairloopitem *) get_array_element(nodes, i);
            atom = item->atom;
            if (!match_transition_fair(event,atom)) {
                return 0;
            }
        }
        return 1;
    } else {
        fprintf(stderr, "No atom in the loop (match_transition_everywhere).");
        exit(1);
    }
}

static int match_transition_enabled(SP_term_ref event, ltlarray *nodes, int cycle_start, int size) {
    if (size > 0 && cycle_start > -1) {
        int i;
        fairloopitem *item;
        ltlatom *atom;
        for (i=cycle_start; i < size; i++) {
            item = (fairloopitem *) get_array_element(nodes, i);
            atom = item->atom;
            if (match_transition_fair(event,atom)) {
                return 1;
            }
        }
        return 0;
    } else {
        fprintf(stderr, "No atom in the loop (match_transition_everywhere).");
        exit(1);
    }
}

static int match_executed_transition(ltl_formula *f, SP_term_ref event, ltlarray *nodes,
                                     int cycle_start, fairloopitem *last_item, int size) {
    
    if (size > 0 && cycle_start > -1) {
        fairloopitem *item, *previous_item;
        SP_term_ref op_id;
        unsigned int i = cycle_start;
        previous_item = (fairloopitem*) get_array_element(nodes,cycle_start);
        while (i < size-1) {
            item = (fairloopitem*) get_array_element(nodes,i+1);
            op_id = (item->executed_transition).op_id;
            if (match_operations(previous_item->atom,&event,&op_id)) {
                return 1;
            }
            previous_item = item;
            i++;
        }
        // at last we need to check the execution of the transition
        // by which the entry is reached
        op_id = (last_item->executed_transition).op_id;
        return match_operations(previous_item->atom,&event,&op_id);
    } else {
        fprintf(stderr, "match_executed_transition: Loop array must have at least one element.");
        exit(1);
    }
}

itemtransition create_itemtransition(unsigned int transition_id, SP_term_ref *op) {
    itemtransition transition;
    transition.transition_id = transition_id;
    transition.op_id = *op;
    return transition;
}

itemtransition get_item_transition(ltl_formula *f, unsigned int transition_id) {
    SP_term_ref op;
    unsigned int nodetransid;
    get_nodetransop_id(f, transition_id, &nodetransid);
    get_nodetransop(f, nodetransid, NULL, NULL, &op);
    return create_itemtransition(transition_id,&op);
}

int equal_items(fairloopitem *item1, fairloopitem *item2, int comapre_executed_bitfields) {
    int result =
                (item1->atom == item2->atom) &&
                ((item1->executed_transition).transition_id == (item2->executed_transition).transition_id);
    if (comapre_executed_bitfields) {
        return result && equal_bitfields(item1->wf_executed,item2->wf_executed,item1->wf_size) &&
                         equal_bitfields(item1->sf_executed,item2->sf_executed,item1->sf_size);
    }
    return result;
}

int item_in_array(fairloopitem *item, ltlarray *array, int size) {
    int i;
    fairloopitem *array_item;
    for (i=0; i<size; i++) {
        array_item = (fairloopitem*) get_array_element(array,i);
        if (equal_items(item,array_item,1)) {
            return 1;
        }
    }
    return 0;
}

int get_first_position_of_item(fairloopitem *item, ltlarray *array, int size) {
    int i;
    int result = -1;
    fairloopitem *array_item;
    for (i=0; i<size; i++) {
        array_item = (fairloopitem*) get_array_element(array,i);
        if (equal_items(item,array_item,0)) {
            result = i;
            break;
        }
    }
    if (result < 0) {
        fprintf(stderr,
                "item could not be find. there is probably internal state error\n");
        exit(1);
    } else {
        return result;
    }
}

// taking the maximal loop execution
// TODO: consider whether we could not "somehow" decide to pick shorter loops execution
//       for the sake of minimal counter-examples.
int item_in_stack(fairloopitem *item, loop_data *l, int *pos) {

    unsigned int size = l->stacksize;
    if (item_in_array(item,&(l->stack),size)) {
        *pos = get_first_position_of_item(item, &(l->stack),size);
        return 1;
    } else {
        return 0;
    }
}

int item_visited(fairloopitem *item, loop_data *stack) {
    int size = stack->number_of_nodes;
    return item_in_array(item,&(stack->nodes),size);
}

void init_stack(ltl_formula *f, loop_data *l) {
    /* we store for each fairloopitem a bitfield of the encountered goals of UNTILs */
    const size_t fairloopitem_size = sizeof(fairloopitem) + 2*get_bitfield_memsize(f->number_of_untils);

    init_array(&(l->stack), fairloopitem_size, NULL, "loop_data (stack)");
    init_array(&(l->nodes), fairloopitem_size, NULL, "loop_data (nodes)");
    
    l->stacksize = 0;
    l->arraysize = 0;
    l->number_of_nodes = 0;
}

int non_empty_stack(loop_data *stack) {
    return (stack->stacksize==0) ? 0 : 1;
}

int fairness_assumption_enabled_in_atom(ltlatom *src, SP_term_ref *fairness_args, SP_term_ref *enabled_op) {
    int i,size;
    int result = 0;
    
    SP_term_ref *arguments = prolog_list_to_array(*fairness_args, &size);
    for (i=0; i<size; i++) {
        if (match_operations(src,&arguments[i],enabled_op)) {
            result = 1;
            break;
        }
    }
    ltl_free(arguments);
    return result;
}

//itemtransition *assumptions_enabled(ltl_formula *f, ltlatom *atom, SP_term_ref *assumptions) {
//    int i,size = 0;
//    unsigned int transid, nodetransid;
//    itemtransition* wf_enabled = (itemtransition*) ltl_malloc(sizeof(itemtransition)*size);
//    SP_term_ref enabled_op;
//    for (i=0; i<atom->number_of_transitions; i++) {
//        transid = atom->first_transition_id+i;
//        get_nodetransop_id(f, transid, &nodetransid);
//        get_nodetransop(f, nodetransid, NULL, NULL, &enabled_op);
//        if (fairness_assumption_enabled_in_atom(atom,assumptions,&enabled_op)) {
//            size++;
//            wf_enabled = (itemtransition*) ltl_realloc(wf_enabled,sizeof(itemtransition)*size);
//            wf_enabled[size-1] = create_itemtransition(transid,&enabled_op);
//        }
//    }
//    return wf_enabled;
//}

bitfield *create_new_bitfield(ltl_formula *f, bitfield* field, ltlatom *previous_atom, unsigned int atransition_id,
                              SP_term_ref *fairness_args, int field_length) {

    int i,size;
    SP_term_ref *arguments = prolog_list_to_array(*fairness_args, &size);

    SP_term_ref op;
    unsigned int nodetransid;
    
    get_nodetransop_id(f, atransition_id, &nodetransid);
    get_nodetransop(f, nodetransid, NULL, NULL, &op);

    bitfield *new_bitfield = (bitfield*) ltl_malloc(get_bitfield_memsize(field_length));
    for (i=0; i<field_length; i++) {
        if (match_operations(previous_atom,&arguments[i],&op)) {
            setbit(new_bitfield,i,1);
        } else {
            setbit(new_bitfield,i,getbit(field,i));
        }
    }
    
    ltl_free(arguments);
    return new_bitfield;
}

fairloopitem *initialise_item(ltl_formula *f, found_scc_type *found,
                     ltlatom *atom, unsigned int transition_id, fairloopitem *previous_item) {
    
    fairloopitem *item = (fairloopitem*) ltl_malloc(sizeof(fairloopitem) + 2*get_bitfield_memsize(f->number_of_untils));
    
    bitfield *sf_executed, *wf_executed;
    int wf_size = (found->fairness_assumptions).weak_size;
    int sf_size = (found->fairness_assumptions).strong_size;

    item->atom = atom;
    item->executed_transition = get_item_transition(f,transition_id);
//    item->wf_enabled = assumptions_enabled(f,atom,&(found->fairness_assumptions.weak));
//    item->sf_enabled = assumptions_enabled(f,atom,&(found->fairness_assumptions.strong));
    /* current fairness assumptions evaluation */
    if (previous_item == NULL) {
        wf_executed = (bitfield *) ltl_malloc(get_bitfield_memsize(wf_size));
        sf_executed = (bitfield *) ltl_malloc(get_bitfield_memsize(sf_size));
        clear_bitfield(wf_executed,wf_size);
        clear_bitfield(sf_executed,sf_size);
    } else {
        wf_executed = create_new_bitfield(f,previous_item->wf_executed, previous_item->atom, transition_id,
                                          &(found->fairness_assumptions).weak, wf_size);
        sf_executed = create_new_bitfield(f,previous_item->sf_executed, previous_item->atom, transition_id,
                                          &(found->fairness_assumptions).strong, sf_size);
    }
    
    item->wf_executed = wf_executed; // executed weak fairness arguments so far
    item->wf_size = wf_size;
    item->sf_executed = sf_executed; // executed strong fairness arguments so far
    item->sf_size = sf_size;
    
    store_used_untils(f, item);

    return item;
}

void delete_item(fairloopitem *item) {
    ltl_free(item->wf_executed);
    ltl_free(item->sf_executed);
}

void delete_loop_data(loop_data *l) {
    fairloopitem *item;
    unsigned int i;
    for (i=0; i<l->number_of_nodes; i++) {
        item = (fairloopitem*) get_array_element(&(l->nodes),i);
//        ltl_free(item->wf_enabled);
//        ltl_free(item->sf_enabled);
        ltl_free(item->wf_executed);
        ltl_free(item->sf_executed);
    }
    delete_array(&(l->nodes));
    delete_array(&(l->stack));
}

static void push_to_stack(ltl_formula *f, loop_data *l, fairloopitem *new_item) {
    
    if (l->stacksize == l->arraysize) {
        assure_arraysize(&(l->stack),l->arraysize+1);
        l->arraysize++;
    }
    fairloopitem *item =(fairloopitem*) get_array_element(&(l->stack),l->stacksize);

    if (item != NULL) {
        /* copy new_item to the item in stack being pushed */
        item->atom = new_item->atom;
        item->executed_transition = new_item->executed_transition;
//        item->wf_enabled = new_item->wf_enabled;
//        item->sf_enabled = new_item->sf_enabled;
        item->wf_executed=new_item->wf_executed;
        item->wf_size = new_item->wf_size;
        item->sf_executed=new_item->sf_executed;
        item->sf_size = new_item->sf_size;
        
        l->stacksize++;
        
        store_used_untils(f, item);
    } else {
        fprintf(stderr, "save_item (fair_counterexample.c): nextitem is NULL.\n");
    }
}

static void add_to_nodes(ltl_formula *f, loop_data *l, fairloopitem *new_item) {
    if(!item_in_array(new_item,&(l->nodes),l->number_of_nodes)) {
        assure_arraysize(&(l->nodes),l->number_of_nodes+1);
        fairloopitem *item =(fairloopitem*) get_array_element(&(l->nodes),l->number_of_nodes);
        if (item != NULL) {
            /* copy new_item to the item in stack being pushed */
            item->atom = new_item->atom;
            item->executed_transition = new_item->executed_transition;
//            item->wf_enabled = new_item->wf_enabled;
//            item->sf_enabled = new_item->sf_enabled;
            item->wf_size = new_item->wf_size;
            item->wf_executed=new_item->wf_executed;
            item->sf_size = new_item->sf_size;
            item->sf_executed=new_item->sf_executed;
            
            l->number_of_nodes++;
            store_used_untils(f, item);
        } else {
            fprintf(stderr, "save_item (fair_example.c): nextitem is NULL.\n");
        }
    }
}

fairloopitem *top_stack(loop_data *l) {
    if (l->stacksize>0) {
        return (fairloopitem*) get_array_element(&(l->stack),l->stacksize-1);
    } else {
        fprintf(stderr, "top_stack() (fair loop search): empty stack!");
        exit(1);
    }
}

void pop_stack(loop_data *l) {
    if (l->stacksize > 0) {
        l->stacksize--;
    } else {
        fprintf(stderr, "pop_stack() (fair loop search): empty stack!");
        exit(1);
    }
}

void rewind_loop(found_scc_type *fscc, loop_data *stack, modelitem **result, int *resultlength) {
    *resultlength = fscc->number_of_path_elements + stack->stacksize;
    *result = (modelitem*) ltl_malloc(sizeof(modelitem)*(*resultlength));
    /* copy the path into the SCC to the beginning of the counter-example */
    unsigned int i;
    for (i=0; i<fscc->number_of_path_elements; i++) {
        (*result)[i] = create_model_item(fscc->path[i],0);
    }
    fairloopitem *item;
//    print_stack_content(stack);
    for (i=0; i<stack->stacksize; i++) {
        item = (fairloopitem*) get_array_element(&(stack->stack),(stack->stacksize-1-i));
        (*result)[*resultlength-1-i] = create_model_item(item->atom,(item->executed_transition).transition_id);
    }
}

int check_enabled_executed(ltl_formula *f, SP_term_ref fairness_args,
                           loop_data *l, int cycle_start, fairloopitem* last_item,
                           int (*enabled_fun)(SP_term_ref,ltlarray*,int,int)) {
    int result = 1;
    int args_size;
    bitfield *enabled, *executed;
    SP_term_ref *arguments = prolog_list_to_array(fairness_args, &args_size);
    if (args_size > 0) {
        enabled  = (bitfield *) ltl_malloc(get_bitfield_memsize(args_size));
        executed = (bitfield *) ltl_malloc(get_bitfield_memsize(args_size));
        unsigned int i;
        for (i=0; i<args_size; i++) {
            if ((*enabled_fun)(arguments[i],&(l->stack),cycle_start,l->stacksize)) {
                setbit(enabled,i,1);
            } else {
                setbit(enabled,i,0);
            }
            if (match_executed_transition(f,arguments[i],&(l->stack),cycle_start,last_item,l->stacksize)) {
                setbit(executed,i,1);
            } else {
                setbit(executed,i,0);
            }
        }
        IFFAIRNESSDEBUG(printf("  Bitfield of enabled arguments : ");
                        show_bitfield(enabled,args_size);
                        printf("\n  Bitfield of executed arguments: ");
                        show_bitfield(executed,args_size);
                        printf("\n") );
        result = is_sub_bitfield(enabled,executed,args_size);
        ltl_free(enabled);
        ltl_free(executed);
    }
    ltl_free(arguments);
    return result;
}

int inspect_cycle(ltl_formula *f, found_scc_type *found, loop_data *l,
                  int cycle_start, fairloopitem *last_item) {
    /* check whether the loop is fair */
    int result;
    // weak fairness
    int weak_fair_result = check_enabled_executed(f,(found->fairness_assumptions).weak,l,cycle_start,last_item,&match_transition_everywhere);
    
    if (!weak_fair_result) {
        return 0;
    } else {
        // strong fairness
        result = check_enabled_executed(f,(found->fairness_assumptions).strong,l,cycle_start,last_item,&match_transition_enabled);
        return (result && check_if_untils_are_matched_in_fair_loop(f,&(l->nodes),cycle_start,l->number_of_nodes));
    }
}

/* The function performs a depth-first search for finding valid loops i.r.t.
   the fairness constraints and which is a self-fulfilling loop.
   An item (or node) in the search graph has the following fields:
    - atom: a pointer to the respective atom of the state space of ltl_formula
    - excuted_transition: the transition id and SP_term of the represented transition,
                          note that transition_id is the tranisition id of the atom search graph
                          and not of that of the origin state space of the checked model
    - wf_executed: a pointer to the bit array keeping track of the executed weak fair fairness arguments
                   along the current search path
    - wf_size: the size of the respective weka fair bit array
    - sf_executed: a pointer to the bit array keeping track of the executed strong fair fairness arguments
                   along the current search path
    - sf_size: the size of the respective strong fair bit array
 */
void get_fair_loop(ltl_formula *f, found_scc_type *found,
                          modelitem **result, int *resultlength, int *loop_length) {
//    print_number_of_open_mallocs("on the start");
    ltlatom *atom, *nextatom;
    int found_valid_loop = 0;
    unsigned int transition_id;
    fairloopitem *item, *nextitem;
    ltlatom **scc = found->scc;
    unsigned int atomnumber = found->number_of_scc_atoms;
    int valid_cycle=0;
    int cycle_start = -1;
    
    loop_data df_stack;
    init_stack(f,&df_stack);
    
    nextitem = initialise_item(f,found,found->scc_entry,0,NULL);
    push_to_stack(f,&df_stack,nextitem);
    add_to_nodes(f,&df_stack,nextitem);
    ltl_free(nextitem);

    unsigned int i;
    int pushed_to_stack;
    while (!found_valid_loop && non_empty_stack(&df_stack)) {
        item = top_stack(&df_stack);
        atom = item->atom;
        pushed_to_stack=0;
        
        // TODO: consider a more efficient loop search where we prefer executing of fairness arguments
        for (i=0; i<atom->number_of_transitions; i++) {
            transition_id = atom->first_transition_id + i;
            nextatom = get_atom_trans(f, transition_id);
            /* considering transitions only to atoms in the stack */
            if (is_in_scc(nextatom, atomnumber, scc)) {
                nextitem = initialise_item(f,found,nextatom,transition_id,item);
                if (item_in_stack(nextitem,&df_stack,&cycle_start)) {
                    /* cycle found
                       check found loop and if the loop fulfils all requirements
                       return the loop as a result */
//                    push_to_stack(f,&df_stack,nextitem);
                    valid_cycle = inspect_cycle(f,found,&df_stack,cycle_start,nextitem);
                    if (valid_cycle) {
                        found_valid_loop = 1;
                        rewind_loop(found,&df_stack,result,resultlength);
                        *loop_length = df_stack.stacksize - cycle_start;
                    }
                    // freeing just the bitfields of nextitem
                    // nextitem was already encountered during the search and thus
                    // would not be added to the array of item-nodes. Therefore,
                    // we need to delete the bitfields of the nextitem at that place
                    // as they wont be deleted later
                    delete_item(nextitem);
                } else {
                    if (!item_visited(nextitem,&df_stack)) {
                        /* nextitem encountered for first time,
                           push to stack and continue exploration in df manner */
                        push_to_stack(f,&df_stack,nextitem);
                        add_to_nodes(f,&df_stack,nextitem);
                        pushed_to_stack = 1;
                    } else {
                        delete_item(nextitem);
                    }
                }
                ltl_free(nextitem);
                if (pushed_to_stack || found_valid_loop) {
                    // we continue to the next item or
                    // terminate because a valid loop was found (in both cases we exit the for-loop)
                    break;
                }
            }
        }
        // in case that all successor states are visited
        if (!pushed_to_stack) {
            pop_stack(&df_stack);
        }
    }
    if (!found_valid_loop) {
        fprintf(stderr,
                "no loop found in SCC, this is a bug in the model checker!\n");
        fprintf(stderr,
                "A file containing a dump called 'ltl_graphdump.pl' has been created\n");
        exit(1);
    }
    delete_loop_data(&df_stack);
//    print_number_of_open_mallocs("On the end");
}
