:- module(ltsmin_trace, [csv_to_trace/3]).

:- use_module(library(lists)).
:- use_module(library(fastrw)).
:- use_module(probsrc(tools), [split_chars/3, split_complex_sep/3]).

:- use_module(extension('ltsmin/ltsmin_c_interface'), [read_hex_str/2, free_str/1, loadfr/0,atom_charptr/2]).
:- use_module(extension('ltsmin/ltsmin'), [generate_bindings/3]).


read_all_lines(Acc, R) :-
    read_line(L),
    (L == end_of_file
      -> Acc = R
      ; read_all_lines([L|Acc], R)).


read_trace(Filename, TypeInfo, Lines, ButLastLines) :-
    open(Filename, read, Stream),
    see(Stream),
    read_all_lines([], ReversedLines),
    seen,
    ReversedLines = [_LastLine|ButLastLinesRev],
    reverse(ReversedLines, [TypeInfo|Lines]),
    reverse(ButLastLinesRev, [TypeInfo|ButLastLines]).


%% HACK: some values are not pretty printed into the #hexcode# representation
%% TODO: either fix LTSmin to do so or handle these cases separately (pk, 30.11.2017)
fast_buf_hex_to_term("\"D]\"", []) :- !.
fast_buf_hex_to_term(HexStr, Term) :-
    read_hex_str(HexStr, Tmp),
    fast_buf_read(Term, Tmp),
    free_str(Tmp).



line_to_state(VariableList, Line, State) :-
    split_chars(Line, ",", Vals),
    length(VariableList, Len),
    length(BinState, Len),
    prefix(Vals, BinState),
    maplist(fast_buf_hex_to_term, BinState, UnboundState),
    generate_bindings(UnboundState, VariableList, State).

butlast(L, BL) :-
    reverse(L, [_Last|Rev]),
    reverse(Rev, BL).

extract_op(Line, Op) :-
    split_complex_sep(Line, ",,", Vals),
    reverse(Vals, [QuotedOpCodesFRW|_]),
    butlast(QuotedOpCodesFRW, [_|OpCodesFRW]),
    atom_charptr(OpCodesFRW, OpFRW),
    fast_buf_read(Op, OpFRW),
    free_str(OpFRW).

:- meta_predicate take_while(1,*,*).
:- meta_predicate take_while1(*,1,*).
take_while(Pred, List, Res) :-
    take_while1(List, Pred, Res).
take_while1([], _, []).
take_while1([H|T], Pred, R) :-
    (call(Pred, H)
     -> R = [H|RT], take_while1(T, Pred, RT)
      ; R = []).


starts_with(Pre, Str) :-
  append(Pre, _, Str).

split_at_colon(S, L, R) :-
    split_chars(S, ":", [L, R]).

extract_variable_list(TypeInfoLine, Variables, _Types) :-
    split_complex_sep(TypeInfoLine, ",,", TypeInfo),
    take_while(starts_with("DA"), TypeInfo, TypedVariables),
    maplist(split_at_colon, TypedVariables, DAVariables, _TypesAsStrings),
    maplist(append("DA"), VariablesStrs, DAVariables),
    maplist(atom_codes, Variables, VariablesStrs).

csv_to_trace(FileName, StateList, ['$init_state'|OpList]) :-
    loadfr,
    read_trace(FileName, TypeInfo, [DummyState|Lines], [DummyState|ButLastLines]),
    extract_variable_list(TypeInfo, VariableList, _TypeList),
    extract_op(DummyState, '$init_state'),
    maplist(line_to_state(VariableList), Lines, StateList),
    maplist(extract_op, ButLastLines, OpList).
