/**
bugfix: between two graphs the number of nodes with the same colour were not compared
all previous versions may contain this bug!

bugfix: for the cell_sizes were only compared the first n entries, which can lead to
an error if there are more colours than nodes
all previous versions may contain this bug! 
*/

#include "interface.h" 
#include<stdio.h>
#include<stdlib.h>

int counter = 0; //for testing counts every call of function
int tree_node_counter;
int global_number_of_colours;

/* compares new graph (with n nodes) with an already stored graph
 * returns 0 if new graph equal to the stored graph in the tree, 
 * 1 if number of nodes or canonical label of new graph is bigger than the stored one
 * -1 if number of nodes or canonical label of new graph is smaller than the stored one
 * ERROR_VALUE if an error occured
*/

int compare_to_internal(graph *canon, int *cell_sizes, int n, struct Node *internalNode){

	//int comp = ERROR_VALUE; // error value = -20
	
	if(internalNode){

		/* compare first the number of nodes of the new graph and stored graph 
			only if they are equal then the canonical forms need to be compared
		*/
		if(n > internalNode->number_of_nodes){
			// comp = 1;
			return 1;
		}
		else if(n < internalNode->number_of_nodes){
			// comp = -1;
			return -1;
		}
		else{
			/* number of nodes are equal when progam comes here and so the sizes of the arrays
			   (cell_sizes is only bigger when static memory allocation is used)
			   global_number_of_colours is the number of colours that can occur in any graph,
				 so compare only so many colours.
			*/
			int i;
			for(i=0;i<global_number_of_colours;i++){
				if(DEBUG_VERSION){
					// printf("compare cell sizes for colour %d \n", i);
				}
				if(cell_sizes[i] > internalNode->cell_sizes[i]) { //first -> then [] 
					// comp = 1;
					return 1; 
				} 
				else if(cell_sizes[i] < internalNode->cell_sizes[i]) {
					// comp = -1;
					return -1;
				}
				//else continue for loop
			}
				
			 /* both graphs have same number of nodes: n and same cell sizes for each colour;
			 have to compare canonical labels (label is pointer) */
			return compare(canon, internalNode->label, n);
		}
	
	}
	else{
		printf("Error: internalNode is Nullpointer; can't compare");
		return ERROR_VALUE;
	}
	//should return berfore program gets here
	return ERROR_VALUE;
}



/* insert canonical form of graph in tree, if canonical form  is not in the tree yet
   returns number of tree node if canonical form is already in tree, 
  negative of number of tree node if it is newly inserted, 0 otherwise
*/
int insert_canon(graph *canon, int *cell_sizes, struct Node **treeNode, int n_local){

	if(DEBUG_VERSION){ 
		printf("insert_canon \n");
	}
	int m_local = (n_local + WORDSIZE - 1) / WORDSIZE;
	int compare_graph = ERROR_VALUE;  //initialized with error value

	 /* canonical form yet to be found or inserted in tree */
	int insert = 0;

	if(*treeNode == NULL){ //insert new node here (*treeNode os still pointer)
		
		//{n_o_n , &n_o_n_w_c, &label, NULL, NULL};
		*treeNode = (struct Node*) malloc(sizeof(struct Node)); 

		/* the label is like the graph itself an adjacency matrix stored in a 1-dim array*/
		graph *label = (graph*) malloc(sizeof(graph) * n_local * m_local );

		/* the local number of cells is maximum n_local but the number for the 
		colour can be bigger, so use MAXN as for global_cell_sizes*/
		int *sizes = (int*) malloc( sizeof(int) * MAXN);
		
		//free(label);
		//free(sizes);

		if(*treeNode && label && sizes){ //memory allocated
			if(DEBUG_VERSION){
				printf("now insert \n");
			}
			/* size of each cell is default 0 -> initialise allocated memory with zeros */
			init_cell_sizes(sizes, n_local);
			//free(label);
			//free(sizes);

			/* initialise label (which is a graph) with empty graph with n_local nodes */
			initGraph(label, n_local);
			/* store the canonical form as label*/
			copy_label(canon, label, n_local);	
			/* store number of nodes with same colour for each colour*/
			copy_cell_sizes(cell_sizes, sizes, global_number_of_colours);
				
			/* give any node (graph) in state space a number, if there are already more than 
			INT_MAX states then nodes(graphs) get number INT_MAX */
			if(tree_node_counter < INT_MAX){
				(*treeNode)->tree_node_number = tree_node_counter; 
				tree_node_counter++;
			}
			else{
				printf("warning: number of stored graphs exceeds INT_MAX \n");
				printf("all following nodes have number INT_MAX");
				(*treeNode)->tree_node_number = INT_MAX; 
			}
			(*treeNode)->number_of_nodes = n_local; 
			(*treeNode)->label = label; 
			(*treeNode)->cell_sizes = sizes;
			(*treeNode)->left=NULL;
			(*treeNode)->right=NULL;
			//canonical form is correctly inserted 
			// return the number of that node (negativ for not existing before)
			insert = (-1) * ((*treeNode)->tree_node_number); 
		}else{
			printf("in insert_canon: no memory \n");
		}
	}
	else{
		compare_graph = compare_to_internal(canon, cell_sizes, n_local, *treeNode);
		switch(compare_graph){
		// canonical form is already in tree (exists = TRUE)
		// return the number of that node (positiv)
		case 0: insert = (*treeNode)->tree_node_number; 
			break;
		case 1: // go down left branch
			insert = insert_canon(canon, cell_sizes, &((*treeNode)->left), n_local);  
			//use pointer because treeNode->left is changed
			break;
		case -1: //go down right branch
			insert = insert_canon(canon, cell_sizes, &((*treeNode)->right), n_local);
			break;
		default: printf("error during comparisson in function insert_canon");
			break;
		}
	}

	// return tree node number if canonical form was already in the tree and 
	// negative tree node number if it was inserted 
	// 0 indicates error
	return insert;
}

/* nedded in insert_canon to store the cell_sizes of the current graph 
* the parameter n must be the number of nodes of that graph otherwise the cell sizes 
* may not stored correctly
*/
void copy_cell_sizes(int *cell_sizes, int *sizes, int n){
	int k = 0;

	while(k < n){ // for max. every node one cell 
		sizes[k] = cell_sizes[k];
		k++;
	}
}

/* nedded in insert_canon to store the canonical form as label */
void copy_label(graph *canon, graph *label, int n){

	int m = (n + WORDSIZE - 1) / WORDSIZE;
	int k = 0;

	while(k < n*m){ // for every node n g has m sets that need to be considered as one row

		int l=0;
		for(l=0; l<m; l++){ //consider one row
			label[k+l] = canon[k+l]; //consider one setword
		}	
		k=k+m;
	}

}


/*free memory that is stored in the tree */
void free_tree(struct Node **treeNode){
	
	if(*treeNode != NULL){

		struct Node *pleft;
		struct Node *pright;

		pleft = (*treeNode)->left; //pointers to subtrees
		pright = (*treeNode)->right;

		free_tree(&pleft); // call function for subtrees
		free_tree(&pright);

		// can free headnode when subtrees are freed
		//if (pleft == NULL && pright == NULL)
		// make sure that subtrees are empty (not necessary)
		if(!pleft && !pright) 
		{
		
			free((*treeNode)->cell_sizes);
			(*treeNode)->cell_sizes = NULL;

			free((*treeNode)->label);
			(*treeNode)->label = NULL;

			free(*treeNode);	
			*treeNode = NULL;	
			counter++;
		}
	}
}
