/*
* Exposes Glucose API to SICStus Prolog
*
* Based on SWI Prolog Minisat Integration by
* Michael Codish, Vitaly Lagoon, Peter J. Stuckey
*/

#include <sicstus/sicstus.h>
#include <cstdlib>
#include <vector>
#include "core/Solver.h"

using namespace Glucose;
using namespace std;

#define val(id,i) ((solvers[id]->model[i] != l_Undef) ? ((solvers[id]->model[i]==l_True)? i+1:-1*(i+1)):0)

vector<Solver*> solvers;

extern "C" {
    void delete_solver(SP_integer id) {
        if (solvers[id]) {
            delete solvers[id];
            solvers[id] = NULL;
        }
    }

    SP_integer new_solver() {
        for (int i=0; i<solvers.size(); i++) {
            if(!solvers[i]) {
                solvers[i] = new Solver;
                return i;
            }
        }

        solvers.push_back(new Solver);
        return solvers.size()-1;
    }

    static inline Lit pl2lit(SP_integer solver_id, SP_term_ref pl_literal) {
        SP_integer pl_lit_int;
        long var;
        bool negated = false;
        
        // literal / var is wrapped in neg(...) is case it stems from the cnf conversion
        if(SP_is_compound(pl_literal)) {
            SP_get_arg(1,pl_literal,pl_literal);
            negated = true;
        }
        
        SP_get_integer(pl_literal,&pl_lit_int);
        // liteal / var is a negative integer if its a negated model value
        if(pl_lit_int < 0) {
            negated = true;
            pl_lit_int = -pl_lit_int;
        }
        
        // offset due to fix / given variables for true and false
        var = pl_lit_int - 1;

        while (var >= solvers[solver_id]->nVars()) {
            solvers[solver_id]->newVar();
        }

        return mkLit(var,negated);
    }

    void add_clause(SP_integer solver_id, SP_term_ref list) {
        SP_term_ref head = SP_new_term_ref();
        vec<Lit> lits;

        while(SP_get_list(list, head, list) ) {
            lits.push(pl2lit(solver_id, head));
        }

        if (!(solvers[solver_id]->addClause(lits))) {
            SP_fail();
        }
    }

    void solve(SP_integer solver_id) {
        if (!(solvers[solver_id]->solve())) {
            SP_fail();
        }
    }

    void get_model(SP_integer solver_id, SP_term_ref model) {
        SP_term_ref l = SP_new_term_ref();
        SP_term_ref pl_lit = SP_new_term_ref();
        int i=solvers[solver_id]->nVars();

        while(--i >= 0) {
            SP_put_integer(pl_lit,val(solver_id, (int)i));
            SP_cons_list(l, pl_lit, l);
        }

        SP_unify(model,l);
    }

    void assign_model(SP_integer solver_id, SP_term_ref asgnList) {
        SP_term_ref pred_true = SP_new_term_ref();
        SP_put_atom(pred_true,SP_atom_from_string("pred_true"));

        SP_term_ref pred_false = SP_new_term_ref();
        SP_put_atom(pred_false,SP_atom_from_string("pred_false"));

        SP_term_ref asgnVar = SP_new_term_ref();

        int indx=1;
        while(SP_get_list(asgnList, asgnVar, asgnList)) {
            if(solvers[solver_id]->model[indx]==l_True) {
                SP_unify(asgnVar,pred_true);
            }
            else {
                SP_unify(asgnVar,pred_false);
            }
            indx++;
        }
    }
}
