#ifndef QUEUE_H_INCLUDED
#define QUEUE_H_INCLUDED

#include "hash.h"
#include "../libs/leveldb-1.20/include/leveldb/c.h"

typedef struct Cell {
    char *term;
    char digest[HASH_LEN];
    size_t term_length;
    struct Cell* next; 
} tCell;

struct db {
    leveldb_t *db;
    leveldb_options_t *opts;
    leveldb_readoptions_t *read_opts;
    leveldb_writeoptions_t *write_opts;
    char *name;
};

typedef struct workQueue {
    tCell *first;
    tCell *last;
    size_t c;
    size_t terms_available;
    size_t thresh;
    struct db db;
    int strategy;
} wQueue;


/* initialises a new queue */
wQueue *init_queue(const int strategy, const size_t thresh, char const * const tmpdir);

/* enqueue a term with length len and hash code digest into the given queue;
 * function takes ownership of the term */
void enqueue(wQueue *q, char * const term, const size_t len, const char digest[HASH_LEN]);

/* dequeues a cell from given queue;
 * returns NULL, if queue was empty.
 * note that the next pointer will be NULL. */
tCell *dequeue(wQueue * const q);

/* returns 1 if the queue is empty,
 * 0 otherwise */
int is_empty(wQueue const * const q);

/* returns the current queue size of the given queue */
size_t q_size(wQueue const * const q);

/* tears down a queue and all its queued items */
void teardown_queue(wQueue **q);

/* tears down a cell and its term */
void teardown_cell(tCell **c);

#endif
