#ifndef __GNU_SOURCE
#define __GNU_SOURCE
#endif

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <semaphore.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include "shm_cln.h"
#include "shm_common.h"
#include "hashtrie.h"

#include "assert_helper.h"


struct semap_env open_semaphores(const int id) {
    struct semap_env r = {0};
    init_semaphore_names(&r, id);

    r.semap_meta   = sem_open(r.semap_meta_s, 0);
    r.semap_hashes = sem_open(r.semap_hashes_s, 0);
    r.semap_hashwrite = sem_open(r.semap_hashwrite_s, 0);
    assert_es(r.semap_meta      != SEM_FAILED, "Opening a semaphore failed. This might be caused by a synchronisation issue. Please report this error.");
    assert_es(r.semap_hashes    != SEM_FAILED, "Opening a semaphore failed. This might be caused by a synchronisation issue. Please report this error.");
    assert_es(r.semap_hashwrite != SEM_FAILED, "Opening a semaphore failed. This might be caused by a synchronisation issue. Please report this error.");

    int i;
    for (i = 0; i < 32; i++) {
        r.semap_nodes[i] = sem_open(r.semap_nodes_s[i], 0);
        assert_es(r.semap_nodes[i]  != SEM_FAILED, "Opening a semaphore failed. This might be caused by a synchronisation issue. Please report this error.");

        r.semap_multiwrite[i] = sem_open(r.semap_multiwrite_s[i], 0);
        assert_es(r.semap_multiwrite[i]  != SEM_FAILED, "Opening a semaphore failed. This might be caused by a synchronisation issue. Please report this error.");
    }
    return r;
}

void teardown_shm(struct shm_env *env) {
    int rc = sem_close(env->semaphores.semap_meta);
    assert(rc == 0);
    rc = sem_close(env->semaphores.semap_hashes);
    assert(rc == 0);
    rc = sem_close(env->semaphores.semap_hashwrite);
    assert(rc == 0);
    int i;
    for (i = 0; i < 32; i++) {
        rc = sem_close(env->semaphores.semap_nodes[i]);
        assert(rc == 0);
        rc = sem_close(env->semaphores.semap_multiwrite[i]);
        assert(rc == 0);
    }
    free_semaphore_names(&(env->semaphores));
    env->semaphores = (struct semap_env) {0};
    shm_cleanup(env, 0);
    #ifdef BLOCKS
    printf("%u readblocks\n", blocks_read);
    #endif
}


struct shm_env *init_shm(const int id) {
    struct semap_env senv = open_semaphores(id);
    return init_shm_common(senv,
                           10000000,
                           10000000,
                           10000 + id * 20000,
                           10001 + id * 20000,
                           20000 + id * 20000);
}


