#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <zmq.h>
#include <czmq.h>

#include <stdlib.h>
#include <assert.h>

#include "../common/hashtrie.h"
#include "mid.h"

/* creates a TCP id reply zocket on port number;
 * sets the model to given model with size len;
 * returns the new zocket */
zsock_t *master_id_setup(const int portnumber) {
    char *endpoint;
    int rc = asprintf(&endpoint, "tcp://*:%d", portnumber);
    assert(rc != -1);

    zsock_t *zocket = zsock_new_rep(endpoint);
    free(endpoint);

    return zocket;
}

/* tears down an id map */
void teardown_id_map(struct id_map *ips) {
    if (ips->len) {
        int i;
        for (i = 1; i <= ips->len; i++) {
            free(ips->id_map[i]);
        }
        free(ips->id_map);
    }
    *ips = (struct id_map) {0};
}

/* receives exactly one id request and replies with an id and a model
 * returns the cmd it received */
int master_h_id(zsock_t * const zocket,
                char * const master_ip,
                const int port_pub,
                struct id_map * const ips,
                zchunk_t const * const model,
                const int send_hashes,
                hashtrie trie) {
    int cmd = 0;
    char *ip = NULL;
    int proxy_port_pub = 0;
    int rc = zsock_recv(zocket, "isi", &cmd, &ip, &proxy_port_pub);
    assert(rc == 0);

    ips->len++;
    ips->id_map = realloc(ips->id_map, sizeof(char**) * (ips->len + 1));
    ips->id_map[ips->len] = ip;
    ips->ports = realloc(ips->ports, sizeof(int*) * (ips->len + 1));
    ips->ports[ips->len] = proxy_port_pub;

    char *parent_endpoint;
    int parent_id = (ips->len-1)/2;
//    parent_id = 0; // uncomment so that everyone subscribes to master
    if (parent_id) {
        rc = asprintf(&parent_endpoint, "tcp://%s:%d", ips->id_map[parent_id], ips->ports[parent_id]);
        assert(rc != 1);
    } else {
        rc = asprintf(&parent_endpoint, "tcp://%s:%d", master_ip, port_pub);
        assert(rc != 1);
    }



    if (!send_hashes) {
        rc = zsock_send(zocket, "icbs", ips->len, model, NULL, 0, parent_endpoint);
    } else {
        size_t blocksize;
        char *block = trie_iterate(trie, &blocksize);
        rc = zsock_send(zocket, "icbs", ips->len, model, block, blocksize, parent_endpoint);
        free(block);
    }

    assert(rc == 0);
    free(parent_endpoint);
    return cmd;
}


/* reader function for master_h_id */
int mid_reader(zloop_t *loop, zsock_t *reader, void *args) {
    struct mid_args *arg = (struct mid_args*) (args);
    (void) master_h_id(reader, arg->master_ip, arg->port_pub, arg->ips, arg->model, arg->send_hashes, arg->trie);
    return 0;
}
