#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "../master/mstats.h"
#include "../common/msg_headers.h"

#define NR_PROXIES 36

zmsg_t *build_need_msg(int index, int amount) {
    zmsg_t *msg = zmsg_new();
    int i;
    for (i = 0; i < amount; i++) {
        zmsg_addstrf(msg, "proxy %d nr %d", index, i);
    }
    return msg;
}

pDigest get_pdigest(int index,
                    int header,
                    int amount,
                    int frozen) {
    zmsg_t *msg = NULL;
    if (header == PSTATS_HAS_EMPTY ||
            header == PSTATS_COMPLETELY_EMPTY) {
        msg = build_need_msg(index, amount);
    }
    pDigest res;
    res.header = header;
    res.full = amount;
    res.need_msg = msg;
    res.frozen = frozen;
    return res;
}

mStats setup_mstats(void) {
    pDigest *arr = calloc(sizeof(pDigest), NR_PROXIES + 1);
    mStats stats;
    stats.pdigests = arr;
    stats.len = NR_PROXIES;

    int i;
    for (i = 1; i <= NR_PROXIES; i++) {
        if (i == 5) {
            arr[i] = (pDigest) {0};
        } else if (i == 3) {
            arr[i] = get_pdigest(i, PSTATS_HAS_PMS, 3, 0);
        } else {
            arr[i] = get_pdigest(i, PSTATS_COMPLETELY_EMPTY, 14, 0);
        }
    }
    return stats;
}

zmsg_t *get_stats_msg(int id, int header, int amount) {
    zmsg_t *msg = zmsg_new();
    zmsg_addstrf(msg, "%d", id);
    switch (header) {
        case PSTATS_HAS_PMS:
            break;
        case PSTATS_HAS_EMPTY:
        case PSTATS_COMPLETELY_EMPTY:
            zmsg_addstrf(msg, "%d", amount);
            int i;
            for (i = 0; i < amount; i++) {
                zmsg_addstrf(msg, "proxy %d nr %d", id, i);
            }
            break;
        case PSTATS_HAS_FULL:
            zmsg_addstrf(msg, "%d", amount);
            break;
    }
    return msg;
}

zmsg_t *get_transfer_done(int index, int header, int amount) {
    zmsg_t *msg = get_stats_msg(index, header, amount);
    zmsg_pushstrf(msg, "%d", P_TRANSFER_DONE);
    return msg;
}

char const * const *setup_idmap(int amount) {
    char **res = calloc(sizeof(char*), amount + 1);
    int i;
    for (i = 1; i <= amount; i++) {
        asprintf(&res[i], "%d", i);
    }
    return (char const * const *) res;
}

int main(int argc, char *argv[]) {
    mStats stats = setup_mstats();
    char const * const *idmap = setup_idmap(NR_PROXIES);

    zsock_t *pub = zsock_new_pub("inproc://foobar");
    zsock_t *sub = zsock_new_sub("inproc://foobar", "");

    zmsg_t *statmsg = get_stats_msg(3, PSTATS_HAS_FULL, 1);
    zmsg_print(statmsg);
    update_queue_digests(&stats, PSTATS_HAS_FULL, statmsg, STDOUT_FILENO);

    statmsg = get_stats_msg(3, PSTATS_HAS_FULL, 2);
    zmsg_print(statmsg);
    update_queue_digests(&stats, PSTATS_HAS_FULL, statmsg, STDOUT_FILENO);

    m_do_balance(&stats, pub, idmap, STDOUT_FILENO);

    zmsg_t *master_msg;
    zsock_recv(sub, "m", &master_msg);

    zmsg_print(master_msg);
    zmsg_destroy(&master_msg);
    

    assert(stats.pdigests[1].header == PSTATS_HAS_PMS);
    assert(stats.pdigests[1].frozen == 2);
    assert(stats.pdigests[3].header == PSTATS_HAS_PMS);
    assert(stats.pdigests[3].frozen == 2);

    statmsg = get_stats_msg(3, PSTATS_HAS_FULL, 3);
    zmsg_print(statmsg);
    update_queue_digests(&stats, PSTATS_HAS_FULL, statmsg, STDOUT_FILENO);

    assert(stats.pdigests[3].header == PSTATS_HAS_PMS);
    assert(stats.pdigests[3].frozen == 2);

    print_mstats(&stats, STDOUT_FILENO);

    
    zsock_destroy(&pub);
    zsock_destroy(&sub);
    teardown_mstats(&stats);
    return 0;
}
