#include <zmq.h>
#include <czmq.h>

#include <pthread.h>

#include "../master/mid.h"
#include "../proxy/pid.h"
#include "../worker/wid.h"

#include "../common/shm_common.h"
#include "../common/shm_srv.h"

const int PROXIES_PER_MASTER = 2;
const int WORKERS_PER_PROXY = 2;
char const * const MODEL = "DA\0le_model\0";
const int MODEL_LEN = 12;

hashtrie trie;

void *master(void *args) {
    zchunk_t *model_chunk = zchunk_new(MODEL, MODEL_LEN);
    zsock_t *zocket = master_id_setup(5000);

    struct id_map proxy_ips = {0};
    int i;
    for (i = 0; i < PROXIES_PER_MASTER; i++) {
        master_h_id(zocket, &proxy_ips, model_chunk, 0, trie);
    }

    for (i = 0; i < PROXIES_PER_MASTER; i++) {
        assert(0 == strcmp(proxy_ips.id_map[i+1], "10.0.0.100"));
    }

    teardown_id_map(&proxy_ips);
    zsock_destroy(&zocket);
    zchunk_destroy(&model_chunk);
    return NULL;
}

void *proxy(void *args) {
    int index = *((int*) args);
    struct pid_response res = proxy_request_id("localhost", 5000, "10.0.0.100", 0);
    assert(res.id == index+1);
    assert(memcmp(zchunk_data(res.model), MODEL, MODEL_LEN) == 0);
    free(res.digest_block);

    zsock_t *zocket = proxy_setup_id_rep_zocket(0);

    int i;
    for (i = 0; i < WORKERS_PER_PROXY; i++) {
        proxy_h_id(zocket, res.model);
    }
    proxy_teardown_id();
    zsock_destroy(&zocket);
    return NULL;
}


int main(int argc, char *argv[]) {
    struct semap_env senv = create_semaphores(0);
    trie = get_trie(senv, 10000);

    pthread_t master_thread;
    pthread_create(&master_thread, NULL, master, NULL);

    int i, j;
    // start master thread
    for (i = 0; i < PROXIES_PER_MASTER; i++) {
        pthread_t proxy_thread;
        pthread_create(&proxy_thread, NULL, proxy, &i);
        // start proxy thread
        for (j = 0; j < WORKERS_PER_PROXY; j++) {
            struct id_response res = worker_request_id("ipc:///tmp/zmq-proxy-0-id");
            assert(memcmp(zchunk_data(res.model), MODEL, MODEL_LEN) == 0);
            unsigned int expected_id = j + 1;
            assert(res.id == expected_id); 
            free(res.model);
        }
        // wait for proxy thread
        pthread_join(proxy_thread, NULL);
    }
    pthread_join(master_thread, NULL);
    printf(".");
    teardown_hashtrie(&trie, 1, 0);
    puts("");
    return 0;
}
