#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/msg_headers.h"
#include "../proxy/p_pub_sub.h"
#include "../proxy/pstats.h"

static void test_proxy_balance(zsock_t *pub, zsock_t *sub) {
    struct worker_queues wqs;
    size_t arr[] =  {0, 0, 1234, 1500, 0, 34, 0};
    wqs.queuesizes = arr; 
    wqs.len = 6;
    wqs.empty_workers = 3;
    wqs.balanced_workers = 1;
    wqs.full_workers = 2;

    int rc = local_balance(pub, &wqs, 100, 5000, STDOUT_FILENO, 0);
    assert(rc == 2);

    int header;
    char *id_has;
    char *endpoint;
    // first local balancing msg
    rc = zsock_recv(sub, "iss", &header, &id_has, &endpoint);
    assert(rc == 0);

    assert(header == L_TRANSFER_TO);
    assert(0 == strcmp("3", id_has));
    assert(0 == strcmp("tcp://localhost:5011", endpoint));
    free(id_has);
    free(endpoint);


    // second local_balancing msg
    rc = zsock_recv(sub, "iss", &header, &id_has, &endpoint);
    assert(rc == 0);

    assert(header == L_TRANSFER_TO);
    assert(0 == strcmp("2", id_has));
    assert(0 == strcmp("tcp://localhost:5014", endpoint));
    free(id_has);
    free(endpoint);

    assert(arr[1] & ((size_t) 1 << 63));
    assert(arr[2] & ((size_t) 1 << 63));
    assert(arr[3] & ((size_t) 1 << 63));
    assert(arr[4] & ((size_t) 1 << 63));
    assert(!(arr[5] & ((size_t) 1 << 63)));
    printf(".");
}


int main(int argc, char *argv[]) {
    zsock_t *pub = zsock_new_pub("inproc://foo-pubsub");
    zsock_t *sub = zsock_new_sub("inproc://foo-pubsub", "");

    test_proxy_balance(pub, sub);

    zsock_destroy(&pub);
    zsock_destroy(&sub);
    puts("");
    return 0;
}
