#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/queue.h"
#include "../common/magic_constants.h"
#include "../common/msg_headers.h"
#include "../common/helpers.h"

#include "wpush.h"
#include "wsink.h"

zsock_t *init_work_sink(const int port_start, const int worker_id, const int proxynumber) {
    char *endpoint_rep;
    int rc = asprintf(&endpoint_rep,"tcp://*:%d", wsink_port(port_start, proxynumber, worker_id));
    assert(rc != -1);

    //puts(endpoint_rep);
    zsock_t *res = zsock_new_rep(endpoint_rep);
    free(endpoint_rep);
    return res;
}


void handle_wsink(zsock_t * const rep,
                  wQueue * const queue,
                  zsock_t * const push,
                  const int id,
                  const int fd_logfile) {
    int header;
    uint64_t timestamp;
    zmsg_t *msg;

    int rc = zsock_recv(rep, "i8m", &header, &timestamp, &msg);
    assert(rc == 0);
    assert(zmsg_size(msg) % 2 == 0);

    rc = zsock_send(rep, "i8", TRANSFER_ACK, timestamp);
    assert(rc == 0);

    int msg_size = zmsg_size(msg) / 2;
    //puts("I got it!");
    //char *headers = header == LOCAL_TRANSFER ? "LOCAL" : "GLOBAL";
    //printf("%d items (header: %s) @time %lu\n", msg_size, headers, timestamp);

    //printf("before: qsize = %lu\n", q_size(queue));
    int i;
    for (i = 0; i < msg_size; i++) {
        char *digest = zmsg_popstr(msg);
        char buffi[READABLE_HASH_LEN];
        print_readable_digest((unsigned char*) digest, buffi);
        dprintf(fd_logfile, "{:type :recvd-state, :hash :%s}\n", buffi);

        zframe_t *frame = zmsg_pop(msg);
        char *term = zframe_strdup(frame);
        size_t term_length = zframe_size(frame);
        zframe_destroy(&frame);

        enqueue(queue, term, term_length, digest);
        free(digest);
    }
    zmsg_destroy(&msg);

    //printf("after: qsize = %lu\n", q_size(queue));

    // FIXME: this is (maybe?) duplicated code
    int ack_header;
    switch (header) {
        case LOCAL_TRANSFER:
            ack_header = L_TRANSFER_DONE;
            break;
        case CHECK_THIS:
        case GLOBAL_TRANSFER:
            ack_header = G_TRANSFER_DONE;
            break;
        default:
            assert(0);
    }
    signal_transfer_done(push, ack_header, id, queue, timestamp);
}

int wsink_reader(zloop_t *loop, zsock_t *reader, void *args) {
    puts("receive work?");
    struct wsink_args *arg = (struct wsink_args*) args;
    handle_wsink(reader, arg->q, arg->push, arg->worker_id, arg->fd_logfile);
    return 0;
}
