% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bvisual_any_maxsolver, [any_maxsolver/5]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,dot).
:- module_info(description,'Tool for generating dot visualization of B predicates and expressions. Finds a maximally satisfiable subset for predicates.').

%:- use_module(library(ordsets)).
%:- use_module(library(timeout)).

:- use_module(bsyntaxtree, [conjunction_to_list/2, conjunct_predicates/2]).
:- use_module(bvisual, [get_any/5]).
:- use_module(maxsolver).

:- use_module(error_manager, [add_error/2, time_out_with_enum_warning_one_solution/3, is_time_out_result/1]).

any_maxsolver(IDs, Condition, State, LocalState, NewLocalState):-
    conjunction_to_list(Condition, Set),!,
    length(Set, Count),
    %Check = bvisual_any_maxsolver:x_check(IDs, LocalState, State), % inlined to enable InfoLog flow analysis
    x_time_out_by_size(Count, TimeOut),
    %print(time_out(maxsolver(Check),TimeOut,Res)),nl,
    time_out_with_enum_warning_one_solution(maxsolver(bvisual_any_maxsolver:x_check(IDs, LocalState, State), Set, Solver0), TimeOut, Res),
    !,
    ( 
        is_time_out_result(Res) -> 
            add_error(bvisual, 'Time out at trying to find a solver of maximal length.'),
            longest_satisfiable_prefix(bvisual_any_maxsolver:x_check(IDs, LocalState, State), Set, Solver, _)
            ;
        Solver0 = time_out(Solver) ->
            add_error(bvisual, 'Time out at trying to find a solver of maximal length.')
            ;
        Solver = Solver0
    ),
    conjunct_predicates(Solver, SolverExpr),
    get_any(IDs, SolverExpr, State, LocalState, NewLocalState).


:- use_module(preferences).

x_time_out_by_size(0, 10) :- !. % starting from sicstus 4.3, msb(0) throws an error
x_time_out_by_size(Size, TimeOut):-
    preferences:get_preference(time_out, BaseTimeOut),
    TimeOut is max(10,min(msb(Size), 3) * BaseTimeOut). /* Timeout must be at least 1 */


%
% Helpers
%
:- public x_check/4.
x_check(_, _, _, []) :- !.
x_check(IDs, LocalState, State, Subset):-
    conjunct_predicates(Subset, LHS),!,
    get_any(IDs, LHS, State, LocalState, _).