#!/usr/bin/env python

import subprocess

target_file = 'perf_overview.tex'

cases = [
    ("Who Killed Agatha?",                "WhoKilledAgatha.mch",              "perf_agatha.txt"),
    ("Crew Allocation",                   "CrewAllocationConstantsLarge.mch", "perf_crew_alloc.txt"),
    ("20--Queens",                        "NQueens20.mch",                    "perf_nqueens20.txt"),
    ("Graph Colouring (integer sets)",    "GraphColouringNAT.mch",            "perf_graph_col_nat.txt"),
    ("Graph Colouring (enumerated sets)", "GraphColouringDS.mch",             "perf_graph_col_es.txt"),
    ("Graph Isomorphism",                 "CheckGraphIsomorphismConst2.mch",  "perf_graphiso2.txt"),
    ("Loop detection in control flow",    "Loop.mch",                         "perf_loop.txt"),
    ("SAT instance",                      "flat200-90.mch",                   "perf_flat200-90.txt"),
    ("Send More Money",                   "SendMoreMoney.mch",                "perf_sendmoremoney.txt"),
    ("Aristotle's sieve (1 step)",        "SieveStep1.mch",                   "perf_sieve.txt"),
    ("Simple Computation",                "SimpleComputation2.mch",           "perf_simplecomp.txt")
    ]

prob_directory = "/home/plagge/prob2"
prob_binary = prob_directory + "/probcli"
example_directory = prob_directory + "/examples/B/Kodkod"

iterations = 3
timeout = 60000
max_solutions = 50

extract = { "Kodkod total time": 'kodkod_total',
            "Kodkod Java time": 'kodkod_java',
            "Kodkod preparation time": 'kodkod_prep',
            "Kodkod computation time": 'kodkod_comp',
            "ProB total time": 'prob_total'
            }

def do_tests():
    sumfile = open(target_file, 'w')
    for case in cases:
        (name, filename, logfile) = case
        filename = example_directory + '/' + filename
        do_test(name, filename, logfile, sumfile)
    sumfile.close()

def do_test(name, filename, logfile, sumfile) :
    print "test case: " + name
    resultcode = subprocess.call([prob_binary,
                                  '-p', 'CLPFD', 'true',
                                  '-p', 'MAX_INITIALISATIONS', str(max_solutions),
                                  '-p', 'TIME_OUT', str(timeout),
                                  '-kodkod_performance', logfile, str(iterations),
                                  filename])
    if resultcode != 0:
        raise Exception("calling ProB failed")
    times = extract_times(logfile)
    sumfile.write("  %s & %s & %s & %s \\\\\n" % (name, times['prob_total'], times['kodkod_total'], times['kodkod_java']) )

def extract_times(logfile):
    infile = open(logfile)
    lines = infile.readlines()
    infile.close()
    times = {}
    for l in lines:
        splitted = l.split(":")
        if len(splitted)==2:
            key = splitted[0].strip()
            if extract.has_key(key):
                times[extract[key]] = splitted[1].strip()
    return times

if __name__ == '__main__':
    do_tests()
