% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(reduce_graph_state_space,	[print_state_merge_for_dot/1,print_dot_for_dfa_from_nfa/1,
					print_min_dfa_for_dot/1,
					reduce_graph_reset/0,
					%assert_desired_user_transitions/1,
					%state_merge_separate_iviolations/1,
					print_subgraph_associated_with_invariant_violations/1,
					%assert_goal_nodes/0,
					%jump_to_next_goal_node/1,
					print_subgraph_of_goal_nodes/1,
					set_signature_merge_show_all_transitions/1,
					signature_merge_show_all_transitions/1,
					%time_stamp/1, time_between/3, mc_duration/1,
					%total_visited_states/1, total_transitions/1,
					%reset_use_all_ops_args/0,

					extract_graph_of_state_merge/1,


                                        % tcltk gui
                                        user_made_op_selection/0, user_made_arg_selection/0, set_use_all_operations/2

					/* symmetry reduction */
				%%	cleanup_symmetry/0 /* remove all stored representatives */,
				%%	representative/3 /* get the representative state for this state */,
				%%%	reset_relation_types/0 /* resetting this module for new machine */
				]).

/* reduce_graph_state_space.pl - GRAPH REDUCTION FUNCTIONS */

/* List of functions that work and (often) reduce the current state space
 - print_dot_for_dfa_from_nfa(File).
 - print_min_dfa_for_dot(File).
 - print_state_merge_for_dot(File).

 NB: print_min_dfa_for_dot will perform the nfa -> dfa conversion first, and then
     try and reduce it. However, this doesn't seem to further reduce the dfa ==>
     dfa is minimal already
*/
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,dot).
:- module_info(description,'Various minimization/reduction algorithms for the state space.').

:- use_module(debug).

:- use_module(library(lists)).
%:- use_module(library(system)).
:- use_module(preferences).
%:- use_module(graph_canon).
%:- use_module(library(terms)).

:- use_module(visualize_graph,[print_graph_header/1,
                    print_graph_footer/0]).
:- use_module(state_space,[visited_expression/2, visited_expression_id/1,
                          transition/3, store_transition/4,
                          invariant_violated/1, not_all_transitions_added/1]).
:- use_module(specfile,[csp_mode/0]).
:- use_module(translate,[translate_bvalue/2]).

reduce_graph_reset :- assert_desired_user_transitions(alphabet),
        reset_use_all_ops_args, reset_counter.

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reduce_graph_reset,
                           'Reset reduce_graph_state_space.').

/*==============================================================================
		Utility Functions
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */


% different types of graphs
graph_type(signature_merge).
graph_type(dfa_abstraction).


% flatten list
ent03r_flatten([],[]) :- !.
ent03r_flatten([X|L],R) :- !, ent03r_flatten(X,XR), ent03r_flatten(L,LR), append(XR,LR,R).
ent03r_flatten(X,[X]).


ent03r_union([],X,X).
ent03r_union([X|R],Y,Z) :- member(X,Y),!,ent03r_union(R,Y,Z).
ent03r_union([X|R],Y,[X|Z]) :- ent03r_union(R,Y,Z).

:- volatile curr_equivalence_class_id/1.
:- dynamic curr_equivalence_class_id/1.

% initialise id counter
init_equivalence_class_id :-
	(curr_equivalence_class_id(X) ->
		retract(curr_equivalence_class_id(X))
		;true),
	assert(curr_equivalence_class_id(0)).

% get a unique id
next_equivalence_class_id(ID) :-
	(curr_equivalence_class_id(X) ->
		ID is X+1,
		retract(curr_equivalence_class_id(X)),
		assert(curr_equivalence_class_id(ID))
		; ID=0, assert(curr_equivalence_class_id(ID))).

number_of_equivalence_classes(X) :- curr_equivalence_class_id(X).
% a counter; to do: use extension(counter)
increment_counter(C) :-
	(counter(D) -> C is D + 1,retract(counter(_)),assert(counter(C));C is 0,
	assert(counter(C))).
reset_counter :- (counter(_) -> retract(counter(_));true),assert(counter(0)).




transition(State,TranEncoded,IndividualSig,Succ) :-
	transition(State,TranEncoded,Succ),
	transition_info(TranEncoded,_,_,IndividualSig).

transition_info('-->'(FAction,Args), Name, Arity, Signature) :-
	transition_name_arity('-->'(FAction,Args),Name,Arity), %Arity is -(Arity+1),
	compute_signature(Name,Arity,Signature),!.
transition_info(io(V,Channel,_Span),Name,Arity,Signature) :-
     specfile:csp_mode,
     atomic(Channel), length(V,Arity), Name = Channel,
	 compute_signature(Name,Arity,Signature),!.
transition_info(Action, Name, Arity, Signature) :- nonvar(Action), functor(Action,Name,Arity),
	compute_signature(Name,Arity,Signature),!.
transition_info(_, Name, Arity, Signature) :- Name = 'err:', Arity = ('?'), Signature = 'err'.

:- volatile transition_info/3.
:- dynamic transition_info/3.

retractall_transition_info :-
	retractall(transition_info(_,_,_)).

% leuschel Jan 2013: I don't understand this piece of code below:
compute_signature(Name, Arity, Signature) :-
	(transition_info(Name,Arity,Signature) ->
		true
		; name('/', FS_cl),
		/* Arity for operations that return a value is one less than it
		actually is: so that later predicates can test for a return
		value by the arity being a negative number; and being one less
		enables coping with returning operations with 0 arity. */
		((Arity < 0) ->
			RealArity is (-Arity-1), name('-->', Arrow_cl),
			name(1, Return_Len),name('(', LP_cl), name(')', RP_cl)
			; RealArity is Arity, Arrow_cl = [],
			Return_Len = [], LP_cl = [], RP_cl = []),
		name(Name, Nn), name(RealArity, Aa),
		NnAa = [Nn, FS_cl, Aa],
		ent03r_flatten(NnAa, F_cl),
		RES = [F_cl, Arrow_cl, LP_cl, Return_Len, RP_cl],
		ent03r_flatten(RES, Flat_RES), name(Signature, Flat_RES),
			assert(transition_info(Name,Arity,Signature))),
		!.

% find the name and arity of a transition
transition_name_arity('-->'(FAction,_), Name, Arity) :- nonvar(FAction),
	functor(FAction,Name,Arity),!.
transition_name_arity(Action, Name, Arity) :- nonvar(Action), functor(Action,Name,Arity),!.
transition_name_arity(_, Name, Arity) :- Name = 'err:', Arity = 'err'.

% get the transition's arguments from it
get_transition_arguments('-->'(FAction,_),Args) :-
	nonvar(FAction),FAction =.. [_|Args],!.
get_transition_arguments(Action, Args) :- nonvar(Action), Action =.. [_|Args],!.
get_transition_arguments(_, Args) :- Args = 'err'.


% convert the transition's arguments to a list of Strings
get_args_as_strings([],[]).
get_args_as_strings([First|SecondPlus],[String|Tail]) :-
	translate:translate_bvalue(First,String),
	get_args_as_strings(SecondPlus,Tail).

% convert a list of atoms to a flat list of codes representing them separated
% by a Delimitor string.
convert_string_to_code_list(ListA, CodeListA, Delimitor) :-
	name(Delimitor,DelimitorCodeList),
	convert_string_to_code_list_helper(ListA, CodeListA,DelimitorCodeList).
convert_string_to_code_list_helper([],[],_).
convert_string_to_code_list_helper([H|T],[Hhh|Tt],DelimitorCodeList_) :-
	((T == []) ->
		DelimitorCodeList = [];DelimitorCodeList = DelimitorCodeList_),
	name(H,Hh),append(Hh,DelimitorCodeList,Hhh),convert_string_to_code_list_helper(T,Tt,DelimitorCodeList).

% get the alphabet for the state space, for a particular graph type, taking into
% account the "don't cares" for transitions.
get_alphabet(GT, Alphabet) :-
	findall(Letter,(transition(_,L,_), %transition_for_user(L,LU),
	  remove_transition_arg_dont_cares(GT,L,Letter)),Alphabet_),
	  remove_dups(Alphabet_,Alphabet).


% get the signatures of alphabet for the state space.
get_alphabet_signatures(Alphabet) :-
	findall(Letter,(transition(_,L,_),transition_info(L,_,_,Letter)),Alphabet_),remove_dups(Alphabet_,Alphabet).


:- volatile transition_arg_dont_cares/3.
:- dynamic transition_arg_dont_cares/3.

retractall_transition_arg_dont_cares :-
	retractall(transition_arg_dont_cares(_,_,_)).


% make a list of size, Length containing ascending integers, except at indices
% specified by DontCares, where the element is a '_'. Result is List.
make_counting_list(Length,DontCares,List) :-
	make_counting_list(0,Length,DontCares,List).
make_counting_list(End,End,_,[]) :- !.
make_counting_list(Counter,Length,DontCares,[Head|Tail]) :-
	(member(Counter,DontCares) ->
		Head = '_'
		; Head is Counter),
	CounterAddOne is (Counter + 1),
	make_counting_list(CounterAddOne,Length,DontCares,Tail).

% Assert "don't cares" format for a transition: '_' => don't care, and
% anything but this implies that we're intested in this element: GT is the
% type of graph we're asserting for.
assert_transition_dont_cares(GT,TransitionSig, DontCareIndices) :-
	retractall(transition_arg_dont_cares(GT,TransitionSig,_)),
	assert(transition_arg_dont_cares(GT,TransitionSig, DontCareIndices)).

% Assert that all arguments of all UNKNOWN transitions are not cared about -
% so this can be used to set dont cares for new transitions of the state space,
% without changing current dont cares.
assert_all_transition_arg_dont_cares :-
	graph_type(GT),
	transition(_,Trans,_),transition_info(Trans,_,ArgsLength,Sig),
	(transition_arg_dont_cares(GT,Sig,_) -> fail;true),
	% check if argslength is < 0, then add one if it is.
	((ArgsLength < 0) ->
		ArgLength is (-ArgsLength-1)
		; ArgLength is ArgsLength),
	make_counting_list(ArgLength,[],DontCares),
	make_counting_list(ArgLength,DontCares,ArgFormat),
	assert_transition_dont_cares(GT,Sig,ArgFormat),fail.
assert_all_transition_arg_dont_cares.

% Assert that ALL arguments of ALL transitions are not cared about -
% this writes over whatever dont cares had been set before.
assert_all_transition_arg_dont_cares(ignore_previous_dont_cares) :-
	retractall_transition_arg_dont_cares,
	graph_type(GT),
	transition(_,Trans,_),transition_info(Trans,_,ArgsLength,Sig),
	% check if argslength is < 0, then add one if it is.
	((ArgsLength < 0) ->
		ArgLength is (-ArgsLength-1)
		; ArgLength is ArgsLength),
	make_counting_list(ArgLength,[],DontCares),
	make_counting_list(ArgLength,DontCares,ArgFormat),
	assert_transition_dont_cares(GT,Sig,ArgFormat),fail.
assert_all_transition_arg_dont_cares(ignore_previous_dont_cares).

% predicate to filter out the arguments from a transition we don't care about.
remove_transition_arg_dont_cares(_GT, Tran, FilteredTran) :-
    specfile:csp_mode, Tran = io(V,Ch,_),!,
    length(V,Len), gen_underscores(Len,US),
    FilteredTran =.. [Ch|US].
remove_transition_arg_dont_cares(GT, Tran, FilteredTran) :-
	get_transition_arguments(Tran,Args),transition_info(Tran,N,Arity,Sig),name(N,Name),
	transition_arg_dont_cares(GT,Sig,DontCares),
	get_args_as_strings(Args,ArgsS),
	remove_transition_arg_dont_cares3(DontCares,ArgsS,Filtered),
	convert_string_to_code_list(Filtered,FilteredStrings,','),
	((Arity < 0) ->
		name('-->(1)',Arrow_cl)
		; Arrow_cl = []),name('(',LP),name(')',RP),
	ent03r_flatten([Name,LP,FilteredStrings,RP,Arrow_cl],FilteredTranCodeList),
	name(FilteredTranName,FilteredTranCodeList),
	FilteredTranName = FilteredTran. % put unification afterwards in case FilteredTran is already instantiated and not atomic

gen_underscores(0,R) :- !,R=[].
gen_underscores(N,['_'|T]) :- N1 is N-1, gen_underscores(N1,T).
% given 3 lists, DontCares, Original, Result, put each element of Original
% into the same index in Result, unless the same element in DontCares is '_', in
% which case put '_' in Result.
remove_transition_arg_dont_cares3([],[],[]) :- !.
remove_transition_arg_dont_cares3(['_'|Tail1], [_|Tail2], ['_'|Rest]) :-
	!, remove_transition_arg_dont_cares3(Tail1,Tail2,Rest).
remove_transition_arg_dont_cares3([_|Tail1], [G|Tail2], [G|Rest]) :-
	!, remove_transition_arg_dont_cares3(Tail1,Tail2,Rest).

:- volatile desired_user_transition/2.
:- dynamic desired_user_transition/2.

% Assert that the user wants to see all of the transitions in the alphabet, and
% remove information stored about transition argument dont cares. (use on loading machine).
assert_desired_user_transitions(alphabet) :-
	retractall(desired_user_transition(_,_)),
	retractall_transition_info,retractall_transition_arg_dont_cares,
	assert_all_transition_arg_dont_cares,get_alphabet_signatures(Ab),/* print(Ab),nl, */
	graph_type(GT),member(X,Ab),
	assert(desired_user_transition(GT,X)),fail.
assert_desired_user_transitions(alphabet).

% Assert that the user wants to see all of the transitions in the alphabet.
assert_desired_user_transitions(only_alphabet) :-
	retractall(desired_user_transition(_,_)),get_alphabet_signatures(Ab),
	graph_type(GT),member(X,Ab),
	assert(desired_user_transition(GT,X)),fail.
assert_desired_user_transitions(only_alphabet).

% Assert that the user wants to see all of the transitions in the alphabet, for
% a particular graph type.
assert_desired_user_transitions(GT,only_alphabet) :-
	retractall(desired_user_transition(GT,_)),get_alphabet_signatures(Ab),
	graph_type(GT),member(X,Ab),
	assert(desired_user_transition(GT,X)),fail.
assert_desired_user_transitions(_,only_alphabet).


:- dynamic goal_nodes/1.

assert_goal_nodes :-
	retractall(goal_nodes(_)),findall(X,(visited_expression_id(X),
	user:test_goal_in_node(X)),R),sort(R,Rr),assert(goal_nodes(Rr)).


:- dynamic use_all_operations/2,use_no_arguments/2,user_has_made_op_selection/1,
	user_has_made_arg_selection/1.
reset_use_all_ops_args :-
	retractall(use_all_operations(_,_)),
	retractall(use_no_arguments(_,_)),
	retractall(user_has_made_op_selection(_)),
	retractall(user_has_made_arg_selection(_)),
	assert(use_all_operations(signature_merge,yes)),
	assert(use_no_arguments(signature_merge,yes)),
	assert(use_all_operations(dfa_abstraction,yes)),
	assert(use_no_arguments(dfa_abstraction,yes)),
	assert(user_has_made_op_selection(no)),
	assert(user_has_made_arg_selection(no)).

user_made_op_selection :-
	(user_has_made_op_selection(yes) ->
		true
		; retractall(user_has_made_op_selection(no)),print('change just made'),nl,
		assert(user_has_made_op_selection(yes))).
user_made_arg_selection :-
	(user_has_made_arg_selection(yes) ->
		true
		; retractall(user_has_made_arg_selection(no)),
		assert(user_has_made_arg_selection(yes))).

set_use_all_operations(Type, Yes) :-
	retractall(use_all_operations(Type,_)),
	((Yes == yes) ->
		assert(use_all_operations(Type,yes)),
		assert_desired_user_transitions(Type,only_alphabet)
		; assert(use_all_operations(Type,no))).
set_use_no_arguments(Type, Yes) :-
	retractall(use_no_arguments(Type,_)),
	((Yes == yes) ->
		assert(use_no_arguments(Type,yes)),
		assert_all_transition_arg_dont_cares(ignore_previous_dont_cares)
		; assert(use_no_arguments(Type,no))).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		End utility functions
==============================================================================*/



/*==============================================================================
	DFA minimisation applied and modified to prob state space graphs
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

:- use_module(library(ordsets)).

:- dynamic 	distinguished/3,table_marked/1,equivalence_clazz/1,dfa_states_/2,
		dfa_tran_/3,final_states/1,min_dfa_state/2.

dfa_cleanup :-
	retractall(equivalence_clazz(_)),retractall(distinguished(_,_,_)),
	retractall(table_marked(_)),retractall(min_dfa_state(_,_)).

init :-
	dfa_cleanup,assert(table_marked(no)),get_dfa_states(States),init(States).

init([_|[]]).
init([H|T]) :-
	setup_distinguished_table(H, T),
	init(T).

setup_distinguished_table(_, []).
setup_distinguished_table(X, [H|T]) :-
	final_states(Final_States),
	(((member(X, Final_States), \+member(H, Final_States)) | (\+member(X, Final_States), member(H, Final_States))) ->
		assert(distinguished(X,H,yes)),retractall(table_marked(_)), assert(table_marked(yes))
		; assert(distinguished(X,H,no))),
	setup_distinguished_table(X,T).


construct_distinguished_table :-
	distinguished(A,B,no),
	dfa_tran_(A,X,Aprime),
	dfa_tran_(B,X,Bprime),
	%print('('),print(A),print(','),print(B),print(' = no)'),print(' '),print(X),print(' ('),print(Aprime),print(','),print(Bprime),print(')'),nl,
	((distinguished(Aprime,Bprime, yes) | distinguished(Bprime,Aprime,yes)) ->
		retract(distinguished(A,B,_)), assert(distinguished(A,B, yes)),
		retractall(table_marked(_)), assert(table_marked(yes))
		,print(A),print(' '),print(B),print(' '),print(yes),nl
		;true),
	fail.
construct_distinguished_table :-
	(table_marked(yes) ->
		retractall(table_marked(_)), assert(table_marked(no)), construct_distinguished_table;true).

assert_equivalence_clazz(Class, []) :-
	((\+equivalence_clazz(Class),equivalence_clazz(Any),ord_intersection(Class,Any,[]))
	   -> assert(equivalence_clazz(Class))
		; true).
assert_equivalence_clazz(Class, [H|T]) :-
	((ord_intersection(Class, H, [_|_]) ->
		ord_union(Class, H, Res),assert_equivalence_clazz(Res, T))
	 ; assert_equivalence_clazz(Class, T)).

assert_equivalence_clazzes_([]).
assert_equivalence_clazzes_([H|T]) :-
	assert_equivalence_clazz(H, T),
	assert_equivalence_clazzes_(T).

assert_equivalence_clazzes :-
	findall([X,Y],distinguished(X,Y,'no'),Res),
	sort(Res, Result),
	assert_equivalence_clazzes_(Result),
	fail.
assert_equivalence_clazzes.

unflatten([], []).
unflatten([H|T], [[H]|Res]) :-
	unflatten(T, Res).

minimize_dfa(Res) :-
	init, construct_distinguished_table, assert_equivalence_clazzes,
	findall(X,equivalence_clazz(X),ReS),% retractall(equivalence_clazz(_)),
	findall([X,Y],distinguished(X,Y,'yes'),Result),
	%print(Result),nl,
	ent03r_flatten(Result, Flat_Re),remove_dups(Flat_Re, Resu),
	sort(Resu, Result_),
	%print('*'),print(Result_),nl,
	ent03r_flatten(ReS, Flat_Res),sort(Flat_Res,Flat_Res_),
	%print(Flat_Res_),nl,
	length(Result_,LR),length(Flat_Res_,FR),
	(LR =< FR ->
		ord_intersection(Result_, Flat_Res_,_,Diff)
		; ord_intersection(Flat_Res_, Result_,_,Diff)),
	%print(Diff),nl,
	unflatten(Diff, Fat_Diff),append(ReS, Fat_Diff, Res),assert_min_dfa_states(Res).

get_representative_from_list(L,Representative) :-
	member(Representative, L),!.

assert_min_dfa_states([]).
assert_min_dfa_states([H|T]) :-
	get_representative_from_list(H,Represen),
	assert(min_dfa_state(H,Represen)),
	assert_min_dfa_states(T).

print_min_dfa_states_representatives_for_dot :-
	min_dfa_state(_,ID),
	map_dfa_state_to_id(RealRep,ID),
	(RealRep == [] -> fail;true),
	(RealRep == [root] ->
		print(ID),print('[shape=invtriangle, color=green, label=""];'),nl,fail %"#0d9d03"
		; print(ID), print('[shape=ellipse, color=black, label=\"'),print(ID),
		print('\", id='),print(ID),print('];'),nl,fail).
print_min_dfa_states_representatives_for_dot.

print_min_dfa__representatives_transitions_ :-
	min_dfa_state(_,Representative),
	dfa_tran_(Representative,Letter,Succ),
	min_dfa_state(State,SuccRepresentative),

	map_dfa_state_to_id(RealRep,Representative),
	RealRep \== [],
	map_dfa_state_to_id(RealSuccRep,SuccRepresentative),
	RealSuccRep \== [],

	member(Succ,State),
	print(Representative),
	print(' -> '),
	print(SuccRepresentative),
	(RealRep == [root] ->
		print('[style=dotted, color=black, label=\"')
		; print('[color=blue, label=\"')),
	print(Letter),
	print('\"];'),
	nl,
	fail.
print_min_dfa__representatives_transitions_.

print_min_dfa_for_dot(File) :-
	nfa_to_dfa,
	minimize_dfa(_),
	% for each min_dfa_state, get a representative state, for each letter of the alphabet, print the transition
	% and the representative state to which the transition leads
	tell(File),
	print_graph_header(minimum_dfa),
	print_min_dfa_states_representatives_for_dot,
	print_min_dfa__representatives_transitions_,
	print_graph_footer,
	told.
print_min_dfa_for_dot(_).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 		End DFA minimisation
==============================================================================*/

/*==============================================================================
		NFA to DFA
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

% Use this to find out about all of the transitions - instead of using transition/3 directly. This predicate filters
% out transitions which the user isn't interested in.
nfa_to_dfa_transition_filter(A,B,Sig,D) :-
	transition(A,B,D),
	transition_info(B,_,_,SmallSig),
	remove_transition_arg_dont_cares(dfa_abstraction, B, Sig),
	desired_user_transition(dfa_abstraction,SmallSig).

:- volatile transi_/2,map_dfa_state_to_id/2.
:- dynamic transi_/2,map_dfa_state_to_id/2.
findall_transitions(State,TransName,_) :-
	nfa_to_dfa_transition_filter(State,_,TransName,Succ),
	(transi_(State,Succ) ->
		true
		; asserta(transi_(State,Succ))),
	fail.
findall_transitions(State,_,Result) :-
	findall(Succ,transi_(State,Succ),Result),retractall(transi_(_,_)).

nfa_cleanup :-
	retractall(dfa_states_(_,_)),retractall(dfa_tran_(_,_,_)),
	retractall(final_states(_)),retractall(map_dfa_state_to_id(_,_)).

move_([],_,[]).
move_([Head|Tail],T,R) :-
	move_(Head,T,R1),
	move_(Tail,T,R2),ent03r_union(R1,R2,R).
move_(S,T,R) :-
	findall_transitions(S,T,R).
move(S,T,R) :- move_(S,T,Rp),
    retractall(transi_(_,_)),ent03r_flatten(Rp,R),!.

nfa_to_dfa_worker(Alphabet) :-
	repeat,
	(dfa_states_(S,no) -> true;!,fail),
	%print('testing S = '),print(S),nl,
	retract(dfa_states_(S,no)),
	asserta(dfa_states_(S,yes)),
	% print(treating_dfa_state(S)),nl,
	(map_dfa_state_to_id(S,_) ->
		true
		; next_equivalence_class_id(ID0),
		  % print(new_equiv_class(ID0)),nl,
		  asserta(map_dfa_state_to_id(S,ID0))),
	%print('*'),
	member(Letter,Alphabet),
	% print(computing_move(S,Letter)),nl,
	move(S,Letter,Res),
	sort(Res,Result),
	% print(move(S,Letter,Res)),nl,
	(dfa_states_(Result,_) ->
		true /* state already exists */
		; asserta(dfa_states_(Result,no)), % register new, untreated state
		  % print(new_successor_state(Result)),nl,
		 (map_dfa_state_to_id(Result,_) ->
			true
			; next_equivalence_class_id(ID1),
		      % print(new_equiv_class(ID1)),nl,
			  asserta(map_dfa_state_to_id(Result,ID1)))),
		%print('asserting '),print(Result),nl),
	map_dfa_state_to_id(S,ID2),
	map_dfa_state_to_id(Result,ID3),
	% print(dfa_tran(ID2,Letter,ID3)),nl,
	asserta(dfa_tran_(ID2,Letter,ID3)),fail.
nfa_to_dfa_worker(_).

get_dfa_states(States) :-
	findall(S,map_dfa_state_to_id(_,S),Statess),
	sort(Statess,States).



nfa_to_dfa :-
	nfa_cleanup,
	init_equivalence_class_id,
	/* findall(X,desired_user_transition(dfa_abstraction,X),Alphabet), */
	get_alphabet(dfa_abstraction,Alphabet),
	print('% Alphabet: '), print(Alphabet),nl,
	asserta(dfa_states_([root],no)),
	nfa_to_dfa_worker(Alphabet).
nfa_to_dfa :-
	get_dfa_states(States),
	map_dfa_state_to_id([root],ID1),
	(map_dfa_state_to_id([],ID2) ->
		List = [ID2,ID1], %add other final states here
		sort(List,SList),
		ord_subtract(States,SList,Finals)
	  ; ord_subtract(States,[ID1],Finals)),

	assert(final_states(Finals)).



print_dfa_from_nfa_states_id :-
	map_dfa_state_to_id(Reals,ID),
	((Reals == []) -> fail;true),

	% work out outline colour
	((member(A,Reals),transition(A,_,_)) ->	% this node not is open (**in this dfa graph)
		get_preference(dot_normal_node_colour,OutColour)
		; get_preference(dot_open_node_colour,OutColour)), % this node is open

	% work out fill colour
	((invariant_violated(C), member(C,Reals)) ->
		get_preference(dot_invariant_violated_node_colour,InColour)
		; true),

	% work out shape
	((Reals == [root]) -> % its the root
		get_preference(dot_root_shape,Shape),
		print(root),print('[id="root", shape='),print(Shape),print(', color=green, label=""];'),nl,fail
		; ((state_space:current_state_id(B),member(B,Reals)) -> % it contains the current state
			get_preference(dot_current_node_shape,Shape)
			; get_preference(dot_normal_node_shape,Shape))), % it doesn't contain the current state

	% do the printing
	print(ID),print('[id='),print(ID),print(', shape='),print(Shape),print(', color=\"'),print(OutColour),print('\"'),
	(nonvar(InColour) -> print(', style=filled, fillcolor='),print(InColour); true),print('];'),nl,
	fail.
print_dfa_from_nfa_states_id.

print_dot_for_dfa_from_nfa_ :-
	dfa_tran_(ID1,Letter,ID2),
	map_dfa_state_to_id(ID1_reals, ID1),
	map_dfa_state_to_id(ID2_reals, ID2),
	((ID2_reals == []) -> fail;true),
	((ID1_reals == [root]) ->
		print(root)
		; print(ID1)),
	print(' -> '),
	print(ID2),
	(map_dfa_state_to_id([root], ID1) ->
		print('[color=black, label=\"') %style=dotted,
		; ((ID1_reals \== ID2_reals,ord_subset(ID2_reals, ID1_reals)) ->
			print('[style=dotted,')
			; print('[')),
		print('color=blue, label=\"')),
	print(Letter),
	print('\"];'),
	nl,
	fail.
print_dot_for_dfa_from_nfa_.

print_dot_for_dfa_from_nfa(File) :-
	% use all operations if desired
	(use_all_operations(dfa_abstraction,yes) ->
		set_use_all_operations(dfa_abstraction,yes)
		; true),
	% use no arguments if desired
	(use_no_arguments(dfa_abstraction,yes) ->
		set_use_no_arguments(dfa_abstraction,yes)
		; true),
    print('% Converting NFA to DFA'),nl,
	nfa_to_dfa,
	number_of_equivalence_classes(Nr),
	format('% Writing DFA with ~w states to file: ~w~n',[Nr,File]),
	tell(File),
	print_graph_header(nfa_to_dfa),
	print_dfa_from_nfa_states_id,
	nl,
	print_dot_for_dfa_from_nfa_,
	print_graph_footer,
	told,
	(Nr>100 -> format('% Warning: viewing ~w states with dot can take time!~n',[Nr]) ; true).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 		End of NFA to DFA code
==============================================================================*/



/*==============================================================================
		Signature Merge based on a state's transitions
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  */

:- dynamic counter/1,state_signature/4,
	state_merge_equivalence_class/3,transition_printed/1,state_printed/1,
	%state_merge_filter_mode/1,
	state_merge_separate_invariant_violations/1,non_det_trans/3,
	signature_merge_show_all_transitions/1,non_def_trans/2.

% Either no or yes. The former means we don't separate states that break the
% invariant violation into their own equivalence class. The latter means we do.
state_merge_separate_invariant_violations(no).

% Either no or yes. The former means we don't view every outgoing transition for
% each equivalence class. The latter means we do.
signature_merge_show_all_transitions(yes).

set_signature_merge_show_all_transitions(H) :-
	(signature_merge_show_all_transitions(H) ->
		true
		; retractall(signature_merge_show_all_transitions(_)),
		  assert(signature_merge_show_all_transitions(H))).

%state_merge_separate_iviolations(V) :-
%	(((V == yes) ; (V == no)) ->
%		retractall(state_merge_separate_invariant_violations(_)),
%		assert(state_merge_separate_invariant_violations(V))
%		; print('invalid state_merge_separate_iviolations/1 mode: '),print(V),
%		  print(' should be either, yes, or no'),nl).

% Either normal ==> perform reduction on original state space
% OR nfa_to_dfa ==> perform reduction on state space outputted from nfa_to_dfa function
% NB: SPELL YOUR OPTION CORRECTLY OR GET WEIRD RESULTS!
%state_merge_filter_mode(normal).

% clean up the database
cleanup_state_merge :-
	retractall(state_signature(_,_,_,_)),retractall(state_merge_equivalence_class(_,_,_)),
	retractall(counter(_)),
	retractall(non_det_trans(_,_,_)),
	retractall(non_def_trans(_,_))
	/*,retractall(desired_user_transition(_,_))*/.

% get the transitions: take care that we regard the different modes
state_merge_transition_filter(A,B,Sig,D) :-
	transition(A,B,D),transition_info(B,_,_,Sig),
	desired_user_transition(signature_merge,Sig).

% get the states: take care that we regard the different modes
state_merge_state_filter(A,B,B) :- visited_expression(A,B).

% assert the transitions stemming from each state
assert_state_merges :-
	state_merge_state_filter(ID,State,_),
	compute_signature_for_state(ID,State,Res),
	% check whether to separate each invariant violation in it's own
	% equivalence class.
	(state_merge_separate_invariant_violations(no) ->
		assert_state_signature(ID,Res,ok)
		; (invariant_violated(ID) ->
			assert_state_signature(ID,Res,invariant_violated)
			; assert_state_signature(ID,Res,ok))),
	fail.
assert_state_merges.

assert_state_signature(ID,Signature,OK) :-
   (state_merge_equivalence_class(Signature,ClassID,OK)
    -> true
    ;  increment_counter(ClassID),
       assert(state_merge_equivalence_class(Signature,ClassID,OK))
   ),
   assert(state_signature(ID,Signature,OK,ClassID)).

%:- dynamic observing_variable/1.
/* set this predicate if we want to observe a particular variable instead of the signature */
%observing_variable('display_st').

%compute_signature_for_state(_,State,Res) :-
%   observing_variable(Var),store:lookup_value(Var,State,Val),!, Res=Val.
compute_signature_for_state(ID,_,Res) :-
    findall(Tr,state_merge_transition_filter(ID,_,Tr,_),Res_),
	sort(Res_,Res).

% find out the state equivalence class for a given state
find_state_merge_equivalence_class_for_state(State,ClassID) :-
	state_signature(State,_TransSig,_Status,ClassID),!.

% find out a state corresponding to a given equivalence class
find_state_for_state_merge_equivalence_class(ClassID,State) :-
	 state_signature(State,_SIG,_OK,ClassID).

% apply the state merge algorithm
state_merge :-
	cleanup_state_merge,
	% work on the correct state space e.g. the original one as opposed to
	% one outputted by another algorithm
	/*assert_desired_user_transitions,*/
	print_message('Computing Signatures'),
	assert_state_merges.

% the state merge state's colour
state_merge_state_colour(State, Colour) :-
	(state_merge_equivalence_class([],State,_) ->
		get_preference(dot_open_node_colour,Colour)
		; get_preference(dot_normal_node_colour,Colour)).

% print a state's id for the state merge - print 'root' for this if the id maps to the root
print_state_merge_state_id(State) :-
	(find_state_merge_equivalence_class_for_state(root,State) -> print(root)
				; print(State)).

% print a transition for dot
print_transition_for_dot(A,IndividualSig,C) :- Trans = [A,IndividualSig,C],
	(transition_printed(Trans) ->
		true
		; assert(transition_printed(Trans)),
		  print_state_merge_state_id(A),print('->'),print(C),
		  (find_state_merge_equivalence_class_for_state(root, A) ->
				/* Style = 'solid', Color='blue' % dealing with the root */
				(non_det_trans(A,IndividualSig,true) -> % not dealing with root
					Style = 'dotted' % there's more than one of these transitions
					; Style = 'solid'),
				  (non_def_trans(A,IndividualSig) ->
				  	Color = 'purple';Color = 'black')
				; (non_det_trans(A,IndividualSig,true) -> % not dealing with root
					Style = 'dotted' % there's more than one of these transitions
					; Style = 'solid'),
				  (non_def_trans(A,IndividualSig) ->
				  	Color = 'purple';Color = 'blue')),
		  print('[style='),print(Style),print(', color=\"'),print(Color),
		  print('\", label=\"'),print(IndividualSig),print('\"];'),nl).

% print a state for dot
print_class_as_dotstate(ClassID,OK) :-
	(state_printed(ClassID) ->
		true
		; assert(state_printed(ClassID)),
		  state_merge_state_colour(ClassID, StateColour),
		  get_preference(dot_invariant_violated_node_colour,IColor),
		  print_state_merge_state_id(ClassID),
		  print(' ['),
		  	(find_state_merge_equivalence_class_for_state(root,ClassID) ->
				print('shape=invtriangle, color=green, '),
				print('label=\"\", id=\"'),print(root)
				; ((OK = invariant_violated) ->
					print('shape=ellipse, style=filled, fillcolor="'),
					print(IColor),print('\", label=\"'),
					print_node_label(ClassID),print('\" , id=\"'),print_node_label(ClassID)
					; print('shape=ellipse, color=\"'),print(StateColour),print('\", label=\"'),
					  print_node_label(ClassID),print('\" , id=\"'),print_node_label(ClassID))),
		print('\"];'),nl).

print_node_label(ClassID) :-
   (preference(dot_print_node_ids,false)
     -> state_merge_equivalence_class(Signature,ClassID,_OK),
        print(Signature)
    ; preference(dot_print_node_info,true)
     -> state_merge_equivalence_class(Signature,ClassID,_OK),
        print(ClassID),print('\\n'), print(Signature)
     ;  print(ClassID)
   ).

% print the merged states
print_state_merge_states_for_dot :-
	state_merge_equivalence_class(_,ClassID,OK),
	print_class_as_dotstate(ClassID,OK),fail.
print_state_merge_states_for_dot.

% assert whether there is more than one transition from an equivalence class
% with the same label.
assert_non_det_trans(_State,StateID,IndividualTransSignature,_To1ID) :-
      non_det_trans(StateID,IndividualTransSignature,_Status),!. % already computed
assert_non_det_trans(State,StateID,TransSignature,To1ID) :-
	  find_state_for_state_merge_equivalence_class(StateID,AnotherState),
	  State \= AnotherState,
	  \+ not_all_transitions_added(AnotherState), % this is an open node; not yet computed
	  /* state_merge_transition_filter */
	  transition(AnotherState,_,TransSignature,To2),
	   \+(find_state_merge_equivalence_class_for_state(To2,To1ID)),!,
	   %print_message(non_det(StateID,TransSignature,To1ID,AnotherState,To2)),
       asserta(non_det_trans(StateID,TransSignature,true)).
assert_non_det_trans(_State,StateID,IndividualTransSignature,_To1ID) :-
     assert(non_det_trans(StateID,IndividualTransSignature,false)).

% print the transitions of the merged states.
print_state_merge_transitions_for_dot :-
	transition(State,_,IndividualSig,To),
	find_state_merge_equivalence_class_for_state(State,ID1),
	find_state_merge_equivalence_class_for_state(To,ID2),
	assert_non_det_trans(State,ID1,IndividualSig,ID2),
	(desired_user_transition(signature_merge,IndividualSig) ->
		print_transition_for_dot(ID1,IndividualSig,ID2)
		; (signature_merge_show_all_transitions(yes) ->
			(non_def_trans(ID1,IndividualSig) ->
				true
				; assert(non_def_trans(ID1,IndividualSig))),
			print_transition_for_dot(ID1,IndividualSig,ID2)
			; true)
	),
	fail.
print_state_merge_transitions_for_dot.

% print the merged state graph for dot
print_state_merge_for_dot :-
	% check if user has specified all transitions
	(use_all_operations(signature_merge,yes) ->
		set_use_all_operations(signature_merge,yes)
		; true),
	clean_up_after_printing,
	print_message('Performing State Merge'),
	state_merge,
	print_graph_header(state_merge),
	print_message('Generating Dot States'),
	print_state_merge_states_for_dot,
	print_message('Generating Dot Transitions'),
	print_state_merge_transitions_for_dot,
	print_graph_footer.

print_state_merge_for_dot(File) :- time(print_state_merge_for_dot2(File)).
print_state_merge_for_dot2(File) :-
	tell(File),print_state_merge_for_dot,told.


% For ProB 2.0 Graphical Visualizations
extract_graph_of_state_merge(Graph) :-
	% check if user has specified all transitions
	(use_all_operations(signature_merge,yes) ->
		set_use_all_operations(signature_merge,yes)
		; true),
	clean_up_after_printing,
	print_message('Performing State Merge'),
	state_merge,
	print_message('Generating States'),
	extract_state_merge_states(States),
	print_message('Generating Transitions'),
	extract_state_merge_transitions(Transitions),
	Graph = [States,Transitions].

extract_state_merge_states(States) :-
	findall(State,extract_state_for_graph(State),States).

extract_state_for_graph(State) :-
	state_merge_equivalence_class(Sig,Id,OK),
	State = state(Sig,Id,OK).

extract_state_merge_transitions(Transitions) :-
	findall(Trans,extract_state_merge_transition(Trans),Transitions).

extract_state_merge_transition(Transition) :-
	transition(State,_,IndividualSig,To),
	find_state_merge_equivalence_class_for_state(State,ID1),
	find_state_merge_equivalence_class_for_state(To,ID2),
	assert_non_det_trans(State,ID1,IndividualSig,ID2),
	(desired_user_transition(signature_merge,IndividualSig) ->
		extract_transition_representation(ID1,IndividualSig,ID2,Color,Style)
		; (signature_merge_show_all_transitions(yes) ->
			(non_def_trans(ID1,IndividualSig) ->
				true
				; assert(non_def_trans(ID1,IndividualSig))),
			extract_transition_representation(ID1,IndividualSig,ID2,Color,Style)
			; true)
	),
	Transition = trans(ID1,IndividualSig,ID2,Color,Style).


extract_transition_representation(A,IndividualSig,C,Color,Style) :-
	Trans = [A,IndividualSig,C],
	(transition_printed(Trans) ->
		fail
		; assert(transition_printed(Trans)),
		  	(find_state_merge_equivalence_class_for_state(root, A) ->
				/* Style = 'solid', Color='blue' % dealing with the root */
				(non_det_trans(A,IndividualSig,true) -> % not dealing with root
					Style = 'dotted' % there's more than one of these transitions
					; Style = 'solid'),
				  (non_def_trans(A,IndividualSig) ->
				  	Color = 'purple';Color = 'black')
				; (non_det_trans(A,IndividualSig,true) -> % not dealing with root
					Style = 'dotted' % there's more than one of these transitions
					; Style = 'solid'),
				  (non_def_trans(A,IndividualSig) ->
				  	Color = 'purple';Color = 'blue'))
		).



% clean up the database from assertions used during graph printing.
clean_up_after_printing :-
	retractall(state_printed(_)),retractall(transition_printed(_)).
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 		End of signature merge reduction
==============================================================================*/


:- volatile things/1.
:- dynamic things/1.
assert_associated_nodes([]).
assert_associated_nodes([H|T]) :-
	(things(H) ->
		assert_associated_nodes(T)
		; assert(things(H)),findall(X,transition(X,_,H),Res),
		  append(T,Res,Ress),assert_associated_nodes(Ress)).

check_against_list(_,[]).
check_against_list(Element,[H|T]) :-
	(state_printed(Element) ->
		true
		; assert(state_printed(Element)),
		  visualize_graph:tcltk_print_node_for_dot(Element)),
	/* (transition_printed([Element,H]) ->
		true
		; assert(transition_printed([Element,H])), */
		(transition(Element,_,H) -> visualize_graph:tcltk_print_transitions_for_dot(Element,H)
			; true),
	check_against_list(Element,T).

print_dot_from_node_list(_,File) :-
	tell(File),
	retractall(state_printed(_)),
	retractall(transition_printed(_)),
	print_graph_header(general_graph),fail.
print_dot_from_node_list(Llist,_) :-
	sort(Llist,List),
	member(N1,List),
	check_against_list(N1,List),fail.
print_dot_from_node_list(_,_) :-
	print_graph_footer,told.

print_subgraph_associated_with_node(Node,File) :-
	retractall(things(_)),
	(is_list(Node) ->
		assert_associated_nodes(Node)
		; assert_associated_nodes([Node])),
	findall(X,reduce_graph_state_space:things(X),Result),
	print_dot_from_node_list(Result,File).

print_subgraph_associated_with_invariant_violations(File) :-
	findall(X,invariant_violated(X),R),remove_dups(R,Rr),
	print_subgraph_associated_with_node(Rr,File).

print_subgraph_of_goal_nodes(File) :-
	assert_goal_nodes,goal_nodes(L),print_dot_from_node_list(L,File).

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 		End of Iterative Deepening search for a node
==============================================================================*/






