% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(state_space_dijkstra,[find_shortest_path_from_to/4,
  find_shortest_path_from_to_found_predicate/4,
  set_target_id/1, set_found_predicate/2]).

:- use_module(tools).
:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'Tools for finding shortest paths between visited states.').

% can find the shortest path from a given node to
%  any node satisfying a genering "found_predicate"

:- use_module(state_space).
:- use_module(library(heaps)).
:- use_module(library(lists)).


:- dynamic found_predicate/1. % a predicate that should return true if we have found a target node
:- dynamic mark_id/3.

set_target_id(TargetID) :- 
   retractall(found_predicate(_)),
   assert(found_predicate(TargetID)).
   
set_found_predicate(ID,COND) :-
   retractall(found_predicate(_)),
   assert((found_predicate(ID) :- COND)).
 
find_shortest_path_from_to(From,To,ResPath,ResIDL) :-
   set_target_id(To),
   find_shortest_path_from_to_found_predicate(From,ResPath,ResIDL,_).
find_shortest_path_from_to_found_predicate(From,ResPath,ResIDL,TargetID) :-
   retractall(mark_id(_,_,_)),
   empty_heap(H),
   add_to_heap(H,0,node(From,null,null),NH),
   dijkstra(NH,RP,RI,TargetID,10),
   reverse(RP,ResPath), reverse(RI,ResIDL).


dijkstra(Heap,ResOpIDPath,ResIDList,TargetID,Cnt) :-
  (Cnt>0 -> C1 is Cnt-1 ; C1=1000, write('.'),flush_output(user)),
  get_from_heap(Heap,Dist,node(ID,PrevID,TransID),NewHeap),!,
  (found_predicate(ID) -> TargetID=ID, %print(extracting_path(ID)),nl,
       assert(mark_id(ID,PrevID,TransID)),nl,
       extract_path(ID,ResIDList,ResOpIDPath) %,print(res(ResIDList)),nl
    ;  treat_node(ID,PrevID,TransID,Dist,NewHeap,NewHeap1),
       dijkstra(NewHeap1,ResOpIDPath,ResIDList,TargetID,C1)
  ).
dijkstra(_,_,_,_,_) :-
   print('NO PATH EXISTS TO A NODE WITH found_predicate TRUE ! '),nl,
   fail.

extract_path(ID,IDL,Path) :-
  mark_id(ID,PrevID,TransID),
  (PrevID=null -> IDL=[], Path=[]
   ;  any_transition(PrevID,TransID,ID), IDL=[ID|TI],Path=[TransID|TP],
      extract_path(PrevID,TI,TP)).

treat_node(ID,_,_,_,H,RH) :- mark_id(ID,_,_),!,RH=H. % node already treated
treat_node(ID,PrevID,TransID,Dist,H,RH) :- assert(mark_id(ID,PrevID,TransID)),
    findall(node(NewID,ID,NewTransID),
            any_transition(ID,NewTransID,NewID), Succs),
    D1 is Dist+1,
    add_nodes(Succs,D1,H,RH).
    
add_nodes([],_,H,H).
add_nodes([N|T],Prio,H,RH) :- add_to_heap(H,Prio,N,H1),
  add_nodes(T,Prio,H1,RH).
  
 % use_module(state_space_dijkstra), find_shortest_path_to(root,22,P,I)
  