% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)
:- module(state_viewer_images,[tcltk_get_image_list/2,

             tcltk_can_react_to_item_right_click/2,
             tcltk_react_to_item_right_click/2,

             tcltk_can_react_to_item_drag/3,
             tcltk_react_to_item_drag/2,

            get_animation_images/1, % only get animation image list
            get_animation_image_grid/6,
            get_react_to_item_right_click_options/4,
            react_to_item_right_click/6,

            html_print_history/1, html_print_history/2
            ]).

/* ------------------------------------------------------------------ */

:- use_module(module_information).
:- module_info(group,visualization).
:- module_info(description,'This module can provide a Tk graphical representation of a B state.').



:- use_module(bmachine,[b_get_machine_animation_function/2,
                        b_get_definition_string_from_machine/2,
                        get_animation_image/2]).
:- use_module(specfile,[state_corresponds_to_fully_setup_b_machine/2]).

/* First argument: image list of pairs: ImageName, ImageFilePath
   Second argument: list of triplets {type framename label}
   */

:- use_module(error_manager).
:- use_module(debug).
:- use_module(state_space,[current_expression/2]).


eval_definition_funs([],_,[]).
eval_definition_funs([anim_fun(_Nr,AnimFunction1)|T],BState,Result) :-
    eval_definition_fun(AnimFunction1, BState, FunctionRes1),
    eval_definition_funs(T,BState,FunctionRes0),
    override_anim_fun(FunctionRes1,FunctionRes0,Result).

:- use_module(xtl_interface,[xtl_animation_function_result/2, xtl_animation_image/2]).
:- use_module(specfile,[animation_mode/1, b_or_z_mode/0]).


tcltk_get_image_list(list(ImageList),list(FramesAndItems)) :-
       animation_mode(Mode),
       get_animation_images(Mode,ImageList),
       current_expression(_CurID,State),
       evaluate_animation_function(Mode,State,FunctionRes),
       display_animation_function_result(FunctionRes,FramesAndItems).


evaluate_animation_function(xtl,State,FunctionRes) :- !,
       xtl_animation_function_result(State,FunctionRes).
evaluate_animation_function(_,State,FunctionRes) :- % Mode will be 'b'
       specfile:state_corresponds_to_fully_setup_b_machine(State,BState),
       findall(anim_fun(Nr,AF),b_get_machine_animation_function(AF,Nr),AnimFuns),
       AnimFuns \= [],
       sort(AnimFuns,SAnimFuns),
       %print(animation_images(ImageList)),
      % print('Animation Function: '),
      % translate:print_bexpr(SAnimFuns),
       eval_definition_funs(SAnimFuns, BState, FunctionRes).
       %  print('full_fun'(FunctionRes)),nl.

display_animation_function_result(FunctionRes,FramesAndItems) :-
       get_min_max_row(FunctionRes,Min,Max),
       debug_println(9,visualization_matrix_min_max(Min,Max)),
       setup_frames(Min,Max,Frames), %debug_println(9,frames(Frames)),
       sort(FunctionRes,SortedFR),
       get_frame_items(SortedFR,Items),
       % print(items(Items)),nl,
       pad_text_frame_items(Items,PItems),
       append(Frames,PItems,FramesAndItems).


:- use_module(tools,[ split_last/4, split_chars/3 ]).
tcltk_can_react_to_item_right_click(TkPath,list(Names)) :-
   convert_tk_path_to_coordinates(TkPath,XCoord,YCoord),
   animation_mode(M),
   current_state_id(CurID),
   print(start_right_clicked_object(XCoord,YCoord,mode(M),state(CurID))),nl,
   findall(Name,can_react_to_item_right_click(CurID,M,XCoord,YCoord,_,Name),Names).

% for ProB 2:
get_react_to_item_right_click_options(CurID,XCoord,YCoord,Names) :-
    animation_mode(M),
    findall(Name,can_react_to_item_right_click(CurID,M,XCoord,YCoord,_,Name),Names).

tcltk_react_to_item_right_click(TkPath,ActionStr) :-
   % print(right_clicked_tk_path(TkPath)),nl,
   convert_tk_path_to_coordinates(TkPath,XCoord,YCoord),
   current_state_id(CurID),
    % print(right_clicked_object(XCoord,YCoord,mode(M),state(CurID))),nl,
   react_to_item_right_click(CurID,XCoord,YCoord,ActionStr,_,_).

:- use_module(state_space,[current_state_id/1, transition/4]).
:- use_module(xtl_interface,[xtl_animation_image_right_click_transition/3]).
:- use_module(translate,[translate_event_with_target_id/3]).
:- use_module(bmachine,[b_get_definition/5,type_with_errors/4]).
:- use_module(library(lists),[maplist/3]).
:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).

can_react_to_item_right_click(CurID,Mode,X,Y,TransitionTemplateList,Str) :-
   get_right_click_template(Mode,CurID,X,Y,TransitionTemplateList),
   get_template(TransitionTemplateList,Template), % deals with list templates
   transition(CurID,Template,_ActId,DestID),
   translate_event_with_target_id(Template,DestID,Str).

get_right_click_template(xtl,_,X,Y,Template) :-
   xtl_animation_image_right_click_transition(X,Y,Template).
get_right_click_template(b,CurID,X,Y,Template) :-
   Def_Name = 'ANIMATION_RIGHT_CLICK',
   b_get_definition(Def_Name,substitution,[_,_],RawSubst,_Deps),
   arg(1,RawSubst,Pos),
   temporary_set_preference(allow_local_operation_calls,true,CHNG),
   Scope = [prob_ids(visible),operation_bodies], % do not include variables; otherwise clash warnings
   % type definition call:
   type_with_errors(definition(Pos,'ANIMATION_RIGHT_CLICK',[integer(Pos,X),integer(Pos,Y)]),Scope,subst,Typed),
   reset_temporary_preference(allow_local_operation_calls,CHNG),
   get_b_template_for(CurID,Typed,[],Template).

% ANIMATION_RIGHT_CLICK can use the following in B:
% ANY, LET: to introduce wild cards; predicates will not (yet) be evaluated !!
% IF-ELSIF-ELSE: conditions have to be evaluable using the parameters only
% CHOICE: to provide multiple right click actions

% Operation Calls: the arguments must evaluable without access to variables and constants at the moment

:- use_module(specfile, [state_corresponds_to_set_up_constants/2]).
get_b_template_for(CurID,Subst,LV,Template) :-
       visited_expression(CurID,State),
       state_corresponds_to_set_up_constants(State,BState),
       % expand_const_and_vars_to_full_store
       get_b_template(Subst,LV,BState,Template).


get_b_template(b(E,subst,_),LocalVars,State,Template) :- get_b_template_aux(E,LocalVars,State,Template).

:- use_module(bsyntaxtree, [get_texpr_id/2,   def_get_texpr_id/2]).
get_b_template_aux(any(Paras,_Pred,Body),LocalVars,State,Template) :- !,
    append(Paras,LocalVars,LV2),
    get_b_template(Body,LV2,State,Template).
get_b_template_aux(let(Paras,_Pred,Body),LocalVars,State,Template) :- !,
    append(Paras,LocalVars,LV2),
    get_b_template(Body,LV2,State,Template).
get_b_template_aux(choice(Options),LocalVars,State,Template) :- !,
    member(A,Options),
    get_b_template(A,LocalVars,State,Template).
get_b_template_aux(if(Options),LocalVars,State,Template) :- !,
    get_b_template_for_if(Options,LocalVars,State,Template).
get_b_template_aux(operation_call(TOpID,_Result,Parameters),LocalVars,State,Template) :- !,
    def_get_texpr_id(TOpID,op(OpID)),
    % TO DO: deal with Result
    maplist(eval_parameter(LocalVars,State),Parameters,TemplateArgs),
    Template =.. [OpID|TemplateArgs].
get_b_template_aux(Subst,_LocalVars,_State,_) :-
    add_error('ANIMATION_RIGHT_CLICK','Uncovered substitution: ',Subst),fail.

get_b_template_for_if([b(if_elsif(Test,Body),_,_)|_], LocalVars,State,Template) :-
    b_interpreter:b_test_boolean_expression_wf(Test,[],State),!,
    get_b_template(Body,LocalVars,State,Template).
get_b_template_for_if([_|T],LV,State,Template) :- get_b_template_for_if(T,LV,State,Template).

eval_parameter(LocalVars,_State,Parameter,Result) :-
   get_texpr_id(Parameter,PID),
   member(LV,LocalVars), get_texpr_id(LV,PID),
   !,
   Result = _.
eval_parameter(_LocalVars,State,Parameter,Result) :-
    b_interpreter:b_compute_expression_nowf(Parameter,[],State,Result).


react_to_item_right_click(CurID,X,Y,ActionStr,ResTransitionID,ResNewID) :-  %print(ActionStr),nl,
   animation_mode(Mode),
   ((can_react_to_item_right_click(CurID,Mode,X,Y,TransitionTemplateList,ActionStr),
     get_template(TransitionTemplateList,TransitionTemplate),
     user:tcltk_perform_action(CurID,TransitionTemplate,TransitionID,NewID))
      -> format('Performed: ~w~n',[TransitionTemplate]),
         (get_template_rest(TransitionTemplateList,Rest)
           -> % there are other actions to execute
              execute_rest(Rest,TT), % currently only happens in XTL mode
              ResTransitionID = [TransitionID|TT], ResNewID = unknown
              % NewID points to first NewID, not to target; TO DO: modify command in case we support XTL in ProB2 UI
            ; ResTransitionID = TransitionID, ResNewID=NewID
         )
     ; format('Cannot execute transition ~w~n Right-Click on (~w,~w)~n',[ActionStr,X,Y]),
       fail
    ).

% note: a single click is also a item_drag from an object to itself
tcltk_can_react_to_item_drag(FromTkPath,ToTkPath,Image) :-
   animation_mode(M),
   convert_tk_path_to_coordinates(FromTkPath,FromX,FromY),
   convert_tk_path_to_coordinates(ToTkPath,ToX,ToY),
   can_react_to_item_drag(M,FromX,FromY,ToX,ToY,ImageNr),
   (ImageNr=none -> Image=''
     ; get_animation_image_and_label(M,ImageNr,_,Lbl) -> Image=Lbl
     ; Image='').

tcltk_react_to_item_drag(FromTkPath,ToTkPath) :-
   convert_tk_path_to_coordinates(FromTkPath,FromX,FromY),
   convert_tk_path_to_coordinates(ToTkPath,ToX,ToY),
   debug_println(20,dragged_object(FromX,FromY,ToX,ToY)),
   animation_mode(M),
   react_to_item_drag(M,FromX,FromY,ToX,ToY).

:- use_module(xtl_interface,[xtl_animation_image_click_transition/6]).
can_react_to_item_drag(Mode,FromX,FromY,ToX,ToY,ImageNr) :-
   current_state_id(CurID),
   get_item_drag_template_and_image(Mode,FromX,FromY,ToX,ToY,TransitionTemplateList,ImageNr),
   get_template(TransitionTemplateList,TransitionTemplate),
   transition(CurID,TransitionTemplate,_ActId,_DestID).

% transition templates can also be lists; we just check if first action is feasible
get_template([H|_],Template) :- !, % in XTL mode we allow lists of actions
   Template=H.
get_template(Template,Template). % in B,TLA,Z,... mode we only execute a single event

get_template_rest([_|T],T).

get_item_drag_template_and_image(xtl,FromX,FromY,ToX,ToY,TransitionTemplate,ImageNr) :- !,
   xtl_animation_image_click_transition(FromX,FromY,ToX,ToY,TransitionTemplate,ImageNr).
get_item_drag_template_and_image(b,FromX,FromY,ToX,ToY,Template,1000) :-
   % TO DO: provide a way to return image number
   Def_Name = 'ANIMATION_CLICK',
   b_get_definition(Def_Name,substitution,[_,_,_,_],RawSubst,_Deps),
   arg(1,RawSubst,Pos),
   temporary_set_preference(allow_local_operation_calls,true,CHNG),
   Scope = [prob_ids(visible),operation_bodies], % do not include variables; otherwise clash warnings; variables already included in operation_bodies
   type_with_errors(definition(Pos,'ANIMATION_CLICK',
        [integer(Pos,FromX),integer(Pos,FromY),integer(Pos,ToX),integer(Pos,ToY)]),Scope,subst,Typed),
   reset_temporary_preference(allow_local_operation_calls,CHNG),
   State = [], % TO DO: get full state like for right click
   get_b_template(Typed,[],State,Template).

react_to_item_drag(Mode,FromX,FromY,ToX,ToY) :-
   get_item_drag_template_and_image(Mode,FromX,FromY,ToX,ToY,Templ1,_),
   !,
   ((get_item_drag_template_and_image(Mode,FromX,FromY,ToX,ToY,TransitionTemplateList,_),
     get_template(TransitionTemplateList,TransitionTemplate),
     user:tcltk_perform_action(TransitionTemplate,_))
      -> (get_template_rest(TransitionTemplateList,Rest)
           -> % there are other actions to execute
              execute_rest(Rest,_)
            ; true
         )
     ; format('Cannot execute transition ~w~n Mouse Drag (~w,~w)->(~w,~w)~n',[Templ1,FromX,FromY,ToX,ToY]),
       fail
    ).
react_to_item_drag(_,_,_,_,_) :- fail.

% execute remaining templates of action
execute_rest([],Res) :- !, Res=[].
execute_rest([H|T],[TransitionID|TT]) :- !,
           current_state_id(CurID), %print(try_auto_play(CurID,H)),nl,
           user:compute_all_transitions_if_necessary,
           current_state_id(CurID),
           user:tcltk_perform_action(CurID,H,TransitionID,_NewID), %print(perf(H,_NewID)),nl,
           execute_rest(T,TT).
execute_rest(T,[]) :- format('~n*** Cannot execute rest of transition template ~w~n',[T]).

% if successful convert a Tk Widget Path like .stateviewer.sviewFrame.frame1.item1_3
% to coordinates of the state_viewer module: (1,3)
convert_tk_path_to_coordinates(TkPath,XCoord,YCoord) :-
   split_last(TkPath,'.',_,NameOfItemClicked),
   atom_codes(NameOfItemClicked,Codes),
   get_suffix(Codes,Rest),
   !,
   split_chars(Rest,"_",[YCC,XCC]),  % note: coordinates reversed in label
   number_codes(XCoord,XCC),
   number_codes(YCoord,YCC).
convert_tk_path_to_coordinates(TkPath,_,_) :- format('*** Cannot convert TkPath ~w~n',[TkPath]),fail.

get_suffix(Codes,Rest) :- append("item",Rest,Codes),!.
get_suffix(Codes,Rest) :- append("text",Rest,Codes),!.

:- use_module(custom_explicit_sets,[try_expand_custom_set/2, is_set_value/2]).

eval_definition_fun(AnimFunction,BState, FunctionRes) :-
       b_interpreter:b_compute_expression_nowf(AnimFunction,[],BState,FunctionResCl),
       %%  nl, print('FunctionResult'(FunctionResCl)),nl,
       is_set_value(FunctionResCl,eval_definition_fun),
       try_expand_custom_set(FunctionResCl, ExpCl),
       (translate_to_ints(ExpCl, FunctionRes)
          -> true /* already translate to ints so that override will work properly */
          ; check_anim_function(ExpCl),fail
       )  % ,print('fres'(FunctionRes)),nl
       .

check_anim_function([]) :- !.
%check_anim_function(int(_)) :- !. % just a single picture; now gets converted by translate_to_ints
check_anim_function([H|_]) :- !,
  (H = ((Row,Col),_Img) ->
         (((get_int(Row,_),get_int(Col,_)) -> true
          ;  add_error(state_viewer_images,'Elements of ANIMATION_FUNCTION should be convertible to (INT*INT)*INT: ',H))
         )
   ; H = (Row,(Col,_Img)) ->
         (((get_int(Row,_),get_int(Col,_)) -> true
          ;  add_error(state_viewer_images,'Elements of ANIMATION_FUNCTION should be convertible to INT*(INT*INT): ',H))
        )
     ;  add_error(state_viewer_images,
            'Elements of the ANIMATION_FUNCTION should be of type (INT*INT)*INT: ',H)
   ).
check_anim_function((A,B)) :- !,
   add_error(state_viewer_images,'ANIMATION_FUNCTION should be a set of pairs, not a pair: ',(A,B)).
check_anim_function(X) :- !,
   add_error(state_viewer_images,'ANIMATION_FUNCTION should be of type (INT*INT) +-> INT (or similar): ',X).

:- use_module(library(avl)).

translate_to_ints(In,Out) :- empty_avl(E),translate_to_ints(In,E,Out).
translate_to_ints([],_,[]) :-!.
translate_to_ints(X,_,Res) :- get_int(X,IX),!,Res=[((1,1),int(IX))]. % a single picture
translate_to_ints([H|T], AVL, Res) :-
   translate_element(H,AVL, NewAVL, Res, NewRes),
   translate_to_ints(T,NewAVL,NewRes).

translate_element(((Row,Col),Img),AVL, NewAVL, Res, NewRes) :- !,
   get_int(Row,RI), get_int(Col,CI), (Img=II),
   add_image(RI,CI,II, AVL, NewAVL, Res, NewRes).
translate_element((Row,(Col,Img)),AVL, NewAVL, Res, NewRes) :- !,
   translate_element(((Row,Col),Img), AVL, NewAVL, Res, NewRes).
translate_element((Row,Set),AVL,NewAVL,Res,NewRes) :- % Set can be relation or sequence
   % expand Set and associate elements with Row
   is_set_value(Set,translate_element),
   try_expand_custom_set(Set, ESet),
   findall((Row,El),member(El,ESet),All),
   l_translate(All,AVL,NewAVL,Res,NewRes).



l_translate([],AVL,AVL,Res,Res).
l_translate([H|T],AVL,NewAVL,Res,NewRes) :- translate_element(H,AVL,NewAVL1,Res,NewRes1),
   l_translate(T,NewAVL1,NewAVL,NewRes1,NewRes).

add_image(RI,CI,II, AVL, NewAVL, Res, NewRes) :-
   (avl_fetch((RI,CI),AVL,ExistingValue)
    -> add_error(state_viewer_images,
          'Animation function defines two images for (Row,Col):[img1,img2] = ',((RI,CI):[ExistingValue,II])),
       NewAVL = AVL, Res=NewRes
     ; avl_store((RI,CI),AVL,II,NewAVL), Res=[((RI,CI),II)|NewRes]
   ).

/* a version of override which does not worry about correct typing */
% override_anim_fun(Fun1, Fun2, Res)  :: Fun2 overrides Fun1
override_anim_fun([],Fun2,Res) :- !, Res=Fun2.
override_anim_fun(int(_),Fun2,Res) :- !, Res=Fun2.
override_anim_fun([((RowT,Col),Img)|T],Fun2,Res) :- !,
   (member(((RowT,Col),_),Fun2)
     -> Res = ResT % we use version from Fun2
      ; Res = [((RowT,Col),Img)|ResT]
   ), override_anim_fun(T,Fun2,ResT).
override_anim_fun(X,F,R) :-
   add_error(state_viewer_images,'Unknown Format for ANIMATION_FUNCTION_DEFAULT', X),
   R=F.

% get minimum and maximum index for rows
get_min_max_row([((Row,_Col),_Img)|T],Min,Max) :- !,
   number(Row), % this is already in a format for Tcl/Tk; not int(.)
   get_min_max_row2(T,Row,Row,Min,Max).
get_min_max_row([],_,_) :- !,
   print('**** Empty ANIMATION_FUNCTION result'),nl, fail.
get_min_max_row(Res,_,_) :-
   add_error(state_viewer_images,
             'Unknown Format for ANIMATION_FUNCTION result, must be of type INT or (INT*INT)+->INT',Res),
   fail.

get_min_max_row2([],Min,Max,Min,Max).
get_min_max_row2([((Row,_Col),_Img)|T],Min,Max,MinR,MaxR) :-
    (Row<Min -> Min1=Row ; Min1=Min),
    (Row>Max -> Max1=Row ; Max1=Max),
    get_min_max_row2(T,Min1,Max1,MinR,MaxR).

% get minimum and maximum index for columns
get_min_max_col([((_Row,Col),_Img)|T],Min,Max) :- !,
   number(Col), % this is already in a format for Tcl/Tk; not int(.)
   get_min_max_col2(T,Col,Col,Min,Max).
get_min_max_col2([],Min,Max,Min,Max).
get_min_max_col2([((_Row,Col),_Img)|T],Min,Max,MinR,MaxR) :-
    (Col<Min -> Min1=Col ; Min1=Min),
    (Col>Max -> Max1=Col ; Max1=Max),
    get_min_max_col2(T,Min1,Max1,MinR,MaxR).

get_int(Nr,R) :- number(Nr),!,R=Nr.
get_int(int(R),R).
get_int(fd(R,_Name),R).
get_int(pred_true /* bool_true */,1).
get_int(pred_false /* bool_false */,0).
get_int(string(A),Integer) :-  % convert something like "Train1" -> 1
   atom_codes(A,Codes), reverse(Codes,RC),
   get_number_prefix(RC,Prefix),
   Prefix \= [], reverse(Prefix,RP),
   number_codes(Integer,RP).

get_number_prefix([N|T],[N|RT]) :- N >= 48, N =< 57, % Ascii of a digit 0-9
   !, get_number_prefix(T,RT).
get_number_prefix(_,[]).

:- use_module(bmachine,[b_get_true_expression_definition/1]).
% add extra whitespace to make all text fields the same length
% TO DO: maybe only pad within a column; not overall
pad_text_frame_items(FrameItems,NewFrameItems) :- b_or_z_mode,
   (b_get_true_expression_definition('ANIMATION_STR_JUSTIFY_LEFT') ;
    b_get_true_expression_definition('ANIMATION_STR_JUSTIFY_RIGHT') ),
   longest_text(FrameItems,0,Len),
   Len>0,
   !,
   %print(longest(Len)),nl,
   pad_text_frame_items_to(FrameItems,Len,NewFrameItems).
pad_text_frame_items(F,F).


pad_text_frame_items_to([],_,[]).
pad_text_frame_items_to([ItemLabel,Frame,Text|TR],Len,[ItemLabel,Frame,NewText|NTR]) :-
   (is_text_type(ItemLabel)
      -> atom_codes(Text,Codes),
         length(Codes,TLen),
         AdditionalSpaces is Len-TLen, gen_spaces(AdditionalSpaces,AS),
         (b_get_true_expression_definition('ANIMATION_STR_JUSTIFY_RIGHT')
           -> append(AS,Codes,NewCodes)
           ;  append(Codes,AS,NewCodes) % 34 -> ' 39 -> "
         ),
         atom_codes(NewText,NewCodes)
       ; NewText=Text),
   pad_text_frame_items_to(TR,Len,NTR).

gen_spaces(Nr,R) :- Nr<1,R=[]. % 34 -> '
gen_spaces(Nr,[32|T]) :- N1 is Nr-1, gen_spaces(N1,T).


% find longest text label; so that we can pad the other ones with spaces
longest_text([],R,R).
longest_text([ItemLabel,_Frame,Text|TR],Acc,R) :-
   is_text_type(ItemLabel),!,
   atom_length(Text,Len),
   (Len>Acc -> NAcc = Len ; NAcc=Acc),
   longest_text(TR,NAcc,R).
longest_text([_,_,_|TR],Acc,R) :- longest_text(TR,Acc,R).

is_text_type(ItemLabel) :- atom_codes(ItemLabel,Codes), append("text",_,Codes).

get_frame_items([],[]).
get_frame_items([((Row,Col),Img)|T],[ItemLbl,FrameLbl,ImageLbl|TR]) :-
    convert_nr_into_tk_label(Row,"frame",FrameLbl),
    ((get_int(Img,ImgNr),images(ImgNr,_,ImageLbl))
      -> convert_2nr_into_tk_label(Row,Col,"item",ItemLbl)
      ;  convert_2nr_into_tk_label(Row,Col,"text",ItemLbl),
         ((get_int(Img,ImgNr),get_animation_string(ImgNr,ImageLbl)) -> true
            ; translate:translate_bvalue(Img,ImageLbl)
         )
    ),
    get_frame_items(T,TR).

setup_frames(Min,Max,[]) :- Min>Max,!.
setup_frames(Min,Max,[frame, FrameLbl, top |T]) :-
    convert_nr_into_tk_label(Min,"frame",FrameLbl), M1 is Min+1,
    setup_frames(M1,Max,T).

:- dynamic images/3.


get_animation_images(Mode,Res) :- clear_animation_images,
    findall( img(Nr,S,Lbl), get_animation_image_and_label(Mode,Nr,S,Lbl), List),
    flatten_img_list(List,Res).
clear_animation_images :- retractall(images(_,_,_)).

flatten_img_list([],[]).
flatten_img_list([img(Nr,S,Lbl)|T],[Lbl,S|FT]) :- assert(images(Nr,S,Lbl)), flatten_img_list(T,FT).


get_animation_image_and_label(Mode,Nr,S,Lbl) :-
    get_animation_image_in_mode(Mode,Nr,S),
    convert_nr_into_tk_label(Nr,"IMG",Lbl).

get_animation_image_in_mode(xtl,Nr,S) :- !, xtl_animation_image(Nr,S).
get_animation_image_in_mode(b,Nr,S) :- !, get_animation_image(Nr,S).
get_animation_image_in_mode(csp_and_b,Nr,S) :- get_animation_image(Nr,S).
% csp not covered

get_animation_string(Nr,S) :- b_or_z_mode,
         number_codes(Nr,TailCodes),
         /* ANIMATION_STR */
         atom_codes(Def_Name,[65,78,73,77,65,84,73,79,78,95,83,84,82|TailCodes]),
         b_get_definition_string_from_machine(Def_Name,S).


convert_nr_into_tk_label(Nr,Prefix,Label) :- number_codes(Nr,Codes),
        append(Prefix,Codes,C),
        atom_codes(Label,C).

convert_2nr_into_tk_label(Nr1,Nr2,Prefix,Label) :-
        number_codes(Nr1,Codes1), number_codes(Nr2,Codes2),
        append("_",Codes2,C2), append(Codes1,C2,Codes),
        append(Prefix,Codes,C),
        atom_codes(Label,C).

% --------------------------

% a function for Java FX or other graphical frameworks
% | ?- get_animation_images(L).
% L = [image_file(0,'images/empty_box_white.gif'),...,image_file(6,'images/F.gif')] ?

get_animation_images(SList) :-
    animation_mode(Mode),
    findall( image_file(Nr,File), get_animation_image_in_mode(Mode,Nr,File), List),
    sort(List,SList).

% a function for Java FX or other graphical frameworks
% | ?- get_animation_image_grid(1,G,M,X).
%G = [entry(1,1,image(0)),entry(1,2,image(0)),entry(1,3,image(0)),entry(1,4,image(0)),...,entry(Row,Col,image(...)),(...,...)|...],
%M = 1,
%X = 6 ?

:- use_module(state_space,[visited_expression/2]).
:- use_module(library(lists)).
get_animation_image_grid(ID,Matrix,MinRow,MaxRow,MinCol,MaxCol) :-
       visited_expression(ID,State),
       animation_mode(Mode),
       evaluate_animation_function(Mode,State,FunctionRes),
       get_min_max_row(FunctionRes,MinRow,MaxRow),
       get_min_max_col(FunctionRes,MinCol,MaxCol),
       maplist(convert_animation_function(Mode),FunctionRes,Matrix).


convert_animation_function(Mode,((Row,Col),Res),
                           entry(Row,Col,TranslatedRes)) :-
      convert_aux(Mode,Res,TranslatedRes).
convert_aux(Mode,Img,image(ImgNr)) :- get_int(Img,ImgNr),get_animation_image_in_mode(Mode,ImgNr,_File),!.
convert_aux(_Mode,Str,text(StringLbl)) :- get_int(Str,StrNr),get_animation_string(StrNr,StringLbl),!.
convert_aux(b,Value,text(PPValue)) :- translate:translate_bvalue(Value,PPValue),!.
convert_aux(_,Value,text(Value)).


% export to HTML
%html_print_animation_image_grid(ID) :- html_print_animation_image_grid(user_output,ID,'').

html_print_animation_image_grid(Stream,ID,FilePrefix) :-
   get_animation_image_grid(ID,Matrix,MinRow,_MaxRow,_MinCol,_MaxCol),
   sort(Matrix,SortedMatrix),
   start_tag(Stream,table),
   start_tag(Stream,tr),animation_mode(Mode),
   html_print(SortedMatrix,Stream,MinRow,Mode,FilePrefix),
   end_tag(Stream,table).

html_print([],Stream,_,_,_FilePrefix) :- end_tag(Stream,tr).
html_print([entry(Row,_Col,Img)|T],Stream,LastRow,Mode,FilePrefix) :-
   (Row>LastRow -> end_tag(Stream,tr),start_tag(Stream,tr) ; true),
  % TO DO: insert empty cells when we jump over Columns
   html_image(Img,Stream,Mode,FilePrefix),
   html_print(T,Stream,Row,Mode,FilePrefix).

html_image(image(ImgNr),Stream,Mode,FilePrefix) :- get_animation_image_in_mode(Mode,ImgNr,File),!,
   ((FilePrefix='' ; is_absolute_path(File))
    -> format(Stream,'  <th><img src="~w"></th>~n',[File])
     ; format(Stream,'  <th><img src="~w~w"></th>~n',[FilePrefix,File])
   ).
html_image(text(T),Stream,_,_) :- !,
   format(Stream,'  <th><tt>~w</tt></th>~n',[T]).
html_image(T,Stream,_,_) :-
   format(Stream,'  <th>ERROR: ~w</th>~n',[T]).

start_tag(Stream,T) :- format(Stream,'<~w>~n',[T]).
end_tag(Stream,T) :- format(Stream,'</~w>~n',[T]).

:- use_module(bmachine,[b_get_machine_animation_expression/1]).
:- use_module(translate,[translate_bvalue/2,translate_bexpression/2]). %translate_event_with_target_id/3]).
html_print_animation_expression(Stream,StateID) :-
       get_animation_expression_label_and_value(StateID,S1,S2),
       format(Stream,'~w=~w~n',[S1,S2]).
html_print_animation_expression(_,_).

get_animation_expression_label_and_value(StateID,S1,S2) :- b_or_z_mode,
       b_get_machine_animation_expression(AExpr),
       visited_expression(StateID,State),
       specfile:state_corresponds_to_fully_setup_b_machine(State,BState),
       b_interpreter:b_compute_expression_nowf(AExpr,[],BState,AValue),
       translate:translate_bexpression(AExpr,S1),
       translate:translate_bvalue(AValue,S2).

:- use_module(state_space,         [transition/4,visited_expression/2,history/1, op_trace_ids/1,
                                    current_expression/2, current_state_id/1,
                                    get_action_term_trace/1, get_state_id_trace/1]).


html_print_history(File) :- html_print_history(File,3).
html_print_history(File,MaxOpDebugLevel) :-
   get_state_id_trace([root|StateIds]),
   get_action_term_trace(RAT), reverse(RAT,Actions),
   % TO DO: adapt path of images so that they work in File
   (currently_opened_file(SpecFile)
    -> get_parent_directory(SpecFile,SParent),
       absolute_file_name(SParent, SAbsParent), %, [relative_to(Directory)]),
       get_parent_directory(File,TParent),
       absolute_file_name(TParent, TAbsParent), %, [relative_to(Directory)]),
       compute_prefix(SAbsParent,TAbsParent,FilePrefix)
     ; SpecFile=unknown, FilePrefix='.'),
   open(File,write,Stream,[encoding('UTF-8')]),
   html_header(Stream,true),
   html_info(Stream,Actions,SpecFile),
   html_print_history2(Stream,1,Actions,StateIds,none,FilePrefix,MaxOpDebugLevel),
   html_footer(Stream),
   close(Stream).

:- use_module(tools_strings,[ajoin_with_sep/3]).
:- use_module(tools,[split_atom/3,is_absolute_path/1]).

% compute a file prefix that needs to be added to relocate relative path for OldPath to paths for NewPath:
compute_prefix(OldPath,NewPath,FilePrefix) :-
   %print(old(OldPath)),nl, print(new(NewPath)),nl,
   split_atom(OldPath,['/'],O1),
   split_atom(NewPath,['/'],N1),
   max_prefix(O1,N1,_CommonPrefix,O2,N2), %print(common(CommonPrefix)),nl,
   maplist(gendotdot,N2,N3),
   append(N3,O2,UpdatePath),
   (UpdatePath=[] -> FilePrefix = ''
     ; append(UpdatePath,[''],U1),
       ajoin_with_sep(U1,'/',FilePrefix)),
    debug_println(9,update_prefix_for_images(FilePrefix)).

gendotdot(_,'..').

max_prefix([H|T],[H|T2],[H|PT],Suffix1,Suffix2) :- !,
   max_prefix(T,T2,PT,Suffix1,Suffix2).
max_prefix(S1,S2,[],S1,S2).

html_print_state(Stream,ID,FilePrefix) :-
   (html_print_animation_image_grid(Stream,ID,FilePrefix)
    -> true ; format(Stream,'No visualisation available for state id ~w~n',[ID])),
   html_print_animation_expression(Stream,ID).

:- use_module(translate,[translate_event/2]). %translate_event_with_target_id/3]).
html_print_history2(_,_,[],[],_,_,_).
html_print_history2(Stream,_,[],[ID|_],_,FilePrefix,_) :-
   html_print_state(Stream,ID,FilePrefix).
html_print_history2(Stream,StepNr,[Action|AT],[ID|T],PreviousID,FilePrefix,MaxOpDebugLevel) :-
   translate_event(Action,AS),
   format(Stream,'~n~n<h2>~w. ~w</h2>~n',[StepNr,AS]), flush_output(Stream),
   html_print_action(Stream,Action,PreviousID,ID,MaxOpDebugLevel),
   html_print_state(Stream,ID,FilePrefix),
   S1 is StepNr+1,
   html_print_history2(Stream,S1,AT,T,ID,FilePrefix,MaxOpDebugLevel).

:- use_module(specfile,[get_operation_name/2]). %get_operation_arguments
:- use_module(bvisual2,[html_debug_operation_for_stateid/4]).
html_print_action(Stream,Action,PreviousID,_ID,MaxLevel) :-
   MaxLevel > 0, PreviousID \= none,
   get_operation_name(Action,OpName),
   !,
   % print operation guards up to MaxLevel nesting, TODO: use parameter values for this action
   (html_debug_operation_for_stateid(Stream,OpName,PreviousID,MaxLevel) % fails for INITIALISATION, SETUP_CONSTANTS
     -> true
      ; true). %add_internal_error('Failed: ',html_debug_operation_for_stateid(OpName,ID,MaxLevel))).
html_print_action(_,_,_,_,_).

html_header(Stream,Border) :-
  format(Stream,'<html>~n<head>~n<style>~ntable, th, td {~n',[]),
  (Border=true -> format(Stream,'    border: 1px solid black;~n',[]) ; true),
  format(Stream,'    border-collapse: collapse;~n}~nth, td {~n    padding: 0px;~n}~n</style>~n</head>~n<body>',[]).


html_footer(Stream) :-
  format(Stream,'</body>~n</html>~n',[]).

:- use_module(version, [version_str/1, revision/1, lastchangeddate/1]).
:- use_module(library(system),[ datime/1]).
:- use_module(specfile,[currently_opened_file/1]).
html_info(Stream,A,SpecFile) :-
   format(Stream,'<h2>Information</h2>~n<ul>~n',[]),
   datime(datime(Year,Month,Day,Hour,Min,_Sec)),
   format(Stream,' <li>Trace generated on ~w.~w.~w at ~w:~w</li>~n',[Day,Month,Year,Hour,Min]),
   length(A,Len),
   format(Stream,' <li>Trace length: ~w</li>~n',[Len]),
   animation_mode(Mode),
   format(Stream,' <li>Specification file (~w): ~w</li>~n',[Mode,SpecFile]),
   ((SpecFile \= unknown,get_file_revision_info(SpecFile,GitRev))
     -> format(Stream,' <li>Git revision of file: ~s</li>~n',[GitRev]) ; true),
   ((SpecFile \= unknown,get_HEAD_revision_info(SpecFile,GitHeadRev))
     -> format(Stream,' <li>Git revision of repository HEAD: ~s</li>~n',[GitHeadRev]) ; true),
   version_str(V), revision(R), lastchangeddate(LD),
   format(Stream,' <li>ProB version: ~w</li>~n',[V]),
   format(Stream,' <li>ProB revision: ~w (~w)</li>~n',[R, LD]),
   format(Stream,'</ul>~n',[]).

% b_get_main_filename -> Project Info:, Model Version/Revision ? Can we ask git ?
% git ls-files -s VSS_TrainStatus_HL3.mch diff
% git rev-parse HEAD

:- use_module(system_call,[get_command_path/2, system_call_env/6]).
:- use_module(tools,[get_parent_directory/2]).

get_file_revision_info(File,GitOutputText) :-
    get_parent_directory(File,CurWorkingDir),
    run_git_command(['ls-files','-s',File],CurWorkingDir,GitOutputText).
get_HEAD_revision_info(ReferenceFile,GitOutputText) :-
    get_parent_directory(ReferenceFile,CurWorkingDir),
    run_git_command(['rev-parse','HEAD'],CurWorkingDir,GitOutputText).


run_git_command(COMMAND,CurWorkingDir,GitOutputText) :-
    get_command_path(git,GitPath),
    (debug_mode(on) -> format(user_output,'Git: ~w with CWD: ~w~n',[GitPath,CurWorkingDir]) ; true),
    on_exception(E,
          system_call_env(GitPath,COMMAND,[cwd(CurWorkingDir)],GitOutputText,GitError,ExitCode),
          (debug_format(9,'Git exception: ~w~n',[E]),fail)),
      (debug_mode(on)
       -> format(user_output,'git command ~w~n:~s~nError: ~s~nExitCode:~w~n',
                              [COMMAND,GitOutputText,GitError,ExitCode]) ; true),
      ExitCode = exit(0), GitError=[].


