:- module(synthesis_util,[get_distinguishing_state_or_trans/8,
                          get_valid_transitions_for_operation/5,
                          get_invariant_violating_transition/6,
                          get_valid_and_invalid_equality_predicates_for_operation/6,
                          get_distinct_states/5,
                          get_valid_and_invalid_equality_predicates_for_invariants/4,
                          get_violated_and_valid_machine_invariants/4,
                          get_invariant_violating_example/4,
                          get_involved_variables/4,
                          get_invariant_violating_vars_from_examples/3,
                          get_single_constant_or_set_operator_id/3,
                          get_input_output_nodes_from_bindings/3,
                          get_output_nodes_from_bindings/2,
                          get_location_var_node_by_info_term/3,
                          get_location_var_node_by_name/3,
                          get_var_node_by_info_term/3,
                          get_operation_params_from_ids/2,
                          set_operation_params_domain/2,
                          get_guard_list_from_operation/2,
                          get_current_vars_from_example/2,
                          get_input_nodes_from_bindings/2,
                          get_varname_from_node_info/2,
                          get_varname_from_id_node/2,
                          get_valid_states/3,
                          get_machinevars_by_name/2,
                          get_binding_for_name_or_value/3,
                          get_valid_inputs_for_operation/3,
                          b_get_typing_predicate_for_vars/2,
                          b_get_typed_invariant_from_machine/1,

                          expand_state_cache_if_guard_and_solution/6,
                          add_distinguishing_states_from_cache/6,
                          conjunct_previous_invariant_or_guard_if_necessary/4,
                          check_transitions_against_machine_operations/4,
                          check_transition_against_machine_operations/5,
                          check_negative_inputs_against_operation_precondition/2,
                          replace_preds_by_subst_if_neccessary/3,
                          replace_precondition_of_operation/3,
                          contains_equivalent_state/3,
                          find_typed_identifier_uses_with_info/2,
                          delete_numbers_from_atom/2,
                          create_equality_nodes_from_example/2,
                          adapt_machine_code_for_operations/2,

                          %b_synthesis
                          b_get_machinevars_with_operation_params/1,
                          remove_prime/2,
                          parallel_to_conjunction/2,
                          prime_or_unprime_variables_in_ast/4,
                          parse_and_typecheck_value/2,
                          keep_asserted_initial_examples/6,
                          prepare_solution/5,
                          finalize_synthesized_action/3,
                          split_input_output_tuples/3,
                          prepare_synthesized_guards_for_action/2,
                          %

                          mark_examples_as_initial/4,
                          reduce_input_to_violating_vars/3,

                          is_ast/1,
                          zip/3]).

:- use_module(synthesis(logging)).

:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(b_interpreter),[b_test_boolean_expression_wf/3]).
:- use_module(probsrc(solver_interface),[solve_predicate/3]).
:- use_module(smt_solvers_interface(smt_solvers_interface),[smt_solve_predicate/4]).
:- use_module(probsrc(bmachine_eventb),[prime_variable/2,is_primed_id/1]).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(before_after_predicates),[before_after_predicate_with_equalities/3]).
:- use_module(probsrc(btypechecker),[btype/7]).
:- use_module(probsrc(parsercall),[parse/3]).
:- use_module(probsrc(translate),[translate_bexpression/2,translate_machine/3,generate_typing_predicates/2]).

:- use_module(library(lists)).
:- use_module(library(sets),[subtract/3,intersection/3]).

:- use_module(extension('plspec/plspec/plspec')).

b_get_machinevars_with_operation_params(MachineVarsWithParams) :- 
    b_get_machine_variables(MachineVars) ,
    b_get_machinevars_with_operation_params_aux(MachineVars,MachineVarsWithParams).
b_get_machinevars_with_operation_params_aux(MachineVars,MachineVarsWithParams) :- 
    on_exception(error(existence_error(_,_),_),b_synthesis:additional_machinevars(AdditionalMachineVars),fail) , ! , 
    append(MachineVars,AdditionalMachineVars,MachineVarsWithParams).
b_get_machinevars_with_operation_params_aux(MachineVars,MachineVars).

parse_and_typecheck_value(Value,ValueAst) :- 
    atom_codes(Value,L) , 
    parse(expression,L,Parsed) , 
    btype(bsynthesis,Parsed,openenv(_,env_empty),_,TempAst,[],_) , 
    set_type_if_necessary(TempAst,ValueAst).

get_machinevars_by_name(VarNames,CurrentVars) :-
    b_get_machinevars_with_operation_params(MachineVarsWithParams) , 
    get_machinevars_by_name_aux(VarNames,MachineVarsWithParams,[],CurrentVars).
get_machinevars_by_name_aux([],_,Acc,Acc).
get_machinevars_by_name_aux([VarName|T],MachineVars,Acc,CurrentVars) :-
    member(b(identifier(VarName),Type,Info),MachineVars) ,
    get_machinevars_by_name_aux(T,MachineVars,[b(identifier(VarName),Type,Info)|Acc],CurrentVars).

% True if the given set of states contains an equivalent state.
contains_equivalent_state(State,States,EquivalentState) :- 
  length(State,VarAmount) , 
  contains_equivalent_state(VarAmount,State,States,EquivalentState).
contains_equivalent_state(_,_,[],_) :- ! , fail.
contains_equivalent_state(VarAmount,State,[ExampleState|_],EquivalentState) :-
  length(ExampleState,ExampleVarAmount) , 
  VarAmount = ExampleVarAmount , 
  create_equality_nodes_from_example(State,StateEqualityNodes) ,
  create_equality_nodes_from_example(ExampleState,ExampleStateEqualityNodes) ,
  get_distinct_independent_types_constraint_for_enums(DistinctEnumSetsConstraint) , 
  append([DistinctEnumSetsConstraint,StateEqualityNodes,ExampleStateEqualityNodes],EqualityNodes) ,
  conjunct_predicates(EqualityNodes,Equality) , 
  solver_interface:solve_predicate(Equality,_,Solution) , 
  Solution = solution(_) ,  ! ,
  EquivalentState = ExampleState.
contains_equivalent_state(VarAmount,State,[_|T],EquivalentState) :-
  contains_equivalent_state(VarAmount,State,T,EquivalentState).

get_distinct_independent_types_constraint_for_enums(Constraint) :- 
  findall(MachineSetName,b_get_machine_set(MachineSetName),MachineSetNames) , 
  get_distinct_independent_types_constraint_for_enums(MachineSetNames,[],ConstraintList) , 
  conjunct_predicates(ConstraintList,Constraint).

get_distinct_independent_types_constraint_for_enums([],Acc,Acc).
get_distinct_independent_types_constraint_for_enums([MachineSetName|T],Acc,ConstraintList) :- 
  b_get_named_machine_set(MachineSetName,SetValues) , 
  findall(SetValueNode,(member(SetValue,SetValues) , SetValueNode = b(identifier(SetValue),global(MachineSetName),[])),SetValueNodes) , 
  findall(Inequality,(member(Node1,SetValueNodes) , member(Node2,SetValueNodes) , Node1 \= Node2 , Inequality = b(not_equal(Node1,Node2),pred,[])),Inequalities) , 
  get_distinct_independent_types_constraint_for_enums(T,[Inequalities|Acc],ConstraintList).

% seq(_), string_set, integer_set, bool_set,..
% Also needed in location_vars_to_program.
get_single_constant_or_set_operator_id(seq-Type,Out,b(seq(Set),set(seq(NodeType)),[])) :-
    get_single_constant_or_set_operator_id(Type,_,Set) ,
    Set = b(_,TInnerType,_) ,
    TInnerType =.. [_|InnerType] ,
    NodeType =.. InnerType ,
    unique_typed_id("o",set(seq(NodeType)),TempOut) ,
    % we add the internal type of infinite set constants here
    % so we are able to call this predicate when converting location variables to ASTs later on
    add_texpr_infos(TempOut,[synthesis(type,seq-Type)],Out).
get_single_constant_or_set_operator_id(constant-Type,Out,ConstantNode) :-
    get_constant_node_or_id(constant-Type,ConstantNode) ,
    unique_typed_id("o",Type,TempOut) ,
    add_texpr_infos(TempOut,[synthesis(type,Type)],Out).

get_constant_node_or_id(string_set,b(string_set,set(string),[])).
get_constant_node_or_id(bool_set,b(bool_set,set(boolean),[])).
get_constant_node_or_id(integer_set,b(integer_set('INTEGER'),set(integer),[])).
get_constant_node_or_id(integer_set_min_max,b(integer_set('INT'),set(integer),[])).
get_constant_node_or_id(natural1,b(integer_set('NATURAL1'),set(integer),[])).
get_constant_node_or_id(natural,b(integer_set('NATURAL'),set(integer),[])).
get_constant_node_or_id(implementable_natural1,b(integer_set('NAT1'),set(integer),[])).
get_constant_node_or_id(implementable_natural,b(integer_set('NAT'),set(integer),[])).
get_constant_node_or_id(constant-Type,Constant) :-
    unique_typed_id("constant",Type,TempConstant) , 
    add_texpr_infos(TempConstant,[synthesis(enumerated_constant),synthesis(type,Type)],Constant).

add_distinguishing_states_from_cache(action,_,PosInput,NegInput,PosInput,NegInput) :- !.
add_distinguishing_states_from_cache(_,(DistIn,DistOut),PosInput,NegInput,NewPosInput,NewNegInput) :-
    append(DistIn,PosInput,TempPosInput) ,
    append(DistOut,NegInput,TempNegInput) ,
    remove_dups(TempPosInput,NewPosInput) ,
    remove_dups(TempNegInput,NewNegInput).

% If the domain is enumerated from 1 to n we suppose set(couple(integer,_)) is a sequence.
is_seq(b(Node,Type,Info)) :-
    (Type = set(couple(integer,_))  ; Type = seq(_)) ,
    b_interpreter:b_test_boolean_expression_wf(b(equal(b(domain(b(Node,Type,Info)),set(integer),[]),
        b(interval(b(integer(1),integer,[]),
            b(size(b(Node,Type,Info)),integer,[])),set(integer),[])),pred,[]),[],[]).

get_distinguishing_state_or_trans(DistTransitionCache,CurrentVars,Type,ValidInvariantsOrGuards,AST1,
                                  AST2,NewDistTransitionCache,Distinguishing) :-
    % add typing of machine variables to the invariant for consistency with AtelierB and to find a solution referring to
    % all machine variables
    b_get_typed_invariant_from_machine(TypedInvariant) ,
    % add WD-constraints for sequences
    find_typed_identifier_uses(AST1,AST1Nodes) , find_typed_identifier_uses(AST2,AST2Nodes) ,
    findall(Seq,(member(Seq,AST1Nodes) , is_seq(Seq)),Seq1) ,
    findall(Seq,(member(Seq,AST2Nodes) , is_seq(Seq)),Seq2) ,
    append(Seq1,Seq2,TSequences) , remove_dups(TSequences,Sequences) ,
    maplist(sequence_well_definedness_constraint,Sequences,SeqConstraintList) ,
    conjunct_predicates(SeqConstraintList,SeqConstraints) ,
    get_distinct_independent_types_constraint_for_enums(DistinctEnumSetsConstraintList) , 
    conjunct_predicates(DistinctEnumSetsConstraintList,DistinctEnumSetsConstraint) , 
    preferences:set_preference(randomise_enumeration_order,true) , 
    get_distinguishing_state_or_trans_aux(DistinctEnumSetsConstraint,DistTransitionCache,Type,ValidInvariantsOrGuards,
                b(conjunct(SeqConstraints,b(conjunct(b(negation(
                    b(equivalence(AST1,AST2),pred,[])),pred,[]),TypedInvariant),pred,[])),pred,[]),Distinguishing) ,
    expand_transition_cache_if_solution(Type,DistTransitionCache,CurrentVars,Distinguishing,NewDistTransitionCache) ,
    synthesis_log(distinguishing_transition_cache_changed(DistTransitionCache,NewDistTransitionCache)).
% When synthesizing an action only search for a distinguishing transition that doesn't violate the
% so far synthesized guard (in case there is one).
% Furthermore, we exclude already visited distinguishing transitions using the cache in order to
% prevent cycles when a distinguishing transition's output is changed by the user
% (otherwise we would find the same transition over and over again because the one with an adapted
% output is added to the examples).
get_distinguishing_state_or_trans_aux(DistinctEnumSetsConstraint,DistTransitionCache,action,ValidInvariantsOrGuards,Constraint,Distinguishing) :-
    ValidInvariantsOrGuards \= any , ValidInvariantsOrGuards \= none , ! ,
    DistConstraint = b(conjunct(DistinctEnumSetsConstraint,b(conjunct(b(conjunct(Constraint,ValidInvariantsOrGuards),pred,[]),DistTransitionCache),pred,[])),pred,[]) ,
    synthesis_log(distinguishing_constraint(action,DistConstraint)) ,
    solve_distinguishing_constraint_with_fallback(DistConstraint,TDistinguishing) ,
    get_solution_or_none(TDistinguishing,Distinguishing).
get_distinguishing_state_or_trans_aux(DistinctEnumSetsConstraint,DistTransitionCache,action,_,Constraint,Distinguishing) :- ! ,
    DistConstraint = b(conjunct(DistinctEnumSetsConstraint,b(conjunct(Constraint,DistTransitionCache),pred,[])),pred,[]) ,
    synthesis_log(distinguishing_constraint(action,Constraint)) ,
    solve_distinguishing_constraint_with_fallback(DistConstraint,TDistinguishing) ,
    get_solution_or_none(TDistinguishing,Distinguishing).
% DistTransitionCache is empty when synthesizing guard/invariant
get_distinguishing_state_or_trans_aux(DistinctEnumSetsConstraint,_,Type,_,Constraint,Distinguishing) :-
    synthesis_log(distinguishing_constraint(Type,Constraint)) ,
    solve_distinguishing_constraint_with_fallback(b(conjunct(DistinctEnumSetsConstraint,Constraint),pred,[]),TDistinguishing) ,
    get_solution_or_none(TDistinguishing,Distinguishing).

% Solving the distinguishing constraint maybe causes an integer overflow due to infinite domains which
% possibly ends in not finding any solution, i.e., neither true nor false.
% Therefore, we try different solvers if no solution can be found.
% Of course, we do not want to enumerate infinite domains, but there is no reasonable domain restriction
% if for instance the invariant defines a variable to be an element of INTEGER without an upper bound.
% The distinguishing constraint is of the form "not(P1 <=> P2) & I", whereat P1,P2 are two synthesized
% programs and I is the current machine invariant. When synthesizing a substitution with already known
% guards G the constraint results to "not(P1 <=> P2) & I & G".
solve_distinguishing_constraint_with_fallback(DistConstraint,Solution) :-
    solve_predicate(DistConstraint,_,ProBSolution) ,
    % ProB either found a solution or a contradiction
    synthesis_log(solve_distinguishing_constraint(DistConstraint,'ProB')) ,
    ProBSolution \= no_solution_found(_) , ! , Solution = ProBSolution.
solve_distinguishing_constraint_with_fallback(DistConstraint,Z3Solution) :-
    % ProB could not make a decision and answered no_solution_found/1, then we ask z3
    synthesis_log(solve_distinguishing_constraint(DistConstraint,'Z3')) ,
    smt_solve_predicate(z3,DistConstraint,_,Z3Solution).
%    Solution \= no_solution_found(_) , ! , Solution = Z3Solution.
%%%%% Maybe too much overhead to ask a third solver for the distinguishing constraint since two solvers were already not able to find a solution.
%solve_distinguishing_constraint_with_fallback(DistConstraint,ProBZ3Solution) :-
%    synthesis_log(solve_distinguishing_constraint(DistConstraint,'Combined ProB and Z3')) ,
%    preferences:set_preference(smt_supported_interpreter,true) ,
    % Z3 also did not find a solution, then we ask ProB/Z3 combined as a last try
%    solve_predicate(DistConstraint,_,ProBZ3Solution) ,
%    preferences:set_preference(smt_supported_interpreter,false).

get_solution_or_none(solution(Solution),solution(Solution)).
get_solution_or_none(Solution,none) :-
    Solution \= solution(_).

b_get_typed_invariant_from_machine(TypedInvariant) :-
    b_get_typing_predicate(TypingPredicate) ,
    b_get_invariant_from_machine(Invariant) ,
    conjunct_predicates([TypingPredicate,Invariant],TypedInvariant).

% Return typing predicate for the machine variables.
b_get_typing_predicate(TypingPredicate) :-
    b_get_machinevars_with_operation_params(MachineVars) ,
    findall(NonSeq,(member(NonSeq,MachineVars) , NonSeq \= b(_,seq(_),_)),NonSeqMachineVars) ,
    b_get_typing_predicate_for_vars(NonSeqMachineVars,TypingPredicate).

% Return typing predicate for specific machine variables.
b_get_typing_predicate_for_vars(CurrentVars,TypingPredicate) :-
    translate:generate_typing_predicates(CurrentVars,TypingPredicatesList) ,
    conjunct_predicates(TypingPredicatesList,TypingPredicate).

:- spec_pre(expand_transition_cache_if_solution/5,[atom,ast,list(ast),one_of([atom,solution_compound]),var]).
:- spec_invariant(expand_transition_cache_if_solution/5,[atom,ast,list(ast),one_of([atom,solution_compound]),ast]).
:- spec_post(expand_transition_cache_if_solution/5,[atom,ast,list(ast),one_of([atom,solution_compound]),var],
                                                   [atom,ast,list(ast),one_of([atom,solution_compound]),ast]).
expand_transition_cache_if_solution(action,DistTransitionCache,CurrentVars,solution(Solution),NewDistTransitionCache) :- ! ,
    get_input_output_nodes_from_bindings(Solution,Input,Output) ,
    exclude_transition_constraint(Input,Output,CurrentVars,Exclusion) ,
    NewDistTransitionCache = b(conjunct(DistTransitionCache,Exclusion),pred,[]).
expand_transition_cache_if_solution(action,DistTransitionCache,_,_,DistTransitionCache).
expand_transition_cache_if_solution(Type,DistTransitionCache,_,_,DistTransitionCache) :-
    Type \= action.

:- spec_pre(expand_state_cache_if_guard_and_solution/6,[atom,solution_compound,pair(ast_examples,ast_examples),
                                                        ast_examples,ast_examples,var]).
:- spec_invariant(expand_state_cache_if_guard_and_solution/6,[atom,solution_compound,pair(ast_examples,ast_examples),
                                                              ast_examples,ast_examples,pair(ast_examples,ast_examples)]).
:- spec_post(expand_state_cache_if_guard_and_solution/6,[atom,solution_compound,pair(ast_examples,ast_examples),
                                                         ast_examples,ast_examples,var],
                                                        [atom,solution_compound,pair(ast_examples,ast_examples),
                                                         ast_examples,ast_examples,pair(ast_examples,ast_examples)]).
expand_state_cache_if_guard_and_solution(guard,solution(_),(UsedPositives,UsedNegatives),PosInput,
                                         NegInput,(PositiveStates,NegativeStates)) :-
    append(UsedPositives,PosInput,TPositiveStates) ,
    append(UsedNegatives,NegInput,TNegativeStates) ,
    remove_dups(TPositiveStates,PositiveStates) ,
    remove_dups(TNegativeStates,NegativeStates).
expand_state_cache_if_guard_and_solution(_,_,DistStateCache,_,_,DistStateCache).

:- spec_pre(exclude_transition_constraint/4,[list(ast),list(ast),list(ast),var]).
:- spec_invariant(exclude_transition_constraint/4,[list(ast),list(ast),list(ast),ast]).
:- spec_post(exclude_transition_constraint/4,[list(ast),list(ast),list(ast),var],
                                             [list(ast),list(ast),list(ast),ast]).
exclude_transition_constraint(Input,Output,CurrentVars,b(negation(EqualityConjunction),pred,[])) :-
    b_get_machinevars_with_operation_params(MachineVars) ,
    maplist(get_varname_from_id_node,CurrentVars,CurrentIDs) ,
    create_equality_nodes_from_examples(CurrentIDs,Input,Output,MachineVars,EqualityNodes) ,
    conjunct_predicates(EqualityNodes,EqualityConjunction).

sequence_well_definedness_constraint(SeqIDNode,WDConstraint) :-
    % well-defined sequence: domain enumerated from 1..n
    WDConstraint = b(equal(b(domain(SeqIDNode),set(integer),[]),
        b(interval(b(integer(1),integer,[]),b(size(SeqIDNode),integer,[])),set(integer),[])),pred,[]).

is_ast(b(_,_,_)).

delete_numbers_from_atom(NumberAtom,Atom) :-
    atom_codes(NumberAtom,Codes) ,
    delete_numbers_from_codelist(Codes,NewCodes) ,
    atom_codes(Atom,NewCodes).
delete_numbers_from_codelist([],[]).
delete_numbers_from_codelist([Code|R],[Code|NR]) :-
    (Code > 57 ; Code < 48) , ! ,
    delete_numbers_from_codelist(R,NR).
delete_numbers_from_codelist([_|R],NR) :-
    delete_numbers_from_codelist(R,NR).

% Returns a list of all guards or 'any' if none is given.
:- spec_pre(get_guard_list_from_operation/2,[atom,var]).
:- spec_invariant(get_guard_list_from_operation/2,[atom,one_of([ast,atom])]).
:- spec_post(get_guard_list_from_operation/2,[atom,var],[atom,one_of([ast,atom])]).
get_guard_list_from_operation(Operation,ListOfGuards) :-
    b_get_machine_operation(Operation,_,OperationParameters,OpBody) ,
    get_guard_list_from_operation_aux(Operation,OpBody,GuardList) ,
    recreate_guards_for_operation_parameters(OperationParameters,OperationParameterGuards) ,
    append(GuardList,OperationParameterGuards,ListOfGuards).
% No given guards.
get_guard_list_from_operation(_,any).

get_guard_list_from_operation_aux(Operation,OpBody,GuardList) :-
    get_guard_list_from_operation_aux2(Operation,OpBody,Guard) ,
    conjunction_to_list(Guard,GuardList).
get_guard_list_from_operation_aux(_,_,[]).

get_guard_list_from_operation_aux2(_,b(precondition(Guard,_),_,_),Guard).
get_guard_list_from_operation_aux2(Operation,b(rlevent(Operation,_,_,_,Guard,_,_,_,_,_,_),_,_),Guard).
get_guard_list_from_operation_aux2(_,b(any(_,Guard,_),subst,_),Guard).

% ProB removes typing guards for operations parameters which we need for synthesis to create constants for operation parameters
recreate_guards_for_operation_parameters(OperationParameters,OperationParameterGuards) :-
    findall(SetAst,b_get_machine_set(_,SetAst),IndependentTypes) ,
    recreate_guards_for_operation_parameters(OperationParameters,IndependentTypes,[],OperationParameterGuards).
recreate_guards_for_operation_parameters([],_,Acc,Acc).
recreate_guards_for_operation_parameters([OperationParameter|T],IndependentTypes,Acc,OperationParameterGuards) :-
    get_texpr_type(OperationParameter,Type) ,
    member(b(SetName,set(Type),Info),IndependentTypes) ,
    Membership = b(member(OperationParameter,b(SetName,set(Type),Info)),pred,[]) ,
    recreate_guards_for_operation_parameters(T,IndependentTypes,[Membership|Acc],OperationParameterGuards).

:- spec_pre(conjunct_previous_invariant_or_guard_if_necessary/4,[atom,one_of([atom,ast,list(ast)]),ast,var]).
:- spec_invariant(conjunct_previous_invariant_or_guard_if_necessary/4,[atom,one_of([atom,ast,list(ast)]),ast,ast]).
:- spec_post(conjunct_previous_invariant_or_guard_if_necessary/4,[atom,one_of([atom,ast,list(ast)]),ast,var],
                                                                 [atom,one_of([atom,ast,list(ast)]),ast,ast]).
conjunct_previous_invariant_or_guard_if_necessary(invariant,ValidInvariants,ProgramAST,SolutionAST) :- ! ,
    % mark synthesized invariants in the info to be able to split synthesized invariants from those that have already been valid beforehand
    % we want to avoid redundancy when generating typing invariants later on
    conjunction_to_list(ProgramAST,ProgramASTList) ,
    findall(MarkedConjunct,(member(Conjunct,ProgramASTList) , add_texpr_infos(Conjunct,[synthesized_code],MarkedConjunct)),MarkedConjunctList) ,
    conjunct_predicates(MarkedConjunctList,MarkedProgramAst) ,
    append(ValidInvariants,[MarkedProgramAst],InvariantsList) , % ValidInvariants prior to synthesized ones since they may contain typing invariants
    conjunct_predicates(InvariantsList,SolutionAST).
conjunct_previous_invariant_or_guard_if_necessary(guard,Guards,ProgramAST,SolutionAST) :-
    Guards \= any , ! ,
    % search for valid guards that are already given by the machine ( NewPreCond => OldPreCond )
    find_typed_identifier_uses(ProgramAST,[],ProgIDs) ,
    findall(Pre,(member(Pre,Guards) , find_typed_identifier_uses(Pre,PreIDs) ,
                 append(ProgIDs,PreIDs,TIDs) , remove_dups(TIDs,IDs) ,
                 b_test_boolean_expression_wf(b(forall(IDs,ProgramAST,Pre),pred,[used_ids([])]),[],[])),InvalidPreConds) ,
    sets:subtract(Guards,InvalidPreConds,ValidPreConds) ,
    append(ValidPreConds,[ProgramAST],PreCondsList) , % ValidPreConds prior to synthesized ones since they may contain predicates setting the type of operation parameters
    conjunct_predicates(PreCondsList,SolutionAST).
conjunct_previous_invariant_or_guard_if_necessary(action,Guard,ProgramAST,SolutionAST) :-
   \+member(Guard,[any,none]) ,  ! ,
   SolutionAST = b(precondition(Guard,ProgramAST),subst,[]).
conjunct_previous_invariant_or_guard_if_necessary(_,_,ProgramAST,ProgramAST).

% Replace equal by assign_single_id and conjunctions by parallel substitution.
:- spec_pre(replace_preds_by_subst_if_neccessary/3,[atom,ast,var]).
:- spec_invariant(replace_preds_by_subst_if_neccessary/3,[atom,ast,ast]).
:- spec_post(replace_preds_by_subst_if_neccessary/3,[atom,ast,var],[atom,ast,ast]).
replace_preds_by_subst_if_neccessary(action,TSolutionAST,SolutionAST) :-
    conjunction_to_list(TSolutionAST,EqualityList) ,
    maplist(equality_to_assignment,EqualityList,AssignmentList) ,
    SolutionAST = b(parallel(AssignmentList),subst,[]).
replace_preds_by_subst_if_neccessary(_,SolutionAST,SolutionAST).

equality_to_assignment(b(equal(Primed,Assignment),pred,Info),b(assign_single_id(Primed,Assignment),subst,Info)).

% Replace parallel substitution by conjunction predicates and assignments by equalities.
:- spec_pre(parallel_to_conjunction/2,[ast,var]).
:- spec_invariant(parallel_to_conjunction/2,[ast,ast]).
:- spec_post(parallel_to_conjunction/2,[ast,var],[ast,ast]).
parallel_to_conjunction(b(Node,Type,Info),b(Node,Type,Info)) :-
    ground(Node) , 
    Type \= subst.
parallel_to_conjunction(b(Node,subst,_),Conjunction) :-
    ground(Node) , 
    Node =.. [parallel,Assignments] ,
    findall(Equality,(member(Assignment,Assignments) , equality_to_assignment(Equality,Assignment)),EqualityNodes) ,
    conjunct_predicates(EqualityNodes,Conjunction).
%%

% Get the variables involved in an invariant violation if currently all machine variables are considered,
% i.e., they have not been reduced by now.
get_involved_variables(_,PosInput,_,VarNames) :-
    examples_already_reduced(PosInput) , ! ,
    PosInput = [Example|_] ,
    % the examples already have been reduced by the user
    maplist(get_varname_from_node_info,Example,VarNames).
get_involved_variables(Guards,PosInput,NegInput,VarNames) :-
    get_involved_variables_aux(Guards,PosInput,NegInput,TempVarNames) ,
    remove_dups(TempVarNames,VarNames) ,
    % only consider machine variables
    %b_get_machinevars_with_operation_params(MachineVars) ,
    %maplist(get_varname_from_id_node,MachineVars,MachineVarNames) ,
    %sets:intersection(VarNamesNoDups,MachineVarNames,VarNames) ,
    synthesis_log(involved_variables(VarNames)).
% check the invariants first
get_involved_variables_aux(_,_,NegInput,VarNames) :-
    get_violated_and_valid_machine_invariants(NegInput,_,ViolatingVars,_) ,
    ViolatingVars \= [] , ! ,
    findall(Name,member(b(identifier(Name),_,_),ViolatingVars),VarNames).
% if no invariant violating vars, check the guards
get_involved_variables_aux(Guards,PosInput,NegInput,VarNames) :-
    Guards \= [] ,
    append(PosInput,NegInput,Input) ,
    check_states_on_guards(Input,Guards,ViolatedGuards) ,
    findall(UsedIDs,(member(Guard,ViolatedGuards) , find_identifier_uses(Guard,[],UsedIDs)),NestedList) ,
    append(NestedList,UsedIDs) ,
    UsedIDs \= [] ,
    remove_dups(UsedIDs,VarNames).
% otherwise, consider all variables used in the examples
get_involved_variables_aux(_,PosInput,_,VarNames) :-
    PosInput = [Example|_] ,
    maplist(get_varname_from_node_info,Example,VarNames).

check_states_on_guards(State,Guards,ViolatedGuards) :-
    maplist(create_equality_nodes_from_example,State,StateEqualities) ,
    findall(Guard,(member(Guard,Guards) , check_states_on_guard(StateEqualities,Guard)),ValidGuards) ,
    subtract(Guards,ValidGuards,ViolatedGuards).
check_states_on_guard([],_).
check_states_on_guard([StateEquality|T],Guard) :-
    conjunct_predicates([Guard|StateEquality],CheckGuard) ,
    b_test_boolean_expression_wf(CheckGuard,_,[]) , ! ,
    check_states_on_guard(T,Guard).

:- spec_pre(get_violated_and_valid_machine_invariants/4,[list(ast),var,var,var]).
:- spec_invariant(get_violated_and_valid_machine_invariants/4,[list(ast),list(ast),list(ast),list(ast)]).
:- spec_post(get_violated_and_valid_machine_invariants/4,[list(ast),var,var,var],
                                                         [list(ast),list(ast),list(ast),list(ast)]).
get_violated_and_valid_machine_invariants(Input,ViolatedInvariants,ViolatingVars,ValidInvariants) :-
    % according to specific operation
    get_invariants_for_current_vars(Input,_,InvariantList) ,
    % consider the machine properties in case an invariant refers to machine constants
    b_get_properties_from_machine(MachineProperties) ,
    maplist(get_violated_and_valid_machine_invariants_aux(MachineProperties,InvariantList),Input,TempViolatedInvariants) ,
    tools:flatten(TempViolatedInvariants,TempViolatedInvariants2) ,
    remove_dups(TempViolatedInvariants2,ViolatedInvariants) ,
    conjunct_predicates(ViolatedInvariants,TempPred) ,
    find_typed_identifier_uses_with_info(TempPred,TempViolatingVars) ,
    b_get_machinevars_with_operation_params(MachineVars) ,
    % filter only machine variables from TempViolatingVars and do not consider machine constants that may occur within an invariant
    % but do not define the machine state which only refers to machine variables
    findall(MachineVar,(member(MachineVar,TempViolatingVars) , get_varname_from_id_node(MachineVar,MachineVarName) ,
                        member(b(identifier(MachineVarName),_,_),MachineVars)),ViolatingVars) ,
    % all other machine variables are valid considering the current machine variables
    b_get_invariant_from_machine(MachineInvariant) ,
    conjunction_to_list(MachineInvariant,MachineInvariantList) ,
    subtract(MachineInvariantList,ViolatedInvariants,ValidInvariants).

get_violated_and_valid_machine_invariants_aux(MachineProperties,InvariantList,Input,ViolatedInvariants) :-
    create_equality_nodes_from_example(Input,EqualityNodes) ,
    conjunct_predicates([MachineProperties|EqualityNodes],EqualityPredWithMachineProperties) ,
    get_violated_and_valid_machine_invariants_aux2(EqualityPredWithMachineProperties,InvariantList,ViolatedInvariants).

get_violated_and_valid_machine_invariants_aux2(_,[],[]).
get_violated_and_valid_machine_invariants_aux2(EqualityNodes,[SingleInv|T],[SingleInv|NT]) :-
    conjunct_predicates([EqualityNodes,SingleInv],InvPred) ,
    find_identifier_uses(InvPred,[],UsedIDs) ,
    findall(Binding,(member(ID,UsedIDs) , Binding = bind(ID,_)),Bindings) ,
    \+ b_test_boolean_expression_wf(InvPred,Bindings,[]) , ! ,
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,T,NT).
get_violated_and_valid_machine_invariants_aux2(EqualityNodes,[_|T],NT) :-
    get_violated_and_valid_machine_invariants_aux2(EqualityNodes,T,NT).

% Get the machine invariants considering at least one of the current variables derived from the examples.
get_invariants_for_current_vars(Examples,TypedIDs,InvariantList) :-
    Examples = [Example|_] ,
    maplist(get_varname_from_node_info,Example,VarNames) ,
    get_machinevars_by_name(VarNames,TypedIDs) ,
    b_get_invariant_from_machine(MachineInvariant) ,
    conjunction_to_list(MachineInvariant,CompleteInvariantList) ,
    findall(Invariant,(member(Invariant,CompleteInvariantList) , find_identifier_uses(Invariant,[],UsedIDs) ,
                       intersection(UsedIDs,VarNames,Intersection) , Intersection \= []),InvariantList).

% Check if the given examples are already reduced by the user, i.e., they refer to less variables than
% there are machine variables.
examples_already_reduced(Examples) :-
    Examples = [Example|_] ,
    length(Example,VarAmount) ,
    b_get_machinevars_with_operation_params(MachineVars) ,
    length(MachineVars,MachineVarAmount) ,
    VarAmount < MachineVarAmount.

% The examples are tuples of variable name and pretty value from the java ui.
get_invariant_violating_vars_from_examples([],[],[]) :- !.
get_invariant_violating_vars_from_examples(Positive,Negative,ViolatingVarNames) :-
    % split examples that are tuples of input and output state
    split_input_output_tuples(Positive,PosInput,PosOutput) ,
    split_input_output_tuples(Negative,NegInput,NegOutput) ,
    append([PosInput,PosOutput,NegInput,NegOutput],Input) ,
    get_violated_and_valid_machine_invariants(Input,_,ViolatingVars,_) , ! ,
    maplist(get_varname_from_id_node,ViolatingVars,ViolatingVarNames).

% Get another transition that violates the invariant, exclude given input/output example.
:- spec_pre(get_invariant_violating_transition/6,[atom,list(ast),list(ast),list(ast),list(ast),var]).
:- spec_invariant(get_invariant_violating_transition/6,[atom,list(ast),list(ast),list(ast),list(ast),pair(ast,ast)]).
:- spec_post(get_invariant_violating_transition/6,[atom,list(ast),list(ast),list(ast),list(ast),var],
                                                  [atom,list(ast),list(ast),list(ast),list(ast),pair(ast,ast)]).
get_invariant_violating_transition(Operation,UsedVars,ViolatedInvariants,Input,Output,Counter) :-
    b_get_machine_operation(Operation,_,_,TOpBody) ,
    % don't create contradiction in case the guard is equivalent
    % to an invariant which we will negate in the following
    remove_pre_condition_from_body(TOpBody,OpBody) ,
    before_after_predicate_with_equalities(OpBody,UsedVars,BABody) ,
    %find_identifier_uses(OpBody,[],BodyIDs) ,
    findall(V,(member(b(identifier(V),_,_),UsedVars)),BodyIDs) ,
    b_get_machinevars_with_operation_params(MachineVars) ,
    create_equality_nodes_from_examples(BodyIDs,Input,Output,MachineVars,TNodeList) ,
    % negate equalities to exclude this transition
    findall(Inequality,(create_texpr(negation(Equality),pred,[],Inequality) ,
        member(Equality,TNodeList)),NodeList) ,
    conjunct_predicates(ViolatedInvariants,Invariant) ,
    maplist(get_varname_from_id_node,MachineVars,MachineVarNames) ,
    prime_or_unprime_variables_in_ast(prime,MachineVarNames,Invariant,PrimedInvariant) ,
    create_texpr(negation(Invariant),pred,[],NInvariant) ,
    create_texpr(negation(PrimedInvariant),pred,[],NPrimedInvariant) ,
    conjunct_predicates([BABody,NInvariant,NPrimedInvariant|NodeList],BAExclusion) ,
    solve_predicate(BAExclusion,_,Solution) ,
    get_invariant_violating_transition_aux(Solution,Counter).

get_invariant_violating_transition_aux(Solution,none) :-
    Solution \= solution(_).
get_invariant_violating_transition_aux(solution(Solution),(CounterInput,CounterOutput)) :-
    % split by input and output state
    get_input_output_nodes_from_bindings(Solution,CounterInput,CounterOutput).

get_invariant_violating_example(_,[],[],_) :- fail.
get_invariant_violating_example(ViolatedInvariants,[PosInput|TIn],[PosOutput|TOut],Example) :-
  maplist(check_invariant_on_state(PosInput,PosOutput),ViolatedInvariants) , ! ,
  % transition doesn't violate an invariant
  get_invariant_violating_example(ViolatedInvariants,TIn,TOut,Example).
% invariant violating transition found
get_invariant_violating_example(_,[PosInput|_],[PosOutput|_],(PosInput,PosOutput)).

% Is true if the given invariant is true for the pair of input/output states.
check_invariant_on_state(Input,Output,InvariantAst) :-
  create_equality_nodes_from_examples(Input,Output,EqualityNodes) ,
  conjunct_predicates([InvariantAst|EqualityNodes],InvariantTransitionPred) ,
  b_test_boolean_expression_wf(InvariantTransitionPred,[],_).

%% Get the input and/or output nodes from a solution of the distinguishing constraint.
:- spec_pre(get_input_output_nodes_from_bindings/3,[solution_bindings,var,var]).
:- spec_invariant(get_input_output_nodes_from_bindings/3,[solution_bindings,list(ast),list(ast)]).
:- spec_post(get_input_output_nodes_from_bindings/3,[solution_bindings,var,var],[solution_bindings,list(ast),list(ast)]).
get_input_output_nodes_from_bindings(Solution,Input,Output) :-
    get_input_nodes_from_bindings(Solution,Input) ,
    get_output_nodes_from_bindings(Solution,Output).

:- spec_pre(get_input_nodes_from_bindings/2,[solution_bindings,var]).
:- spec_invariant(get_input_nodes_from_bindings/2,[solution_bindings,list(ast)]).
:- spec_post(get_input_nodes_from_bindings/2,[solution_bindings,var],[solution_bindings,list(ast)]).
get_input_nodes_from_bindings(Solution,Input) :-
    findall(Binding,(member(Binding,Solution) , get_varname_from_binding(Binding,VarName) ,
        \+is_primed_id(VarName)),Bindings) ,
    example_bindings_to_ast_nodes(Bindings,Input).

:- spec_pre(get_output_nodes_from_bindings/2,[solution_bindings,var]).
:- spec_invariant(get_output_nodes_from_bindings/2,[solution_bindings,list(ast)]).
:- spec_post(get_output_nodes_from_bindings/2,[solution_bindings,var],[solution_bindings,list(ast)]).
get_output_nodes_from_bindings(Solution,Output) :-
    get_output_nodes_from_bindings_aux(Solution,NewSolution) ,
    example_bindings_to_ast_nodes(NewSolution,Output).
get_output_nodes_from_bindings_aux([],[]).
get_output_nodes_from_bindings_aux([Binding|T],[Binding|NT]) :-
    get_varname_from_binding(Binding,VarName) ,
    is_primed_id(VarName) , ! ,
    get_output_nodes_from_bindings_aux(T,NT).
get_output_nodes_from_bindings_aux([_|T],NT) :-
    get_output_nodes_from_bindings_aux(T,NT).

get_varname_from_binding(binding(VarName,_,_),VarName).
get_varname_from_binding(bind(VarName,_),VarName).

% rlevent/11
remove_pre_condition_from_body(b(Node,_,_),Body) :-
    Node =.. [_,_,_,_,_,_,_,Actions|_] , ! ,
    create_texpr(parallel(Actions),subst,[],OpBody) ,
    remove_pre_condition_from_body(OpBody,Body).
remove_pre_condition_from_body(b(any(_,_,TActions),subst,_),Body) :-
    (is_list(TActions) -> Actions = TActions ; Actions = [TActions]) , ! ,
    create_texpr(parallel(Actions),subst,[],OpBody) ,
    remove_pre_condition_from_body(OpBody,Body).
remove_pre_condition_from_body(PreBody,Body) :-
    PreBody = b(precondition(_,Body),_,_) , !.
remove_pre_condition_from_body(Body,Body).

% prime/unprime all identifiers in VarsToConsider within an ast, if VarsToConsider is an empty list prime/unprime all variables
:- spec_pre(prime_or_unprime_variables_in_ast/4,[atom,list(ast),ast,var]).
:- spec_invariant(prime_or_unprime_variables_in_ast/4,[atom,list(ast),ast,ast]).
:- spec_post(prime_or_unprime_variables_in_ast/4,[atom,list(ast),ast,var],[atom,list(ast),ast,ast]).
prime_or_unprime_variables_in_ast(prime,VarsToConsider,b(identifier(ID),Type,Info),PrimedID) :-
    (member(ID,VarsToConsider) ; VarsToConsider = []) , ! ,
    prime_variable(b(identifier(ID),Type,Info),PrimedID).
prime_or_unprime_variables_in_ast(prime,VarsToConsider,b(identifier(ID),Type,Info),Ast) :- ! ,
    \+member(ID,VarsToConsider) , Ast = b(identifier(ID),Type,Info).
prime_or_unprime_variables_in_ast(unprime,VarsToConsider,b(identifier(ID),Type,Info),UnprimedID) :-
    (member(ID,VarsToConsider) ; VarsToConsider = []) , ! ,
    (is_primed_id(ID)
    ->  remove_prime(b(identifier(ID),Type,Info),UnprimedID)
    ;   UnprimedID = b(identifier(ID),Type,Info)).
prime_or_unprime_variables_in_ast(unprime,VarsToConsider,b(identifier(ID),Type,Info),Ast) :- ! ,
    \+member(ID,VarsToConsider) , Ast = b(identifier(ID),Type,Info).
prime_or_unprime_variables_in_ast(_,_,b(Node,Type,Info),Ast) :-
    (Node = integer(_) ; Node = value(_) ; Node = integer_set(_) ; Node = string(_) ; Node =.. [_]) , ! ,
    Ast = b(Node,Type,Info).
prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [parallel,Assignments] ,
    maplist(prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider),Assignments,NewAssignments) ,
    NewNode =.. [parallel,NewAssignments].
% skip nodes like comprehension_set, set_extension or sequence_extension
prime_or_unprime_variables_in_ast(_,_,b(Node,Type,Info),b(Node,Type,Info)) :-
    Node =.. [OpType|_] ,
    (OpType = comprehension_set ; OpType = set_extension ; OpType = sequence_extension) , !.
prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Op,In] ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,In,NewIn) ,
    NewNode =.. [Op,NewIn].
prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,b(exists(Ids,Pred),Type,Info),b(exists(NewIds,NewPred),Type,Info)) :-
    maplist(prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider),Ids,NewIds) ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,Pred,NewPred).
prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    Node =.. [Op,In1,In2] ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,In1,NewIn1) ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,In2,NewIn2) ,
    NewNode =.. [Op,NewIn1,NewIn2].
prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,b(forall(Ids,Left,Right),Type,Info),b(forall(NewIds,NewLeft,NewRight),Type,Info)) :-
    maplist(prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider),Ids,NewIds) ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,Left,NewLeft) ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,Right,NewRight).
prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,b(Node,Type,Info),b(NewNode,Type,Info)) :-
    % e.g. if_then_else
    Node =.. [Op,In1,In2,In3] ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,In1,NewIn1) ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,In2,NewIn2) ,
    prime_or_unprime_variables_in_ast(PrimeType,VarsToConsider,In3,NewIn3) ,
    NewNode =.. [Op,NewIn1,NewIn2,NewIn3].

% Check that all negative inputs are invalid for the operation's precondition.
check_negative_inputs_against_operation_precondition(_,[]). % no negative input states are present
check_negative_inputs_against_operation_precondition(Operation,NegInput) :-
    NegInput \= [] ,
    b_get_machine_operation(Operation,_,_,OpBody) ,
    OpBody = b(precondition(PreCondition,_),_,_) ,
    check_negative_inputs_against_operation_precondition_aux(PreCondition,NegInput).

% fails if any negative input state is valid on the current precondition, that means, the precondition needs to be strengthened
check_negative_inputs_against_operation_precondition_aux(_,[]).
check_negative_inputs_against_operation_precondition_aux(PreCondition,[NegState|T]) :-
    create_equality_nodes_from_example(NegState,EqualityNodes) ,
    conjunct_predicates([PreCondition|EqualityNodes],EqualityPred) ,
    \+b_test_boolean_expression_wf(EqualityPred,_,_) ,
    check_negative_inputs_against_operation_precondition_aux(PreCondition,T).

%% User adds a missing state transition.
% Search for a machine operation that completely supports the missing state
% or one where we just have to synthesize a relaxed guard.
% VarNames refers to the variables used in the added states.
check_transitions_against_machine_operations(VarNodes,InputExamples,OutputExamples,Result) :-
    % check each example againt the existing machine operations
    maplist(check_transition_against_machine_operations(VarNodes),InputExamples,OutputExamples,SingleResults) ,
    check_transitions_against_machine_operations_aux(SingleResults,Result) , !.
check_transitions_against_machine_operations_aux(SingleResults,Result) :-
    % an operation satisifies the transition but its guard needs to be relaxed
    member((OpName,guard),SingleResults) ,
    % and all other example transitions are also satisfied by this operation,
    maplist(operation_match(OpName),SingleResults) ,
    % therefore we synthesize a relaxed guard
    Result = (OpName,guard).
check_transitions_against_machine_operations_aux(SingleResults,Result) :-
    % an operation satisifies all transitions, and thus, we don't have to synthesize anything
    member((OpName,done),SingleResults) ,
    maplist('='((OpName,done)),SingleResults) ,
    Result = (OpName,done).
check_transitions_against_machine_operations_aux(_,no_match).

operation_match(OpName,(OpName,_)).

:- spec_pre(check_transition_against_machine_operations/4,[list(ast),list(ast),list(ast),var]).
:- spec_invariant(check_transition_against_machine_operations/4,[list(ast),list(ast),list(ast),
                                                                 one_of([atom,pair(atom,atom)])]).
:- spec_post(check_transition_against_machine_operations/4,[list(ast),list(ast),list(ast),var],
                                                           [list(ast),list(ast),list(ast),
                                                            one_of([atom,pair(atom,atom)])]).
% Check a single transition against the existing machine operations.
check_transition_against_machine_operations(VarNodes,Input,Output,Result) :-
    findall((Op,Res,Par,Body),b_get_machine_operation(Op,Res,Par,Body),Operations) ,
    check_transition_against_machine_operations(Operations,Input,Output,VarNodes,Result).
check_transition_against_machine_operations([],_,_,_,no_match).
check_transition_against_machine_operations([(_,_,_,b(skip,subst,_))|T],Input,Output,VarNodes,Result) :-
    check_transition_against_machine_operations(T,Input,Output,VarNodes,Result).
% Skip if the desired action's and the given ones' variables doesn't match.
check_transition_against_machine_operations([(_,_,_,OpBody)|T],Input,Output,VarNodes,Result) :-
    find_identifier_uses(OpBody,[],BodyIDs) ,
    findall(Name,member(b(identifier(Name),_,_),VarNodes),VarNames) ,
    % find distinguishing variables
    findall(DistName,(member(DistName,BodyIDs) , \+member(DistName,VarNames)),DistNames) ,
    DistNames \= [] , ! ,
    check_transition_against_machine_operations(T,Input,Output,VarNodes,Result).
check_transition_against_machine_operations([(OpName,_,_,OpBody)|T],Input,Output,VarNodes,Result) :-
    b_get_machinevars_with_operation_params(MachineVars) ,
    before_after_predicate_with_equalities(OpBody,MachineVars,BABody) ,
    find_identifier_uses(OpBody,[],BodyIDs) ,
    create_equality_nodes_from_examples(BodyIDs,Input,Output,VarNodes,NodeList) ,
    b_get_typed_invariant_from_machine(Invariant) ,
    conjunct_predicates([BABody,Invariant|NodeList],BABodyPred) ,
    check_transition_against_machine_operations_aux(BABodyPred,NodeList,[(OpName,_,_,OpBody)|T],Input,Output,VarNodes,Result).

% Found operation that already executes the transition.
check_transition_against_machine_operations_aux(BABodyPred,_,[(OpName,_,_,_)|_],_,_,_,(OpName,done)) :-
    solve_predicate(BABodyPred,_,solution(_)) , !.
% Only the guard of an operation does not fit.
check_transition_against_machine_operations_aux(_,NodeList,[(OpName,_,_,OpBody)|_],_,_,_,(OpName,guard)) :-
    b_get_machinevars_with_operation_params(MachineVars) ,
    % remove guard and try if action itself supports the missing state
    OpBody = b(precondition(_PreCondition,Action),_,_) ,
    before_after_predicate_with_equalities(Action,MachineVars,BAAction) ,
    conjunct_predicates([BAAction|NodeList],BAActionPred) ,
    solve_predicate(BAActionPred,_,solution(_)) , !.
% The action also does not match so that we search for another operation.
check_transition_against_machine_operations_aux(_,_,[_|T],Input,Output,VarNodes,Result) :-
    check_transition_against_machine_operations(T,Input,Output,VarNodes,Result).
%%

% Get valid transitions for an operation and return tuples holding pretty printed equality predicates of before/after states.
% Furthermore, get invalid states if a precondition is given and return pretty printed equality predicates for this states.
% We can directly use these pretty printed predicates to compute the states in ProB2 using FindStateCommand.
get_valid_and_invalid_equality_predicates_for_operation(OperationName,ValidAmount,InvalidAmount,ValidPrettyEqualityTuples,InvalidPrettyEqualities,IgnoredIDs) :-
    preferences:set_preference(randomise_enumeration_order,true) ,
    b_get_machine_operation(OperationName,_,_,Substitution) ,
    find_identifier_uses(Substitution,[],TempUsedIDs) ,
    % filter only machine variables, an operation may access machine constants that we need to discard here
    b_get_machinevars_with_operation_params(MachineVars) ,
    findall(UsedID,(member(UsedID,TempUsedIDs) , member(b(identifier(UsedID),_,_),MachineVars)),UsedIDs) ,
    findall(IgnoredID,(member(b(identifier(IgnoredID),_,_),MachineVars) , \+member(IgnoredID,UsedIDs)),IgnoredIDs) ,
    get_machinevars_by_name(UsedIDs,UsedMachineVars) ,
    get_valid_transitions_for_operation(UsedMachineVars,ValidAmount,OperationName,BeforeStates,AfterStates) ,
    maplist(create_equality_predicate_from_example,BeforeStates,BeforeEqualities) ,
    maplist(create_equality_predicate_from_example,AfterStates,AfterEqualities) ,
    maplist(translate_bexpression,BeforeEqualities,PrettyBeforeEqualities) ,
    maplist(translate_bexpression,AfterEqualities,PrettyAfterEqualities) ,
    zip(PrettyBeforeEqualities,PrettyAfterEqualities,ValidPrettyEqualityTuples) ,
    get_invalid_state_equalities_for_operation(OperationName,UsedMachineVars,InvalidAmount,Substitution,InvalidEqualities) ,
    maplist(translate_bexpression,InvalidEqualities,InvalidPrettyEqualities) , !.

get_invalid_state_equalities_for_operation(OperationName,MachineVars,Amount,b(precondition(Precondition,_),subst,_),InvalidEqualities) :- ! ,
    % additionally consider the machine properties since a precondition may access machine constants
    b_get_properties_from_machine(MachineProperties) ,
    % filter predicates that set the types of operation parameters from the actual preconditions, since we do not want to negate the typing of operation parameters in the following
    b_get_machine_operation(OperationName,_,OperationParams,_) ,
    split_typing_predicates_from_precondition(OperationParams,Precondition,TypingPredicates,PreconditionPredicate) ,
    conjunct_predicates([MachineProperties,TypingPredicates,b(negation(PreconditionPredicate),pred,[])],NegatedPrecondition) ,
    get_invalid_state_equalities_for_operation_aux(MachineVars,Amount,NegatedPrecondition,[],InvalidStates) ,
    maplist(create_equality_predicate_from_example,InvalidStates,InvalidEqualities).
% no precondition is given
get_invalid_state_equalities_for_operation(_,_,_,_,[]).

get_invalid_state_equalities_for_operation_aux(_,0,_,Acc,Acc).
get_invalid_state_equalities_for_operation_aux(MachineVars,Amount,PreconditionWithProperties,Acc,InvalidStates) :-
    exclude_accumulated_states(MachineVars,Acc,[],TExclusion) ,
    (TExclusion = [] -> Exclusion = b(truth,pred,[]) ; Exclusion = TExclusion) ,
    create_texpr(conjunct(Exclusion,PreconditionWithProperties),pred,[],Predicate) ,
    solve_predicate(Predicate,_,Solution) ,
    Solution = solution(SolutionBindings) , ! ,
    % solutions may also refer to machine constants that we do not want to consider here since we search for machine states
    findall(Binding,(member(Binding,SolutionBindings) , Binding = binding(Name,_,_) , member(b(identifier(Name),_,_),MachineVars)),MachineVarBindings) ,
    example_bindings_to_ast_nodes(MachineVarBindings,InvalidState) ,
    NAmount is Amount - 1 ,
    get_invalid_state_equalities_for_operation_aux(MachineVars,NAmount,PreconditionWithProperties,[InvalidState|Acc],InvalidStates).
get_invalid_state_equalities_for_operation_aux(_,_,_,Acc,Acc).

get_valid_transitions_for_operation(CurrentVars,Amount,OperationName,BeforeStates,AfterStates) :-
    Amount \= 0 ,
    get_valid_inputs_for_operation(OperationName,Amount,ValidInputStates) ,
    maplist(create_equality_nodes_from_example,ValidInputStates,ValidInputStateEqualities) ,
    before_after_predicates:before_after_predicate_for_operation(OperationName,BeforeAfterPredicate) ,
    get_valid_transitions_for_operation_aux(CurrentVars,ValidInputStateEqualities,BeforeAfterPredicate,[],[],BeforeStates,AfterStates).
get_valid_transitions_for_operation(_,_,_,[],[]).

get_valid_transitions_for_operation_aux(_,[],_,BeforeAcc,AfterAcc,BeforeAcc,AfterAcc).
get_valid_transitions_for_operation_aux(CurrentVars,[ValidInputStateEquality|T],BeforeAfterPredicate,BeforeAcc,AfterAcc,BeforeStates,AfterStates) :-
    conjunct_predicates(ValidInputStateEquality,InputEquality) ,
    solve_predicate(b(conjunct(InputEquality,BeforeAfterPredicate),pred,[]),_,Solution) ,
    Solution = solution(SolutionBindings) , ! ,
    % solutions may also refer to machine constants that we do not want to consider here since we search for machine states
    findall(Primed,(member(Primed,SolutionBindings) , Primed = binding(Name,_,_) ,
                    bmachine_eventb:is_primed_id(Name) , remove_prime(b(identifier(Name),_,_),UnprimedAst) ,
                    member(UnprimedAst,CurrentVars)),AfterBindings) ,
    sets:subtract(SolutionBindings,AfterBindings,TempBeforeBindings) ,
    findall(Unprimed,(member(Unprimed,TempBeforeBindings) , Unprimed = binding(Name,_,_) ,
                      \+bmachine_eventb:is_primed_id(Name) , member(b(identifier(Name),_,_),CurrentVars)),BeforeBindings) ,
    example_bindings_to_ast_nodes(BeforeBindings,Before) ,
    example_bindings_to_ast_nodes(AfterBindings,After) ,
    get_valid_transitions_for_operation_aux(CurrentVars,T,BeforeAfterPredicate,[Before|BeforeAcc],[After|AfterAcc],BeforeStates,AfterStates).
get_valid_transitions_for_operation_aux(CurrentVars,[_|T],BeforeAfterPredicate,BeforeAcc,AfterAcc,BeforeStates,AfterStates) :-
    get_valid_transitions_for_operation_aux(CurrentVars,T,BeforeAfterPredicate,BeforeAcc,AfterAcc,BeforeStates,AfterStates).

exclude_accumulated_states(_,[],Acc,Exclusion) :-
    conjunct_predicates(Acc,Exclusion).
exclude_accumulated_states(CurrentVars,[BeforeState|T],Acc,Exclusion) :-
    maplist(exclude_variable_state(CurrentVars),BeforeState,ExclusionList) ,
    conjunct_predicates(ExclusionList,Inequalities) ,
    exclude_accumulated_states(CurrentVars,T,[Inequalities|Acc],Exclusion).

exclude_variable_state(MachineVars,b(Node,Type,Info),Inequality) :-
    member(synthesis(machinevar,VarName),Info) ,
    member(b(identifier(VarName),MachineVarType,MachineVarInfo),MachineVars) ,
    create_texpr(not_equal(b(Node,Type,Info),b(identifier(VarName),MachineVarType,MachineVarInfo)),pred,[],Inequality).

% Filter predicates that set the types of operation parameters from the actual preconditions.
split_typing_predicates_from_precondition(OperationParams,Precondition,TypingPredicate,PreconditionPredicate) :-
    maplist(get_texpr_id,OperationParams,OperationParamNames) ,
    conjunction_to_list(Precondition,PreconditionList) ,
    findall(TypingPred,(member(TypingPred,PreconditionList) , find_identifier_uses(TypingPred,[],UsedIDs) ,
            sets:intersection(UsedIDs,OperationParamNames,Intersection) , Intersection \= []),TypingPredicatesList) ,
    sets:subtract(PreconditionList,TypingPredicatesList,ActualPreconditionList) ,
    conjunct_predicates(TypingPredicatesList,TypingPredicate) ,
    conjunct_predicates(ActualPreconditionList,PreconditionPredicate).

% Can be used in both directions: Either VarName or Value has to be ground.
:- spec_pre(get_binding_for_name_or_value/3,[solution_bindings,one_of([var,atom]),one_of([var,atom,compound])]).
:- spec_invariant(get_binding_for_name_or_value/3,[solution_bindings,one_of([var,atom]),one_of([var,atom,compound])]).
:- spec_post(get_binding_for_name_or_value/3,[solution_bindings,one_of([var,atom]),one_of([var,atom,compound])],
                                             [solution_bindings,one_of([var,atom]),one_of([var,atom,compound])]).
get_binding_for_name_or_value(SolutionBindings,VarName,Value) :-
    member(binding(VarName,Value,_),SolutionBindings) ;
    member(bind(VarName,Value),SolutionBindings).

% search identifier nodes within an ast and keep each nodes information
find_typed_identifier_uses_with_info(AST,UsedIDs) :-
    find_typed_identifier_uses_with_info_aux(AST,TempUsedIDs) ,
    remove_dups(TempUsedIDs,UsedIDs) , !.

find_typed_identifier_uses_with_info_aux(List,IDs) :-
    is_list(List) ,
    findall(ID,(member(ID,List) , ID = b(identifier(_),_,_)),IDs).
find_typed_identifier_uses_with_info_aux(b(identifier(IDName),Type,Info),[b(identifier(IDName),Type,Info)]).
find_typed_identifier_uses_with_info_aux(b(Node,_,_),UsedIDs) :- 
  Node =.. [Name,List] , 
  (Name = set_extension ; Name = parallel) , 
  maplist(find_typed_identifier_uses_with_info_aux,List,UsedIDsNested) , 
  tools:flatten(UsedIDsNested,UsedIDs).
find_typed_identifier_uses_with_info_aux(b(assign_single_id(_ID,Ast),_,_),UsedIDs) :- 
  find_typed_identifier_uses_with_info_aux(Ast,UsedIDs).
find_typed_identifier_uses_with_info_aux(b(Node,_,_),UsedIDs) :-
    Node =.. [if_then_else,Condition,TrueOut,FalseOut] ,
    find_typed_identifier_uses_with_info_aux(Condition,ConditionIDs) ,
    find_typed_identifier_uses_with_info_aux(TrueOut,TrueOutIDs) ,
    find_typed_identifier_uses_with_info_aux(FalseOut,FalseOutIDs) ,
    append([ConditionIDs,TrueOutIDs,FalseOutIDs],UsedIDs).
find_typed_identifier_uses_with_info_aux(b(Node,_,_),UsedIDs) :-
    Node =.. [forall,_,Left,Right] ,
    find_typed_identifier_uses_with_info_aux(Left,LeftIDs) ,
    find_typed_identifier_uses_with_info_aux(Right,RightIDs) ,
    append([LeftIDs,RightIDs],UsedIDs).
find_typed_identifier_uses_with_info_aux(b(Node,_,_),UsedIDs) :-
    Node =.. [_,Arg] ,
    find_typed_identifier_uses_with_info_aux(Arg,UsedIDs).
find_typed_identifier_uses_with_info_aux(b(Node,_,_),UsedIDs) :-
    Node =.. [_,Arg1,Arg2] ,
    find_typed_identifier_uses_with_info_aux(Arg1,UsedIDs1) ,
    find_typed_identifier_uses_with_info_aux(Arg2,UsedIDs2) ,
    append(UsedIDs1,UsedIDs2,UsedIDs).
% rlevent
find_typed_identifier_uses_with_info_aux(b(Node,_,_),UsedIDs) :-
    Node =.. [_,_,_,_,_,Body|_] ,
    find_typed_identifier_uses_with_info_aux(Body,UsedIDs).
find_typed_identifier_uses_with_info_aux(_,[]).

% Get current identifier names from an example before creating equality nodes to the example.
:- spec_pre(create_equality_nodes_from_examples/3,[ast_example,ast_example,var]).
:- spec_invariant(create_equality_nodes_from_examples/3,[ast_example,ast_example,ast_example]).
:- spec_post(create_equality_nodes_from_examples/3,[ast_example,ast_example,var],
                                                   [ast_example,ast_example,ast_example]).
create_equality_nodes_from_examples(Input,Output,EqualityNodes) :-
    b_get_machinevars_with_operation_params(MachineVars) ,
    maplist(get_varname_from_node_info,Input,CurrentIDs) ,
    create_equality_nodes_from_examples(CurrentIDs,Input,Output,MachineVars,EqualityNodes).
% We need to set variables to values from example by creating equality nodes
% that we can add to the before after predicate.
create_equality_nodes_from_examples([],_,_,_,[]).
create_equality_nodes_from_examples([VarName|T],Input,Output,VarNodes,[Node1,Node2|NodeList]) :-
    member(VarNode,VarNodes) ,
    VarNode = b(identifier(VarName),_,_) ,
    member(InputNode,Input) , get_varname_from_node_info(InputNode,VarName) ,
    member(OutputNode,Output) , get_varname_from_node_info(OutputNode,VarName) ,
    Node1 = b(equal(VarNode,InputNode),pred,[]) ,
    prime_variable(VarNode,PrimedVarNode) ,
    Node2 = b(equal(PrimedVarNode,OutputNode),pred,[]) ,
    create_equality_nodes_from_examples(T,Input,Output,VarNodes,NodeList).

create_equality_nodes_from_example(ExampleState,EqualityNodes) :-
  create_equality_nodes_from_example(ExampleState,[],EqualityNodes).
create_equality_nodes_from_example([],Acc,Acc).
create_equality_nodes_from_example([ValueAst|T],Acc,EqualityNodes) :-
    get_varname_from_node_info(ValueAst,VarName) ,
    get_machinevars_by_name([VarName],[MachineVar]) ,
    Equality = b(equal(MachineVar,ValueAst),pred,[]) ,
    create_equality_nodes_from_example(T,[Equality|Acc],EqualityNodes).

create_equality_predicate_from_example(ExampleState,EqualityPredicate) :-
    create_equality_nodes_from_example(ExampleState,EqualityNodes) ,
    conjunct_predicates(EqualityNodes,EqualityPredicate).

get_location_var_node_by_info_term(LocationVars,SynthesisTerm,LocationVar) :-
    member(LocationVar,LocationVars) , LocationVar = b(identifier(_),integer,Info) ,
    member(SynthesisTerm,Info).

get_var_node_by_info_term(LocationVars,SynthesisTerm,LocationVar) :-
    member(LocationVar,LocationVars) , get_texpr_info(LocationVar,Info) ,
    member(SynthesisTerm,Info).

get_location_var_node_by_name(LocationVarName,LocationVars,LocationVarNode) :-
    member(LocationVarNode,LocationVars) ,
    LocationVarNode = b(identifier(LocationVarName),integer,_).

% Compute some valid input states for a given operation by considering the invariant as well as the operation's precondition.
get_valid_inputs_for_operation(OperationName,C,ValidStates) :-
    preferences:set_preference(randomise_enumeration_order,true) ,
    b_get_machine_operation(OperationName,_,_,Substitution) ,
    get_precondition_from_operation_body(Substitution,Precondition) ,
    b_get_typed_invariant_from_machine(Invariant) ,
    % consider the machine properties since the guard may access machine constants
    b_get_properties_from_machine(MachineProperties) ,
    conjunct_predicates([Invariant,Precondition,MachineProperties],InvariantWithPrecondition) ,
    solve_predicate(InvariantWithPrecondition,_,solution(Solution)) ,
    find_identifier_uses(Substitution,[],TempUsedIDs) ,
    % filter only machine variables, an operation may access machine constants that we need to discard here
    b_get_machinevars_with_operation_params(MachineVars) ,
    findall(UsedID,(member(UsedID,TempUsedIDs) , member(b(identifier(UsedID),_,_),MachineVars)),UsedIDs) ,
    get_machinevars_by_name(UsedIDs,UsedMachineVars) ,
    get_states_aux(UsedMachineVars,C,Solution,InvariantWithPrecondition,ValidStates).

get_precondition_from_operation_body(b(select([b(select_when(Condition,_),_,_)]),_,_),Condition).
get_precondition_from_operation_body(b(precondition(Precondition,_),_,_),Precondition).
get_precondition_from_operation_body(_,b(truth,pred,[])).

get_valid_and_invalid_equality_predicates_for_invariants(ValidAmount,InvalidAmount,ValidPrettyEqualities,InvalidPrettyEqualities) :-
    b_get_machinevars_with_operation_params(MachineVars) ,
    maplist(get_texpr_id,MachineVars,MachineVarNames) ,
    get_valid_states(MachineVarNames,ValidAmount,ValidStates) ,
    maplist(create_equality_predicate_from_example,ValidStates,ValidEqualities) ,
    maplist(translate_bexpression,ValidEqualities,ValidPrettyEqualities) ,
    get_invalid_states(MachineVarNames,InvalidAmount,InvalidStates) ,
    maplist(create_equality_predicate_from_example,InvalidStates,InvalidEqualities) ,
    maplist(translate_bexpression,InvalidEqualities,InvalidPrettyEqualities).

% Get invalid states for visualizing the invariants.
:- spec_pre(get_invalid_states/3,[list(atom),int,var]).
:- spec_invariant(get_invalid_states/3,[list(atom),int,list(ast)]).
:- spec_post(get_invalid_states/3,[list(atom),int,var],[list(atom),int,list(ast)]).
get_invalid_states(CurrentVarNames,C,ValidStates) :-
    preferences:set_preference(randomise_enumeration_order,true) ,
    b_get_invariant_from_machine(Invariant) ,
    Invariant \= b(truth,pred,_) , % only typing invariants are given, we then do not provide any invalid states
    NegInvariant = b(negation(Invariant),pred,[]) ,
    solve_predicate(NegInvariant,_,solution(Solution)) ,
    get_machinevars_by_name(CurrentVarNames,CurrentVarNodes) ,
    get_states_aux(CurrentVarNodes,C,Solution,NegInvariant,ValidStates).
get_invalid_states(_,_,[]).

% Get valid states for synthesis of invariants.
:- spec_pre(get_valid_states/3,[list(atom),int,var]).
:- spec_invariant(get_valid_states/3,[list(atom),int,list(ast)]).
:- spec_post(get_valid_states/3,[list(atom),int,var],[list(atom),int,list(ast)]).
get_valid_states(CurrentVarNames,C,ValidStates) :-
    preferences:set_preference(randomise_enumeration_order,true) ,
    b_get_typed_invariant_from_machine(Invariant) ,
    solve_predicate(Invariant,_,solution(Solution)) ,
    get_machinevars_by_name(CurrentVarNames,CurrentVarNodes) ,
    get_states_aux(CurrentVarNodes,C,Solution,Invariant,ValidStates).

:- spec_pre(get_states_aux/5,[list(atom),int,solution_bindings,ast,var]).
:- spec_invariant(get_states_aux/5,[list(atom),int,solution_bindings,ast,list(ast)]).
:- spec_post(get_states_aux/5,[list(atom),int,solution_bindings,ast,var],[list(atom),int,solution_bindings,ast,list(ast)]).
get_states_aux(CurrentVarNodes,C,Solution,Invariant,[ValidState|NT]) :-
    C \= 1 , C1 is C - 1 ,
    findall(Exclusion,(member(Var,CurrentVarNodes) , Var = b(identifier(VarName),Type,_) ,
                       member(binding(VarName,Value,_),Solution) ,
                       create_texpr(not_equal(Var,b(value(Value),Type,[])),pred,[],Exclusion)),ExclusionList) ,
    disjunct_predicates(ExclusionList,Exclusion) , % use a disjunction since it is sufficient if any value is different to the other states
    create_texpr(conjunct(Invariant,Exclusion),pred,[],NewPred) ,
    solve_predicate(NewPred,_,solution(NewSolution)) , ! ,
    example_bindings_to_ast_nodes(Solution,TempValidState) ,
    reduce_state_to_current_vars(CurrentVarNodes,TempValidState,ValidState) ,
    get_states_aux(CurrentVarNodes,C1,NewSolution,NewPred,NT).
get_states_aux(CurrentVarNodes,_,Solution,_,[ValidState]) :-
    is_list(Solution) ,
    example_bindings_to_ast_nodes(Solution,TempValidState) ,
    reduce_state_to_current_vars(CurrentVarNodes,TempValidState,ValidState).
get_states_aux(_,_,_,_,[]).

% A state is a list of ast value nodes each having the term synthesis/2 with the corresponding machine var name in its info.
reduce_state_to_current_vars(CurrentVarNodes,State,ReducedState) :-
    findall(ValueNode,(member(ValueNode,State) , get_texpr_info(ValueNode,ValueNodeInfo) ,
            member(synthesis(machinevar,MachineVar),ValueNodeInfo) , member(b(identifier(MachineVar),_,_),CurrentVarNodes)),ReducedState).

%% Only for synthesis of guards or invariants:
% In case we synthesize guard/invariant it is possible that a state is graduated as positive and negative
% because we replace output states with TRUE/FALSE. Thus, the constraint would have a contradiction.
% E.g. synthesizing an invariant: If we graduate [0]-->[-1] as valid and [-1]-->[-2] as invalid,
% we don't want to assume [-1] to be a negative state from the second example.
% Synthesizing a guard it is the opposite, i.e., we don't want to assume [-1] to be a positive state.
get_distinct_states(invariant,Positive,TempNegative,Positive,Negative) :-
    get_distinct_states(Positive,TempNegative,Negative).
get_distinct_states(guard,TempPositive,Negative,Positive,Negative) :-
    get_distinct_states(Negative,TempPositive,Positive).

% Duplicates are subtracted from the examples given in the second argument.
:- spec_pre(get_distinct_states/3,[ast_examples,ast_examples,var]).
:- spec_invariant(get_distinct_states/3,[ast_examples,ast_examples,ast_examples]).
:- spec_post(get_distinct_states/3,[ast_examples,ast_examples,var],[ast_examples,ast_examples,ast_examples]).
get_distinct_states(Positive,TempNegative,Negative) :-
    findall(Neg,(member(Neg,TempNegative) , \+contains_equivalent_state(Neg,Positive,_)),Negative).

% Mark initial examples to be able to restrict constant domains when defining the synthesis
% constraint. Otherwise we would expand the constant domains for each new example derived from
% distinguishing interactions.
mark_examples_as_initial(TempPositive,TempNegative,Positive,Negative) :-
    maplist(mark_examples_as_initial_aux,TempPositive,Positive) ,
    maplist(mark_examples_as_initial_aux,TempNegative,Negative).
mark_examples_as_initial_aux([],[]).
mark_examples_as_initial_aux([ValueNode|T],[b(Node,Type,NewInfo)|NT]) :-
    ValueNode = b(Node,Type,Info) ,
    mark_examples_as_initial_aux2(Info,NewInfo) ,
    mark_examples_as_initial_aux(T,NT).
mark_examples_as_initial_aux2(Info,Info) :-
    member(synthesis(initial_example),Info).
mark_examples_as_initial_aux2(Info,[synthesis(initial_example)|Info]) :-
    \+member(synthesis(initial_example),Info).

% InputList contains all machine variables, but for synthesizing an invariant we only need to
% observe variables from violated invariants.
% Furthermore we accept hints from the user on variables that are probably involved in an action's
% guard so we also need to reduce examples there.
reduce_input_to_violating_vars(ViolatingVars,InputExample,ReducedExample) :-
    % find values of matching variable names located in ast nodes info term synthesis/2
    findall(Value,(member(Value,InputExample) , get_varname_from_node_info(Value, VarName) ,
                   member(VarName,ViolatingVars)),ReducedExample).

% Get the current variables from an example (info contains synthesis/2 with var name).
get_current_vars_from_example(Example,CurrentVars) :-  
    b_get_machinevars_with_operation_params(MachineVars) ,
    findall(VarName,(member(Node,Example) , get_varname_from_node_info(Node,VarName)),CurrentVarNames) ,
    findall(VarNode,(member(VarNode,MachineVars) , VarNode = b(identifier(VarName),_,_) ,
                     member(VarName,CurrentVarNames)),CurrentVars).

value_to_ast(Value,b(value(Value),Type,[])) :-
    kernel_objects:infer_value_type(Value,Type).

get_varname_from_node_info(b(_,_,Info),VarName) :-
    member(synthesis(machinevar,VarName),Info).

get_varname_from_id_node(b(identifier(VarName),_,_),VarName).

% translate prob values to asts node and add the machine variable name to each node's info
example_bindings_to_ast_nodes([],[]).
example_bindings_to_ast_nodes([Binding|T],[Ast|NT]) :-
    (Binding = binding(TVarName,In,_) ; Binding = bind(TVarName,In)) ,
    value_to_ast(In,TAst) ,
    remove_prime(TVarName,VarName) ,
    add_texpr_infos(TAst,[synthesis(machinevar,VarName)],TTAst) ,
    set_type_if_necessary(TTAst,Ast) , ! ,
    example_bindings_to_ast_nodes(T,NT).

remove_prime(b(identifier(VarName),Type,Info),b(identifier(UnPrimed),Type,Info)) :- ! ,
    remove_prime(VarName,UnPrimed).
remove_prime(VarName,UnPrimed) :-
    atom_concat(UnPrimed,'\'',VarName) , !.
remove_prime(VarName,VarName).

% we have to set the types of untyped ast nodes like empty sets
set_type_if_necessary(b(Node,Type,Info),b(TypedNode,GroundType,Info)) :-
    Node \= identifier(_) , Node \= couple(_,_) ,
    untyped(Type) ,
    b_get_machinevars_with_operation_params(MachineVars) ,
    member(synthesis(machinevar,VarName),Info) ,
    member(b(identifier(VarName),GroundType,_),MachineVars) ,
    set_nested_types_if_necessary(Node,GroundType,TypedNode).
% or couples refering to enumerated machine sets
set_type_if_necessary(b(couple(Key,Value),Type,Info),b(couple(TypedKey,TypedValue),couple(KeyType,ValueType),Info)) :-
    untyped(Type) , ! ,
    set_type_if_necessary(Key,TypedKey) ,
    set_type_if_necessary(Value,TypedValue) ,
    get_texpr_type(TypedKey,KeyType) ,
    get_texpr_type(TypedValue,ValueType).
% or elements of enumerated machine sets
set_type_if_necessary(b(identifier(IdName),Type,Info),b(identifier(IdName),GroundType,Info)) :-
    untyped(Type) , ! ,
    b_get_machine_enumerated_sets_with_names(SetNamesTuples) ,
    % get the ast node of the enumerated set the identifier belongs to
    findall(SetNode,(member((SetNode,ElmNames),SetNamesTuples) , member(IdName,ElmNames)),[SetNode]) ,
    SetNode = b(_,set(GroundType),_).
set_type_if_necessary(b(Node,Type,Info),b(Node,Type,Info)).

% set the type of the elements of sets or sequences
set_nested_types_if_necessary(Node,set(InnerType),TypedNode) :-
    Node =.. [SetOrSeq,Elements] ,
    (SetOrSeq = set_extension ; SetOrSeq = sequence_extension) ,
    maplist(set_explicit_type_if_necessary(InnerType),Elements,TypedElements) ,
    TypedNode =.. [SetOrSeq,TypedElements].
set_nested_types_if_necessary(Node,_,Node).

set_explicit_type_if_necessary(InnerType,b(Node,Type,Info),TypedNode) :-
    untyped(Type) , ! ,
    TypedNode = b(Node,InnerType,Info).
set_explicit_type_if_necessary(_,Ast,Ast).

b_get_machine_enumerated_sets_with_names(SetNamesTuples) :-
    findall((SetName,SetNode),b_get_machine_set(SetName,SetNode),NameNodeTuples) ,
    findall((SetName,ElmNames),bmachine:b_get_named_machine_set(SetName,ElmNames),NameElmsTuples) ,
    findall((SetNode,ElmNames),(member((SetName,SetNode),NameNodeTuples) , member((SetName,ElmNames),NameElmsTuples)),SetNamesTuples).

untyped(Type) :-
    var(Type).
untyped(set(Type)) :-
    untyped(Type).
untyped(set(couple(Key,Value))) :-
    untyped(Key) ; untyped(Value).
untyped(seq(Type)) :-
    untyped(Type).
untyped(seq(couple(Key,Value))) :-
    untyped(Key) ; untyped(Value).
% ProB sometimes sets the type of empty sets to set(any)
untyped(any).

% Zip two lists of the same size, i.e., create tuples of elements having the same index respectively.
zip(L1,L2,Zipped) :-
    zip_aux(L1,L2,[],Zipped).
zip_aux([],[],Acc,Acc).
zip_aux([Elm1|T1],[Elm2|T2],Acc,Zipped) :-
    zip_aux(T1,T2,[(Elm1,Elm2)|Acc],Zipped).

% Used in b_synthesis
% An operation satisfies the provided behavior, and thus, we do not have to synthesize anything.
prepare_solution(_,_,operation_satisfied(OperationName),operation_satisfied(OperationName),none).
% Synthesis returned a distinguishing state or transition. The machine code is not changed.
prepare_solution(_,_,state(State),none,state(TranslatedState)) :-
    maplist(translate_example_ast,State,TranslatedState).
prepare_solution(_,_,transition(Input,Output),none,transition(TranslatedInput,TranslatedOutput)) :-
    maplist(translate_example_ast,Input,TranslatedInput) ,
    maplist(translate_example_ast,Output,TranslatedOutput).
% Succeeded synthesizing a program fitting the desired behavior. Apply the solution to the machine code.
prepare_solution(Operation,SynthesisType,Synthesized,NewMachineAtom,distinguishing(none)) :-
    full_b_machine(Machine) ,
    apply_solution_to_machine_and_type_invariant(Machine,Operation,SynthesisType,Synthesized,NewMachineAst) ,
    translate_machine(NewMachineAst,Codes,_) , atom_codes(NewMachineAtom,Codes).

% Return a tuple of machine variable name and value for a given state's value ast.
translate_example_ast(ExampleAst,(MachineVarName,TranslatedValue)) :-
    translate_bexpression(ExampleAst,TranslatedValue) ,
    get_varname_from_node_info(ExampleAst,MachineVarName).

% We derive several operations when implicitly synthesizing if-statements which we here apply to the machine code.
adapt_machine_code_for_operations(Operations,NewMachineAtom) :-
    full_b_machine(Machine) ,
    adapt_machine_code_for_operations_aux(Operations,Machine,TempNewMachine) ,
    % type invariants which may have been optimized by ProB
    b_get_invariant_from_machine(Invariant) ,
    apply_solution_to_machine_and_type_invariant(TempNewMachine,_,invariant,Invariant,NewMachineAst) ,
    translate_machine(NewMachineAst,Codes,_) , atom_codes(NewMachineAtom,Codes).

adapt_machine_code_for_operations_aux([],MachineCode,MachineCode).
adapt_machine_code_for_operations_aux([Operation|T],MachineCode,NewMachineCode) :-
    apply_solution_to_machine(MachineCode,_,action,Operation,TempNewMachine) ,
    adapt_machine_code_for_operations_aux(T,TempNewMachine,NewMachineCode).

apply_solution_to_machine_and_type_invariant(Machine,Operation,Type,Synthesized,NewMachine) :-
    Type \= invariant ,
    apply_solution_to_machine(Machine,Operation,Type,Synthesized,TempNewMachine) ,
    % the invariant has not been changed but typing predicates may be removed by the solver optimization,
    % therefore, we add typing predicates for each machine variable at the beginning of the invariant
    b_get_invariant_from_machine(Invariant) ,
    apply_solution_to_machine(TempNewMachine,_,invariant,Invariant,NewMachine).
apply_solution_to_machine_and_type_invariant(Machine,Operation,invariant,SynthesizedInvariant,NewMachine) :-
    apply_solution_to_machine(Machine,Operation,invariant,SynthesizedInvariant,NewMachine).

apply_solution_to_machine(Machine,_,_,none,Machine).
apply_solution_to_machine(machine(MachineName,Content),_,action,Synthesized,NewMachine) :-
    member(operation_bodies/OperationBodies,Content) ,
    subtract(Content,[operation_bodies/_],TempContent) ,
    apply_solution_to_machine_aux(action,OperationBodies,Synthesized,NewOperationBodies) ,
    append(TempContent,[operation_bodies/NewOperationBodies],NewContent) ,
    NewMachine = machine(MachineName,NewContent).
apply_solution_to_machine(machine(MachineName,Content),Operation,guard,Synthesized,NewMachine) :-
    member(operation_bodies/OperationBodies,Content) ,
    subtract(Content,[operation_bodies/_],TempContent) ,
    apply_solution_to_machine_aux(guard,Operation,Synthesized,OperationBodies,NewOperationBodies) ,
    append(TempContent,[operation_bodies/NewOperationBodies],NewContent) ,
    NewMachine = machine(MachineName,NewContent).
apply_solution_to_machine(machine(MachineName,Content),_,invariant,Synthesized,NewMachine) :-
    subtract(Content,[invariant/_],TempContent) ,
    conjunction_to_list(Synthesized,SynthesizedList) ,
    % split synthesized code from previous invariants that have been valid beforehand
    findall(SynthesizedCode,(member(SynthesizedCode,SynthesizedList) , get_texpr_info(SynthesizedCode,Info) , member(synthesized_code,Info)),SynthesizedCodeList) ,
    conjunct_predicates(SynthesizedCodeList,TempSynthesizedCode) ,
    find_typed_identifier_uses(TempSynthesizedCode,UsedVars) ,
    % create typing invariants only for variables that are involved in the synthesized code, the other typing invariants have already been conjuncted in ValidInvariants
    % we hereby want to avoid redundancy and conflicts in typing invariants, like the valid invariant has been i : NATURAL but generate_typing_predicates creates i : INTEGER afterwards
    % whilst i has not even been affected by the invariant violation
    b_get_typing_predicate_for_vars(UsedVars,TypingPredicate) ,
    conjunct_predicates([TypingPredicate,Synthesized],TypedSynthesizedInvariant) ,
    add_invariant_to_machine_content(TempContent,TypedSynthesizedInvariant,NewContent) ,
    NewMachine = machine(MachineName,NewContent).

% Existing operations are replaced by the synthesized ones.
apply_solution_to_machine_aux(action,OperationBodies,Synthesized,NewOperationBodies) :-
    is_list(Synthesized) ,
    maplist(get_name_from_operation_ast,Synthesized,OperationNames) ,
    remove_operations_from_list(OperationBodies,OperationNames,[],UninvoledOperations) ,
    append(Synthesized,UninvoledOperations,NewOperationBodies).
apply_solution_to_machine_aux(action,OperationBodies,Synthesized,[Synthesized|UninvoledOperations]) :-
    \+is_list(Synthesized) ,
    get_name_from_operation_ast(Synthesized,OperationName) ,
    remove_operations_from_list(OperationBodies,[OperationName],[],UninvoledOperations).

% Given a list of operation bodies, i.e. operation/4 ASTs, remove those whose name is in OperationNames.
remove_operations_from_list([],_,Acc,Acc).
remove_operations_from_list([OperationBody|T],OperationNames,Acc,NewOperationBodies) :-
    get_name_from_operation_ast(OperationBody,OperationName) ,
    \+member(OperationName,OperationNames) ,
    remove_operations_from_list(T,OperationNames,[OperationBody|Acc],NewOperationBodies).
remove_operations_from_list([_|T],OperationNames,Acc,NewOperationBodies) :-
    remove_operations_from_list(T,OperationNames,Acc,NewOperationBodies).

apply_solution_to_machine_aux(guard,OperationName,Synthesized,[Operation|T],[NewOperation|T]) :-
    replace_precondition_if_operation_fits(OperationName,Operation,Synthesized,NewOperation) , !.
apply_solution_to_machine_aux(guard,OperationName,Synthesized,[Operation|T],[Operation|NT]) :-
    apply_solution_to_machine_aux(guard,OperationName,Synthesized,T,NT).

get_name_from_operation_ast(b(operation(ID,_,_,_),_,_),OperationName) :-
    ID = b(identifier(op(OperationName)),_,_).

% insert the synthesized invariant prior to the linking_invariant
add_invariant_to_machine_content([linking_invariant/OperationBodies|T],Invariant,[invariant/Invariant,linking_invariant/OperationBodies|T]).
add_invariant_to_machine_content([ContentType/Content|T],Invariant,[ContentType/Content|NewContent]) :-
    ContentType \= operation_bodies ,
    add_invariant_to_machine_content(T,Invariant,NewContent).

replace_precondition_if_operation_fits(OperationName,Operation,Synthesized,NewOperation) :-
    Operation = b(Node,Type,Info) ,
    Node = operation(OperationNameId,X,Y,Body) ,
    get_varname_from_id_node(OperationNameId,IdName) ,
    IdName = op(OperationName) , % operation fits
    replace_precondition_if_operation_fits_aux(Body,Synthesized,NewBody) ,
    NewOperation = b(operation(OperationNameId,X,Y,NewBody),Type,Info).
% Replace or create a new precondition if the given operation name is equal to the one of the operation.
replace_precondition_if_operation_fits_aux(Body,Synthesized,NewBody) :-
    Body = b(precondition(_,Substitutions),subst,Info) , ! ,
    NewBody = b(precondition(Synthesized,Substitutions),subst,Info).
replace_precondition_if_operation_fits_aux(Body,Synthesized,NewBody) :-
    Body = b(Substitutions,subst,Info) ,
    NewBody = b(precondition(Synthesized,Substitutions),subst,Info).

replace_precondition_of_operation(Operation,none,FinalOperation) :- ! ,
    FinalOperation = Operation.
replace_precondition_of_operation(Operation,Precondition,FinalOperation) :-
    Operation = b(operation(OperationName,Res,Par,Body),subst,_) ,
    Body = b(precondition(_,Subst),_,_) , ! ,
    FinalOperation = b(operation(OperationName,Res,Par,b(precondition(Precondition,Subst),subst,[])),subst,[]).
replace_precondition_of_operation(Operation,Precondition,FinalOperation) :-
    Operation = b(operation(OperationName,Res,Par,Subst),subst,_) ,
    FinalOperation = b(operation(OperationName,Res,Par,b(precondition(Precondition,Subst),subst,[])),subst,[]).

% The Java UI provides tuples of input and output examples.
split_input_output_tuples(InputOutputTupleList,InputList,OutputList) :- 
    split_input_output_tuples(InputOutputTupleList,[],[],InputList,OutputList).
split_input_output_tuples([],InputListAcc,OutputListAcc,InputListAcc,OutputListAcc).
split_input_output_tuples([(Input,Output)|T],InputListAcc,OutputListAcc,InputList,OutputList) :-
    maplist(variable_parsedvalue_tuple_to_ast,Input,InputAsts) ,
    maplist(variable_parsedvalue_tuple_to_ast,Output,OutputAsts) ,
    accumulate_input_output_if_non_empty(InputAsts,InputListAcc,NewInputListAcc) ,
    accumulate_input_output_if_non_empty(OutputAsts,OutputListAcc,NewOutputListAcc) ,
    split_input_output_tuples(T,NewInputListAcc,NewOutputListAcc,InputList,OutputList).

accumulate_input_output_if_non_empty([],Acc,Acc).
accumulate_input_output_if_non_empty(Asts,Acc,[Asts|Acc]) :- Asts \= [].

% Given a tuple of variable name and parsed term. Create the ast using the typechecker and
% add synthesis/2 term to the node's info describing the corresponding machine variable.
variable_parsedvalue_tuple_to_ast((VarName,ParsedValue),TypedAst) :-
    btype(bsynthesis,ParsedValue,openenv(_,env_empty),_,TempAst,[],_) ,
    add_texpr_infos(TempAst,[synthesis(machinevar,VarName)],TempAst2) ,
    remove_set_of_couple_type(TempAst2,TempAst3) ,
    set_type_if_necessary(TempAst3,TempAst4) , ! ,
    get_texpr_type(TempAst4,Type) ,
    add_texpr_infos(TempAst4,[synthesis(type,Type)],TypedAst).

% Remove set(couple(_,_)) type since a sequence with size < 3 may be interpreted as a set of couples although it should be a seq(_).
% The type is added afterwards according to the corresponding machine variable.
remove_set_of_couple_type(b(Node,Type,Info),b(Node,_,Info)) :-
    Type == set(couple(_,_)) , !.
remove_set_of_couple_type(AST,AST).


keep_asserted_initial_examples(Input,Output,OldInput,OldOutput,NewInput,NewOutput) :-
    keep_asserted_initial_states(Input,OldInput,NewInput) ,
    keep_asserted_initial_states(Output,OldOutput,NewOutput).
keep_asserted_initial_states([],_,[]).
keep_asserted_initial_states([State|T],OldStates,[NewState|NT]) :-
    contains_equivalent_state(State,OldStates,EquivalentState) , ! ,
    NewState = EquivalentState ,
    keep_asserted_initial_states(T,OldStates,NT).
keep_asserted_initial_states([State|T],OldStates,[State|NT]) :-
    keep_asserted_initial_states(T,OldStates,NT).

get_operation_params_from_ids(TypedIdentifiers,Params) :- 
    get_operation_params_from_ids(TypedIdentifiers,[],Params).
get_operation_params_from_ids([],Acc,Acc).
get_operation_params_from_ids([TypedIdentifier|T],Acc,Params) :- 
    get_texpr_info(TypedIdentifier,Info) , 
    ((member(synthesis(ParamType,_),Info) , (ParamType = param_single ; ParamType = param_set)) 
    ; member(introduced_by(operation),Info)) , 
    TypedIdentifier = b(Node,Type,_) , 
    \+ member(b(Node,Type,_),Acc) , ! , 
    get_operation_params_from_ids(T,[TypedIdentifier|Acc],Params).
get_operation_params_from_ids([_|T],Acc,Params) :- 
    get_operation_params_from_ids(T,Acc,Params).

set_operation_params_domain(Params,Membership) :- 
  set_operation_params_domain(Params,[],MembershipList) , 
  conjunct_predicates(MembershipList,Membership).

set_operation_params_domain([],Acc,Acc).
set_operation_params_domain([Param|T],Acc,MembershipList) :- 
  get_texpr_info(Param,ParamInfo) , 
  (member(synthesis(type,set(global(IndependentType))),ParamInfo) ; Param = b(_,global(IndependentType),_)) , 
  bmachine:b_get_named_machine_set(IndependentType,SetAst,_) , 
  set_operation_params_domain(T,[b(member(Param,SetAst),pred,[])|Acc],MembershipList).

finalize_synthesized_action(_,none,none).
finalize_synthesized_action(OperationName,b(precondition(PreCondition,Substitutions),subst,[]),
                            b(operation(OpNameId,_,Params,b(precondition(NewPreCondition,UnprimedSubstitutions),subst,[])),subst,[])) :-
    OpNameId = b(identifier(op(OperationName)),op([],[]),[]) ,
    find_typed_identifier_uses_with_info(Substitutions,TypedIdentifiers) , 
    get_operation_params_from_ids(TypedIdentifiers,Params) , 
    Params \= [] , ! , 
    set_operation_params_domain(Params,ParamTypeDefsPred) , 
    conjunct_predicates([ParamTypeDefsPred,PreCondition],NewPreCondition) , 
    prime_or_unprime_variables_in_ast(unprime,[],Substitutions,UnprimedSubstitutions).
finalize_synthesized_action(OperationName,b(precondition(PreCondition,Substitutions),subst,[]),
                            b(operation(OpNameId,_,_,b(precondition(PreCondition,UnprimedSubstitutions),subst,[])),subst,[])) :-
    OpNameId = b(identifier(op(OperationName)),op([],[]),[]) ,
    prime_or_unprime_variables_in_ast(unprime,[],Substitutions,UnprimedSubstitutions).
finalize_synthesized_action(OperationName,Action,b(operation(OpNameId,_,Params,b(precondition(ParamTypeDefsPred,UnprimedAction),subst,[])),subst,[])) :-
    OpNameId = b(identifier(op(OperationName)),op([],[]),[]) ,
    find_typed_identifier_uses_with_info(Action,TypedIdentifiers) , 
    get_operation_params_from_ids(TypedIdentifiers,Params) , 
    Params \= [] , ! , 
    set_operation_params_domain(Params,ParamTypeDefsPred) , 
    prime_or_unprime_variables_in_ast(unprime,[],Action,UnprimedAction).
finalize_synthesized_action(OperationName,Action,b(operation(OpNameId,_,_,UnprimedAction),subst,[])) :-
    OpNameId = b(identifier(op(OperationName)),op([],[]),[]) ,
    prime_or_unprime_variables_in_ast(unprime,[],Action,UnprimedAction).

% Synthesizing a guard for an action simultaneously can fail. We memorize distinguishing transitions
% so even without a guard we won't find a distinguishing transitions twice. However,
% finding a new distingushing transition that leads to synthesizing a guard simultaneously may
% succeed with the new example later on (previous distinguishing states are also memorized and will be reused).
prepare_synthesized_guards_for_action(Guards,Guards) :-
    is_ast(Guards).
prepare_synthesized_guards_for_action(_,any).


% Get all the location vars from a list of operators. The result is a map with entries like
% Operator:ListOfLocationVars. Just one iteration of the list of location variables for all operators.
%get_all_location_vars_for_operators(Operators,LocationVars,OperatorLVarsMap) :-
%    findall(Op:[],member(Op,Operators),EmptyAcc) ,
%    get_all_location_vars_for_operators_aux(LocationVars,EmptyAcc,OperatorLVarsMap).
%get_all_location_vars_for_operators_aux([],Acc,Acc).
%get_all_location_vars_for_operators_aux([LocationVar|T],Acc,OperatorLVarsMap) :-
%    LocationVar = b(_,_,Info) ,
%    member(synthesis(ComponentName,LocationType),Info) ,
%    member(LocationType,[left_input,right_input,output]) ,
%    delete_numbers_from_atom(ComponentName,Operator) ,
%    accumulate_locationvar_for_operator(LocationVar,Operator,Acc,NewAcc) ,
%    get_all_location_vars_for_operators_aux(T,NewAcc,OperatorLVarsMap).
%get_all_location_vars_for_operators_aux([_|T],Acc,OperatorLVarsMap) :-
%    get_all_location_vars_for_operators_aux(T,Acc,OperatorLVarsMap).

% If the operator the location variable refers to is one that we search for, we accumulate the variable respectively.
%accumulate_locationvar_for_operator(LocationVar,Operator,Acc,NewAcc) :-
%    member(Operator:OperatorLocationVars,Acc) ,
%    delete(Acc,Operator:_,TempAcc) ,
%    NewAcc = [Operator:[LocationVar|OperatorLocationVars]|TempAcc].
% Otherwise skip the location variable.
%accumulate_locationvar_for_operator(_,Operator,Acc,Acc) :-
%    \+member(Operator:_,Acc).

% Adjust the distinguishing state to the current variables (synthesis type is guard or invariant).
%:- spec_pre(adjust_distinguishing_state_binding/3,[list(ast),solution_bindings,var]).
%:- spec_invariant(adjust_distinguishing_state_binding/3,[list(ast),solution_bindings,solution_bindings]).
%:- spec_post(adjust_distinguishing_state_binding/3,[list(ast),solution_bindings,var],
%                                                   [list(ast),solution_bindings,solution_bindings]).
%adjust_distinguishing_state_binding(CurrentVars,Distinguishing,Distinguishing) :-
%    length(CurrentVars,VarAmount) ,
%    length(Distinguishing,VarAmount).
% Shrink distinguishing state.
%adjust_distinguishing_state_binding(CurrentVars,Distinguishing,ShrunkenInput) :-
%    length(CurrentVars,VarAmount) ,
%    length(Distinguishing,DistAmount) ,
%    VarAmount < DistAmount ,
%    findall(Binding,(member(Binding,Distinguishing) , Binding = binding(VarName,_,_) ,
%                     member(b(identifier(VarName),_,_),CurrentVars)),ShrunkenInput).
% Expand distinguishing state.
%adjust_distinguishing_state_binding(CurrentVars,Distinguishing,ShrunkenInput) :-
%    length(CurrentVars,VarAmount) ,
%    length(Distinguishing,DistAmount) ,
%    VarAmount > DistAmount ,
%    findall(Binding,(member(b(identifier(VarName),_,_),CurrentVars) ,
%                     \+member(Binding,Distinguishing) , Binding = binding(VarName,_,_)),ShrunkenInput).
