% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(tools_printing,[
          print_term_summary/1, print_term_summary_hash/1,
          print_term_summary_wo_nl/1,
          print_functor/1,
          print_var/1, print_vars/1,
          watch/3,
          trace_unify/2, trace_print/1, trace_print/2,
          print_error/1,
          print_red/2, print_red/1, print_green/1,
          format_with_colour/4,
          start_terminal_colour/2,
          reset_terminal_colour/1,
          print_time_stamp/1,
          print_dynamic_pred/3, print_dynamic_fact/1,
          print_goal/1,
          nested_print_term/2
          ]).


:- use_module(module_information).
:- module_info(group,infrastructure).
:- module_info(description,'This module contains printing/debug helper predicates.').

:- use_module(library(terms),[term_hash/2]).
print_term_summary_hash(T) :- print_term_summary(T),
   term_hash(T,H),  %% (H=179896960 -> trace ; true), %%
   format('Hash = ~w~n',H).

%:- use_module(library(terms),[term_size/2]).
print_term_summary(T) :- print_term_summary_wo_nl(T),
  %term_size(T,Sz), format(' term_size: ~w',[Sz]),
  nl.

print_term_summary_wo_nl(N) :- var(N),!,print('VARIABLE: '),print(N).
print_term_summary_wo_nl((A,B)) :- print('('), print_term_summary_wo_nl(A),
    print(','), print_term_summary_wo_nl(B),print(')').
print_term_summary_wo_nl(clpfd:Call) :- !, print(clpfd:Call).
print_term_summary_wo_nl(Module:Call) :- !,
    print(Module), print(':'),print_term_summary_wo_nl(Call).
print_term_summary_wo_nl(bind(Var,Val)) :- !,print_arg(bind(Var,Val)).
print_term_summary_wo_nl(N) :- functor(N,F,Arity), print(F), print('/'),print(Arity),
                     N=..[F|Args], l_print_arg(Args).

l_print_arg(V) :- var(V),!, print('...').
l_print_arg([]).
l_print_arg([H|T]) :- print(': '), print_arg(H), l_print_arg(T).



:- use_module(library(avl)).
% USE MODULE NOT PERFORMED TO AVOID LOADING translate and all dependent modules
%:- use_module(translate,[print_bexpr_with_limit/2, print_bstate/1]).

print_arg(N) :- var(N),!,print_var(N).
print_arg(int(N)) :- !, (var(N) -> print_var_integer(N) ; print(int(N))).
print_arg(float(N)) :- !, print(float(N)).
print_arg(fd(N,GS)) :- !,
  ((number(N),b_global_sets:is_b_global_constant(GS,N,Res)) -> print(Res)
    ; print('fd('),print_integer(N),format(',~w)',GS)).
print_arg(rec(Fields)) :- !, print('rec('), print_arg(Fields),print(')').
print_arg(bind(Var,Val)) :- atomic(Var), !, print(Var),print('/'),print_arg(Val).
print_arg(field(Var,Val)) :- atomic(Var), !, print('field('),print(Var),print('/'),print_arg(Val),print(')').
print_arg(typedvalc(Val,Type,VarID,EnumWarning,Card)) :-
   atomic(VarID), !, format('typedvalc for ~w :(',[VarID]),print_arg(Val),
   format(',~w,~w,~w) ',[Type,EnumWarning,Card]).
print_arg(wfx(A,B,C,D)) :- !, print(wfx(A,B,C,D)).
print_arg(Module:Call) :- atomic(Module),!,
    print(Module), print(':'),print_arg(Call).
print_arg(A) :- atomic(A),!, print(A).
print_arg(A) :- A=node(_,_,_,_,_),ground(A), !, print('AVLnode.size='),avl_size(A,Sz), print(Sz).
print_arg(avl_set(A)) :- ground(A), !,
  (custom_explicit_sets:singleton_set(avl_set(A),El) -> print('AVL('),print_arg(El),print(')')
   ;  print('AVL.size='),avl_size(A,Sz), print(Sz)).
print_arg((A,B)) :- !, print('('),print_arg(A), print(','), print_arg(B), print(')').
print_arg(string(A)) :- !, print('string('),print_arg(A), print(')').
print_arg([]) :- !, print('[]').
print_arg([H|T]) :- T==[], !, print('['),print_arg(H),print(']').
print_arg([H|T]) :- nonvar(T),T=[H2|T2],T2==[], !,
     print('['),print_arg(H),print(','),print_arg(H2),print(']').
print_arg([H|T]) :- nonvar(H), H=bind(_,_),!, get_list_up_to([H|T],10,S,ALL),
   translate:print_bstate(S),
   (ALL=all -> true ; print(', ... ')).
print_arg([H|T]) :- list_skeleton_size([H|T],Sz,Type), !,
   print(Type),
   ((Type=closed,ground([H|T])) -> print('&ground') ; true),
   print(' LIST.size='), print(Sz),
   (Sz<6 -> print(' ['), l_print_arg([H|T]), print(']') ; true).
print_arg(closure(P,T,B)) :- !, print('CLOSURE '),
  (custom_explicit_sets:is_interval_closure(P,T,B,Low,Up) -> print([Low:Up])
    ; print(P)). %, print(' | '), print(_B).
print_arg(mnf(Call)) :- !, print('mnf('),print_term_summary_wo_nl(Call), print(')').
print_arg(pp_mnf(Call)) :- !, print('mnf('),print_term_summary_wo_nl(Call), print(')').
print_arg(pp_cll(Call)) :- !, print('mnf('),print_term_summary_wo_nl(Call), print(')').
print_arg(b(Expr,T,I)) :- !,
       (T==pred -> print('PRED(') ; print('BEXPR(')),
       print_functor(Expr), print(','),
       translate:print_bexpr_with_limit(b(Expr,T,I),200), print(')').
print_arg(N) :- print_functor(N).

:- use_module(library(clpfd),[fd_dom/2, fd_degree/2]).
print_var(V) :- fd_dom(V,D), D \= inf..sup, fd_degree(V,Dg),!,
   format('INT VAR: ~w:~w [~w] ',[V,D,Dg]).
   %print(':('),frozen(V,Goal),write_term(Goal,[max_depth(3)]),print(')').
print_var(V) :- print('VARIABLE: '),print(V), print_frozen_var_info(V).
print_var_integer(X) :- print('int(?:'),fd_dom(X,Dom),print(Dom),
    %print(':('),frozen(X,Goal),print(Goal),print(')')
    print(')').

% try determine variable type from frozen info
print_frozen_var_info(V) :- frozen(V,G), (print_frozen_var_info2(G,V) -> true ; true).
print_frozen_var_info2((A,B),V) :- (print_frozen_var_info2(A,V) -> true ; print_frozen_var_info2(B,V)).
print_frozen_var_info2(b_interpreter_check:imply_true(_,_,_),_) :- print(' : BOOL').
print_frozen_var_info2(bool_pred:blocking_force_eq(V,_,_),Var) :- Var==V, print(' : BOOL').
print_frozen_var_info2(kernel_equality:eq_empty_set(V,_),Var) :- Var==V, print(' : SET').
% TO DO: add other co-routines for boolean values,...

:- use_module(library(lists),[maplist/2]).
print_vars(V) :- maplist(tools_printing:print_var,V),nl.

print_integer(I) :- (var(I) -> fd_dom(I,Dom), print('?:'),print(Dom) ; print(I)).

get_list_up_to([],_,[],all).
get_list_up_to([H|T],N,R,ALL) :-
  (N<1 -> R=[],ALL=no ; N1 is N-1, R=[H|TR], get_list_up_to(T,N1,TR,ALL)).

list_skeleton_size(X,Sz,Type) :- var(X),!,Sz=0,Type=open.
list_skeleton_size([],0,closed).
list_skeleton_size([_|T],N,Type) :- list_skeleton_size(T,NT,Type), N is NT+1.


print_functor(N) :- var(N),!,print('VARIABLE: '),print(N).
print_functor(N) :- functor(N,F,Arity), print(F), print('/'),print(Arity).

% ---------------

:- block watch(-,?,?).
watch([],N,Pos) :- !,print_data([],N,Pos).
watch([H|T],N,Pos) :- !,print_data(H,N,Pos), P1 is Pos+1, watch(H,N,Pos), watch(T,N,P1).
watch((P1,P2),N,Pos) :- print_data((P1,P2),N,Pos), watch(P2,N,Pos).
watch(Other,N,Pos) :- print_data(Other,N,Pos).
print_data(Data,N,Pos) :- print(N), print(' @ '), print(Pos),print(' : '), print(Data),nl.

% ---------------

:- use_module(library(lists),[maplist/3]).
trace_unify(A,B) :- trace_unify_aux(A,B,0).
trace_unify_aux(A,B,Lvl) :- (var(A);var(B)),!,indent(Lvl),print(A),print(' <-> '), print(B),
   if(A=B,true,(print(' FAILS'),nl,fail)),nl.
trace_unify_aux(A,B,Lvl) :- functor(A,FA,FAN), functor(B,FB,FBN),
   indent(Lvl),print(FA/FAN),
   A=..[_|As], B=..[_|Bs],
   ((FA,FAN)=(FB,FBN)
     -> (As=[] -> nl ; print('('),nl,l_trace_unify(As,Bs,Lvl))
     ;  format(' FUNCTOR MISMATCH : ~w/~w <-> ~w/~w~n',[FA,FAN,FB,FBN]),
        maplist(abstract_top_level,As,AAs), AA =..[FA|AAs],
        maplist(abstract_top_level,Bs,ABs), AB =..[FB|ABs],
        indent(Lvl),
        format('  TOP: ~w <-> ~w~n',[AA,AB]),
        fail
   ).
indent(0) :- !.
indent(X) :- X>0, print('+-'), X1 is X-1, indent(X1).

abstract_top_level(X,R) :- simple(X),!,R=X.
abstract_top_level(X,AX) :- functor(X,F,N), functor(AX,F,N).


l_trace_unify([],[],L) :- indent(L), print(')'),nl.
l_trace_unify([A|TA],[B|TB],Lvl) :- L1 is Lvl+1, trace_unify_aux(A,B,L1),
   l_trace_unify(TA,TB,Lvl).

% ---------------

trace_print(A) :- trace_print(A,0,4).
trace_print(A,Max) :- trace_print(A,0,Max).

trace_print(A,Lvl,MaxLvl) :- Lvl >= MaxLvl,!, indent(Lvl), print(A),nl.
trace_print(A,Lvl,_MaxLvl) :- treat_as_atomic(A),!, indent(Lvl), print(A),nl.
trace_print([H|T],Lvl,MaxLvl) :- !, indent(Lvl), print('['),nl,
    l_trace_print([H|T],Lvl,MaxLvl),
    indent(Lvl), print(']'),nl.
%trace_print(A,Lvl,_MaxLvl) :- functor(A,_,1), arg(1,A,AN), atomic(AN),!, indent(Lvl), print(A),nl.
trace_print(A,Lvl,_MaxLvl) :- A=..[_|Args], maplist(atomic,Args),!,
    indent(Lvl), print(A),nl.
trace_print(A,Lvl,MaxLvl) :- functor(A,FA,FAN),
   indent(Lvl),print(FA/FAN),print('('),nl,
   A=..[_|As], l_trace_print(As,Lvl,MaxLvl).

treat_as_atomic(A) :- var(A).
treat_as_atomic(A) :- atomic(A).
treat_as_atomic(nodeid(pos(_,_,_,_,_,_))).
treat_as_atomic(sharing(_,_,_,_)).
treat_as_atomic(ID/V) :- treat_as_atomic(ID),treat_as_atomic(V).
treat_as_atomic([H|T]) :- T==[], treat_as_atomic(H).

l_trace_print([],L,_MaxLvl) :- indent(L), print(')'),nl.
l_trace_print([A|TA],Lvl,MaxLvl) :- L1 is Lvl+1, trace_print(A,L1,MaxLvl),
   l_trace_print(TA,Lvl,MaxLvl).

% -----------------------------------

% print on error stream
print_error(Error) :-
    start_terminal_colour([red,bold],user_error),
    write(user_error,'! '),
    (var(Error)  -> write(user_error,'_')
    ;  write_term(user_error,Error,[max_depth(20),numbervars(true)])),
    nl(user_error),
    reset_terminal_colour(user_error).
    %% flush_output(user_error), %%.

format_with_colour(Stream,Colour,Str,Args) :-
   start_terminal_colour(Colour,Stream),
   format(Stream,Str,Args),
   reset_terminal_colour(Stream).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(no_terminal_colors,true)).
no_color.
print_red(X) :- write(X).
print_red(Stream,T) :- write(Stream,T).
print_green(X) :- write(X).
reset_terminal_colour(_).
start_terminal_colour(_,_).
:- else.
:- use_module(library(system),[environ/2]).
no_color :- environ('NO_COLOR',_). % see http://no-color.org
print_red(Stream,Term) :- no_color,!,
   format(Stream,'~w',[Term]).
print_red(Stream,Term) :-
   format(Stream,'\e[31m~w\e[0m',[Term]).

print_red(Term) :- no_color,!,
   format('~w',[Term]).
print_red(Term) :-
   format('\e[31m~w\e[0m',[Term]).

print_green(Term) :- no_color,!,
   format('~w',[Term]).
print_green(Term) :-
   format('\e[32m~w\e[0m',[Term]).

reset_terminal_colour(_) :- no_color,!.
reset_terminal_colour(Stream) :- write(Stream,'\e[0m').

% see https://misc.flogisoft.com/bash/tip_colors_and_formatting
start_terminal_colour(_,_) :- no_color,!.
start_terminal_colour(red,Stream) :- !, write(Stream,'\e[31m').
start_terminal_colour(green,Stream) :- !, write(Stream,'\e[32m').
start_terminal_colour(yellow,Stream) :- !, write(Stream,'\e[33m').
start_terminal_colour(blue,Stream) :- !, write(Stream,'\e[34m').
start_terminal_colour(magenta,Stream) :- !, write(Stream,'\e[35m').
start_terminal_colour(cyan,Stream) :- !, write(Stream,'\e[36m').
start_terminal_colour(light_gray,Stream) :- !, write(Stream,'\e[37m').
start_terminal_colour(dark_gray,Stream) :- !, write(Stream,'\e[90m').
start_terminal_colour(light_red,Stream) :- !, write(Stream,'\e[91m').
start_terminal_colour(light_green,Stream) :- !, write(Stream,'\e[92m').
start_terminal_colour(white,Stream) :- !, write(Stream,'\e[97m').
start_terminal_colour(bold,Stream) :- !, write(Stream,'\e[1m').
start_terminal_colour(underline,Stream) :- !, write(Stream,'\e[4m').
start_terminal_colour(dim,Stream) :- !, write(Stream,'\e[2m').
start_terminal_colour(black_background,Stream) :- !, write(Stream,'\e[49m').
start_terminal_colour(red_background,Stream) :- !, write(Stream,'\e[41m').
start_terminal_colour(green_background,Stream) :- !, write(Stream,'\e[42m').
start_terminal_colour(yellow_background,Stream) :- !, write(Stream,'\e[43m').
start_terminal_colour(white_background,Stream) :- !, write(Stream,'\e[107m').
start_terminal_colour(blink,Stream) :- !, write(Stream,'\e[5m').
start_terminal_colour(reverse,Stream) :- !, write(Stream,'\e[7m').
start_terminal_colour(hidden,Stream) :- !, write(Stream,'\e[8m').
start_terminal_colour(reset,Stream) :- !, write(Stream,'\e[0m').
start_terminal_colour([],_Stream) :- !.
start_terminal_colour([H|T],Stream) :- !, start_terminal_colour(H,Stream),start_terminal_colour(T,Stream).
start_terminal_colour(C,_Stream) :- format(user_error,'*** UNKNOWN COLOUR: ~w',[C]).
:- endif.

% -----------------------------------

:- use_module(library(system),[datime/1]).

print_time_stamp(Stream) :-
    datime(datime(Year,Month,Day,Hour,Min,Sec)),
    (Min<10
      -> format(Stream,'~w/~w/~w - ~wh0~w ~ws',[Day,Month,Year,Hour,Min,Sec])
       ; format(Stream,'~w/~w/~w - ~wh~w ~ws',[Day,Month,Year,Hour,Min,Sec])
    ).

% -----------------------------------

print_dynamic_pred(InModule,Pred,N) :- nl,
   write(':- dynamic '), print(Pred), write('/'),print(N), write('.'),nl,
   functor(Call,Pred,N),
   call(InModule:Call), write_term(Call,[quoted(true)]),write('.'),nl,fail.
print_dynamic_pred(InModule,Pred,N) :-
   functor(Call,Pred,N),
   (call(InModule:Call) -> true ; (write_with_numbervars(Call), write(' :- fail.'),nl)),
   nl.

write_with_numbervars(Term) :- copy_term(Term,T), numbervars(T,0,_),
   write_term(T,[quoted(true),numbervars(true)]).

print_dynamic_fact(Fact) :-
   functor(Fact,Pred,N),
   print(':- dynamic '), print(Pred), write('/'),print(N), write('.'),nl,
   write_term(Fact,[quoted(true)]),write('.'),nl.

% -----------------------------------

print_goal(V) :- var(V),!, format(' ~w~n',[V]).
print_goal([A|B]) :- !,print_goal(A), print_goal(B).
print_goal((A,B)) :- !,print_goal(A), print_goal(B).
print_goal(A) :- format('  ~w~n',[A]).


% -----------------------------------

% possible applicaton: inspect size of terms transmitted to ProB2:
% prob2_interface:get_machine_formulas(Formulas), tell('formulasnest.pl'), nested_print_term(Formulas,6),told.

nested_print_term(Term,MaxNest) :- nested_print_term(0,MaxNest,Term).

indentws(0) :- !.
indentws(X) :- X>0, print('  '), X1 is X-1, indentws(X1).

:- use_module(library(terms),[term_size/2]).
nested_print_term(CurNesting,_Max,T) :- var(T),!, indent(CurNesting),write(T),nl.
nested_print_term(CurNesting,_,T) :- atom(T), atom_length(T,Len),
    indent(CurNesting),
    (Len > 100
     -> atom_codes(T,Codes), length(Prefix,50), append(Prefix,_,Codes),
        format('atom: len(~w): ~s...~n',[Len,Prefix])
      ; format('~w~n',[T])).
nested_print_term(CurNesting,_,T) :- number(T), indent(CurNesting), format('~w~n',[T]).
nested_print_term(CurNesting,MaxNesting,T) :- CurNesting >= MaxNesting,!,
    term_size(T,Sz), functor(T,F,N),
    indentws(CurNesting),
    (Sz>25000 -> Msg='****' ; Msg=''), format('~w/~w :: sz(~w)~w~n',[F,N,Sz,Msg]).
nested_print_term(CurNesting,MaxNesting,T) :-
    list_skeleton_size(T,Len,closed), term_size(T,Sz),!,
    indentws(CurNesting),format('[ list_len_sz(~w,~w)~n',[Len,Sz]),
    C1 is CurNesting+1,
    maplist(nested_print_term(C1,MaxNesting),T),
    indentws(CurNesting),format('] list_len(~w)~n',[Len]).
nested_print_term(CurNesting,MaxNesting,T) :-
    term_size(T,Sz), functor(T,F,N),
    indentws(CurNesting),format('~w/~w( sz(~w)~n',[F,N,Sz]),
    T =.. [_|Args], C1 is CurNesting+1,
    maplist(nested_print_term(C1,MaxNesting),Args),
    indentws(CurNesting),format('~w/~w)~n',[F,N]).


