% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(b_interpreter,
   [b_test_boolean_expression/4, b_test_boolean_expression/6,
    b_test_boolean_expression_wf/3,
    % b_det_test_boolean_expression/6,
    b_not_test_boolean_expression/4, b_not_test_boolean_expression/6,
    b_not_test_boolean_expression_wf/3,

    b_compute_expression/5,
  %  b_execute_statement/6,
    b_execute_top_level_statement/7,

    b_compute_expression_nowf/4,

    b_set_up_concrete_constants/1, b_partial_set_up_concrete_constants/1,
    b_initialise_machine/4,

    b_execute_top_level_operation_update/5, b_execute_top_level_operation_wf/6,

    convert_list_of_expressions_into_set/2,
    %state_satisfies_negation_of_invariant/1,

    b_state_violates_invariant/2,
    state_violates_assertions/2,

    insert_before_substitution_variables/5,  % add $0 values to a store

    try_lookup_value_in_store_and_global_sets/3, lookup_value_in_store_and_global_sets/4,
    set_up_typed_localstate/6, set_up_typed_localstate2/7,
    %set_up_localstate_for_let/6,

    b_get_machine_operation_for_animation/6,
    set_projection_on_static_assertions/1, reset_b_interpreter/0, split_names_and_types/3,

    tcltk_unsatisfiable_components_exist/0, tcltk_quick_describe_unsat_properties/1,
    get_unsat_component_predicate/2,
    b_generate_for_all_list_domain_nolwf/8 % for predicate_evaluator
    ]).

%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling(b_test_boolean_expression/4).
%:- enable_profiling(b_test_boolean_expression_wf/3).

:- use_module(kernel_objects).
:- use_module(bsets_clp).
:- use_module(delay).


:- use_module(kernel_mappings).
:- use_module(b_global_sets).
:- use_module(store).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).

:- use_module(b_global_sets).
:- use_module(bmachine).
:- use_module(b_enumerate, [b_enumerate_values_in_store/5,
      b_tighter_enumerate_values/2]).
:- use_module(debug).
:- use_module(performance_messages).
:- use_module(self_check).
:- use_module(tools).
:- use_module(tools_printing,[print_error/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides the basic interpreter for expressions, predicates and substitutions.').

:- use_module(preferences).
:- use_module(error_manager).
:- use_module(typechecker).
:- use_module(specfile,[state_corresponds_to_initialised_b_machine/2
                       ,state_corresponds_to_set_up_constants_only/2]).

:- use_module(btypechecker,[couplise_list/2]).
:- use_module(bsyntaxtree).

:- use_module(translate).

:- use_module(kernel_objects).
:- use_module(kernel_waitflags).
:- use_module(kernel_records).
:- use_module(kernel_z).

:- use_module('kodkod/kodkod').

:- use_module(custom_explicit_sets).

:- use_module(state_space).
:- use_module(b_interpreter_eventb).
:- use_module(tools_printing,[print_term_summary/1,print_functor/1]).
:- use_module(value_persistance,[load_partial_constants/3,lookup_cached_transitions/3]).
:- use_module(b_expression_sharing,[is_lazy_let_identifier/1]).
:- use_module(bool_pred,[negate/2]).
:- use_module(smt_solvers_interface(smt_solvers_interface),[smt_add_predicate/5]).

:- use_module(debug).
:- use_module(external_functions).
:- if(current_prolog_flag(dialect,sicstus)).
:- if((current_prolog_flag(version_data,sicstus(4,X,_,_,_)),X<3)).
:- use_module(library(terms),[term_variables/2]). % is built-in in SICSTUS 4.3
:- endif.
:- endif.

/* ----------------------------------*/

/* For partial evaluation */

%:- use_module('~/cvs_root/cogen2/logen_source/runtime/logen_dispatcher.pl').

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling_naming(b_interpreter:b_state_violates_invariant_aux/2,name).

%name(b_interpreter:b_state_violates_invariant_aux,_,INVERIANT).

%b_state_violates_invariant(ID,State) :-
%   b_intepreter:b_state_violates_invariant_aux(ID,State).

% old profiler
:- use_module(runtime_profiler,[profile_single_call/3]).
b_state_violates_invariant(ID,State) :-
   profile_single_call('INVARIANT',ID,b_interpreter:b_state_violates_invariant_aux(ID,State)).

b_state_violates_invariant_aux(ID,CurBState) :-
   retract(specialized_inv(ID,AVL)),  %print(specialized_inv(ID,AVL)),nl,
   intersect(AVL,SPO), !,
   \+ check_invariant_predicate_with_time_out(ID,SPO,CurBState).
b_state_violates_invariant_aux(ID,CurBState) :-
  %%  print(checking_NORMAL_inv(_ID)),nl, %,print_bexpr(CurBState),nl,
   state_violates_invariant(ID,CurBState).

intersect(A,Inv) :-
 % trace,
  avl_domain(A,List),
  get_predicates(List,PList),
  ord_intersection(PList,InvList),
  conjunct_predicates(InvList,Inv).

get_predicates([],[]).
get_predicates([H|T],[OHL|TR]) :-
   b_specialized_invariant_for_op(H,HR),
   conjunction_to_list(HR,HL),
   list_to_ord_set(HL,OHL),
   get_predicates(T,TR).
/* ----------------------------------*/

 :- load_files(library(system), [when(compile_time), imports([environ/2])]).

:- if(environ(use_pe_cache, true)).

 % just use a previously generate pe_cache file defining
  %  state_satisfies_invariant_aux and
  %  b_execute_top_level_operation_update(Name,FullOp,InState,Updates,Path)
:- include(pe_cache).
reset_partial_evaluator. % nothing to be done

:- elif(\+ environ(partially_evaluate, true)).

reset_partial_evaluator. % nothing to be done

state_satisfies_invariant_aux(ID,State) :-
   check_invariant_with_time_out(ID,State).  % version without partial evaluation

:- else.
/* partial evaluation enabled */
:- use_module(partial_evaluator(pe)).
% ANNOTATIONS moved to unused_code/

reset_partial_evaluator :- clear_specialized_code.
state_satisfies_invariant_aux(ID,State) :-
    MemoCode=generalize_call(list(term(bind,[atomic,dynamic])),State,StateSkel),
    fast_partially_evaluate_and_load(check_invariant_with_time_out(PID,StateSkel),
                                     ResCall,MemoCode,[ground/StateSkel]),
    StateSkel=State, PID=ID,
    set_error_context(checking_invariant),
    call(ResCall),
    clear_error_context.

%abstract_state([],[]).
%abstract_state([bind(ID,_)|T],[bind(ID,_)|AT]) :- abstract_state(T,AT).
:- endif.

:- use_module(library(timeout)).


check_invariant_with_time_out(ID,State) :-
   b_get_invariant_from_machine(Invariant),
   check_invariant_predicate_with_time_out(ID,Invariant,State).

check_invariant_predicate_with_time_out(_ID,b(truth,_,_),_State) :- !. % only typing in invariant
check_invariant_predicate_with_time_out(ID,Invariant,State) :-
    set_error_context(checking_invariant),
    %debug_println(10,checking_invariant), % translate:print_bexpr(Invariant),nl, print(State),nl,
    preferences:get_computed_preference(debug_time_out,DebugTimeOut),
    % print(checking_invariant(ID,State)),nl,
    % print_bexpr(Invariant),nl,
   % statistics(runtime,_),
    call_cleanup(
      (time_out_with_enum_warning_one_solution_no_new_error_scope(
            b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],State ),
            DebugTimeOut,TimeOutRes)
         -> check_timeout(TimeOutRes,ID)),
     clear_error_context).
    %% , statistics(runtime,[_,DeltaT]), print(delta_time(inv,DeltaT)),nl.

check_timeout(time_out,ID) :- !,
         nl,print_message('**** TIMEOUT during invariant evaluation !'),nl,
         print_message(node_id(ID)),nl,
         assert_time_out_for_invariant(ID),
         fail.
check_timeout(virtual_time_out(_WARNING),ID) :- !,
         nl,print_message('**** VIRTUAL TIMEOUT during invariant evaluation !'),nl,
         print_message(node_id(ID)),nl,
         assert_time_out_for_invariant(ID),
         fail.
check_timeout(_,_).

% b_test_boolean_expression_for_ground_state:
%     idea: check each conjunct individually; avoids multiplication of choice points between conjuncts
b_test_boolean_expression_for_ground_state(b(Expr,_,Infos),LS,S) :- !,
	b_test_boolean_expression_for_ground_state2(Expr,Infos,LS,S).
:- if(\+ environ(partially_evaluate, true)).
b_test_boolean_expression_for_ground_state2(conjunct(LHS,RHS),_,LocalState,State) :- !,
 % usually LHS is the nested conjunct and RHS atomic: it would be better to swap LHS and RHS: but does not seem to buy anything
        b_test_boolean_expression_for_ground_state(LHS,LocalState,State),
        !, % do we need this cut; it makes partial evaluation more tricky
        b_test_boolean_expression_for_ground_state(RHS,LocalState,State).
:- else.
b_test_boolean_expression_for_ground_state2(conjunct(LHS,RHS),_,LocalState,State) :- !,
        b_test_boolean_expression_for_ground_state(LHS,LocalState,State),
        %!, % do we need this cut; it makes partial evaluation more tricky <--- commented out for PE
        b_test_boolean_expression_for_ground_state(RHS,LocalState,State).
:- endif.
% TO DO: treate lazy_let_pred ?
b_test_boolean_expression_for_ground_state2(BE,Infos,LS,S) :-
 %   (preferences:get_preference(provide_trace_information,true)
 %     -> nl,print(' ====> '),print_bexpr_with_limit(BE,150),nl ; true),!,
% print_bexpr_with_limit(BE,150),nl, statistics(runtime,_),
    b_test_boolean_expression_wf(b(BE,pred,Infos),LS,S)
    % , statistics(runtime,[_,DeltaT]), print(delta_time(inv,DeltaT)),nl
     .

state_violates_invariant(I,const_and_vars(ID,S)) :- !,
  add_internal_error('State not expanded: ',state_violates_invariant(I,const_and_vars(ID,S))).
state_violates_invariant(I,expanded_const_and_vars(ID,S,FullState)) :- !,
  add_internal_error('State not expanded: ',state_violates_invariant(I,expanded_const_and_vars(ID,S,FullState))),
  state_violates_invariant(ID,FullState).
state_violates_invariant(ID,State) :- State \== root, State \= concrete_constants(_),
   \+(state_satisfies_invariant_aux(ID,State)).


% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling_naming(b_interpreter:state_violates_assertions_aux/2,name2).

%name2(b_interpreter:state_violates_assertions_aux,_,ASSERTIONS).

%state_violates_assertions(ID,State) :-
%   b_interpreter:state_violates_assertions_aux(ID,State).

% old profiler
:- use_module(runtime_profiler,[profile_single_call/3]).
:- use_module(tools_meta,[safe_time_out_or_virtual_time_out/3]).
:- use_module(bmachine,[get_assertions_from_machine/2, b_machine_has_dynamic_assertions/0, b_machine_has_static_assertions/0]).
state_violates_assertions(ID,State) :-
   profile_single_call('ASSERTIONS',ID,b_interpreter:state_violates_assertions_aux(ID,State)).
state_violates_assertions_aux(ID,State) :-
    b_machine_has_dynamic_assertions,
    state_corresponds_to_initialised_b_machine(State,BState), !,
    \+(state_satisfies_assertions(ID,dynamic,BState)).
state_violates_assertions_aux(ID,State) :-
    b_machine_has_static_assertions,
    state_corresponds_to_set_up_constants_only(State,BState),
    \+(state_satisfies_assertions(ID,static,BState)).
state_satisfies_assertions(ID,Type,State) :-
   ( get_assertions_from_machine(Type,Ass) ->
       set_error_context(checking_assertions),
       preferences:get_computed_preference(debug_time_out,DebugTimeOut),
       safe_time_out_or_virtual_time_out(b_test_list_of_boolean_expression(Ass,[],State),
            DebugTimeOut,TimeOutRes),
       clear_error_context,
       (TimeOutRes = time_out
       ->  nl,print_message('**** TIMEOUT during assertion checking !'),nl,
           print_message(node_id(ID)),nl,
           assert_time_out_for_assertions(ID)
       ; true %debug_println(10,assertions_ok)
      )
   ; true
   ).



/* ----------------------------------*/
/*     b_test_boolean_expression     */
/* ----------------------------------*/


:- type variable_id_list +--> list(type(variable_id)).

:- type boolean_expression +--> call(bsyntaxtree:check_if_typed_predicate).


b_test_list_of_boolean_expression(List,LS,S) :-
      init_wait_flags(WF),
      b_test_list_of_boolean_expression_wf(List,LS,S,WF),
      ground_wait_flags(WF).

b_test_list_of_boolean_expression_wf([],_,_,_WF).
b_test_list_of_boolean_expression_wf([B1|T],LS,S,WF) :-
   ~~pp_cll(b_test_boolean_expression(B1,LS,S,WF)),
   b_test_list_of_boolean_expression_wf(T,LS,S,WF).


:- use_module(b_interpreter_components).


:- assert_pre(b_interpreter:b_test_boolean_expression_wf(E,LS,S),
	      (bsyntaxtree:check_if_typed_predicate(E),type_check(LS,store),type_check(S,store) )).
:- assert_post(b_interpreter:b_test_boolean_expression_wf(_,_,_), true ).

%:- dynamic concount/1.
%inc_cc(N) :- ( concount(OldCount) -> retractall(concount(OldCount)), NewCount is OldCount + N; NewCount = N), assert(concount(NewCount)).




% --------------------- BOOLEAN EXPRESSIONS ----------------------------

:- use_module(b_interpreter_check).
:- use_module(kernel_propagation,  [do_not_enumerate_binary_boolean_operator/3]).

b_test_boolean_expression_wf(E,LS,S) :-
      init_wait_flags(WF),
      % TO DO: call b_trace_test_components ?
      b_test_boolean_expression(E,LS,S,WF),
      %(debug_mode(on) -> portray_waitflags(WF) ; true),
      %print('GROUNDING DET'),nl,trace,
      %ground_det_wait_flag(WF), nl,print('DET'),nl,
      %portray_waitflags(WF),nl,
      ground_wait_flags(WF).
b_test_boolean_expression(b(Expr,_,Infos),LS,S,WF) :- !,
  (b_interpreter_check:composed(Expr) -> empty_avl(Ai)
    ; Ai = no_avl), % simple expression: no sharing is possible: no need to register expressions
	b_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,_).
b_test_boolean_expression(E,LS,S,WF) :- % will generate error message
      empty_avl(Ai), b_test_boolean_expression(E,LS,S,WF,Ai,_).

:- assert_pre(b_interpreter:b_test_boolean_expression(E,LS,S,WF),
	      (nonvar(E),bsyntaxtree:check_if_typed_predicate(E),type_check(LS,store),type_check(S,store),
		  type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_test_boolean_expression(_,_,_,WF), type_check(WF,wait_flag)).


%b_test_boolean_expression(Expr,_,_,_) :- print('test: '),translate:print_bexpr(Expr),nl,print(Expr),nl,fail. %%

:- if(debug:global_debug_flag).
b_test_boolean_expressiond(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	b_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao).
b_test_boolean_expression(Expr,LS,S,WF,Ai,Ao) :- !,
	(waitflag0_is_set(WF)
	 -> debug:watch(300,b_interpreter:b_test_boolean_expressiond(Expr,LS,S,WF,Ai,Ao))
	 ;  debug:watch_det(300,b_interpreter:b_test_boolean_expressiond(Expr,LS,S,WF,Ai,Ao)) ).
:- else.
b_test_boolean_expression(b(Expr,Type,Infos),LS,S,WF,Ai,Ao) :- !,
    (preference(smt_supported_interpreter,true)
     -> b_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao),
        get_wait_flag1(smt_call,WF,BeforeEnumWF),
        gensym:gensym(smt_assertion_name,Symbol),
        smt_solvers_interface:smt_add_predicate(BeforeEnumWF,b(Expr,Type,Infos),LS,S,Symbol)
     ;  b_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao)).
:- endif.
b_test_boolean_expression(E,LS,S,WF,Ai,Ao) :-
  add_internal_error('Boolean expression not properly wrapped: ',
                     b_test_boolean_expression(E,LS,S,WF,Ai,Ao)),
  b_test_boolean_expression2(E,[],LS,S,WF,Ai,Ao).


b_test_boolean_expression2(truth,_,_,_,_WF,Ai,Ao) :- !,Ai=Ao.
b_test_boolean_expression2(falsity,_,_,_,_WF,_Ai,_Ao) :- !,fail.
b_test_boolean_expression2(negation(BExpr),_,LocalState,State,WF,Ai,Ao) :- !,
        b_not_test_boolean_expression(BExpr,LocalState,State,WF,Ai,Ao).
%b_test_boolean_expression2(conjunct(LHS,RHS),[try_smt|Info],LocalState,State,WF,Ai,Ao) :-
%       print('TRY SMT for Conjunct'),nl,
%       b_check_boolean_expression(b(conjunct(LHS,RHS),pred,[try_smt|Info]),LocalState,State,WF,PredRes,Ai,Ao),!,
%       print('SMT Check of Conjunct: '),nl, translate:print_bexpr(LHS),nl, translate:print_bexpr(RHS),nl,
%       PredRes=pred_true.
b_test_boolean_expression2(conjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
%%   print('check lhs of conjunct: '), translate:print_bexpr(LHS),nl,
        b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
%%   print('check rhs of conjunct: '), translate:print_bexpr(RHS),nl,
        b_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_test_boolean_expression2(implication(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :-
    %b_compiler:b_compile(LHS,[],LocalState,State,CLHS), % this slows down Cruise benchmark from 3.75 s to 4.3
    !,
    b_test_implication(LHS,RHS,LocalState,State,WF,Ai,Ao).
b_test_boolean_expression2(equivalence(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    b_test_equivalence(LHS,RHS,LocalState,State,WF,Ai,Ao).
b_test_boolean_expression2(disjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    % print(disjunct),nl, print_bexpr(LHS),nl, print_bexpr(RHS),nl,
   % TO DO: normalise disjunct first ?
   copy_wf_start(WF,CWF),
   if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii),
      % we can check lhs
       if(b_interpreter_check:b_check_boolean_expression0(pred_false,PredRes,RHS,LocalState,State,CWF,RR,Aii,Ao),
            ( /* we can also check rhs */
             b_interpreter_check:disjoin(PredRes,RR,pred_true,LHS,RHS,WF),
             copy_wf_finish(WF,CWF)
            ),
           (
            Ao=Aii, % print(disjunct_rhs_cannot_be_checked),nl, print_bexpr(LHS),nl, print_bexpr(RHS),nl,
            copy_wf_finish(WF,CWF),
            imply_test_boolean_expression(PredRes,pred_false,RHS,LocalState,State,WF,Aii),
            %,print(predres(PredRes)),nl,nl,
            enumerate_bool(PredRes,WF)
           )
         ),
      % we cannot check lhs
      if((always_well_defined(RHS), %print(check_rhs(RHS)),nl,trace,
          b_check_boolean_expression(RHS,LocalState,State,CWF,PredRes,Ai,Aii)),
           (
            Ao=Aii,   %print(disjunct_lhs_cannot_be_checked),nl, print_bexpr(LHS),nl, print_bexpr(RHS),nl,
            copy_wf_finish(WF,CWF),
            imply_test_boolean_expression(PredRes,pred_false,LHS,LocalState,State,WF,Aii),
            %,print(predres(PredRes)),nl,nl,
            enumerate_bool(PredRes,WF)
           ),
          % we cannot check neither lhs nor rhs:
		  (Ai=Ao,
		   get_priority_of_boolean_expression2(disjunct(LHS,RHS),StartPrio),
		   %get_wait_flag(Prio,disjunct,WF,WF2),
		   get_binary_choice_wait_flag_exp_backoff(StartPrio,disjunct,WF,WF2),
		  % be sure that e.g. for (x=2 or x=3) both branches will be tested once deterministic propagation is over
		  %print(prio(Prio)), print(' : '),print_bexpr(LHS), print(' or '), print_bexpr(RHS),nl,
		  b_test_disjunction(LHS,RHS,LocalState,State,WF,WF2,Ai)
		  )
	  )
    ).
b_test_boolean_expression2(exists(Parameters,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
        b_test_exists(Parameters,RHS,Infos,LocalState,State,WF).
%b_test_boolean_expression2(forall(PARS,LHS,RHS),_Inf,LocalState,State,WF) :-
%   PARS = [b(identifier(R1),integer,_),b(identifier(R2),integer,_)],
%   LHS = b(conjunct(b(conjunct(MEM1,MEM2),pred,_I1),NEQ),pred,_I2),
%   NEQ = b(not_equal(b(identifier(RR1),integer,_I3),b(identifier(RR2),integer,_I4)),pred,_I5),
%   (RR1=R1, RR2=R2 ; RR1=R2, RR2=R1),
%   MEM1 = b(member(b(identifier(RRR1),integer,_),b(SET,set(integer),_)),pred,_),
%   MEM2 = b(member(b(identifier(RRR2),integer,_),b(SET,set(integer),_)),pred,_),
%   (RRR1=R1, RRR2=R2 ; RRR1=R2, RRR2=R1),
%   /* TO DO: check that RHS is symmetric ! */  /* TO DO : generalize symmetry breaking */
%   !, %   print(symmetry_breaking_forall(R1,R2)),nl,
%   GT = b(less(b(identifier(R1),integer,_I3),b(identifier(R2),integer,_I4)),pred,_I5),
%   NewLHS = b(conjunct(b(conjunct(MEM1,MEM2),pred,_I1),GT),pred,_I2),
%   b_test_boolean_expression2(forall(PARS,NewLHS,RHS),_Inf,LocalState,State,WF).
b_test_boolean_expression2(forall(Parameters,LHS,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
                             /* !ID.(LHS => RHS) */
       % print_message(forall(Parameters,LHS,RHS)),
      %  get_wait_flag0(WF,WF0),  when(nonvar(WF0),
         b_for_all(Parameters,Infos,LHS,RHS,LocalState,State,WF). %).
b_test_boolean_expression2(let_predicate(Ids,AssignmentExprs,Pred),_,LocalState,State,WF,Ai,Ao) :- !,Ai=Ao,
	set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_test_boolean_expression(Pred,LetState,State,WF). % we could pass Ai in if Ids fresh

b_test_boolean_expression2(lazy_let_pred(Id,AssignmentExpr,Expr),_I,LocalState,State,WF,Ai,Ao) :- !,
	add_lazy_let_id_and_expression(Id,AssignmentExpr,LocalState,State,NewLocalState,WF,Ai),
	%when(ground(IdValue),(print(computed(Id)),nl, translate:print_bvalue(IdValue),nl)),
	b_test_boolean_expression(Expr,NewLocalState,State,WF,Ai,Ao). % we could use Ai,Ao version as lazy lets never interfere with each other and never invalidate existing expressions
b_test_boolean_expression2(lazy_lookup_pred(Id),_Info,LocalState,_State,_WF,Ai,Ao) :- !, Ai=Ao,
        %print(lazy_lookup_pred(Id,LocalState)),nl,
        lookup_value_for_existing_id(Id,LocalState,(Trigger,Value)), % should normally only occur in LocalState; value introduced by lazy_let
        (Trigger,Value) = (pred_true,pred_true). % force evaluation
b_test_boolean_expression2(value(PredVal),_Info,_LocalState,_State,_WF,Ai,Ao) :- !,   % this can occur when lazy_lookup_pred gets compiled by b_compiler
   is_pred_true(PredVal),Ai=Ao.
b_test_boolean_expression2(Pred,Info,LocalState,State,WF,Ai,Ao) :-
  % TO DO: we could put some of the operators into the Ai store (comparison operators)
   b_interpreter_check:register_predicate(Pred,Info,pred_true,Reused,Ai,Ao),
   %portray_avl(Ai),nl,
   (Reused==true
    ->   %% print('REUSED: '), translate:print_bexpr(Pred),nl
         true
    ;    b_test_atomic_boolean_expression2(Pred,Info,LocalState,State,WF)).

:- block is_pred_true(-).
is_pred_true(X) :- (X=pred_true -> true ; X=pred_false -> fail ; add_internal_error('Illegal call: ',is_pred_true(X)),fail).
:- block is_pred_false(-).
is_pred_false(X) :- (X=pred_false -> true ; X=pred_true -> fail ; add_internal_error('Illegal call: ',is_pred_false(X)),fail).

%:- block is_pred_true(-).
%is_pred_true(X) :- (X=pred_true -> true ; X=pred_false -> fail ; add_internal_error('Illegal call: ',is_pred_true(X)),fail).
%:- block is_pred_false(-).
%is_pred_false(X) :- (X=pred_false -> true ; X=pred_true -> fail ; add_internal_error('Illegal call: ',is_pred_false(X)),fail).

b_test_implication(LHS,RHS,LocalState,State,WF,Ai,Ao) :-
   copy_wf_start(WF,CWF),
   if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii),
  %% print('checking => '), print_bexpr(LHS), print(' '), print(PredRes),nl,
  %% The following improves constraint propagation; but can be tricky with undefinedness
	   if((var(PredRes),always_well_defined(RHS),
		   b_check_boolean_expression(RHS,LocalState,State,CWF,RHSPredRes,Aii,Ao)),
		  (/* also propagate from right to left */
		  % print('checking inverse => '), print_bexpr(RHS), print(' '), print(RHSPredRes),nl,
		   b_interpreter_check:imply_true(PredRes,RHSPredRes),
		   copy_wf_finish(WF,CWF)),
		 (Aii=Ao,
		  copy_wf_finish(WF,CWF),
		  imply_test_boolean_expression(PredRes,pred_true,RHS,LocalState,State,WF,Aii))
	   ),
	   (Ai=Ao,
	    %nl,print('*** CANNOT CHECK BOOLEAN EXPRESSION: '), translate:print_bexpr(LHS),nl,
	     perfmessagecall(cannot_reify_lhs_of_implication,translate:print_bexpr(LHS),LHS),
	    %add_error(cannot_check,'LHS cannot be checked',LHS),
		get_binary_choice_wait_flag(implication,WF,WF2),
		block_test_implication(LHS,RHS,LocalState,State,WF,WF2))
	).

% ------------

:- discontiguous b_test_atomic_boolean_expression2/5.
%b_test_atomic_boolean_expression2(Expr,T,LocalState,State,WF) :- print(eval(Expr,T)),nl,fail.
b_test_atomic_boolean_expression2(partition(LHS,RHS),_,LocalState,State,WF) :- !,
    b_compute_expression(LHS,LocalState,State,SV1,WF), %print(lhs(SV1)),nl,
    l_compute_expression(RHS,LocalState,State,SV2,WF), % RHS is Prolog list of expressions
   %%  print(partition_wf(SV1,SV2,WF)),nl, %%
    kernel_call_predicate(partition_wf(SV1,SV2,WF),kernel_objects,WF,partition(LHS,RHS)).
b_test_atomic_boolean_expression2(member(TElem,TSet),_,LocalState,State,WF) :- !,
	b_test_member_expression(TElem,TSet,LocalState,State,WF).
b_test_atomic_boolean_expression2(not_member(TElem,TSet),_,LocalState,State,WF) :- !,
	get_texpr_expr(TSet,Set),
	b_test_notmember_expression(Set,TElem,TSet,LocalState,State,WF).
b_test_atomic_boolean_expression2(freetype_case(Freetype,Case,Expr),_,LocalState,State,WF) :-
	!,equal_object_wf(Value,freeval(Freetype,Case,_),b_test_atomic_boolean_expression2,WF),
	b_compute_expression(Expr,LocalState,State,Value,WF). % NOT COVERED (16)
b_test_atomic_boolean_expression2(finite(SetExpr),_,LocalState,State,WF) :-
	!, b_compute_expression(SetExpr,LocalState,State,Value,WF),
	kernel_call_predicate(is_finite_set_wf(Value,WF),kernel_objects,WF,finite(SetExpr)).
/* Extension for Kodkod */
b_test_atomic_boolean_expression2(kodkod(ProblemId,_),_,LocalState,State,WF) :-
	!,get_wait_flag1(kodkod_call,WF,WF0),
    kodkod_request(ProblemId,pos,LocalState,State,WF0). % NOT COVERED (17)
%b_test_atomic_boolean_expression2(equal(Arg1,Arg2),_,LocalState,State,WF) :- get_texpr_type(Arg1,integer),!,
%  b_compute_expression(Arg1,LocalState,State,ExprRes,WF),
%  b_compute_expression(Arg2,LocalState,State,ExprRes,WF). /* avoids one extra variable compared to code below + could fail earlier ; but problem as b_compute_expression still uses = in some places */

:- include(b_arithmetic_expressions).
%The following clause improves Nqueens performance by about 50 %; by avoiding intermediate variables
% TO DO: do the same for >, <, >= ...
b_test_atomic_boolean_expression2(equal(Arg1,Arg2),Info,LocalState,State,WF) :- !,
   b_test_atomic_equal_boolean_expression(Arg1,Arg2,Info,LocalState,State,WF).
b_test_atomic_boolean_expression2(not_equal(Arg1,Arg2),Info,LocalState,State,WF) :- !,
   b_test_atomic_not_equal_boolean_expression(Arg1,Arg2,Info,LocalState,State,WF).
b_test_atomic_boolean_expression2(less_equal(Arg1,Arg2),_Info,LocalState,State,WF) :-
   %get_texpr_type(Arg1,integer), % type checker already checks this
   preferences:preference(use_clpfd_solver,true),
    %\+ member(no_clpfd_arith_integer_expression,Info),
    !,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   clpfd_interface:clpfd_leq_expr(CLPFD_Expr1,CLPFD_Expr2).
b_test_atomic_boolean_expression2(greater_equal(Arg1,Arg2),_Info,LocalState,State,WF) :-
   %get_texpr_type(Arg1,integer), % type checker already checks this
   preferences:preference(use_clpfd_solver,true),
    %\+ member(no_clpfd_arith_integer_expression,Info),
    !,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   clpfd_interface:clpfd_leq_expr(CLPFD_Expr2,CLPFD_Expr1).
b_test_atomic_boolean_expression2(less(Arg1,Arg2),_Info,LocalState,State,WF) :-
   %get_texpr_type(Arg1,integer), % type checker already checks this
   preferences:preference(use_clpfd_solver,true),
   % \+ member(no_clpfd_arith_integer_expression,Info),
   !,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   %print(cplfd_lt_expr(CLPFD_Expr1,CLPFD_Expr2)),nl,
   clpfd_interface:clpfd_lt_expr(CLPFD_Expr1,CLPFD_Expr2).
b_test_atomic_boolean_expression2(greater(Arg1,Arg2),_Info,LocalState,State,WF) :-
   %get_texpr_type(Arg1,integer), % type checker already checks this
   preferences:preference(use_clpfd_solver,true),
    %\+ member(no_clpfd_arith_integer_expression,Info),
    !,
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   %print(cplfd_gt_expr(CLPFD_Expr1,CLPFD_Expr2)),nl,
   clpfd_interface:clpfd_lt_expr(CLPFD_Expr2,CLPFD_Expr1).
b_test_atomic_boolean_expression2(external_pred_call(FunName,Args),Info,LocalState,State,WF) :-
      !,
      (do_not_evaluate_args(FunName) -> EvaluatedArgs=[]
        ;  b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF)),
      % print('EXTERNAL PRED CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_predicate(FunName,Args,EvaluatedArgs,LocalState,State,pred_true,Info,WF).
b_test_atomic_boolean_expression2(Expression,Info,LocalState,State,WF) :-
     preferences:preference(use_clpfd_solver,false),
     special_binary_predicate(Expression,Module:Kernel_predicate,Arg1,T1,Arg2,T2),
     (nonmember(contains_wd_condition,Info) -> true ; preference(disprover_mode,true)),
     !,
     % print('Special call: '), print_bexpr(Expression),nl, print(Kernel_predicate),print(' : '),nl,
     %(Arg1==list_of_expressions -> b_compute_expressions ... % no longer needed
     b_compute_expression(Arg1,LocalState,State,SV1,WF),
     b_compute_expression(Arg2,LocalState,State,SV2,WF),
     KernelCall =.. [Kernel_predicate,SV1,T1,SV2,T2,WF],
     kernel_call_predicate(KernelCall,Module,WF,Expression).
b_test_atomic_boolean_expression2(Expression,_,LocalState,State,WF) :-
      functor(Expression,BOP,2),
      kernel_mappings:binary_boolean_operator(BOP,Module:Kernel_predicate,WFEnabled),!,
      %% print(bop(BOP)),nl, %%
      arg(1,Expression,Arg1), arg(2,Expression,Arg2),
      b_compute_expression(Arg1,LocalState,State,SV1,WF),
      %% print_term_summary(bop_arg1(BOP,Arg1,State,SV1)),nl, %%
      (binary_boolean_operator_trivially_true(BOP,SV1)
          -> check_well_defined(Arg2,LocalState,State,WF)
     ; b_compute_expression(Arg2,LocalState,State,SV2,WF),
    %% print_term_summary(bop_arg2(BOP,Arg2,State,SV2)),nl, %%
     %% print(binary_bool_kernel_call(KernelCall)),nl,  %%
    ( WFEnabled==yes ->
       KernelCall =.. [Kernel_predicate,SV1,SV2,WF2],
       (do_not_enumerate_binary_boolean_operator(BOP,Arg1,Arg2) % we assume: this is only true for WFEnabled predicates
        -> add_wait_flag_info(WF,wfx_no_enumeration,WF2)  %,print(no_enum(BOP,Arg1)),nl
         ; WF2=WF)
     ; KernelCall =.. [Kernel_predicate,SV1,SV2], WF2=WF  % note: WF2 still used in kernel_call_predicate
    ),
    kernel_call_predicate(KernelCall,Module,WF2,Expression)).
b_test_atomic_boolean_expression2(E,Info,_LS,_S,_WF) :-
    add_internal_error('Uncovered boolean expression: ',b(E,pred,Info)),
    print_functor(E),nl,
    %print(local_state(LS)),nl, print(state(S)),nl,
    fail.


binary_boolean_operator_trivially_true(subset,X) :- X==[].

check_well_defined(Arg,_LocalState,_State,_WF) :-
   always_well_defined(Arg),!.
check_well_defined(Arg,LocalState,State,WF) :-
   b_compute_expression(Arg,LocalState,State,_,WF).

%:- use_module(specfile,[animation_minor_mode/1]).
:- use_module(b_machine_hierarchy,[abstract_constant/2,concrete_constant/2]).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
   Arg1 = b(_,integer,_), %get_texpr_type(Arg1,integer),
   preferences:preference(use_clpfd_solver,true),!,
   b_test_arith_equal_boolean_expression(Arg1,Arg2,LocalState,State,WF).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,Info,LocalState,State,WF) :-
    member(lambda_result,Info),
    \+ simple_expression(Arg2),
    is_lambda_result_id(Arg1), %get_texpr_id(Arg1,'_lambda_result_'), % ensure we have not renamed it to was_lambda_result ?
    !, % this is an equality for the result of a lambda abstraction; only evaluate at the end
    b_compute_expression(Arg1,LocalState,State,SV1,WF), % this is the lambda result identifier _lambda_result_
    %get_enumeration_finished_wait_flag(WF,EWF),
    %get_wait_flag(10000,lambda_result,WF,EWF),
    % compute deterministic parts of Arg2 in wf0 phase: avoid computing this stuff over again
    %print('DELAY: '), translate:print_bexpr(Arg2),nl,
    kernel_waitflags:copy_wf01e_wait_flags(WF,WF2), b_compute_expression(Arg2,LocalState,State,SV2,WF2),
    get_last_wait_flag(lambda_result,WF,EWF),
    ground_value_check(SV2,G2),
    when((nonvar(G2);nonvar(SV1);nonvar(EWF)),b_lambda_result_equal_boolean_expression(SV1,SV2,WF,EWF,WF2)).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
     (LocalState = [] -> true % only check if we are at the outer level (not inside a quantifier,...) or if we have CSE introduzed lazy_lets:
      ; preferences:preference(use_common_subexpression_elimination,true)
        % TO DO: check that we have only added @nr to local state from lazy_let; ADD INFO FIELD
      ),
     get_texpr_type(Arg1,Type1), is_set_type(Type1,_SetType1),
     % previously we required SetType1 = couple(_,_)
     get_texpr_id(Arg1,ID), % check if ABSTRACT_CONSTANT
     (abstract_constant(ID,_) -> true
      %; b_is_unused_constant(ID) -> ABSTRACT=true % TO DO: check if ID not used somewhere else in Properties ??!!
      %; specfile:animation_minor_mode(eventb),concrete_constant(ID,_),ABSTRACT=false ),  % maybe z or tla minor_mode: disable this check ??
      ),
     kernel_objects:max_cardinality(Type1,MaxCard1),
     (number(MaxCard1) -> MaxCard1>100 ; true), % check that type is large enough to warrant recursive treatment
     % Should we maybe even restrict it to infinite types ??
     % in future: we may do size change analysis, or something like that to check that we have properly defined
     % recursive function
     %print(try_symbolic(ID,ABSTRACT)),nl,
     (debug_mode(on) -> print('Equation defining abstract constant: '), print(ID),nl ; true),
     mark_bexpr_as_symbolic(Arg2,SymArg2),
    !,
     b_compute_expression(Arg1,LocalState,State,SV1,WF), % the identifier
     b_compute_expression(SymArg2,LocalState,State,SV2,WF),
     (definitely_expand_this_explicit_set(SV2) % check for a few patterns which should definitely be expanded; in case user puts a set accidentally into the ABSTRACT_CONSTANTS section
      -> b_expand_compute_comprehension_set(SV2,SV3,WF)
      ; custom_explicit_sets:mark_closure(SV2,['SYMBOLIC'],SV3)), % necessary ?
     kernel_call_predicate_equal_object_optimized(SV1,SV3,WF).
b_test_atomic_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
    b_compute_expression(Arg1,LocalState,State,SV1,WF),
    b_compute_expression(Arg2,LocalState,State,SV2,WF),
    kernel_call_predicate_equal_object_optimized(SV1,SV2,WF).


b_test_atomic_not_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
   get_texpr_type(Arg1,boolean),!,  % special treatment : not_equal_object has no type information
   b_compute_expression(Arg1,LocalState,State,B1,WF),
   b_compute_expression(Arg2,LocalState,State,B2,WF),
   %print(not_eq(B1,B2)),nl,
   bool_pred:negate(B1,B2).
b_test_atomic_not_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
   get_texpr_type(Arg1,integer),
   preferences:preference(use_clpfd_solver,true),
    %\+ member(no_clpfd_arith_integer_expression,Info),
    !,
    % Note: calling b_compute_arith_expression will also instantiate a variable to at least the int(_) skeleton; thereby enabling propagation
    %       this is potentially better than calling the default versions of the predicates, which may wait until the int(_) skeleton is set up before propagation
   b_compute_arith_expression(Arg1,LocalState,State,CLPFD_Expr1,WF),
   b_compute_arith_expression(Arg2,LocalState,State,CLPFD_Expr2,WF),
   clpfd_interface:clpfd_neq_expr(CLPFD_Expr1,CLPFD_Expr2).
b_test_atomic_not_equal_boolean_expression(Arg1,Arg2,_Info,LocalState,State,WF) :-
    b_compute_expression(Arg1,LocalState,State,SV1,WF),
    b_compute_expression(Arg2,LocalState,State,SV2,WF),
    kernel_call_predicate_not_equal_object(SV1,SV2,WF,not_equal(Arg1,Arg2)).


simple_expression(b(A,_,_)) :- simple_expression_aux(A).
simple_expression_aux(boolean_true).
simple_expression_aux(boolean_false).
simple_expression_aux(empty_set).
simple_expression_aux(empty_sequence).
simple_expression_aux(integer(_)).
simple_expression_aux(max_int).
simple_expression_aux(min_int).
simple_expression_aux(string(_)).
simple_expression_aux(value(_)).


is_lambda_result_id(b(identifier('_lambda_result_'),_,_)).
% at the moment we only detect those identifiers; the info field is currently not updated for _was_lambda_result_
% and not available in other treatmens of _lambda_result_ (in b_enumerate)
%is_lambda_result_id(b(identifier(X),_,Info)) :-
%   memberchk(lambda_result,Info).
%   (X=='_lambda_result_' -> (memberchk(lambda_result,Info) -> true ; format('~n*** _lambda_result_ without info: ~w~n~N',[Info])) ; memberchk(lambda_result,Info)).

%:- block b_lambda_result_equal_boolean_expression(-,?,?, -,?). % we now use when above
% only compute lambda result when finished finding domain element, i.e., _EWF is grounded
b_lambda_result_equal_boolean_expression(SV1,SV2,WF,EWF,WF2) :-
    kernel_call_predicate_equal_object_optimized(SV1,SV2,WF),
     % print(' LAMBDA : '), print((SV1,_EWF)),nl,
    % the inner waitflags are for evaluating the lambda_result Expression; unless there is a WD issue this should
    % never fail; we can delay the enumeration as much as possible
    % NOTE: even if SV2 is ground, Arg2 may not be (e.g., if Arg2 = prj2(..,..)(NotGr,Ground))
    blocking_ground_inner_wait_flags(WF2,EWF).    % without when we will ground WF0 which is shared with outer WF-store
:- block blocking_ground_inner_wait_flags(?,-).
blocking_ground_inner_wait_flags(WF2,_) :- kernel_waitflags:ground_inner_wait_flags(WF2).

% TO DO: add avl_set+closure expansion
%b_test_member_expression(Expr,El,_Set,_LocalState,_State,_WF) :-
%        print('test member: '),print_bexpr(Expr),nl,fail.
b_test_member_expression(TElem,b(Set,Type,Info),LocalState,State,WF) :-
        b_test_member_expression2(Set,Type,Info,TElem,LocalState,State,WF).

% new profiler
%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling(b_test_member_expression2/7).

% old profiler
%:- use_module(hit_profiler).

b_test_member_expression2(comprehension_set(Par,Cond),_Type,_Info,El,LocalState,State,WF) :-
    % used to call in next clause below: b_generate_closure_if_necessary(Par,Cond,LocalState,State,SetValue,WF) /* will not expand closure */
/* optimized case for dealing with comprehension_set ; however, does not seem to buy anything over older version with b_generate_closure_if_necessary in test 1079 */
/* new version buys about 6 % for {x| x:1..400000 & x:{n|n >= 9000 & n<= 600000 & n mod 100 = 0}} (Sep 3 2014) */
    set_up_typed_localstate(Par,ValueList,_TypedVals,LocalState,NewLocalState,positive),
    !,
    convert_list_into_pairs(ValueList,ElementVal),
    %print(chk(Par,ElementVal)),nl,
    b_compute_expression(El,LocalState,State,ElementVal,WF),
    b_test_boolean_expression(Cond,NewLocalState,State,WF).
b_test_member_expression2(SymbolicSetConstant,_Type,_Info,El,LocalState,State,WF) :-
        %hit_profiler:add_profile_hit(SymbolicSetConstant),
        cst_in_boolean_type(SymbolicSetConstant,Module:Kernel_predicate),!,  % two cases : string_set and integer_set
        b_compute_expression(El,LocalState,State,ElValue,WF),
	    functor(KernelCall,Kernel_predicate,2),
	    arg(1,KernelCall,ElValue),
	    arg(2,KernelCall,WF),
     %  print_message(cst_in_boolean_type_call(KernelCall)),
        kernel_call_predicate(KernelCall,Module,WF,SymbolicSetConstant).  % should be member(... SymbolicSetConstant)
b_test_member_expression2(Expression,_Type,_Info,El,LocalState,State,WF) :-
        functor(Expression,BOP,2),
        binary_in_boolean_type(BOP,Module:Kernel_predicate),!,
        b_compute_expression(El,LocalState,State,ElValue,WF),
        arg(1,Expression,Arg1),arg(2,Expression,Arg2),
        (binary_in_definitely_true(BOP,ElValue)
         -> % no need to compute arguments; unless required for WD:
            check_well_defined(Arg1,LocalState,State,WF),
            check_well_defined(Arg2,LocalState,State,WF)
         ;  b_compute_expression(Arg1,LocalState,State,SV1,WF),
            b_compute_expression(Arg2,LocalState,State,SV2,WF),
            binary_in_kernel_call(Module,Kernel_predicate,ElValue,SV1,SV2,WF,Expression)
        ).
b_test_member_expression2(Expression,_Type,_Info,El,LocalState,State,WF) :-
        functor(Expression,UnOP,1),
        unary_in_boolean_type(UnOP,Module:Kernel_predicate),!,
        b_compute_expression(El,LocalState,State,ElValue,WF),
        arg(1,Expression,Arg1),
        (unary_in_definitely_true(UnOP,ElValue)
         -> % no need to compute arguments; unless required for WD:
            check_well_defined(Arg1,LocalState,State,WF)
         ;  b_compute_expression(Arg1,LocalState,State,SV1,WF),
            unary_in_kernel_call(Module,Kernel_predicate,ElValue,SV1,WF,Expression)
        ).
b_test_member_expression2(Expression,Type,Info,El,LocalState,State,WF) :-
    b_compute_expression2(Expression,Type,Info,LocalState,State,SetValue,WF),
	(always_well_defined_or_disprover_mode(El)
		  -> SetValue \== [] % if we have the empty set: check_element will always fail
		   ;  true % we could check SetValue \== [] if find_abort_values=false ??
	),
	b_compute_expression(El,LocalState,State,Element,WF),
	% TO DO: for certain things like external_function_call which are not invertible: delay setting up element of check
	(not_invertible(El) % then no sense in setting up member constraint; we cannot used the info anyway; but we could enumerate
	 -> %print(not_inv(El,Element)),nl,
	    % we wait until Element is fully known before doing the check
	    when(ground(Element),kernel_call_predicate_check_element_of_wf(Element,SetValue,WF))
	 ; kernel_call_predicate_check_element_of_wf(Element,SetValue,WF)).

:- use_module(external_functions,[external_fun_can_be_inverted/1]).
not_invertible(b(E,_,_)) :- not_inv_aux(E).
not_inv_aux(external_function_call(FunName,_Args)) :- \+ external_fun_can_be_inverted(FunName).
% TO DO: capture other cases like function applications
% TO DO: also do this in b_interpreter_check ?
% TO DO: We should compute this information in ast_cleanup; e.g., what if we do STRING_TO_INT(s)+1 : SET

% TO DO: >>> 2:{x,v} --> special case for set_extension([b(identifier(x),integer,[nodeid(none)]),b(identifier(v),integer,[nodeid(none)])]) --> 2=x or 2=v --> evaluate [x,v] -> sort and then use list member (or should b_ast_cleanup do this) ?

b_test_notmember_expression(SymbolicSetCst,El,_Set,LocalState,State,WF) :-
        kernel_mappings:cst_not_in_boolean_type(SymbolicSetCst,Module:Kernel_predicate),!,
        b_compute_expression(El,LocalState,State,ElValue,WF),
	    functor(KernelCall,Kernel_predicate,1),
	    arg(1,KernelCall,ElValue),
     %  debug_print(cst_not_in_boolean_type(9,KernelCall)),
        kernel_call_predicate(KernelCall,Module,WF,SymbolicSetCst).
b_test_notmember_expression(Expression,El,_Set,LocalState,State,WF) :-
        functor(Expression,BOP,2),
        kernel_mappings:binary_not_in_boolean_type(BOP,Module:Kernel_predicate),!,
        b_compute_expression(El,LocalState,State,ElValue,WF),
        \+ binary_in_definitely_true(BOP,ElValue),
        arg(1,Expression,Arg1),arg(2,Expression,Arg2),
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
        b_compute_expression(Arg2,LocalState,State,SV2,WF),
      %  KernelCall =.. [Kernel_predicate,ElValue,SV1,SV2,WF],
        kernel_call_predicate3(Kernel_predicate,ElValue,SV1,SV2,Module,WF,Expression).
b_test_notmember_expression(Expression,El,_Set,LocalState,State,WF) :-
        functor(Expression,UnOP,1),
        unary_not_in_boolean_type(UnOP,Module:Kernel_predicate),!,
        b_compute_expression(El,LocalState,State,ElValue,WF),
        \+ unary_in_definitely_true(UnOP,ElValue),
        arg(1,Expression,Arg1),
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
        %KernelCall =.. [Kernel_predicate,ElValue,SV1,WF],
        %kernel_call_predicate(KernelCall,Module,WF,Expression).
        kernel_call_predicate2(Kernel_predicate,ElValue,SV1,Module,WF,Expression).
b_test_notmember_expression(Expr,Elem,Set,LocalState,State,WF) :-
	(Expr=comprehension_set(Par,Cond)
          -> b_generate_closure_if_necessary(Par,Cond,LocalState,State,SetValue,WF) /* will not expand closure */
          ; b_compute_expression(Set,LocalState,State,SetValue,WF)
    ),
    (SetValue == []
      -> /* nothing can be member of empty set */
         check_well_defined(Elem,LocalState,State,WF)
         % we could disable this check if find_abort_values=true
      ;  b_compute_expression(Elem,LocalState,State,ElemValue,WF),
	     kernel_call_predicate(not_element_of_wf(ElemValue,SetValue,WF),kernel_objects,WF,Expr)
	).

:- block block_test_implication(?,?,?,?,?,-).
block_test_implication(LHS,RHS,LocalState,State,WF,_WF2) :-
    %% print(impl(LHS)),nl, %%
	b_test_inner_boolean_expression(LHS,LocalState,State,WF),
	b_test_inner_boolean_expression(RHS,LocalState,State,WF).
block_test_implication(LHS,_RHS,LocalState,State,WF,_WF2) :- %% print(impl_not(LHS)),nl, %%
        b_not_test_inner_boolean_expression(LHS,LocalState,State,WF).


b_test_equivalence(LHS,RHS,LocalState,State,WF,Ai,Ao) :-
  copy_wf_start(WF,CWF),
  if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii),
      (b_check_test_equivalence(PredRes,LHS,RHS,LocalState,State,CWF,Aii,Ao),
       copy_wf_finish(WF,CWF)),
      (% check LHS fails, We could check if RHS can be checked with b_check_boolean_expression
       Ai=Ao, get_binary_choice_wait_flag(equivalence,WF,WF2),
       b_enum_test_equivalence(LHS,RHS,LocalState,State,WF,WF2,Ai))).
b_not_test_equivalence(LHS,RHS,LocalState,State,WF,Ai,Ao) :-
  copy_wf_start(WF,CWF),
  if(b_check_boolean_expression(LHS,LocalState,State,CWF,PredRes,Ai,Aii),
      (bool_pred:negate(PredRes,NegPredRes),
       b_check_test_equivalence(NegPredRes,LHS,RHS,LocalState,State,CWF,Aii,Ao),
       copy_wf_finish(WF,CWF)),
      (% check LHS fails, We could check if RHS can be checked with b_check_boolean_expression
       Ai=Ao, get_binary_choice_wait_flag(not_equivalence,WF,WF2),
       b_enum_not_test_equivalence(LHS,RHS,LocalState,State,WF,WF2,Ai))).
b_check_test_equivalence(PredRes,LHS,RHS,LocalState,State,WF,Aii,Ao) :-
  % print('checking <=> '), print_bexpr(LHS), print(' '), print(PredRes),nl,
  ((var(PredRes),b_check_boolean_expression(RHS,LocalState,State,WF,PredRes2,Aii,Ao))  % TO DO <---- REPLACE CUT
    -> /* also propagate from right to left */
      %print('checking inverse <=> '), print_bexpr(RHS), print(' '), print(PredRes2),nl,
     equiv_bidrectional_test_boolean_expression(PredRes,PredRes2,LHS,RHS,LocalState,State,WF)
   ;  Aii=Ao,
      equiv_test_boolean_expression(PredRes,pred_true,RHS,LocalState,State,WF,Aii)
  ). % should we also add a binary choice flag ?? what about a=2 <=> b/=3 ??

:- block b_enum_test_equivalence(?,?,?,?,?,-,?).
b_enum_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),  % TO DO: we could call b_test_inner_boolean_expression with Ai
	b_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).
b_enum_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
    b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).

%enumerate_bool(X,WF) :- !.
enumerate_bool(X,WF) :- (nonvar(X) -> true /* already set */
       ; get_wait_flag0(WF,WF0),
         enumerate_bool0(X,WF,WF0)).

:- block enumerate_bool0(?,?,-).
enumerate_bool0(X,WF,_) :- (nonvar(X) -> true /* already set */
      ;  %print(enum_bool(X)), translate:print_bexpr(LHS),nl,
         % get_binary_choice_wait_flag(enumerate_bool,WF,WF2),
         get_wait_flag(15,enumerate_bool,WF,WF2), % what priority should we use ?? give time for other enumerators to decide the lhs of the disjunct; actions_cbtc is slower to setup with low priority
         enumerate_bool_aux(X,WF2)).
:- block enumerate_bool_aux(-,-).
%enumerate_bool_aux(P,WF) :- var(P),print(forcing_bool(P,WF)),nl,fail. %
enumerate_bool_aux(pred_true,_).
enumerate_bool_aux(pred_false,_).

:- block b_test_disjunction(?,?,?,?,?,-,?).
b_test_disjunction(LHS,_RHS,LocalState,State,WF,_WF2,Ai) :-
    %% print(disjunct_left),nl, print_bexpr(LHS), print(' or '), print_bexpr(_RHS),nl, %%
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,_).
b_test_disjunction(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
    %% print(disjunct_right),nl, print_bexpr(RHS),nl, %%
	b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).

:- block b_not_test_conjunction(?,?,?,?,?,-,?).
b_not_test_conjunction(LHS,_RHS,LocalState,State,WF,_WF2,Ai) :-
    %%  print(' making_lhs_conjunct_false '), print_bexpr(LHS),nl, %% portray_waitflags(WF),
	b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,_).
b_not_test_conjunction(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
    %%  print(' making_lhs_conjunct_true '), print_bexpr(LHS), print(' ::: '), print_bexpr(RHS),nl, %%
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).

:- block b_enum_not_test_equivalence(?,?,?,?,?,-,?).
b_enum_not_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
	b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),  % we could call b_test_inner_boolean_expression with Ai
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).
b_enum_not_test_equivalence(LHS,RHS,LocalState,State,WF,_WF2,Ai) :-
	b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_test_boolean_expression(RHS,LocalState,State,WF,Aii,_).



/* --------------------------------------*/
/*     b_not_test_boolean_expression     */
/* --------------------------------------*/

b_not_test_list_of_boolean_expression([],_,_,_WF). % :- nl, print(done_not),nl,nl,flush_output.
b_not_test_list_of_boolean_expression([B1|T],LS,S,WF) :-
   %print('not '),translate:print_bexpr(B1),nl,flush_output,
   ~~pp_cll(b_not_test_boolean_expression(B1,LS,S,WF)),
   b_not_test_list_of_boolean_expression(T,LS,S,WF).

:- assert_pre(b_interpreter:b_not_test_boolean_expression_wf(E,LS,S),
                  (bsyntaxtree:check_if_typed_predicate(E),type_check(LS,store),type_check(S,store))).
:- assert_post(b_interpreter:b_not_test_boolean_expression_wf(_,_,_), true).


b_not_test_boolean_expression_wf(E,LS,S) :-
      init_wait_flags(WF),b_not_test_boolean_expression(E,LS,S,WF),
      ground_wait_flags(WF).

b_not_test_boolean_expression(E,LS,S,WF) :-
      empty_avl(Ai), b_not_test_boolean_expression(E,LS,S,WF,Ai,_).

:- assert_pre(b_interpreter:b_not_test_boolean_expression(E,LS,S,WF),
                  (type_check(E,boolean_expression),type_check(LS,store),type_check(S,store), type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_not_test_boolean_expression(_,_,_,_), true).

:- if(debug:global_debug_flag).
b_not_test_boolean_expression(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
	(waitflag0_is_set(WF)
	 -> debug:watch(400,b_interpreter:b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao))
	 ;  debug:watch_det(400,b_interpreter:b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao)) ).
:- else.
b_not_test_boolean_expression(b(Expr,Type,Infos),LS,S,WF,Ai,Ao) :- !,
    (preference(smt_supported_interpreter,true)
     -> b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao),create_negation(b(Expr,Type,Infos),Neg),
        get_wait_flag1(smt_call,WF,BeforeEnumWF),
        gensym:gensym(smt_assertion_name,Symbol),
        smt_solvers_interface:smt_add_predicate(BeforeEnumWF,Neg,LS,S,Symbol)
     ;  b_not_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao)).
:- endif.
b_not_test_boolean_expression(E,LS,S,WF,Ai,Ao) :-
  add_internal_error('Boolean (not) expression not properly wrapped: ',
                     b_not_test_boolean_expression(E,LS,S,WF,Ai,Ao)),
  b_not_test_boolean_expression2(E,[],LS,S,WF,Ai,Ao).

b_not_test_boolean_expression2(truth,_,_,_,_WF,_Ai,_Ao) :- !,fail.
b_not_test_boolean_expression2(falsity,_,_,_,_WF,Ai,Ao) :- !,Ai=Ao. % NOT COVERED (2)
b_not_test_boolean_expression2(negation(BExpr),_,LocalState,State,WF,Ai,Ao) :- !,
        b_test_boolean_expression(BExpr,LocalState,State,WF,Ai,Ao).
% following makes test 349 fail:
%b_not_test_boolean_expression2(conjunct(LHS,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
%   nl, print(' NOT: '),translate:print_bexpr(b(conjunct(LHS,RHS),pred,Infos)),nl,
%   create_negation(LHS,NegLHS), % TO DO: avoid this construction by generalising code for disjunct
%   create_negation(RHS,NegRHS),
%   b_test_boolean_expression2(disjunct(NegLHS,NegRHS),Infos,LocalState,State,WF,Ai,Ao).
b_not_test_boolean_expression2(conjunct(LHS,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
    b(conjunct(LHS,RHS),pred,Infos) = Conj,
    copy_wf_start(WF,CWF),
    if(b_check_boolean_expression(Conj,LocalState,State,CWF,PredRes,Ai,Ao),
        (PredRes=pred_false,
         copy_wf_finish(WF,CWF) %, print(' check not conjunct: '), print_bexpr(Conj),nl
         ),
        if(b_check_boolean_expression(LHS,LocalState,State,CWF,LHSPredRes,Ai,Ao), % some redundant work with call above: TODO: avoid !
          (copy_wf_finish(WF,CWF),
           %print(' CHECK : '),translate:print_bexpr(LHS),nl, translate:print_bexpr(RHS),nl,
           get_last_wait_flag(not_conjunct_rhs,WF,WF2), % before starting to enumerate infinite types: better try both possibilities:
           b_not_test_conjunction_rhs(LHSPredRes,RHS,LocalState,State,WF,Ai,WF2)
          ),
          (Ai=Ao,
           get_priority_of_boolean_expression2(disjunct(LHS,RHS),StartPrio),
           %get_wait_flag(Prio,not_conjunct,WF,WF2),
		   get_binary_choice_wait_flag_exp_backoff(StartPrio,not_conjunct,WF,WF2),
	        %print(not_conjunct),nl, translate:print_bexpr(LHS),nl,translate:print_bexpr(RHS),nl,nl,
	        % TO DO: we could check if LHS or RHS are registered in Ai as pred_true/pred_false ??
	       b_not_test_conjunction(LHS,RHS,LocalState,State,WF,WF2,Ai))
	      )
      ).

b_not_test_boolean_expression2(implication(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    b_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
	b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_not_test_boolean_expression2(equivalence(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
    b_not_test_equivalence(LHS,RHS,LocalState,State,WF,Ai,Ao).
b_not_test_boolean_expression2(disjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
        b_not_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
        b_not_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
b_not_test_boolean_expression2(lazy_let_pred(Id,AssignmentExpr,Expr),_I,LocalState,State,WF,Ai,Ao) :- !,
	add_lazy_let_id_and_expression(Id,AssignmentExpr,LocalState,State,NewLocalState,WF,Ai),
	b_not_test_boolean_expression(Expr,NewLocalState,State,WF,Ai,Ao). % we could use Ai,Ao version as lazy lets never interfere with each other and never invalidate existing expressions
b_not_test_boolean_expression2(lazy_lookup_pred(Id),_Info,LocalState,_State,_WF,Ai,Ao) :- !, Ai=Ao,
        %print(lazy_lookup_pred(Id,LocalState,_State)),nl,
        lookup_value_for_existing_id(Id,LocalState,(Trigger,Value)), % should normally only occur in LocalState; value introduced by lazy_let
        (Trigger,Value) = (pred_true,pred_false). % force evaluation
b_not_test_boolean_expression2(value(PredVal),_Info,_LocalState,_State,_WF,Ai,Ao) :- !,    % this can occur when lazy_lookup_pred gets compiled by b_compiler
   is_pred_false(PredVal),Ai=Ao.
b_not_test_boolean_expression2(Expr,Info,LocalState,State,WF,Ai,Ao) :-
  % TO DO: we could put some of the operators into the Ai store (comparison operators)
   b_not_test_atomic_boolean_expression2(Expr,Info,LocalState,State,WF,Ai,Ao).

% Other boolean expressions, not involving boolean connectives
b_not_test_atomic_boolean_expression2(exists(Parameters,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
                             /* #ID.(RHS) */
        Ai=Ao,
        b_not_test_exists(Parameters,RHS,Infos,LocalState,State,WF).
b_not_test_atomic_boolean_expression2(forall(Parameters,LHS,RHS),Infos,LocalState,State,WF,Ai,Ao) :- !,
                             /* !ID.(LHS => RHS) */
	Ai=Ao,
	safe_create_texpr(negation(RHS),pred,Negation),
	conjunct_predicates([LHS,Negation],Condition),
	% print('neg: '),print_bexpr(Condition),nl,
	 % conjunct_predicates merges the used_ids Info in Condition
        b_test_exists(Parameters,Condition,Infos,LocalState,State,WF).
/* TODO(DP, 2.8.2008): Extension for Z: a let-statement as predicate */
b_not_test_atomic_boolean_expression2(let_predicate(Ids,AssignmentExprs,Pred),_,LocalState,State,WF,Ai,Ao) :-
	!,Ai=Ao,
	set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_not_test_boolean_expression(Pred,LetState,State,WF).  % NOT COVERED (10)
/* TODO(DP, 2.8.2008): */
b_not_test_atomic_boolean_expression2(freetype_case(Freetype,Case,Expr),_,LocalState,State,WF,Ai,Ao) :-
	!,Ai=Ao,b_compute_expression(Expr,LocalState,State,freeval(Freetype,RealCase,_),WF),
	dif(RealCase,Case).  % NOT COVERED (11)
b_not_test_atomic_boolean_expression2(finite(SetExpr),_,LocalState,State,WF,Ai,Ao) :-
	!, Ai=Ao,b_compute_expression(SetExpr,LocalState,State,Value,WF),
	kernel_call_predicate(is_infinite_set_wf(Value,WF),kernel_objects,WF,finite(SetExpr)).
/* Extension for Kodkod */
b_not_test_atomic_boolean_expression2(kodkod(ProblemId,_),_,LocalState,State,WF,Ai,Ao) :-
	!,Ai=Ao,get_wait_flag1(kodkod_negative_call,WF,WF0),
    kodkod_request(ProblemId,neg,LocalState,State,WF0). % NOT COVERED (12)
b_not_test_atomic_boolean_expression2(partition(LHS,RHS),Info,LocalState,State,WF,Ai,Ao) :- !,
   b_interpreter_check:register_predicate(partition(LHS,RHS),Info,pred_false,Reused,Ai,Ao),
   (Reused==true -> true
    ;  %print(not_partition(LHS,RHS)),nl,
       b_compute_expression(LHS,LocalState,State,SV1,WF), %print(lhs(SV1)),nl,
       l_compute_expression(RHS,LocalState,State,SV2,WF), % RHS is Prolog list of expressions
       %print(not_partition_wf(SV1,SV2,WF)),nl,
       kernel_call_predicate(not_partition_wf(SV1,SV2,WF),kernel_objects,WF,partition(LHS,RHS))
    ).
b_not_test_atomic_boolean_expression2(external_pred_call(FunName,Args),Info,LocalState,State,WF,Ai,Ao) :-
      !,  Ai=Ao,
      (do_not_evaluate_args(FunName) -> EvaluatedArgs=[]
        ;  b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF)),
      % print('EXTERNAL PRED CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_predicate(FunName,Args,EvaluatedArgs,LocalState,State,pred_false,Info,WF).
b_not_test_atomic_boolean_expression2(Pred,Info,LocalState,State,WF,Ai,Ao) :-
   b_interpreter_check:register_predicate(Pred,Info,pred_false,Reused,Ai,Ao),
   %portray_avl(Ai),nl,
   (Reused==true
    ->   %% print('REUSED (not): '), translate:print_bexpr(Pred),nl,
         true
    ;   b_not_test_atomic_boolean_expression3(Pred,Info,LocalState,State,WF)).

b_not_test_atomic_boolean_expression3(Expression,Info,LocalState,State,WF) :-
        functor(Expression,BOP,2),
        kernel_mappings:negate_binary_boolean_operator(BOP,NBOP),
        arg(1,Expression,Arg1),arg(2,Expression,Arg2),!,
        NExpression =.. [NBOP,Arg1,Arg2],
      %% print_message(negating(BOP,NBOP)), %%
	   %%create_texpr(NExpression,pred,[generated(negation)],TNExpression),
        b_test_atomic_boolean_expression2(NExpression,Info,LocalState,State,WF).
b_not_test_atomic_boolean_expression3(E,_,_,_,_WF) :-
    add_internal_error('Uncovered boolean expression (not): ',E),
    print_functor(E),nl,fail.


% test RHS of not-conjunction depending on outcome of LHS
:- block b_not_test_conjunction_rhs(-,?,?,?,?,?,-).
b_not_test_conjunction_rhs(pred_false,_RHS,_LocalState,_State,_WF,_Ai,_LWF).
b_not_test_conjunction_rhs(pred_true,RHS,LocalState,State,WF,Ai,_LWF) :-
    b_not_test_boolean_expression(RHS,LocalState,State,WF,Ai,_).

/* -----------------------------*/
/*     b_compute_expression     */
/* -----------------------------*/



:- assert_must_succeed((b_interpreter:b_compute_expression(b(pow_subset(
                      b(identifier(nm),set(integer),[])),set(set(integer)),[]),
                      [], [bind(nm,[])],R,_WF),
                             nonvar(R), custom_explicit_sets:expand_custom_set_to_list(R,[[]]))).
:- assert_must_succeed((b_interpreter:b_compute_expression(b(pow_subset(
                      b(identifier(nm),set(boolean),[])),set(set(boolean)),[]),[],
                                        [bind(nm,[pred_false /* bool_false */])],R,_WF),
                             nonvar(R), custom_explicit_sets:expand_custom_set_to_list(R,ER),
                             kernel_objects:equal_object(ER,[[pred_false /* bool_false */],[]]))).

:- assert_pre(b_interpreter:b_compute_expression(Exp,LS,State,_Val,WF),
                (nonvar(Exp),
		  bsyntaxtree:check_if_typed_expression(Exp),
                  type_check(LS,store),type_check(State,store), type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_compute_expression(_E,_LS,_State,Val,WF),
                 (b_interpreter:value_type(Val), type_check(WF,wait_flag))). %, nonvar(Val)

b_compute_expression_nowf(E,LS,S,R) :-
      %add_message(b_compute_expression,'Called Version wo WFLAGS: ', b_compute_expression(E,LS,S,R)),
      init_wait_flags(WF),
      b_compute_expression(E,LS,S,R,WF),
      ground_wait_flags(WF), % this will add abort errors and fail before enumerating the result:
      init_wait_flags(WF2),
      get_texpr_type(E,Type),
      b_enumerate:b_tighter_enumerate_single_value(R,Type,b_compute_expression_nowf,WF2), %should rarely be necessary
      % a non-ground result only occurs when WD error appears and in this case ground_wait_flags will fail after
      % adding the abort error (and thus no value is returned)
      % exceptions are test 1066 with a complicated set comprehension used in a function application
      ground_wait_flags(WF2).

l_compute_expression([],_LS,_S,[],_WF).
l_compute_expression([H|T],LS,S,[CH|CT],WF) :-
   b_compute_expression(H,LS,S,CH,WF),
   l_compute_expression(T,LS,S,CT,WF).

%b_compute_expression(Expr,_,_,_,_) :-
%      print('expr: '),translate:print_bexpr(X),nl,fail.
:- if(debug:global_debug_flag).
b_compute_expressiond(b(Expr,Type,Info),LS,S,R,WF) :-  !,
    ((ground(Type), Type  \== pred, Type \== subst) -> true
     ; add_error(b_compute_expression,'Expression has illegal type: ',b(Expr,Type,Info))),
     check_value(R,b(Expr,Type,Info)),
     b_compute_expression2(Expr,Type,Info,LS,S,R,WF).
:- block check_value(-,?).
check_value(pred_true,_) :- !. check_value(pred_false,_) :- !.
check_value(string(_),_) :- !.
check_value([],_) :- !.
check_value([H|T],E) :- !, check_value(H,E), check_value(T,E).
check_value((A,B),E) :- !, check_value(A,E), check_value(B,E).
check_value(fd(N,T),E) :- !, (nonvar(T) -> check_fd(N,T,E) ; check_err(fd(N,T),E)).
check_value(global_set(GS),E) :- !,(nonvar(GS) -> true ; check_err(global_set(GS),E)).
check_value(avl_set(AS),E) :- !,((nonvar(AS),AS=node(_,_,_,_,_)) -> true ; check_err(global_set(AS),E)).
check_value(int(X),E) :- !,check_int(X,E).
check_value(closure(P,T,B),E) :- !,((nonvar(P),nonvar(T),nonvar(B)) -> true ; check_err(closure(P,T,B),E)).
check_value(X,E) :- check_err(uncovered(X),E).
% TO DO: add records + more thorough checking of global_set and closures and avl_sets
:- block check_int(-,?).
check_int(X,E) :- (number(X) -> true ; check_err(int(X),E)).
:- block check_fd(-,?,?).
check_fd(X,T,E) :- ((number(X),b_fd_type(T,Low,Up),X>=Low,X=<Up) -> true ; check_err(fd(X,T),E)).

check_err(V,E) :- add_internal_error('Illegal value: ',V:E).
b_compute_expression(E,LS,S,R,WF) :- !,
    (waitflag0_is_set(WF)
	 -> debug:watch(100,b_interpreter:b_compute_expressiond(E,LS,S,R,WF))
	 ;  debug:watch_det(100,b_interpreter:b_compute_expressiond(E,LS,S,R,WF))).
:- else.
b_compute_expression(b(Expr,Type,Info),LS,S,R,WF) :-  !,
    ((ground(Type), Type  \== pred, Type \== subst) -> true
     ; add_internal_error('Expression has illegal type: ',b(Expr,Type,Info))),
     %% check_value(R,b(Expr,Type,Info)), %% comment in to check values at runtime <---
     b_compute_expression2(Expr,Type,Info,LS,S,R,WF).
:- endif.
b_compute_expression(Expr,LS,S,R,WF) :-
  add_internal_error('Expression not properly wrapped: ',b_compute_expression(Expr,LS,S,R,WF)),
  b_compute_expression2(Expr,unknown,[],LS,S,R,WF).


:- use_module(bsyntaxtree,[get_texpr_set_type/2]).

%:- use_module(b_global_sets,[b_type2_set/2]).
%b_compute_expression2(V,T,I,LS,S,R,_WF) :- nonvar(R), print('result_instantiated: '(R:V)),nl,fail.
b_compute_expression2(value(Val),_T,_I,_LS,_S,V,WF) :- !, equal_object_wf(Val,V,WF). %Val=V.
b_compute_expression2(boolean_true,_T,_I,_LState,_State,Res,_WF) :- !, Res = pred_true /* bool_true */. % for simple types we can use ordinary unification
b_compute_expression2(boolean_false,_T,_I,_LState,_State,Res,_WF) :- !, Res = pred_false /* bool_false */.
b_compute_expression2(integer_set(NSET),_T,_I,_LState,_State,Res,WF) :- !, equal_object_wf(Res,global_set(NSET),WF). %Res = global_set(NSET).
b_compute_expression2(bool_set,_T,_I,_LState,_State,Res,WF) :- !, % b_type2_set(bool,Res).
    equal_object_wf(Res,avl_set(node(pred_false,true,1,empty,node(pred_true,true,0,empty,empty))),WF).
 	% was equal_object(Res,[pred_true /* bool_true */,pred_false /* bool_false */]) % changing this to avl used to break test 800 Bosch Deadlock v9; probably because of identity closure detection instantiation
b_compute_expression2(string_set,_T,_I,_LState,_State,Res,WF) :- !, equal_object_wf(Res,global_set('STRING'),WF).
	% was all_strings(Res).  %% NOT COVERED (6)
b_compute_expression2(string(S),_T,_I,_LState,_State,Res,_WF) :- !,
     % TODO(ML,26.8.2008): Check if we should use a different functor for syntax tree vs data
     Res = string(S).
b_compute_expression2(empty_set,_T,_I,_LState,_State,Res,_WF) :- !, empty_set(Res).
b_compute_expression2(empty_sequence,_T,_I,_LState,_State,Res,_WF) :- !, empty_sequence(Res).
b_compute_expression2(event_b_identity,EType,_I,_LState,_State,Res,WF) :-
    get_set_type(EType,couple(Type,Type)),!,
    event_b_identity_for_type(Type,Res,WF).
b_compute_expression2(integer(Val),_T,_I,_LocalState,_State,Res,_WF) :- !, Res = int(Val).
b_compute_expression2(max_int,_T,_I,_LS,_S,Res,_WF) :-
	preferences:get_preference(maxint,CVal),!, Res = int(CVal).
b_compute_expression2(min_int,_T,_I,_LS,_S,Res,_WF) :-
	preferences:get_preference(minint,CVal),!, Res = int(CVal).
b_compute_expression2(identifier(Id),_Type,Info,LocalState,State,Res,WF) :- !,
        lookup_value_in_store_and_global_sets(Id,LocalState,State,Value,Info),
        equal_object_wf(Res,Value,WF).
b_compute_expression2(lazy_lookup_expr(Id),_Type,_Info,LocalState,_State,Res,WF) :- !,
        %print(lazy_lookup_expr(Id,LocalState,_State)),nl,
        lookup_value_for_existing_id(Id,LocalState,(Trigger,Value)), % should normally only occur in LocalState; value introduced by lazy_let
        Trigger = pred_true, % force evaluation
        equal_object_wf(Res,Value,WF).
b_compute_expression2(lazy_let_expr(Id,AssignmentExpr,Expr),_T,_I,LocalState,State,Value,WF) :-
	!,
	add_lazy_let_id_and_expression(Id,AssignmentExpr,LocalState,State,NewLocalState,WF),
	%Trigger = pred_true, % expressions cannot be delayed !? (TO DO: more refined check)
	%block_lazy_compute_expression(Trigger,_,AssignmentExpr,LocalState,State,IdValue,WF,Ai),
	%when(ground(IdValue),(print(computed(Id)),nl, translate:print_bvalue(IdValue),nl)),
	b_compute_expression(Expr,NewLocalState,State,Value,WF).
b_compute_expression2(rec(Fields),_T,_I,LocalState,State,Record,WF) :- !,
        ~~mnf(l_compute_field(Fields,LocalState,State,FieldValues,WF)),
        construct_record_wf(FieldValues,Record,WF).
b_compute_expression2(record_field(RecEx,Name),_T,_I,LocalState,State,Value,WF) :- !,
        b_compute_expression(RecEx,LocalState,State,RecValue,WF),
        access_record(RecValue,Name,Value).
b_compute_expression2(freetype_set(Id),_T,_I,_LState,_State,Val,_WF) :- !, Val=freetype(Id). %% NOT COVERED (17)
b_compute_expression2(function(Function,FunArg),Type,Info,LocalState,State,Res,WF) :- !,
       b_compute_expression_function(Function,FunArg,Type,Info,LocalState,State,Res,WF).
b_compute_expression2(card(b(Arg1,Type,Info)),integer,_I,LocalState,State,Card,WF) :- !,
       b_compute_card(Arg1,Type,Info,LocalState,State,Card,WF).
b_compute_expression2(max(b(Arg1,Type,Info)),integer,_I,LocalState,State,Max,WF) :- !,
       b_compute_max(Arg1,Type,Info,LocalState,State,Max,WF).
b_compute_expression2(min(b(Arg1,Type,Info)),integer,_I,LocalState,State,Min,WF) :- !,
       b_compute_min(Arg1,Type,Info,LocalState,State,Min,WF).
b_compute_expression2(set_extension(Ex),Type,_I,LocalState,State,Res,WF) :- !,
       % print_bexpr(b(set_extension(Ex),none,[])),nl,
        b_compute_expressions(Ex,LocalState,State,ExValue,WF),
        % print_term_summary(b_compute_expressions(Ex,LocalState,State,ExValue,WF)),
        kernel_call(b_interpreter:convert_list_of_expressions_into_set_wf(ExValue,ValueSet,Type,WF),ExValue,WF, set_extension(Ex)),
        equal_object_wf(Res,ValueSet,WF).
        %kernel_call_convert_list_of_expressions_into_set_wf(ExValue,ValueSet,Type, WF).
%b_compute_expression('ListTermes'(ListOfEx),LocalState,State,ValueSet,WF) :- !,
%        /* convert list of expressions of CASE statement into  set  of elements */
%        b_compute_expression(ListOfEx,LocalState,State,Value,WF),
%        convert_list_of_expressions_into_set(Value,ValueSet).
b_compute_expression2(sequence_extension(Ex),_T,_I,LocalState,State,ValueSeq,WF) :- !,
        b_compute_expressions(Ex,LocalState,State,Value,WF),
        kernel_call(b_interpreter:convert_list_of_expressions_into_sequence(Value,ValueSeq),Value,WF,sequence_extension(Ex)).
b_compute_expression2(convert_bool(PRED),_T,_I,LocalState,State,Val,WF) :- !, /* B bool(.) operator */
        b_convert_bool(PRED,LocalState,State,WF,Val).
b_compute_expression2(couple(El1,El2),_T,_I,LocalState,State,Result,WF) :- !, Result = (Val1,Val2),
        b_compute_expression(El1,LocalState,State,Val1,WF),
        b_compute_expression(El2,LocalState,State,Val2,WF).
b_compute_expression2(comprehension_set(Parameters,Condition),_T,Info,LocalState,State,Result,WF) :- !,
        b_compute_comprehension_set(Parameters,Condition,Info,LocalState,State,Result,WF).
b_compute_expression2(recursive_let(TId,TSet),_T,_I,LocalState,State,Result,WF) :- !,
    % LET TId = TSet in TSet
    ( get_texpr_expr(TSet,comprehension_set(Parameters,Condition)) ->
        true                    % if the argument TSet is a comprehension set,
                                % we just add a recursive parameter
    ; get_texpr_set_type(TSet,Type) ->
                                % if not, we construct a new comprehension set {x | x:TSet}
        unique_typed_id("_tmparg.",Type,TArg),
        Parameters=[TArg],
        safe_create_texpr(member(TArg,TSet),pred,Condition)
    ; otherwise ->
        add_internal_error('Expected set as argument to recursive_let',
                           b_compute_expression2(recursive_let(TId,TSet))),fail
    ),
    % Generate closure where the references to the recursion are kept
    % add recursion ID to parameters to prevent removal of references during compilation
    generate_recursive_closure(TId,recursion,[TId|Parameters],Condition,LocalState,State,Closure1,WF),
    % remove recursion ID from parameters after compilation
    Closure1=closure([_|P],[_|T],Cond),RClosure=closure(P,T,Cond),
    % Generate closure where the references to the recursion are kept
    generate_recursive_closure(TId,RClosure,Parameters,Condition,LocalState,State,Result,WF).
b_compute_expression2(general_sum(Ids,Condition,Expression),_T,_I,LocalState,State,SumResult,WF) :- !,
       b_general_sum_or_mul(Ids,Condition,Expression,LocalState,State,SumResult,WF,sum).
b_compute_expression2(general_product(Ids,Condition,Expression),_T,_I,LocalState,State,MulResult,WF) :- !,
       b_general_sum_or_mul(Ids,Condition,Expression,LocalState,State,MulResult,WF,mul).
/* Begin: Extensions for Z */
b_compute_expression2(if_then_else(If,Then,Else),T,_I,LocalState,State,Value,WF) :- !,
    b_try_check_boolean_expression_wf(If,LocalState,State,WF,PredRes),
    ((var(PredRes), can_get_full_fd_value(T), preferences:preference(use_clpfd_solver,true),
      always_well_defined_or_disprover_mode(Then), always_well_defined_or_disprover_mode(Else)
     )
      -> b_clpfd_if_then_else(PredRes,T,Then,Else,LocalState,State,Value,WF)
      ;  %(nonvar(PredRes) -> true ; print(normal_if(PredRes,T)),nl, translate:print_bexpr(Then),nl, translate:print_bexpr(Else),nl),
         b_compute_if_then_else(PredRes,Then,Else,LocalState,State,Value,WF)
    ).
b_compute_expression2(let_expression(Ids,AssignmentExprs,Expr),_T,_I,LocalState,State,Value,WF) :-
	!, set_up_localstate_for_let(Ids,AssignmentExprs,LocalState,State,LetState,WF),
	b_compute_expression(Expr,LetState,State,Value,WF). %% NOT COVERED (30)
b_compute_expression2(freetype_constructor(Id,Case,Expr),_T,_I,LocalState,State,FreeValue,WF) :-
	!, FreeValue = freeval(Id,Case,Value),
	b_compute_expression(Expr,LocalState,State,Value,WF). %% NOT COVERED (32)
b_compute_expression2(freetype_destructor(Id,Case,Expr),_T,Info,LocalState,State,Value,WF) :-
	!, b_compute_expression(Expr,LocalState,State,freeval(Id,VCase,VValue),WF),
    check_freetype_case(Id,Case,VCase,VValue,Value,Info,WF).
/* End: Extensions for Z */
b_compute_expression2(assertion_expression(Cond,ErrMsg,Expr),_T,_Info,LocalState,State,Value,WF) :- !,
    b_try_check_boolean_expression_wf(Cond,LocalState,State,WF,PredRes),
     %nl,translate:print_bexpr(Cond),nl,  translate:print_bexpr(Expr),nl, print(predRes(PredRes)),nl,
    ( PredRes==pred_false ->
      add_wd_error_span(ErrMsg,pred_false,span_predicate(Expr,LocalState,State),WF),
      get_last_wait_flag(assertion_expression,WF,WFC),
      % print(wd_error(Expr)),nl,
      when(nonvar(WFC),
            (ground_value(Value) -> true % we have a WD Error, any Value goes
              ;  b_compute_expression(Expr,LocalState,State,Value,WF) % compute Value; avoid infinite enumeration of possible values; useful e.g. for prj1/prj2 assertion expressions
            ))
     ; b_compute_expression(Expr,LocalState,State,Value,WF),
       add_wd_error_if_true(PredRes,ErrMsg,span_predicate(Expr,LocalState,State),WF)
     ). %% NOT COVERED (34)
b_compute_expression2(multiplication(Arg1,Arg2),integer,Info,LocalState,State,Value,WF) :-
        same_texpr(Arg1,Arg2), !,  % is less useful when CLPFD is already turned on, calls square(Arg1,Value,WF)
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
        unary_kernel_call(kernel_objects,square,SV1,Value,WF,multiplication(Arg1,Arg1),integer,Info).
b_compute_expression2(image(Relation,Set),_EType,_Info,LocalState,State,Value,WF) :-
        % print(image(Relation,Set)),nl,
        Relation = b(closure(Rel),_,_CInfo),!, % we have closure1(Rel)[Set] -> avoid computing full closure
        b_compute_expression(Rel,LocalState,State,RV,WF), %print(rel(RV)),nl,
        b_compute_expression(Set,LocalState,State,SV,WF), %print(set(SV)),nl,
        kernel_call(bsets_clp:image_for_closure1_wf(RV,SV,Value,WF),(RV,SV),WF,image(Relation,Set)).
% seems to have issue with things like NATURAL1 \/ {0} = NATURAL -> probably add_element needs improving; fix first and comment back in later
%b_compute_expression2(union(Arg1,Arg2),_Type,Info,LocalState,State,Value,WF) :-
%         !,
%         b_compute_expression(Arg1,LocalState,State,SV1,WF),
%         b_compute_union_expression(SV1,Arg2,Info,LocalState,State,Value,WF).
b_compute_expression2(external_function_call(FunName,Args),EType,Info,LocalState,State,Value,WF) :-
      !,
      b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF),
      /*  (external_function_argument_list(FunName, ArgList)
       -> b_compute_expressions_symbolic(Args, ArgList, LocalState,State, EvaluatedArgs, WF)
       ; b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF)  ), */
      % print('EXTERNAL FUNCTION CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_function(FunName,EvaluatedArgs,Value,EType,Info,WF).
b_compute_expression2(Expression,EType,Info,LocalState,State,Res,WF) :-
       is_set_type(EType,Type),
      /* avoid evaluating certain expensive expressions:
         convert them into symbolic closures and expand only if really required */
        functor(Expression,UnOp,1),
        symbolic_closure_unary_operator(UnOp),
            /* TO DO: UNFOLD THIS OR WRAP in CONSTRUCTOR */
        arg(1,Expression,BType),
        !, %preferences:get_preference(convert_types_into_closures,true),!,
        %  print_message(delayed_un_op(UnOp,BType)),
        b_compute_expression(BType,LocalState,State,ArgValue,WF),
	( do_not_keep_symbolic_unary(UnOp,ArgValue), unary_function(UnOp,Module,KernelFunction) ->
	    /* special treatment for empty set: no need to keep symbolic */
	    unary_kernel_call(Module,KernelFunction,ArgValue,Res,WF,Expression,EType,Info)
	; otherwise ->
	    get_texpr_type(BType,ArgType),
	    create_texpr(value(ArgValue),ArgType,[],TValue),
	    ClosureSetExpression =.. [UnOp,TValue],
	    construct_member_closure('_zzzz_unary',Type,ClosureSetExpression,Value),
	    (unop_to_be_marked_as_symbolic(UnOp,ArgValue) ->
	      mark_closure_as_symbolic(Value,SValue),
	      equal_object_wf(Res,SValue,WF)
	      ; equal_object_wf(Res,Value,WF))

	).
b_compute_expression2(Expression,EType,_Info,LocalState,State,Res,WF) :-
       is_set_type(EType,Type),
      /* avoid evaluating certain expensive expressions:
         convert them into symbolic closures and expand only if really required */
        functor(Expression,BinOp,2),
        arg(1,Expression,BType1),
        arg(2,Expression,BType2),
        kernel_mappings:symbolic_closure_binary_operator(BinOp),
        !,
        %preferences:get_preference(convert_types_into_closures,true),!,
        b_compute_expression(BType1,LocalState,State,ArgValue1,WF),
        (binary_arg1_determines_value(BinOp,ArgValue1,DetResult)
         -> /* print(det(BinOp,ArgValue1)),nl, */
            equal_object_wf(Res,DetResult,b_compute_expression2(BinOp),WF)
         ;  b_compute_expression(BType2,LocalState,State,ArgValue2,WF),
			% print_message(delayed_bin_op(BinOp,ArgValue1,ArgValue2)),
			(is_definitely_empty(BinOp,ArgValue1,ArgValue2)
			  -> %print(empty_symbolic(BinOp,ArgValue1,ArgValue2)),nl,
				 empty_set(Res)
			  ; if(do_not_keep_symbolic(BinOp,ArgValue1,ArgValue2,ERes,WF),
			      % then:
			       %print('Not keeping symbolic: '),print(BinOp), nl,print(ArgValue1),nl, print(ArgValue2),nl,
				   equal_object_wf(Res,ERes,b_compute_expression2(BinOp),WF),
				   % else:
					 (get_texpr_type(BType1,TypeArg1), create_texpr(value(ArgValue1),TypeArg1,[],TArg1),
					 get_texpr_type(BType2,TypeArg2), create_texpr(value(ArgValue2),TypeArg2,[],TArg2),
					 ClosureSetExpression =.. [BinOp,TArg1,TArg2],
					 construct_member_closure('_zzzz_unary',Type,ClosureSetExpression,Value),
					 % print('Constructed Symbolic Closure : '), translate:print_bvalue(Value),nl,
					 (binop_to_be_marked_as_symbolic(BinOp,ArgValue1,ArgValue2) ->
						  mark_closure_as_symbolic(Value,SValue),
						  equal_object_wf(Res,SValue,WF)
						  ; equal_object_wf(Res,Value,WF))
					 )
				) % if
			)
         ).
b_compute_expression2(Expression,Type,Info,LocalState,State,Value,WF) :-
        functor(Expression,Op,1),
        unary_function(Op,Module,KernelFunction),
        arg(1,Expression,Arg1),!,
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
     %  print(unary_fun(Op,SV1)),nl,
        unary_kernel_call(Module,KernelFunction,SV1,Value,WF,Expression,Type,Info).
b_compute_expression2(Expression,Type,Info,LocalState,State,Value,WF) :-
        functor(Expression,Op,2),
        binary_function(Op,Module,KernelFunction),
        arg(1,Expression,Arg1),
        arg(2,Expression,Arg2),!,
        b_compute_expression(Arg1,LocalState,State,SV1,WF),
        (binary_arg1_determines_value(Op,SV1,Result)  % we could disable this if find_abort_values=true or if SV1 not guaranteed to be wd
          -> /* print(det(Op,SV1,Result)),nl, */
             equal_object_wf(Value,Result,b_compute_expression2(Op),WF)
          ;  b_compute_expression(Arg2,LocalState,State,SV2,WF),
             binary_kernel_call(Module,KernelFunction,SV1,SV2,Value,WF,Expression,Type,Info)
        ).
b_compute_expression2(E,T,Info,_L,_S,_R,_WF) :-
   add_internal_error('Uncovered expression: ',b(E,T,Info)),
   print_functor(E), print(' : '), print(T),nl,fail.

% ---


% check if a symbolic_closure_unary_operator result should be marked with prob_annotation(SYMBOLIC)
% because direct sub-args are marked SYMBOLIC
unop_to_be_marked_as_symbolic(UnOp,Value) :- nonvar(Value),
     to_be_marked_as_symbolic_aux(UnOp,Value).
to_be_marked_as_symbolic_aux(struct,rec(Fields)) :- !,
     member(field(_,SymbolicClosure),Fields),
     is_symbolic_closure(SymbolicClosure).
%to_be_marked_as_symbolic_aux(pow_subset,SymbolicClosure) :- is_symbolic_closure(SymbolicClosure).
%to_be_marked_as_symbolic_aux(pow1_subset,SymbolicClosure) :- is_symbolic_closure(SymbolicClosure).
to_be_marked_as_symbolic_aux(_,SymbolicClosure) :- is_symbolic_closure(SymbolicClosure).

% check if a symbolic_closure_binary_operator result should be marked with prob_annotation(SYMBOLIC)
% because direct sub-args are marked SYMBOLIC
binop_to_be_marked_as_symbolic(interval,_,_) :- !, fail.
binop_to_be_marked_as_symbolic(set_subtraction,ArgVal1,_) :- !, is_symbolic_closure(ArgVal1).
% what about intersection ?
binop_to_be_marked_as_symbolic(_,ArgVal1,_) :- is_symbolic_closure(ArgVal1).
binop_to_be_marked_as_symbolic(_,_,ArgVal2) :- is_symbolic_closure(ArgVal2).
% ---

lazy_compute_expression(TRIGGER,AssignmentExpr,LocalState,State,IdValue,WF,Ai) :-
  get_enumeration_finished_wait_flag(WF,F),
  block_lazy_compute_expression(TRIGGER,F,AssignmentExpr,LocalState,State,IdValue,WF,Ai).

:- block block_lazy_compute_expression(-,-,?,?,?, ?,?,?).
block_lazy_compute_expression(TRIGGER,_F,_AssignmentExpr,_LocalState,_State,_IdValue,_WF,_Ai) :-
   var(TRIGGER),!. %,print(unused_let),nl.
block_lazy_compute_expression(pred_true,_,AssignmentExpr,LocalState,State,IdValue,WF,Ai) :-
   % print(computing(LocalState,State)), nl, translate:print_bexpr(AssignmentExpr),nl,
   % get_texpr_type(AssignmentExpr,Type),
   AssignmentExpr = b(Expr,Type,Infos),
   (Type \== pred -> b_compute_expression2(Expr,Type,Infos,LocalState,State,IdValue,WF)
    ; IdValue == pred_true -> b_test_boolean_expression2(Expr,Infos,LocalState,State,WF,Ai,_)
    ; IdValue == pred_false -> b_not_test_boolean_expression2(Expr,Infos,LocalState,State,WF,Ai,_)
    ; b_check_boolean_expression(AssignmentExpr,LocalState,State,WF,PredVal,Ai,_) -> IdValue = PredVal
    ; % reification of check_boolean_expression failed:
      b_try_check_failed(AssignmentExpr,LocalState,State,WF,IdValue,'$GENERATE_ON_DEMAND')
   ).
block_lazy_compute_expression(pred_false,_,_AssignmentExpr,_LocalState,_State,[],_WF,_Ai). % maybe generate any value ??

% ---

:- block add_wd_error_if_true(-,?,?,?).
add_wd_error_if_true(pred_true,_,_,_).
add_wd_error_if_true(pred_false,ErrMsg,Span,WF) :- %print(wd_error(ErrMsg)),nl,
        add_wd_error_span(ErrMsg,pred_false,Span,WF).

% ---


:- use_module(kernel_tools,[ground_value/1]).
b_compute_expression_function(Function,FunArg,Type,Info,LocalState,State,Res,WF) :-
       preferences:preference(find_abort_values,false), % the treatment below will not check, e.g., if the set extension is really a function
       is_extension_function(Function,FInfo,ExFun), !,
      /* a set extension applied to an argument: in some cases we can can first evaluate the argument,
        and then avoid having to compute the set_extension for all non-matching args */
         %  print(is_set_extension_function),nl,  print(set_extension(ExFun)),nl, %print(funarg(FunArg)),nl,
        b_compute_expression(FunArg,LocalState,State,ArgValue,WF),
         %  print(funval(ArgValue)),nl,
        % TO DO : should we only do this if ArgValue is known ?? otherwise the standard treatment is better ??
        %(ground(ArgValue) -> true ; true),
        ((memberchk(contains_wd_condition,FInfo) ;
          preferences:preference(use_clpfd_solver,false) ; ground_value(ArgValue))
         -> %print(apply_function_set_extension(FInfo,ArgValue)),nl,
            b_apply_function_set_extension(ExFun,ArgValue,LocalState,State,Res,WF,Info)
         ;  /* the standard treatment can infer information about the result,... in clpfd_mode */
            %print(non_ground_fun_arg(ArgValue,FInfo,LocalState,State,Res)),nl, translate:print_bexpr(b(function(Function,FunArg),Type,Info)),nl,
            b_compute_expression(Function,LocalState,State,FValue,WF),
            Span = span_predicate(b(function(Function,FunArg),Type,Info),LocalState,State),
            %kernel_call(bsets_clp:apply_to(FValue,ArgValue,Res,Span,WF),(FValue,ArgValue),WF,function(Function,FunArg))).
            get_texpr_type(Function,FunctionType),
            kernel_call_apply_to(FValue,ArgValue,Res,FunctionType,Span,WF)).
b_compute_expression_function(Function,Arg,Type,Info,LocalState,State,Value,WF) :-
     %  print_message(call_function(Function,Arg)),
        b_compute_expression(Function,LocalState,State,FValue,WF),
     %  print_message(function(FValue)),
        b_compute_expression(Arg,LocalState,State,ArgValue,WF),
        Span = span_predicate(b(function(Function,Arg),Type,Info),LocalState,State),
        %kernel_call(bsets_clp:apply_to(FValue,ArgValue,Value,Span,WF),(FValue,ArgValue),WF,function(Function,Arg)).
        get_texpr_type(Function,FunctionType),
        kernel_call_apply_to(FValue,ArgValue,Value,FunctionType,Span,WF).
    %   print_message(kernel_res(Value)).

:- use_module(bsyntaxtree,[get_set_type/2,is_set_type/2]).
b_compute_card(Arg1,Type,Info,LocalState,State,Card,WF) :-
      b_compute_card2(Arg1,Type,Info,LocalState,State,Card,WF),
      % maybe we should move this code below inside kernel_objects ??
      (ground(Card) -> true
        ; get_set_type(Type,SetElType),
          kernel_objects:max_cardinality(SetElType,MaxCard), % this could be precomputed statically
          %print(check_card(Card,MaxCard)),nl,
          (number(MaxCard) -> kernel_objects:in_nat_range(Card,int(0),int(MaxCard))
           ; kernel_objects:greater_than_equal(Card,int(0)) % not sure this is required
          )
      ).
b_compute_card2(set_extension(Ex),_,_,LocalState,State,Card,WF) :- !,
       b_compute_expressions(Ex,LocalState,State,ExValue,WF),
       kernel_objects:cardinality_of_set_extension_list(ExValue,Card,WF).
b_compute_card2(comprehension_set(Parameters,Condition),_Type,Info,LocalState,State,Card,WF) :- !,
       % ensure that we keep the closure and don't expand it:
       b_compute_comprehension_set_symbolic(Parameters,Condition,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:finite_cardinality_as_int(SV1,Card,WF),SV1,WF,card).
% TO DO: treat comprehension set + things like card({x|x>1 & x<2**e & x mod 2 = 0}) = 0
b_compute_card2(Arg1,Type,Info,LocalState,State,Card,WF) :-
       b_compute_expression2(Arg1,Type,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:finite_cardinality_as_int(SV1,Card,WF),SV1,WF,card).
b_compute_max(set_extension(Ex),_,Info,LocalState,State,Max,WF) :- !,
       b_compute_expressions(Ex,LocalState,State,ExValue,WF),
       %print(max_set_extension(ExValue)),nl,
       kernel_objects:maximum_of_set_extension_list(ExValue,Max,Info,WF).
b_compute_max(Arg1,Type,Info,LocalState,State,Max,WF) :-
       b_compute_expression2(Arg1,Type,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:maximum_of_set(SV1,Max,Info,WF),SV1,WF,max).
b_compute_min(set_extension(Ex),_,Info,LocalState,State,Min,WF) :- !,
       b_compute_expressions(Ex,LocalState,State,ExValue,WF),
       kernel_objects:minimum_of_set_extension_list(ExValue,Min,Info,WF).
/* doesnot really buy a lot: */
/*b_compute_min(union(Arg1,Arg2),_,Info,LocalState,State,Min,WF) :- %print(union(Arg1,Arg2)),nl,
       Arg1 = b(set_extension(Ex),Type1,Info1),
       !, %print(start_min_union),nl,
       b_compute_min(set_extension(Ex),Type1,Info1,LocalState,State,Min1,WF), %print(arg1(Min1)),nl,
       b_compute_expression(Arg2,LocalState,State,SV2,WF),
         print(min_union(Min1,SV2)),nl,
       b_compute_min2(SV2,Min1,Min,Info,WF).
:- block b_compute_min2(-,?,?,?,?).
b_compute_min2([],Min1,Res,_,_WF) :- !, Res=Min1.
b_compute_min2(SV1,Min1,Res,Info,WF) :-
       kernel_mappings:must_succ_kernel_call(kernel_objects:minimum_of_set(SV1,Min2,Info,WF),SV1,WF,min),
       min(Min1,Min2,Res).
min(int(A),int(B),int(Res)) :- kernel_objects:minimum(A,B,Res). */
b_compute_min(Arg1,Type,Info,LocalState,State,Min,WF) :-
       b_compute_expression2(Arg1,Type,Info,LocalState,State,SV1,WF),
       kernel_mappings:must_succ_kernel_call(kernel_objects:minimum_of_set(SV1,Min,Info,WF),SV1,WF,min).

generate_recursive_closure(TRecId,RecValue,Parameters,Condition,LocalState,State,Result,WF) :-
    get_texpr_id(TRecId,RecId),
    add_var_to_localstate(RecId,RecValue,LocalState,State1),
    b_generate_closure(Parameters,Condition,State1,State,RClosure1,WF),
    mark_closure_as_recursive(RClosure1,RClosure2),
    mark_closure(RClosure2,[recursive(TRecId)],Result),
    (  debug_mode(off) -> true
     ; ground_value(Result) -> true
     ; otherwise      -> term_variables(Result,TV),add_warning(b_interpreter,'non-ground closure generated: ',TV:Result)
    ).

% check if a freetype value has the expected case, if not
% generate a WD-error
check_freetype_case(Id,Case,VCase,Inner,Value,Info,WF) :-
    get_enumeration_finished_wait_flag(WF,F),
    check_freetype_case2(Id,Case,VCase,Inner,Value,Info,F,WF).
:- block check_freetype_case2(?,?,-,?,?,?,-,?).
check_freetype_case2(Id,Case,VCase,Inner,Value,Info,_F,WF) :-
    ( nonvar(VCase) ->
        ( Case=VCase -> equal_object_wf(Inner,Value,check_freetype_case2,WF)
        ; otherwise ->
            ajoin(['unexpected case for user defined type ',Id,
                   ' (expected ',Case,')'],Msg),
            add_wd_error_set_result(Msg,VCase,Inner,Value,Info,WF))
    ; otherwise ->       % the computation has finished without value,
        true).           % just drop the check

% check if we have a set_extension argument representing an explicit function
% {a |-> f1, b |-> f2, ...}  or  [f1,f2]
is_extension_function(b(A,_,Info),Info,C) :- is_extension_function_aux(A,C).
is_extension_function_aux(set_extension(Ex),CaseList) :- is_set_extension_function(Ex,CaseList).
is_extension_function_aux(sequence_extension(Ex),CaseList) :- is_sequence_extension_function(Ex,1,CaseList).
is_set_extension_function([],[]).
is_set_extension_function([b(Couple,TC,_)|T],[case(A,B)|ST]) :- is_couple(Couple,TC,A,B),
   is_set_extension_function(T,ST).
is_sequence_extension_function([],_,[]).
is_sequence_extension_function([B|T],Nr,[case(b(value(int(Nr)),integer,[]),B)|ST]) :- N1 is Nr+1,
    is_sequence_extension_function(T,N1,ST).

is_couple(couple(A,B),_,A,B).
is_couple(value((VA,VB)),couple(TA,TB), b(value(VA),TA,[]), b(value(VB),TB,[])).

:- use_module(kernel_equality,[equality_objects_wf/4]).
b_apply_function_set_extension([],FunArgVal,LocalState,State,_Res,WF,FunctionPos) :-
    add_wd_error_span('Function applied outside of domain: ',FunArgVal,span_predicate(FunctionPos,LocalState,State),WF).
b_apply_function_set_extension([case(A,B)|T],FunArgVal,LocalState,State,Res,WF,FunctionPos) :-
    b_compute_expression(A,LocalState,State,AVal,WF),
    %print(' --> '),translate:print_bexpr(B),nl,
    equality_objects_wf(AVal,FunArgVal,EqRes,WF),
    % TO DO: if Res is known, we can call equality_objects(BVal,Res,ResEqRes)
    b_apply_function_set_extension_aux(EqRes,B,T,FunArgVal,LocalState,State,Res,WF,FunctionPos).

:- block b_apply_function_set_extension_aux(-,?,?,?, ?,?,?,?,?).
b_apply_function_set_extension_aux(pred_true,B,_T,_FunArgVal,LocalState,State,Res,WF,_FunctionPos) :-
    b_compute_expression(B,LocalState,State,BVal,WF), %print(value(BVal)),nl,
    equal_object_wf(BVal,Res,b_apply_function_set_extension_aux,WF).
b_apply_function_set_extension_aux(pred_false,_B,T,FunArgVal,LocalState,State,Res,WF,FunctionPos) :-
    b_apply_function_set_extension(T,FunArgVal,LocalState,State,Res,WF,FunctionPos).


% --------------------- SIGMA & PI ------------------------
/* first clause: try to do a specific treatment for formulas of the form SIGMA(ID).(ID:SET|EXPR) */
/* Advantage: will not delay as long before computing comprehension_set SET ; see test 1312 */
b_general_sum_or_mul([TID],b(member(b(identifier(ID),_TYPE,_),SET),pred,_),
              Expression,LocalState,State,SumResult,WF,SUMorMUL) :- get_texpr_id(TID,ID),!,
       % nl,print(b_general_sum_or_mul_set(ID,SET,SUMorMUL)),nl, %%
       b_compute_expression(SET,LocalState,State,SETResult,WF),
       b_general_sum_or_mul_over_set([TID],SETResult,Expression,LocalState,State,SumResult,WF,SUMorMUL).
b_general_sum_or_mul(Ids,Condition,Expression,LocalState,State,SumMulResult,WF,SUMorMUL) :-
       b_compute_comprehension_set(Ids,Condition,[],LocalState,State,SETResult,WF),
        %print(set(SETResult)),nl, trace,
        (var(SETResult) -> perfmessagecall(sigma_pi_not_expanded_or_reified(SUMorMUL),
                                                 translate:print_bexpr(Condition),Condition) ; true),
       b_general_sum_or_mul_over_set(Ids,SETResult,Expression,LocalState,State,SumMulResult,WF,SUMorMUL).


:- block b_general_sum_or_mul_over_set(?,-, ?,?,?, ?,?,?).
b_general_sum_or_mul_over_set([TID],SETResult,Expression,_LS,_S,SumResult,WF,SUMorMUL) :-
       get_texpr_id(Expression,ID), % check if we just sum the elements of the list  SIGMA(x).(x:S|x)
       get_texpr_id(TID,ID),
       sum_or_mul_of_explicit_set(SETResult,SUMorMUL,R),!,
        %%  print_term_summary(sum_or_mul_explicit(R,SETResult)),nl, %%
       equal_object_wf(SumResult,R,b_general_sum_or_mul_over_set,WF).
b_general_sum_or_mul_over_set(IDs,SETResult,Expression,LocalState,State,SumResult,WF,SUMorMUL) :-
        % print_term_summary(sum_or_mul(SETResult)),nl,%%
       expand_custom_set_to_list(SETResult,ESETResult,_Done,b_general_sum_or_mul_over_set),
       %print(sum(SETResult,ESETResult)),nl,
        % print(expanded(ESETResult,SumResult)),nl,
       b_sum_or_mul_over_list(ESETResult,IDs,Expression,LocalState,State,SumResult,WF,SUMorMUL).

b_sum_or_mul_over_list(L,IDs,Expr,LS,State,Result,WF,SUMorMUL) :-
   get_acc_base(SUMorMUL,Acc),
   get_wait_flag0(WF,WF0),
   b_sum_or_mul_over_list_acc(L,IDs,Expr,LS,State,Acc,Result,WF,SUMorMUL,WF0).

:- block b_sum_or_mul_over_list_acc(-, ?, ?,?,?,?, ?,?,?,?).
% better if list is fully known: no constraint propagation performed during computation !
% for example: SIGMA(x).(x:{-1} \/ 1..30000|x+1) goes down from 2050 ms to 1210 ms
b_sum_or_mul_over_list_acc([],_,_,_,_,Acc,Result,_WF,_SUMorMUL,_) :- Result=int(Acc).
b_sum_or_mul_over_list_acc([H|T],IDs,Expr,LocalState,State,Acc,Result,WF,SUMorMUL,WF0) :-
   %  print(sum_over_acc(H,IDs,Acc,Result)),nl,
   add_typed_vars_to_localstate(IDs,H,LocalState,NewLocalState),
   %print(added(NewLocalState)),nl,
   b_compute_expression(Expr,NewLocalState,State,HExprVal,WF),
   b_sum_or_mul_over_list_acc2(HExprVal,T,IDs,Expr,LocalState,State,Acc,Result,WF,SUMorMUL,WF0).

:- block b_sum_or_mul_over_list_acc2(?,-, ?, ?,?,?,?, ?,?,?,?),
         b_sum_or_mul_over_list_acc2(?,?, ?, ?,?,?,-, ?,?,?,-),
         b_sum_or_mul_over_list_acc2(-,?, ?, ?,?,?,?, ?,?,?,-). % wait until T nonvar and either Acc & HExprVal known or WF0 set
% purpose: try and remain in non-CLPFD mode as long as possible to avoid overflows,... (see eg. test 1642, 1708)
b_sum_or_mul_over_list_acc2(HExprVal,T,IDs,Expr,LocalState,State,Acc, Result,WF,SUMorMUL,WF0) :-
   (nonvar(T),nonvar(Acc),ground(HExprVal)
     -> HExprVal=int(HVal), % we can continue computing the system without constraint propagation; is faster
        compose_acc(SUMorMUL,Acc,HVal,NewAcc),
        b_sum_or_mul_over_list_acc(T,IDs,Expr,LocalState,State,NewAcc,Result,WF,SUMorMUL,WF0)
    ; T==[],Acc==0 -> /* we have just a single element; no need for clpfd treatment */
        equal_object_wf(HExprVal,Result,b_sum_or_mul_over_list_acc)
    ;   % print(init_clp_acc(HExprVal,T,Acc)),nl,
        init_clp_acc(SUMorMUL,HExprVal,int(Acc),CLPAccumulator),
        b_sum_or_mul_over_list_clp(T,IDs,Expr,LocalState,State,CLPAccumulator,Result,WF,SUMorMUL)
   ).

% add set comprehension solution ValueForTypedIDs as individual values for every TypedIDs to LocalState
add_typed_vars_to_localstate(TypedIDs,ValueForTypedIDs,LocalState,NewLocalState) :-
   (flatten_pairs(TypedIDs,ValueForTypedIDs,ValAsList),
    add_typed_vars_to_localstate_aux(TypedIDs,ValAsList,LocalState,R)
    -> NewLocalState = R
    ; add_internal_error('Call failed: ',add_typed_vars_to_localstate(TypedIDs,ValueForTypedIDs,LocalState,NewLocalState)),
      fail
   ).
% set comprehension results for {a,b,c|P} are nested like this ((a,b),c)
% first argument only used for length (to know how deep the pairs are nested)
flatten_pairs(L,Pairs,FlatList) :- flatten_pairs(L,Pairs,FlatList,[]).
flatten_pairs([_],A) --> !, [A].
flatten_pairs([_|T],(A,B)) --> flatten_pairs(T,A) ,[B].
add_typed_vars_to_localstate_aux([],_,LS,LS).
add_typed_vars_to_localstate_aux([TID],[H],LocalState,NewLocalState) :- !,
   get_texpr_id(TID,ID), get_texpr_type(TID,TYPE),
   add_typed_var_to_localstate(ID,H,TYPE,LocalState,NewLocalState).
add_typed_vars_to_localstate_aux([TID1|Ids],[H1|Hs],LocalState,NewLocalState) :- !,
   get_texpr_id(TID1,ID1),  get_texpr_type(TID1,TYPE),
   add_typed_var_to_localstate(ID1,H1,TYPE,LocalState,LS1),
   add_typed_vars_to_localstate_aux(Ids,Hs,LS1,NewLocalState).


:- block b_sum_or_mul_over_list_clp(-, ?, ?,?,?, ?, ?,?,?).
b_sum_or_mul_over_list_clp([],_, _,_,_, Acc, Result,_WF,SUMorMUL) :- %print(acc(Acc)),nl,
   finalise_clp(SUMorMUL,Acc,Result).
b_sum_or_mul_over_list_clp([H|T],IDs,Expr,LocalState,State,Acc, Result,WF,SUMorMUL) :-
    % print(sum_over(H,IDs,Result)),nl,
   add_typed_vars_to_localstate(IDs,H,LocalState,NewLocalState),
   b_compute_expression(Expr,NewLocalState,State,HExprVal,WF),
   compose_clp(SUMorMUL,HExprVal,Acc,NewAcc),
   b_sum_or_mul_over_list_clp(T,IDs,Expr,LocalState,State,NewAcc,Result,WF,SUMorMUL).

get_acc_base(sum,0).
get_acc_base(mul,1).
compose_acc(sum,A,B,C) :- C is A+B.
compose_acc(mul,A,B,C) :- C is A*B.
init_clp_acc(sum,int(A),int(B),[A,B]).
init_clp_acc(mul,A,B,C) :- times(A,B,C).
%compose_clp(sum,A,B,C) :- int_plus(A,B,C).  % should we use clpfd_sum ??
compose_clp(sum,int(A),Acc,NewAcc) :- NewAcc = [A|Acc].
compose_clp(mul,A,B,C) :- times(A,B,C).
:- use_module(clpfd_interface,[clpfd_sum/2]).
finalise_clp(sum,Acc,int(Result)) :- %print(clpfd_sum(Acc,Result)),nl,
    clpfd_sum(Acc,Result). % using clpfd_sum can be much more efficient: no intermediate variables set up
finalise_clp(mul,int(Acc),int(Acc)).

/* for if-then-else in Z: compute Then or Else,
   depending on the outcome of the predicate */
:- block b_compute_if_then_else(-,?,?,?,?,?,?).
b_compute_if_then_else(pred_true,Then,_Else,LocalState,State,Value,WF) :-
    b_compute_expression(Then,LocalState,State,Value,WF).
b_compute_if_then_else(pred_false,_Then,Else,LocalState,State,Value,WF) :-
    b_compute_expression(Else,LocalState,State,Value,WF).

:- use_module(clpfd_interface,[clpfd_if_then_else/4]).
% a more sophisticated treatment, can also work backwards forcing PredRes
% examples:
% x = IF a<10 THEN 0 ELSE 5 END & x:6..10 & a:1..23
% see must_fail_clpfd_det 132 and 133
b_clpfd_if_then_else(PredRes,Type,Then,Else,LocalState,State,Value,WF) :-
    get_full_fd_value(Type,Value,FDValue),
    b_compute_expression(Then,LocalState,State,ThenValue,WF),
    get_full_fd_value(Type,ThenValue,FDThenValue),
    b_compute_expression(Else,LocalState,State,ElseValue,WF),
    get_full_fd_value(Type,ElseValue,FDElseValue),
    % TO DO: catch CLPFD overflow and use equality_objects
    % print(clpfd_if_then_else(PredRes,ThenValue,FDThenValue,ElseValue,FDElseValue,Value,FDValue)),nl,
    clpfd_if_then_else(PredRes,FDThenValue,FDElseValue,FDValue).

% try and get full_fd_value; true if full value can be translated to CLP(FD) value
get_full_fd_value(global(T),fd(X,T),X).
get_full_fd_value(integer,int(X),X).
get_full_fd_value(boolean,B,FD) :- kernel_equality:prop_eq_01(B,FD).
% record with single field ?

can_get_full_fd_value(global(_)).
can_get_full_fd_value(integer).
can_get_full_fd_value(boolean).

% a slightly improved version over set_up_localstate (not sure it is really much faster):
add_lazy_let_id_to_local_state(b(identifier(ID),_,_),Trigger,IdValue,LocalState,NewState) :-
    NewState = [bind(ID,(Trigger,IdValue))|LocalState].

add_lazy_let_id_and_expression(Id,IdExpr,LocalState,InState,NewLocalState,WF) :-
    empty_avl(Ai),
    add_lazy_let_id_and_expression(Id,IdExpr,LocalState,InState,NewLocalState,WF,Ai).
add_lazy_let_id_and_expression(Id,IdExpr,LocalState,InState,NewLocalState,WF,Ai) :-
  	add_lazy_let_id_to_local_state(Id,Trigger,IdValue,LocalState,NewLocalState),
  	% print(setting_up_lazy_let_subst(Id,NewLocalState,State)),nl,
  	lazy_compute_expression(Trigger,IdExpr,LocalState,InState,IdValue,WF,Ai).

/* for let-statements in Z and now in B: expand localstate by new variables and assign
   values to them ;
   now Expressions are evaluated in state *with* the new identifiers !
*/
%set_up_localstate_for_let_no_reuse(Ids,Exprs,LocalState,State,LetState,WF) :-
%	set_up_localstate(Ids,Vars,LocalState,LetState),
%	compute_let_expressions(Exprs,Vars,LocalState,State,WF).
set_up_localstate_for_let(Ids,Exprs,LocalState,State,LetState,WF) :-
	set_up_localstate(Ids,Vars,LocalState,LetState),
	compute_let_expressions(Exprs,Vars,LetState,State,WF).
compute_let_expressions([],[],_,_,_).
compute_let_expressions([Expr|RestExprs],[Var|RestVars],LocalState,State,WF) :-
	b_compute_expression(Expr,LocalState,State,Value,WF),
	equal_object_optimized(Var,Value,compute_let_expressions),
	compute_let_expressions(RestExprs,RestVars,LocalState,State,WF).

l_compute_field([],_LS,_S,[],_WF).
l_compute_field([Field|T],LS,S,[FVal|CT],WF) :-
   b_compute_field(Field,LS,S,FVal,WF),
   l_compute_field(T,LS,S,CT,WF).

b_compute_field(field(Name,TypeOrVal),LocalState,State,field(Name,TypeOrFieldValue),WF) :-
        b_compute_expression(TypeOrVal,LocalState,State,TypeOrFieldValue,WF).



:- use_module(b_compiler).

/* :- type bsets_closure +--> closure(list(type(variable_id)),
                                   list(type(basic_type_descriptor)),type(boolean_expression)). */

:- assert_pre(b_interpreter:b_compute_comprehension_set(Parameters,Cond,_Info,LS,State,_R,_WF),
     (ground_check(Parameters),
      bsyntaxtree:check_if_typed_predicate(Cond),type_check(LS,store),type_check(State,store)) ).

:- assert_post(b_interpreter:b_compute_comprehension_set(_P,_C,_Info,_LS,_State,Val,_WF),
                 (b_interpreter:value_type(Val))). %, nonvar(Val)



b_compute_comprehension_set_symbolic(Parameters,Condition,_Info,LocalState,State,Result,WF) :-
   % a version which never expands the generated closure
    b_generate_closure_if_necessary(Parameters,Condition,LocalState,State,CResult,WF),
    Result=CResult.
% we could catch Parameters = [TID], Condition = b(member(b(identifier(ID),_TYPE,_),SET)
b_compute_comprehension_set(Parameters,Condition,Info,LocalState,State,Result,WF) :-
    b_generate_closure_if_necessary(Parameters,Condition,LocalState,State,CResult,WF),
    %print(gen(Parameters,Condition,CResult)),nl,
    b_compute_comprehension_set_aux(CResult,Info,Result,WF).

:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
b_compute_comprehension_set_aux(CResult,_,Result,_WF) :-
  (var(CResult) ; \+ functor(CResult,closure,3) ),!,
  %% print(not_generated_closure(CResult)),nl, %%
  Result=CResult.
b_compute_comprehension_set_aux(CResult,Info,Result,_WF) :-
   member(prob_annotation('SYMBOLIC'),Info),!,
   % print('SYMBOLIC'(Info)),nl,
   custom_explicit_sets:mark_closure(CResult,['SYMBOLIC'],CRS),Result=CRS.
b_compute_comprehension_set_aux(CResult,_Info,Result,_WF) :-
   (preferences:get_preference(convert_comprehension_sets_into_closures,true),
    \+ definitely_expand_this_explicit_set(CResult)
     ;
    %% print(check(CResult)),nl,trace, %%
    dont_expand_symbolic_explicit_set(CResult)),!,
    %((CResult=closure(P,_,_), print(dont_expand(P)),nl) -> true ; true),
   Result=CResult.
b_compute_comprehension_set_aux(CResult,_Info,Result,WF) :- b_expand_compute_comprehension_set(CResult,Result,WF).

b_expand_compute_comprehension_set(CResult,Result,WF) :-  % print(expand(CResult,Result)),nl,
     % using just get_wait_flag1(b_compute_comprehension_set,WF,WF0), makes many benchmarks faster ; but e.g., test 268 fails; maybe we sholud pre-compute bexpr_variables ? and pass this to expand_only_custom_closure_global
     (preferences:preference(use_smt_mode,false)
      -> get_wait_flag(1000,b_compute_comprehension_set,WF,WF0) % ideally we should say WF0 = _ ; but for some benchmarks Rule_DB_SIGAREA_0024_ori.mch this is important to avoid expanding closures before other stuff fails
      % only in SMT mode can there be any benefit in expanding a non-ground closure (when it is a lambda closure)
      % => hence delay expansion longer, possibly detecting enumeration warnings while expanding
      % TO DO: for lambda closures with known domain: expand earlier !
      ; get_wait_flag(2,b_compute_comprehension_set,WF,WF0)),
     %(CResult=closure([kpb,kpe,_],_,_) -> when(ground(CResult),(print(ground(CResult,WF0,WF)),nl,trace)) ; true),
     kernel_objects:mark_as_to_be_computed(Result), % avoid that we instantiate this set (e.g., in check_element_of_wf: we will get a full description of the set later; see test 1353
     (var(WF0),var(Result) -> ground_value_check(CResult,CGr) ; true),
     block_compute_comprehension_set_aux(CResult,Result,WF,WF0,CGr).

:- block block_compute_comprehension_set_aux(?,-,?,-,-).
block_compute_comprehension_set_aux(CResult,Result,WF,_WF0,_CGr) :-
	       ( %%((\+ ground(CResult),nonvar(WF0))-> print_term_summary(expand_comprehension_set_closure(CResult)),nl ; true), %%
	        nonvar(Result)
	         -> %% print_term_summary(equal_object(CResult,Result,WF0)), nl, %%
	            equal_object_wf(CResult,Result,WF) % this may expand CResult or do other tests; like emptyness test
	        ; %print_term_summary(try_expanding(CResult,cgr(CGr),wf0(WF0),Result)),nl,
	          catch_clpfd_overflow_call2(
	            on_enumeration_warning_with_continue(expand_only_custom_closure_global(CResult,Expansion,check_no_inf), /* we already know it is not infinite; and we do not want warnings for virtual time-outs  */
					 %(print(not_expanding(CResult,Result)),nl,
					 % an enumeration warning occured in the expansion scope: keep closure symbolic after all
					 % NOTE: for this detection to work it is better to delay the when above and not use a priority too low for WF0 above
					  b_compute_comprehension_set_sym_msg(CResult,Result)
					 ,
		             equal_object_wf(Result,Expansion,WF)  % instantiating Result can trigger other code which fails and confuses on_enumeration_warning
		             ),
		          b_compute_comprehension_set_sym(CResult,Result)
		        ) % catch_clpfd_overflow_call2
		    %,print(done_expand_custom_set(Result)),nl
		   ).

b_compute_comprehension_set_sym_msg(CResult,Result) :- CResult = closure(P,_T,B),
    % TO DO: should we check that this closure has not been constructed by ProB, e.g., for UNION ??
    % e.g., P = ['__RANGE_LAMBDA__'] what about '_zzzz_unary', '_zzzz_binary', '_prj_arg1__', ...
    % also: what if the enumeration warning has a pending_abort_error
    !,
    Msg = 'Keeping comprehension-set symbolic (you may want to use the /*@symbolic*/ pragma to prevent this message, unless it was due to a WD-Error), identifiers: ',
    add_message(b_compute_comprehension_set,Msg,P,B),
    b_compute_comprehension_set_sym(CResult,Result).
b_compute_comprehension_set_sym_msg(CResult,Result) :- b_compute_comprehension_set_sym(CResult,Result).


b_compute_comprehension_set_sym(CResult,Result) :-
    custom_explicit_sets:mark_closure(CResult,['SYMBOLIC'],CRS), % prevent doing the same mistake again
	equal_object(Result,CRS).

split_names_and_types([],[],[]).
split_names_and_types([Identifier|IdRest],[Name|NRest],[Type|TRest]) :-
    %get_texpr_expr(Identifier,identifier(Name)),
    def_get_texpr_id(Identifier,Name),
    get_texpr_type(Identifier,Type),
    split_names_and_types(IdRest,NRest,TRest).

:- use_module(closures,[construct_closure_if_necessary/4]).
b_generate_closure(Parameters,Condition,LocalState,State,Result,WF) :-
    split_names_and_types(Parameters,Names,Types),
    construct_closure(Names,Types,ClosurePred,Result),
    ((LocalState==[],State==[])
     -> ClosurePred = Condition % no lookups can be performed anyway; no need to compile
      ; b_compiler:b_compile(Condition,Names,LocalState,State,ClosurePred,WF) ).


b_generate_closure_if_necessary(Parameters,Condition,LocalState,State,Result,WF) :-
    split_names_and_types(Parameters,Names,Types),
    b_compiler:b_compile(Condition,Names,LocalState,State,ClosurePred,WF),
    %print('COMPILED: '),translate:print_bexpr(ClosurePred),nl,
    construct_closure_if_necessary(Names,Types,ClosurePred,Result).


:- public convert_list_of_expressions_into_sequence/2.
convert_list_of_expressions_into_sequence(List,SeqValue) :-
   convert_list_of_expressions_into_sequence(List,1,Seq),
   equal_object_optimized(Seq,SeqValue,convert_list_of_expressions_into_sequence).

convert_list_of_expressions_into_sequence([],_,[]).
convert_list_of_expressions_into_sequence([H|T],Cur,[(int(Cur),H)|ConvT]) :-
  C1 is Cur+1,
  convert_list_of_expressions_into_sequence(T,C1,ConvT).

convert_list_of_expressions_into_set(List,Set) :-
    init_wait_flags(WF),
    convert_list_of_expressions_into_set_wf(List,Set,set(any),WF),
    ground_wait_flags(WF).

convert_list_of_expressions_into_set_wf(ListOfElements,Set,Type,WF) :-
   list_of_expressions_is_ground(ListOfElements,Ground,OrderedListOfElements),
   %print(convert_list_of_expressions_into_set1(Ground,OrderedListOfElements,Set,Type)),nl,
   convert_list_of_expressions_into_set1(Ground,OrderedListOfElements,Set,Type,WF).
convert_list_of_expressions_into_set1(true,ListOfElements,Set,_Type,WF) :- %var(Set),
   !,  construct_avl_from_lists(ListOfElements,AVLSet),
   equal_object_wf(AVLSet,Set,convert_list_of_expressions_into_set1,WF).
convert_list_of_expressions_into_set1(_,Seq,Set,Type,WF) :-
   kernel_objects:max_cardinality(Type,TypeMaxCard),
   convert_list_of_expressions_into_set2(Seq,1,[],Set,TypeMaxCard,WF). % , print(converted(Seq,Set)),nl.

% TO DO: sort elements : ground elements first
convert_list_of_expressions_into_set2([],_,Set,Res,_,WF) :- %print(equal_object(Set,Res)),nl,
   equal_object_wf(Set,Res,convert_list_of_expressions_into_set2,WF).
%  debug:timer_call(kernel_objects:equal_object(Set,Res)).
convert_list_of_expressions_into_set2([H|T],MaxSze,SetSoFar,OutSet,TypeMaxCard,WF) :-
   %kernel_objects:basic_type_set(Type,IntSet,MaxSze),
   check_set_lists_bounded(IntSet,TypeMaxCard,MaxSze),
   % TO DO: optimize; avoid traversing or typing elements over and over again
   % maybe add a max. sze argument to add_element
   add_element_wf(H,SetSoFar,IntSet,WF), % print(add_el(H,SetSoFar,IntSet)),nl,
   % TO DO: check if H in OutSet and keep track of which elements have already been covered in OutSet
   %check_element_of_wf(H,OutSet,WF), makes tests 114, 115, 118-120, 1063 fail ! % we could also call lazy_try_check_element_of if this turns out to be too expensive
   kernel_objects:lazy_try_check_element_of(H,OutSet,WF),
   M1 is MaxSze+1,
   convert_list_of_expressions_into_set2(T,M1,IntSet,OutSet,TypeMaxCard,WF).


:- use_module(inf_arith).
% check that a set does not exceed a certain bound (to prevent infinite
%  or overeager enumeration of sets); Warning: will not check custom_explicit_set sizes
check_set_lists_bounded(Set,TypeMaxCard,Bound) :-
   inf_arith:infmin(TypeMaxCard,Bound,Card),
   (Card=inf -> true
     ; Card<1 -> empty_set(Set)
     ; check_set_bounded_aux(Set,Card)).

:- block check_set_bounded_aux(-,?).
check_set_bounded_aux(CustomSet,_) :- is_custom_explicit_set(CustomSet,check_set_bounded_aux),!.
check_set_bounded_aux([],_).
check_set_bounded_aux([_|T],Card) :-
   C1 is Card-1,
   (C1=0 -> empty_set(T) ; check_set_bounded_aux(T,C1)).



% check if the list can be converted into AVL form; if not: re-order list so that ground elements are present first
list_of_expressions_is_ground([],true,[]).
list_of_expressions_is_ground([E|T],Ground,ResList) :-
   (should_be_converted_to_avl(E) -> ResList=[E|RT], list_of_expressions_is_ground(T,Ground,RT)
    ; Ground=false,
      bring_ground_elements_forward(T,[E],ResList)
    ).

bring_ground_elements_forward([],Acc,Acc).
bring_ground_elements_forward([H|T],Acc,Res) :-
  (should_be_converted_to_avl(H) -> Res=[H|RT], bring_ground_elements_forward(T,Acc,RT)
    ; bring_ground_elements_forward(T,[H|Acc],Res)).

% compute cardinality of a list of parameters, such as for forall
parameter_list_cardinality([],1).
parameter_list_cardinality([b(identifier(_),BasicType,_)|T],Res) :-
   parameter_list_cardinality(T,TCard),
   (TCard=inf -> Res=inf
    ; kernel_objects:max_cardinality(BasicType,Card),
      kernel_objects:safe_mul(Card,TCard,Res)
   ).
% see also basic_type_list_cardinality, now moved to custom_explicit_sets

b_for_all(Parameters,Infos,LHS,RHS,LocalState,State,WF) :-
     get_wait_flag1(for_all(Parameters),WF,LWF), % hack: give total_function,... priority to set up domain/range
     %def_get_texpr_ids(Parameters,PN), print(compiling(forall(PN))),nl,b_compiler:b_compile(RHS,PN,LocalState,State,CompiledRHS,WF),
     b_for_all_aux(LWF,Parameters,Infos,LHS,RHS,LocalState,State,WF).

:- block b_for_all_aux(-,?,?,?,?,?,?,?).
b_for_all_aux(_,Parameters,Infos,LHS,RHS,LocalState,State,WF) :-
    split_names_and_types(Parameters,Names,Types),
    b_compile(LHS,Names,LocalState,State,CLHS,WF),
    b_compile(RHS,Names,LocalState,State,CRHS,WF),
    b_for_all_aux2(Parameters,Names,Types,Infos,CLHS,CRHS,WF).

:- use_module(kernel_tools,[ground_bexpr/1, ground_bexpr_check/2]).
b_for_all_aux2(_,_,_,_,LHS,RHS,_) :-
        (is_falsity(LHS) ; is_truth(RHS)),
        !. % quantifier always true
b_for_all_aux2(Parameters,CParameters,CParameterTypes,Infos,LHS,RHS,WF) :-
      ground_bexpr(LHS),
      \+ obvious_enumeration(LHS,CParameters), % if we have an obvious (small) enumeration in the LHS it is often better simply to enumerate and check the RHS for every instance
      ground_bexpr(RHS),  %print(not_exists_forall(Parameters,LHS,RHS)),nl,
      !,
      b_for_all_find_counter_example(Parameters,CParameters,CParameterTypes,Infos,LHS,RHS,WF).
b_for_all_aux2(Parameters,CParameters,CParameterTypes,Infos,LHS,RHS,WF) :-
    parameter_list_cardinality(Parameters,ParCard),
    peel_implications(Parameters,ParCard,LHS,RHS,  LHS2,RHS2),
    b_for_all_aux3(Parameters,ParCard,CParameters,CParameterTypes,Infos,LHS2,RHS2,WF).

b_for_all_aux3(Parameters,ParCard,CParameters,CParameterTypes,Infos,LHS,RHS,WF) :-
     %large_domain(Parameters,ParCard,LHS),
     is_infinite_or_very_large_explicit_set(closure(CParameters,CParameterTypes,LHS),10000000),
     % the domain is so large that we cannot possibly check all instances
     % in this case: we prefer to look for a counter example in the hope that the constraint solver can narrow down the search by combining LHS and not(RHS)
     !,
     debug_println(9,large_forall_domain(ParCard)),
     ground_bexpr_check(LHS,LG),
     ground_bexpr_check(RHS,RG),
     when((nonvar(LG),nonvar(RG)),
        b_for_all_find_counter_example(Parameters,CParameters,CParameterTypes,Infos,LHS,RHS,WF)
        ).
b_for_all_aux3(Parameters,ParCard,_CP,_,Infos,LHS,RHS,WF) :-
    if(expand_forall_quantifier(Parameters,LHS,RHS,Infos,WF,ReificationVariable),
      (  % print(expanded_forall(_CP)),nl,
       ReificationVariable=pred_true),
      (   % print(normal_forall(Parameters)),nl,
       b_generate_for_all_list_domain1(Parameters,ParCard,LHS,RHS,[],[],AllSolList,NewRHS,WF),
       %debug_println(9,'generated_for_all_domain'),
      % translate:print_bexpr(NewRHS),nl,
       expand_forall1(AllSolList,NewRHS,Parameters,WF)
      )).

% a treatment of for all where we look for counter examples satisfying LHS  & not(RHS)
b_for_all_find_counter_example(_Parameters,CParameters,CParameterTypes,_Infos,LHS,RHS,WF) :-
 % if LHS is Parameters: avl_set(V)  then we probably should not apply this rule ?
     create_negation(RHS,NegRHS),
     conjunct_predicates([LHS,NegRHS],Pred0),
	   % conjunct_predicates merges the used_ids Info in Condition
     %(member(used_ids(Used),Infos)
     % -> add_texpr_infos(Pred0,[used_ids(Used)],ClosureBody) ,print(used_ids(Used)),nl
     %  ; print(no_used_ids(Parameters,Infos)),nl,
     ClosureBody = Pred0,
     % format('looking for: ~w ',[CParameters]), translate:print_bexpr(LHS),nl, translate:print_bexpr(NegRHS),nl,
     not_with_enum_warning_and_possible_delay(
        ( %print(test_not_exists(CParameters)),nl,
         custom_explicit_sets:b_test_closure(CParameters,CParameterTypes,ClosureBody,ParValues,negative),
         performance_messages:perfmessage(forall_counter_example(CParameters,ParValues),LHS),
         %translate:print_bexpr(ClosureBody),nl,
         debug_println(9,forall_counter_example(CParameters,ParValues)
         )
           %preferences:temporary_set_preference(expand_avl_upto,-1),maplist(translate:print_bvalue,ParValues),
         ),WF).

:- use_module(library(timeout),[time_out/3]).
not_with_enum_warning_and_possible_delay(C,WF) :-
      get_preference(disprover_mode,true),
      !,
      % delay potentially expensive quantifications in the hope of finding other contradictions earlier
      if( (time_out(not_with_enum_warning(C),20,TRes),TRes \== time_out),
          true,
          (get_wait_flag(2,not_with_enum_warning_and_possible_delay,WF,LWF), % which priority to use ??
           debug_println(19,delaying_not),
           not_with_enum_warning_delay(C,LWF))).
not_with_enum_warning_and_possible_delay(C,_) :- not_with_enum_warning(C).

:- block not_with_enum_warning_delay(?,-).
not_with_enum_warning_delay(C,_) :- debug_println(19,trigger_not_with_enum_warning_delay),
   not_with_enum_warning(C).

:- use_module(avl_tools,[quick_avl_approximate_size/2]).
obvious_enumeration(LHS,[Id]) :-% TO DO  : use multiple ids
    % TO DO: unify with is_for_all_set_membership_predicate code below
      LHSMembership = b(member(b(identifier(Id),_,_),b(value(VAL),_,_)),pred,_),
      member_conjunct(LHSMembership,LHS,_),
      nonvar(VAL), VAL = avl_set(AVL), % TO DO: also support interval closures ?
      quick_avl_approximate_size(AVL,Size),
      %print(forall_mem_val(Id,Size)),nl,
      Size<256.
% also check equality ?

% expands forall quantifier and thus enables that we also propagate from right to left
     % This helps for test 1397 (HandshakeSimple_TotFun)
     % TO DO: directly call the code which tries to expand a forall quantifier
expand_forall_quantifier(Parameters,LHS,RHS,Infos,WF,ReificationVariable) :-
    (preferences:preference(use_smt_mode,true) -> true
     ; preferences:preference(solver_strength,SS), SS>9),  % sometimes contrapositive checking maybe not useful; see test 1368
    % TO DO: investigate and provide more principled heuristic
    b_interpreter_check:b_check_forall_wf(Parameters,LHS,RHS,Infos,[],[],WF,ReificationVariable).

% rewrite !(x,..).( TRUTH => (LHS2 => RHS2))   into  !(x,...).(LHS2 => RHS2)
peel_implications(Parameters,ParCard,LHS,RHS,  NewLHS,NewRHS) :-
  %print(forall),nl, print(Parameters),nl,print(LHS),nl,
  % this patterns occurs often in Event-B translations to classical B
   LHS=b(truth,_,_),
   RHS=b(implication(LHS2,RHS2),_,_), % TO DO: maybe also accept nested forall and then add to Parameters?
   (ParCard=inf -> true ; ParCard>50), % larger bound than below: maybe the user explicitly wants to expand it
   !,
   peel_implications(Parameters,ParCard,LHS2,RHS2, NewLHS,NewRHS).
peel_implications(_,_,LHS,RHS,  LHS,RHS).

b_generate_for_all_list_domain_nolwf(Parameters,LHS,RHS,LS,S,ForAllDomain,NewRHS,WF) :-
    parameter_list_cardinality(Parameters,ParCard),
    peel_implications(Parameters,ParCard,LHS,RHS,  LHS2,RHS2),
    b_generate_for_all_list_domain1(Parameters,ParCard,LHS2,RHS2,LS,S,ForAllDomain,NewRHS,WF).

% generate the domain of a Universally Quantified Formula
b_generate_for_all_list_domain1(Parameters,_ParCard,LHS,RHS,LS,S,ForAllDomain,NewRHS,WF) :-
   % optimized treatment for formulas of the form !x.(x:S => RHS) or !(x,y).(x|->y:S => RHS)
   % to do: expand to more formulas, such as !(x,y).(x:S & y:T => RHS)
   % or !y.(y:1..(n-1) & f(y)=TRUE => f(y+1)=TRUE)  -->  !y.(y:1..(n-1) => (f(y)=TRUE => f(y+1)=TRUE))
   % or we could try and reify set membership for small domains and trigger implication
    is_for_all_set_membership_predicate(LHS,Parameters,Set,Pattern,ParameterValues),!,
    %print(forall_set(Parameters)),nl,
    NewRHS=RHS,
    b_generate_set_for_all_domain(Set,LS,S,Pattern,ParameterValues,ForAllDomain,WF).
b_generate_for_all_list_domain1(Parameters,ParCard,LHS,RHS,LS,S,ForAllDomain,NewRHS,WF) :-
     b_generate_closure(Parameters,LHS,LS,S,Closure,WF), % POSSIBLY NO NEED TO COMPILE
     %print(forall_2(Closure,ParCard)),nl,
     b_generate_for_all_domain2(Closure,Parameters,ParCard,LHS,RHS,ForAllDomain,NewRHS,WF).

b_generate_for_all_domain2(Closure,_Parameters,_,LHS,RHS,ForAllDomain,NewRHS,_WF) :-
      (LHS=b(truth,pred,_) ; ground_value(Closure)),
      !, NewRHS=RHS,  % nl,print('FORALL EXPAND: '),(ground_value(Closure) -> print(ground(_Parameters)),nl, print_bexpr(LHS),nl ; print('NONGROUND!!'),nl,print_bexpr(RHS),nl), %%
      b_expand_closure_forall_to_list(Closure,ForAllDomain).
b_generate_for_all_domain2(_,Parameters,_,LHS,RHS,ForAllDomain,NewRHS,WF) :-
     is_a_conjunct(LHS,_,_), % otherwise we have already checked for set_membership above !
     LHS_Membership = b(member(_,_),_,_), % TO DO: also allow conjuncts
     member_conjunct(LHS_Membership,LHS,RestLHS),
     is_for_all_set_membership_predicate(LHS_Membership,Parameters,Set,Pattern,ParameterValues),
     max_cardinality_ok_for_expansion(Set,100,[],[]), % TO DO: simplify LS,S argument <----
     !, %% maximum cardinality of Set (could be infinite !!) !x.(x:NATURAL1 & x<N => P)
     % print('Splitting for all conjunct: '), print(Parameters),nl, print_bexpr(LHS_Membership),nl,print_bexpr(RestLHS),nl, %%
     NewRHS = b(implication(RestLHS,RHS),pred,[]), % TO DO: extract & join info
     b_generate_set_for_all_domain(Set,[],[],Pattern,ParameterValues,ForAllDomain,WF).
b_generate_for_all_domain2(Closure,_Parameters,ParCard,_LHS,RHS,ForAllDomain,NewRHS,_WF) :-
     (ParCard=inf -> true ; ParCard>8),
     !,
     NewRHS=RHS,  %print(large_parcard(ParCard,_Parameters)),nl,translate:print_bexpr(_LHS),nl, print(' => '),translate:print_bexpr(RHS),nl,
     b_expand_closure_forall_to_list(Closure,ForAllDomain).
b_generate_for_all_domain2(_Closure,Parameters,_,LHS,RHS,ForAllDomain,NewRHS,WF) :-
     % print('Inlining Forall '),parameter_list_cardinality(Parameters,PC), print(PC),nl,print(Parameters),nl,
     NewLHS=b(truth,pred,[]), NewRHS=b(implication(LHS,RHS),pred,[]),
     b_generate_closure(Parameters,NewLHS,[],[],Closure,WF),
     b_expand_closure_forall_to_list(Closure,ForAllDomain).

% code to decide whether a for all should be expanded if the
% predicate contains a set membership predicate : !x.( (P & x:S & P') => Q)
max_cardinality_ok_for_expansion(b(Expr,Type,_Info),Limit,LS,S) :- % which value to choose for Limit ??
  %print('Checking Max Card: '), print_bexpr(b(Expr,Type,_Info)),nl,
  kernel_objects:max_cardinality(Type,TypeCard), %print(type_card(TypeCard)),nl,
  ( (number(TypeCard),TypeCard =< Limit) -> true ;
   max_cardinality_ok_for_expansion2(Expr,Limit,TypeCard,LS,S)).

max_cardinality_ok_for_expansion1(b(E,_,_),Limit,TC,LS,S) :-
   max_cardinality_ok_for_expansion2(E,Limit,TC,LS,S).

:- use_module(custom_explicit_sets,[efficient_card_for_set/3]).

max_cardinality_ok_for_expansion2(set_extension(E),Limit,_TC,_,_) :- !, length(E,Len), Len =< Limit.
max_cardinality_ok_for_expansion2(interval(Up,Low),Limit,_TC,LS,S) :- !,
  %print(interval(Up,Low,Limit)),nl,
  get_integer(UpI,Up,LS,S), get_integer(LowI,Low,LS,S),
  UpI-LowI<Limit.
max_cardinality_ok_for_expansion2(integer_set(GS),Limit,_TC,_LS,_S) :- !, b_global_set_cardinality(GS,Card),
  number(Card), Card =< Limit.
max_cardinality_ok_for_expansion2(set_subtraction(A,_),Limit,TC,LS,S) :- !,
    max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(image(RelA,_),Limit,TC,LS,S) :- !,
    max_cardinality_ok_for_expansion1(RelA,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(function(_RelA,_),_Limit,_TC,_LS,_S) :- !,
    fail. % TO DO: we should look at the size of the range elements
%    max_cardinality_ok_for_expansion1(_RelA,_Limit,_TC,_LS,_S).
max_cardinality_ok_for_expansion2(intersection(A,B),Limit,TC,LS,S) :- !,
 (max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S) -> true ; max_cardinality_ok_for_expansion1(B,Limit,TC,LS,S)).
max_cardinality_ok_for_expansion2(union(A,B),Limit,TC,LS,S) :- !,
  max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S), max_cardinality_ok_for_expansion1(B,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(range(A),Limit,TC,LS,S) :- !,
  max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(domain(A),Limit,TC,LS,S) :- !,
  max_cardinality_ok_for_expansion1(A,Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(direct_product(A,B),Limit,TC,LS,S) :- !,
   max_cardinality_ok_for_expansion2(cartesian_product(A,B),Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(parallel_product(A,B),Limit,TC,LS,S) :- !,
   max_cardinality_ok_for_expansion2(cartesian_product(A,B),Limit,TC,LS,S).
max_cardinality_ok_for_expansion2(cartesian_product(A,B),Limit,TC,LS,S) :- !,
  ILimit is Limit//2, ILimit>0, %use sqrt ?
  % print(cartesian(Limit,ILimit,A,B)),nl,
  max_cardinality_ok_for_expansion1(A,ILimit,TC,LS,S),
  max_cardinality_ok_for_expansion1(B,ILimit,TC,LS,S).
max_cardinality_ok_for_expansion2(value(S),Limit,_TC,_LS,_S) :- !,
   efficient_card_for_set(S,Card,C),call(C),
   number(Card), Card =< Limit.
max_cardinality_ok_for_expansion2(identifier(_ID),_Limit,TC,_LS,_S) :- !, number(TC).
  % TO DO: check if ID is fully defined; if so use card else accept as it will
  % probably be represented by a list and the forall constraints could be useful in finding ID's value
  % we have to store/compute ID anyway (could be infinite closure; hence check that TC is a number);  a bit of a hack; is there a more principled way?
max_cardinality_ok_for_expansion2(external_function_call(_,_),_,_,_,_) :- !,fail.
max_cardinality_ok_for_expansion2(lazy_lookup_expr(_),_,_,_,_) :- !,fail.
max_cardinality_ok_for_expansion2(record_field(_,_),_Limit,_TC,_LS,_S) :- !, fail.
max_cardinality_ok_for_expansion2(E,Limit,TC,_LS,_S) :- functor(E,F,N),
   print(unknown_max_card(F/N,Limit,TC)),nl,fail.

get_integer(X, b(Expr,integer,_),LS,S) :- get_integer2(Expr,X,LS,S).
get_integer2(integer(X),X,_,_) :- number(X).
get_integer2(identifier(Id),Val,LS,S) :- lookup_value(Id,LS,S,int(Val)), number(Val).

member_conjunct(Predicate,Conjunction,Rest) :-
  conjunction_to_list(Conjunction,List),
  select(Predicate,List,RestList),
  conjunct_predicates(RestList,Rest).
% member_conjunct(Conj,Conj,b(truth,pred,[])). covered by case above

b_generate_set_for_all_domain(Set,LS,S,Pattern,ParameterValues,ForAllDomain,WF) :-
    b_compute_expression(Set,LS,S,EvSet,WF),
    expand_custom_set_to_list(EvSet,ESet,_Done,b_generate_set_for_all_domain),    %% print(b_set_forall_domain(ESet)),nl, %%%
    convert_list_to_parameter_tuples(ESet,Pattern,ParameterValues,ForAllDomain).


% check if it is a predicate of the form x:S or x:Sx & y:Sy ... which we can optimize
%  (i.e. compute SET before all values inside it are fully known)
is_for_all_set_membership_predicate(LHS,Parameters,Set,Pattern,ParameterValues) :-
    % print('CHECK FORALL : '), translate:print_bexpr(LHS),nl,nl,
    get_texpr_ids(Parameters,ParIDs), % TO DO: call get_texpr_ids before
    same_length(ParIDs,ParameterValues),
    is_for_all_set_membership_predicate2(LHS,ParIDs,ParIDs,UnmatchedParIDs,Set,Pattern,ParameterValues,_UPV),
    UnmatchedParIDs=[].
    %(UnmatchedParIDs=[] -> debug_println(9,set_membership_forall(ParIDs)) ; debug_println(9,unmatched(UnmatchedParIDs)),fail).

is_for_all_set_membership_predicate2(LHS,All,ParIDs,UnmatchedParIDs,Set,Pattern,ParameterValues,UPV) :-
    is_member_test(LHS,Element,Set),
    !,
    identifiers_not_used_in_bexp(All,Set),
    non_failing_pattern_match(Element,ParIDs,UnmatchedParIDs,Pattern,ParameterValues,UPV). % TO DO: also allow failing pattern matches ?!; at least when splitting
is_for_all_set_membership_predicate2(LHS,All,ParIDs,ParIDs2,Set,
                                    (Pat1,Pat2),ParameterValues,UPV) :-
    is_a_conjunct(LHS,LHS1,LHS2), % we could have something of the form x:Sx & y:Sy ...
    Set = b(cartesian_product(Set1,Set2),couple(T1,T2),[]),
    is_for_all_set_membership_predicate2(LHS1,All,ParIDs, ParIDs1,Set1,Pat1,ParameterValues,PV1),
    get_texpr_type(Set1,T1),
    !,
    is_for_all_set_membership_predicate2(LHS2,All,ParIDs1,ParIDs2,Set2,Pat2,PV1,UPV),
    get_texpr_type(Set2,T2).

% TO DO: maybe use a more general test to be reused in other places
is_member_test(b(P,_,_),Element,Set) :- is_member_test_aux(P,Element,Set).
is_member_test_aux(member(Element,Set),Element,Set).
% also detect equalities: important for tests 28,29 after b_compiler can introduce equalities for singleton set memberships
is_member_test_aux(equal(Element,SetVal),Element,Set) :- % TO DO : also other direction + other kinds of values ?
   SetVal = b(value(SV),T,I),
   Set = b(value([SV]),set(T),I).

identifiers_not_used_in_bexp(IDList,Expr) :-
   find_identifier_uses(Expr,[],Used),
   id_check(IDList,Used).
id_check([],_).
id_check([ID|T],Used) :- %get_texpr_id(IDE,ID),
   nonmember(ID,Used), !,
   id_check(T,Used).

non_failing_pattern_match(IDE,IDsIn,IDsOut,PatVariable,VarsIn,VarsOut) :-
   get_texpr_id(IDE,ID),
   !,
   (nth1(Nr,IDsIn,ID,IDsOut) -> nth1(Nr,VarsIn,PatVariable,VarsOut)),!.
%   select(ID,IDsIn,IDsOut).
non_failing_pattern_match(b(couple(ID1,ID2),_,_),IDS,OutIDS,Pattern,VarsIn,VarsOut) :-
   Pattern = (Pat1,Pat2),
   non_failing_pattern_match(ID1,IDS,RemIDS,Pat1,VarsIn,VarsInt),
   non_failing_pattern_match(ID2,RemIDS,OutIDS,Pat2,VarsInt,VarsOut).

%large_parameter_list_cardinality(Parameters,Limit) :-
%  parameter_list_cardinality(Parameters,ParCard),  (ParCard=inf -> true; ParCard > Limit).

:- block convert_list_to_parameter_tuples(-,?,?,?).
convert_list_to_parameter_tuples([],_,_,[]).
convert_list_to_parameter_tuples([H|T],Pattern,ParameterValues,Res) :-
   copy_term((Pattern,ParameterValues),(CPat,CPV)),
   (CPat=H -> Res = [CPV|RT]
     ; print(error_could_not_unify_pattern(H,Pattern)),nl,
       Res=RT),
    convert_list_to_parameter_tuples(T,Pattern,ParameterValues,RT).

:- use_module(library(lists),[maplist/3]).
expand_forall1(List,RHS,Parameters,_WF) :-
   ground_bexpr(RHS), % because compiled this means we cannot do any constraint propagation on variables in RHS
   ground_value(List), % ditto here: no variables on which we could do constraint propagation
   %% preferences:get_preference(double_evaluation_when_analysing,false), % causes test 1112 to fail TO DO:fix
   % otherwise we do not have a real redundancy wrt to not exists check
   debug_println(15,ground_forall_translating_into_not_exists(Parameters)),
   !,
   kernel_waitflags:init_wait_flags(LocalWF),
   % Advantage over treatment below: NegRHS is interpreted only once
   % but we cannot provide any instantiations and there must not be any links with variables outside of the forall !!
   create_negation(RHS,NegRHS),
   not_with_enum_warning( (
         b_interpreter:test_forall_instance(ParResult,NegRHS,Parameters,LocalWF,negative), % this can fail; if inconsistency detected in WF0
         kernel_waitflags:ground_det_wait_flag(LocalWF),
         member(ParResultInList,List), % check the body for every solution
         lists:maplist(kernel_objects:equal_object,ParResultInList,ParResult), % match the parameters with the solutions of the domain
         kernel_waitflags:ground_wait_flags(LocalWF)
        )
     ).
expand_forall1(List,RHS,Parameters,WF) :- expand_forall_aux(List,RHS,Parameters,WF,0).

:- block expand_forall_aux(-,?,?, ?,?).
expand_forall_aux([],_RHS,_Parameters,_WF,_Nr) :- !.
      %% :- print('finished_expanding forall bodies: '),print_bexpr(_RHS),nl , print(number_of_instances(_Nr)),nl,kernel_waitflags:portray_waitflags(_WF). %%
expand_forall_aux([ParResult|T],RHS,Parameters,WF,Nr) :- !,
       %% print(test_forall_instance(Nr,ParResult)),nl, %%
   test_forall_instance(ParResult,RHS,Parameters,WF,positive),
   N1 is Nr+1,
   expand_forall_aux(T,RHS,Parameters,WF,N1).
expand_forall_aux(Set,RHS,Parameters,WF,Nr) :-
    add_internal_error('Illegal call: ', expand_forall_aux(Set,RHS,Parameters,WF,Nr)),
    fail.

test_forall_instance(ParResult,RHS,Parameters,WF,NegationContext) :-
   LocalState=[],State=[],
   %(Parameters == [b(identifier(ss),global('SESSION'),[nodeid(pos(307,1,79,6,79,7))])] -> trace ; true),
   if(set_up_typed_localstate(Parameters,ParResult,_TypedVals,LocalState,NewLocalState,NegationContext),true,
      (nl,print(set_up_typed_localstate_failed(Parameters,ParResult,_,LocalState,NewLocalState)),nl,
     % it can happen that ParResult has constraints on it and that setting up the typing info detects the error
     % add_internal_error('Call failed: ', set_up_typed_localstate(Parameters,ParResult,_,LocalState,NewLocalState)),
     fail)),
   %  print('test: '),print_bexpr(RHS), print(' : '), print(ParResult),nl,
   b_test_boolean_expression(RHS,NewLocalState,State,WF).
    % using b_test_inner_boolean_expression considerably slows down NQueens, Sudoku,...
 %  WF=InnerWF. /* give the boolean expression a chance to do ground propagations first */




% use this when the test boolean expression could be called with WF0 already set
% give the inner expression a chance to do the deterministic stuff first
b_test_inner_boolean_expression(Pred,LocalState,State,WF) :-
   kernel_waitflags:clone_wait_flags_from1(WF,WF2),
   % set WF0 to variable to propagate deterministic infos first:
   b_test_boolean_expression(Pred,LocalState,State,WF2),
   kernel_waitflags:clone_wait_flags_from1_finish(WF,WF2).
% the same for testing negation of a predicate:
b_not_test_inner_boolean_expression(Pred,LocalState,State,WF) :-
   kernel_waitflags:clone_wait_flags_from1(WF,WF2),
   % set WF0 to variable to propagate deterministic infos first:
   b_not_test_boolean_expression(Pred,LocalState,State,WF2),
   kernel_waitflags:clone_wait_flags_from1_finish(WF,WF2).

   /* now the same as above; but without converting list of values into pairs */
:- assert_pre(b_interpreter:b_expand_closure_forall_to_list(Closure,_Result),
                (nonvar(Closure), custom_explicit_sets:is_closure(Closure,_Parameters,_ParameterTypes,ClosureBody),
                 nonvar(ClosureBody),
                 bsyntaxtree:check_if_typed_predicate(ClosureBody))).
:- assert_post(b_interpreter:b_expand_closure_forall_to_list(_Closure,Result),
                 b_interpreter:value_type(Result)).
b_expand_closure_forall_to_list(Closure,Result) :- nonvar(Closure),
    Closure = closure(Parameters,ParameterTypes,ClosureBody),
    !,
    % print('Expand closure forall: '), translate:print_bexpr(ClosureBody),nl,
    % at the moment we may wait too long before expanding this closure
    % what if we have  f(1)  in the ClosureBody: in principle we only need to wait until f(1) known or even f sekeleton set up ? compilation now solves this issue ?
    ((performance_monitoring_on, term_variables(ClosureBody,Vars), Vars \= []) -> %nl,print(Closure),nl,
      perfmessagecall(delaying_forall(Parameters,vars(Vars),out(ParValues)),translate:print_bexpr(ClosureBody),ClosureBody)
    ; true),
    delay_setof_list( ParValues,
                     %% (print_bt_message(expanding_forall_closure(Parameters,ClosureBody,ParValues)), %%
                      custom_explicit_sets:b_test_closure(Parameters,ParameterTypes,ClosureBody,ParValues,all_solutions)
                      %% ,print_bt_message(expanded(Parameters,ParValues)) ) %%
                      ,
                     Result,
                     ParValues).
b_expand_closure_forall_to_list(Closure,Result) :-
       add_internal_error('Illegal Call: ', b_expand_closure_forall_to_list(Closure,Result)),
       fail.
/*  replace last clause by following if you want to use  b_generate_closure_if_necessary
b_expand_closure_forall_to_list(Set,LESet) :- % not generated a closure; could be global_set, avl_set or list
   expand_custom_set_to_list(Set,ESet,_,b_expand_closure_forall_to_list),
   wrap_setelements_into_list(ESet,LESet). % the expand_forall expects a list of values; one for each parameter
:- block wrap_setelements_into_list(-,?).
wrap_setelements_into_list([],[]).
wrap_setelements_into_list([H|T],[[H]|WT]) :- wrap_setelements_into_list(T,WT).
*/

:- assert_pre(b_interpreter:set_up_typed_localstate(V,S,_), (list_skeleton(V),list_skeleton(S))).
:- assert_post(b_interpreter:set_up_typed_localstate(_,_,R),list_skeleton(R)).

set_up_typed_localstate(Identifiers,InState,OutState) :-
	set_up_typed_localstate(Identifiers,_FreshVars,_TypedVals,InState,OutState,positive).

set_up_typed_localstate([],[],[],State,State,_).
set_up_typed_localstate([Identifier|IdRest],[Val|ValRest],[typedval(Val,Type,Var,TriggersEnumWarning)|TRest],
                        InState,OutState,NegationContext) :-
    % The other way round (first add Identfier, then IdRest) would be more efficient,
    % but we do it this way to keep compatibility with a previous version.
    % In specfile:compute_operation_effect, the arguments of an operation are in the same order than in the store
    set_up_typed_localstate(IdRest,ValRest,TRest,InState,InterState,NegationContext),
    def_get_texpr_id(Identifier,Var), get_texpr_type(Identifier,Type),
    get_texpr_info(Identifier,Info),
    triggers_enum_warning(Var,Type,Info,NegationContext,TriggersEnumWarning),
    add_typed_var_to_localstate(Var,Val,Type,InterState,OutState).

% same version; but with list of atomic ids and list of types
set_up_typed_localstate2([],[],[],[],State,State,_).
set_up_typed_localstate2([Var|IdRest],[Type|TyRest],[Val|ValRest],[typedval(Val,Type,Var,TriggersEnumWarning)|TRest],
                        InState,OutState,NegationContext) :-
    % The other way round (first add Identfier, then IdRest) would be more efficient,
    % but we do it this way to keep compatibility with a previous version.
    % In specfile:compute_operation_effect, the arguments of an operation are in the same order than in the store
    set_up_typed_localstate2(IdRest,TyRest,ValRest,TRest,InState,InterState,NegationContext),
    triggers_enum_warning(Var,Type,[],NegationContext,TriggersEnumWarning),
    add_typed_var_to_localstate(Var,Val,Type,InterState,OutState).

% possible values for NegationContext: positive, negative, all_solutions
triggers_enum_warning(Id,_Type,_Info,all_solutions,Res) :- !,
    Res = trigger_throw(Id).
triggers_enum_warning(Id,_Type,Info,NegationContext,trigger_false(Id)) :-
    NegationContext=positive, % when negated then success will occur if all possibilities have been tried out
    memberchk(introduced_by(exists),Info),
    %%((preferences:get_preference(disprover_mode,true),type_contains_unfixed_deferred_set(Type))-> fail % otherwise finding a solution may depend on the size of the deferred set; true),
    %% The Disprover now checks whether unfixed_deferred sets were involved; keeping track of implicit enumerations of deferred sets is just too difficult
    !.
triggers_enum_warning(Id,_,_,_,trigger_true(Id)).


add_typed_var_to_localstate(Var,Val,Type,InState,OutState) :-
   %% print(adding(Var,Val,Type)),nl, %%
    add_var_to_localstate(Var,Val,InState,OutState),
    %%when(ground(Val), (print(' ++ '),print(Var),print(' ---> '), print(Val),nl)),
    %%print(basic_type(Val,Type,InState,OutState)),nl, %translate:print_bvalue(Val),nl,
    kernel_objects:basic_type(Val,Type). %% ,print(ok(Val)),nl.

/*
leftmost_conjunct(b(conjunct(A,B),pred,Info),Left,Right) :- !,
    (leftmost_conjunct(A,AL,AR)
      -> Left = AL, Right = b(conjunct(AR,B),pred,Info)
      ;  Left=A, Right = B ).

xxb_test_exists(Parameters,Condition,Infos,LocalState,State,WF) :-
     b_check_boolean_expression(b(exists(Parameters,Condition),pred,Infos),LocalState,State,WF,NR),
     get_exists_used_ids(Parameters,Condition,Infos,Used),lookup_values_if_vars(Used,LocalState,State,WaitVars),
     \+ ground(WaitVars),
     !,
     debug_println(9,unfolding_exists(Parameters)), print(unfolding(Parameters)),nl,
     %translate:print_bexpr(Condition),nl, print(State),nl, print(LocalState),nl,nl,
     NR=pred_true. %, print(set_res(pred_true)),nl.
     % if enabled it slows down: ./probcli examples/EventBPrologPackages/SET_Game/SET_GAM_Sym_NoSet20_mch.eventb -mc 10000000 -p SYMMETRY_MODE hash -p TIME_OUT 7000 -p CLPFD TRUE -df -goal "n=18" -p MAX_OPERATIONS 20 -strict -expcterr goal_found
*/

b_test_exists(Parameters,Condition,Infos,LocalState,State,WF) :-
        get_wait_flag0(WF,WF0),
	 	b_test_exists(WF0,Parameters,Condition,Infos,LocalState,State,WF).

:- block b_test_exists(-,?,?, ?,?,?,?).
b_test_exists(_LWF,Parameters,Condition,Infos,LocalState,State,WF) :-
    if((preferences:preference(use_smt_mode,true),
        b_interpreter_check:b_check_exists_wf(Parameters,Condition,Infos,LocalState,State,WF,ReificationVariable)),
        % try and expand quantifiers of small cardinality; see also expand_forall_quantifier
        % tested in test 1452
        (debug_println(expanded_exists(Parameters)),
         ReificationVariable=pred_true),
        b_test_exists_wo_expansion(Parameters,Condition,Infos,LocalState,State,WF)).

b_test_exists_wo_expansion(Parameters,Condition,Infos,LocalState,State,WF) :-
    member(dom_range_generated_exists,Infos),
    !, % we enumerate the exists normally; treat it just like an ordinary predicate
    % this will generate multiple solutions; but avoids delaying enumeration of quantified parameters
    %print('Lifting existential quantifier: '), translate:print_bexpr(Condition),nl,
    set_up_typed_localstate(Parameters,_FreshOutputVars,TypedVals,LocalState,NewLocalState,positive),
    b_test_boolean_expression(Condition,NewLocalState,State,WF),
    b_enumerate:b_tighter_enumerate_values(TypedVals,WF).
b_test_exists_wo_expansion(Parameters,Condition,Infos,LocalState,State,WF) :-
    get_exists_used_ids(Parameters,Condition,Infos,Used),
    % IT IS IMPORTANT THAT THE USED INFO IS CORRECT ; otherwise the WaitVars will be incorrect
    % The used_ids info does not include the Parameters of the existential quantifier !
     %  print(b_test_exists(Parameters)),nl, print(state(State)),nl,
    set_up_typed_localstate(Parameters,_FreshOutputVars,TypedVals,LocalState,NewLocalState,positive),
    lookup_values_if_vars(Used,LocalState,State,WaitVarsState), % TO DO: we could use b_compiler instead
    % What if the Condition was compiled and now contains a previously used_variable inside a closure/value ?
    term_variables((Condition,WaitVarsState),WaitVars),
    %(ground(Condition) -> true ; print(waitvars(WaitVars,Condition)),nl,
    create_inner_wait_flags(WF,b_test_exists(Parameters),LocalWF), % we are creating inner waitflags here as the exist will wait anyway; LocalWF shares just WFE flag with WF
    b_test_boolean_expression(Condition,NewLocalState,State,LocalWF), /* check Condition */
    %% TO DO: check if Condition is deterministic; if so we do not need to use delay_once_wait_vars
      %   print('tested_exists: '), print_bexpr(Condition),nl, % print(local(NewLocalState)),nl, %%
     %  print(waitvars(WaitVars)),nl,kernel_waitflags:portray_waitflags(LocalWF),nl,
    ground_det_wait_flag(LocalWF),  % print('det no contradiction'),nl,flush_output,
    b_enumerate:b_tighter_enumerate_values(TypedVals,LocalWF), % will not yet enumerate
    get_enumeration_starting_wait_flag(b_test_exists(Parameters),WF,ESWF),  %% TO DO: get flag when enumeration of infinite type starts; maybe if domain of exists is small we can start enumerating earlier ?? <-> relation with lifting exists in closure expansion heuristic
    when((nonvar(ESWF);ground(WaitVars)),b_enumerate_exists(ESWF,WaitVars,Parameters,LocalWF,WF)).

b_enumerate_exists(ESWF,WaitVars,Parameters,LocalWF,WF) :- %print(enum_exists(ESWF,WaitVars,Parameters,LocalWF,WF)),nl,
    %print('LOCAL WF: '),portray_waitflags(LocalWF), nl, print('WF: '),portray_waitflags(WF),nl,
    ((var(ESWF) ; ground(WaitVars)) % if ESWF is var then WaitVars must be ground due to when above
      -> kernel_waitflags:get_idle_wait_flag(b_enumerate_exists(Parameters),WF,LWF),
         /* allow all other co-routines waiting on WaitVars to complete */
         b_enumerate_exists_aux_ground(LWF,WaitVars,Parameters,LocalWF,WF)
       ; b_enumerate_exists_aux_non_ground(WaitVars,Parameters,LocalWF,WF)
    ).

:- block b_enumerate_exists_aux_ground(-,?,?,?,?).
b_enumerate_exists_aux_ground(_,_WaitVars,Parameters,LocalWF,_WF) :-
   /* Note: this does a local cut: so it is important
       that all conditions are fully evaluated before performing the cut ; otherwise
       solutions will be lost */
         % print_bt_message('ENUMERATE EXISTS'(_Parameters)),nl, % portray_waitflags(_WF),
         if(ground_wait_flags_for_exists(LocalWF),true,
              (perfmessage(late_bound_exists_failed(Parameters),Parameters),fail)).


ground_wait_flags_for_exists(LocalWF) :-
         enter_new_clean_error_scope(Level), %print(enter_exists(Level)),nl,
         % NOTE: if this engenders an enumeration warning: it may have been better to enumerate the outer WF first ?! : but as WaitVars are ground, so only failure of outer WF could help us here in preventing enumeration warning
         call_cleanup(ground_wait_flags_for_exists2(LocalWF,Level),
                      exit_error_scope(Level,_,ground_wait_flags_for_exists)).

ground_wait_flags_for_exists2(LocalWF,Level) :-
         (kernel_waitflags:ground_constraintprop_wait_flags(LocalWF)
           -> % local cut: one solution is sufficient.
              clear_events_in_error_scope(Level,enumeration_warning(_,_,_,_,_))
         %  ;  %print(fail),nl, % do not clear enumeration warnings    fail
         ).

b_enumerate_exists_aux_non_ground(WaitVars,_Parameters,LocalWF,WF) :-
         %%  print(b_test_exists_enumerating(_WaitVars,_Parameters,LocalWF)),nl,
         kernel_waitflags:ground_wait_flag_to_enumeration(LocalWF),
         % print(waiting_for_ground(WaitVars)),nl, portray_waitflags(LocalWF),nl,
              % TO DO: use kernel_tools
         when(ground(WaitVars),b_enum_exists_with_cut(LocalWF,WF)).

b_enum_exists_with_cut(LocalWF,WF) :-
     kernel_waitflags:get_idle_wait_flag(b_enumerate_exists,WF,LWF),
     % allow all other co-routines waiting on WaitVars to complete
     b_enum_exists_with_cut_aux(LWF,LocalWF).
:- block b_enum_exists_with_cut_aux(-,?).
b_enum_exists_with_cut_aux(_LWF,LocalWF) :-
     %  print(b_enum_exists_with_cut(LocalWF)),nl,
     ground_wait_flags_for_exists(LocalWF).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
:- use_module(b_ast_cleanup,[check_used_ids_info/4]).
get_exists_used_ids(Parameters,Condition,Infos,Used) :-
    member(used_ids(Used),Infos)
    -> check_used_ids_info(Parameters,Condition,Used,exists) %% comment in to check used_ids field
    ; otherwise ->
	     add_internal_error(
		  'Expected information of used identifiers in exists operation information',Parameters:Infos),
		  bsyntaxtree:find_identifier_uses(Condition, [], Used).
:- else.
get_exists_used_ids(Parameters,Condition,Infos,Used) :-
    member(used_ids(Used),Infos) -> true
    ; otherwise ->
	     add_internal_error(
		  'Expected information of used identifiers in exists operation information',Parameters:Infos),
		  bsyntaxtree:find_identifier_uses(Condition, [], Used).
:- endif.
% what if we call test_exists for a negated universal quantifier ??



is_an_equality(b(equal(A,B),_,_),A,B).

:- use_module(kernel_tools,[ground_state/1]).
/* the following causes performance problems with examples/B/Mathematical/GraphIso/CheckGraphIsomorphism
   and with lausanne.mch in SMT mode */
b_not_test_exists(Parameters,Condition,Infos,LocalState,State,WF) :-
  preferences:preference(use_smt_mode,true),
  is_a_conjunct(Condition,LHS,RHS),
  % to do: use kernel_tools
  (\+ ground_state(LocalState) ; \+ ground_state(State)), % TO DO: check if any of the variables in Condition are not ground
  % TO DO: check if used_ids exist in Infos ? memberchk(used_ids(_),Infos),
  !,
  %print(translating_not_exists_into_forall),nl,print('  '),print_bexpr(LHS),nl, print('=> '),print_bexpr(RHS),nl,
  (is_for_all_set_membership_predicate(RHS,Parameters,_Set,_Pattern,_ParameterValues),
   \+ is_for_all_set_membership_predicate(LHS,Parameters,_,_,_)
    -> LLHS=RHS, RRHS=LHS % swap LHS and RHS; TO DO: search for membership predicate inside LHS,RHS !!
       %, print(swap_lhs_rhs),nl
     ; LLHS=LHS, RRHS=RHS
  ),
  %print(state(State)),nl, print(local(LocalState)),nl,
  NegRHS = b(negation(RRHS),pred,[]),
  b_for_all(Parameters,Infos,LLHS,NegRHS,LocalState,State,WF).
b_not_test_exists(Parameters,Condition,_Infos,LocalState,State,WF) :-
   % print(not_exists),nl,
    % compile to remove dependency on unused parts of State+LocalState; will make it delay less below
    def_get_texpr_ids(Parameters,ParaIDs),
     %%  print(b_not_test_exists(ParaIDs)),nl, print(state(LocalState,State)),nl, %%
    ~~mnf( b_compiler:b_compile(Condition,ParaIDs,LocalState,State,CompiledCond,WF) ),
   %% print('compiled: '),print_bexpr(CompiledCond),nl, %%
    delay_not( (
                 b_interpreter:set_up_typed_localstate(Parameters,FreshOutputVars,TypedVals,[],NewLocalState,negated),
                kernel_waitflags:create_inner_wait_flags(WF,b_not_test_exists(Parameters),LocalWF), % delay generation of WD error messages until outer WFE flag set
                  % print(testing_not_exists_condition(Parameters,FreshOutputVars)),nl, %%
                 b_interpreter:b_test_boolean_expression(CompiledCond,NewLocalState,[],LocalWF),  % check Condition
                 % print(enumerating(TypedVals)),nl,
                 b_enumerate:b_tighter_enumerate_values(TypedVals,LocalWF),
                 kernel_waitflags:ground_constraintprop_wait_flags(LocalWF) % do not ground WFE
                 %  , print(finished_enum_exists(TypedVals,NewLocalState)),nl, kernel_waitflags:portray_waitflags(LocalWF) %%
              ), [WF,LocalWF,TypedVals,NewLocalState|FreshOutputVars] ).




/* --------------------------------- */

value_type(_X). % only used in post_conditions (assert_post)
/*
value_type(X) :- nonvar(X),
 (X = value(_) ; X = global_set(_) ;
  X = global_constant(_) ; X = term(_)).*/


% look up identifiers in store, if the identifier
% is not in the store, assume it's a constant and ignore it
% used for determining wait variables for existential quantifier
% TO DO: better deal with CSE @identifiers ?
lookup_values_if_vars([],_,_,[]).
lookup_values_if_vars([Id|IRest],LocalStore,Store,Values) :-
	( lookup_value(Id,LocalStore,Store,Val)
	  -> (force_evaluation_of_lazy_value(Id,Val,RealVal)
	      -> Values = [RealVal|VRest]
	       ; Values = [Val|VRest])
	; Values = VRest),
	lookup_values_if_vars(IRest,LocalStore,Store,VRest).

force_evaluation_of_lazy_value(Id,Value,RealVal) :- nonvar(Value), Value=(Trigger,RealVal),
    var(Trigger),
    b_expression_sharing:is_lazy_let_identifier(Id),
    !,
    Trigger=pred_true. % force evaluation, as the exists will otherwise delay/flounder
    % Warning: this means that outer lazy_let_expressions inside an exists should be well-defined (WD)


 /* will not generate error messages; used for CSP||B when we don't know if something is a B variable or not */
   /* TO DO: add peel_prefix for global_sets and constants ?? */
try_lookup_value_in_store_and_global_sets(Id,State,Val) :-
 (lookup_value(Id,State,RVal) -> RVal=Val
    ; (b_global_set(Id)
       -> Val=global_set(Id)  /* global set evaluates to itself */
        ; (b_global_sets:lookup_global_constant(Id,VGC)
              -> Val = VGC
              ;  fail
          ))
  ).


:- assert_pre(b_interpreter:lookup_value_in_store_and_global_sets(Id,LS,State,_Val),
                (nonvar(Id),type_check(LS,store),type_check(State,store))).
:- assert_post(b_interpreter:lookup_value_in_store_and_global_sets(_Id,_LS,_State,_Val), true).

lookup_value_in_store_and_global_sets(Id,LocalState,State,Val) :-
  lookup_value_in_store_and_global_sets(Id,LocalState,State,Val,unknown).
lookup_value_in_store_and_global_sets(Id,LocalState,State,Val,Span) :-
 (lookup_value_with_span(Id,LocalState,State,RVal,Span) -> RVal=Val
    ; lookup_value_in_global_sets(Id,Val,LocalState,State,Span)
  ).

lookup_value_in_global_sets(Id,Val,LocalState,State,Span) :-
   (b_global_sets:lookup_global_constant(Id,VGC)
        -> Val=VGC
        ; (b_global_set(Id)
              ->  Val = global_set(Id)  /* global set evaluates to itself */
              ;   %translate_environment(LocalState,LS),
                  %translate_environment(State,S),
                  %nl,print(local(LS)),nl,  print(global(S)),nl,nl,
                  translate:translate_bstate_limited(LocalState,LS), %context(
                  translate:translate_bstate_limited(State,S),
                  ((member(bind(Id,_),LocalState) ; member(bind(Id,_),State))
                    -> Msg = 'Cannot determine value for identifier: '
                    ;  Msg = 'Cannot find identifier: '
                  ),
                  (LocalState=[] -> StateInfo=global(S) ; StateInfo=local(LS):global(S)),
                  add_error(identifier_not_found,Msg,Id:StateInfo,Span),
                  fail % Val = term(Id)  /* Val=fail */
          )
   ).

%translate_environment([],[]) :- !.
%translate_environment([bind(V,_)|T],[V|TT]) :- !, translate_environment(T,TT).
%translate_environment(A,A).

/* -----------------------------*/
/*    b_compute_expressions     */
/* -----------------------------*/

b_compute_expressions([], _, _, [],_WF).
b_compute_expressions([EXPRsHd|EXPRsTl],LocalState,InState,[ValHd|ValTl],WF) :-
   ~~pp_cll(b_compute_expression(EXPRsHd,LocalState,InState,ValHd,WF)),
   b_compute_expressions(EXPRsTl,LocalState,InState,ValTl,WF).

/*
b_compute_expressions_symbolic([],[], _, _, [],_WF).
b_compute_expressions_symbolic([EXPRsHd|EXPRsTl],[SYMBOLIC|TS],LocalState,InState,[ValHd|ValTl],WF) :-
   (SYMBOLIC==symbolic -> EXPRsHd=b(E,T,I), EXPR = b(E,T,[prob_annotation('SYMBOLIC')|I])
     ; EXPR = EXPRsHd),
   ~~pp_cll(b_compute_expression(EXPR,LocalState,InState,ValHd,WF)),
   b_compute_expressions_symbolic(EXPRsTl,TS,LocalState,InState,ValTl,WF).
*/

% a variation of above that delays computing the expressions until Waitflag 1/2 is set
b_compute_assign_expressions_when(EXPRsTl,LocalState,InState,ValTl,WF,OR) :-
   get_assign_expr_priority_wf(OR,WF,LWF),
   b_compute_expressions_when2(EXPRsTl,LocalState,InState,ValTl,WF,LWF).
b_compute_expressions_when2([], _, _, [],_WF,_).
b_compute_expressions_when2([EXPRsHd|EXPRsTl],LocalState,InState,[ValHd|ValTl],WF,LWF) :-
  b_compute_expression_when2(EXPRsHd,LocalState,InState,ValHd,WF,LWF),
  b_compute_expressions_when2(EXPRsTl,LocalState,InState,ValTl,WF,LWF).

get_assign_expr_priority_wf(output_required,WF,LWF) :- !,
  % the Output of the statement is required; do not delay computing it
  get_wait_flag0(WF,LWF).
  %get_wait_flag1(get_assign_expr_priority_wf,WF,LWF).
get_assign_expr_priority_wf(_,WF,LWF) :-
 % the Output is not directly required; compute it later (when we are sure that the statement actually succeeds)
  (preferences:preference(use_smt_mode,true)
   -> Prio = 2.0 % SMT Mode: we are probably doing CBC checking, we may have set up constraints on ValTl
   ; Prio=4000), % unless the expression is inside an sequential composition: nothing can depend on ValTl; TO DO: check if we are inside a sequential composition; if not use inf priority; otherwise use 2.0 or 4
   get_wait_flag(Prio,get_assign_expr_priority_wf,WF,LWF).
   % LHS/RHS of assignments do not affect constraint propagation (at least in normal model checking mode; in CBC or bounded MC this would be different) ??!!

:- block b_compute_expression_when2(?,?,?,?,?,-).
b_compute_expression_when2(eval_must_not_fail(ErrMsg,EXPRsHd),LocalState,InState,Res,WF,_) :- !,
 % catch failure; in case the interpreter fails to catch wd-error itself (e.g., try_find_abort is false)
    if(b_compute_expression(EXPRsHd,LocalState,InState,ValHd,WF),equal_object_wf(ValHd,Res,WF),
        (translate:translate_bexpression_with_limit(EXPRsHd,TS),
         add_wd_error_span(ErrMsg,TS,span_predicate(EXPRsHd,LocalState,InState),WF))).
b_compute_expression_when2(EXPRsHd,LocalState,InState,ValHd,WF,_) :-
    %% translate:print_bexpr(EXPRsHd),nl, print(value(ValHd)),nl, frozen(ValHd,G), print('Frozen: '),print(G),nl, kernel_waitflags:portray_waitflags(WF), %%
    %%debug:time(b_interpreter:b_compute_expression(EXPRsHd,LocalState,InState,ValHd,WF)).
    b_compute_expression(EXPRsHd,LocalState,InState,ValHd,WF).
/* -----------------------------*/
/*     b_execute_statement      */
/* -----------------------------*/
/* b_execute_statement(XMLStatements:input, StateofLocalVariables:input,
                      StateOfGlobalVariables:input,
                      ListOfUpdateBindingsForGlobalVariablesAndResultVariables:output) */



:- assert_pre(b_interpreter:b_execute_statement(Stmt,LS,IS,_OS,_WF,_Path,_OR),
	      (bsyntaxtree:check_if_typed_substitution(Stmt),type_check(LS,store),type_check(IS,store) )).
:- assert_post(b_interpreter:b_execute_statement(_Stmt,_LS,_IS,OS,_WF,_Path,_OR),
	      (type_check(OS,store))).

%b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path) :- !,
%  b_execute_statement(Body,LocalState,InState,OutState,WF,Path). % code below has still problems in INITIALISATION
b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,output_required) :-
   get_texpr_expr(Body,BE),
   BE \= while(_COND,_STMT,_INV,_VARIANT), % also other non-det substitutions ??
   BE \= select(_),
   BE \= select(_,_),
   BE \= case(_,_,_),
   !,
   b_execute_statement(Body,LocalState,InState,OutState,WF,Path,output_required).
b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,OR) :-
   get_wait_flag(2,inner_stmt,WF,LWF), % was 2.0
   clone_wait_flags_from1(WF,InnerWF),
   b_execute_inner_statement(Body,LocalState,InState,OutState,InnerWF,Path,LWF,OR).
:- block b_execute_inner_statement(?,?,?,?,?,?,-,?).
b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,_,OR) :-
   b_execute_statement(Body,LocalState,InState,OutState,WF,Path,OR),
   ground_wait_flag0(WF).

:- use_module(source_profiler,[add_source_location_hits/2]).

b_execute_statement(Stmt, LS, S, OS, WF,Path) :-
   b_execute_statement(Stmt, LS, S, OS, WF,Path,output_not_required).
b_execute_statement(b(Stmt,_,Infos), LS, S, OS, WF,Path,OR) :- !,
   %% check_valid_store(LS,Infos), check_valid_store(S,Infos), %%
    %%  print_term_summary(Stmt), %%
    b_execute_statement2(Stmt, Infos, LS, S, OS, WF,Path,OR).
b_execute_statement(Stmt, LS, S, OS, WF,Path,OR) :-
  add_internal_error('Statement not properly wrapped: ',b_execute_statement(Stmt, LS, S, OS, WF,Path,OR)),
  b_execute_statement2(Stmt, [], LS, S, OS, WF,Path,OR).

% blocks should be removed by type-checker
b_execute_statement2(block(Stmt),_,LS,IS,OS,WF,Path,OR) :- !,
    b_execute_statement(Stmt,LS,IS,OS,WF,Path,OR).  %% NOT COVERED
/*  Var1, ..., VarN := Expr1, ..., ExprN */
/*  also: Var1, fun1(arg), ... := Expr1, Expr2, ... */
b_execute_statement2(assign(LHS,Exprs),Info,LocalState,InState,OutState,WF,assign,OR) :- !,
  %% print(assign(LHS,Exprs)),nl,
   add_source_location_hits(Info,1),
   b_compute_assign_expressions_when(Exprs,LocalState,InState,VALs,WF,OR),% VALs - a list of computed values
   b_assign_values_or_functions(LHS,VALs,LocalState,InState,OutState,WF,OR).
b_execute_statement2(assign_single_id(IDE,Expr),Info,LocalState,InState,OutState,WF,assign,_OR) :-
   get_texpr_id(IDE,ID),!,
   add_source_location_hits(Info,1),
   b_compute_expression(Expr,LocalState,InState,Value,WF), % b_compute_assign_expressions_when
   %b_compute_assign_expressions_when([Expr],LocalState,InState,[Value],WF),
   OutState=[bind(ID,Value)].
b_execute_statement2(choice(ChoiceList),_,LocalState,InState,OutState,WF,Path,OR) :- !,
   /*  Choice LHS Or RHS Or ... End */
   get_wait_flag(2,choice,WF,WFC), %  % was 2.0, maybe use length of ChoiceList as Priority ??
   b_execute_choice(ChoiceList,LocalState,InState,OutState,WF,Path,WFC,OR).
b_execute_statement2(becomes_element_of(LHS,RHS),  /* LHS :: RHS */
   Info,LocalState,InState,OutState,WF,becomes_element_of,_OR) :- !,
   % LHS must be of the form Var1,..,Vark where all variables are disjoint
%% print_message('becomes_element_of'(RHS)),
   add_source_location_hits(Info,1),
   b_compute_expression(RHS,LocalState,InState,ValueSet,WF),
   check_element_of_wf(SingleValue,ValueSet,WF),
      ground_det_wait_flag(WF),
   convert_pairs_into_list(LHS,SingleValue,ValList),
   set_up_typed_localstate(LHS,ValList,TypedVals,[],OutState,positive),
   b_tighter_enumerate_values(TypedVals,WF).
b_execute_statement2(becomes_such(Vars,Condition),  /* Vars : Condition */
   Info,LocalState,InState,OutState,WF,becomes_such,_OR) :- !,
   add_source_location_hits(Info,1),
   b_execute_becomes_such(Vars,Condition,LocalState,InState,OutState,_,WF).
b_execute_statement2(sequence(Seq),   /* LHS ; RHS */
   _,LocalState,InState,OutUpdates,WF,Path,OR) :- !, %pe_print(sequence(Seq,InState)),
   b_execute_sequence(Seq,LocalState,InState,OutUpdates,WF,Path,OR).
b_execute_statement2(precondition(PreCond,Body),_,LocalState,InState,OutState,WF,Path,OR) :- !,
  %print('*** Non-outermost PRE construct'),nl,
  Path=pre(PreRes,IPath),
  b_try_check_boolean_expression(PreCond,LocalState,InState,WF,PreRes),
  ground_det_wait_flag(WF),
  check_precondition_result(PreRes,PreCond,LocalState,InState,WF),
  (PreRes == pred_true -> b_execute_statement(Body,LocalState,InState,OutState,WF,IPath,OR)
    ; PreRes == pred_false -> IPath = not_executed
    ; b_execute_inner_statement(Body,LocalState,InState,OutState,WF,IPath,OR) % with extra waitflag
   ).
b_execute_statement2(assertion(PreCond,Body),_,LocalState,InState,OutState,WF,Path,OR) :- !,
   get_last_wait_flag(assertion,WF,WFC),
  b_execute_assertion(PreCond,Body,LocalState,InState,OutState,WF,Path,WFC,OR).
b_execute_statement2(select(Whens),_, LocalState, InState, OutState,WF,select(Nr,Path),OR) :- !,
  % print(select(Whens)),nl,
   get_texpr_expr(TExpr,select_when(PreCond, Body)),
   nth1(Nr, Whens, TExpr), % THIS IS A CHOICE POINT if length(Whens)>1 ! <-------- TO DO guard
   b_test_boolean_expression(PreCond,LocalState,InState,WF),
   ground_det_wait_flag(WF), % print(try_select_inner_statement(Nr)), translate:print_bexpr(PreCond),nl,
   b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,OR).
b_execute_statement2(select(Whens,Else), Infos, LocalState, InState, OutState,WF,Path,OR) :- !,
  (%print(select_with_else),nl,
   b_execute_statement2(select(Whens), Infos, LocalState, InState, OutState,WF,Path,OR)
  ; % THIS IS A CHOICE POINT  ! <-------- TO DO guard
   findall(PC,
      (member(SW,Whens),get_texpr_expr(SW,select_when(PC,_))), AllPreconds),
    % nl, print(select_else),nl, %%
    b_not_test_list_of_boolean_expression(AllPreconds,LocalState,InState,WF),
    Path = select(else,IPath),
    % print(exec_else),nl, portray_waitflags(WF),nl, %%
    b_execute_inner_statement(Else,LocalState,InState,OutState,WF,IPath,OR)
   % ,print(finished_inner_statement_else_SELECT),nl %, portray_waitflags(WF)
    ).
b_execute_statement2(var(Parameters,Body),   /* VAR _ IN _ substitution */
   _,LocalState,InState,NewOutState,WF,var(Path),OR) :- !,
   /* introduces a new variable that can be assigned to */
   ~~mnf( split_names_and_types(Parameters,Names,_) ),
   ~~pp_mnf(store:delete_variables_from_state(Names,InState,InState1)), % in case any local parameter name is also a constant/variable: remove it % but if we do an operation_call: we leave context and it becomes visible again !!
   % otherwise we may think there are two updates, e.g., in WHILE loop; see Rule_DB_SIGAREA_0030 from test 1303
   ~~mnf( set_up_undefined_localstate(Parameters,InState1,NewInState) ),
   ~~pp_mnf(store:delete_variables_from_state(Names,LocalState,NewLocalState)),
   %check_valid_store(InState,var_in_state), check_valid_store(LocalState,var_local_state),
   b_execute_statement(Body,NewLocalState,NewInState,OutState,WF,Path,OR),
 %%print_message(var_res(OutState)),
   % check_valid_store(OutState,var_out_state),
   filter_results(Parameters,OutState,NewOutState,WF)
   . %, print('Finished get results'(Names,_ResultValues)),nl,  print(new_out(NewOutState)),nl.
   /* it can be ok not to assign to a  local variable */

%TODO(DP, 5.8.2008):
b_execute_statement2(let(Parameters,Defs,Body),   /* LET _ BE _ IN _ */
   _,LocalState,InState,OutState,WF,let(Path),OR) :- !,
   set_up_typed_localstate(Parameters,LocalState,NewLocalState),
   ~~mnf(b_execute_let_definitions(Defs,NewLocalState,InState,WF)),
   b_execute_statement(Body,NewLocalState,InState,OutState,WF,Path,OR).
b_execute_statement2(lazy_let_subst(Id,IdExpr,Body),_,LocalState,InState,OutState,WF,lazy_let(Path),OR) :-
	!,
	add_lazy_let_id_and_expression(Id,IdExpr,LocalState,InState,NewLocalState,WF),
    b_execute_statement(Body,NewLocalState,InState,OutState,WF,Path,OR).
b_execute_statement2(any(Parameters,PreCond,Body),   /* ANY _ WHERE */
   _,LocalState,InState,OutState,WF,any(Bindings,Path),OR) :- !, %statistics(runtime,_),
   %print('ANY: '), print_bexpr(PreCond),nl, %%
   set_up_typed_localstate(Parameters,FreshVars,TypedVals,LocalState,NewLocalState,positive),
   create_any_bindings(Parameters,FreshVars,Bindings),
   b_test_inner_boolean_expression(PreCond,NewLocalState,InState,WF), /* check WHERE */
   %%observe_state(TypedVals,WF,0), b_tracetest_boolean_expression(PreCond,NewLocalState,InState,WF,one), %% comment in for debugging
   b_tighter_enumerate_values(TypedVals,WF), % moved enumeration after test; in case WF already instantiated
  %% print(any_body(Body)),nl,
   b_execute_statement(Body,NewLocalState,InState,OutState,WF,Path,OR). /* should we use b_execute_inner_statement/LWF here ?? */
b_execute_statement2(skip,    /* SKIP */
   Info,_LocalState,_InState,OutState,_WF,skip,_OR) :- !,
   add_source_location_hits(Info,1),
   OutState = [].
b_execute_statement2(parallel(Statements),  /*   LHS  || RHS  */ %@@@
   _,LocalState,InState,OutState,WF,parallel(Paths),OR) :-  !,
   b_execute_statements_in_parallel(Statements,LocalState,InState,OutState,WF,Paths,OR).
b_execute_statement2(init_statement(Stmt),_,LocalState,InState,OutState,WF,Path,OR) :- !,
  /*   Like block: but
   writes error messages if statement fails */ %@@@
     (preferences:get_preference(provide_trace_information,true)
    -> nl,print(' =INIT=> '),translate:print_subst(Stmt),nl, flush_output(user_output),
       %hashing:my_term_hash(Stmt,Hash),print(hash(Hash)),nl,
       statistics(walltime,[Tot1,_])
     ; true),
  if(b_execute_initialisation_statement(Stmt,LocalState,InState,OutState,WF,Path,OR),
     (
      (preferences:get_preference(provide_trace_information,true)
       -> statistics(walltime,[Tot2,_]), Tot is Tot2-Tot1,
          format(' [=OK= ~w ms]~n',[Tot]), flush_output(user_output)
        ; true)
     ),
     (translate:translate_substitution(Stmt,TStmt),
      add_error(initialisation_fails,'Initialisation statement fails: ',TStmt,Stmt), fail
      )
    ).
b_execute_statement2(if(IfList),_,LocalState,InState,OutState,WF,if,OR) :- !,
  %LWF=1,
   LWF='$GENERATE_ON_DEMAND', %
   b_execute_else_list(IfList,LocalState,InState,OutState,WF,LWF,OR).
b_execute_statement2(operation_call(Operation,Results,Parameters),_,LocalState,InState,OutState,WF,
                     operation_call(OperationName,ParamValues,ResultValues,InnerPath),OR) :- !,
    def_get_texpr_id(Operation,op(OperationName)),
    b_execute_operation_with_parameters(OperationName,LocalState,InState,Results,
                                        Parameters,OpOutState,ParamValues,ResultValues,InnerPath,WF,OR),
    %print_message(result_values(OperationName,ResultValues,InnerPath)), %%
   split_names_and_types(Results,ResultingIDs,_),
   store_values(ResultingIDs, ResultValues, ResultsOutState),
     %print(store_values(ResultingIDs, ResultValues, ResultsOutState)),nl,
   combine_updates(ResultsOutState,OpOutState,OutState).

b_execute_statement2(while(COND,STMT,INV,VARIANT),
                    Info,LocalState,InState,OutState,WF,while,_OR) :- !, %pe_print(while(InState)),nl,
  %print_message(while(LocalState,InState)),
  %hit_profiler:add_profile_hit(while(COND,WF),2),
  % we may go around the loop many times: avoid carrying around unused variables
   get_while_reads_write_info_and_filter_state(Info,LocalState,InState,LS,IS,ModVars),
   %print_message(execute_while(LocalState,InState)),
   b_compiler:b_optimize(INV,ModVars,LS,IS,CINV,WF),
   b_compiler:b_optimize(COND,ModVars,LS,IS,CCOND,WF),
   b_compiler:b_optimize(VARIANT,ModVars,LS,IS,CVARIANT,WF),
   %,b_compiler:b_compile(STMT,ModVars,LS,IS,CSTMT,WF)
   % TO DO: maybe exclude in one go; use ordered approach ?
   %exclude(unused_variable(ModVars),LS,LS2), % only keep modified vars; all other compiled
   %exclude(unused_variable(ModVars),IS,IS2), % only keep modified vars; all other compiled
   %  print(while_loop(ModVars,LS,IS)),nl,
   b_compute_expression(CVARIANT,LS,IS,VarVal,WF),
   b_execute_while(CCOND,STMT,CINV,CVARIANT,ModVars,VarVal,LS,IS,OutState,WF).
b_execute_statement2(external_subst_call(FunName,Args),Info,LocalState,State,OutState,WF,external(FunName),_OR) :-
      !, b_compute_expressions(Args,  LocalState,State, EvaluatedArgs, WF),
      % print('EXTERNAL SUBST CALL: '), print(FunName), print(' '), print(EvaluatedArgs),nl,
      call_external_substitution(FunName,EvaluatedArgs,State,OutState,Info,WF).
b_execute_statement2(E,I,LS,S,O,WF,Path,OR) :-
   add_internal_error('Uncovered statement: ',b_execute_statement2(E,I,LS,S,O,WF,Path,OR)),
   print_functor(E),nl,fail.

% -------------------

b_execute_becomes_such(Vars,Condition,LocalState,InState,OutState,Values,WF) :- !,
   % create the LHS variables in the local state
   set_up_typed_localstate(Vars,Values,TypedVals,LocalState,TempLocalState1,positive),
   get_texpr_ids(Vars,Ids),
   % store those values also the outgoing state
   store_values(Ids,Values,OutState),
   % there might be some variables of the form x$0 to access the
   % values before the substitution. We put them into the local state
   insert_before_substitution_variables(Vars,LocalState,InState,TempLocalState1,TempLocalState),
   b_test_inner_boolean_expression(Condition,TempLocalState,InState,WF),
   b_tighter_enumerate_values(TypedVals,WF). % moved enumeration after test, to avoid overhead in case Condition is deterministic (see e.g. WhileAssignTupleChoose_v2)


b_execute_sequence([],_LocalState,_InState,[],_WF,[],_OR).
b_execute_sequence([First|Rest],LocalState,InState,OutUpdates,WF,Path,OR) :-
      (Rest == [] -> b_execute_statement(First,LocalState,InState,OutUpdates,WF,Path,OR)
        ; Path = sequence(Path1,RestPath),
          b_execute_statement(First,LocalState,InState,LHSUpdates,WF,Path1,output_required),
          ground_det_wait_flag(WF), % maybe better to delay next execute_statement ?
          %  print('===> RHS: '), translate:print_subst(b(UnwrappedRest,subst,[])),nl,
          store_intermediate_updates(InState,LHSUpdates,IntermediateState,First),
          % TO DO: statically determine what the Rest can modify and copy over LHSUpdates not modified by Rest into OutUpdates
          b_execute_statement_list_when_state_set(Rest,LocalState,IntermediateState,
                                                  LHSUpdates,OutUpdates,WF,RestPath,OR)
      ).


:- block check_precondition_result(-,?,?,?,?).
check_precondition_result(pred_true,_,_,_,_).
check_precondition_result(pred_false,PreCond,LS,S,WF) :-
      translate_bexpression(PreCond,PreString),
      add_abort_error_span(precondition_error,'Precondition violated: ',PreString,span_predicate(PreCond,LS,S),WF).

create_any_bindings([],[],[]).
create_any_bindings([Param|Prest],[Value|Vrest],[bind(Id,Value)|Brest]) :-
    get_texpr_id(Param,Id),
    create_any_bindings(Prest,Vrest,Brest).

% insert the values of variables before the substitution under an
% alternative name. The alternative name is given as an optional extra
% information in the variable (see type-checking of becomes_such for
% details). Usually (i.e. always) the alternative name is x$0 for the variable x.
insert_before_substitution_variables([],_LocalState,_InState,State,State).
insert_before_substitution_variables([Var|VRest],LocalState,InState,StateA,StateB) :-
	get_texpr_info(Var,Infos),
	( member(before_substitution(_,Alternative),Infos) ->
	    def_get_texpr_id(Var,Id),
	    lookup_value(Id,LocalState,InState,Value),
	    store_value(Alternative,Value,StateA,StateC)
	; otherwise ->
	    %debug_println(19,no_before_subst(Var)),
	    StateA = StateC),
	insert_before_substitution_variables(VRest,LocalState,InState,StateC,StateB).



% treat the bool(.) operator
b_convert_bool(Pred,LocalState,State,WF,PredRes) :-
    get_wait_flag0(WF,WF0), % first give a chance for PredRes to be instantiated
    b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,WF0).

:- block b_convert_bool_wf0(?,?,?,?,-,-).
b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,_) :- PredRes==pred_true,!,
    %% print('bool(.)=TRUE '), translate:print_bexpr(Pred),nl, %%
    b_test_inner_boolean_expression(Pred,LocalState,State,WF).
b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,_) :- PredRes==pred_false,!,
    %% print('bool(.)=FALSE '), translate:print_bexpr(Pred),nl, %%
    b_not_test_inner_boolean_expression(Pred,LocalState,State,WF).
b_convert_bool_wf0(Pred,LocalState,State,WF,PredRes,_) :-
    %% print('bool(.) check : '), translate:print_bexpr(Pred),nl, % obsv(LocalState),
    b_try_check_boolean_expression_wf(Pred,LocalState,State,WF,PredRes ).
    %% print('finished bool(.) check : '), translate:print_bexpr(Pred),nl,translate:print_bstate(LocalState),nl.

%%obsv([]).
%%obsv([bind(Var,Val) | T] ) :- when(ground(Val), format("~w = ~w~n",[Var,Val])), obsv(T).

b_try_check_boolean_expression_wf(Pred,LocalState,State,WF,PredRes) :-
   b_try_check_boolean_expression_lwf(Pred,LocalState,State,WF,PredRes,'$GENERATE_ON_DEMAND' ).
% try to do check_boolean expression; if not possible create choice point if LWF is grounded
b_try_check_boolean_expression_lwf(Pred,LocalState,InState,WF,PredRes,LWF) :-
    %print('Compile: '), translate:print_bexpr(Pred),nl,
    b_compiler:b_optimize(Pred,[],LocalState,InState,CPRED,WF), % this increases chance that check can be applied+avoids multiple computations in two branches
    b_try_check_boolean_expression_lwf_c(CPRED,LocalState,InState,WF,PredRes,LWF).
b_try_check_boolean_expression_lwf_c(b(truth,_,_),_LocalState,_InState,_WF,PredRes,_LWF) :- !,
    % typical case in e.g. IF-THEN-ELSE
    PredRes= pred_true.
b_try_check_boolean_expression_lwf_c(Pred,LocalState,InState,WF,PredRes,LWF) :-
    copy_wf_start(WF,CWF),
    if(b_check_boolean_expression_with_enum(Pred,LocalState,InState,CWF,Res),
     % Note: we do not use a cut here ! usually b_check will not enumerate; but on rare occasions it can (apply function in inverse,...)
        % (print(check(Res)), print(' : '), translate:print_bexpr(Pred),nl,
         (PredRes=Res,
          copy_wf_finish(WF,CWF)
         ),
         b_try_check_failed(Pred,LocalState,InState,WF,PredRes,LWF)
      ).
b_try_check_failed(Pred,LocalState,InState,WF,PredRes,LWF) :-
    %print(choice_point(LWF)),nl, translate:print_bexpr(Pred),nl,
    perfmessagecall(reification_of_check_predicate_failed,translate:print_bexpr(Pred),Pred),
    (LWF=='$GENERATE_ON_DEMAND' -> get_binary_choice_wait_flag('$GENERATE_ON_DEMAND',WF,WF1) ; WF1=LWF),
    b_try_check_boolean_expression_lwf_block(Pred,LocalState,InState,WF,PredRes,WF1).

% a version of b_check_boolean_expression which will enumerate its result as last resort
% should we only do this in SMT mode : preferences:preference(use_smt_mode,true) ?
b_check_boolean_expression_with_enum(Pred,LocalState,State,WF,Res) :-
   b_check_boolean_expression(Pred,LocalState,State,WF,Res),
   (nonvar(Res) -> true
    ; get_last_wait_flag(b_check_boolean_expression_with_enum,WF,LWF),
      enum_pred_result(Res,LWF)). % before doing infinite enumerations: better do case-distinction here

:- block enum_pred_result(-,-).
%enum_pred_result(X,Y) :- print(enum_pred_result(X,Y)),nl,fail. %%
enum_pred_result(pred_true,_).
enum_pred_result(pred_false,_).

:- block b_try_check_boolean_expression_lwf_block(?,?,?, ?,-,-).
b_try_check_boolean_expression_lwf_block(Pred,LocalState,InState,WF,PREDRES,_) :-
    PREDRES \== pred_false,
    % print(' TRUE +++++> '), translate:print_bexpr(Pred),nl,
    b_test_inner_boolean_expression(Pred,LocalState,InState,WF),
    PREDRES = pred_true. % set it only after we have actually checked the predicate; e.g., for IF we may execute rest of while loop etc... before actually having set up the constraint Pred; See Hansen23_WhilePerformance
b_try_check_boolean_expression_lwf_block(Pred,LocalState,InState,WF,PREDRES,_) :-
    PREDRES \== pred_true,
    % print(' FALSE +++++> '), translate:print_bexpr(Pred),nl,
    b_not_test_inner_boolean_expression(Pred,LocalState,InState,WF),
    PREDRES = pred_false.

% try to do check_boolean expression; if not possible create choice point *directly*
b_try_check_boolean_expression(Pred,LocalState,InState,WF,PredRes) :-
    b_compiler:b_optimize(Pred,[],LocalState,InState,CPRED,WF), % this increases chance that check can be applied+avoids multiple computations in two branches
    b_try_check_boolean_expression_c(CPRED,LocalState,InState,WF,PredRes).
b_try_check_boolean_expression_c(Pred,LocalState,InState,WF,PredRes) :-
    copy_wf_start(WF,CWF),
    if(b_check_boolean_expression(Pred,LocalState,InState,CWF,Res),
        (PredRes=Res,copy_wf_finish(WF,CWF)),
        b_try_check_boolean_expression_c2(Pred,LocalState,InState,WF,PredRes)
      ).
b_try_check_boolean_expression_c2(Pred,LocalState,InState,WF,pred_true) :-
    %hit_profiler:add_profile_hit(b_try_check_failed(Pred),3),
    b_test_inner_boolean_expression(Pred,LocalState,InState,WF).
b_try_check_boolean_expression_c2(Pred,LocalState,InState,WF,pred_false) :-
    b_not_test_inner_boolean_expression(Pred,LocalState,InState,WF).


% WHILE SUBSTITUTION TREATMENT:

:- if(environ(prob_noopt_mode,true)).
get_while_reads_write_info_and_filter_state(Info,LocalState,InState,LS,IS,ModVars) :-
    member(modifies(ModVars),Info),!,
    LS=LocalState, IS=InState. % do not filter state
:- endif.
get_while_reads_write_info_and_filter_state(Info,LocalState,InState,LS,IS,ModVars) :-
    member(reads(ReadVars),Info),
    member(modifies(ModVars),Info),!,
  % we may go around the loop many times: avoid carrying around unused variables
    ord_union(ReadVars,ModVars,ReadOrModVars),
    %print(read_mod(ReadVars,ModVars)),nl,
    exclude(unused_localvariable(ReadOrModVars),LocalState,LS), %print(filter(LocalState)),nl,print(new(LS)),nl,
    exclude(unused_variable(ReadOrModVars),InState,IS).
       % TO DO: investigate whether we should use ord_intersect; ReadVars is already sorted
get_while_reads_write_info_and_filter_state(Info,LocalState,InState,LS,IS,ModVars) :-
    add_internal_error('No reads & modifies info for while loop: ',get_while_reads_write_info_and_filter_state(Info,LocalState,InState,LS,IS,ModVars)),
    ModVars=_, % all variables assumed to be modified
    LS=LocalState, IS=InState.

unused_variable(List,bind(Var,_)) :-
    % b_is_variable(Var), % TO DO: also include constants in reads Info ?
    ord_nonmember(Var,List).
unused_localvariable(List,bind(Var,Val)) :- ord_nonmember(Var,List),
    (nonvar(Val),Val=(_,_) -> atom_codes(Var,Codes), Codes \= [64|_]  % not a lazy_let ID starting with @
     % TO DO: either ensure lazy-let ids are also included in reads/writes ! or mark Val not as pair but using a special functor in add_lazy_let_id_to_local_state ?
      ; true).


:- block lazy_exclude(-,?,?).
lazy_exclude([],_,[]).
lazy_exclude([bind(Var,Val)|T],ModVars,Res) :- lazy_exclude_aux(Var,Val,T,ModVars,Res).
:- block lazy_exclude_aux(-,?,?,?,?).
lazy_exclude_aux(Var,Val,T,ModVars,Res) :-
   (member(Var,ModVars) -> Res = [bind(Var,Val)|RT]
     ; Res = RT),
   lazy_exclude(T,ModVars,RT).


:- use_module(kernel_tools,[ground_state_check/2]).
% outermost while code:; at next iteration b_execute_while1 will be called.
:- block b_execute_while(?,?,?,?,?,-, ?,?,?, ?). % we block on Variant Value int(VarVal)
b_execute_while(COND,STMT,INV,VARIANT,ModVars,int(VarVal),LocalState,InState,OutState,WF) :-
   %kernel_waitflags:get_idle_wait_flag(b_execute_while,WF,LWF),
   (number(VarVal),VarVal =< 50
    -> LWF=VarVal % we will wait anyway on GrInState below; no danger for WD issues in in b_optimize
    ; kernel_waitflags:get_last_wait_flag(b_execute_while1,WF,LWF)
      % see also tests 979, 981
    ),
   % before entering while loop, we wait for co-routines to complete that may have activated b_execute_while
   %(debug_mode(off) -> true ;  gensym:gensym(while,WNr), print(execute_while_idle(VarVal,LWF,WNr)),nl, error_manager:print_error_span(COND)),
   b_execute_while_idle(COND,STMT,INV,VARIANT,ModVars,int(VarVal),LocalState,InState,OutState,WF,LWF).

:- block b_execute_while_idle(?,?,?,?,?,?, ?,?,?, ?,-). % we block on enumeration starting wait_flag
b_execute_while_idle(COND,STMT,INV,VARIANT,ModVars,int(VarVal),LocalState,InState,OutState,WF,_) :-
   %ground_constraintprop_wait_flags(WF),
   (number(VarVal), VarVal>50
   % if VarVal > 50 -> worthwhile to compile STMT
      -> b_compiler:b_optimize(STMT,ModVars,LocalState,InState,CSTMT,WF),
         % print(compiled_while(ModVars)),nl, translate:print_subst(CSTMT),nl,
         OutState=WhileOutState,
         exclude(unused_variable(ModVars),LocalState,LS), % only keep modified vars; all other compiled
         exclude(unused_variable(ModVars),InState,IS) % only keep modified vars; all other compiled
      ;  CSTMT=STMT, LS=LocalState, IS=InState,
         lazy_exclude(WhileOutState,ModVars,OutState) % only copy over those variables (potentially) modified by While loop
    % Note: it is important that we only have one copy of each modified variable in OutState
    ),
    ground_state_check(InState,GrInState),
    when(nonvar(GrInState),
        b_execute_while_loop(COND,CSTMT,INV,VARIANT,none,VarVal,LS,IS,WhileOutState,WF)).

:- block b_execute_while1(?,?,?,?,?,-,?,?,?,?).
b_execute_while1(COND,STMT,INV,VARIANT,PrevVariantValue,int(VarVal),LocalState,InState,OutState,WF) :-
   %  get_last_wait_flag(while(VarVal),WF,LWF),
   ground_constraintprop_wait_flags(WF),
   b_execute_while_loop(COND,STMT,INV,VARIANT,PrevVariantValue,VarVal,LocalState,InState,OutState,WF).

% block on PrevVarVal and VarVal not yet needed here; but we need it for b_execute_while_body below
:- block b_execute_while_loop(?,?,?,?,-,?,?,?,?,?), b_execute_while_loop(?,?,?,?,?,-,?,?,?,?).
b_execute_while_loop(COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF) :-
        b_try_check_boolean_expression_c(INV,LocalState,InState,WF,INVRes),
        %ground_constraintprop_wait_flags(WF), % seems to slow down interpretation
        b_check_execute_while_loop1(INVRes,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF).

:- block b_check_execute_while_loop1(-,?,?,?,?,?,?,?,?,?,?).
b_check_execute_while_loop1(pred_false,_COND,_STMT,INV,_VAR,_PrevVarVal,VarVal,LocalState,InState,_OutState,WF) :-
        add_abort_error_span(while_invariant_violation,
         'While INVARIANT VIOLATED, current VARIANT value: ',VarVal,span_predicate(INV,LocalState,InState),WF).
b_check_execute_while_loop1(pred_true,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF) :-
        b_try_check_boolean_expression_c(COND,LocalState,InState,WF,CondRes),
        %ground_constraintprop_wait_flags(WF), % seems to slow down interpretation
        b_check_execute_while_loop2(CondRes,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF).

:- block b_check_execute_while_loop2(-,?,?,?,?,?,?,?,?,?,?).
b_check_execute_while_loop2(pred_false,_COND,_STMT,_INV,_VARIANT,_PrevVarVal,_VarVal,_LS,InState,OutState,_WF) :-
      % print(while_finished(_VarVal)), print(' '), translate:print_bexpr(_VARIANT),nl, trace,
        OutState=InState. % we copy over the while InState: the updates will be computed at the end of the while loop
b_check_execute_while_loop2(pred_true,COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF) :-

        %hit_profiler:add_profile_hit(b_execute_while_body(STMT,VarVal,PrevVarVal,LocalState),2),
        % TO DO: maybe only compute VARIANT now and not above ??
        % b_compute_expression(VARIANT,LocalState,InState,VarVal,WF),
        % ground_constraintprop_wait_flags(WF), %% moved here because VARIANT computed later
        b_execute_while_body(COND,STMT,INV,VARIANT,PrevVarVal,VarVal,LocalState,InState,OutState,WF).

%:- block b_execute_while_body(?,?,?,?,-,?,?,?,?,?), b_execute_while_body(?,?,?,?,?,-,?,?,?,?).
% we must ensure VarVal and PrevVaraintValue known
% the block declaration above for b_execute_while_loop already ensures this
b_execute_while_body(_COND,_STMT,_INVARIANT,VAR,_PV,VarVal,LocalState,InState,_OutState,WF) :-
      VarVal<0, !, % more efficient than is_not_natural(int(VarVal)),
      ground_constraintprop_wait_flags(WF),
      % add_abort_error_span will fail; hence no need to set _OutState
      add_abort_error_span(while_variant_error,'While VARIANT not NATURAL:',VarVal,span_predicate(VAR,LocalState,InState),WF).
b_execute_while_body(_COND,_STMT,_INVARIANT,VAR,PrevVariantValue,VarVal,LocalState,InState,_OutState,WF) :-
      % check no longer required VarVal>=0,%is_natural(int(VarVal),WF),
      PrevVariantValue \= none,
      PrevVariantValue =< VarVal, !, %less_than_equal(int(PrevVariantValue),int(VarVal)),
      ground_constraintprop_wait_flags(WF),
      % add_abort_error_span will fail; hence no need to set _OutState
      ajoin(['VARIANT = ',VarVal,', previous value = ',PrevVariantValue],Details),
      add_abort_error_span(while_variant_error,'While VARIANT not decreasing:',
                           Details,span_predicate(VAR,LocalState,InState),WF).
b_execute_while_body(COND,STMT,INV,VARIANT,_PrevVariantValue,VarVal,LocalState,InState,OutState,WF) :-
      % check no longer required due to cuts above %VarVal>=0, %is_natural(int(VarVal),WF),
      %(PrevVariantValue==none -> true  ; less_than(int(VarVal),int(PrevVariantValue))),
      %ground_constraintprop_wait_flags(WF), % seems to slow down interpretation
      % print(executing_while(VarVal)), print(' '), translate:print_bexpr(VARIANT),nl,
      copy_wf_start(WF,CWF),
      b_execute_statement(STMT,LocalState,InState,LHSUpdates,CWF,_Path1,output_required),
      copy_wf_finish(WF,CWF),
      ground_det_wait_flag(WF),
      store_intermediate_updates(InState,LHSUpdates,IntermediateState,STMT),
      check_state_skeleton_bound(LHSUpdates,Bound), % otherwise variable lookups will cause problems; ensures IntermediateState skeleton bound
      b_execute_while_when_state_set(COND,STMT,INV,VARIANT,VarVal,LocalState,IntermediateState,OutState,WF,Bound).

:- block b_execute_while_when_state_set(?,?,?,?,?,?,?,?,?,-).
b_execute_while_when_state_set(COND,STMT,INV,VARIANT,PrevVariantValue,LocalState,State,Res,WF,_) :-
    %format('~nComputing VARIANT (prev ~w)~n',[PrevVariantValue]), translate:print_bstate(State),nl, portray_waitflags(WF),nl,
    copy_wf_start(WF,CWF),
    b_compute_expression(VARIANT,LocalState,State,VarVal,CWF),
    copy_wf_finish(WF,CWF),
    b_execute_while1(COND,STMT,INV,VARIANT,PrevVariantValue,VarVal,LocalState,State,Res,WF).

% ------------- IF-THEN-ELSE

:- block b_execute_else_list2(-,?,?,?, ?,?,?,-,?). % should we also unblock if LWF set ? needed for seq. composition to set up state ??
b_execute_else_list2(PredRes,Body,Rest,LocalState,InState,OutState,InnerWF,LWF,OR) :-
   b_execute_else_list3(PredRes,Body,Rest,LocalState,InState,OutState,InnerWF,LWF,OR).

b_execute_else_list3(pred_true,Body,_,LocalState,InState,OutState,WF,_,OR) :-
     % print('IF TEST: '), nl, %print_subst(Body), nl,
  b_execute_statement(Body,LocalState,InState,OutState,WF,_Path,OR).
b_execute_else_list3(pred_false,_Body,Rest,LocalState,InState,OutState,WF,LWF,OR) :-
     % print('IF - ELSE NEG TEST: '), print(LWF),nl, % print(LocalState),nl, print(InState),nl,
  b_execute_else_list(Rest,LocalState,InState,OutState,WF,LWF,OR).

b_execute_else_list([],_,_State,[],_WF,_LWF,_OR). /* skip */
b_execute_else_list([TExpr|Rest],LocalState,InState,OutState,WF,LWF,OR) :-
  get_texpr_expr(TExpr,if_elsif(Test,Body)),
  % get_enumeration_finished_wait_flag(WF,LWF),
  InnerWF=WF, %clone_wait_flags_from1(WF,InnerWF),
  b_try_check_boolean_expression_lwf(Test,LocalState,InState,InnerWF,PredRes,LWF),
  %ground_wait_flag0(InnerWF),
  (nonvar(PredRes) -> b_execute_else_list3(PredRes,Body,Rest,LocalState,InState,OutState,InnerWF,LWF,OR)
   ; (LWF == '$GENERATE_ON_DEMAND' ->
          (preferences:preference(use_smt_mode,true) -> Prio = 2 % was 2.0
           ; OR == output_required -> Prio = 2000
           ; Prio = 200000 %last_finite_priority(Prio)
          ),  % TO DO: check if there is a guard inside; if not we can wait much longer; unless we are in CBC mode
          get_wait_flag(Prio,if_else,WF,LWF2)
       ; LWF2=LWF),
     b_execute_else_list2(PredRes,Body,Rest,LocalState,InState,OutState,InnerWF,LWF2,OR)
   ).



:- block b_execute_choice(?,?,?,?,?,?,-,?).
b_execute_choice(ChoiceList,LocalState,InState,OutState,WF,choice(Nr,Path),_WFC,OR) :-
   nth1(Nr,ChoiceList,Choice),
   b_execute_statement(Choice,LocalState,InState,OutState,WF,Path,OR).

:- block b_execute_assertion(?,?,?,?,?,?,?,-,?).
b_execute_assertion(PreCond,Body,LocalState,InState,OutState,WF,Path,_,OR) :-
   (  ~~pp_cll(b_test_boolean_expression(PreCond,LocalState,InState,WF)),
         /* ASSERT SATISFIED */
      ground_det_wait_flag(WF),
      Path = assertion(IPath),
      b_execute_statement(Body,LocalState,InState,OutState,WF,IPath,OR)
   ;
      b_not_test_boolean_expression(PreCond,LocalState,InState,WF),
          /* ASSERT VIOLATED */
      ground_det_wait_flag(WF),
      Path = assertion_violated,
      translate_bexpression(PreCond,PreString),
      add_abort_error_span(assert_error,'ASSERT violated: ',PreString,span_predicate(PreCond,LocalState,InState),WF)
      %,store_value(abort__flag__info,term(abort(assertion_violated(PreCond))),[],OutState)
  ).


:- block check_state_skeleton_bound(-,?).
check_state_skeleton_bound([],true).
check_state_skeleton_bound([bind(Var,_)|T],Res) :-
  check_state_skeleton_bound_aux(Var,T,Res).

:- block check_state_skeleton_bound_aux(-,?,?).
check_state_skeleton_bound_aux(Var,T,Res) :-
   (atomic(Var) -> true ; print('*** variable name not atomic: '), print(Var),nl),
   check_state_skeleton_bound(T,Res).


b_execute_statement_list_when_state_set(StmtList,LS,State,USoFar,Res,WF,Path,OR) :-
   check_state_skeleton_bound(State,Bound), % otherwise variable lookups will cause problems
   b_execute_statement_when_state_set2(StmtList,LS,State,USoFar,Res,WF,Path,Bound,OR).
:- block b_execute_statement_when_state_set2(?,?,?,?,?,?,?,-,?).
b_execute_statement_when_state_set2(StmtList,LS,State,UpdatesSofar,Res,WF,Path,_,OR) :-
   copy_wf_start(WF,CWF),
   b_execute_sequence(StmtList,LS,State,StmtUpdates,CWF,Path,OR),
   copy_wf_finish(WF,CWF),
   merge_updates(UpdatesSofar,StmtUpdates,Res).


store_values([], [], []).
store_values([IDsHd|IDsTl], [VALsHd|VALsTl], [bind(IDsHd,VALsHd)|OutStateTl]) :-
  store_values(IDsTl, VALsTl, OutStateTl).




    /*  the B syntax  does allow multiple definitions in a LET: */

b_execute_let_definitions(Pred,LocalState,State,WF) :-
   is_a_conjunct(Pred,LHS,RHS),!,
   b_execute_let_definitions(LHS,LocalState,State,WF),
   b_execute_let_definitions(RHS,LocalState,State,WF).
b_execute_let_definitions(b(equal(LHS,RHS),_,_),LocalState,State,WF) :- !,
   b_compute_expression(LHS,LocalState,State,LHSVal,WF),
   b_compute_expression(RHS,LocalState,State,RHSVal,WF),
   % TO DO: ideally we should check if LHS is a declared identifier and is only defined once
   kernel_objects:equal_object_optimized(LHSVal,RHSVal,b_execute_let_definitions).
b_execute_let_definitions(b(lazy_let_pred(Id,AssignmentExpr,Body),_,_),LocalState,State,WF) :- !,
   add_lazy_let_id_and_expression(Id,AssignmentExpr,LocalState,State,NewLocalState,WF),
   b_execute_let_definitions(Body,NewLocalState,State,WF).
b_execute_let_definitions(P,_L,_S,_WF) :-
   translate:translate_bexpression_with_limit(P,70,TP),
   add_internal_error('Let contains a predicate which is not an equality: ',TP),fail.


/* -----------------------------------------*/
/*             b_assign_function            */
/* -----------------------------------------*/

b_assign_values_or_functions([],[],_LS,_S,[],_WF,_OR).
b_assign_values_or_functions([TExpr|ERest],[Value|VRest],LS,S,[bind(Id,New)|NRest],WF,OR) :-
    get_texpr_expr(TExpr,Expr),
    b_assign_value_or_function(Expr,Value,LS,S,Id,New,WF,OR),
    b_assign_values_or_functions(ERest,VRest,LS,S,NRest,WF,OR).

:- use_module(kernel_tools,[ground_value_check/2]).
b_assign_value_or_function(identifier(Id),Value,_LS,_S,LHS_Id,FullValue,WF,_OR) :- !,
    LHS_Id=Id,
    equal_object_wf(FullValue,Value,b_assign_value_or_function(Id),WF).
b_assign_value_or_function(function(Fun,Arg),Value,LocalState,InState,Id,FullValue,WF,OR) :- !,
    %b_compute_expression(Arg,LocalState,InState,ArgVal,WF),
    %expand_global_sets_but_not_closures(ArgVal,EArgVal),
    FunMNF = eval_must_not_fail('Assignment left-hand-side (or part thereof) could not be evaluated: ',Fun), % could be f(..)(..)
    b_compute_assign_expressions_when([Arg,FunMNF],LocalState,InState,[ArgVal,FunVal],WF,OR), % or should we pass the OR info around ?
    /* we do not examine local state: these vars cannot be modified by operations !? */
    (OR=output_required -> override(FunVal,ArgVal,Value,NewFun,WF)
       ; ground_value_check((FunVal,ArgVal,Value),GRV),
         % wait until all arguments known so that we can effiently compute override
         when(nonvar(GRV),override(FunVal,ArgVal,Value,NewFun,WF))),
    get_texpr_expr(Fun,Lhs),
    b_assign_value_or_function(Lhs,NewFun,LocalState,InState,Id,FullValue,WF,OR).
b_assign_value_or_function(LHS,Value,LocalState,InState,Id,FullValue,WF,OR) :-
    add_internal_error('Illegal LHS for assignment (must be function(identifier) or identifier): ',
                       b_assign_value_or_function(LHS,Value,LocalState,InState,Id,FullValue,WF,OR)),
    fail.



/* -----------------------------------------*/
/*     b_execute_statements_in_parallel     */
/* -----------------------------------------*/
/* execute a parallel composition of statements */
:- assert_pre(b_interpreter:b_execute_statements_in_parallel(Ss,LS,State,_Os,WF,_Paths,_OR),
                (ground_check(Ss),type_check(LS,store),
                 type_check(State,store),type_check(WF,wait_flag))).
:- assert_post(b_interpreter:b_execute_statements_in_parallel(_Ss,_LS,_State,Os,WF,_Paths,_OR),
                (type_check(Os,store),type_check(WF,wait_flag))).

b_execute_statements_in_parallel([],_,_,[],_WF,[],_OR).
b_execute_statements_in_parallel([Stmt|TS],LocalState,InState,OutState,WF,[Path1|TPath],OR) :-
  % print('PAR: '),translate:print_subst(Stmt),nl,
  b_execute_statement(Stmt,LocalState,InState,OutState1,WF,Path1,OR),
  combine_updates(OutState1,OutState2,OutState), /* merge updates */
  % print('PAR 1st branch '), print(OutState),nl,
  b_execute_statements_in_parallel(TS,LocalState,InState,OutState2,WF,TPath,OR).


%:- block combine_updates(-,-,?).
%combine_updates(A,B,C) :- (B==[] -> A=C ; combine_updates_aux(A,B,C)).

% TO DO: investigate whether we could also unblock on second argument and copy it to result
:- block combine_updates(-,?,?). % can also be used to merge states
combine_updates([],X,R) :- !,R=X.
combine_updates([H|T],Y,Res) :- !, Res=[H|Z], combine_updates(T,Y,Z).
combine_updates(X,Y,Z) :- add_internal_error('Illegal call: ',combine_updates(X,Y,Z)),fail.



b_execute_initialisation_statement(Stmt,LS,In,Out,_OUTERWF,Path,OR) :-
    init_wait_flags(WF), % each initialisation can be executed in isolation (except in CBC mode)
    b_execute_statement(Stmt,LS,In,Out,WF,Path,OR),
    ground_wait_flags(WF).

% replace certain parallels by init_parallel for better error reporting in INITIALISATION
% will not be used in CBC mode (by set_up_initialisation or tc_initialise)
translate_init_statements(b(Stmt,subst,I),b(TStmt,subst,I)) :- %print_term_summary(Stmt),
  tp2(Stmt,I,TStmt).
tp2(assign(LHS,RHS),I,parallel(S)) :-  LHS=[_,_|_],
   % translate multiple assignment into parallel assignments
   maplist(generate_assignment(I),LHS,RHS,S),
   !.
tp2(Stmt,I,init_statement(b(Stmt,subst,I))) :- %get_preference(provide_trace_information,true),
   wrap_stmt(Stmt),
   !.
% TO DO: cover WHILE
tp2(parallel(S),_,parallel(TS)) :- maplist(translate_init_statements,S,TS),
   !.
tp2(sequence(A),_,sequence(LA)) :- !,maplist(translate_init_statements,A,LA).
tp2(if(A),_,if(LA)) :- !, maplist(translate_init_statements,A,LA).
tp2(if_elsif(Test,Body),_,if_elsif(Test,LBody)) :- !,
   translate_init_statements(Body,LBody).
tp2(X,_,X). %  :- functor(X,F,A), nl,nl,print(uncovered(F/A)),nl,nl.

wrap_stmt(assign(_,_)).
wrap_stmt(becomes_element_of(_,_)).
wrap_stmt(becomes_such(_,_)).
wrap_stmt(assign_single_id(_,_)).

generate_assignment(I,V,RHS,b(init_statement(b(Assign,subst,I)),subst,I)) :-
   Assign = assign([V],[RHS]).



/* ---------------------------------------*/
/*     b_execute_top_level_statement      */
/* ---------------------------------------*/

/* this is just like b_execute_statement with the
 only difference that a PRE-condition is treated like
 a select, i.e., it does not generate an abort if used
 outside of its condition */
b_execute_top_level_statement(TExpr,LocalState,InState,OutState,WF,Path,OR) :-
   get_texpr_expr(TExpr,Stmt),
   b_execute_top_level_statement2(Stmt,TExpr,LocalState,InState,OutState,WF,Path,OR).
b_execute_top_level_statement2(precondition(PreCond,Body),TExpr,
   LocalState,InState,OutState,WF,Path,OR) :-
   preferences:preference(treat_outermost_pre_as_select,true),!,
   %debug_println(9,'Top level PRE treated as SELECT'),
   (pge_algo:do_not_evaluate_guard ->
	    debug_println(9,pge_guard_will_not_be_evaluated)
   ; (preference(ltsmin_do_not_evaluate_top_level_guards,true),
      get_texpr_info(TExpr,Info),member(prob_annotation('LTSMIN-GUARD'),Info)) ->
	    debug_println(9,ltsmin_guard_will_not_be_evaluated)
   ; b_test_boolean_expression(PreCond,LocalState,InState,WF)
   ),
  % b_tracetest_boolean_expression(PreCond,LocalState,InState,WF,one),  % comment in for debugging
   b_execute_statement(Body,LocalState,InState,OutState,WF,Path,OR).
b_execute_top_level_statement2(lazy_let_subst(Id,IdExpr,Body),_,
  LocalState,InState,OutState,WF,lazy_let(Path),OR) :-
  	!,
  	add_lazy_let_id_and_expression(Id,IdExpr,LocalState,InState,NewLocalState,WF),
    b_execute_top_level_statement(Body,NewLocalState,InState,OutState,WF,Path,OR).
b_execute_top_level_statement2(_,TExpr,LocalState,InState,OutState,WF,Path,OR) :-
  b_execute_statement(TExpr,LocalState,InState,OutState,WF,Path,OR).


/* -----------------------------*/
/*      b_execute_operation     */
/* -----------------------------*/
/* execute a single operation */

try_get_op_name(FullOperation,_Name) :- var(FullOperation),!.
try_get_op_name('-->'(FO,_),Name) :- !,try_get_op_name(FO,Name).
try_get_op_name(FullOperation,Name) :- functor(FullOperation,Name,_).

b_get_machine_operation_for_animation('$initialise_machine',[],Parameters,Body,OType,_TopLevel) :-
     b_get_machine_variables(DeclaredVars), Parameters=DeclaredVars,
     b_get_initialisation_from_machine(Body,OType).
b_get_machine_operation_for_animation(Name,Results,Parameters,Body,OType,TopLevel) :-
    preference(show_eventb_any_arguments,EVENTB),
    b_get_machine_operation(Name,Results,RealParameters,RealBody,OType,_OpPos),
    % TO DO: check whether we should only keep b_top_level_operation(Name) in case TopLevel==true !
    (Results=[],TopLevel==true,EVENTB=true, RealParameters=[],
     translate_any_into_parameters(RealBody,FakeParameters,FakeBody)
     % Then pretend the any parameters are parameters of the event
      -> Parameters=FakeParameters, Body=FakeBody
        % , print(faked(Name,Parameters,Body)),nl
      ;  Parameters=RealParameters, Body=RealBody
    ).

translate_any_into_parameters(b(E,Type,Info),Parameters,NewSubst) :-
   translate_any_into_parameters_aux(E,Type,Info,Parameters,NewSubst).
translate_any_into_parameters_aux(any(Parameters,PreCond,ABody),Type,Info,Parameters,NewSubst) :-
    NewSubst=b(select([b(select_when(PreCond,ABody),Type,Info)]),Type,Info). % TO DO: go deeper ? (nested ANY ?)
translate_any_into_parameters_aux(let(Parameters,Defs,LBody),Type,Info,Parameters,NewSubst) :-
    NewSubst=b(select([b(select_when(Defs,LBody),Type,Info)]),Type,Info). % TO DO: go deeper ? ANY could have been translated into several LETs ?
translate_any_into_parameters_aux(lazy_let_subst(Id,IdExpr,Body),Type,Info,Parameters,NewSubst) :-
    NewSubst=b(lazy_let_subst(Id,IdExpr,NewInnerSubst),Type,Info),
    translate_any_into_parameters(Body,Parameters,NewInnerSubst).


:- assert_pre(b_interpreter:b_execute_top_level_operation_update(_Name,_FullOperation,InState,_NewState,_P),
                (ground_check(InState),type_check(InState,store))).
:- assert_post(b_interpreter:b_execute_top_level_operation_update(Name,_FullOperation,_InState,NewState,_Path),
                (type_check(NewState,store),ground_check(Name))).

:- if( environ(use_pe_cache, true)).
% the specialized code is in pe_cache.pl
:- elif(\+ environ(partially_evaluate, true)).

b_execute_top_level_operation_update(Name,FullOperation,InState,Updates,Path) :-
    b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF),
    ground_wait_flags(WF).
    %store:store_updates_and_normalise(Updates,InState,NewState).

:- else.
/* partial evaluation enabled */

:- unfold(pe_execute_top_level_operation_update(_Name,_FullOperation,_InState,_Updates,_Path)).
pe_execute_top_level_operation_update(Name,FullOperation,InState,Updates,Path) :-
    b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF),
    ground_wait_flags(WF).

b_execute_top_level_operation_update(Name,FullOp,InState,Updates,Path) :-
    %print(compile_op(Name,FullOp)),nl,
    % THIS SHOULD BE DONE BY AN ANNOTATION: start_specialization(p(static,dynamic,...)) rather than by hand
    MemoCode=generalize_call(list(term(bind,[atomic,dynamic])),InState,StateSkel),
    %nl,print(partially_evaluating_operation(Name)),nl,
    %(Name==new -> trace ; true),
    fast_partially_evaluate_and_load(pe_execute_top_level_operation_update(Name,FullOp,StateSkel,Updates,Path),
                                     ResCall,MemoCode,[ground/StateSkel]),
    StateSkel=InState, call(ResCall).
:- endif.

b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF) :-
    init_wait_flags(WF),
    try_get_op_name(FullOperation,Name), b_top_level_operation(Name),
    b_execute_operation2(Name,FullOperation,InState,Updates,_,[],_,true,Path,WF,output_not_required).

% for use by CBC:
b_execute_top_level_operation_wf(Name,FullOperation,InState,NewState,Path,WF) :-
    try_get_op_name(FullOperation,Name), b_top_level_operation(Name),
    % call before execute operation to copy over results immediately in case b_execute_operation2 generates choice points:
    store_intermediate_updates(InState,NewOutState,NewState,operation(Name)), % no normalising is better for constraint propagation
    % used to call: store:lazy_store_updates_and_normalise(NewOutState,InState,NewState),
    b_execute_operation2(Name,FullOperation,InState,NewOutState,_,[],_,true,Path,WF,output_required).
     % we should ensure that Operation part is executed to instantiate NewOutState skeleton



% Note : if the argument OR (OutputRequired) has value output_required this indicates that the output
% an operation or its effect is required (e.g., because of sequential composition) and the constraint solver
% should not delay computing the effect of the operation
% Atelier-B Handbook 6.16: [R <- op (E)]P   =   [X := E ; S ; R :=Y] P   where op defined by  X <- op(Y) = S
b_execute_operation_with_parameters(Name,LocalState,InState,Results,Parameters,OutState,ParamValues,ResultValues,Path,WF,OR) :-
     (b_get_machine_operation(Name,_,_,_)
      -> b_compute_expressions(Parameters,LocalState,InState,ParamValues,WF),
         % TO DO : remove local vars from InState
         (b_operation_reads_output_variables(Name,_Reads,TReadsInfo)
           -> % format('Transferring outputs ~w ~w~n  ~w~n',[Name,_Reads,TReadsInfo]),
              maplist(output_parameters_value(LocalState,InState,Results),TReadsInfo,ResultInputValues)
              %,print(ResultInputValues),nl
           ; ResultInputValues=[]),
         b_execute_operation2(Name,_,InState,OutState,ParamValues,ResultInputValues,ResultValues,false,Path,WF,OR)
      ;  add_internal_error('Unknown B operation, cannot call: ',
                            b_execute_operation_with_parameters(Name,LocalState,InState,Results,Parameters,
                                                                OutState,ParamValues,ResultValues,Path,WF,OR)),
         fail
     ).

output_parameters_value(LocalState,InState,Results,reads_result(Index,ID,_Type),bind(ID,Val)) :-
    nth1(Index,Results,ReadResultTID),
    def_get_texpr_id(ReadResultTID,ReadID),
    (lookup_value_without_check(ReadID,LocalState,InState,Val)
     -> true %check_not_undefined(Val,ReadID,ReadResultTID)
      ; add_error(output_parameters_value,'Could not obtain value for read output parameter: ',ReadID,ReadResultTID),
        Val = term(undefined)).

% TO DO: use this ?
%:- block check_not_undefined(-,?,?).
%check_not_undefined(term(undefined),ReadID,ReadResultTID) :- !,
%   add_error(output_parameters_value,'Output parameter not assigned to: ',ReadID,ReadResultTID).
%check_not_undefined(_,_,_).

% ResultInputValues: in Atelier-B result parameters can also be read !
b_execute_operation2(Name,Operation,InState,Updates,ParaValues,ResultInputValues,ResultValues,_TopLevel,TransInfo,_WF,_OR) :-
    ResultInputValues = [], % in this case the operation's behaviour also depends on current value of output variables
    lookup_cached_transitions(Name,InState,Solutions),!,
    member(transition(ParaValues,ResultValues,Updates,TransInfo),Solutions),
    create_full_operation(Name,ResultValues,ParaValues,ResultValues,Operation).

b_execute_operation2(Name,Operation,InState,NewOutState,ParaValues,ResultInputValues,ResultValues,
                     TopLevel,TransInfo,WF,OutputRequired) :-
    b_get_machine_operation_for_animation(Name,Results,Parameters,Body,OType,TopLevel),
    %% % we do not want skip the evaluation of a guard of an operation with parameters
    %% (Parameters==[] ->true; retractall(pge_algo:do_not_evaluate_guard)),
    b_execute_operation3(OType,Name,Operation,
			 InState,NewOutState,
			 Body,Parameters,ParaValues,Results,ResultInputValues,ResultValues,TopLevel,TransInfo,WF,OutputRequired).

create_full_operation(Name,Results,ParaValues,ResultValues,FullOperation) :-
    Operation =.. [Name|ParaValues],
    ( Results == [] -> /* we have an ordinary operation */
        FullOperation = Operation
    ; FullOperation = '-->'(Operation,ResultValues)
    ).

b_execute_operation3(classic,Name,Operation,InState,NewOutState,
		             Body,Parameters,ParaValues,Results,ResultInputValues,NormalisedResultValues,TopLevel,
		             [],  % [path(Path)],  % as path is currently filtered out anyway and not stored !
		             WF,OR) :-
    %print('Attempting Operation ----> '),print(Name),nl,
    empty_state(EmptyState),
    set_up_typed_localstate(Parameters,ParaValues,ParamTypedVals,EmptyState,LocalState,positive),
    l_expand_and_normalise_values(ParaValues,NormalisedParaValues,Parameters),
    create_full_operation(Name,Results,NormalisedParaValues,
                          NormalisedResultValues,Operation),
    %%  statistics(runtime,[Time1,_]), %%%
    set_up_undefined_localstate_with_predefined(Results,ResultInputValues,InState,NewInState),
     /* to avoid error messages with sequential composition (store_intermediate_updates) */
    check_additional_guard(TopLevel,LocalState,NewInState,WF),
   (TopLevel==true
       % TO DO: if Precondition or Body reads output variable, then add dummy output variable to LocalState
    -> b_execute_top_level_statement(Body,LocalState,NewInState,OutState,WF,_Path1,OR)
    ;  b_execute_statement(Body,LocalState,NewInState,OutState,WF,_Path2,OR)
   ),
    b_tighter_enumerate_values(ParamTypedVals,WF), % moved afer execute for performance reasons; see PerformanceTests/ParameterEnumeration.mch
       /* first look in OutState then in InState */
  %% statistics(runtime,[Time2,_]), Time is Time2-Time1, %%%
  %  print_message(operation(Time,Name)), %%%
     % print_bt_message('Operation Successful ------> '),
     (preferences:preference(treat_outermost_pre_as_select,true),
      preferences:preference(require_operations_to_assign_all_outputs,true)
       -> ReportError = true ; ReportError = false),
     get_results(Results,OutState,NormalisedResultValues,NewOutState,ReportError,normalise_results,WF).
b_execute_operation3(eventb_operation(ChangeSet,ParaValues,Operation),
		     _Name,Operation,  InState,Updates,
		     TBody,Parameters,ParaValues,[],[],[],true,TransInfo,_WF,_OutputRequired) :-
    %debug_println(9,attempting(Name)),
    % set up environment
    %copy_variable_store(InState,ChangeSet,Updates,FullOutState), % we now delay this until after guard evaluated
    % TO DO: check if need to pass _OutputRequired to b_event
    b_event_with_change_set(TBody,Parameters,ParaValues,InState,ChangeSet,Updates,TransInfo).  % note: WF not passed

:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
% additional guards can be set by the user to constrain/direct the execution of operations
check_additional_guard(TopLevel,LocalState,State,WF) :-
   (TopLevel==true,bmachine:b_machine_temp_predicate(TempPred)
      /* an additional precondition typed in by the user */
     -> (debug_mode(on) -> print(check_additional_guard),nl, print_bexpr(TempPred),nl ; true),
        % print_bstate(State),nl,  print_bstate(LocalState),nl,
        add_prob_deferred_set_elements_to_store(State,BState1,visible),
        b_test_boolean_expression(TempPred,LocalState,BState1,WF)
        %,print(checked),nl
     ;  true
   ).

% additional properties can either be set by user or by command-line
add_additional_properties(OldProperty,NewProperty) :-
   (bmachine:b_machine_temp_predicate(TempPred) % this overrides additional property !
     -> conjunct_predicates([TempPred,OldProperty],NewProperty)
      ; findall(AddPred,b_machine_additional_property(AddPred),L),
        (L=[] -> NewProperty = OldProperty
         ; conjunct_predicates([OldProperty|L],NewProperty)
        )
   ).

% get removes results from OutState
filter_results(Results,OutState,NewOutState,WF) :-
  get_results(Results,OutState,_ResultValues,NewOutState,false,do_not_normalise,WF).
get_results(Results,OutState,ResultValues,NewOutState,ReportError,Normalise,WF) :-
  (Results==[] -> (ResultValues,NewOutState) = ([],OutState)
    ; get_results1(Results,OutState,ResultValues,NewOutState,ReportError,Normalise,WF)).
:- block get_results1(?,-,?,?,?,?,?).
get_results1([],OutState,[],OutState,_ReportError,_Normalise,_WF).
get_results1([R|Results],OutState,[NRV|ResultValues],NewOutState,ReportError,Normalise,WF) :-
    def_get_texpr_id(R,ResultId),
    ~~mnf(lookup_and_delete_value(ResultId,RV,OutState,OutState2,ReportError,WF,Finished)),
    %% print_message(get_result(ResultId,RV,OutState)), %%
    get_results2(Finished,R,Results,RV,NRV,OutState2,ResultValues,NewOutState,ReportError,Normalise,WF).
:- block get_results2(-,?,?,?,?,?,?,?,?,?,?).
get_results2(_,R,Results,RV,NRV,OutState2,ResultValues,NewOutState,ReportError,Normalise,WF) :-
    (Normalise==normalise_results -> normalise_value_for_var(R,true,RV,NRV)
      ; %print_term_summary(no_norm(R,RV)),nl,
        NRV=RV), % Maybe we should not normalise in CBC mode ?
    get_results1(Results,OutState2,ResultValues,NewOutState,ReportError,Normalise,WF).



  /* ------------------------------------------------------------------ */
  /*                       SET UP CONSTANTS                             */
  /* ------------------------------------------------------------------ */

:- use_module(static_ordering).

:- volatile no_solution_found_for_constants/0, unsat_properties_component/1, unsat_properties_enumeration_warning/1, unsat_properties_abort_error/1,
 unsat_properties_conjunct_inside_component/3.
:- dynamic no_solution_found_for_constants/0, unsat_properties_component/1, unsat_properties_enumeration_warning/1, unsat_properties_abort_error/1,
 unsat_properties_conjunct_inside_component/3.

reset_unsat_component_infos :-
  retractall(unsat_properties_component(_)),
  retractall(unsat_properties_enumeration_warning(_)),
  retractall(unsat_properties_abort_error(_)),
  retractall(unsat_properties_conjunct_inside_component(_,_,_)).

% store information about unsatisfiable components for later usage
assert_unsat_component_infos :-
  unsat_component(X),  % from b_interpreter_components
  assert(unsat_properties_component(X)),fail.
assert_unsat_component_infos :-
  unsat_component_enumeration_warning(X),  % from b_interpreter_components
  assert(unsat_properties_enumeration_warning(X)),fail.
assert_unsat_component_infos :-
  unsat_component_abort_error(X),  % from b_interpreter_components
  assert(unsat_properties_abort_error(X)),fail.
assert_unsat_component_infos :-
  unsat_conjunct_inside_component(X,BE,Status), % from b_interpreter_components
  assert(unsat_properties_conjunct_inside_component(X,BE,Status)),fail.
assert_unsat_component_infos.

tcltk_unsatisfiable_components_exist :- unsat_component(_),!.

%tcltk_unsatisfiable_components(list(UnsatComponents)) :-
%  findall(C,unsat_properties_component(C),UnsatComponents), UnsatComponents\=[].

get_property_components(Components) :-
    b_machine_has_constants_or_properties,
    b_get_properties_from_machine(Properties),
    bsyntaxtree:predicate_components(Properties,Components),!.
get_property_components([]).

:- dynamic uses_implementable_integers/0.
:- dynamic min_max_integer_value_used/2.

tcltk_quick_describe_unsat_properties(list(FullDescr)) :-
   retractall(uses_implementable_integers),
   retractall(min_max_integer_value_used(_,_)), assert(min_max_integer_value_used(none,none)),
   get_property_components(Components),
   findall(TP, (unsat_properties_conjunct_inside_component(CompNr,P,_Status), % find individual conjuncts
               % Status can be false or unknown
               nth1(CompNr,Components,component(FullPred,_Vars)),
               check_uses_implementable_integers(FullPred), % for the min-max one may need to look at the entire component !
               translate_bexpression_with_limit(P,TrP),
               get_component_enum_warning_message(CompNr,EnumWarningAbortMsg),
               (EnumWarningAbortMsg = '' -> Prefix = '' ; Prefix = '\n    '),
               get_span_msg(P,'\n    ',SpanMsg),
               ajoin([' * ',TrP,Prefix,EnumWarningAbortMsg,SpanMsg,'\n'],TP)),
               %ajoin(['Component:',CompNr,'; ',TrP,'; == false;'],TP)),
               D1),
   findall([Msg1,PS|DescrVars],
                  (get_unsat_component_description(Components,CompNr,PS,DescrVars,EnumWarningAbortMsg,SpanMsg),
                   ajoin([' * Predicate Component ',EnumWarningAbortMsg,CompNr, SpanMsg, ':'],Msg1)),
               UC),
   append(UC,UCL), % join all component descriptions
   (D1=[]
    % TO DO: adapt message for enumeration warning
    -> Descr = D2,
       (UCL=[] -> D2 = ['The properties were satisfiable']
        ; D2 = ['No solution found for the following predicates: '|UCL]
       )
    ; Descr = ['The following properties prevented finding a solution:\n'|D12],
      (UCL=[] -> D2=[]
       ; D2 = ['\nIn addition no solution was found for the following predicates: '|UCL]
       ),
      append(D1,D2,D12)),
    findall(W,other_warning(W),Warnings),
    (Warnings = [] -> FullDescr=Descr ; append(Descr,['\nWARNINGS/MESSAGES:'|Warnings],FullDescr)).

% called by predicate_debugger:
get_unsat_component_predicate(Predicate,Vars) :-
      unsat_properties_component(CompNr),
      get_property_components(Components),
      nth1(CompNr,Components,component(Predicate,Vars)).

get_unsat_component_description(Components,CompNr,PS,DescrVars,EnumWarningAbortMsg,SpanMsg) :-
      unsat_properties_component(CompNr),
      \+ unsat_properties_conjunct_inside_component(CompNr,_,_), % we have not yet dealt with this component in first findall above
      nth1(CompNr,Components,component(Pred,Vars)),
      describe_constants(Vars,DescrVars),
      check_uses_implementable_integers(Pred), % store information in case it does use implementable integers for later user messages
     translate_bexpression_with_limit(Pred,150,PS),
     get_component_enum_warning_message(CompNr,EnumWarningAbortMsg),
     % note: sometimes we have a conjunction with empty span info, but different conjuncts (possibly from different files):
     get_span_msg(Pred,'\n   ',SpanMsg).

get_span_msg(Pred,Prefix,SpanMsg) :-
     (extract_span_description(Pred,SpanStr) -> ajoin([Prefix,SpanStr],SpanMsg) ; SpanMsg = '').

get_component_enum_warning_message(CompNr,EnumWarningAbortMsg) :-
    (unsat_properties_enumeration_warning(CompNr)
      ->(unsat_properties_abort_error(CompNr)
          -> EnumWarningAbortMsg = '[** with Enumeration Warning and WD-ERROR **] '
           ; EnumWarningAbortMsg = '[* with Enumeration Warning *] ')
       ; (unsat_properties_abort_error(CompNr)
          -> EnumWarningAbortMsg = '[** with WD-ERROR **] '
           ; EnumWarningAbortMsg =  '')
     ).

describe_constants([],R) :- !,R=[].
describe_constants(Vars,['* over identifiers : '|Vars]).

check_uses_implementable_integers(_) :- specfile:eventb_mode,!. % cannot use implementable integers
check_uses_implementable_integers(_) :- uses_implementable_integers,!. % no need to check another component
check_uses_implementable_integers(Pred) :-
   (uses_implementable_integers(Pred)
    -> assert(uses_implementable_integers),
       retract(min_max_integer_value_used(Min,Max)),
       min_max_integer_value_used(Pred,Min,Max,NewMin,NewMax),
       assert(min_max_integer_value_used(NewMin,NewMax))
     ; true).

other_warning(W) :- uses_implementable_integers,
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   min_max_integer_value_used(Min,Max), %print(minmax(Min,Max)),nl,
   ( (number(Min),Min<MININT ; number(Max),Max>MAXINT)
     -> ajoin(['\n MININT...MAXINT only set to ',MININT,'..',MAXINT,' but integers used at least from ',Min,'..',Max],W)
    ; (MAXINT < 127 ; MININT > -128)
     -> ajoin(['\n MININT...MAXINT possibly too small: ',MININT,'..',MAXINT],W)).
    % TO DO: we could also extract explicit integer values used in the component and compare them against MININT/MAXINT
other_warning(W) :- bmachine_eventb:deferred_set_equality_without_enumeration_axioms(X,_Set),
   translate_bexpression_with_limit(X,XS),
   ajoin(['\n Not recognized as enumerated set: ',XS],W).
% TO DO: generate warning if MAXINT/MININT small and NAT,NAT1 or INT was used in component(s) above

:- use_module(tools,[bt_start_ms_timer/1, bt_stop_ms_timer/1]).
:- use_module(static_ordering,[sort_ids_by_usage/4,reorder_state/3]).

b_set_up_concrete_constants(NormalisedConstantsState) :-
    reset_unsat_component_infos,
    retractall(no_solution_found_for_constants),assert(no_solution_found_for_constants),
    % preference(allow_incomplete_partial_setup_constants,AllowSkipping),
    reset_component_info(true),
    b_machine_has_constants_or_properties,
    b_get_machine_constants(UnsortedConstants),
    !,
    bt_start_ms_timer('SETUP_CONSTANTS'),
    b_get_properties_from_machine(MProperties),
    add_additional_properties(MProperties,Properties),
    project_onto_static_assertions(Properties,UnsortedConstants,PProperties,PUnsortedConstants),
    (preferences:preference(use_static_ordering,true)
     -> sort_ids_by_usage(PUnsortedConstants,PProperties,Constants,no_warnings), % no_warnings as Ids could be used in operations, assertions, ...
        set_up_typed_localstate(Constants,_FreshVars,TypedVals,[],SortedState,positive),
        reorder_state(PUnsortedConstants,SortedState,ConstantsState)
     ;  Constants =  PUnsortedConstants,
        set_up_typed_localstate(Constants,_FreshVars,TypedVals,[],ConstantsState,positive)
    ),
    load_partial_constants(ConstantsState,PProperties,LPProperties),  % check if some values have been computed and stored in a file
     b_global_sets:static_symmetry_reduction_for_global_sets(ConstantsState),
    %% print_bt_message(checking_properties), %%
    b_trace_test_components(LPProperties,ConstantsState,TypedVals),
    (unsat_component(_) -> (!,fail) ; true), % otherwise some expressions were skipped; use partial_set_up
    (silent_mode(on) -> true ; bt_stop_ms_timer('SETUP_CONSTANTS')),
    normalise_store(ConstantsState,NormalisedConstantsState),
    retractall(no_solution_found_for_constants). /* ensure that b_partial_set_up_concrete_constants fails */
b_set_up_concrete_constants([]).

:- volatile project_properties_on_identifiers/1.
:- dynamic project_properties_on_identifiers/1.
set_projection_on_static_assertions(ALL) :-
  % set interpreter in mode which projects out all constants not needed
  % for static assertion checking; all variables are also projected out
   retractall(project_properties_on_identifiers(_)),
   (b_get_assertions(ALL,dynamic,[]) -> true
     ; nl,print('*** WARNING: DYNAMIC ASSERTIONS NOT EMPTY'),nl,nl),
   b_get_assertions(ALL,static,L),
   (L=[] -> nl,print('*** WARNING: NO STATIC ASSERTIONS ***'),nl,nl ; true),
   conjunct_predicates(L,Conj),
   (debug_mode(on) -> print('Static Assertion: '),translate:print_bexpr(Conj),nl ; true),
   bsyntaxtree:predicate_identifiers(Conj,IDs),
   (debug_mode(on) -> print('IDS: '), print(IDs),nl ; true),
   assert(project_properties_on_identifiers(IDs)).

project_onto_static_assertions(Properties,Constants,NewProperties,NewConstants) :-
    %% set_projection_on_static_assertions(main), %% main or all
    project_properties_on_identifiers(ProjIDs),
    bsyntaxtree:project_predicate_on_identifiers(Properties,ProjIDs,NewProperties,PIDs,_),
    sort(PIDs,SIPDs),
    debug_println(9,projecting(SIPDs)),
    include(keep_constant(SIPDs),Constants,NewConstants),!. % Constants is a typed identifier list.
project_onto_static_assertions(Properties,Constants,Properties,Constants).

keep_constant(SIPDs,TypedID) :- get_texpr_id(TypedID,ID),ord_member(ID,SIPDs).



/* can be called if b_set_up_concrete_constants failed; will partially set up the state using values found in first ground value propagation phase */
:- use_module(specfile,[get_specification_description/2]).
b_partial_set_up_concrete_constants(NormalisedConstantsState) :-
    assert_unsat_component_infos,
    no_solution_found_for_constants,
    b_get_machine_constants(Constants),
    Constants \= [],
    (det_solution_for_constant(_,_) -> true), % we have found a solution for at least one constant
    set_up_typed_localstate(Constants,[],ConstantsState),
    print('Solutions found for the following constants:'),nl,
    fill_in_det_solutions(ConstantsState,FC),
    (preference(allow_incomplete_partial_setup_constants,true) -> true
      ; get_specification_description(properties,PS),
        % TO DO: determine if time-out occurred or not
        ajoin([PS,' are not satisfied or TIME-OUT occurred (but all CONSTANTS valued)'],Msg),
        add_error(setup_constants_inconsistent,Msg)
    ),
    normalise_store(FC,NormalisedConstantsState).


:- use_module(store,[no_value_for_variable/2]).
fill_in_det_solutions([],[]).
fill_in_det_solutions([bind(Var,Val)|T],[bind(Var,FVal)|FT]) :- %print(bind(Var,Val)),nl,
  (det_solution_for_constant(Var,Val) ->   print_term_summary(bind(Var,Val)),FVal=Val
     ; print(no_sol_for(Var)),nl,
       preference(allow_incomplete_partial_setup_constants,true),
       no_value_for_variable(FVal,Var)
  ),
  fill_in_det_solutions(T,FT).



/* --------------------------------- */

:- use_module(b_state_model_check, [b_check_valid_state/1]).

% note: set_up_initialisation for CBC checks does not use this predicate
b_initialise_machine(ConstantsState,NormalisedInitialVarsState,FullInitialState,Path) :-
   if(b_initialise_machine2(ConstantsState,NormalisedInitialVarsState,FullInitialState,Path),
      true,
      %(add_message(initialisation_fails,'INITIALISATION FAILS'),fail)).
      (\+ logged_error(initialisation_fails,_,_,_), % we have already provided a more precise error message
       add_error_and_fail(initialisation_fails,'INITIALISATION FAILS'))).

b_initialise_machine2(ConstantsState,NormalisedInitialVarsState,FullInitialState,Path) :-
/* InitialVarsState is an ouput */
   b_get_initialisation_from_machine(InitStatement,OType),
   ( InitStatement=[] ->
	   Path = [],
       add_message(b_initialise_machine,'Machine has no INITIALISATION statement(s)!'),
       NormalisedInitialVarsState = [],
       FullInitialState = ConstantsState
   ; OType == classic ->
	   Path = [path(PathInfo)],
       b_get_machine_variables(DeclaredVars),
       set_up_undefined_localstate(DeclaredVars,ConstantsState,NewInState),
       /* this is if the INITIALISATION has a sequential composition; to avoid error messages */
       init_wait_flags(WF),
       (bmachine:b_machine_temp_predicate(_) -> InitStatement2=InitStatement  /* avoid enumerating each init statement in parallel in case additional guard links them */
        ; translate_init_statements(InitStatement,InitStatement2)
        ),
       b_execute_statement(InitStatement2,[],NewInState,InitialVarsState,WF,PathInfo),
    %  print_message(init(InitialVarsState)), %%
       check_additional_guard(true,InitialVarsState,ConstantsState,WF),
       %print(init(InitialVarsState,InitStatement2)),nl,
       % Warning: in Z mode (e.g. test 565) the variable skeleton may not yet be set up !
       sort_variable_binding(InitialVarsState,NormalisedInitialVarsState),
       append(ConstantsState,NormalisedInitialVarsState,FullInitialState),
       split_names_and_types(DeclaredVars,DeclaredNames,DeclaredTypes),
    %  print_message(grounding_constraint_waitflags(InitialVarsState)), %%
       ground_wait_flags(WF),

       b_enumerate_values_in_store(DeclaredNames,DeclaredTypes,_,InitialVarsState,WF)  %% MOVED HERE TO AVOID that target values are enumerated before initialisation effect computed
   ; OType = eventb_operation(_ChangeSet,ParaValues,_Operation) ->
       prepare_eventb_initialisation(ConstantsState,InitialVarsUnNormState,FullInitialUnNormState),
       get_texpr_expr(InitStatement,
                      rlevent(_Name,_Sec,_Stat,Parameters,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_AbsEv)),
       b_event(InitStatement,Parameters,ParaValues,ConstantsState,FullInitialUnNormState,Path),
       normalise_store(InitialVarsUnNormState,NormalisedInitialVarsState),
       normalise_store(FullInitialUnNormState,FullInitialState) % a bit of redundant work here; we do the work twice and try to normalise constants again !
   ),
   b_check_valid_state(FullInitialState).

% ensure that the variables always appear in the same order as reported by b_get_machine_variables:
% this is important for e.g., state_packing  TO DO: we could sort them according to the variable names, but would there be a gain ??
sort_variable_binding(Store,SortedStore) :-
   b_get_machine_variables(Variables),
   sort_aux(Variables,Store,[],SortedStore). %, print(sorted(SortedStore)),nl.
sort_aux([],S,Acc,NormRes) :- reverse(Acc,Res),
   normalise_store(Res,NormRes),
   (S=[] -> true ; add_internal_error('Unknown variable bindings: ',S)).
sort_aux([TVar|T],Store,Acc,Res) :- def_get_texpr_id(TVar,ID),
   bselect(ID,Val,Store,Rest),
   sort_aux(T,Rest,[bind(ID,Val)|Acc],Res).

:- block bselect(?,?,-,?). % in Z mode (e.g. test 565) the variable skeleton may not yet be set up !
bselect(ID,Val,[bind(ID1,Val1)|T],Rest) :- % we assume ID1 is nonvar
   (ID=ID1 -> Val=Val1,T=Rest
     ; Rest = [bind(ID1,Val1)|RT],
       bselect(ID,Val,T,RT)).

prepare_eventb_initialisation(ConstantsState,InitialVarsState,FullInitialState) :-
    b_get_machine_variables(Variables),
    empty_state(Empty),
    set_up_typed_localstate(Variables,_,_,Empty,InitialVarsState,positive),
    % NormalisedInitialVarsState = InitialVarsState, % TO DO : check that b_event normalises this
    combine_updates(ConstantsState,InitialVarsState,FullInitialState).



reset_b_interpreter :-
  reset_partial_evaluator,
  retractall(project_properties_on_identifiers(_)),
  reset_unsat_component_infos,
   custom_explicit_sets:reset_memo_table,
  reset_unsat_component_infos,
  retractall(no_solution_found_for_constants).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_b_interpreter,
                           'Reset B-Interpreter Memo Table & Unsat Components.').
