% (c) 2009-2018 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_objects,[basic_type/2,
                     enumerate_basic_type/2, enumerate_basic_type_wf/3, enumerate_basic_type_wf/4,
                     all_objects_of_type/2,
                     max_cardinality/2,
                     enumerate_type/3, % last argument basic or tight
                     enumerate_basic_type/3, enumerate_type/4, % last argument false/true disables/enables enum warning
                     enumerate_tight_type/2, enumerate_tight_type/3,
                     enumerate_int/3,
                     enum_warning/4,
                     all_strings/1, is_string/2, is_not_string/1,

                     top_level_dif/2,
                     equal_object_optimized/2, equal_object_optimized/3, equal_object_optimized_wf/4,
                     equal_object/2, equal_object/3, equal_object_wf/3, equal_object_wf/4,
                     not_equal_object/2, not_equal_object_wf/3,
                     equal_cons/3,  equal_cons_wf/4, equal_cons_lwf/4,
                     get_next_element/3,
                     is_marked_to_be_computed/1, mark_as_to_be_computed/1,

                     %equality_objects/3,
                     membership_test/3, membership_test_wf/4,

                     %is_a_set/1,
                     empty_set/1, empty_set_wf/2,
                     not_empty_set/1, not_empty_set_wf/2,
                     exact_element_of/2,
                     check_element_of/2, check_element_of_wf/3,
                     not_element_of/2, not_element_of_wf/3,

                     add_element/3, add_element/4, add_element_wf/4, add_element_wf/5,
                     add_new_element_wf/4,
                     delete_element_wf/4,
                     remove_element/3, remove_element_wf/4,remove_element_wf/5,remove_element_wf_if_not_infinite_or_closure/6,
                     remove_exact_first_element/3,

                     partition_wf/3, not_partition_wf/3,
                     %all_different/2,
                     disjoint_sets/3, not_disjoint_sets/3,

                     union/3, union_wf/4, union_generalized/2, union_generalized_wf/3,
                     intersection/3, intersection_generalized_wf/4,
                     difference_set/3, difference_set_wf/4,
                     in_difference_set_wf/4, not_in_difference_set_wf/4,
                     in_union_set_wf/4, not_in_union_set_wf/4,
                     in_intersection_set_wf/4, not_in_intersection_set_wf/4,

                     strict_subset_of/2, strict_subset_of_wf/3,
                     check_subset_of/2, check_subset_of_wf/3, check_finite_subset_of_wf/3,
                     check_non_empty_subset_of_wf/3, check_finite_non_empty_subset_of_wf/3,
                     not_subset_of/2, not_subset_of_wf/3, not_both_subset_of/5,
                     not_finite_subset_of_wf/3,
                     not_strict_subset_of/2, not_strict_subset_of_wf/3,
                     not_non_empty_subset_of_wf/3, not_non_empty_finite_subset_of_wf/3,
                     both_global_sets/4,check_subset_of_global_sets/2, check_not_subset_of_global_sets/2,

	                   first_of_pair/2, second_of_pair/2,
                     minimum_of_set_extension_list/4,
                     maximum_of_set_extension_list/4,
                     minimum_of_set/4, maximum_of_set/4,
                     is_finite_set_wf/2, is_infinite_set_wf/2, test_finite_set_wf/3,
                     finite_cardinality_as_int/3, cardinality_as_int_for_wf/2,
                     cardinality_as_int_wf/3,
                     cardinality_as_int/2, cardinality_peano_wf/3, card_convert_int_to_peano/2,
                     % card_geq/2,
                     cardinality_greater/5, cardinality_greater_equal/5,
                     cardinality_of_set_extension_list/3,

                     cartesian_product/3, %  removed
                     is_cartesian_pair_wf/4, not_is_cartesian_pair/4,

                     power_set/2, non_empty_power_set/2,

                     % is_boolean/1, %is_not_boolean/1,
                     is_integer/2, is_not_integer/1,
                     is_natural/2, is_natural1/2,
                     is_implementable_int/2,is_implementable_nat/2, is_implementable_nat1/2,
                     is_not_natural/1, is_not_natural1/1,
                     is_not_implementable_int/1,is_not_implementable_nat/1, is_not_implementable_nat1/1,

                     less_than/2, less_than_equal/2,
                     less_than_direct/2, less_than_equal_direct/2,
                     safe_less_than_equal/2, safe_less_than_equal/3,
                     safe_pow2/2, safe_mul/3, safe_add/3, safe_pown/3,
                     greater_than/2, greater_than_equal/2,
                     int_plus/3,
                     division/5, floored_division/5,
                     modulo/5,
                     int_minus/3, unary_minus_wf/3,
                     % nat_range/3, % removed
                     in_nat_range_wf/4, not_in_nat_range/3, not_in_nat_range_wf/4, test_in_nat_range_wf/5,
                     in_nat_range/3, % version without enumeration
                     times/3, square/3,
                     int_power/5,
                     % pred/2, succ/2, removed
                     integer_global_set/1,

                     element_of_global_set/2,element_of_global_set_wf/3,not_element_of_global_set/2,

                     exhaustive_kernel_check/1, exhaustive_kernel_check_wf/2, exhaustive_kernel_check_wf/3,
                     exhaustive_kernel_check_wfdet/2,
                     exhaustive_kernel_succeed_check/1, exhaustive_kernel_fail_check/1,
                     exhaustive_kernel_fail_check_wf/2, exhaustive_kernel_fail_check_wfdet/2,
                     exhaustive_kernel_check/2, exhaustive_kernel_succeed_check/2, exhaustive_kernel_fail_check/2,

                     singleton_set_element/3
                     ]).


%:- use_module('../extensions/profiler/profiler.pl').
%:- use_module('../extensions/profiler/profiler_te.pl').
%:- enable_profiling(enumerate_basic_type/3).
%:- enable_profiling(enumerate_type/3).
%:- enable_profiling(enumerate_tight_type/2).

%:- print(loading_kernel_objects),nl.

%portray_message(informational, _).
:- use_module(library(terms)).
:- use_module(self_check).

:- use_module(debug).
:- use_module(tools_printing,[print_term_summary/1]).
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides operations for the basic datatypes of ProB (equal, not_equal, enumeration).').

:- use_module(typechecker).
:- use_module(error_manager).

:- use_module(b_global_sets). %,[global_type/2, b_global_set_cardinality/2, b_empty_global_set/1]).

:- use_module(kernel_waitflags).
:- use_module(library(lists)).
:- use_module(library(avl),[avl_min/2, avl_max/2]).

%:- use_module(library(clpfd)).
%:- use_module(fd_utils).
:- use_module(fd_utils_clpfd).

:- use_module(kernel_freetypes).

:- use_module(custom_explicit_sets).


:- use_module(typechecker).

%:- use_module(clpfd_off_interface). %
% on a 32 bit system: use clpfd_off_interface; on 64 bit system clpfd_interface should be ok (integer overflows)
:- use_module(clpfd_interface). %


:- type atomic_type +--> (term(integer,[]) ; term(string,[]) ; constant(list(atomic)) ; abort ; boolean ; global(atomic)).
:- type atomic_any_type +--> (type(atomic_type) ; term(any,[]) ).
:- type basic_type_descriptor +--> (type(atomic_any_type) ; set(basic_type_descriptor) ;
                                    seq(basic_type_descriptor) ;
                                    couple(basic_type_descriptor,basic_type_descriptor) ;
                                    record(list(type(field_type))) ;
                                    freetype(atomic)).

:- type inferred_basic_type_descriptor +--> (var ; type(atomic_type) ; set(inferred_basic_type_descriptor) ;
                                    seq(inferred_basic_type_descriptor) ;
                                    couple(inferred_basic_type_descriptor,inferred_basic_type_descriptor)).

:- type fd_index +--> (integer ; var).
:- type fd_set +--> (atomic ; var).
:- type fd_term +--> fd(fd_index,fd_set).
:- type bsets_integer +--> int((integer ; var)).
:- type bsets_string +--> string((atomic ; var)).
:- type bsets_bool +--> (pred_false /* bool_false */ ; pred_true /* bool_true */).
:- type field_type +--> field(atomic,basic_type_descriptor).

%:- type bsets_sequence +--> (nil_seq ; cons(type(bsets_object),type(bsets_sequence))).
%:- type bsets_set +--> vlist(type(bsets_object)).
:- type bsets_set +--> (term([],[]) ; var ; term('.',[type(bsets_object),type(bsets_set)]) ;
                       avl_set( ground ) ;
                       closure(list(type(variable_id)),
                                   list(type(basic_type_descriptor)),type(boolean_expression))
                           ; closure_x(list(type(variable_id)),
                                   list(type(basic_type_descriptor)),type(boolean_expression),any)).
:- type bsets_couple +--> term(',',[type(bsets_object),type(bsets_object)]).
:- type bsets_global +--> global_set((atomic ; var)).
:- type bsets_field +--> field(atomic,type(bsets_object)).
:- type bsets_record +--> rec((var ; list(bsets_field))).
:- type bsets_freetype +--> freeval(atomic,(atomic ; var),type(bsets_object)).

:- type bsets_object +--> (fd_term ; bsets_integer ; bsets_bool ; term(term,[any]) ; bsets_set ;
                           % abort(any) ;  % deprecated
                           bsets_couple ; bsets_string ; bsets_global ; var;
                           bsets_record ; bsets_freetype).


:- assert_must_succeed(kernel_waitflags:set_silent(true)). % disable waitflag store not init msgs




% a predicate to exhaustively check a kernel predicate with all possible modes

:- use_module(tools_timeout,[time_out_call/1]).
exhaustive_kernel_check_opt(C,Cond) :- (Cond -> exhaustive_kernel_check(C) ; true).
exhaustive_kernel_check(C) :- exhaustive_kernel_check4([],C,true,true).
exhaustive_kernel_check(Opts,C) :- exhaustive_kernel_check4(Opts,C,true,true).
exhaustive_kernel_check_wf(C,WF) :- exhaustive_kernel_check_wf([],C,WF).
exhaustive_kernel_check_wf(Opts,C,WF) :-
   exhaustive_kernel_check4(Opts,C,kernel_waitflags:init_wait_flags(WF),
                                 kernel_waitflags:ground_wait_flags(WF)).
exhaustive_kernel_check_wfdet(C,WF) :-
   exhaustive_kernel_check4([],C,kernel_waitflags:init_wait_flags(WF),
                                 kernel_waitflags:ground_det_wait_flag(WF)).

exhaustive_kernel_check4(Opts,Call,Pre,Post) :- enumerate_kernel_call(Call,Opts,ECall,Code),
   debug_println(9,exhaustive_kernel_check(ECall,Code)),
   flatten_call((Pre,ECall,Code,Post),FullCall), %print(FullCall),nl,
   time_out_call(must_succeed_without_residue(FullCall)),debug_println(9,ok),
   fail.
exhaustive_kernel_check4(_,_,_,_).

flatten_call((A,B),Res) :- !,flatten_call(A,FA), flatten_call(B,FB), conjoin_call(FA,FB,Res).
flatten_call(Module:Call,Res) :- !, flatten_call(Call,F), Res=Module:F.
flatten_call(X,X).

conjoin_call(true,X,R) :- !,R=X.
conjoin_call(X,true,R) :- !, R=X.
conjoin_call(X,Y,(X,Y)).

exhaustive_kernel_succeed_check(C) :- exhaustive_kernel_succeed_check([],C).
exhaustive_kernel_succeed_check(Opts,Call) :- enumerate_kernel_call(Call,Opts,ECall,Code),
   debug_println(9,exhaustive_kernel_succeed_check(ECall,Code)),
   flatten_call((ECall,Code),FullCall),
   time_out_call(must_succeed(FullCall)),debug_println(9,ok),
   fail.
exhaustive_kernel_succeed_check(_,_).

exhaustive_kernel_fail_check_opt(C,Cond) :- (Cond -> exhaustive_kernel_fail_check(C) ; true).
exhaustive_kernel_fail_check(C) :- exhaustive_kernel_fail_check4([],C,true,true).
exhaustive_kernel_fail_check(Opts,C) :- exhaustive_kernel_fail_check4(Opts,C,true,true).
exhaustive_kernel_fail_check_wf(C,WF) :-
   exhaustive_kernel_fail_check4([],C,kernel_waitflags:init_wait_flags(WF),
                                      kernel_waitflags:ground_wait_flags(WF)).
exhaustive_kernel_fail_check_wfdet(C,WF) :-
   exhaustive_kernel_fail_check4([],C,kernel_waitflags:init_wait_flags(WF),
                                      kernel_waitflags:ground_det_wait_flag(WF)).
exhaustive_kernel_fail_check4(Opts,Call,Pre,Post) :- enumerate_kernel_call(Call,Opts,ECall,Code),
   debug_println(9,exhaustive_kernel_fail_check(ECall,Code)),
   flatten_call((Pre,ECall,Code,Post),FullCall),
   time_out_call(must_fail(FullCall)),debug_println(9,ok),
   fail.
exhaustive_kernel_fail_check4(_,_,_,_).

% enumerate_kernel_call(Call, OptionList, NewCall, CodeAfter)
enumerate_kernel_call((A,B),Opts,(EA,EB),(CA,CB)) :- !,
    enumerate_kernel_call(A,Opts,EA,CA), enumerate_kernel_call(B,Opts,EB,CB).
enumerate_kernel_call(Module:Call,Opts,Module:ECall,Code) :- !, enumerate_kernel_call(Call,Opts,ECall,Code).
enumerate_kernel_call(Call,Opts,ECall,Code) :- Call=..[KernelPred|CArgs],
   (member(commutative,Opts)
     -> (Args=CArgs ; CArgs=[A1,A2|T], Args=[A2,A1|T])
    ; Args=CArgs
   ),
   l_enumerate_kernel_args(Args,EArgs,Code,KernelPred,1), ECall=..[KernelPred|EArgs].
l_enumerate_kernel_args([],[],true,_,_).
l_enumerate_kernel_args([H|T],[EH|ET],Code,KernelPred,Nr) :-
   enumerate_kernel_args(H,EH,C1,KernelPred/Nr),
   N1 is Nr+1,
   l_enumerate_kernel_args(T,ET,C2,KernelPred,N1),
   permute_code((C1,C2),Code).

permute_code((true,C),R) :- !,R=C.
permute_code((C,true),R) :- !, R=C.
permute_code((C1,C2),(C1,C2)).
permute_code((C1,C2),(C2,C1)).

enumerate_kernel_args(Var,Res,Code,_) :- var(Var),!, Res=Var, Code=true.
enumerate_kernel_args(X,Res,Code,KP_Nr) :- do_not_delay(X,KP_Nr),!, Res=X, Code=true.
enumerate_kernel_args(Arg,Arg,true,_). % just keep the argument
enumerate_kernel_args(Arg,NewArg,Code,_) :- % delay the argument fully
   (term_is_of_type(Arg,bsets_object,no)
    -> Code = equal_object(NewArg,Arg,enumerate_kernel_args)
     ; Code = '='(NewArg,Arg)).
enumerate_kernel_args(int(X),int(XX),Code,_) :- nonvar(X), Code = '='(X,XX). % delay setting number content
enumerate_kernel_args(string(X),string(XX),Code,_) :- nonvar(X), Code = '='(X,XX). % delay setting string content
enumerate_kernel_args((A,B),(AA,BB),(CodeA,CodeB),KP_Nr) :-
   enumerate_kernel_args(A,AA,CodeA,KP_Nr),enumerate_kernel_args(B,BB,CodeB,KP_Nr),
   (AA,BB) \== (A,B). % avoid re-generating case 3 above (just keep argument)
enumerate_kernel_args(freeval(ID,Case,A),freeval(ID,Case,AA),CodeA,KP_Nr) :-
   enumerate_kernel_args(A,AA,CodeA,KP_Nr),
   AA \== A. % avoid re-generating case 3 above (just keep argument)
enumerate_kernel_args([H|T],[H|NewT],Code,_) :- Code = equal_object(NewT,T).
enumerate_kernel_args([H|T],Res,Code,KP_Nr) :- try_expand_and_convert_to_avl([H|T],AVL),
    AVL \= [H|T], enumerate_kernel_args(AVL,Res,Code,KP_Nr).
enumerate_kernel_args([H|T],Res,Code,KP_Nr) :- ground([H|T]),generate_member_closure([H|T],Closure),
    enumerate_kernel_args(Closure,Res,Code,KP_Nr).

do_not_delay(b(_,_,_),_). % do not delay B predicates and expressions
do_not_delay(global_set(G),KP/ArgNr) :- custom_explicit_sets:is_infinite_global_set(G),
    %print(inf_nr(KP,ArgNr)),nl,
    do_not_delay_arg(KP,ArgNr).
% these arguments cause difficulty if infinite sets are delayed
do_not_delay_arg(partial_function_wf,2).
do_not_delay_arg(partial_function_wf,3).
do_not_delay_arg(subset_test,2).
do_not_delay_arg(subset_strict_test,2).

generate_member_closure(ExplicitSet,closure(['_zzzz_unit_tests'],[Type],Pred)) :-
   infer_type(ExplicitSet,set(Type)),
   Pred =
   b(member(b(identifier('_zzzz_unit_tests'),Type,[generated]),
    b(value(ExplicitSet),set(Type),[])),pred,[]).

infer_type(Value,Type) :- (infer_value_type(Value,Type)
  -> true %,print(inferred(Type,Value)),nl
  ;  print('### Could not infer type: '), print(Value),nl,fail).

:- use_module(btypechecker,[couplise_list/2]).
infer_value_type([],set(any)).
infer_value_type([H|T],set(ResType)) :- infer_value_type(H,Type),
   ((contains_any(Type),T=[H2|_], % try H2; maybe we can infer a better type here
     infer_value_type(H2,Type2), \+ contains_any(Type2))
     -> ResType = Type2
      ; ResType = Type).
infer_value_type(avl_set(node(H,_True,_,_,_)),set(Type)) :- infer_value_type(H,Type).
infer_value_type(int(_),integer).
infer_value_type(string(_),string).
infer_value_type((A,B),couple(TA,TB)) :- infer_value_type(A,TA), infer_value_type(B,TB).
infer_value_type(fd(_,T),global(T)).
infer_value_type(pred_true /* bool_true */,boolean).
infer_value_type(pred_false /* bool_false */,boolean).
infer_value_type(rec(Fields),record(FieldTypes)) :- infer_field_types(Fields,FieldTypes).
infer_value_type(freeval(Id,_Case,_Val),freetype(Id)).
infer_value_type(closure(_,Types,_),set(Res)) :- couplise_list(Types,Res).
infer_value_type(global_set('STRING'),R) :- !, R=set(string).
infer_value_type(global_set(X),R) :- b_integer_set(X),!,R=set(integer).
infer_value_type(global_set(Name),set(global(Name))) :- b_global_set(Name).

infer_field_types([],[]).
infer_field_types([field(Name1,Val)|T],[field(Name1,VT)|TT]) :-
     infer_value_type(Val,VT),
     infer_field_types(T,TT).

contains_any(any).
contains_any(couple(A,B)) :- (contains_any(A) -> true ; contains_any(B)).
contains_any(set(A)) :- contains_any(A).
% to do: fields

:- assert_pre(kernel_objects:basic_type(Obj,Type), (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:basic_type(Obj,_), type_check(Obj,bsets_object)).

%:- block basic_type(-,-).

basic_type(FD,global(T)) :- !, global_type(FD,T). % will set up CLP(FD) domain for X
% TO DO:  Also: what about global(T) inside other structures (pairs) ?
basic_type(Rec,record(FieldTypes)) :- !, Rec=rec(Fields),  %print(basic_field(FieldTypes)),nl,
    basic_field_types(Fields,FieldTypes).
%basic_type(Set,set(Type)) :- !, basic_type_set(Type,Set,inf).
basic_type(_X,_TY). %basic_type2(TY,X) %basic_symbreak(TY,X)
      %print(ignore_basic_type(X,Y)),nl %, basic_type2(TY,X) %%STILL REQUIRED ?????

basic_field_types([],[]).
basic_field_types([field(Name1,Val)|T],[field(Name2,VT)|TT]) :-
     basic_field_types2(Name1,Val,T,Name2,VT,TT).

basic_field_types2(Name,Val,T,Name,VT,TT) :-
  basic_type(Val,VT),basic_field_types(T,TT).

/*
basic_type2(set(Type),Set) :- !,basic_type_set(Type,Set,inf).
basic_type2(_,_).

:- use_module(inf_arith).
basic_type_set(Type,Set,Bound) :-
  (kernel_objects:max_cardinality(Type,C)
    -> inf_arith:infmin(C,Bound,Card)
     ; Card=inf),
   %print(basic_type_set(Set,Card,Type)),nl,nl,
   basic_type_set(Set,[],Type,Card).

:- block basic_type_set(-,?,?,?).
%basic_type_set(S,Sofar,T) :- print_message(basic_type_set(S,Sofar,T)),fail.
basic_type_set(CustomSet,_,_Type,_) :- is_custom_explicit_set(CustomSet,basic_type_set),!.
basic_type_set([],_,_,_).
basic_type_set([H|T],ElementsSoFar,Type,Card) :-
   inf_arith:infdec_if_positive(Card,C1),
   basic_type(H,Type),
   not_element_of(H,ElementsSoFar), %consistent([H|ElementsSoFar]),
   (C1=0 -> empty_set(T) ; basic_type_set(T,[H|ElementsSoFar],Type,C1)).

:- use_module(kernel_sym_break_order).
basic_symbreak(set(_),Set) :- !,sym_break_set(Set).
basic_symbreak(seq(_),Set) :- !,sym_break_set(Set).
basic_symbreak(_,_). */

/*
basic_type2(global(T),fd(X,T)) :- global_type(fd(X,T),T). % will set up CLP(FD) domain for X
basic_type2(set(Type),Set) :- basic_type_set(Set,[],Type).
basic_type2(seq(X),Seq) :- basic_type(Seq,set(couple(integer,X))).
basic_type2(boolean,_).
%% basic_type2(boolean,pred_true). basic_type2(boolean,pred_false).
basic_type2(couple(XT,YT),(X,Y)) :- basic_type(X,XT), basic_type(Y,YT).
basic_type2(string,string(_)).
basic_type2(constant(List),term(T)).
basic_type2(integer,int(_X)). %  :- clpfd_interface:clpfd_inrange(_X,-999999,999999). %%%%
basic_type2(string,string(_)).
basic_type2(record(FieldTypes),rec(Fields)) :-
  basic_field_types(Fields,FieldTypes,rec(Fields),record(FieldTypes)).
basic_type2(freetype(Id),freeval(Id,Case,Val)) :- basic_type_freetype(Id,Case,Val).
basic_type2(any,_).

:- block basic_type_set(-,?,?).
%basic_type_set(S,Sofar,T) :- print_message(basic_type_set(S,Sofar,T)),fail.
basic_type_set(CustomSet,[],_Type) :- is_custom_explicit_set(CustomSet,basic_type_set),!.
basic_type_set([],_,_).
basic_type_set([H|T],ElementsSoFar,Type) :-
   basic_type(H,Type),
   not_element_of(H,ElementsSoFar), %consistent([H|ElementsSoFar]),
   basic_type_set(T,[H|ElementsSoFar],Type).



:- block basic_type_freetype(?,-,?).
basic_type_freetype(Id,Case,Val) :-
   registered_freetype(Id,Cases),
   member(case(Case,CT),Cases),!,
   basic_type(Val,CT).

:- block basic_field_types(-,-,?,?).
basic_field_types([],[],_,_).
basic_field_types([field(Name1,Val)|T],[field(Name2,VT)|TT],Orig1,Orig2) :-
     basic_field_types2(Name1,Val,T,Name2,VT,TT,Orig1,Orig2).

:- block basic_field_types2(?,?,?, -,?,?,?,?).
basic_field_types2(Name1,Val,T,Name2,VT,TT,Orig1,Orig2) :-
  (Name1=Name2
   -> basic_type(Val,VT),
      basic_field_types(T,TT,Orig1,Orig2)
    ;  add_error(kernel_objects,'Record field names do not match: ',(Name1,Name2,Orig1,Orig2)),
       fail
  ).


*/

/* ------------------------- */
/*   enumerate_basic_type/2  */
/* ------------------------- */
/* a version of basic_type that enumerates */

:- assert_must_succeed(enumerate_basic_type([],set(couple(integer,integer)) )).
:- assert_must_succeed(enumerate_basic_type([([],int(2)), ([int(3)],int(4))],
                                         set(couple(set(integer),integer)) )).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                             set(couple(integer,integer)) )).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                             seq(integer) )).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                             seq(integer) )).
:- assert_must_succeed((enumerate_basic_type(X,global('Name')),
                        equal_object(X,fd(1,'Name')) )).
:- assert_must_succeed((enumerate_basic_type(X,global('Name')),
                        equal_object(X,fd(2,'Name')) )).
:- assert_must_succeed((enumerate_basic_type(X,global('Name')),
                        X==fd(2,'Name')) ).
:- assert_must_succeed((enumerate_basic_type(X,record([field(a,global('Name'))])),
                        equal_object(X,rec([field(a,fd(1,'Name'))])) )).
:- assert_must_succeed((enumerate_basic_type(X,record([field(a,integer),field(b,global('Name'))])),
                        equal_object(X,rec([field(a,int(1)),field(b,fd(1,'Name'))])) )).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc1,[case(a,constant([a])),case(b,integer)]),
			kernel_freetypes:set_freetype_depth(2),
			enumerate_basic_type(X,freetype(selfc1)),equal_object(X,freeval(selfc1,a,term(a))),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc5,[case(a,constant([a])),case(b,integer)]),
			kernel_freetypes:set_freetype_depth(2),
			enumerate_basic_type(X,freetype(selfc5)),equal_object(X,freeval(selfc5,b,int(1))),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc7,[case(nil,constant([nil])),case(node,couple(freetype(selfc7),freetype(selfc7)))]),
			kernel_freetypes:set_freetype_depth(3),
			findall(X,enumerate_basic_type(X,freetype(selfc7)),Solutions),
			length(Solutions,5),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc2,[case(a,constant([a])),case(b,freetype(selfc3))]),
			kernel_freetypes:add_freetype(selfc3,[case(c,constant([c])),case(d,freetype(selfc2))]),
			kernel_freetypes:set_freetype_depth(4),
			enumerate_basic_type(X,freetype(selfc2)),
			equal_object(X,freeval(selfc2,b,freeval(selfc3,d,freeval(selfc2,b,freeval(selfc3,c,term(c)))))),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(1,'Name'),fd(1,'Code'))])) ).
:- assert_must_succeed((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(2,'Name'),fd(1,'Code')), (fd(1,'Name'),fd(2,'Code'))])) ).
:- assert_must_succeed((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(1,'Name'),fd(2,'Code')), (fd(2,'Name'),fd(1,'Code'))])) ).
:- assert_must_succeed_any((enumerate_basic_type(X,set(couple(global('Name'),global('Code'))) ),
                        equal_object(X,[(fd(1,'Name'),fd(2,'Code')), (fd(2,'Name'),fd(1,'Code'))])) ).
:- assert_must_succeed(enumerate_basic_type([(int(2),(int(1),int(2))),
                                                  (int(1),(int(3),int(4)))],
                                             set(couple(integer,couple(integer,integer))) )).
:- assert_must_succeed(enumerate_basic_type([(int(2),(int(1),int(2))),
                                                  (int(55),(int(3),int(4)))],
                                             set(couple(integer,couple(integer,integer))) )).
:- assert_must_succeed(enumerate_basic_type([term('err')],set(constant([err])))).
:- assert_must_succeed(enumerate_basic_type([(int(1),int(2)),(int(3),int(4))],
                                               set(couple(integer,integer)))).

:- assert_must_succeed_multiple(enumerate_basic_type([(int(2),fd(_A,'Name')),(int(3),fd(_B,'Name')),
           (int(4),fd(_C,'Name')),(int(5),fd(_D,'Name')),(int(6),fd(_E,'Name')),(int(7),fd(_F,'Name')),
           (int(8),fd(_G,'Name')),(int(9),fd(_H,'Name')),(int(10),fd(_I,'Name')),
           (int(11),fd(_,'Name')),(int(12),fd(_,'Name')),(int(13),fd(_,'Name')),
           (int(14),fd(_,'Name'))],set(couple(integer,global('Name'))))).

:- assert_must_fail(( findall(XX,enumerate_basic_type(XX, set(set(global('Code')))) ,S), member(X,S), remove(S,X,R), member(X2,R), equal_object(X,X2)  )).

:- assert_must_succeed((  enumerate_basic_type(global_set('Code'),
                   set(global('Code'))) )).

:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([(fd(1,'Name'),fd(2,'Code')), (fd(2,'Name'),fd(1,'Code'))],set(couple(global('Name'),global('Code')))))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([(fd(1,'Name'),pred_true), (fd(2,'Name'),pred_false), (fd(2,'Name'),pred_true)],set(couple(global('Name'),boolean))))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([pred_true,pred_false],set(boolean)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(enumerate_basic_type([[],[pred_true,pred_false]],set(set(boolean))))).

:- assert_pre(kernel_objects:enumerate_basic_type(Obj,Type),
             (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:enumerate_basic_type(Obj,_), (type_check(Obj,bsets_object),ground_check(Obj))).

enumerate_basic_type_wf(Obj,Type,WF) :-
   enumerate_basic_type_wf(Obj,Type,enumerate_basic_type,WF).
:- block enumerate_basic_type_wf(?,-,?,?).
enumerate_basic_type_wf(Obj,Type,EnumWarning,WF) :-
   enumerate_basic_type4(Type,Obj,basic,trigger_true(enum_wf_context(WF,EnumWarning))). % add WF context info

:- block enumerate_basic_type(?,-).
enumerate_basic_type(Obj,Type) :- %print_message(call_enumerate_basic_type(Obj,Type)),
   %enumerate_basic_type2(Obj,Type).
   enumerate_basic_type4(Type,Obj,basic,trigger_true(enumerate_basic_type)).
   %(ground(Obj) -> true ; enumerate_basic_type3(Type,Obj,basic)).

:- block enumerate_basic_type(?,-,-).
enumerate_basic_type(Obj,Type,EnumWarning) :-
   enumerate_basic_type4(Type,Obj,basic,EnumWarning).


:- block enumerate_type(?,-,?). % last argument: basic or tight
enumerate_type(Obj,Type,Tight) :- %print_message(call_enumerate_basic_type(Obj,Type)),
   %enumerate_basic_type2(Obj,Type).
   enumerate_basic_type4(Type,Obj,Tight,trigger_true(enumerate_type_3)).

:- block enumerate_type(?,-,?,-).
enumerate_type(Obj,Type,Tight,EnumWarning) :-
   enumerate_basic_type4(Type,Obj,Tight,EnumWarning).

%enumerate_basic_type2(X,Type) :-
%  (ground(X) -> (basic_type(X,Type) -> true
%                    ; add_internal_error('Type error: ',enumerate_basic_type2(X,Type)))
%             ; enumerate_basic_type3(Type,X)).

enumerate_basic_type4(global(T),R,_Tight,EnumWarning) :-
     enumerate_global_type_with_enum_warning(R,T,EnumWarning).
enumerate_basic_type4(set(X),Set,Tight,EnumWarning) :-
     enumerate_basic_type_set(Set,X,Tight,EnumWarning).
enumerate_basic_type4(seq(SeqRanType),Seq,Tight,EnumWarning) :-
     (Tight = tight -> enumerate_seq_type(Seq,SeqRanType,EnumWarning) % might trigger warning. push flag.
       ; enumerate_basic_type4(set(couple(integer,SeqRanType)),Seq,basic,EnumWarning)).
enumerate_basic_type4(couple(XT,YT),(X,Y),Tight,EnumWarning) :-
     enumerate_type(X,XT,Tight,EnumWarning),enumerate_type(Y,YT,Tight,EnumWarning).
enumerate_basic_type4(boolean,B,_Tight,_EnumWarning) :- enumerate_bool(B).
enumerate_basic_type4(string,string(S),_Tight,EnumWarning) :- enumerate_string(S,EnumWarning).
enumerate_basic_type4(constant([V]),term(V),_Tight,_EnumWarning).
enumerate_basic_type4(record(FT),rec(F),Tight,EnumWarning) :- %print(enum_rec(FT,F)),nl,
   enumerate_basic_field_types(F,FT,Tight,EnumWarning). %, print(rec(F)),nl.
enumerate_basic_type4(freetype(Id),freeval(Id,C,Value),Tight,_EnumWarning) :-
	 enumerate_freetype(Tight,freeval(Id,C,Value),freetype(Id)).
enumerate_basic_type4(freetype(Id,Depth),freeval(Id,C,Value),Tight,_EnumWarning) :-
	 enumerate_freetype(Tight,freeval(Id,C,Value),freetype(Id,Depth)).
enumerate_basic_type4(integer,int(N),Tight,EnumWarning) :-
   (nonvar(N)
    -> (number(N) -> true ; add_internal_error('Illegal value: ',enumerate_basic_type4(integer,int(N),Tight,EnumWarning)))
     ; enumerate_int(N,EnumWarning)).
enumerate_basic_type4(abort,V,Tight,EnumWarning) :-
    add_internal_error(deprecated_abort_type,enumerate_basic_type4(abort,V,Tight,EnumWarning)).
enumerate_basic_type4(constant,V,Tight,EnumWarning) :-
    add_internal_error(deprecated_abort_type,enumerate_basic_type4(constant,V,Tight,EnumWarning)).
enumerate_basic_type4(any,Obj,_Tight,EnumWarning) :- enumerate_any(Obj,EnumWarning).

:- use_module(library(random),[random/3]).
enumerate_bool(X) :- preferences:preference(randomise_enumeration_order,true),
      random(1,3,1),!,
      (X=pred_false ; X=pred_true).
enumerate_bool(pred_true). /* was bool_true */
enumerate_bool(pred_false).

max_cardinality_string(inf). % was 2
all_strings(AS) :- findall(string(S),enumerate_string(S,trigger_throw(all_strings)),AS).
:- use_module(btypechecker,[machine_string/1]).
enumerate_string(S,_EnumWarning) :- atomic(S),!.
enumerate_string(S,EnumWarning) :- %print('### WARNING, Enumerating STRING'),nl,
  % frozen(S,Goal), print(enum(S,Goal)),nl,
  %  MAYBE TO DO: we could check if prolog:dif(S,'"STR1"') are in frozen Goal and then enumerate more?
  % if we do this we need to adapt dont_expand(global('STRING')) :- ... further below
  enum_warning('STRING',inf,'"STRING1","STRING2",...',EnumWarning),
  (S = 'STRING1', \+ btypechecker:machine_string(S)  % used to be '"STR1"'
  ; S = 'STRING2', \+ btypechecker:machine_string(S) % used to be '"STR2"'
  ; btypechecker:machine_string(S)).

is_string(string(_),_WF).
is_not_string(X) :- top_level_dif(X,string).

:- block enumerate_any(-,-).
enumerate_any(fd(X,T),EnumWarning) :- !,
    when(nonvar(T),enumerate_global_type_with_enum_warning(fd(X,T),T,EnumWarning)).
enumerate_any(int(N),EnumWarning) :- !,enumerate_basic_type4(integer,int(N),basic,EnumWarning).
enumerate_any(term(X),_EnumWarning) :- !, print_message(could_not_enumerate_term(X)).
enumerate_any(string(S),EnumWarning) :- !, enumerate_string(S,EnumWarning).
enumerate_any(pred_true /* bool_true */,_EnumWarning) :- !.
enumerate_any(pred_false /* bool_false */,_EnumWarning) :- !.
enumerate_any([],_EnumWarning) :- !.
enumerate_any([H|T],EnumWarning) :- !, enumerate_any(H,EnumWarning), enumerate_any(T,EnumWarning).
enumerate_any((H,T),EnumWarning) :- !, enumerate_any(H,EnumWarning), enumerate_any(T,EnumWarning).
enumerate_any(T,_EnumWarning) :- print_message(could_not_enumerate(T)).


:- use_module(preferences,[preference/2]).
:- use_module(library(clpfd),[labeling/2]). %, indomain/1]).
% enumerate an INTEGER variable
enumerate_int(N,EnumWarning) :-
   clpfd_domain(N,FDLow,FDUp),  % print(enum(N,FDLow,FDUp)),nl,
   (finite_domain(FDLow,FDUp)
     -> label(N,FDLow,FDUp)
      ; enum_unbounded(FDLow,FDUp,N,EnumWarning)
   ).
label(N,FDLow,FDUp) :-
    gen_enum_warning_if_large(N,FDLow,FDUp),
    clpfd_interface:clpfd_randomised_labeling([],[N]).
    % Note:  CLP(FD) labeling  does not necessarily try all values in range (disjunctive domains)

% when in CLP(FD) mode; try and do a case-split and see if that narrows down the possible ranges
enum_unbounded(X,Y,N,EnumWarning) :- preferences:preference(use_clpfd_solver,true),!,
    enum_unbounded_clp(X,Y,N,EnumWarning).
enum_unbounded(_FDLow,_FDUp,N,_EnumWarning) :- is_wdguarded_result_variable(N),!.
enum_unbounded(FDLow,FDUp,N,EnumWarning) :-
    make_domain_finite(FDLow,FDUp,Min,Max),
    enum_warning('INTEGER',FDLow:FDUp,Min:Max,EnumWarning),
    enumerate_int1(N,Min,Max). % will also do a case split, but without posting constraints

% check if this variable is marked as being assigned to by currently not-well-defined construct such as min,max,...:
is_wdguarded_result_variable(N) :-
    frozen(N,FrozenGoal), %print(frozen(N,FrozenGoal)),nl,
    is_wdguarded_result_variable_aux(FrozenGoal,N). %, print(not_enumerating(N)),nl.
is_wdguarded_result_variable_aux(kernel_waitflags:is_wd_guarded_result(V),N) :- !, N==V.
is_wdguarded_result_variable_aux((A,B),N) :-
   is_wdguarded_result_variable_aux(A,N) ; is_wdguarded_result_variable_aux(B,N).

% enumerate unbounded integer variable N in a CLP(FD) fashion:
enum_unbounded_clp(0,Y,N,EnumWarning) :- (Y=sup ; Y>0),
   % we span 0 and positive numbers
   !,
   %print(case_split_0(0,Y,N)),nl,
   (N=0
     % for division/modulo... 0 is often a special case
    ; post_constraint(N #>0), force_enumerate_int_wo_case_split(N,'INTEGER',EnumWarning)
    ).
enum_unbounded_clp(X,Y,N,EnumWarning) :-
   (X=inf -> true ; X<0), (Y=sup ; Y>0),
   % we span both negative and positive numbers
   !,
   % do a case split
   %print(case_split(X,Y,N)),nl,
   (N=0
    % Instead of doing a case-split on 0; we could try and detect other relevant values (e.g., what if we have x / (y-1)
    ; post_constraint(N #>0), force_enumerate_int_wo_case_split(N,'INTEGER',EnumWarning)
    ; post_constraint(N #<0), force_enumerate_int_wo_case_split(N,'INTEGER',EnumWarning)
    ).
enum_unbounded_clp(FDLow,FDUp,N,EnumWarning) :-
   % we cover only negative or only positive numbers
   force_enumerate_with_warning(N,FDLow,FDUp,'INTEGER',EnumWarning).

% force enumeration without case split:
force_enumerate_int_wo_case_split(N,Msg,EnumWarning) :-
   clpfd_domain(N,FDLow,FDUp),  % print(enum(N,FDLow,FDUp)),nl,
   (finite_domain(FDLow,FDUp)
     ->  label(N,FDLow,FDUp)
     ; %print(force_enumerate_int_wo_case_split(FDLow,FDUp)),nl,
       force_enumerate_with_warning(N,FDLow,FDUp,Msg,EnumWarning)
   ).

force_enumerate_with_warning(N,_FDLow,_FDUp,_Msg,_EnumWarning) :- % check if we should enumerate at all
    is_wdguarded_result_variable(N),!.
force_enumerate_with_warning(N,FDLow,FDUp,Msg,EnumWarning) :-
    make_domain_finite(FDLow,FDUp,Min,Max),
    enum_warning(Msg,FDLow:FDUp,Min:Max,EnumWarning),
    %post_constraint(N in Min..Max),  % I am not sure whether this is useful or not
    %print(posted_in(N,Min,Max)),nl,trace,
    enumerate_int2(N,Min,Max).

%enum_warning3(TYPE,RANGE,RESTRICTED_RANGE) :- enum_warning(TYPE,RANGE,RESTRICTED_RANGE,trigger_true(unknown)).

enum_warning(TYPE,RANGE,RESTRICTED_RANGE,Trigger) :-
   Warning = enumeration_warning(enumerating(Info),TYPE,RANGE,RESTRICTED_RANGE,critical),
   (   get_trigger_info(Trigger,Info)
    -> true
    ;  Info=unknown),
   (add_new_event_in_error_scope(Warning,print_enum_warning(Trigger,TYPE,RANGE,RESTRICTED_RANGE)) % may also throw(Warning)
     ->
       (preference(allow_enumeration_of_infinite_types,false)
           -> print('### VIRTUAL TIME-OUT generated because ENUMERATE_INFINITE_TYPES=false'),nl, % trace,
              print_pending_abort_error(Info),
              throw(Warning)
         ; Trigger = trigger_throw(Source)
           -> print('### VIRTUAL TIME-OUT generated for '),print(Source),nl, % trace,
              print_pending_abort_error(Info),
              throw(Warning)
         ; true)
      ;  true).

get_trigger_info(trigger_false(I),Info) :- get_trigger_info2(I,Info). % was non_critical ; TO DO: simplify !
get_trigger_info(trigger_true(I),Info)  :- get_trigger_info2(I,Info).
get_trigger_info(trigger_throw(I),Info) :- get_trigger_info2(I,Info).
get_trigger_info2(enum_wf_context(_,Info),Res) :- !,Res=Info.
get_trigger_info2(Info,Info).

get_trigger_info_wf_context(enum_wf_context(WF,_),WF).

% TO DO: pass WF explicitly rather than extracting it from enumeration warning terms
:- use_module(translate,[translate_span/2, translate_error_term/2]).
print_pending_abort_error(Info) :-
    get_trigger_info_wf_context(Info,WF),
    pending_abort_error(WF,Msg,Term,Span),!, % just print one error
    translate:translate_span(Span,TSpan), translate:translate_error_term(Term,TT),
    format('   (could be due to WD-Error ~w: ~w ~w)~n',[TSpan,Msg,TT]).
print_pending_abort_error(_).

print_enum_warning(Trigger,TYPE,RANGE,RESTRICTED_RANGE,THROWING) :-
        print_throwing(THROWING),
        print('### Warning: enumerating '), % error_manager:trace_if_user_wants_it,
        print_trigger_var(Trigger,Info),
        print(TYPE), print(' : '),
        print(RANGE), print(' ---> '),  print(RESTRICTED_RANGE),nl,
        (THROWING=throwing -> print_pending_abort_error(Info)
         ; true). %%, trace.

print_throwing(THROWING) :-
   (preference(strict_raise_enum_warnings,true)
    -> add_error(strict_raise_enum_warnings,'Enumeration warning occured') ; true),
   (THROWING=throwing -> print('VIRTUAL TIME-OUT caused by: ') ; true).

print_trigger_var(trigger_true(Info),Info)  :- !, print_trigger_var_aux(Info), print(' : ').
print_trigger_var(trigger_throw(Info),Info) :- !, print_trigger_var_aux(Info), print(' : (all_solutions) : ').
print_trigger_var(trigger_false(Info),Info) :- !, print_trigger_var_aux(Info), print(' (not critical [unless failure]) : ').
print_trigger_var(X,X) :- print(' UNKNOWN TRIGGER: '),print(X), print(' : ').

print_trigger_var_aux(enum_wf_context(_WF,VarID)) :- !, print(VarID).
print_trigger_var_aux(VarID) :- print(VarID).

% generate a warning if a large range is enumerated
gen_enum_warning_if_large(Var,FDLow,FDUp) :-
   (FDUp>FDLow+8388608  /* 2**23 ; {x|x:1..2**23 & x mod 2 = x mod 1001} takes about 2 minutes */
    % however the domain itself could be very small ! we could use clpfd_size instead
            -> enum_warning_large(Var,'INTEGER',FDLow:FDUp)
            ;  true).
enum_warning_large(_Var,TYPE,RANGE) :-
   Warning = enumeration_warning(enumerating,TYPE,RANGE,RANGE,non_critical),
   (add_new_event_in_error_scope(Warning,print_enum_warning_large(TYPE,RANGE))
     -> %(b_enumerate:get_prolog_variable_name(_Var,Name,_Type) -> print(' : Identifier = '), print(Name) ; true),
        %nl,frozen(_Var,G), print(frozen(G)), % comment in to get info about variable
        %% trace,
        (preference(allow_enumeration_of_infinite_types,true) -> true
          ;  debug_println(9,'### THROWING EXCEPTION'), %trace,
             throw(Warning))
      ;  true).
print_enum_warning_large(TYPE,RANGE,THROWING) :-
        print_throwing(THROWING),
        print('### Warning: enumerating large range '),
        print(TYPE), print(' : '),
        print(RANGE),nl.

:- block finite_warning(-,?,?,?,?).
finite_warning(_,Par,Types,Body,Source) :-
     add_new_event_in_error_scope(enumeration_warning(checking_finite_closure,Par,Types,finite,critical),
                                 print_finite_warning(Par,Types,Body,Source) ),
     fail. % WITH NEW SEMANTICS OF ENUMERATION WARNING WE SHOULD PROBABLY ALWAYS FAIL HERE !
print_finite_warning(Par,Types,Body,Source,THROWING) :-
        print_throwing(THROWING),
        print('### Warning: could not determine set comprehension to be finite: '),
        translate:print_bvalue(closure(Par,Types,Body)),nl,
        print('###  Source: '), print(Source),nl.

:- block enumerate_natural(-,?,-).
enumerate_natural(N,From,_) :- nonvar(N) -> true ; enumerate_natural(N,From).
enumerate_natural(N,From) :- clpfd_domain(N,FDLow,FDUp),   % print(enumerate_nat(N,From,FDLow,FDUp)),nl,
   (finite_domain(FDLow,FDUp)
     ->  label(N,FDLow,FDUp)
     ;    (number(FDLow),FDLow>=From -> FDLow1=FDLow ; FDLow1=From),
         (FDLow1=0
            -> (N=0 ; /* do a case split */
                post_constraint(N #>0), % this can sometimes make the domain finite
                force_enumerate_int_wo_case_split(N,'NATURAL',trigger_true('NATURAL'))
               )
             ;  force_enumerate_with_warning(N,FDLow1,FDUp,'NATURAL(1)',trigger_true('NATURAL(1)'))
         )
   ).


% assumes one of FDLow and FDUp is not a number
make_domain_finite(FDLow,_FDUp,Min,Max) :- number(FDLow),!,Min=FDLow,
   preferences:preference(maxint,MaxInt),
   (MaxInt>=FDLow -> Max=MaxInt ; Max=FDLow). % ensure that we try at least one number
make_domain_finite(_FDLow,FDUp,Min,Max) :- number(FDUp),!,Max=FDUp,
   preferences:preference(minint,MinInt),
   (MinInt=<FDUp -> Min=MinInt ; Min=FDUp).
make_domain_finite(_FDLow,_FDUp,Min,Max) :-
  ((preferences:preference(maxint,Max),
    preferences:get_preference(minint,Min))->true). % ensure that we try at least one number

enumerate_int1(N,Min,Max) :-
     (Min<0 /* enumerate positive numbers first; many specs only use NAT/NATURAL */
         ->  (enumerate_int2(N,0,Max) ; enumerate_int2(N,Min,-1))
          ;  enumerate_int2(N,Min,Max)
       ).
enumerate_int(X,Low,Up) :- get_int_domain(X,Low,Up,RL,RU),
   %% print(enumerate_int(X,Low,Up, RL,RU)),nl, %%
   enumerate_int2(X,RL,RU).

get_int_domain(X,Low,Up,RL,RU) :- clpfd_domain(X,FDLow,FDUp),
    fd_max(FDLow,Low,RL),fd_min(FDUp,Up,RU).

finite_domain(Low,Up) :- \+ infinite_domain(Low,Up).
infinite_domain(inf,_) :- !.
infinite_domain(_,sup).

fd_max(inf,L,R) :- !,R=L.
fd_max(FDX,Y,R) :- ((nonvar(FDX),nonvar(Y),FDX>Y) -> R=FDX ; R=Y).
fd_min(sup,L,R) :- !,R=L.
fd_min(FDX,Y,R) :- ((nonvar(FDX),nonvar(Y),FDX<Y) -> R=FDX ; R=Y).

:- use_module(extension('random_permutations/random_permutations')).
:- use_module(library(random),[random/1]).
enumerate_int2(N,X,Y) :-
    preferences:get_preference(randomise_enumeration_order,true)
    -> enumerate_int_random(N,X,Y) ; enumerate_int2_linear(N,X,Y).

enumerate_int2_linear(N,X,Y) :- X=<Y,
    (N=X ; X1 is X+1, enumerate_int2_linear(N,X1,Y)).

enumerate_int_random(N,X,Y) :-
    init_random_permutations,
    %format('enumerate_int2: Enumerating ~w from ~w to ~w~n',[N,X,Y]),
    IntervalLength is Y - X + 1,
    get_num_bits(IntervalLength,MaxIdx,NumBits),
    get_masks(NumBits,LeftMask,RightMask),
    % the seed relies on the random predicate, not on now/1, thus prob can be made deterministic by setting a central random seed
    random(TempSeed),
    Seed is floor(TempSeed * 10000),
    enumerate_int_random_aux(N,0,MaxIdx,X,Y,Seed,NumBits,LeftMask,RightMask).

enumerate_int_random_aux(N,CurIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask) :-
    random_permutation_element(CurIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask,Drawn,NextIdx),
    %format('enumerate_int2: Setting variable ~w to ~w~n',[N,Drawn]),
    ( N=Drawn
    ; enumerate_int_random_aux(N,NextIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask)).

enumerate_basic_type_set(X,Type,Tight,EnumWarning) :- var(X),!,
   max_cardinality_with_check(Type,Card),
   enumerate_basic_type_set2(X,[],Card,Type,none,Tight,EnumWarning).
enumerate_basic_type_set([],_,_,_EnumWarning) :- !.
enumerate_basic_type_set(avl_set(_),_,_,_EnumWarning) :- !.
enumerate_basic_type_set(freetype(_),_,_,_EnumWarning) :- !.
enumerate_basic_type_set(global_set(GS),Type,_Tight,_EnumWarning) :- !,
     (Type = global(GT)
      -> (GS = GT -> true
          ; nonvar(GS), add_error_and_fail(enumerate_basic_type_set,'Type error in global set: ',GS:GT))
      ;  Type = integer,integer_global_set(GS)
      ;  Type = string, string_global_set(GS)
      ).
enumerate_basic_type_set(closure(_Parameters, _ParameterTypes, Body),_Type,_Tight,EnumWarning) :- !,
      (ground(Body) -> true
        ;  print('### Enumerating non-ground closure-body: '), print_term_summary(Body),nl,
           %translate:print_bexpr(Body),nl,
           term_variables(Body,Vars), print('### Variables: '), print(Vars),nl, %trace,
           print('### EnumWarning Info: '), print(EnumWarning),nl,
           enumerate_values_inside_expression(Body)
      ).
enumerate_basic_type_set([H|T],Type,Tight,EnumWarning) :- !,
   % collect bound elements; avoid enumerating initial elements with elements that already appear later
   collect_bound_elements([H|T], SoFar,Unbound,Closed),
   (Closed=false -> max_cardinality_with_check(Type,Card)
     ; Card = Closed),
     % print(enum(Card,Unbound,SoFar,[H|T],Closed)),nl,
    enumerate_basic_type_set2(Unbound,SoFar,Card,Type,none,Tight,EnumWarning).
%enumerate_basic_type_set([H|T],Type,Tight) :- !,
%  (is_list_skeleton([H|T],Card) -> true
%     ;  max_cardinality_with_check(Type,Card)
%  ),
%   enumerate_basic_type_set2([H|T],[],Card,Type,none,Tight).
enumerate_basic_type_set(S,Type,Tight,EnumWarning) :-
  add_internal_error('Illegal set: ',enumerate_basic_type_set(S,Type,Tight,EnumWarning)).

enumerate_basic_type_set2(HT,ElementsSoFar,_Card,_Type,_Last,_Tight,_EnumWarning) :- nonvar(HT),
   is_custom_explicit_set(HT,enumerate_basic_type),!,
   disjoint_sets(HT,ElementsSoFar). % I am not sure this is necessary; probably other constraints already ensure this holds
enumerate_basic_type_set2(HT,ElementsSoFar,Card,Type,Last,Tight,EnumWarning) :- var(HT),
   preferences:preference(randomise_enumeration_order,true),!,
   (random(1,3,1)
     -> (enumerate_basic_type_set_cons(HT,ElementsSoFar,Card,Type,Last,Tight,EnumWarning)
          ; HT = [])
     ;  (HT = [] ;
         enumerate_basic_type_set_cons(HT,ElementsSoFar,Card,Type,Last,Tight,EnumWarning))
   ).
enumerate_basic_type_set2([],_,_,_,_,_Tight,_EnumWarning).
enumerate_basic_type_set2(HT,ElementsSoFar,Card,Type,Last,Tight,EnumWarning) :-
   enumerate_basic_type_set_cons(HT,ElementsSoFar,Card,Type,Last,Tight,EnumWarning).

enumerate_basic_type_set_cons(HT,ElementsSoFar,Card,Type,Last,Tight,EnumWarning) :-  positive_card(Card),
   %debug:trace_point(enum(HT,ElementsSoFar,Card,Type,Last,Tight)),
   (var(HT) -> HT=[H|T], NewLast=NormH /* the enumerator has completely determined H */
    % Note: HT=[H|T] may wake up co-routines and then attach infos to H; but these should hold indpendently for all elements
    ; HT=[H|T],
      (unbound_value(H) -> NewLast=NormH /* the enumerator has completely determined H */
           ;  NewLast=Last)  /* H was not freely chosen by the enumerator */
   ),
   not_element_of(H,ElementsSoFar), % this is only needed for elements generated by the enumerator itself
   enumerate_type(H,Type,Tight,EnumWarning),
   % TO DO: extract normal form from add_new_element
   val_greater_than(H,NormH,Last),
   %(val_greater_than(H,NormH,Last) -> print(ok(H)),nl ; print(pruned(H,NormH,Last)),nl,fail),
   %(ground(NormH) -> true ; print(not_ground(NormH)),nl),
   C1 is Card-1,
   add_new_element(NormH,ElementsSoFar,SoFar2), % we could use avl_store(NormH,A,true,A2) if ElementsSoFar=avl_set(A)
   %debug:trace_point(add_new_element(NormH,ElementsSoFar,SoFar2)),
   enumerate_basic_type_set2(T,SoFar2,C1,Type,NewLast,Tight,EnumWarning).

:- assert_must_succeed((collect_bound_elements([int(1),int(2),int(4),X,int(5)|T],_,U,C),U==[X|T],C==false)).
:- assert_must_succeed((collect_bound_elements([int(1),int(2),int(4),X,int(5)],_,U,C),U==[X],C==1)).
:- assert_must_succeed(exhaustive_kernel_succeed_check(collect_bound_elements([int(1),int(2),int(4),int(5)],_,_,_))).

% collect the bound and unbound elements in a list; also return if the list is closed (then return length) or return false
collect_bound_elements(T, SoFar,Unbound,Closed) :- var(T),!, SoFar=[],Unbound=T,Closed=false.
collect_bound_elements([],[],[],0).
collect_bound_elements(avl_set(A),avl_set(A),[],0).
collect_bound_elements(global_set(GS),SoFar,Unbound,Closed) :- expand_custom_set(global_set(GS),ES),
   collect_bound_elements(ES,SoFar,Unbound,Closed).
collect_bound_elements(freetype(FS),SoFar,Unbound,Closed) :- expand_custom_set(freetype(FS),ES),
   collect_bound_elements(ES,SoFar,Unbound,Closed).
collect_bound_elements(closure(P,T,B),SoFar,Unbound,Closed) :- expand_custom_set(closure(P,T,B),ES),
   collect_bound_elements(ES,SoFar,Unbound,Closed).
collect_bound_elements([H|T],SoFar,Unbound,Closed) :-
   collect_bound_elements(T,TSoFar,TUnbound,TClosed),
   (ground(H) -> add_new_element(H,TSoFar,SoFar), Unbound=TUnbound, TClosed=Closed
     ;  SoFar = TSoFar, Unbound = [H|TUnbound],
       (TClosed=false -> Closed=false ; Closed is TClosed+1)
    ).

% use_module(kernel_objects),findall(A,enumerate_tight_type(A,set(set(boolean))),B),length(B,C).
% use_module(kernel_objects),findall(A,enumerate_tight_type(A,set(set(global('Name')))),B),length(B,C).
% use_module(kernel_objects),findall(A,enumerate_tight_type(A,set(global('Name'))),B),length(B,C).

% perform order checking on terms, normalising them first
% val_greater_than(A,NormA,NormB)
val_greater_than(A,NormA,NormB) :- !,
   (nonvar(A),custom_explicit_sets:convert_to_avl_inside_set(A,NormA)
     -> (NormB==none -> true ; NormA @> NormB)
     ;  add_internal_error('Call failed: ',custom_explicit_sets:convert_to_avl_inside_set(A,NormA)),
        NormA = A).

positive_card(inf) :- !, print('$').
positive_card(C) :- (number(C) -> C>0
                        ; add_internal_error('Not number: ',positive_card(C)),fail).



:- block enumerate_basic_field_types(?,-,?,-).
enumerate_basic_field_types([],[],_Tight,_EnumWarning).
enumerate_basic_field_types(Fields,[field(Name,VT)|TT],Tight,EnumWarning) :-
  enumerate_basic_field_types2(Fields,Name,VT,TT,Tight,EnumWarning).

:- block enumerate_basic_field_types2(?,-,?,?,?,?).
enumerate_basic_field_types2([field(Name,V)|T], Name,VT,TT,Tight,EnumWarning) :-
  enumerate_type(V,VT,Tight,EnumWarning),
  enumerate_basic_field_types(T,TT,Tight,EnumWarning).


:- block all_objects_of_type(-,?).
all_objects_of_type(Type,Res) :-
  findall(O,enumerate_basic_type(O,Type),Res).

list_length(X,L,Type,MaxIndex) :- var(X),!,L=0, Type=open, MaxIndex=0.
list_length([],0,closed,0).
list_length([H|T],C1,Type,MaxIndex1) :- list_length(T,C,Type,MaxIndex),
  C1 is C+1,
  ((nonvar(H),H=(I,_),nonvar(I),I=int(Idx),number(Idx),Idx>MaxIndex)
    -> MaxIndex1 = Idx  ; MaxIndex1 = MaxIndex).


:- assert_must_succeed((max_cardinality(set(couple(global('Name'),global('Code'))),64))).
:- assert_must_succeed((max_cardinality(set(set(set(couple(global('Name'),global('Code'))))),_))).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc4,[case(a,boolean),case(b,couple(boolean,boolean))]),
			max_cardinality(freetype(selfc4),6),
			kernel_freetypes:reset_freetypes)).
:- assert_must_succeed((kernel_freetypes:add_freetype(selfc6,[case(a,boolean),case(b,freetype(selfc6)),case(c,constant([c]))]),
			kernel_freetypes:set_freetype_depth(3),
			findall(X,enumerate_tight_type(X,freetype(selfc6)),Solutions),
			length(Solutions,NumberOfSolutions),
			max_cardinality(freetype(selfc6),NumberOfSolutions),
			kernel_freetypes:reset_freetypes)).

:- use_module(tools_printing,[print_error/1]).
max_cardinality_with_check(Set,CCard) :-
   (max_cardinality(Set,Card) ->
     (Card=inf
      -> debug_println(9,very_large_cardinality(Set,Card)),
         CCard = 20000000
      ;  CCard=Card,
         (Card>100 -> debug_println(9,large_cardinality(Set,Card)) ; true)
     )
     ; print_error(failed(max_cardinality(Set,CCard))), CCard = 10
   ).
max_cardinality(global(T),Card) :- b_global_set_cardinality(T,Card).
max_cardinality(boolean,2).
max_cardinality(constant([_V]),1).
max_cardinality(any,inf) :- print_message(dont_know_card_of_any). /* what should we do here ? */
max_cardinality(string,MC) :- max_cardinality_string(MC). % is inf now
%max_cardinality(abort,1).
max_cardinality(integer,Card) :- Card=inf. %b_global_set_cardinality('INTEGER',Card).
max_cardinality(seq(X),Card) :- % Card=inf, unless a freetype can be of cardinality 0
   max_cardinality(set(couple(integer,X)),Card).
max_cardinality(couple(X,Y),Card) :-
   max_cardinality(X,CX), max_cardinality(Y,CY), safe_mul(CX,CY,Card).
max_cardinality(record([]),1).
max_cardinality(record([field(_,T1)|RF]),Card) :-
   max_cardinality(record(RF),RC),
   max_cardinality(T1,C1),
   safe_mul(C1,RC,Card).
max_cardinality(set(X),Card) :- max_cardinality(X,CX),
  safe_pow2(CX,Card).
  % RealCard is 2**CX, (RealCard is inf -> Card is inf ; Card is integer(RealCard)).
max_cardinality(freetype(Id),Card) :- max_cardinality_freetype(freetype(Id),Card).
max_cardinality(freetype(Id,Depth),Card) :- max_cardinality_freetype(freetype(Id,Depth),Card).

:- assert_must_succeed((safe_pow2(3,R),R==8)).
:- assert_must_succeed((safe_pow2(inf,R),R==inf)).
:- assert_must_succeed((safe_pow2(3072,R),R==inf)).
:- assert_must_succeed((safe_pow2(18446744073709551616,R),R==inf)).
:- assert_must_succeed((safe_pow2(500,X), safe_pow2(501,X2), X2 is 2*X)).
% :- assert_must_succeed((kernel_objects:safe_pow2(1022,X), kernel_objects:safe_pow2(1023,X2), X2 is 2*X)).

safe_pow2(Exp,Res) :- (Exp==inf -> Res=inf
   ; Exp>1023 -> Res=inf  /* the limit where SICStus 4.2.1 reported inf; 4.2.3 goes further but uses a huge amount of memory */
   ; Res is 2^Exp % ^ integer exponentiation operator new in SICStus 4.3
   ).
% this seems to be either precise or give inf (at 1023 on 64 bit system)

:- assert_must_succeed((safe_pown(3,2,R),R==9)).
:- assert_must_succeed((safe_pown(2,3072,R),R==inf)).
:- assert_must_succeed((safe_pown(3,647,R),R==inf)).
:- assert_must_succeed((safe_pown(2,500,X), safe_pown(2,501,X2), X2 is 2*X)).
:- assert_must_succeed((safe_pown(2,500,X), safe_pow2(500,X))).
:- assert_must_succeed((kernel_objects:safe_pown(2,1022,X), kernel_objects:safe_pown(2,1023,X2), X2 is 2*X)).
:- assert_must_succeed((kernel_objects:safe_pown(3,500,X), kernel_objects:safe_pown(3,501,X3), X3 is 3*X)).
safe_pown(Base,Exp,Res) :-
  (Base=inf -> (Exp=0 -> Res=1 ; Res=inf)
    ; Exp=inf -> (Base=0 -> Res=0 ; Base=1 -> Res=1 ; Res=inf)
    ; infinite_pown(Base,Exp) -> Res=inf /* SICStus 4.2.1 reported inf; 4.2.3 goes further but uses a huge amount of memory */
    ; Res is Base ^ Exp % ^ integer exponentiation operator new in SICStus 4.3
  ).

/* SICStus 4.2.1 reported inf; 4.2.3 goes further but uses a huge amount of memory */
infinite_pown(Base,Exp) :- Exp > 1023, !, Base >= 2.
infinite_pown(Base,Exp) :- Exp > 646, !, Base >= 3. % was not really necessary when using overflow_float_pown
infinite_pown(Base,Exp) :- Exp > 511, !, Base >= 4.
infinite_pown(Base,Exp) :- Exp > 441, !, Base >= 5.

:- assert_must_succeed((safe_mul(3,2,R),R==6)).
:- assert_must_succeed((safe_mul(inf,2,R),R==inf)).
:- assert_must_succeed((safe_mul(2,inf,R),R==inf)).
:- assert_must_succeed((safe_mul(0,inf,R),R==0)).
:- assert_must_succeed((safe_mul(inf,0,R),R==0)).
% safe_multiplication for positive numbers
safe_mul(0,_,R) :- !, R=0. % true for cartesian product: card({}*INTEGER)=0
safe_mul(_,0,R) :- !, R=0. % ditto
safe_mul(inf,_,R) :- !, R=inf.
safe_mul(_,inf,R) :- !, R=inf.
safe_mul(X,Y,R) :- is_overflowcheck(R,X*Y).

safe_add(inf,_,R) :- !, R=inf.
safe_add(_,inf,R) :- !, R=inf.
safe_add(X,Y,R) :- is_overflowcheck(R,X+Y).


/* is with overflow check */
is_overflowcheck(Var,Expr) :-  %print(is_with_overflow(Var,Expr)),nl,
  (Expr=inf -> Var=inf
    ; on_exception(error(_,_), Var is Expr, Var=inf)).

%overflow_float_pown(Base,Exp,Res) :-
%  on_exception(error(_,_),
%     (R1 is Base**Exp,  /* separate into two steps; SICStus 4.2.3 otherwise seems to use integer exponentation */
%      Res is integer(R1)), Res=inf).

/*
% this code below would sometimes fail in spld generated code:
is_overflowcheck_old(Var,Expr) :- print(is_with_overflow(Var,Expr)),nl,
  (Expr=inf -> Var=inf
    ;  on_exception(error(type_error(evaluable,inf/0),_),
        on_exception(error(evaluation_error(float_overflow),_),
                     Var is Expr,
                     (print(float_overflow),nl,Var = inf)),
                     (print(evaluable_inf),nl,Var=inf))).
% catches any exception (e.g. also representation_error when converting inf to integer)
% could be unsafe with timeout !!
is_errc(Var,Expr) :-
  (Expr=inf -> Var=inf
            ;  safe_on_exception(_E, Var is Expr,  Var = inf)).
*/

/* ---------------------------- */


/* use a cleverer, better enumeration than enumerate_basic_type */
/* can only be used in certain circumstances: operation preconditions,
  properties,... but not for VARIABLES as there is no guarantee that
  something declared as a sequence will actually turn out to be a sequence */

:- assert_pre(kernel_objects:enumerate_tight_type(Obj,Type),
             (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:enumerate_tight_type(Obj,_), (type_check(Obj,bsets_object),ground_check(Obj))).
:- assert_pre(kernel_objects:enumerate_tight_type(Obj,Type,_),
             (type_check(Obj,bsets_object),type_check(Type,basic_type_descriptor))).
:- assert_post(kernel_objects:enumerate_tight_type(Obj,_,_), (type_check(Obj,bsets_object),ground_check(Obj))).

:- assert_must_succeed(enumerate_tight_type([(int(1),int(2)),(int(2),int(4))],
                                             seq(integer) )).
:- assert_must_succeed(enumerate_tight_type([(int(1),int(2))],seq(integer) )).
:- assert_must_succeed(enumerate_tight_type([],seq(integer) )).
:- assert_must_succeed((enumerate_tight_type(X,record([field(a,integer),field(b,global('Name'))])),
                        equal_object(X,rec([field(a,int(1)),field(b,fd(1,'Name'))])) )).
:- assert_must_fail(enumerate_tight_type([(int(1),int(2)),(int(3),int(_))],
                                             seq(integer) )).
:- assert_must_fail(enumerate_tight_type([(int(3),int(_))],seq(integer) )).
:- assert_must_succeed((bsets_clp:is_sequence(X,global_set('Name')),
                        enumerate_tight_type(X,seq(global('Name')) ),
                        X = [(int(1),fd(2,'Name'))]     )).
:- assert_must_succeed(( enumerate_tight_type(XX, record([field(balance,integer),field(name,global('Name'))])) ,
  XX = rec([field(balance,int(1)),field(name,fd(3,'Name'))]) )).
:- assert_must_succeed(( enumerate_tight_type(XX, set(record([field(balance,global('Name')),field(name,global('Name'))]))) ,                /* STILL TAKES VERY LONG !! */
  XX = [rec([field(balance,fd(3,'Name')),field(name,fd(3,'Name'))])] )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(balance,global('Name')),field(name,global('Name'))]))) ,S),
  length(S,Len), Len = 512 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(name,global('Code'))]))) ,S),
  length(S,Len), Len = 4 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(fname,global('Code')),field(name,global('Code'))]))) ,S),
  length(S,Len), Len = 16 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(record([field(fname,global('Code')),field(name,global('Name'))]))) ,S),
  length(S,Len), Len = 64 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(global('Name'))) ,S),
  length(S,Len), Len = 8 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(boolean))) ,S),
  length(S,Len), Len = 16 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(global('Name')))) ,S),
  length(S,Len), Len = 256 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(global('Code')))) ,S),
  length(S,Len), Len = 16 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(set(boolean))) ,S),
  length(S,Len), Len = 16 )).
:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(couple(global('Code'),global('Name')))) ,S),
  length(S,Len), Len = 64 )).
%:- assert_must_succeed(( findall(XX,enumerate_tight_type(XX, set(couple(global('Code'),integer))) ,S),
%  length(S,Len), Len = 64 )).
:- assert_must_succeed(( enumerate_tight_type(XX, set(record([field(balance,integer)]))) ,
  XX = [rec([field(balance,int(1))])] )).
:- assert_must_succeed(( enumerate_tight_type(global_set('Code'),set(global('Code'))) )).

:- block enumerate_tight_type(?,-).
enumerate_tight_type(Obj,Type) :- %enumerate_tight_type2(Type,Obj).
  %%print_message(call_enumerate_tight_type(Obj,Type)),
  (ground_value(Obj) -> true ;  % print(enumerate_tight_type(Obj,Type)),nl,
    enumerate_basic_type4(Type,Obj,tight,trigger_true(enumerate_tight_type))
  ).

:- block enumerate_tight_type(?,-,-).
enumerate_tight_type(Obj,Type,EnumWarning) :- %enumerate_tight_type2(Type,Obj).
  %%print_message(call_enumerate_tight_type(Obj,Type)),
  (ground_value(Obj) -> true ;  % print(enumerate_tight_type(Obj,Type)),nl,
    enumerate_basic_type4(Type,Obj,tight,EnumWarning)
  ).


/* TO DO: provide tight enumerators for nat, functions, ... ?? */



:- assert_must_succeed((X=[(int(I1),pred_true /* bool_true */),Y], dif(I1,1),
                        kernel_objects:enumerate_seq_type(X,boolean,true),I1==2,Y=(int(1),pred_false /* bool_false */))).

enumerate_seq_type(X,Type,EnumWarning) :-
   list_length(X,Len,ListType,MaxIndex), % ListType can be open or closed
   (ListType=closed
     -> MaxIndexForEnum=Len, EW = no_enum_warning,
        MaxIndex =< Len % otherwise this is obviously not a sequence
     ;  (MaxIndex>Len -> Card = MaxIndex ; Card=Len), % in case we already have an explicit index which is higher than the length we use that as index
        b_global_set_cardinality('NAT1',NatCard),
        (NatCard<Card -> Max1=Card ; Max1=NatCard),
        (Max1<1 -> MaxIndexForEnum = 1 ; MaxIndexForEnum=Max1), % ensure that we generate enumeration warning
        EW = EnumWarning
   ),
   enumerate_seq(X,range(1,MaxIndexForEnum),MaxIndexForEnum,Type,EW).

enumerate_seq([],_,_,_,_).
enumerate_seq(Seq,Indexes,Card,Type,EnumWarning) :-
   (unbound_variable_for_cons(Seq) -> positive_card(Card),VarSeq=true ; VarSeq=false),
   Seq = [El|TSeq],
   (unbound_variable(El) -> VarEl=true ; VarEl=VarSeq),
   El = (int(Index),Element),
   (VarEl=true -> get_next_index(Indexes,Index,RemIndexes)
      ; number(Index) -> remove_index_ground(Indexes,Index,RemIndexes) % this can fail if Index > MaxIndex found above ! but not first time around, i.e., we will generate enum warning anyway
      ; remove_index(Indexes,Index,RemIndexes)),
   (EnumWarning==no_enum_warning -> true
    ; enum_warning('seq (length)',inf,Card,EnumWarning)), % delay enum_warning until we have made the first case-split (sometimes instantiating the sequence to at least one element will trigger an inconsistency)
   enumerate_tight_type(Element,Type),
   C1 is Card-1,
   enumerate_seq(TSeq,RemIndexes,C1,Type,no_enum_warning).

get_next_index([Index1|RestIndexes],Index1,RestIndexes).
get_next_index(range(I1,I2),I1,Res) :-
   I11 is I1+1, (I11>I2 -> Res=[] ; Res=range(I11,I2)).

remove_index_ground(Indexes,X,Res) :- get_next_index(Indexes,H,T),
   (X=H -> Res=T ; Res=[H|R2], remove_index_ground(T,X,R2)).

remove_index(Indexes,X,Res) :- get_next_index(Indexes,H,T),
   (X=H,Res=T ; X\==H, Res=[H|R2], remove_index(T,X,R2)).



/* a few more unit tests: */

:- assert_must_succeed(( findall(X,enumerate_type(X,set(couple(boolean,boolean)),tight) ,L), length(L,16) )).
:- assert_must_succeed(( findall(X,enumerate_type(X,set(couple(boolean,boolean)),basic) ,L), length(L,16) )).

:- assert_must_succeed(( enumerate_tight_type(
  [rec([field(balance,int(0)),field(name,fd(2,'Name'))])],[
   rec([field(balance,int(1)),field(name,fd(3,'Name'))]),
   rec([field(balance,int(1)),field(name,fd(2,'Name'))]),
   rec([field(balance,int(0)),field(name,fd(1,'Name'))]),
   rec([field(balance,int(-1)),field(name,fd(1,'Name'))])],
   set(record([field(balance,integer),field(name,global('Name'))]))) )).
:- assert_must_succeed(( enumerate_tight_type([
   rec([field(balance,int(1)),field(name,fd(2,'Name'))]),
   rec([field(balance,int(1)),field(name,fd(1,'Name'))]),
   rec([field(balance,int(0)),field(name,fd(1,'Name'))]),
   rec([field(balance,int(-1)),field(name,fd(1,'Name'))])|X],
   set(record([field(balance,integer),field(name,global('Name'))]))) ,
   X = [rec([field(balance,int(1)),field(name,fd(3,'Name'))])] )).

:- assert_must_succeed((not_element_of(X,[(pred_true /* bool_true */,pred_true /* bool_true */),
 (pred_true /* bool_true */,pred_false /* bool_false */),(pred_false /* bool_false */,pred_false /* bool_false */)]),
  enumerate_tight_type(X,couple(boolean,boolean)))).

:- assert_must_succeed(( not_equal_object(X,(pred_true /* bool_true */,pred_false /* bool_false */)),
                         not_equal_object(X,(pred_false /* bool_false */,pred_false /* bool_false */)),
                         not_equal_object(X,(pred_true /* bool_true */,pred_true /* bool_true */)),
                         enumerate_tight_type(X,couple(boolean,boolean)))).

:- assert_must_succeed(( X = [fd(3,'Name')|T],enumerate_tight_type(X,set(global('Name'))),
  T == [fd(1,'Name'),fd(2,'Name')] )).



unbound_value(V) :-
   (var(V) -> unbound_variable(V)
           ;  V = (V1,W1),unbound_value(V1), unbound_value(W1)).

:- use_module(bsyntaxtree,[syntaxtraversion/6]).
enumerate_values_inside_expression(TExpr) :-
   syntaxtraversion(TExpr,Expr,Type,_Infos,Subs,_),
   nonvar(Expr),!,
   enumerate_expr(Expr,Type,Subs).
enumerate_values_inside_expression(X) :-
   add_internal_error('Unexpected B expression: ',enumerate_values_inside_expression(X)).

%:- block enumerate_expr(-,?,?).
enumerate_expr(value(X),Type,Subs) :- !,
       (ground(Type) -> enumerate_value(X,Type)
         ; add_internal_error('Value type not ground: ',enumerate_expr(value(X),Type,Subs))).
enumerate_expr(_,_,Subs) :- l_enumerate_values_inside_expression(Subs).

:- use_module(bsyntaxtree,[is_set_type/2]).
% catch a few type errors:
enumerate_value(X,Type) :- X==[], !,
  (is_set_type(Type,_) -> true ; add_internal_error('Illegal type: ',enumerate_value(X,Type))).
enumerate_value(X,Type) :- enumerate_basic_type(X,Type).

:- block l_enumerate_values_inside_expression(-).
l_enumerate_values_inside_expression([]).
l_enumerate_values_inside_expression([H|T]) :-
     enumerate_values_inside_expression(H),
     l_enumerate_values_inside_expression(T).


/* --------------- */
/* top_level_dif/2 */
/* --------------- */
/* checks whether two terms have a different top-level functor */

:- assert_must_succeed(top_level_dif(a,b)).
:- assert_must_succeed(top_level_dif(f(_X),g(_Z))).
:- assert_must_fail(top_level_dif(f(a),f(_Z))).
:- assert_must_fail(top_level_dif(f(a),f(b))).

:- block top_level_dif(-,?),top_level_dif(?,-).
top_level_dif(X,Y) :-
    functor(X,FX,_),functor(Y,FY,_), FX\=FY.   /* check arities ? */


/* ------------------------------------------------------------------- */
/*                             EQUAL OBJECT                            */
/* ------------------------------------------------------------------- */

sample_closure(C) :-
   construct_closure([xx],[integer],Body,C),
   Body = b(conjunct(b(conjunct(
    b(member(b(identifier(xx),integer,[]),b(integer_set('NAT'),set(identifier(xx)),[])),pred,[]),
    b(greater(b(identifier(xx),integer,[]),b(integer(0),integer,[])),pred,[])),pred,[]),
    b(less(b(identifier(xx),integer,[]),b(integer(3),integer,[])),pred,[])),pred,[]).

:- assert_must_succeed(equal_object([int(3),int(1)],
     closure([zz],[integer],b(member(b(identifier(zz),integer,[]),b(value([int(1),int(3)]),set(integer),[])),pred,[])))).
:- assert_must_succeed(( equal_object( (fd(1,'Name'),fd(1,'Name')) , (fd(1,'Name'),fd(1,'Name')) ) )).
:- assert_must_succeed(( equal_object( (X,Y) , (fd(2,'Name'),fd(2,'Name')) ) , X = fd(2,'Name'), Y=fd(2,'Name') )).
:- assert_must_fail(equal_object(term(a),term(b))).
:- assert_must_fail(equal_object(int(1),int(2))).
:- assert_must_fail(equal_object([term(a),term(b)],[term(a),term(c)])).
:- assert_must_fail((equal_object([(int(1),[Y])],[(int(X),[Z])]),
                    Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[int(2)])).
:- assert_must_fail(equal_object(rec([field(a,int(1))]),rec([field(a,int(2))]))).
:- assert_must_fail(equal_object(rec([field(a,int(2)),field(b,int(3))]),
                       rec([field(a,int(2)),field(b,int(4))]))).
:- assert_must_succeed(equal_object(rec([field(a,int(2))]),rec([field(a,int(2))]))).
:- assert_must_succeed(equal_object(rec([field(a,int(2)),field(b,[int(3),int(2)])]),
                                        rec([field(a,int(2)),field(b,[int(2),int(3)])]) )).
:- assert_must_succeed(equal_object([(term(a),[])],[(term(a),[])])).
:- assert_must_succeed(equal_object(_X,[int(1),int(2)])).
:- assert_must_succeed(equal_object([int(1),int(2)],_X)).
:- assert_must_succeed((equal_object([(int(1),[Y])],[(int(X),[Z])]),
                    Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[])).
:- assert_must_succeed(equal_object([int(1),int(2)],[int(2),int(1)])).
:- assert_must_succeed(equal_object(global_set('Name'),[fd(2,'Name'),fd(3,'Name'),fd(1,'Name')])).
:- assert_must_succeed(equal_object(global_set('Name'),[fd(1,'Name'),fd(3,'Name'),fd(2,'Name')])).
:- assert_must_succeed((equal_object([fd(3,'Name'),fd(2,'Name'),fd(1,'Name')],global_set('Name')))).
%:- assert_must_succeed((equal_object([fd(3,'Name'),fd(2,'Name'),fd(1,'Name')],X),X=global_set('Name'))).
:- assert_must_succeed((equal_object(Y,X),X=global_set('Name'),equal_object(Y,[fd(3,'Name'),fd(2,'Name'),fd(1,'Name')]))).
:- assert_must_succeed((equal_object(X,X),X=global_set('Name'))).
:- assert_must_succeed((equal_object(_,X),X=global_set('Name'))).
:- assert_must_succeed((equal_object(X,global_set('Name')),X=global_set('Name'))).
:- assert_must_succeed((equal_object([_A,_B],[int(2),int(1)]))).
:- assert_must_fail((equal_object(X,global_set('Code')),X=global_set('Name'))).
:- assert_must_fail((equal_object(Y,global_set('Name')),Y=[fd(3,'Name'),fd(1,'Name')])).
:- assert_must_fail((equal_object(Y,global_set('Name')),Y=[_,_])).
:- assert_must_succeed((equal_object(X,closure([xx],[integer],b(truth,pred,[]))),X==closure([xx],[integer],b(truth,pred,[])))).
:- assert_must_succeed((sample_closure(C), equal_object([int(1),int(2)],C))).
:- assert_must_succeed((sample_closure(C), equal_object(C,[int(1),int(2)]))).
:- assert_must_fail((sample_closure(C), equal_object(C,[int(1),int(0)]))).
:- assert_must_fail((sample_closure(C), equal_object(C,global_set('NAT')))).
:- assert_must_succeed((equal_object(freeval(selfcx,a,int(5)),freeval(selfcx,a,int(5))))).
:- assert_must_fail((equal_object([int(1),int(2),int(3)],global_set('NATURAL1')))).
:- assert_must_fail((equal_object(X,global_set('NATURAL1')),equal_object(X,[int(1),int(2),int(3)]))).
:- assert_must_fail((equal_object(X,[int(1),int(2),int(3)]),equal_object(X,global_set('NATURAL1')))).
:- assert_must_fail((equal_object(X,global_set('NATURAL')),equal_object(X,global_set('NATURAL1')))).
:- assert_must_succeed((equal_object(X,global_set('NATURAL')),equal_object(X,global_set('NATURAL')))).
% :- assert_must_fail((equal_object(freeval(selfcx,a,int(5)),freeval(selfcy,a,int(5))))). % is a type error
:- assert_must_fail((equal_object(freeval(selfcx,b,int(5)),freeval(selfcx,a,int(5))))).
:- assert_must_fail((equal_object(freeval(selfcx,a,int(5)),freeval(selfcx,a,int(6))))).
:- assert_must_succeed((equal_object(
 [[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
    [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(3,'Name'),fd(2,'Name')]]
,[[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
     [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(2,'Name'),fd(3,'Name')]])
     )).
:- assert_must_succeed(exhaustive_kernel_check( (equal_object([int(3),int(2),int(1)],[int(2)|T]),
equal_object(T,[int(1),int(3)])))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],equal_object([int(3),int(1)],[int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],equal_object([int(3),int(4),int(1)],[int(4),int(1),int(3)]))).

%:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],equal_object([int(1),int(2),int(3)],global_set('NATURAL1')))).
:- assert_must_succeed((     equal_object([int(0),int(5)|T],avl_set(node(int(1),true,1,node(int(0),true,0,empty,empty),node(int(3),true,1,empty,node(int(5),true,0,empty,empty))))), nonvar(T),equal_object(T,[int(_A),int(_B)]) )).
% NOTE: had multiple solutions; after solving Ticket #227 it no longer has :-)
:- assert_must_succeed((     equal_object([int(0),int(5)|T],avl_set(node(int(1),true,1,node(int(0),true,0,empty,empty),node(int(3),true,1,empty,node(int(5),true,0,empty,empty))))), nonvar(T),equal_object(T,[_A,_B]) )).

:- assert_must_succeed((equal_object([_X,_Y],[int(1),int(2)]))).
:- assert_must_succeed((equal_object([(int(1),X),(int(2),Y),(int(3),Z),(int(4),A),(int(5),B),(int(6),C),(int(7),D),(int(8),E),(int(9),F),(int(10),G)],avl_set(node((int(5),int(25)),true,0,node((int(2),int(4)),true,1,node((int(1),int(1)),true,0,empty,empty),node((int(3),int(9)),true,1,empty,node((int(4),int(16)),true,0,empty,empty))),node((int(8),int(64)),true,0,node((int(6),int(36)),true,1,empty,node((int(7),int(49)),true,0,empty,empty)),node((int(9),int(81)),true,1,empty,node((int(10),int(100)),true,0,empty,empty)))))),
 A == int(16), B == int(25),C == int(36),D == int(49),E == int(64),F == int(81),G == int(100),X == int(1),Y == int(4), Z == int(9))).

:- use_module(bool_pred).

equal_object(V1,V2) :- equal_object_wf(V1,V2,no_wf_available).
equal_object(V1,V2,Origin) :- equal_object_wf(V1,V2,Origin,no_wf_available).
equal_object_optimized(V1,V2,Origin) :- equal_object_optimized_wf(V1,V2,Origin,no_wf_available).
equal_object_optimized(V1,V2) :- equal_object_optimized(V1,V2,unknown).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
/* a version of equal_object which will convert lists to avl if possible */
equal_object_optimized_wf(V1,V2,Origin,WF) :-
    ( var(V1) -> (var(V2) -> V1=V2 ; equal_object_opt3(V2,V1,WF))
      ; equal_object_opt3(V1,V2,WF)),
    check_value(V1,Origin), check_value(V2,Origin).
equal_object_wf(V1,V2,Origin,WF) :- ( (var(V1);var(V2)) -> V1=V2
                          ; nonvar(V1) -> equal_object3(V1,V2,WF)
                          ;  equal_object3(V2,V1,WF)),
                          check_value(V1,val1(Origin)), check_value(V2,val2(Origin)).
equal_object_wf(V1,V2,WF) :- ( (var(V1);var(V2)) -> V1=V2
                          ; nonvar(V1) -> equal_object3(V1,V2,WF)
                          ;  equal_object3(V2,V1,WF)),
                          check_value(V1,equal_object1), check_value(V2,equal_object2).
check_value(X,Origin) :- nonvar(X) -> check_value_aux(X,Origin) ; true.
check_value_aux((A,B),Origin) :- !, check_value(A,pair1(Origin)), check_value(B,pair2(Origin)).
check_value_aux([H|T],Origin) :- !, check_value(H,head(Origin)), check_value(T,tail(Origin)).
check_value_aux(avl_set(X),Origin) :- !,
   (var(X) -> add_warning(Origin,'Variable avl_set')
   ; X=empty -> add_warning(Origin,'Empty avl_set') ; true).
check_value_aux(closure(P,T,B),Origin) :- !,
    (ground(P),ground(T),nonvar(B) -> true
      ;  add_warning(Origin,illegal_closure(P,T,B))).
check_value_aux(_,_Origin).
:- else.
/* a version of equal_object which will convert lists to avl if possible */
equal_object_optimized_wf(V1,V2,_Origin,WF) :-
   ( var(V1) -> (var(V2) -> V1=V2 ; equal_object_opt3(V2,V1,WF))
     ; equal_object_opt3(V1,V2,WF)).

equal_object_wf(V1,V2,_Origin,WF) :- ( (var(V1);var(V2)) -> V1=V2
                          ; nonvar(V1) -> equal_object3(V1,V2,WF)
                          ;  equal_object3(V2,V1,WF)).
equal_object_wf(V1,V2,WF) :- ( (var(V1);var(V2)) -> V1=V2
                          ; nonvar(V1) -> equal_object3(V1,V2,WF)
                          ;  equal_object3(V2,V1,WF)).
:- endif.


equal_object_opt3(int(X),Y,_WF) :- !, Y=int(X).
equal_object_opt3(fd(X,T),Y,_WF) :- !, Y=fd(X,T).
equal_object_opt3(string(X),Y,_WF) :- !, Y=string(X).
equal_object_opt3(pred_false,Y,_WF) :- !, Y=pred_false.
equal_object_opt3(pred_true,Y,_WF) :- !, Y=pred_true.
equal_object_opt3(X,S2,_WF) :-  var(S2), %unbound_variable(S2), % is it ok to assing an AVL set in one go ?!
   should_be_converted_to_avl_from_lists(X), !, % does a ground(X) check
   construct_avl_from_lists(X,S2).
%equal_object_opt3([H|T],S2) :-  var(S2),ground(H),ground(T), !, construct_avl_from_lists([H|T],S2).
equal_object_opt3(X,Y,WF) :- equal_object3(X,Y,WF).


%%equal_object3c(X,Y) :- if(equal_object3(X,Y),true,
%%   (print_message(equal_object3_failed(X,Y)),equal_object3(X,Y),fail)). %%
:- if(environ(prob_safe_mode,true)).
equal_object3(X,Y,_WF) :- (nonvar(Y) -> type_error(X,Y) ; illegal_value(X)),
    add_internal_error('Internal Typing Error (please report as bug !) : ',equal_object(X,Y)),fail.
:- endif.
%%equal_object3(X,Y,_WF) :- print(eq(X,Y)),nl,trace,fail.
equal_object3(closure(Par,ParTypes,Clo),Y,WF) :- var(Y),!,
     ( closure_occurs_check(Y,Par,ParTypes,Clo)
       -> print(occurs_check(Y,Par)),nl,
          expand_custom_set(closure(Par,ParTypes,Clo),Expansion),
          equal_object_optimized_wf(Y,Expansion,equal_object3,WF)
       ;  Y = closure(Par,ParTypes,Clo)).
equal_object3(closure(Parameters,PT,Cond),Y,WF) :-
    equal_object_custom_explicit_set(closure(Parameters,PT,Cond),Y,WF).
%equal_object3(Obj,Y) :- is_custom_explicit_set(Obj,equal_object3_Obj),
%   equal_object_custom_explicit_set(Obj,Y,WF). % inlined below for performance
equal_object3(global_set(X),Y,WF) :- equal_object_custom_explicit_set(global_set(X),Y,WF).
equal_object3(freetype(X),Y,WF) :- equal_object_custom_explicit_set(freetype(X),Y,WF).
equal_object3(avl_set(X),Y,WF) :- equal_object_custom_explicit_set(avl_set(X),Y,WF).
equal_object3(pred_true /* bool_true */,pred_true /* bool_true */,_WF).
equal_object3(pred_false /* bool_false */,pred_false /* bool_false */,_WF).
equal_object3(term(X),term(X),_WF).
equal_object3(string(X),string(X),_WF).
equal_object3(rec(F1),rec(F2),_WF) :- equal_fields(F1,F2).
equal_object3(freeval(Id,C,F1),freeval(Id,C,F2),WF) :- equal_object_wf(F1,F2,WF).
equal_object3(int(X),int(X),_WF).
equal_object3(fd(X,Type),fd(Y,Type),_WF) :- eq_fd(X,Y).
equal_object3((X,Y),(X2,Y2),WF) :-  equal_object_wf(Y,Y2,WF), equal_object_wf(X,X2,WF).
equal_object3([],X,WF) :- empty_set_wf(X,WF).
equal_object3([H|T],S2,_WF) :- equal_cons(S2,H,T).


equal_object_custom_explicit_set(Obj,Y,WF) :-  % print(eq(Obj,Y)),nl,
    (var(Y) -> Y = Obj
     ; (is_custom_explicit_set_nonvar(Y) -> equal_explicit_sets_wf(Obj,Y,WF)
         ; (Y=[] -> is_empty_explicit_set_wf(Obj,WF)
             ; Y=[H|T] -> equal_custom_explicit_set_cons(Obj,H,T)
             ; add_internal_error('Illegal set: ',equal_object_custom_explicit_set(Obj,Y,WF)),fail
           )
     )).

equal_custom_explicit_set_cons(CS,H,T) :- CS \= avl_set(_),
   var(H),var(T), % TO DO: should we move this treatment below ? to equal_cons_lwf
  % YES, I THINK WE CAN DELETE THIS NOW for avl_sets; but not yet for global_set,...
   %  print_term_summary(equal_custom_explicit_set_cons(CS,H,T)),nl, (debug_mode(on) -> trace ; true),
   unbound_variable(H),
   unbound_variable_for_cons(T),
   !,
   remove_minimum_element_custom_set(CS,Min,NewCS),
   %print_term_summary(remove_min(CS,Min,H,T)),nl,
   (H,T) = (Min,NewCS).
equal_custom_explicit_set_cons(avl_set(AVL),H,T) :- var(H),
     %frozen(H,VH), print_term_summary(check_unbound(H,VH,T)),nl,
     is_unbound_ordered_list_skeleton(H,T),!,  % TO DO: provide this also for global_set(_)
     remove_minimal_elements([H|T],avl_set(AVL),SkeletonToUnify),
     %print_term_summary(remove_min(H,T,SkeletonToUnify)),nl,
     [H|T] =  SkeletonToUnify.

equal_custom_explicit_set_cons(Obj,H,T) :- equal_cons(Obj,H,T).%was expand_custom_set(Obj,Expansion), equal_sets(Expansion,Y)
              % TO DO: handle [int(1),...int(n)] = interval closure

:- block equal_fields(-,-).
equal_fields([],[]).
equal_fields([field(Name,V1)|T1],[field(Name,V2)|T2]) :-
  equal_object(V1,V2,field),
  equal_fields(T1,T2).


% is just like equal_cons, but H and T are guaranteed by the caller to be free
% this just gives one next element of the set; can be used to iterate over sets.
get_next_element(R,H,T) :- var(R),!,R=[H|T].
get_next_element([H1|T1],H,T) :- !,(H1,T1)=(H,T).
get_next_element(R,H,T) :- equal_cons(R,H,T).


equal_cons_wf(R,H,T,WF) :- WF == no_wf_available,!, equal_cons_lwf(R,H,T,2).
equal_cons_wf(R,H,T,WF) :-
   %get_cardinality_wait_flag(R,equal_cons_wf,WF,LWF),
   %get_binary_choice_wait_flag(equal_cons_wf,WF,LWF), %old version
   LWF = lwf_card(R,equal_cons_wf,WF), % will be instantiated by instantiate_lwf
   equal_cons_lwf(R,H,T,LWF).

% a deterministic version; will never instantiate non-deterministically:
% probably better to use equal_cons_wf if possible
%equal_cons_det(R,H,T) :-  equal_cons_lwf(R,H,T,_).

equal_cons(R,H,T) :- %print(eq_cons(R,H,T)),nl,
   equal_cons_lwf(R,H,T,2). %lwf_first(2)).

:- block blocking_equal_cons_lwf(-,?,?,?).
blocking_equal_cons_lwf(E,H,T,LWF) :- equal_cons_lwf(E,H,T,LWF).

equal_cons_lwf(R,H,T,_) :- var(R),!,add_new_el(T,H,R).
equal_cons_lwf([HR|TR],H,T,_) :- ground_value(H), delete_exact_member([HR|TR],H,Rest),!,
   equal_object(Rest,T,equal_cons_lwf_1).
equal_cons_lwf([HR|TR],H,T,LWF) :- !, equal_cons_cons(HR,TR,H,T,LWF).
equal_cons_lwf(avl_set(AVL),H,T,LWF) :- !,
 (is_one_element_custom_set(avl_set(AVL),El)
  -> empty_set(T), % was T=[], but T could be an empty closure !
     equal_object(El,H,equal_cons_lwf_2)
  ;  T==[] -> fail % we have a one element set and AVL is not
  ;  element_can_be_added_or_removed_to_avl(H) ->
           remove_element_from_explicit_set(avl_set(AVL),H,AR), %print(removed(H)),nl,
           equal_object(AR,T,equal_cons_lwf_3)
  ;  nonvar(T),T=[H2|T2],element_can_be_added_or_removed_to_avl(H2) ->  %print(removed_next(H2)),nl,
           remove_element_from_explicit_set(avl_set(AVL),H2,AR),
           equal_object(AR,[H|T2],equal_cons_lwf_4)
           % TO DO: move all such H2 to the front ??
  ;  expand_custom_set(avl_set(AVL),ES),  % length(ES,LenES),print(expanded(LenES,T)),nl,
     % before attempting unification quickly look if lengths are compatible:
     quick_check_length_compatible(ES,[H|T]), % not really sure this is worth it: we have propagate_card in equal_cons_cons below
     equal_cons_lwf(ES,H,T,LWF) ).
equal_cons_lwf(C,H,T,LWF) :-
   is_interval_closure_or_integerset(C,Low,Up),
   (T==[] -> true ; finite_bound(Low), finite_bound(Up)),
   !,  %%print(eq_interval(C,Low,Up)),nl,
   equal_cons_interval(H,T,Low,Up,LWF).
equal_cons_lwf(closure(P,Ty,B),H,T,LWF) :- !,
    expand_custom_set_wf(closure(P,Ty,B),ES,equal_cons_lwf,no_wf_available), % TO DO: pass WF
    blocking_equal_cons_lwf(ES,H,T,LWF).
equal_cons_lwf(freetype(ID),H,T,LWF) :- !, expand_custom_set(freetype(ID),ES),
    blocking_equal_cons_lwf(ES,H,T,LWF).
equal_cons_lwf(global_set(G),H,T,LWF) :- equal_cons_global_set(G,H,T,LWF).

finite_bound(I) :- (var(I) -> true /* inf would be created straightaway */ ; number(I)).

% Purpose: treat some specific closures better; e.g., interval closures and constraint a..b = {1,y,5,x,4} or a..b = {x} & x:100..1002
equal_cons_interval(H,T,Low,Up,_LWF) :- T==[],!, %  Low..Up = {H}  -> Low=H & Up=H
   % unification will fail if Low or Up are not numbers (inf)
   (int(Low),int(Up)) = (H,H).
%equal_cons_interval(_H,_T,Low,Up,_LWF) :- (nonvar(Low),\+ number(Low) ; nonvar(Up),\+ number(Up)),!,
%    enum_warning('OPEN INTERVAL',Low:Up,'cannot expand',trigger_throw(equal_cons_interval)),
%    % we could try and instantiate T to an infinite closure
%    fail.
equal_cons_interval(H,T,Low,Up,LWF) :-  % print(equal_cons_interval(H,T,Low,Up)),nl,
    ((number(Low),number(Up)) -> true % we can expand interval fully
      ; %print(prop1(H,Low,Up)),nl,trace,
        propagate_in_interval([H|T],int(Low),int(Up),0)),
    expand_interval_closure_to_avl(Low,Up,ES),
    blocking_equal_cons_lwf(ES,H,T,LWF).

:- block propagate_in_interval(-,?,?,?).
propagate_in_interval([],Low,Up,Sze) :- %print(interval_size(Sze,Low,Up)),nl,
   (Sze > 0 -> S1 is Sze-1, int_plus(Low,int(S1),Up) ; true). % Test should always be true
propagate_in_interval([H|T],Low,Up,Sze) :-  %print(prop(H,Low,Up,tail(T))),nl,trace,
   in_nat_range(H,Low,Up), % without enumeration
   S1 is Sze+1,
   propagate_in_interval(T,Low,Up,S1).
propagate_in_interval(avl_set(_A),_Low,_Up,_). % TO DO: propagate if Low/Up not instantiated
propagate_in_interval(closure(_,_,_),_,_,_).
propagate_in_interval(global_set(_),_,_,_).

quick_check_length_compatible([],R) :- !,
  (var(R) -> R=[] % can we force R=[] here ??
   ; R \= [_|_]). %(R \= [_|_] -> true ; print(incompatible(R)),fail).
quick_check_length_compatible([_|T],R) :-
  (var(R) -> true
    ; R = [] -> fail %print(incompatible),nl,fail
    ; R = [_|RT] -> quick_check_length_compatible(T,RT)
    ; true).

:- block equal_cons_global_set(-,?,?,?).
equal_cons_global_set(G,H,T,LWF) :- is_infinite_global_set(G),!,
 % for maximal sets we could   complement_set([H],global(G),Res),
    /* should normally fail, unless T is not a list but contains closure or global set */
    test_finite_set(T,Finite), dif(Finite,pred_true),
    when((nonvar(Finite);nonvar(LWF)),equal_cons_global_set_warning(LWF,G,H,T)).
    % used to be : expand_custom_set(global_set(G),ES), equal_cons_lwf(ES,H,T,LWF))).
equal_cons_global_set(G,H,T,LWF) :-
    %(is_infinite_global_set(G) -> test_finite_set(T,Finite), Finite \== pred_true ; true),
    expand_custom_set(global_set(G),ES), equal_cons_lwf(ES,H,T,LWF).


:- block equal_cons_global_set_warning(-,?,?,?).
equal_cons_global_set_warning(_,G,H,T) :-
    add_new_event_in_error_scope(enumeration_warning(enumerating(G),G,'{}',finite,critical),
                                  print_equal_cons_warning(G,H,T)),
    fail. % WITH NEW SEMANTICS OF ENUMERATION WARNING WE SHOULD PROBABLY ALWAYS FAIL HERE !

print_equal_cons_warning(G,H,T,THROWING) :-
        print_throwing(THROWING),
        print('### Enumeration Warning: trying to deconstruct infinite set: '), translate:print_bvalue(global_set(G)),nl,
        print('### Source: '), print(equal_cons_global_set(G,H,T)),nl.

add_new_el(T,H,R) :- var(T),!,R=[H|T].
add_new_el(T,H,R) :- nonvar(T), is_custom_explicit_set_nonvar(T),
   add_element_to_explicit_set(T,H,Res), % will fail for closure/3
   !,
   Res=R.
add_new_el([HT|TT],H,R) :- !,R=[H,HT|TT].
add_new_el([],H,R) :- !, R=[H].
add_new_el(Set,H,R) :- expand_custom_set_to_list(Set,ESet,_,add_new_el),
    add_new_el(ESet,H,R).

delete_exact_member(V,_,_) :- var(V),!,fail.
delete_exact_member([H|T],El,Res) :-
 (H==El -> Res=T
   ; Res=[H|TR], delete_exact_member(T,El,TR)).


%var_list(V) :- var(V),!,unbound_variable(V).
%var_list([]).
%var_list([H|T]) :- unbound_variable(H),var_list(T).

%unbound_variable(V) :- !, unbound_variable_check(V).
unbound_variable(V) :- free_var(V), frozen(V,Residue),unbound_residue(Residue,V).
unbound_residue(external_functions:to_string_aux(GrV,_Val,Str),V) :- !, %GrV checks for groundness of _Val
    V==GrV,unbound_variable(Str).
unbound_residue(external_functions:format_to_string_aux(GrV,_Format,_Val,Str),V) :- !, %GrV checks for groundness of _Val
    V==GrV,unbound_variable(Str).
% TO DO: we need to detect other functions (e.g., B function application,...) which result in values which are not used
unbound_residue(kernel_tools:ground_value_check(V1,V2),V) :- !, V1==V, unbound_variable(V2).
unbound_residue(Residue,_) :- unbound_residue(Residue).

unbound_residue(true).
unbound_residue(kernel_objects:mark_as_to_be_computed(_)).
unbound_residue(kernel_tools:ground_value_check_aux(_,_,V)) :- unbound_variable(V).
unbound_residue(custom_explicit_sets:block_copy_waitflag_store(_,_,_,_,_)). % this stems from checking the domain predicate of function application check_element_of_function_closure
unbound_residue((A,B)) :- unbound_residue(A), unbound_residue(B).
%unbound_residue(kernel_objects:ordered_value(V,_)).  % <-- TO DO: treat this and then assign minimal value !
%unbound_residue(kernel_ordering:ordered_value2(V,_)).
%unbound_residue(U) :- print(unbound(U)),nl,fail.

% check if we have an unbound list_skeleton with optionally just ordering constraints
% check if it is safe to assign H minimal value
% TO DO: also accept if all elements have the same co-routines constraints attached (e.g., because of +-> check)
is_unbound_ordered_list_skeleton(H,T) :-
   is_unbound_ordered_list_skeleton3(H,T,[allow_ordered_values]).
is_unbound_list_skeleton(H,T) :-
   is_unbound_ordered_list_skeleton3(H,T,[]).
is_unbound_ordered_list_skeleton(H,T,Ordered) :-
   is_unbound_ordered_list_skeleton3(H,T,List),
   % if List gets instantiated it will become [allow_ordered_values|_]
   (var(List) -> Ordered=unordered ; Ordered=ordered).

is_unbound_ordered_list_skeleton3(H,T,Options) :- % print(chk_is_unbound_ordered_list_skeleton(H,T)),nl,
   free_var(H),
   (var(T) -> unbound_variable(H),
              unbound_ordered_tail(T,Options) % or ? unbound_variable_for_cons(T)
    ; T = [H2|T2],
      unbound_variable_or_ordered(H,'$$',H2,T,Options),
      is_unbound_ordered_list_skeleton3(H,H2,T2,[H|T],Options)).
is_unbound_ordered_list_skeleton3(Prev,H,T,All,Options) :-
   free_var(H),
   (var(T) -> unbound_variable_or_ordered(H,Prev,'$$',All,Options),
              unbound_ordered_tail(T,Options)
     ; T==[] -> unbound_variable_or_ordered(H,Prev,'$$',All,Options)
     ; T = [H2|T2],
       unbound_variable_or_ordered(H,Prev,H2,All,Options),
       is_unbound_ordered_list_skeleton3(H,H2,T2,All,Options)).

% utility: if is_unbound_ordered_list_skeleton is true, extract for every element in the list one minimal element from CS
remove_minimal_elements(T,CS,Res) :- var(T),!,Res=CS.
remove_minimal_elements([],CS,Res) :- !, empty_set(CS),Res=[].
remove_minimal_elements([_H|T],CS,[Min|Rest]) :-
   remove_minimum_element_custom_set(CS,Min,NewCS),  % _H will be unified in one go with Min later
   remove_minimal_elements(T,NewCS,Rest).

% it is unbound or can be assigned the minimal value of a set
% TO DO: merge with var_list_just_card_constraints
unbound_variable_or_ordered(V,Prev,Nxt,All,Options) :-
    free_var(V), frozen(V,Residue), %print(residue(Residue,Prev,Nxt)),nl,
    unbound_ord_residue_aux(Residue,Prev,V,Nxt,All,Options).% , print(ok),nl.
unbound_ord_residue_aux(true,_Prev,_,_Nxt,_All,_Options).
unbound_ord_residue_aux(kernel_objects:mark_as_to_be_computed(_),_,_,_,_,_).
unbound_ord_residue_aux(kernel_tools:ground_value_check(_,V),_,_,_,_,_) :- unbound_variable(V).
unbound_ord_residue_aux(kernel_tools:ground_value_check_aux(_,_,V),_,_,_,_,_) :- unbound_variable(V).
unbound_ord_residue_aux(kernel_objects:ordered_value(A,B),Prev,V,Nxt,_,Options) :- % there is also a bsets_clp version
   ((A,B)==(Prev,V) ; (A,B)==(V,Nxt)),
   member(allow_ordered_values,Options).
unbound_ord_residue_aux(kernel_objects:not_equal_object_wf(A,B,_),_,V,_,All,_) :-
   (A==V -> exact_member_in_skel(B,All) ; B==V, exact_member_in_skel(A,All)). % all diff constraint; e.g., set up by not_element_of_wf(H,SoFar,WF) in cardinality_as_int2; anyway: all elements in a list must be different
unbound_ord_residue_aux(kernel_objects:not_element_of_wf1(Set,Val,_),_,V,_,All,_) :- Val==V,
   open_tail(All,Tail), Tail==Set. % ditto, again just stating that Values are distinct in the list
unbound_ord_residue_aux((A,B),Prev,V,Nxt,All,Options) :- !,
    unbound_ord_residue_aux(A,Prev,V,Nxt,All,Options),
    unbound_ord_residue_aux(B,Prev,V,Nxt,All,Options).
%unbound_ord_residue_aux(A,Prev,V,Nxt,All) :-
%    print(unbound_ord_residue_aux(A,Prev,V,Nxt,All)),nl,fail.

% get tail of an open list:
open_tail(X,Res) :- var(X),!,Res=X.
open_tail([_|T],Res) :- open_tail(T,Res).
% exact member in a possibly open list:
exact_member_in_skel(X,List) :- nonvar(List), List=[Y|T],
   (X==Y -> true ; exact_member_in_skel(X,T)).


unbound_ordered_tail(T,Options) :- free_var(T), frozen(T,Residue),
   unbound_ordered_tail_aux(Residue,T,Options).
unbound_ordered_tail_aux(true,_,_).
unbound_ordered_tail_aux(kernel_objects:propagate_card(A,B,_Eq),V,_) :-
   (V==A ; V==B). % just specifies A and B have same cardinality
unbound_ordered_tail_aux(prolog:dif(X,Y),V,_) :- (V==X,Y==[] ; V==Y,X==[]).
unbound_ordered_tail_aux(kernel_objects:lazy_ordered_value(W,_),T,Options) :-
    W==T, %% difference with just_cardinality_constraints
    member(allow_ordered_values,Options).
unbound_ordered_tail_aux(bsets_clp:propagate_empty_set(_,_),_,_).
unbound_ordered_tail_aux(kernel_objects:prop_non_empty(_,W,_),T,_) :- W==T.
unbound_ordered_tail_aux(kernel_objects:cardinality_as_int2(W,_,_,_,_,_),T,_) :- W==T.
unbound_ordered_tail_aux(kernel_objects:cardinality3(W,_,_),Var,_) :- W==Var.
unbound_ordered_tail_aux((A,B),T,Options) :-
    (unbound_ordered_tail_aux(A,T,Options) -> true ; unbound_ordered_tail_aux(B,T,Options)).

% co-routine used to mark certain values as to be computed; avoid instantiating them
:- block mark_as_to_be_computed(-).
mark_as_to_be_computed(_).

is_marked_to_be_computed(X) :- var(X),frozen(X,G), %nl,print(check_frozen(X,G)),nl,
      marked_aux(G,X).
marked_aux((A,B),V) :- (marked_aux(A,V) -> true ; marked_aux(B,V)).
marked_aux(kernel_objects:mark_as_to_be_computed(M),V) :- V==M.

:- public unbound_variable_check/1.
% currently not used; but can be useful for debugging
unbound_variable_check(V) :- free_var(V), % check no bool_pred attributes
  ((frozen(V,Goal), Goal\=true)
    -> nl,print('### WARNING: goal attached to unbound variable expression'),nl,print(V:Goal),nl, %trace,
       fail
    ; true).

% check if a variable is unbound or only dif(_,[]) attached; we do not need to check for bool_pred attributes as we have a set
unbound_variable_for_cons(Set) :- var(Set),frozen(Set,F),  % print(f(Set,F)),nl,
   \+ contains_problematic_coroutine_for_cons(F,Set). % for equal cons we can allow more co-routines than when we want to freely determine a value in enumeration

% prolog:dif(X,Y) with Y == [] is ok
contains_problematic_coroutine_for_cons(custom_explicit_sets:element_of_avl_set_wf3(Var,_,_,_,_),V) :- V==Var. % occurs in test 1270
contains_problematic_coroutine_for_cons(kernel_objects:non_free(_),_). % has been marked as non-free
contains_problematic_coroutine_for_cons(kernel_objects:mark_as_to_be_computed(_),_). % has been marked to be computed by closure expansion
contains_problematic_coroutine_for_cons((A,B),Var) :-
   (contains_problematic_coroutine_for_cons(A,Var) -> true
     ; contains_problematic_coroutine_for_cons(B,Var)).


% handling equal_object for [HR|TR] = [H|T]

equal_cons_cons(HR,TR,H,T,_LWF) :- TR==[],!,
     empty_set(T), % was T=[], but T could be an empty closure
     equal_object(HR,H,equal_cons_cons_1).
equal_cons_cons(HR,TR,H,T,_LWF) :- T==[],!,
     empty_set(TR), % was TR=[], but TR could be an empty closure
     equal_object(HR,H,equal_cons_cons_2).
equal_cons_cons(HR,TR,H,T,_LWF) :-
  %(is_unbound_list_skeleton(H,T) -> true ; is_unbound_list_skeleton(HR,TR)),
  (is_unbound_ordered_list_skeleton(H,T,Ordered)
    -> %print(ord(Ordered)),nl,
       (Ordered = unordered -> true
         ; is_unbound_ordered_list_skeleton(HR,TR))
    ;  is_unbound_list_skeleton(HR,TR)),
  % if both are ordered: then the first elements must be equal,
  % if one or both are not ordered: the unification HR=H is only ok if the other is unbound
  % beware of tests 1078 and 1101 when allowing ordered lists
   !,    % print(eq_var_cons(HR,TR,H,T)),nl,
  % HR is variable: no constraints/co-routines attached to it; no other element in TR is constrained either
  %(HR,TR)=(H,T). %fails, e.g., if TR=[] and T= empty closure !
   %print((HR,H,TR,T)),nl,
   % at the moment : unbound_check does not allow ordered set skeletons
   HR=H, equal_object(TR,T).
equal_cons_cons(HR,TR,H,T,LWF) :-
   % here we use LWF for the first time
  %(number(LWF) -> LWF2=LWF ; true),
   % print(equality_objects_lwf(HR,H,EqRes,LWF2)),nl,
   equality_objects_lwf(HR,H,EqRes,LWF2),
   % print(equal_cons2(EqRes,HR,TR,H,T,LWF2)),nl,
   (var(EqRes), definitely_not_in_list(TR,H)
    -> %print(notin(TR,H)),nl,
       EqRes=pred_true  % H cannot appear in TR; it must match HR
    ; true),
   instantiate_lwf(LWF,LWF2), % instantiate later to ensure var(EqRes) can hold if LWF already bound
   equal_cons2(EqRes,HR,TR,H,T,LWF2),
   propagate_card(TR,T,EqRes). % prevents tail recursion; move earlier/remove if EqRes nonvar?
      %,instantiate_lwf(LWF,LWF2) % we could instantiate LWF2 later here to give propagate_card a chance to figure out value of EqRes first ? this slows down examples/B/Alstom/CompilatonProject/Regles/Rule_DB_Route_0001ori.his


% this will instantiate LWF if it has not yet been computed
% (Idea: get_cardinality_wait_flag can be expensive; only do it if we really need the wait_flag)
instantiate_lwf(LWF,R) :- var(LWF),!,R=LWF.
instantiate_lwf(lwf_card(Set,Info,WF),LWF) :- !, get_cardinality_wait_flag(Set,Info,WF,LWF).
    %% get_cardinality_powset_wait_flag(Set,Info,WF,_,LWF).
%instantiate_lwf(lwf_first(X),R) :- !, R=X.
instantiate_lwf(LWF,LWF).

:- block equal_cons2(-,?,?,?,?,?).
equal_cons2(pred_true,_HR,TR,_H,T,_) :- equal_object(TR,T,equal_cons2).
equal_cons2(pred_false,HR,TR, H,T,LWF) :- % print(eq_cons2_neq(HR,TR,H,T)),nl,
    % print(equal_cons_lwf(T,HR,TR2,LWF)),nl,
    equal_cons_lwf(T,HR,TR2,LWF), % look for HR inside T
    % print(res(T,HR,TR2,LWF)),nl,
    T2=TR2,
    equal_cons_lwf(TR,H,T2,LWF). %, was instead of T2=TR2: equal_object(TR2,T2).

:- use_module(kernel_tools,[cannot_match/2]).
% TO DO: investigate whether we should not use kernel_equality or at least a blocking version
definitely_not_in_list(V,_) :- var(V),!,fail.
definitely_not_in_list([],_).
definitely_not_in_list([H|T],X) :- cannot_match(H,X), definitely_not_in_list(T,X).


:- block propagate_card(-,-,-).
propagate_card(X,Y,EqRes) :-
  (nonvar(EqRes) -> true % we no longer need to propagate; equal_cons will traverse
   ; nonvar(X) -> propagate_card2(X,Y,EqRes)
   ; propagate_card2(Y,X,EqRes)).
propagate_card2([],Y,_) :- !,empty_set(Y).
propagate_card2([_|TX],Y,EqRes) :- !,
  (var(Y) -> Y= [_|TY], propagate_card(TX,TY,EqRes)
   ; Y=[] -> fail
   ; Y=[_|TY] -> propagate_card(TX,TY,EqRes)
   ; true
  ). % TO DO: add more propagation
propagate_card2(_,_,_).

%same_card_and_expand(A,B,ExpA,ExpB) :- .... + reorder ??

:- if(environ(prob_safe_mode,true)).
% CODE FOR CHECKING FOR TYPE ERRORS AT RUNTIME
:- assert_must_succeed(type_error([],int(1))).
:- assert_must_succeed(type_error((int(1),int(2)),[pred_true])).
:- assert_must_succeed(type_error(string('Name'),global_set('Name'))).
:- assert_must_fail((type_error([],[_]))).
type_error([],Y) :- no_set_type_error(Y).
type_error([_|_],Y) :- no_set_type_error(Y).
%type_error(X,Y) :- is_custom_explicit_set(X,type_error1), no_set_type_error(Y).
type_error(avl_set(A),Y) :- illegal_avl_set(A) -> true ; no_set_type_error(Y).
type_error(global_set(_),Y) :- no_set_type_error(Y).
type_error(freetype(_),Y) :- no_set_type_error(Y).
type_error(closure(P,_,B),Y) :-
   (var(P) -> true ; var(B) -> true ; P=[] -> true ; P=[P1|_], var(P1) -> true ; no_set_type_error(Y)).
type_error((_,_),Y) :- Y \= (_,_).
type_error(fd(_,T1),Y) :- (Y= fd(_,T2) -> nonvar(T1),nonvar(T2),T1 \=T2 ; true).
type_error(int(_),Y) :- Y\= int(_).
type_error(term(_),Y) :- Y\= term(_).
type_error(rec(_),Y) :- Y \= rec(_).
type_error(freeval(ID,_,_),Y) :- Y \= freeval(ID,_,_).
type_error(string(_),Y) :- Y \= string(_).
% Should raise type error: kernel_objects:union([int(1)],[[]],R).

illegal_value(X) :- var(X),!,fail.
illegal_value(avl_set(A)) :- illegal_avl_set(A).
illegal_value([H|T]) :- illegal_value(H) -> true ; illegal_value(T).
illegal_value(global_set(G)) :- \+ ground(G).
illegal_value(N) :- number(N).
illegal_value((A,B)) :- illegal_value(A) -> true ; illegal_value(B).
% TO DO: complete this

illegal_avl_set(X) :- var(X),!.
illegal_avl_set(empty).
illegal_avl_set(X) :- (X=node(_,_,_,_,_) -> \+ ground(X) ; true).

no_set_type_error(int(_)).
no_set_type_error(fd(_,_)).
no_set_type_error((_,_)).
no_set_type_error(rec(_)).
no_set_type_error(pred_true /* bool_true */).
no_set_type_error(pred_false /* bool_false */).
no_set_type_error(term(_)).
no_set_type_error(string(_)).
no_set_type_error(freeval(_,_,_)).
no_set_type_error(avl_set(A)) :- illegal_avl_set(A).
%% END OF CHECKING CODE
:- endif.


:- assert_must_succeed(not_equal_object(term(a),term(b))).
:- assert_must_succeed(not_equal_object(string('a'),string('b'))).
:- assert_must_succeed(not_equal_object(int(1),int(2))).
:- assert_must_succeed(not_equal_object(rec([field(a,int(1))]),rec([field(a,int(2))]))).
:- assert_must_succeed(not_equal_object(rec([field(a,int(1)),field(b,int(2))]),
              rec([field(a,int(1)),field(b,int(3))]))).
:- assert_must_fail(not_equal_object(rec([field(a,int(1))]),rec([field(a,int(1))]))).
:- assert_must_fail(not_equal_object(rec([field(a,int(1)),field(b,int(2))]),
              rec([field(a,int(1)),field(b,int(2))]))).
:- assert_must_fail(not_equal_object(term(msg),int(2))).
:- assert_must_fail(not_equal_object(fd(1,a),term(msg))).
:- assert_must_succeed(not_equal_object(global_set(a),global_set(b))).
:- assert_must_succeed(not_equal_object([term(a),term(b)],[term(a),term(c)])).
:- assert_must_succeed((not_equal_object([(int(1),[Y])],[(int(X),[Z])]),
                     Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[int(2)])).
:- assert_must_succeed(not_equal_object((int(1),int(2)),(int(3),int(4)))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_equal_object((int(1),int(2)),(int(1),int(4))))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_equal_object((int(1),int(4)),(int(3),int(4))))).
:- assert_must_fail(not_equal_object((int(1),int(4)),(int(1),int(4)))).
:- assert_must_succeed(not_equal_object((int(1),string('a')),(int(1),string('b')))).
:- assert_must_fail(not_equal_object((int(1),string('b')),(int(1),string('b')))).
:- assert_must_fail(not_equal_object([(term(a),[])],[(term(a),[])])).
:- assert_must_fail((not_equal_object([(int(1),[Y])],[(int(X),[Z])]),
                     Y=(term(a),Y2), X=1, Z=(term(a),[]), Y2=[])).
:- assert_must_fail(not_equal_object([int(1),int(2)],[int(2),int(1)])).
:- assert_must_succeed(not_equal_object(term(msg),term(another_msg))).
:- assert_must_succeed(not_equal_object([int(1),int(2)],[int(0),int(4)])).
:- assert_must_fail((sample_closure(C),
    not_equal_object(C,[int(1),int(2)]))).
:- assert_must_succeed((sample_closure(C),
    not_equal_object(C,[int(1),int(0)]))).
:- assert_must_succeed((sample_closure(C),
    not_equal_object(C,global_set('NAT')))).
:- assert_must_fail((not_equal_object(
 [[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
    [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(3,'Name'),fd(2,'Name')]]
,[[],[fd(1,'Name')],[fd(1,'Name'),fd(2,'Name')],
     [fd(1,'Name'),fd(2,'Name'),fd(3,'Name')],[fd(2,'Name')],[fd(2,'Name'),fd(3,'Name')]])
     )).
:- assert_must_fail((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,a,int(2))))).
:- assert_must_succeed((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,a,int(3))))).
:- assert_must_succeed((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,b,int(2))))).
:- assert_must_succeed((not_equal_object(freeval(selfcx,a,int(2)),freeval(selfcx,a,int(3))))).

:- assert_must_succeed((not_equal_object(pred_true /* bool_true */,X), X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([],X),X=[_|_])).
%:- assert_must_succeed((not_equal_object([],X), nonvar(X),X=[_|_])).
:- assert_must_succeed((not_equal_object(X,[]), X=[_|_])).
:- assert_must_succeed((not_equal_object(X,pred_false /* bool_false */), X==pred_true /* bool_true */)).

:- assert_must_succeed(not_equal_object([_X],[int(1),int(3)])). % Inefficiency example of setlog
:- assert_must_succeed_any(not_equal_object([_X],[int(1)])). % Inefficiency example of setlog
:- assert_must_succeed((not_equal_object([X],[pred_true /* bool_true */]),X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([pred_true /* bool_true */],[X]),X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([[X]],[[pred_true /* bool_true */]]),X==pred_false /* bool_false */)).
:- assert_must_succeed((not_equal_object([[pred_true /* bool_true */]],[[X]]),X==pred_false /* bool_false */)).
:- assert_must_succeed((custom_explicit_sets:construct_one_element_custom_set(pred_true /* bool_true */, A), kernel_objects:not_equal_object(A,[X]), X==pred_false /* bool_false */)).
:- assert_must_succeed((custom_explicit_sets:construct_one_element_custom_set(pred_true /* bool_true */,A), kernel_objects:not_equal_object([X],A), X==pred_false /* bool_false */)).
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([],[int(3333)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([],[int(2),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([int(3)],[int(2),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([int(3),int(1),int(4)],[int(2),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],not_equal_object([int(2),int(1),int(3),int(5)],[int(2),int(1),int(3)]))).
% X in 3..4, kernel_objects:not_equal_object([int(2),int(3)],[int(2),int(X)]), X==4. in clpfd Mode



:- block not_equal_object_wf(-,-,?).
 /* TO DO: implement a better _wf version ;  use bool_dif if possible */
not_equal_object_wf(X,Y,WF) :- X\==Y, (var(X) -> not_equal_object_wf1(Y,X,WF)
                                               ; not_equal_object_wf1(X,Y,WF)).

not_equal_object_wf1([],R,WF) :- !, not_empty_set_wf(R,WF).
not_equal_object_wf1(R,E,WF) :- E==[],!, not_empty_set_wf(R,WF).
not_equal_object_wf1(X,Y,_) :- not_equal_object2(X,Y).

not_equal_object(X,Y) :-
 (nonvar(X) -> not_equal_object2(X,Y)
  ; nonvar(Y) -> not_equal_object2(Y,X)
  ;   X\==Y, when((?=(X,Y);nonvar(X);nonvar(Y)), not_equal_object0(X,Y))).

not_equal_object0(X,Y) :- X\==Y,(var(X) -> not_equal_object2(Y,X)
                                         ; not_equal_object2(X,Y)).

%not_equal_object2(X,Y) :- print(not_equal_object2(X,Y)),nl,fail.
not_equal_object2(pred_true /* bool_true */,R) :- !, R=pred_false /* bool_false */.
not_equal_object2(pred_false /* bool_false */,R) :- !, R=pred_true /* bool_true */.
not_equal_object2(fd(X,Type),R) :- !, R=fd(Y,Type), neq_fd(X,Y,Type).
not_equal_object2(int(X),R) :- !, R=int(Y), integer_dif(X,Y).
not_equal_object2(string(X),R) :- !, R=string(Y), dif(X,Y).
not_equal_object2(term(X),R) :- !, R=term(Y), dif(X,Y).
not_equal_object2(rec(F1),R) :- !, R=rec(F2),
    not_equal_fields(F1,F2).
%not_equal_object2([],R) :- var(R),!, print(not_empty(R)),nl,R=[_|_].  % Dangerous for equal_object: we generate a non-enumerated variable !
not_equal_object2((X1,X2),R) :- !, R=(Y1,Y2),
    not_equal_couple(X1,Y1,X2,Y2).
not_equal_object2(X,Y) :- is_custom_explicit_set(X,not_equal_object2),!,
  %  print(not_equal_object2(X,Y)),nl, % uncovered case; for efficiency better explicitly written above
    not_equal_explicit_set(X,Y).
not_equal_object2(X,Y) :- not_equal_object3(X,Y).


:- block not_equal_explicit_set(?,-).
not_equal_explicit_set(X,Y) :-
    is_custom_explicit_set_nonvar(Y),!,
    % print(not_equal_explicit_sets(X,Y)),nl,
    not_equal_explicit_sets(X,Y).
not_equal_explicit_set(X,[]) :- !, %print(check_non_empty(X)),nl,
   is_non_empty_explicit_set(X).
not_equal_explicit_set(X,Y) :-  % print_term_summary(expanding(X)),
    expand_custom_set(X,EX), not_equal_object3_block(EX,Y).

:- block not_equal_object3_block(-,?).
not_equal_object3_block(EX,Y) :- not_equal_object3(EX,Y).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- block not_equal_object3(?,-).
:- if(environ(prob_safe_mode,true)).
not_equal_object3(X,Y) :- nonvar(X),type_error(X,Y),
    add_internal_error('Internal Typing Error (please report as bug !) : ',not_equal_object(X,Y)),
    fail.
:- endif.
not_equal_object3(freeval(ID,Case1,Value1),freeval(ID,Case2,Value2)) :-
    when(?=(Case1,Case2), % we first have to be able to decide the case; if cases are different types of values may be different
	     not_equal_freeval(Case1,Value1,Case2,Value2)).
not_equal_object3([],X) :- not_empty_set(X).
not_equal_object3([H|T],Set2) :-
  (Set2=[] -> true
  ; cardinality_peano_wf([H|T],N1,no_wf_available),
    cardinality_peano_wf(Set2,N2,no_wf_available),
     when(?=(N1,N2),  % when we trigger code below, = can be decided:
        (N1=N2 -> neq_cons(Set2,H,T) ; true))).
%       (dif(N1,N2) ;  (N1=N2, neq_cons(Set2,H,T)))).  %not_equal_object_sets(Set1,Set2) )) ).

not_equal_freeval(Case1,Value1,Case2,Value2) :-
   (Case1=Case2 -> not_equal_object(Value1,Value2) ; true).

:- block not_equal_object_sets(-,?), not_equal_object_sets(?,-).
not_equal_object_sets([H|T],Set2) :- !,
    ( Set2=[H2|_T2]
     -> not_equal_object_sets2(H,T,H2,Set2)
     ;  Set2=[] -> true
     ;  not_equal_object2(Set2,[H|T]) % avl_set probably
    ).
not_equal_object_sets(Set1,Set2) :- % Note : if Set1 =[] then we can fail, as both sets have same length
   % we could have empty set or avl_set can sometimes creep into end of lists
   not_equal_object2(Set1,Set2).

:- block not_equal_object_sets2(-,?,?,?), not_equal_object_sets2(?,?,-,?).
not_equal_object_sets2(H,_T,_H2,Set2) :-
  % TO DO: should we not use kernel_equality:membership_test_wf here ??
    not_element_of(H,Set2).
not_equal_object_sets2(H,T,_H2,Set2) :-
    remove_element(H,Set2,Del2), %print_message(rem(H,Set2,Del2)),
    not_equal_object(T,Del2).


:- block neq_cons(-,?,?).
%neq_cons(X,Y,Z) :- print(neq_cons(X,Y,Z)),nl,fail.
neq_cons([],_,_) :- !.
neq_cons([H2|T2],H1,T1) :- !,
   (T2==[],T1==[]
     -> not_equal_object(H1,H2)
      ; check_and_remove([H2|T2],H1,NewSet2,RemoveSuccesful),
        % print(removed(H1,H2,T2,NewSet2,RemoveSuccesful)),
        neq_cons2(RemoveSuccesful,T1,NewSet2)
   ).
neq_cons(avl_set(A),H1,T1) :- element_can_be_added_or_removed_to_avl(H1),!,
    %print(removing(H1)),nl,
    (remove_element_from_explicit_set(avl_set(A),H1,RA)
      ->  not_equal_object(T1,RA)
      ;  true ).
neq_cons(ES,H1,T1) :- is_custom_explicit_set(ES,neq_cons),expand_custom_set(ES,ExpSet),
    neq_cons(ExpSet,H1,T1).

:- block neq_cons2(-,?,?).
neq_cons2(not_successful,_T1,_NewSet2). % one element could not be removed: the sets are different
neq_cons2(successful,T1,NewSet2) :-  not_equal_object_sets(T1,NewSet2).

% kernel_objects:not_equal_couple(int(1),int(Y),B,pred_true).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(1),int(Y),B,pred_true),Y=1, B==pred_false)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),B,pred_true),Y=1, B==pred_false)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),B,pred_false),Y=1, B==pred_true)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),pred_false,B),Y=1, B==pred_true)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(int(Y),int(1),B,pred_true),Y=2, var(B))).
:- assert_must_succeed(( kernel_objects:not_equal_couple(B,pred_true,int(Y),int(1)),Y=1, B==pred_false)).
:- assert_must_succeed(( kernel_objects:not_equal_couple(B,fd(C,'Code'),fd(Y,'Name'),F),F=fd(1,'Name'),Y=1,B=fd(1,'Code'),C=2 )).
:- assert_must_succeed(( kernel_objects:not_equal_couple(B,pred_true,fd(Y,'Name'),F),F=fd(1,'Name'),Y=1, B==pred_false)).
:- block not_equal_couple(-,?,-,?),not_equal_couple(?,-,?,-).
% (X1,X2) /= (Y1,Y2)

% using CLPFD results in less propagation it seems
% e.g. post_constraint((A1 #\= A2 #\/ B1 #\= B2), dif((A1,B1),(A2,B2))) will not propagate if A1=A2 or B1=B2
% we could do something like
%   post_constraint((N*A1 + B1 #\= N*A2 + B2), dif((A1,B1),(A2,B2))). ; but we need to know good value for N
% TO DO: pass typing information when available ?? or not needed because type info extracted ?


not_equal_couple(X1,Y1,X2,Y2) :-
  equality_objects(X1,Y1,EqRes1),
  (var(EqRes1)
   -> equality_objects(X2,Y2,EqRes2),
      % print(call_not_equal_couple4(EqRes1,X1,Y1,EqRes2,X2,Y2)),nl,
      not_equal_couple4(EqRes1,X1,Y1,EqRes2,X2,Y2)
    ; EqRes1=pred_true -> not_equal_object(X2,Y2)
    ; true).

:- block not_equal_couple4(-,?,?,-,?,?).
not_equal_couple4(EqRes1,X1,Y1,EqRes2,X2,Y2) :-
   %print(unblock_not_equal_couple4(EqRes1,X1,Y1,EqRes2,X2,Y2)),nl,
   (var(EqRes1)
    -> not_equal_couple5(EqRes2,X1,Y1,EqRes1)
     ; not_equal_couple5(EqRes1,X2,Y2,EqRes2)).

not_equal_couple5(pred_true,_X2,_Y2,EqResOther) :- EqResOther=pred_false.
not_equal_couple5(pred_false,_,_,_).


 /* To do: provide special support for things like
    couple of fd's [done], list of fd's, set of fd's */

:- block not_equal_fields(-,-).
not_equal_fields([field(ID1,V1)|T1],[field(ID2,V2)|T2]) :-
 % should we wait for ID1 or ID2 to become nonvar?
   check_field_name_compatibility(ID1,ID2),
   (T1==[]
    -> T2=[],not_equal_object(V1,V2)
     ; not_equal_couple(V1,V2,rec(T1),rec(T2))  % would be slightly more efficient to have a custom version of not_equal_couple
   ).

check_field_name_compatibility(Name1,Name2) :-
   nonvar(Name1), nonvar(Name2), Name1 \= Name2, !,
   add_internal_error('Incompatible fields: ',check_field_name_compatibility(Name1,Name2)).
check_field_name_compatibility(_,_).


/* ------------------------------------------- */
/*      equality_objects/3  function     */
/* ------------------------------------------- */

%% :- ensure_loaded(kernel_equality).

% ----------------------------------------------------------
% ----------------------------------------------------------



:- use_module(kernel_equality).

% ----------------------------------------------------------
% ----------------------------------------------------------

/* ---------------> */
/* This should probably be more systematically applied before every kernel call
 + expanded for other symbolic representations !! */



/* underlying assumption: if G is a global set: we get back the
 global_set tag immediately: no need to use when to wait;
  better: ensure that b_compute_expression always returns a nonvar term */

integer_global_set('NAT').
integer_global_set('NATURAL').
integer_global_set('NAT1').
integer_global_set('NATURAL1').
integer_global_set('INT').
integer_global_set('INTEGER').

string_global_set('STRING').


:- assert_must_succeed(( kernel_objects:element_of_global_set(int(0),'NATURAL'))).
:- assert_must_fail(( kernel_objects:element_of_global_set(int(0),'NATURAL1'))).
:- assert_must_fail(( kernel_objects:element_of_global_set(int(-1),'NATURAL'))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(int(-1),'INTEGER'))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(int(0),'NAT'))).
:- assert_must_fail(( kernel_objects:element_of_global_set(int(0),'NAT1'))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(X,'NAT'),X=int(1))).
:- assert_must_succeed(( kernel_objects:element_of_global_set(X,'NATURAL'),X=int(1))).

element_of_global_set(X,GS) :-
    init_wait_flags(WF),element_of_global_set_wf(X,GS,WF),ground_wait_flags(WF).

:- block element_of_global_set_wf(?,-,?).
element_of_global_set_wf(El,Set,WF) :- b_global_set(Set),!,
      global_type_wf(El,Set,WF).
element_of_global_set_wf(X,'STRING',_WF) :- !, X=string(_).
element_of_global_set_wf(int(X),GS,WF) :-
      element_of_global_integer_set_wf(GS,X,WF).

/* what about BOOL ?? */
element_of_global_integer_set_wf('NAT',X,WF) :-
   preferences:get_preference(maxint,MAXINT),
   in_nat_range_wf(int(X),int(0),int(MAXINT),WF).
element_of_global_integer_set_wf('NATURAL',X,WF) :-
   (ground(X) -> X>=0
      ;  is_natural(int(X),WF),
         %get_last_wait_flag(element_of_global_set(int(X),'NATURAL'),WF,LWF),
         get_integer_enumeration_wait_flag(X,'NATURAL',WF,LWF),
         enumerate_natural(X,0,LWF)
   ).
element_of_global_integer_set_wf('NAT1',X,WF) :-
   preferences:get_preference(maxint,MAXINT),
   in_nat_range_wf(int(X),int(1),int(MAXINT),WF).
element_of_global_integer_set_wf('NATURAL1',X,WF) :-
   (ground(X) -> X>=1
      ;  is_natural1(int(X),WF),
         %get_last_wait_flag(element_of_global_set_wf(int(X),'NATURAL1'),WF,LWF),
         get_integer_enumeration_wait_flag(X,'NATURAL1',WF,LWF),
         enumerate_natural(X,1,LWF)
    ).
element_of_global_integer_set_wf('INT',X,WF) :-
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   in_nat_range_wf(int(X),int(MININT),int(MAXINT),WF).
element_of_global_integer_set_wf('INTEGER',X,WF) :-
    (ground(X) -> true
     ;  %get_last_wait_flag(element_of_global_set_wf(int(X),'INTEGER'),WF,LWF),
        get_integer_enumeration_wait_flag(X,'INTEGER',WF,LWF),
        enumerate_int_wf(X,LWF,element_of_global_integer_set_wf) %when((nonvar(LWF);nonvar(X)),(ground(X)->true;enumerate_int(X)))
    ).


get_integer_enumeration_wait_flag(X,SET,WF,LWF) :-
   clpfd_domain(X,FDLow,FDUp), finite_domain(FDLow,FDUp),!,
   Size is 1+FDUp-FDLow,
   get_wait_flag(Size,element_of_global_set_wf(int(X),SET),WF,LWF).
get_integer_enumeration_wait_flag(X,SET,WF,LWF) :-
%   get_integer_enumeration_wait_flag(element_of_global_set_wf(int(X),SET),WF,LWF).
   get_enumeration_starting_wait_flag(element_of_global_set_wf(int(X),SET),WF,LWF).

:- assert_must_succeed((kernel_objects:enumerate_int_wf(X,4,self_check),X==2)).
:- block enumerate_int_wf(-,-,?).
enumerate_int_wf(X,_LWF,Source) :- %print(enum_int(X)),nl,
   (ground(X) -> true ; enumerate_int(X,trigger_true(Source))).

:- assert_must_succeed(not_element_of_global_set(int(-1),'NAT')).
:- assert_must_succeed(not_element_of_global_set(int(-1),'NATURAL')).
:- assert_must_succeed(not_element_of_global_set(int(0),'NAT1')).
:- assert_must_succeed(not_element_of_global_set(int(0),'NATURAL1')).
not_element_of_global_set(int(X),GS) :-
   (var(GS) -> add_error(kernel_objects,var_not_element_of_global_set,(int(X),GS)) ; true),
   not_element_of_global_set2(GS,X).
not_element_of_global_set2('NAT',X) :-
   preferences:get_preference(maxint,MAXINT),
   clpfd_not_in_non_empty_range(X,0,MAXINT). %when(nonvar(X), (X<0 ; X>MAXINT)).
not_element_of_global_set2('NATURAL',X) :- is_not_natural(int(X)).
not_element_of_global_set2('NAT1',X) :-
   preferences:get_preference(maxint,MAXINT),
   clpfd_not_in_non_empty_range(X,1,MAXINT). %when(nonvar(X),(X<1 ; X>MAXINT)).
not_element_of_global_set2('NATURAL1',X) :- is_not_natural1(int(X)).
not_element_of_global_set2('INT',X) :-
   preferences:get_preference(minint,MININT),
   preferences:get_preference(maxint,MAXINT),
   clpfd_not_in_non_empty_range(X,MININT,MAXINT). %when(nonvar(X), (X < MININT ; X > MAXINT)).
%not_element_of_global_set(string(_X),'STRING') :- fail.
%not_element_of_global_set(int(_X),'INTEGER') :- fail.
%not_element_of_global_set(_El,Set) :- b_global_set(Set), fail.



/* ---- */
/* SETS */
/* ---- */

%:- block is_a_set(-).
%is_a_set(X) :- is_a_set2(X).
%is_a_set2([]) :- !.
%is_a_set2([_|_]) :- !.
%is_a_set2(X) :- is_custom_explicit_set(X,is_a_set2).




:- assert_must_succeed(exhaustive_kernel_fail_check(empty_set([int(4),int(3)]))).
:- assert_must_fail((empty_set([int(2),int(1)]))).
:- assert_must_fail((empty_set([int(1)]))).
:- assert_must_fail((empty_set([[]]))).
:- assert_must_fail((empty_set(global_set('Name')))).
:- assert_must_fail((empty_set(X),X=[int(1)])).
:- assert_must_succeed((empty_set([]))).
empty_set(X) :- (var(X) -> X=[]
                  ; X=[] -> true
                 % ; X=[_|_] -> fail
                  ; is_custom_explicit_set_nonvar(X),is_empty_explicit_set(X)).
empty_set_wf(X,WF) :- (var(X) -> X=[]
                  ; X=[] -> true
                 % ; X=[_|_] -> fail
                  ; is_custom_explicit_set_nonvar(X),is_empty_explicit_set_wf(X,WF)).


:- assert_must_succeed(exhaustive_kernel_check(not_empty_set([int(4),int(3)]))).
:- assert_must_succeed((kernel_objects:not_empty_set([int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_empty_set([int(1)]))).
:- assert_must_succeed((kernel_objects:not_empty_set([[]]))).
:- assert_must_succeed((kernel_objects:not_empty_set(global_set('Name')))).
:- assert_must_succeed((kernel_objects:not_empty_set_lwf(X,1),nonvar(X),X=[_|_])).
:- assert_must_succeed((kernel_objects:not_empty_set_lwf([int(1)],_))).
:- assert_must_fail((kernel_objects:not_empty_set([]))).

not_empty_set_wf(S,WF) :- WF==no_wf_available,!, not_empty_set(S).
not_empty_set_wf(S,WF) :- var(S), !,
  (preferences:preference(use_smt_mode,true) -> S=[_|_]
   % ; WF=no_wf_available -> not_empty_set(S)
   ; get_large_finite_wait_flag(not_empty_set_wf,WF,LWF),
     not_empty_set_lwf(S,LWF)).
not_empty_set_wf(closure(P,T,B),WF) :- !, is_non_empty_explicit_set_wf(closure(P,T,B),WF).
not_empty_set_wf(S,_WF) :- not_empty_set(S).

:- block not_empty_set_lwf(-,-).
% the instantiation with a list skeleton can easily cause multiple solutions for the same
% set to be found: hence we guard it by a wait flag
not_empty_set_lwf(S,_LWF) :- var(S),!, %print(setting_not_empty(S)),nl,
                             S=[_|_].
not_empty_set_lwf(S,_) :- not_empty_set(S).

:- use_module(error_manager,[add_warning/2]).
:- block not_empty_set(-).
%not_empty_set(S) :- var(S),!, print(setting_not_empty(S)),nl,S=[_|_].
  % not_empty_set(X) :- not_equal_object([],X).
not_empty_set([_|_]).
not_empty_set(avl_set(A)) :- (A==empty -> add_warning(not_empty_set,'Empty avl_set'),fail ; true).
not_empty_set(closure(P,T,B)) :- is_non_empty_explicit_set(closure(P,T,B)). % TO DO: also use WF
not_empty_set(global_set(Type)) :- b_non_empty_global_set(Type).
not_empty_set(freetype(ID)) :- kernel_freetypes:is_non_empty_freetype(ID).

% there also exists: eq_empty_set  , a reified version, i.e., test_empty_set


:- assert_must_succeed((exact_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed((exact_element_of(int(1),[int(2),int(3),int(4),int(1)]))).
:- assert_must_succeed((exact_element_of(int(4),[int(2),int(3),int(4),int(1)]))).
:- assert_must_succeed((exact_element_of(int(1),[int(2),int(3)|T]), T=[int(4),int(1)])).
:- assert_must_fail((exact_element_of(int(5),[int(2),int(3)|T]), T=[int(4),int(1)])).
:- assert_must_succeed((exact_element_of(fd(1,'Name'),global_set('Name')))).
:- assert_must_succeed((exact_element_of([int(2),int(1)],[[],[int(2),int(1)]]))).
:- assert_must_fail((exact_element_of([int(1),int(2)],[[],[int(2),int(1)]]))).
%:- assert_must_succeed((exact_element_of([(int(1),fd(2,'Name'))],
%      closure([zzzz],[set(couple(integer,global('Name')))], 'In'('ListExpression'(['Identifier'(zzzz)]),
%                        'Seq'(value([fd(1,'Name'),fd(2,'Name')]))))) )).
%:- assert_must_succeed((exact_element_of(XX,
%     closure([zzzz],[set(couple(integer,global('Name')))], 'In'('ListExpression'(['Identifier'(zzzz)]),
%                        'Seq'(value([fd(1,'Name'),fd(2,'Name')]))))),
%                        equal_object(XX,[(int(1),fd(1,'Name'))]) )).
%:- assert_must_succeed((
%exact_element_of(XX,closure([zzzz],[set(couple(integer,global('Name')))],
% 'In'('ListExpression'(['Identifier'(zzzz)]),
%   'Perm'(value([fd(1,'Name'),fd(2,'Name')]))))),
%   equal_object(XX,[(int(1),fd(2,'Name')),(int(2),fd(1,'Name'))]) )).

%:- assert_must_succeed(( exact_element_of(X,
% closure([zzzz],[set(record([field(balance,integer),field(name,global('Code'))]))],
% 'In'('ListExpression'(['Identifier'(zzzz)]),
% 'PowerSet'(value(closure([zzzz],
% [record([field(balance,integer),field(name,global('Code'))])],'In'('ListExpression'(['Identifier'(zzzz)]),
% 'SetOfRecords'(value(cons_expr(field(balance,global_set('NAT')),
% cons_expr(field(name,global_set('Code')),nil_expr))))))))))),
%  X=[rec([field(balance,int(0)),field(name,fd(2,'Code'))])] )).
%:- assert_must_fail(( exact_element_of(X,
%   closure([zzzz],[set(record([field(balance,integer),field(name,global('Code'))]))],
%  'In'('ListExpression'(['Identifier'(zzzz)]),
%  'PowerSet'(value(closure([zzzz],
%  [record([field(balance,integer),field(name,global('Code'))])],'In'('ListExpression'(['Identifier'(zzzz)]),
%  'SetOfRecords'(value(cons_expr(field(balance,global_set('NAT')),
%  cons_expr(field(name,global_set('Code')),nil_expr))))))))))),
%  X=[rec([field(balance,int(-1)),field(name,fd(2,'Code'))])] )).


/* use this to compute elements */
exact_element_of(X,Set) :- %print_message(exact_element_of(X,Set)),
     dif(Set,[]),
     exact_element_of2(Set,X).
:- block exact_element_of2(-,?).
exact_element_of2([H|_],H).
exact_element_of2([_|T],E) :- exact_element_of3(T,E).
exact_element_of2(X,E) :- is_custom_explicit_set_nonvar(X), check_element_of(E,X).
:- block exact_element_of3(-,?).
exact_element_of3([H|_],H).
exact_element_of3([_|T],E) :- exact_element_of3(T,E).


:- assert_must_succeed(exhaustive_kernel_check(check_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(check_element_of(int(3),[int(2),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(check_element_of(int(1),[]))).

/* uses equal_object instead of unification */
:- assert_must_succeed((check_element_of(X,
      [(int(1),(int(1),(int(1),int(1)))),(int(2),(int(1),(int(1),int(1)))),
       (int(1),(int(1),(int(1),int(2)))),(int(2),(int(1),(int(1),int(2))))]),
       equal_object(X, (int(2),(int(1),(int(1),int(2))))) )).
:- assert_must_succeed((check_element_of(X,
      [ (((int(1),int(1)),int(1)),int(1)), (((int(1),int(1)),int(1)),int(2)),
        (((int(1),int(1)),int(1)),int(3)), (((int(1),int(1)),int(1)),int(4)),
        (((int(1),int(1)),int(2)),int(1)), (((int(1),int(1)),int(2)),int(2))
        ]), equal_object(X,  (((int(1),int(1)),int(2)),int(1)))
      )).
:- assert_must_succeed((check_element_of(fd(1,'Name'),global_set('Name')))).
%:- assert_must_succeed_multiple(check_element_of(X,[[fd(1,'Name')],[]])).
:- assert_must_succeed((check_element_of((int(1),int(2)),[(int(1),int(2))]))).
:- assert_must_succeed((check_element_of((_X,_Y),[(fd(2,'Code'),fd(2,'Code'))]))).
:- assert_must_succeed((init_wait_flags(WF),
    check_element_of_wf((X,Y),[(fd(2,'Code'),fd(2,'Code'))],WF),
    ground_det_wait_flag(WF), X= fd(2,'Code'), Y= fd(2,'Code'),
   kernel_waitflags:ground_wait_flags(WF) )).
   :- assert_must_succeed((init_wait_flags(WF),
          check_element_of_wf((Y,X),[(fd(2,'Code'),fd(2,'Code'))],WF),
          ground_det_wait_flag(WF), X= fd(2,'Code'), Y= fd(2,'Code'),
   kernel_waitflags:ground_wait_flags(WF) )).
:- assert_must_succeed((check_element_of([int(1),int(2)],[[int(2),int(1)]]))).

:- assert_must_succeed((check_element_of([int(1),int(2)],[[],[int(2),int(1)]]))).
:- assert_must_succeed((check_element_of(X,[[],[int(2),int(1)]]), X==[] )).
:- assert_must_succeed((check_element_of_wf(X,[[],[int(2),int(1)]],_WF),
    equal_object(X,[int(1),int(2)]) )).
:- assert_must_succeed((check_element_of_wf(XX,global_set('Name'),WF),kernel_waitflags:ground_wait_flags(WF), XX==fd(3,'Name') )).
:- assert_must_fail(check_element_of([fd(2,'Name')],[[fd(1,'Name')],[]])).
:- assert_must_fail((check_element_of([int(2)],[[],[int(2),int(1)]]))).
:- assert_must_succeed((check_element_of(int(1),_X))).
:- assert_must_succeed((check_element_of((int(2),_X),[(int(1),[(int(1),int(22))]),(int(2),[(int(1),int(55))])]))).

check_element_of(X,Set) :- init_wait_flags(WF),
  check_element_of_wf(X,Set,WF),
  ground_wait_flags(WF).

% new test: check_element_of(int(1),X).
% new test: check_element_of(int(1),[int(2)|X]).

check_element_of_wf(X,Set,WF) :- %%print_message(check_element_of_wf(X,Set)),
   dif(Set,[]),
   check_element_of1(X,Set,WF).

%check_element_of1(X,Set,WF) :- var(X),var(Set),unbound_variable_check(Set),!,
%   Set=[_|_], check_element_of2(Set,X,WF).
%:- block check_element_of1(-,-,?). %%
:- if(environ(prob_data_validation_mode,true)).
% avoid instantiating Sets to lists early on: can disturb enumeration and efficient computation of large sets
check_element_of1(X,Set,WF) :- nonvar(Set),!, check_element_of2(Set,X,WF).
check_element_of1(X,Set,WF) :- get_wait_flag(3000,check_element_of1(X,Set),WF,WF2),
   check_element_of1b(X,Set,WF,WF2).

:- block check_element_of1b(?,-,?,-).
check_element_of1b(X,Set,WF,_) :-   % print_term_summary(check_element_of1(X,Set,WF)),
  (unbound_variable_for_cons(Set)
    ->  mark_as_non_free(X),
        Set=[X|_] % Note: X needs to be nonvar so that other code knows X is not free anymore
        % TO DO: normalise X ?
        % TO DO: do this using CHR/attributes rather than by instantiation
    ;  check_element_of2(Set,X,WF)
  ).
:- else.
check_element_of1(X,Set,WF) :-    % print_term_summary(check_element_of1(X,Set,WF)), trace,
  (unbound_variable_for_cons(Set)
    ->  mark_as_non_free(X),
        Set=[X|_] % Note: X needs to be nonvar so that other code knows X is not free anymore
        % TO DO: normalise X ?
        % TO DO: do this using CHR/attributes rather than by instantiation
    ;  check_element_of2(Set,X,WF)
  ).
:- endif.


% attach co-routine to mark a given term as not a real variable
mark_as_non_free(X) :- var(X) -> non_free(X) ; true.
:- block non_free(-).
non_free([H|T]) :- !, mark_as_non_free(H), mark_as_non_free(T).
non_free((A,B)) :- !, mark_as_non_free(A), mark_as_non_free(B).
non_free(rec(Fields)) :- !, mark_as_non_free_fields(Fields).
non_free(_).
:- block mark_as_non_free_fields(-).
mark_as_non_free_fields([]).
mark_as_non_free_fields([field(_,Val)|T]) :- mark_as_non_free(Val),mark_as_non_free_fields(T).

:- use_module(clpfd_lists,[lazy_fd_value_check/4]).

:- block check_element_of2(-,?,?).
check_element_of2(CS,El,WF) :-
              is_custom_explicit_set_nonvar(CS),!, element_of_custom_set_wf(El,CS,WF).
check_element_of2([],_,_) :- !,fail.
%check_element_of2([H|T],El,WF) :- try_expand_and_convert_to_avl([H|T],AVL),AVL=avl_set(_),!, % much better support exists for AVL trees; should we enable this conversion ?? %nl,print(converted_list_to_AVL([H|T])),nl,nl,
%    element_of_custom_set_wf(El,AVL,WF).
check_element_of2([H|T],E,WF) :- !,  %  print(check_element_of4w(E,H,T,WF)),nl,
   % try and transform E : Set  into  clpfd:element(_,FDVals,EFD)  check:
		  clpfd_lists:lazy_fd_value_check([H|T],E,WF,FullyChecked),
		  %get_partial_set_priority([H|T],WF,LWF), %%
		  %get_wait_flag(2,check_element_of2([H|T],E),WF,LWF), %%
		   ((FullyChecked==true,ground(E)) -> true % no need to check
			 ;  get_cardinality_wait_flag([H|T],check_element_of2,WF,LWF),
				check_element_of4w(E,H,T,WF,LWF) % this call is somewhat redundant if FullyChecked=true; but otherwise in_fd_value_list will not enumerate on its own (e.g., self-checks for relation_over will fail)
			).
check_element_of2(freetype(Id),E,WF) :- !, is_a_freetype_wf(E,Id,WF).
check_element_of2(Set,E,WF) :-
   add_internal_error('Illegal argument: ',check_element_of2(Set,E,WF)),fail.


% call if you already have an explicit waitflag (LWF) setup for the cardinality of the set
:- block check_element_of_wf_lwf(?,-,?,?).
check_element_of_wf_lwf(El,CS,WF,_LWF) :-
        is_custom_explicit_set_nonvar(CS),!, element_of_custom_set_wf(El,CS,WF).
check_element_of_wf_lwf(E,[H|T],WF,LWF) :- check_element_of4w(E,H,T,WF,LWF).
check_element_of_wf_lwf(E,freetype(Id),WF,_) :- !, is_a_freetype_wf(E,Id,WF).

:- block check_element_of4w(-,?,-,?,-).
% check_element_of4w(E,H,T,_WF,_LWF) :- print(check_element_of4w(E,H,T,_WF,_LWF)),nl,fail.
check_element_of4w(E,H,T,_WF,_LWF) :- T==[],!,equal_object(E,H,check_element_of4w).
check_element_of4w(E,H,_T,_WF,_LWF) :- E==H ,!. %,print(eq(E,H)),nl.  % added by mal, 17.10 2007
check_element_of4w(E,H,T,WF,LWF) :- T\==[],
   equality_objects_lwf(E,H,Res,LWF),
   check_element_of4(Res,E,T,WF,LWF).

:- block check_element_of4(-,?,?,?,-).
check_element_of4(pred_true,_E,_,_WF,_LWF).
check_element_of4(pred_false,E,T,WF,LWF) :-
   (var(T) -> T = [E|_] ; check_element_of5(E,T,WF,LWF)).

:- block check_element_of5(?,-,?,?).
check_element_of5(E,R,WF,LWF) :-  %print(equal_cons(R,H,T,for(E))),nl,
  get_next_element(R,H,T),
  check_element_of4w(E,H,T,WF,LWF).



:- assert_must_succeed(exhaustive_kernel_check(not_element_of(int(3),[int(2),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_check(not_element_of(int(3),[int(2),int(1),int(4)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_element_of(int(3),[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_element_of(fd(1,'Name'),[]))).
:- assert_must_fail((kernel_objects:not_element_of(fd(1,'Name'),global_set('Name')))).
:- assert_must_succeed((kernel_objects:not_element_of(X,[fd(1,'Name')]),X = fd(2,'Name'))).
:- assert_must_fail((kernel_objects:not_element_of(X,[fd(1,'Name')]),X = fd(1,'Name'))).
:- assert_must_succeed(kernel_objects:not_element_of(term(a),[])).
:- assert_must_fail((kernel_objects:not_element_of(int(1),[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:not_element_of([int(1),int(2)],
         [[int(1)],[int(0),int(4)],[int(0),int(3)],[int(0),int(1)],[int(0)],[]]))).
:- assert_must_fail((kernel_objects:not_element_of(term(3),[int(2),int(1)]))).


not_element_of(X,Set) :- init_wait_flags(WF),
  not_element_of_wf(X,Set,WF),
  ground_wait_flags(WF).

:- use_module(b_global_sets,[b_fd_type/3]).
:- block not_element_of_wf(-,-,?).
not_element_of_wf(_,Set,_) :- Set==[],!.
not_element_of_wf(El,Set,WF) :- nonvar(El),El=fd(X,GS),b_get_fd_type_bounds(GS,N,N),!,
   % we have a global set with a single element; Set must be empty
   % print(not_el_of(El,Set)),nl,
   X=N,empty_set_wf(Set,WF).
not_element_of_wf(El,Set,WF) :- not_element_of_wf1(Set,El,WF).

:- block not_element_of_wf1(-,?,?).
%not_element_of_wf1(E,S,WF) :- print(not_el_of(E,S,WF)),nl,fail.
not_element_of_wf1(X,E,WF) :- is_custom_explicit_set_nonvar(X),!,
   %print(check_not_el(E,X,WF)),nl,
   not_element_of_custom_set_wf(E,X,WF). % , print(ok(E,X,WF)),nl.
not_element_of_wf1([],_E,_WF).
not_element_of_wf1([H|T],E,WF) :- /* print(call(not_element_of_wf1([H|T],E))),nl, */
   not_equal_object_wf(E,H,WF),
   not_element_of_wf1(T,E,WF).


:- assert_must_succeed(exhaustive_kernel_check(add_element(int(3),[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(add_element(int(2),[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(add_element(int(4),[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed((kernel_objects:add_element(int(3),[int(2),int(1)],R),
                        kernel_objects:equal_object(R,[int(1),int(2),int(3)]))).
:- assert_must_succeed((kernel_objects:add_element([int(2)],[[int(2),int(1)],[]],R),
                        kernel_objects:equal_object(R,[[],[int(1),int(2)],[int(2)]]))).
:- assert_must_succeed((kernel_objects:add_element([int(1),int(2)],[[int(2),int(1)],[]],R),
                        kernel_objects:equal_object(R,[[],[int(1),int(2)]]))).
:- assert_must_succeed((kernel_objects:add_element(X,[int(2),int(1)],R),
                        kernel_objects:equal_object(R,[int(1),int(2)]), X = int(1))).
:- assert_must_succeed((kernel_objects:add_element([int(1),int(2)],
         [[int(1)],[int(0),int(4)],[int(0),int(3)],[int(0),int(1)],[int(0)],[]], _R))).

:- assert_must_succeed((kernel_objects:add_element(int(3),[int(X),int(1)],R,D),
        var(D), X=3, R==[int(3),int(1)], D==done)).

:- assert_must_fail((kernel_objects:add_element(term(msg),[int(2),int(1)],_R))).
:- assert_must_succeed((kernel_objects:add_element(int(3),[int(2),int(X)],R),
   nonvar(R), R =[H|T], H==int(2), nonvar(T),T=[_HH|TT],var(TT),
   X=4, T==[int(4),int(3)])).
:- assert_must_succeed((kernel_objects:add_element(int(3),[int(2),int(X)],R),
   nonvar(R), R =[H|T], H==int(2), nonvar(T),T=[_HH|TT],var(TT),
   X=3, T==[int(3)])).
:- assert_must_succeed((kernel_objects:add_element(int(3),X,[int(2),int(3)]),
   kernel_objects:equal_object(X,[int(2)]) )).
:- assert_must_succeed((kernel_objects:add_element(int(3),X,[int(3)]),
   kernel_objects:equal_object(X,[]) )).
:- assert_must_succeed((add_element(X,[int(1)],[int(1)]),X==int(1))).
:- assert_must_succeed((add_element(X,[],[int(1)]),X==int(1))).
% kernel_objects:add_element(E,[H],R,Done), H = int(X), E=int(Y), X in 1..10, Y in 11..20.


add_element(E,Set,NewSet) :- add_element(E,Set,NewSet,_).
add_element(Element,Set,NewSet,Done) :- add_element_wf(Element,Set,NewSet,Done,no_wf_available).
add_element_wf(E,Set,NewSet,WF) :- add_element_wf(E,Set,NewSet,_,WF).

:- block add_element_wf(?,-,?,?,?).
add_element_wf(Element,Set,NewSet,Done,_WF) :- Set==[],!,
   % try and convert to AVL if possible:
    equal_object_optimized(NewSet,[Element]), % we could call equal_object_opt3 directly
    Done=done.
add_element_wf(E,Set,NewSet,Done,WF) :- add_element1_wf(E,Set,NewSet,Done,WF).

:- block %add_element1(-,?,-,?),
         add_element1_wf(?,-,?,?,?).
add_element1_wf(E,Set,NewSet,Done,WF) :- var(E),!, add_element_var(Set,NewSet,E,Done,WF).
add_element1_wf(E,[H|T],NewSet,Done,WF) :- E==H,!, % avoid running [H|T] through expand_custom_set_to_list, in case T is a variable this will create a pending co-routine
     equal_object_wf(NewSet,[H|T],add_element1_1,WF),Done=done.
add_element1_wf(E,Set,NewSet,Done,WF) :- %print(add_el(E,Set,NewSet)),nl,
     nonvar(Set), is_custom_explicit_set_nonvar(Set),
     add_element_to_explicit_set(Set,E,R),!,
     % print(add_to_explicit_set(E)),nl,
      equal_object_wf(R,NewSet,add_element1_2,WF),Done=done.
add_element1_wf(E,Set,NewSet,Done,WF) :-
    expand_custom_set_to_list(Set,ESet,_,add_element1), % we could avoid this expansion by treating avl_set,... below in add_element3
    add_element2_wf(ESet,E,NewSet,Done,WF).


add_element_var([],Res,Element,Done,WF) :- !, % print(add_elvarempty(Element,[],Res)),nl,
   equal_cons_wf(Res,Element,[],WF),Done=done.
add_element_var(Set,Res,Element,Done,WF) :- Set \= [], Set \= closure(_,_,_),
  is_one_element_set(Res,ResEl), !,
  % the result is a one element set; hence Element *must* be the element in that set
   %% print(add_elvar(Element,Set,Res,ResEl)),nl,
   equal_object_wf(Element,ResEl,add_element_var_1,WF),
   equal_object_wf(Set,Res,add_element_var_2,WF), Done=done.
add_element_var(Set,Res,Element,Done,WF) :- %when(nonvar(Element), add_element(Element,Set,Res,Done)).
   expand_custom_set_to_list(Set,ESet,_,add_element_var),
   add_element2_wf(ESet,Element,Res,Done,WF).

is_one_element_set(S,_) :- var(S),!,fail.
is_one_element_set([H|T],H) :- T==[].
is_one_element_set(avl_set(S),El) :- is_one_element_custom_set(avl_set(S),El).

:- block add_element2_wf(-,?,?,?,?).
% add_element2_wf(E,S,R,_) :- print_message(add_el2(E,S,R)),fail.
add_element2_wf([],E,Res,Done,_WF) :- var(Res),should_be_converted_to_avl(E),
         construct_avl_from_lists([E],R),!,
         (R,Done)=(Res,done).
add_element2_wf(S,E,Res,Done,WF) :- copy_list_skeleton(S,Res,WF),
         add_element3_wf(S,E,Res,Done,WF).

% TO DO: use something else, like subset to propagate info that Set1 <: Set1 \/ {New}
:- block copy_list_skeleton(-,?,?).
copy_list_skeleton([],_,_WF) :- !.
copy_list_skeleton([H|T],R,WF) :- !, % H must be in R, but not all elements of R are in [H|T] !; it could be the added element
    ((ground(H) ; unbound_variable_for_cons(R) ;
       custom_explicit_sets:singleton_set(R,_) % if R is a singleton set {EL} then H must be EL and T=[]
       ) -> equal_cons_wf(R,H,RR,WF), copy_list_skeleton(T,RR,WF)
      ; %nl,print(not_copying([H|T],R)),nl,
        true % otherwise equal_cons_wf can backpropagate elements from R into H !! see {x,y| x = {1,2} &  x \/ y  = {1,2,3} & 1:y } test 1535
      ).
copy_list_skeleton(Set,R,WF) :- !,is_custom_explicit_set(Set,copy_list_skeleton),
   expand_custom_set_to_list(Set,ESet,_,copy_list_skeleton), copy_list_skeleton(ESet,R,WF).
copy_list_skeleton(Skel,R,WF) :- add_internal_error('Argument not a set: ',copy_list_skeleton(Skel,R,WF)).

:- block add_element3_wf(-,?,?,?,?).
add_element3_wf([],E,Res,Done,WF) :- % Res must be {E}
   equal_cons_wf(Res,E,[],WF),
   Done=done.
add_element3_wf([H|T],E,Res,Done,WF) :-
   equality_objects_wf(H,E,EqRes,WF),
   equal_cons_wf(Res,H,TailRes,WF), % was: equal_object([H|TailRes],Res),  % use WF?
   (var(EqRes)
    -> has_not_to_be_added([H|T],Res,EqRes,0)
    ; true),
   %(when(nonvar(EqRes),(print(nv(EqRes,H,T,WF)),nl))),
   add_element4_wf(EqRes,T,E,TailRes,Done,WF).


% check if an element has not to be added to arg1 to obtain arg2
:- block has_not_to_be_added(?,-,?,?),has_not_to_be_added(-,?,?,?).
%has_not_to_be_added(A,B,R,Sz) :- print(has_not_to_be_added(A,B,R,Sz)),nl,fail.
has_not_to_be_added([],[],R,Sz) :- !,(Sz=1 -> R=pred_true % we have 1 element: force equality with first element
                                      ; true).
has_not_to_be_added([],[_H|T],R,_Sz) :- !, %(var(R) -> print(add_f([],[_H|T],R,_Sz)),nl ; true),
   empty_set(T),R=pred_false. % R=pred_false means with add an element
has_not_to_be_added([_|_],[],_,_) :- !,fail. % we can either add or not; in both cases we do not obtain []
has_not_to_be_added([_|T1],[_|T2],R,Sz) :- !, S1 is Sz+1, has_not_to_be_added(T1,T2,R,S1).
has_not_to_be_added(_,_,_,_). % to do: support custom explicit sets

:- block add_element4_wf(-,?,?,?,?,?).
add_element4_wf(pred_true, T,_E,TRes,Done,WF) :- equal_object_wf(T,TRes,add_element4_wf,WF), Done=done.
add_element4_wf(pred_false,T, E,TRes,Done,WF) :- add_element3_wf(T,E,TRes,Done,WF).


:- assert_must_succeed((kernel_objects:add_new_element(int(3),[int(2),int(1)],R),
                        kernel_objects:equal_object(R,[int(1),int(2),int(3)]))).
:- assert_must_succeed((kernel_objects:add_new_element([int(2)],[[int(2),int(1)],[]],R),
                        kernel_objects:equal_object(R,[[],[int(1),int(2)],[int(2)]]))).

% TO DO : get rid of need for non-WF version in enumeration basic type:
add_new_element(E,Set,NewSet) :- init_wait_flags(WF),
   add_new_element_wf(E,Set,NewSet,WF), ground_wait_flags(WF).

% use when you are sure the element to add is not in the set
% to be used for adding elements to an accumulator
:- block add_new_element_wf(?,-,?,?).
%%add_new_element(E,Set,NewSet) :- add_element(E,Set,NewSet). % TO DO : Improve
add_new_element_wf(E,Set,NewSet,WF) :-
     is_custom_explicit_set(Set,add_element), add_element_to_explicit_set(Set,E,R),!,
                             %% print(add_new_to_explicit_set(E)),nl, %%
                             equal_object_wf(R,NewSet,add_new_element_wf,WF).
add_new_element_wf(E,Set,NewSet,WF) :- expand_custom_set_to_list(Set,ESet,_,add_new_element_wf),
         add_new_element2(ESet,E,NewSet,WF).

:- block add_new_element2(-,?,?,?).
add_new_element2([],E,Res,WF) :- var(Res),should_be_converted_to_avl(E),
         construct_avl_from_lists([E],R),!,equal_object_wf(R,Res,add_new_element2,WF).
add_new_element2(S,E,Res,WF) :- equal_cons_wf(Res,E,S,WF).

%:- assert_must_succeed(exhaustive_kernel_check(remove_element(int(3),[int(2),int(1),int(3)],[int(1),int(2)]))).
:- assert_must_succeed((kernel_objects:remove_element(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')]),
                        kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:remove_element(X,global_set('Name'),[fd(2,'Name'),fd(3,'Name')]),
                        X = fd(1,'Name'))).
:- assert_must_succeed((kernel_objects:remove_element(int(1),X,[int(2)]),
                        kernel_objects:equal_object(X,[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:remove_element([int(1),int(2)],X,[]),
                        kernel_objects:equal_object([[int(2),int(1)]],X))).
:- assert_must_fail((kernel_objects:remove_element(int(3),X,_),
                        (X = [int(2),int(1)] ; X=[], X = [int(2)]))).
:- assert_must_fail((kernel_objects:remove_element(int(1),X,Res),
                        X = [int(2),int(1)], (Res=[] ; Res = [_,_|_] ; Res = [int(1)]))).

                        /* remove element is currenlty only used in not_equal_sets */

% remove element X from Set, yielding Res
remove_element(X,Set,Res) :- equal_cons(Set,X,Res).



:- assert_must_succeed(exhaustive_kernel_check(remove_element_wf(int(3),[int(3),int(1)],
 [int(1)],_WF))).
:- assert_must_succeed(exhaustive_kernel_check(remove_element_wf(int(1),[int(3),int(1)],
 [int(3)],_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(remove_element_wf(int(1),[int(3),int(1)],
 [int(1)],_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(remove_element_wf(int(11),[int(1)],
 [int(1)],_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(remove_element_wf(int(1),[int(3),int(1)],
 [],_WF))).
:- assert_must_succeed((kernel_objects:remove_element_wf(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')],_WF),
                        kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:remove_element_wf(int(1),X,[int(2)],_WF),
                        kernel_objects:equal_object(X,[int(2),int(1)]))).
:- assert_must_succeed((     kernel_objects:remove_element_wf(int(1),[int(X),int(2)],R,WF), kernel_waitflags:ground_wait_flags(WF),X==1,R==[int(2)] )).
:- assert_must_succeed((     kernel_objects:remove_element_wf(X,[int(1),int(2)],R,WF), kernel_waitflags:ground_wait_flags(WF),X==int(2),R==[int(1)] )).
:- assert_must_succeed((     kernel_objects:remove_element_wf(X,[pred_true /* bool_true */,pred_false /* bool_false */],R,WF), kernel_waitflags:ground_wait_flags(WF),X==pred_false /* bool_false */,R==[pred_true /* bool_true */] )).

remove_element_wf(X,Set,Res,WF) :- remove_element_wf(X,Set,Res,WF,_DONE).

:- block remove_element_wf(?,-, -,?,?).
remove_element_wf(X,Set,Res,WF,_DONE) :- Res==[],!, % we know that X must be the only element in Set
  equal_object_wf(Set,[X],remove_element_wf,WF).
remove_element_wf(X,Set,Res,WF,DONE) :-
  remove_element_wf1(X,Set,Res,WF,DONE).

:- block remove_element_wf1(?,-, ?,?,?).
remove_element_wf1(X,avl_set(A),Res,WF,DONE) :- element_can_be_added_or_removed_to_avl(X),!,
  /* TO DO: try and move the check about whether X can be added to later; when either X is known
     or LWF is instantiated */
    remove_element_from_explicit_set(avl_set(A),X,AR),
    equal_object_wf(AR,Res,remove_element_wf1,WF), DONE=done.
remove_element_wf1(X,Set,Res,WF,DONE) :- /* DONE is ground when element actually removed */
  expand_custom_set_to_list(Set,ESet,_,remove_element_wf1),
  %% nl,print(remove_element_wf1(X,Set,ESet,Res,WF,DONE)),nl,nl, %%
  remove_element_wf2(X,ESet,Res,LWF,DONE),
  %when(nonvar(DONE), print_bt_message(removed(X,ESet,Res,LWF))),
  (DONE==done -> true
   ;  same_card_prop(ESet,[X|Res]), % in case result is instantiated: check compatible with inputs
      get_cardinality_wait_flag(ESet,remove_element_wf1(X,ESet,Res),WF,LWF),
      quick_propagation_element_information(Set,X,WF,_) % use Set rather than ESet; better if still closure or AVL
   ).

:- block same_card_prop(-,?), same_card_prop(?,-).
same_card_prop([],[_|_]) :- !, fail.
same_card_prop([_|T],R) :- !,
   (R=[] ->  fail
    ; R=[_|RT] -> same_card_prop(T,RT)
    ; true). % just ignore
same_card_prop(_,_).

:- block remove_element_wf2(?,-,?,?,?).
remove_element_wf2(H1,[H2|T],Res,LWF,DONE) :-  Res==[],!,
  equal_object(H1,H2,remove_element_wf2),
  remove_element_wf3(pred_true,H1,H2,T,Res,LWF,DONE).
remove_element_wf2(H1,[H2|T],Res,LWF,DONE) :-
  prop_empty_set(T,EqRes),
  equality_objects_lwf(H1,H2,EqRes,LWF),
 %% print(rem(H1,H2,EqRes,T,Res)),nl,trace, %%
  %%((var(EqRes),var(LWF)) -> print(block_remove_element_wf3(EqRes,H1,H2,T,Res)),nl ; true),
  remove_element_wf3(EqRes,H1,H2,T,Res,LWF,DONE).
   /* important for total_bijection that this has higher priority than other expansions */

:- block prop_empty_set(-,?).
% force second argument to pred_true if first arg is empty set
prop_empty_set([],R) :- !,  R=pred_true.
prop_empty_set(_,_).

:- block remove_element_wf3(-,?,?,?,?,-,?).
% remove_element_wf3(EqRes,H1,H2,T,Res,LWF,DONE) :- print(remove_element_wf3(EqRes,H1,H2,T,Res,LWF,DONE)),nl,fail.
remove_element_wf3(pred_true,_H1,_H2,T,Res,_LWF,DONE) :-
     equal_object(T,Res,remove_element_wf3_1),DONE=done.
remove_element_wf3(pred_false,E,H,T,Res,LWF,DONE) :-
     equal_object([H|RT],Res,remove_element_wf3_2),
     remove_element_wf2(E,T,RT,LWF,DONE).

 /* the same as above: but do not remove if infinite or closure */

:- block remove_element_wf_if_not_infinite_or_closure(?,-,?,?,?,?).
remove_element_wf_if_not_infinite_or_closure(X,Set, Res,WF,LWF,Done) :-
        (dont_expand(Set)
         ->  check_element_of_wf(X,Set,WF),
             equal_object_wf(Res,Set,remove_element_wf_if_not_infinite_or_closure,WF),
             Done=true % or should we wait until X known ?
             %(var(Res)->Res=Set ; equal_object(Res,Set))
          ;  expand_custom_set_to_list(Set,ESet,_,remove_element_wf_if_not_infinite_or_closure),
             remove_element_wf2(X,ESet,Res,LWF,Done)
        ).

%:- use_module(bmachine_construction,[external_procedure_used/1]).
%dont_expand(global_set('STRING')) :- !. % s: STRING +-> ...  will generate new strings !
    %(external_procedure_used(_) -> true). % we could check if there is a STRING generating procedure involved
    % unless we use external functions, there is *no* way that new strings can be generated from a B machine !
    % Hence: we can expand STRING safely and thus avoid infinite enumeration of partial functions, ...
    % example:  procs : STRING +-> {"waiting"} &  card( dom(procs) ) = 6  thus fails quickly
dont_expand(avl_set(_)) :- !,fail.
dont_expand(Set) :- is_non_expanded_closure(Set).
dont_expand(Set) :- is_infinite_or_very_large_explicit_set(Set). % should we use a smaller bound than 20000 ? see test 1609




:- assert_must_succeed((kernel_objects:remove_exact_first_element([int(1),int(2)],X,[[]]),
                        X = [[int(1),int(2)],[]])).
:- assert_must_succeed((kernel_objects:remove_exact_first_element(X,global_set('Name'),T),
		         X==fd(1,'Name'),T==[fd(2,'Name'),fd(3,'Name')])).
:- assert_must_fail((kernel_objects:remove_exact_first_element([[]],X,_),
                        X = [[int(1),int(2)],[]])).

:- assert_must_succeed((kernel_objects:remove_exact_first_element(X,C,R),
      kernel_objects:gen_test_interval_closure(1,2,C),
                        X == int(1), R == [int(2)] )).

gen_test_interval_closure(From,To,CL) :-
   CL=closure(['_zzzz_unary'],[integer],b(member( b(identifier('_zzzz_unary'),integer,[]),
 b(interval(b(value(int(From)),integer,[]),
            b(value(int(To)),integer,[])),set(integer),[])),pred,[])).

:- block remove_exact_first_element(?,-,?).
remove_exact_first_element(X,Set,Res) :- remove_exact_first_element1(Set,X,Res).

remove_exact_first_element1([],_,_) :- fail.
remove_exact_first_element1([H|T],H,T).
remove_exact_first_element1(avl_set(A),H,T) :- remove_minimum_element_custom_set(avl_set(A),H,T).
remove_exact_first_element1(global_set(GS),H,T) :-
   remove_minimum_element_custom_set(global_set(GS),H,T).
remove_exact_first_element1(freetype(GS),H,T) :-
   remove_minimum_element_custom_set(freetype(GS),H,T).
remove_exact_first_element1(closure(P,Types,B),H,T) :-
   remove_minimum_element_custom_set(closure(P,Types,B),H,T).


:- assert_must_succeed((kernel_objects:delete_element_wf(fd(1,'Name'),X,[fd(2,'Name'),fd(3,'Name')],_WF),
                        X = global_set('Name'))).
:- assert_must_succeed((kernel_objects:delete_element_wf(int(1),X,[int(2)],_WF),
                        X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:delete_element_wf([int(1),int(2)],X,[],_WF),
                        X = [[int(2),int(1)]])).
:- assert_must_succeed((kernel_objects:delete_element_wf(int(3),X,[int(2),int(1)],_WF),
                        X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:delete_element_wf(int(1),X,X,_WF),
                        X = [])).
:- assert_must_fail((kernel_objects:delete_element_wf(int(X),[int(1)],[int(1)],_WF),
                        X = 1)).

                        /* WARNING: only use when R is not instantiated by something else;
                           (except for R=[]) */


:- block delete_element_wf(?,-,?,?).
delete_element_wf(X,Set,Res,WF) :-
  ground(X),
  try_expand_and_convert_to_avl_with_check(Set,ESet,delete_element_wf),!,
  delete_element0(X,ESet,Res,WF).
delete_element_wf(X,Set,Res,WF) :- delete_element1(X,Set,Res,WF).

:- block delete_element0(?,-,?,?).
delete_element0(X,ESet,Res,WF) :-
  ( is_custom_explicit_set(ESet,delete_element),
    delete_element_from_explicit_set(ESet,X,DS)
     -> equal_object_wf(DS,Res,delete_element0,WF)
     ;  delete_element1(X,ESet,Res,WF)
  ).

delete_element1(X,Set,Res,WF) :- expand_custom_set_to_list(Set,ESet,_,delete_element1),
  %check_is_expanded_set(ESet,delete_element2),
  delete_element2(ESet,X,Res,WF).

:- block delete_element2(-,?,?,?).
delete_element2([],_,[],_). /* same as above, but allow element to be absent */
delete_element2([H2|T],E,R,WF) :-
    equality_objects_wf(H2,E,EqRes,WF),
    delete_element3(EqRes,H2,T,E,R,WF).
    %when_sufficiently_instantiated(E,H2,delete_element3(H1,[H2|T],R)). /* added by Michael Leuschel, 16/3/06 */

:- block delete_element3(-,?,?,?,?,?).
delete_element3(pred_true,_H2,T,_,R,WF) :- equal_object_wf(R,T,delete_element3,WF).
delete_element3(pred_false,H2,T,E,Res,WF) :- equal_cons_wf(Res,H2,RT,WF),delete_element2(T,E,RT,WF).




:- assert_must_succeed(kernel_objects:check_is_expanded_set([int(1)],test)).

:- public check_is_expanded_set/2.
check_is_expanded_set(X,Source) :-
  (nonvar(X),(X=[] ; X= [_|_]) -> true
    ; add_internal_error('Is not expanded set: ',check_is_expanded_set(X,Source))
  ).


/* union/3 */

:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(3)],[int(2),int(1),int(3)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(1)],[int(1),int(2)],[int(1),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(3)],[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],union([int(3),int(2)],[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],union([int(3),int(4)],[int(2),int(1)],[int(1),int(3),int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(1)],[int(2)],Res),kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union([],[int(2)],Res),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(2)],[],Res),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(2)],[int(2)],Res),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([int(1)],Res,[int(1),int(2)]),
                        kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:union([fd(1,'Name')],X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,X))).
:- assert_must_succeed((kernel_objects:union(X,global_set('Name'),Res),X=[fd(2,'Name'),fd(1,'Name')],
                         kernel_objects:equal_object(Res,global_set('Name')))).
:- assert_must_succeed((kernel_objects:union([fd(1,'Name')],[fd(3,'Name'),fd(2,'Name')],Res),
                         kernel_objects:equal_object(Res,global_set('Name')))).
%:- assert_must_succeed((kernel_objects:union([fd(1,'Name')],[fd(3,'Name'),fd(2,'Name')],Res),
%                         kernel_objects:equal_object(Res,X),X=global_set('Name'))).
:- assert_must_fail((kernel_objects:union([int(1)],[int(2)],Res),
            (kernel_objects:equal_object(Res,[_]);kernel_objects:equal_object(Res,[_,_,_|_])))).
:- assert_must_fail((kernel_objects:union([int(1)],[int(1)],Res),(Res=[];kernel_objects:equal_object(Res,[_,_|_])))).
:- assert_must_fail((kernel_objects:union([fd(1,'Name')],[fd(2,'Name')],Res),
                         kernel_objects:equal_object(Res,global_set('Name')))).
% kernel_objects:union([int(1),int(2)],X,[int(1),int(2),int(3)])

union(S1,S2,Res) :- init_wait_flags(WF), union_wf(S1,S2,Res,WF), ground_wait_flags(WF).

:- block union_wf(-,-,-,?).
%union_wf(Set1,Set2,Res,_WF) :- print(union_wf(Set1,Set2,Res)),nl,fail.
union_wf(Set1,Set2,Res,WF) :- Set1==[],!,equal_object_wf(Set2,Res,union_wf_1,WF).
union_wf(Set1,Set2,Res,WF) :- Set2==[],!,equal_object_wf(Set1,Res,union_wf_2,WF).
union_wf(Set1,Set2,Res,WF) :- Res==[],!,empty_set_wf(Set1,WF), empty_set_wf(Set2,WF).
union_wf(Set1,Set2,Res,WF) :- union0(Set1,Set2,Res,WF).

:- block union0(-,-,?,?), union0(-,?,-,?), union0(?,-,-,?). % require two arguments to be known
%union0(Set1,Set2,Res,_WF) :- print(union0(Set1,Set2,Res)),nl,fail.
union0(Set1,Set2,Res,WF) :- Set1==[],!,equal_object_wf(Set2,Res,union0_1,WF).
union0(Set1,Set2,Res,WF) :- Set2==[],!,equal_object_wf(Set1,Res,union0_2,WF).
union0(Set1,Set2,Res,WF) :- Res==[],!,empty_set_wf(Set1,WF), empty_set_wf(Set2,WF).
union0(Set1,Set2,Res,WF) :- nonvar(Res), singleton_set(Res,X),!,
   %print(union0_to_singleton_set(Set2,Set1,X,WF)),nl,
   (var(Set1) -> union0_to_singleton_set(Set2,Set1,X,WF) ; union0_to_singleton_set(Set1,Set2,X,WF)).
union0(Set1,Set2,Res,WF) :- (var(Set1) -> union1(Set2,Set1,Res,WF) ; union1(Set1,Set2,Res,WF)).

% optimized version for Set1 \/ Set2 = {X}
% TO DO: is not triggered when Set1 and Set2 are instantiated first (before result)
%        >>> z:11..12 & {x,y} \/ {v}   = {z}  does not work
union0_to_singleton_set([],Set2,X,WF) :- !, equal_object_wf(Set2,[X],union0_3,WF). % cannot be reached, due to checks above
union0_to_singleton_set([H|T],Set2,X,WF) :- !, empty_set_wf(T,WF), equal_object_wf(H,X,WF),
   check_subset_of_wf(Set2,[X],WF).
union0_to_singleton_set(avl_set(A),Set2,X,WF) :- !, singleton_set(avl_set(A),AEl),
   equal_object_wf(AEl,X,WF),
   check_subset_of_wf(Set2,[X],WF).
union0_to_singleton_set(Set1,Set2,X,WF) :-  % closure or global_set; revert to normal treatment
    union1(Set1,Set2,[X],WF).

union1(Set1,Set2,Res,WF) :-
   %print_term_summary(union1(Set1,Set2,Res,WF)),nl,
   try_expand_and_convert_to_avl_unless_large_or_closure(Set1,ESet1),
   try_expand_and_convert_to_avl_unless_large_or_closure(Set2,ESet2),
   union1e(ESet1,ESet2,Res,WF). %, print(union1e(ESet1,ESet2,Res,WF)),nl.

try_expand_and_convert_to_avl_unless_large_or_closure(Set,ESet) :- (var(Set);Set=closure(_,_,_)),!,ESet=Set.
try_expand_and_convert_to_avl_unless_large_or_closure(Set,ESet) :-
    try_expand_and_convert_to_avl_unless_large(Set,ESet).

union1e(Set1,Set2,Res,WF) :-    % print_term_summary(union1e(Set1,Set2,Res)),
    is_custom_explicit_set(Set1,union1e),  %print(try_union(Set1,Set2)),nl,
    union_of_explicit_set(Set1,Set2,Union), !,
    %print_term_summary(explicit_set_union(Union)),
    equal_object_wf(Union,Res,union1e,WF).
union1e(Set2,Set1,Res,WF) :-  % Set2=avl_set(_),  nonvar(Set1), Set1 \= avl_set(_),
     nonvar(Set1), Set1=avl_set(_), Set2 \= avl_set(_), \+ ground(Set2),
     !, % avoid expanding Set2
     % print_term_summary(union_invert_arguments(Set2,Set1,Res)),
     expand_custom_set_to_list(Set1,ESet1,_,union1e_1),
     union2(ESet1,Set2,Res,WF), lazy_check_subset_of(Set2,Res,WF).
union1e(Set1,Set2,Res,WF) :-
     expand_custom_set_to_list(Set1,ESet1,_,union1e_2), % we could avoid this expansion by treating avl_set,... below in union2
     union2(ESet1,Set2,Res,WF),
     lazy_check_subset_of(Set1,Res,WF), % ADDED to solve {x,y| { x \/ y  } <: {{1} \/ {2}}}
     lazy_check_subset_of(Set2,Res,WF) % could perform additional constraint checking
   %%,nl,  print(union2_result(ESet1,Set2,Res)),nl,nl
    % ,try_prop_card_leq(ESet1,Res), try_prop_card_leq(Set2,Res).   %%% seems to slow down ProB: investigate
    .

:- block lazy_try_check_element_of(?,-,?).
lazy_try_check_element_of(_H,V,_) :- %print(lazy(_H,V)),nl,
     var(V),!.
lazy_try_check_element_of(H,Set,WF) :- lazy_check_element_of_aux(Set,H,WF).

lazy_check_element_of_aux(closure(P,T,B),H,WF) :- !, check_element_of_wf(H,closure(P,T,B),WF).
lazy_check_element_of_aux(avl_set(A),H,WF) :- !, check_element_of_wf(H,avl_set(A),WF).
lazy_check_element_of_aux([X|T],H,WF) :- !, lazy_check_element_of_list(T,X,H,WF).
lazy_check_element_of_aux(_,_,_).

:- block lazy_check_element_of_list(-,?,?,?).
lazy_check_element_of_list([],X,H,WF) :- !, equal_object_wf(X,H,WF).
lazy_check_element_of_list([Y|T],X,H,WF) :- !, %% print(quick_lst(H,[X,Y|T])),nl,
    quick_propagation_element_information([X,Y|T],H,WF,_). % TO DO: check that we loose no performance due to this
lazy_check_element_of_list(_,_,_,_).

% an incomplete subset check without enumeration
:- block lazy_check_subset_of(-,?,?), lazy_check_subset_of(?,-,?).
% lazy_check_subset_of(A,B,_) :- print(lazy_check_subset(A,B)),nl,fail.
lazy_check_subset_of(Set1,Set2,WF) :- nonvar(Set2),
   (Set2=closure(_,_,_) ; Set2=avl_set(_)),!, lazy_check_subset_of2(Set1,Set2,WF).
lazy_check_subset_of(_,_,_). % ignore other set representations
:- block lazy_check_subset_of2(-,?,?).
lazy_check_subset_of2([],_,_WF) :- !.
lazy_check_subset_of2([H|T],Set,WF) :- !, check_element_of_wf(H,Set,WF), lazy_check_subset_of2(T,Set,WF).
lazy_check_subset_of2(_,_,_). % ignore other set representations

:- block union2(-,?,?,?).
%union2(A,B,R,_) :- print(union2(A,B,R)),nl,fail.
union2([],S,Res,WF) :- equal_object_optimized_wf(S,Res,union2,WF).
union2([H|T],Set2,Res,WF) :-
   (T\==[],nonvar(Set2), Set2=[H2|T2], T2==[]  % minor optimisation for improved propagation; e.g., for x:S & S<:1..13 & S \/ {x} = S2 & x/: S2
   % the constraint is not yet detected straight away: x:S & S<:1..12 & S \/ {x} /= S
     -> union3(H2,T2,[H|T],Res,WF)
     ;  union3(H,T,Set2,Res,WF)
   ).
union3(H,T,Set2,Res,WF) :-
      add_element_wf(H,Set2,R,Done,WF), %print(add_element(H,Set2,R,Done)),nl,
      lazy_try_check_element_of(H,Res,WF),  % TO DO: propagate constraint that H is in Res
      (T==[] -> union2(T,R,Res,WF) ; union4(Done,T,R,Res,WF)).
:- block union4(-,?,?,?,?).
union4(_Done,T,R,Res,WF) :- union2(T,R,Res,WF).  % if WF not set to 2 there maybe equality_objects pending  from add_element_wf ! TO DO: investigate; see test 293


:- assert_must_succeed(exhaustive_kernel_check(union_generalized([[int(3)],[int(2),int(1),int(3)]],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_check(union_generalized([[int(3),int(2)],[],[int(2),int(1),int(3)]],[int(1),int(3),int(2)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(union_generalized([[int(3)],[int(3),int(4)],[int(2),int(1),int(3)]],[int(1),int(3),int(2)]))).
:- assert_must_succeed((kernel_objects:union_generalized([[]],Res),Res=[])).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2)]],Res),
         kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2),int(1)]],Res),
         kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2),int(1)],[],[int(2)]],Res),
         kernel_objects:equal_object(Res,[_,_]))).
:- assert_must_succeed((kernel_objects:union_generalized([[int(1)],[int(2)],X],Res),
                        kernel_objects:equal_object(X,Res), X = [int(2),int(1),int(3)])).
:- assert_must_succeed((kernel_objects:union_generalized([global_set('Name'),X,X,X],Res),
                        kernel_objects:equal_object(global_set('Name'),Res), X = [fd(2,'Name'),fd(1,'Name')])).
:- assert_must_succeed((kernel_objects:union_generalized([X,global_set('Name')],Res),
                        kernel_objects:equal_object(global_set('Name'),Res), X = [fd(2,'Name'),fd(1,'Name')])).
:- assert_must_fail((kernel_objects:union_generalized([[int(1)],[int(2)]],Res),(Res=[_];
         kernel_objects:equal_object(Res,[_,_,_|_])))).
:- assert_must_fail((kernel_objects:union_generalized([[int(1)],[int(1)]],Res),(Res=[];
         kernel_objects:equal_object(Res,[_,_|_])))).

union_generalized(S,Res) :- init_wait_flags(WF), union_generalized_wf(S,Res,WF), ground_wait_flags(WF).

:- block union_generalized_wf(-,-,?).
union_generalized_wf(SetsOfSets,Res,WF) :- var(SetsOfSets), Res==[],!,
       expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,union_generalized_wf),
       %print(check_all_empty(SetsOfSets)),nl, when(nonvar(SetsOfSets),(print(check_all_emptynv(SetsOfSets)),nl)),
       all_empty_sets_wf(ESetsOfSets,WF).
union_generalized_wf(SetsOfSets,Res,WF) :-
       union_generalized_wf2(SetsOfSets,Res,WF).

:- block union_generalized_wf2(-,?,?).
union_generalized_wf2(SetsOfSets,Res,WF) :-
       custom_explicit_sets:union_generalized_explicit_set(SetsOfSets,ARes,WF),!,
       %print_term_summary(union_generalized_explicit_set(SetsOfSets,ARes,Res,WF)),nl,
       equal_object_optimized_wf(ARes,Res,union_generalized_avl_set,WF).
union_generalized_wf2(SetsOfSets,Res,WF) :-
        expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,union_generalized_wf2),
        union_generalized2(ESetsOfSets,[],Res,WF). %, print(res(Res)),nl.

:- block union_generalized2(-,?,?,?).
union_generalized2([],S,Res,WF) :-  equal_object_optimized_wf(S,Res,union_generalized2,WF).
union_generalized2([H|T],UnionSoFar,Res,WF) :-  Res==[],!, %print(empty(H,T,UnionSoFar)),nl,
    empty_set_wf(H,WF), empty_set_wf(UnionSoFar,WF), all_empty_sets_wf(T,WF).
union_generalized2([H|T],UnionSoFar,Res,WF) :- union_wf(H,UnionSoFar,UnionSoFar2,WF),
        %print_message(called_union(H,UnionSoFar,UnionSoFar2,res(Res))), %%
        (((var(T);var(UnionSoFar2)),
          nonvar(Res),is_custom_explicit_set(Res,union_generalized2) % check important for Schneider2_Trees/NewSolver_v3_complex.mch and query CHOOSE_MODULES("bk-phi-H-2013", solution)  (0.1 vs 0.9 secs)
         )
         -> check_subset_of_wf(H,Res,WF)
             % this is only a very weak propagation; example, for union(v) = {4444} & v={{x},{y},{z}} we will instantiate v={{4444},...} and z=4444; see also test 1216
          ; true),
     union_generalized2(T,UnionSoFar2,Res,WF).

:- block all_empty_sets_wf(-,?).
all_empty_sets_wf([],_).
all_empty_sets_wf([H|T],WF) :- empty_set_wf(H,WF), all_empty_sets_wf(T,WF).

:- assert_must_succeed(exhaustive_kernel_check([commutative],intersection([int(3)],[int(2),int(1),int(3)],[int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],intersection([int(4),int(3),int(2)],[int(2),int(1),int(3)],[int(2),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],intersection([int(4),int(3),int(2)],[],[]))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],intersection([int(1),int(3)],[int(4),int(3),int(2)],[]))).
:- assert_must_succeed((kernel_objects:intersection(Y,X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,Y), Y =[fd(1,'Name')])).
:- assert_must_succeed((kernel_objects:intersection([int(1)],[int(2)],Res),Res=[])).
:- assert_must_succeed((kernel_objects:intersection([int(1)],[int(2)],[]))).
:- assert_must_fail((kernel_objects:intersection([int(1),int(4),int(3)],[int(2),int(3)],[]))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(1)],_))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(1)],[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(1)],[int(1),int(2)]))).
:- assert_must_succeed((kernel_objects:intersection([int(1),int(2)],[int(2),int(3)],Res),
         kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:intersection([int(2)],[int(2)],Res),
         kernel_objects:equal_object(Res,[int(2)]))).
:- assert_must_succeed((kernel_objects:intersection([int(2),int(3)],[int(3),int(4),int(2)],Res),
         kernel_objects:equal_object(Res,[int(2),int(3)]))).
:- assert_must_fail((kernel_objects:intersection([int(1)],[int(2)],Res),(
         kernel_objects:equal_object(Res,[_|_])))).
:- assert_must_fail((kernel_objects:intersection([int(1)],[int(1)],Res),(Res=[];
         kernel_objects:equal_object(Res,[_,_|_])))).
:- assert_must_fail((kernel_objects:intersection([fd(1,'Name')],X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,X))).


intersection(S1,S2,Res) :- init_wait_flags(WF), intersection(S1,S2,Res,WF), ground_wait_flags(WF).

:- block intersection(-,-,-,?).
intersection(Set1,Set2,Res,WF) :- (Set1==[] ; Set2==[]),!, empty_set_wf(Res,WF).
intersection(Set1,Set2,Res,WF) :- Set1==Set2,!, % print_term_summary(inter_eq_eq(Set1)),nl,
   equal_object_wf(Res,Set1,inter0_equal,WF).
intersection(Set1,Set2,Res,WF) :- Res==[],!,
   disjoint_sets(Set1,Set2,WF).
intersection(Set1,Set2,Res,WF) :-  % now we need to know at least a bit about both Set1 and Set2; at least given the current code below; TO DO: infer that {x} /\ s = {x}  =>  x:s
     %print(intersection0(Set1,Set2,Res,WF)),nl,
   intersection0(Set1,Set2,Res,WF),
   propagate_intersection(Set1,Set2,Res,WF).

:- block propagate_intersection(?,?,-,?).  % propagate constraint that result elements must be in both sets
%propagate_intersection(Set1,Set2,Set3,WF) :- print(inter(Set1,Set2,WF)),nl,print(res(Set3)),nl,fail.
propagate_intersection(Set1,Set2,[H|T],WF) :- !,
   propagate_intersection_aux(Set1,Set2,H,T,WF).
propagate_intersection(Set1,Set2,avl_set(A),WF) :- !,
   ((unknown_set(Set1) ; unknown_set(Set2)) % otherwise intersection0 has already triggered below
    ->  custom_explicit_sets:avl_approximate_size(A,Size),
        %print(prop_inter(Size)),nl,
        (Size<20
          -> expand_custom_set_to_list(avl_set(A),ESet,_,propagate_intersection)
          ;  avl_min(A,Min), avl_max(A,Max), ESet=[Min,Max]
        ),
        propagate_intersection(Set1,Set2,ESet,WF)
     ;  true).
% other cases: Set1,2,3 could be interval closure with unknown bounds,...
propagate_intersection(_,_,_,_).

:- block propagate_intersection_aux(-,-,-,?,?).
propagate_intersection_aux(Set1,Set2,H,T,WF) :-
   ((unknown_set(Set1) ; unknown_set(Set2)) % otherwise intersection0 has already triggered below
    ->  % print(prop(H,T,Set1,Set2)),nl,
        check_element_of_wf(H,Set1,WF),
        check_element_of_wf(H,Set2,WF),
        propagate_intersection(Set1,Set2,T,WF)
     ;  true).

unknown_set(Set) :- var(Set),!.
unknown_set([H|T]) :- (unknown_val(H) -> true ; unknown_set(T)).
unknown_val(Val) :- var(Val),!.
unknown_val(int(X)) :- var(X).
unknown_val(string(X)) :- var(X).
unknown_val(fd(X,_)) :- var(X).
unknown_val((A,B)) :- (unknown_val(A) -> true ; unknown_val(B)).
unknown_val([H|T]) :- (unknown_val(H) -> true ; unknown_set(T)).
:- block intersection0(-,?,?,?), intersection0(?,-,?,?).
intersection0(Set1,Set2,Res,WF) :- (Set1==[] ; Set2==[]),!, empty_set_wf(Res,WF).
intersection0(Set1,Set2,Res,WF) :- Set1==Set2,!, % print_term_summary(inter0_eq_eq(Set1)),nl,
   equal_object_wf(Res,Set1,inter0_equal,WF).
intersection0(Set1,Set2,Res,WF) :- Res==[],!,
   disjoint_sets(Set1,Set2,WF).
intersection0(Set1,Set2,Res,WF) :-
  intersection_with_interval_closure(Set1,Set2,Inter),!, % avoid expanding intervals at all
  %print_term_summary(inter0(Set1,Set2,Inter)),nl,
  equal_object_wf(Inter,Res,intersection0,WF).
intersection0(Set1,Set2,Res,WF) :-
   try_expand_and_convert_to_avl_unless_large(Set1,ESet1),
   try_expand_and_convert_to_avl_unless_large(Set2,ESet2),
   intersection1(ESet1,ESet2,Res,WF).

intersection1(Set1,Set2,Res,WF) :- nonvar(Set1),is_custom_explicit_set(Set1,intersection),
     %% print_term_summary(try_inter(Set1,Set2,Res)),nl, %%
  intersection_of_explicit_set_wf(Set1,Set2,Inter,WF), !,
     %print_term_summary(explicit_set_inter(Set1,Set2,Inter)), %%
  equal_object_wf(Inter,Res,intersection1,WF).
intersection1(Set1,Set2,Res,WF) :-
     %%  print_term_summary(intersection(Set1,Set2,Res) ), %%
    (Res==[] -> % print_term_summary(disjoint(Set1,Set2)),
                disjoint_sets(Set1,Set2,WF)
     ;
       (swap_set(Set1,Set2)  ->  intersection2(Set2,Set1,Res,WF)
                              ;  intersection2(Set1,Set2,Res,WF))
    ).

swap_set(Set1,_Set2) :- var(Set1),!.
swap_set(_Set1,Set2) :- var(Set2),!,fail.
%swap_set(_Set1,Set2) :- is_infinite_explicit_set(Set2),!,fail.
swap_set(avl_set(_),Set2) :- \+ functor(Set2,avl_set,2), %Set2 \= avl_set(_),
   Set2 \= [],
   \+ functor(Set2,closure,3), %Set2 \= closure(_,_,_),
   \+ functor(Set2,global_set,1). %Set2 \= global_set(_). % if it was a small closure,  intersection_of_explicit_set should have triggered
swap_set(closure(_P,_T,_B),Set2) :- ok_to_swap(Set2). % TO DO: for two closures: we could try and use the smallest one as first argument to intersection2
swap_set(global_set(_GS),Set2) :- ok_to_swap(Set2).

ok_to_swap(global_set(GS)) :- !, \+ is_infinite_or_very_large_explicit_set(global_set(GS),1000000).
ok_to_swap(closure(P,T,B)) :- !,\+ is_infinite_or_very_large_explicit_set(closure(P,T,B),1000000).
ok_to_swap(_).
% maybe also use is_efficient_custom_set as below ??
% what about freetype ?


intersection2(Set1,Set2,Res,WF) :- % print_term_summary(expand_inter1(Set1,Set2,Res,WF)),
                                expand_custom_set_to_list(Set1,ESet1,_,intersection2),
                                % print_term_summary(expanded_inter1(Set1)),
                                intersection3(ESet1,Set2,Res,WF).
:- block intersection3(-,?,?,?).
intersection3([],_,Res,WF) :- empty_set_wf(Res,WF).
intersection3([H|T],Set,Res,WF) :-
    (Res==[] -> %print(inter_res_empty(H,T,Set)),nl,
                not_element_of_wf(H,Set,WF),intersection3(T,Set,Res,WF)
      ; membership_test_wf(Set,H,MemRes,WF),
        %print(mem_test(H,MemRes)),nl, %
        intersection4(MemRes,H,T,Set,Res,WF)
    ).

:- block intersection4(-,?,?, ?,?,?).
intersection4(pred_true,H,T,Set,Result,WF) :- %print(inter4_mem_obj(H)),nl,
                              %check_element_of(H,Set),
                              equal_object_wf([H|Res],Result,intersection4,WF),
                              intersection3(T,Set,Res,WF).
intersection4(pred_false,_H,T,Set,Res,WF) :- %not_element_of(H,Set),
                              intersection3(T,Set,Res,WF).


:- assert_must_succeed(exhaustive_kernel_check_wfdet(disjoint_sets([int(5)],[int(2),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(disjoint_sets([int(5)],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(disjoint_sets([int(5),int(2)],[int(6),int(1),int(3)],WF),WF)).

disjoint_sets(S1,S2) :- init_wait_flags(WF),
   disjoint_sets(S1,S2,WF),
   ground_wait_flags(WF).

:- block disjoint_sets(-,?,?), disjoint_sets(?,-,?).
disjoint_sets(S1,S2,WF) :-
    % print_term_summary(disjoint_sets(S1,S2)),
    % TO DO: we could provide faster code for two avl sets or intervals; but probably caught in intersection code above?
        ((S1==[];S2==[]) -> true
         ; is_efficient_custom_set(S2) -> expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_1),
                                          disjoint_sets2(ESet1,S2,WF)
         ; is_efficient_custom_set(S1) -> expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_2),
                                          disjoint_sets2(ESet2,S1,WF)
         ; expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_3),
           %expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_4),
           disjoint_sets2(ESet1,S2,WF)
        ).

 % TO DO: we could infer some constraints on the possible max sizes of the sets
 % for finite types (sum of size must be <= size of type)
:- block disjoint_sets2(-,?,?).
disjoint_sets2([],_,_WF).
disjoint_sets2([H|T],S2,WF) :- not_element_of_wf(H,S2,WF), disjoint_sets2(T,S2,WF).

% NOT YET USED: not_disjoint_sets could be used for S /\ R /= {}
:- assert_must_succeed(exhaustive_kernel_check_wfdet(not_disjoint_sets([int(3)],[int(2),int(1),int(3)],WF),WF)).
:- block not_disjoint_sets(-,?,?), not_disjoint_sets(?,-,?).
not_disjoint_sets(S1,S2,WF) :-
        ((S1==[];S2==[]) -> fail
         ; is_efficient_custom_set(S2) -> expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_1),
                                          not_disjoint_sets2(ESet1,S2,WF)
         ; is_efficient_custom_set(S1) -> expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_2),
                                          not_disjoint_sets2(ESet2,S1,WF)
         ; expand_custom_set_to_list(S1,ESet1,_,disjoint_sets_3),
           %expand_custom_set_to_list(S2,ESet2,_,disjoint_sets_4),
           not_disjoint_sets2(ESet1,S2,WF)
        ).

:- block not_disjoint_sets2(-,?,?).
not_disjoint_sets2([],_,_WF).
not_disjoint_sets2([H|T],S2,WF) :- membership_test_wf(S2,H,MemRes,WF), not_disjoint3(MemRes,T,S2,WF).

:- block not_disjoint3(-,?,?,?).
not_disjoint3(pred_true,_,_,_).
not_disjoint3(pred_false,T,S2,WF) :- not_disjoint_sets2(T,S2,WF).

:- assert_must_succeed(exhaustive_kernel_check_wfdet(intersection_generalized_wf([[int(3)],[int(2),int(1),int(3)]],[int(3)],unknown,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wfdet(intersection_generalized_wf([[int(3),int(2)],[int(2),int(1),int(3)],[int(4),int(3)]],[int(3)],unknown,WF),WF)).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf(avl_set(node(avl_set(node(fd(1,'Name'),true,1,empty,node(fd(2,'Name'),true,0,empty,empty))),
       true,1,empty,node(avl_set(node(fd(2,'Name'),true,1,empty,node(fd(3,'Name'),true,0,empty,empty))),true,0,empty,empty))),
       avl_set(node(fd(2,'Name'),true,0,empty,empty)),unknown,_WF))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(1)],[int(2)]],Res,unknown,_WF),Res=[])).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(1)],[int(2),int(1)]],Res,unknown,_WF),
         kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(1)],X,[int(2),int(3),int(1)]],Res,unknown,_WF),
                       X = [int(2),int(1)],
         kernel_objects:equal_object(Res,[int(1)]))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([X,X,[int(2),int(3),int(1)]],Res,unknown,_WF),
                       X = [int(2),int(1)], kernel_objects:equal_object(Res,[int(1),int(2)]))).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([[int(2),int(1),int(3)],X,[int(1),int(2)],X],Res,unknown,_WF),
                        kernel_objects:equal_object(X,Res), X = [int(2),int(1)])).
:- assert_must_succeed((kernel_objects:intersection_generalized_wf([global_set('Name'),X],Res,unknown,_WF),
                        kernel_objects:equal_object(X,Res), X = [fd(2,'Name'),fd(1,'Name')])).
:- assert_must_fail((kernel_objects:intersection_generalized_wf([[int(1)],[int(2)]],Res,unknown,_WF),(
         kernel_objects:equal_object(Res,[_|_])))).
:- assert_must_fail((kernel_objects:intersection_generalized_wf([[int(1)],[int(1)]],Res,unknown,_WF),(Res=[];
         kernel_objects:equal_object(Res,[_,_|_])))).
:- assert_must_abort_wf(kernel_objects:intersection_generalized_wf([],_R,unknown,WF),WF).

% code for general_intersection
:- block intersection_generalized_wf(-,?,?,?).
intersection_generalized_wf(SetsOfSets,Res,Span,WF) :-
  expand_custom_set_to_list(SetsOfSets,ESetsOfSets,_,intersection_generalized_wf),
  intersection_generalized2(ESetsOfSets,Res,Span,WF).

intersection_generalized2([],Res,Span,WF) :-  /* Atelier-B manual requires argument to inter to be non-empty */
   add_wd_error_set_result('inter applied to empty set','',Res,[],Span,WF).
intersection_generalized2([H|T],Res,_Span,WF) :- intersection_generalized3(T,H,Res,WF).
:- block intersection_generalized3(-,?,?,?).
intersection_generalized3([],SoFar,Res,WF) :- equal_object_optimized_wf(SoFar,Res,intersection_generalized3,WF).
intersection_generalized3([H|T],InterSoFar,Res,WF) :-
      intersection(H,InterSoFar,InterSoFar2,WF),
      intersection_generalized3(T,InterSoFar2,Res,WF).

:- assert_must_succeed(exhaustive_kernel_check(difference_set([int(3),int(2)],[int(2),int(1),int(3)],[]))).
:- assert_must_succeed(exhaustive_kernel_check(difference_set([int(3),int(2)],[int(2),int(1),int(4)],[int(3)]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[int(1),int(2)]],[]),
                        kernel_objects:equal_object(SSS,[[int(2),int(1)]]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[int(1),int(2)]],R), kernel_objects:equal_object(R,[]),
                        kernel_objects:equal_object(SSS,[[int(2),int(1)]]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[fd(1,'Name'),fd(2,'Name')]],R),
                        kernel_objects:equal_object(R,[]),
                        kernel_objects:equal_object(SSS,[[fd(2,'Name'),fd(1,'Name')]]))).
:- assert_must_succeed((kernel_objects:difference_set(SSS,[[int(1),int(2)]],[]),
                        kernel_objects:equal_object(SSS,[[int(1),int(2)]]))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[int(1)],_))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[int(2)],_))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[int(2)],[int(1)]))).
:- assert_must_succeed((kernel_objects:difference_set([int(1),int(2)],[],[int(2),int(1)]))).
:- assert_must_succeed((kernel_objects:difference_set([],[int(1),int(2)],[]))).
:- assert_must_succeed((kernel_objects:difference_set(Y,X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,[]), Y =[fd(1,'Name')])).
:- assert_must_succeed((kernel_objects:difference_set(X,Y,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,[fd(3,'Name'),fd(1,'Name')]), Y =[fd(2,'Name')])).
:- assert_must_fail((kernel_objects:difference_set(X,Y,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,[]), Y =[fd(1,'Name'),fd(2,'Name')])).
:- assert_must_fail((kernel_objects:difference_set(Y,X,Res),X=global_set('Name'),
                         kernel_objects:equal_object(Res,Y), Y =[fd(1,'Name')])).

difference_set(Set1,Set2,Res) :- init_wait_flags(WF),
  difference_set_wf(Set1,Set2,Res,WF),
  ground_wait_flags(WF).

:- block difference_set_wf(-,-,?,?).
difference_set_wf(Set1,_,Res,WF) :- Set1==[],!,empty_set_wf(Res,WF).
difference_set_wf(Set1,Set2,Res,WF) :- Set2==[],!,equal_object_wf(Set1,Res,difference_set_wf,WF).
difference_set_wf(Set1,Set2,Res,WF) :- difference_set1(Set1,Set2,Res,WF).


:- block difference_set1(?,-,-,?), difference_set1(-,?,-,?).
difference_set1(Set1,Set2,Res,WF) :-
   nonvar(Set1),is_custom_explicit_set(Set1,difference_set),
   difference_of_explicit_set_wf(Set1,Set2,Diff,WF), !,
   %print_term_summary(explicit_set_diff(Set1,Set2,Diff)),%
   equal_object_wf(Diff,Res,difference_set1_1,WF).
difference_set1(Set1,Set2,Res,WF) :- Set2==[],!,equal_object_wf(Set1,Res,difference_set1_2,WF).
difference_set1(Set1,Set2,Res,WF) :- Res==[],!, check_subset_of_wf(Set1,Set2,WF).
difference_set1(Set1,Set2,Res,WF) :- %print(diff(Set1,Set2,Res)),nl,%
  expand_custom_set_to_list(Set1,ESet1,_,difference_set1), %print_term_summary(expanded(Set1,ESet1)),
  compute_diff(ESet1,Set2,Res,WF),
  %print_term_summary(computed_diff(ESet1,Set2,Res)),nl,%
  propagate_into2(Res,ESet1,Set2,WF).  %print(propagated(ESet1,Set2)),nl.

:- block compute_diff(-,?,?,?).
compute_diff([],_Set2,Res,WF) :- empty_set_wf(Res,WF).
compute_diff([H|T],Set2,Res,WF) :-
  membership_test_wf(Set2,H,MemRes,WF),compute_diff2(MemRes,H,T,Set2,Res,WF).

:- block compute_diff2(-,?,?,?,?,?).
compute_diff2(pred_true,_H,T,Set2,Res,WF) :- compute_diff(T,Set2,Res,WF).
compute_diff2(pred_false,H,T,Set2,Res,WF) :- equal_object_wf([H|R2],Res,compute_diff2,WF),
  compute_diff(T,Set2,R2,WF).

% propagate all elements from one set into another one; do not use for computation; may skip elements ...
/* this version not used at the moment:
:- block propagate_into(-,?,?).
propagate_into(_,Set2,_WF) :- nonvar(Set2),
    is_custom_explicit_set(Set2,propagate_into),!. % second set already fully known
propagate_into([],_,_WF) :- !.
propagate_into([H|T],Set,WF) :- !,check_element_of_wf(H,Set,WF), propagate_into(T,Set,WF).
propagate_into(Set1,Set2,WF) :- is_custom_explicit_set(Set1,propagate_into),!,
   (is_infinite_explicit_set(Set1) -> true ;
    expand_custom_set_to_list(Set1,ESet1), propagate_into(ESet1,Set2,WF)). */

:- block propagate_into2(-,?,?,?).
propagate_into2(_,Set2,_NegSet,_WF) :- nonvar(Set2),
    is_custom_explicit_set(Set2,propagate_into),!. % second set already fully known
propagate_into2([],_,_,_WF) :- !.
propagate_into2([H|T],PosSet,NegSet,WF) :- !,
  check_element_of_wf(H,PosSet,WF),
  not_element_of_wf(H,NegSet,WF),propagate_into2(T,PosSet,NegSet,WF).
propagate_into2(Set1,PosSet,NegSet,WF) :- is_custom_explicit_set(Set1,propagate_into),!,
   (is_infinite_explicit_set(Set1) -> true ;
    expand_custom_set_to_list(Set1,ESet1,_,propagate_into2), propagate_into2(ESet1,PosSet,NegSet,WF)).

:- assert_must_succeed(exhaustive_kernel_check_wf(in_difference_set_wf(int(33),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).
:- block in_difference_set_wf(-,-,-,?).
in_difference_set_wf(A,X,Y,WF) :-
    (treat_arg_symbolically(X) ; treat_arg_symbolically(Y) ; preference(convert_comprehension_sets_into_closures,true)),
    % symbolic treatment would also make sense when A is nonvar and X var to force A to be in X ?!
    % print_term_summary(symbolic_treatment_in_difference_set_wf(A,X,Y,WF)),
    !,
    check_element_of_wf(A,X,WF), not_element_of_wf(A,Y,WF).
in_difference_set_wf(A,X,Y,WF) :-
    difference_set_wf(X,Y,Diff,WF),
    check_element_of_wf(A,Diff,WF).

treat_arg_symbolically(X) :- var(X),!.
treat_arg_symbolically(global_set(_)).
treat_arg_symbolically(freetype(_)).
treat_arg_symbolically(closure(P,T,B)) :- \+ small_interval(P,T,B).

small_interval(P,T,B) :- is_interval_closure(P,T,B,Low,Up),
   number(Low),number(Up),
   Up-Low < 500. % Magic Constant; TO DO: determine good value


:- assert_must_succeed(exhaustive_kernel_check_wf(not_in_difference_set_wf(int(2),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(not_in_difference_set_wf(int(111),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(not_in_difference_set_wf(int(1),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).

:- block not_in_difference_set_wf(-,-,-,?).
not_in_difference_set_wf(A,X,Y,WF) :-
    (treat_arg_symbolically(X) ; treat_arg_symbolically(Y) ; preference(convert_comprehension_sets_into_closures,true)),
    !, %print(not_in_diff(A,X,Y,WF)),nl,
    % A : (X-Y)  <=> A:X & not(A:Y)
    % A /: (X-Y)  <=>  A/: X or A:Y
    membership_test_wf(X,A,AX_Res,WF),
    (AX_Res==pred_false -> true
     ; bool_pred:negate(AX_Res,NotAX_Res),
       b_interpreter_check:disjoin(NotAX_Res,AY_Res,pred_true,priority(16384),priority(16384),WF), % better: uese a version that does not do a case split ?! or use last wait flag ?
       membership_test_wf(Y,A,AY_Res,WF)
     ).
not_in_difference_set_wf(A,X,Y,WF) :-
    difference_set_wf(X,Y,Diff,WF),
    not_element_of_wf(A,Diff,WF).


:- assert_must_succeed(exhaustive_kernel_check_wf(in_intersection_set_wf(int(2),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).

:- block in_intersection_set_wf(-,-,-,?).
in_intersection_set_wf(A,X,Y,WF) :-
    (treat_arg_symbolically(X) ; treat_arg_symbolically(Y) ; preference(convert_comprehension_sets_into_closures,true)),
    % print_term_summary(symbolic_treatment_in_difference_set_wf(A,X,Y,WF)),
    !,
    Y \== [], % avoid setting up check_element_of for X then
    check_element_of_wf(A,X,WF), check_element_of_wf(A,Y,WF).
in_intersection_set_wf(A,X,Y,WF) :-
    intersection(X,Y,Inter,WF),
    check_element_of_wf(A,Inter,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(not_in_intersection_set_wf(int(3),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).
:- block not_in_intersection_set_wf(-,-,-,?).
not_in_intersection_set_wf(_A,_X,Y,_WF) :- Y == [], !. % intersection will be empty; avoid analysing X
not_in_intersection_set_wf(A,X,Y,WF) :-
    (treat_arg_symbolically(X) ; treat_arg_symbolically(Y) ; preference(convert_comprehension_sets_into_closures,true)),
    !, %print(not_in_intersection(A,X,Y,WF)),nl,
    % A : (X /\ Y)  <=> A:X & A:Y
    % A /: (X /\ Y)  <=>  A/:X or A/:Y
    membership_test_wf(X,A,AX_Res,WF),
    (AX_Res==pred_false -> true
     ; bool_pred:negate(AX_Res,NotAX_Res), bool_pred:negate(AY_Res,NotAY_Res),
       b_interpreter_check:disjoin(NotAX_Res,NotAY_Res,pred_true,priority(16384),priority(16384),WF), % better: uese a version that does not do a case split ?! or use last wait flag ?
       membership_test_wf(Y,A,AY_Res,WF)
     ).
not_in_intersection_set_wf(A,X,Y,WF) :-
    intersection(X,Y,Inter,WF),
    not_element_of_wf(A,Inter,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(in_union_set_wf(int(2),[int(33),int(2)],[int(2),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(in_union_set_wf(int(33),[int(32),int(2)],[int(2),int(1),int(33)],WF),WF)).

:- block in_union_set_wf(-,-,-,?).
in_union_set_wf(A,X,Y,WF) :-
    (treat_arg_symbolically(X) ; treat_arg_symbolically(Y) ; preference(convert_comprehension_sets_into_closures,true)),
    % symbolic treatment would also make sense when A is nonvar and X var to force A to be in X ?!
    % print_term_summary(symbolic_treatment_in_difference_set_wf(A,X,Y,WF)),
    !,
    %print(in_union(A,X,Y)),nl,
    membership_test_wf(X,A,AX_Res,WF),
    (AX_Res==pred_true -> true
     ; b_interpreter_check:disjoin(AX_Res,AY_Res,pred_true,priority(16384),priority(16384),WF), % better: use a version that does not do a case split ?! or use last wait flag ?
       membership_test_wf(Y,A,AY_Res,WF)
     ).
in_union_set_wf(A,X,Y,WF) :-
    union_wf(X,Y,Union,WF),
    check_element_of_wf(A,Union,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(not_in_union_set_wf(int(3),[int(32),int(2)],[int(2),int(1),int(33)],WF),WF)).

:- block not_in_union_set_wf(-,-,-,?).
not_in_union_set_wf(A,X,Y,WF) :- % print(not_in_union(A,X,Y)),nl,
    not_element_of_wf(A,X,WF),
    not_element_of_wf(A,Y,WF).

% ---------------------


strict_subset_of(X,Y) :-
    init_wait_flags(WF),
    strict_subset_of_wf(X,Y,WF),
    ground_wait_flags(WF).

:- assert_must_succeed(exhaustive_kernel_check(strict_subset_of_wf([int(3),int(2)],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_check(strict_subset_of_wf([],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_check(strict_subset_of_wf([],[ [] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([int(3),int(2),int(1)],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([int(1),int(4)],[int(2),int(1),int(3)],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([[]],[],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(strict_subset_of_wf([],[],_))).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:strict_subset_of(Y,X), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [[int(1),int(2)]], X=[[int(2)],[int(2),int(1)]])).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [fd(1,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:strict_subset_of_wf(Y,X,_WF), Y = [fd(3,'Name'),fd(2,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [fd(1,'Name'),fd(3,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(1),int(2)], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(2)], X=[int(2)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [int(2)], X=[int(1)])).
:- assert_must_fail((kernel_objects:strict_subset_of_wf(X,Y,_WF), Y = [], X=[int(1)])).


 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(\+ environ(disable_chr, true)).
:- use_module(chrsrc(chr_set_membership)).
:- else.
chr_subset_strict(_,_).
:- endif.

strict_subset_of_wf(Set1,Set2,WF) :-
   chr_subset_strict(Set1,Set2),
   strict_subset_of_wf_aux(Set1,Set2,WF).

%:- block strict_subset_of_wf(-,-,?).
strict_subset_of_wf_aux(Set1,Set2,WF) :- Set1==[],!,not_empty_set_wf(Set2,WF).
%strict_subset_of_wf_aux(Set1,Set2,WF) :- var(Set2),nonvar(Set1), print(subs(Set1,Set2)),nl,fail.
strict_subset_of_wf_aux(Set1,Set2,WF) :- nonvar(Set2), singleton_set(Set2,_),!, empty_set_wf(Set1,WF).
strict_subset_of_wf_aux(Set1,Set2,WF) :- %print_quoted(check_subset_of(Set1,Set2)),nl,
   not_empty_set_wf(Set2,WF),
   get_cardinality_powset_wait_flag(Set2,strict_subset_of_wf,WF,_,LWF),
   % we could subtract 1 from priority !? (get_cardinality_pow1set_wait_flag)
   when(((nonvar(LWF),(nonvar(Set1);ground(Set2)))  ;  (nonvar(Set1),nonvar(Set2))  ),
      strict_subset_of_aux_block(Set1,Set2,WF,LWF)).

strict_subset_of_aux_block(Set1,_Set2,_WF,_LWF) :-
    Set1==[],
    !. % we have already checked that Set2 is not empty
strict_subset_of_aux_block(Set1,Set2,WF,_LWF) :-
    nonvar(Set2), is_definitely_maximal_set(Set2),
    !,
    not_equal_object_wf(Set1,Set2,WF).
strict_subset_of_aux_block(Set1,Set2,WF,_LWF) :- nonvar(Set2), singleton_set(Set2,_),!, empty_set_wf(Set1,WF).
strict_subset_of_aux_block(Set1,Set2,_WF,_LWF) :-
    both_global_sets(Set1,Set2,G1,G2),
    !, %(print(check_strict_subset_of_global_sets(G1,G2)),nl,
    check_strict_subset_of_global_sets(G1,G2).
strict_subset_of_aux_block(Set1,Set2,_WF,_LWF) :-
    var(Set1),ground(Set2),
    !,  % DO WE STILL NEED THIS VERSION ????
    non_free(Set1), % as we now force order
    expand_custom_set_to_list(Set2,ESet2,_,strict_subset_of_wf),  %print(gen_strict_subsets(ESet2)),nl,
    gen_strict_subsets(Set1,ESet2).
strict_subset_of_aux_block(Set1,Set2,WF,LWF) :-
     strict_subset_of0(Set1,Set2,WF,LWF).

% TO DO (26.10.2014): test 1270 now passes thanks to maximal set check above
% but we should need a better way of ensuring that something like {ssu|ssu<<:POW(elements)} is efficiently computed
% (which it no longer is once the unbound_variable check had been fixed)
% we could also just generally use Set1 <: Set2 & Set1 /= Set2


%:- block strict_subset_of0(-,?,?,?).  % required to wait: we know Set2 must be non-empty, but Set1 could be an avl-tree or closure
% TO DO: deal with infinite Set1
strict_subset_of0(Set1,Set2,WF,LWF) :-
      expand_custom_set_to_list(Set1,ESet1,_,strict_subset_of0),
      (ESet1==[] -> true %not_empty_set(Set2) already checked above
        ; is_infinite_explicit_set(Set2) -> %print(inf(Set2)),nl,
           % Set1 is expanded to a list ESet1 and thus finite: it is sufficient to check subset relation
           check_subset_of_wf(ESet1,Set2,WF)
        ;  try_expand_custom_set_wf(Set2,ESet2,strict_subset_of0,WF),
            %%try_prop_card_lt(ESet1,ESet2), try_prop_card_gt(ESet2,ESet1),
           strict_subset_of2(ESet1,[],ESet2,WF,LWF)
       ).

:- block strict_subset_of2(-,?,?,?,-).
%strict_subset_of2(S,SoFar,Set2,WF) :- print(strict_subset_of2(S,SoFar,Set2,WF)),nl,fail.
strict_subset_of2([],_,RemS,WF,_LWF) :- not_empty_set_wf(RemS,WF). /* we know it must be explicit set */
strict_subset_of2([H|T],SoFar,Set2,WF,LWF) :- var(Set2),!,
    Set2 = [H|Set2R],
    add_new_element_wf(H,SoFar,SoFar2,WF), %was SoFar2 = [H|SoFar],
    strict_subset_of2(T,SoFar2,Set2R,WF,LWF).
strict_subset_of2([H|T],SoFar,Set2,WF,LWF) :-
 % when_sufficiently_for_member(H,Set2,WF,
     remove_element_wf(H,Set2,RS2,WF), not_empty_set_wf(RS2,WF),
     not_element_of_wf(H,SoFar,WF),  /* consistent((H,SoFar)), necessary? */
     when((nonvar(T) ; (ground(LWF),ground(RS2))),
           (add_new_element_wf(H,SoFar,SoFar2,WF), %SoFar2 = [H|SoFar],
            strict_subset_of2(T,SoFar2,RS2,WF,LWF) )).




:- assert_must_succeed(exhaustive_kernel_check(partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_check(partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_check(partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5)],[int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)], [int(3)] ],_))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ], _))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2)],[ [int(2)], [int(1)], [] ], _))).
:- assert_must_fail((kernel_objects:partition_wf([int(1),int(2)],[ [int(2)], [int(1),int(2)] ], _))).
:- assert_must_fail((kernel_objects:partition_wf([int(1),int(3)],[ [int(1)], [int(2)] ], _))).
:- assert_must_fail((kernel_objects:partition_wf([int(1),int(2),int(3)],[ [int(1)], [int(2)] ], _))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1)],[S1,S2],_WF), S1=[H|T], S2==[],T==[],H==int(1))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2)],[S1,S2],_WF), S1=[H|T], S2=[int(1)],(preferences:preference(use_clpfd_solver,true) -> T==[],H==int(2) ; T=[],H=int(2)))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2),int(3)],[S1,S2,S3],_WF), S1=[H2|T], S3=[int(3)],T=[H1|TT],H2=int(2),TT==[],S2==[],H1==int(1))).
:- assert_must_succeed((kernel_objects:partition_wf([int(1),int(2),int(3)],[[int(1)],X,[int(2)]],_WF),
			X==[int(3)])).

:- use_module(bsets_clp,[disjoint_union_generalized_wf/3]).
:- use_module(kernel_tools,[ground_value/1]).
:- block partition_wf(?,-,?).
partition_wf(Set,ListOfSets,WF) :-  % print_term_summary(partition_wf(Set,ListOfSets,_WF)),nl, %
   (ground_value(Set),find_non_ground_set(ListOfSets,NGS,Rest) ->
        % we have: partition(Set, GroundSet1,...,GroundSetk, NGS)
        % print(backwards_computation_of_partition(NGS)),nl, %
       bsets_clp:disjoint_union_generalized_wf(Rest,RestSet,WF),
       check_subset_of_wf(RestSet,Set,WF), % otherwise this is not a partition of Set
       difference_set(Set,RestSet,NGS)
     ; bsets_clp:disjoint_union_generalized_wf(ListOfSets,Set,WF)
   ),
   all_disjoint(ListOfSets,WF).

 :- assert_must_succeed((kernel_objects:find_non_ground_set([int(1),int(2),A,int(5)],B,C), B==A,C==[int(1),int(2),int(5)])).
find_non_ground_set([H|T],NG,Rest) :-
  (ground_value(H) -> Rest=[H|TR], find_non_ground_set(T,NG,TR)
    ; ground_value(T),NG=H, Rest=T).

:- block all_disjoint(-,?).
all_disjoint(closure(P,T,B),WF) :-
   expand_custom_set(closure(P,T,B),ExpandedSet), all_disjoint(ExpandedSet,WF).
all_disjoint(avl_set(AVL),WF) :-
   expand_custom_set(avl_set(AVL),ExpandedSet), all_disjoint(ExpandedSet,WF).
% no case for global_set: cannot be a set of sets
all_disjoint([],_WF).
all_disjoint([H|T],WF) :- all_disjoint_with(T,H,WF),
   all_disjoint(T,WF).

:- block all_disjoint_with(-,?,?).
all_disjoint_with([],_,_WF).
all_disjoint_with([H|T],Set1,WF) :- disjoint_sets(Set1,H,WF), all_disjoint_with(T,Set1,WF).




:- assert_must_succeed(exhaustive_kernel_fail_check(not_partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5)],[int(1)] ],_))).
:- assert_must_succeed(exhaustive_kernel_check(not_partition_wf([int(1),int(2),int(5)],[ [int(2)], [int(5),int(1)], [int(3)] ],_))).
:- assert_must_fail((kernel_objects:not_partition_wf([int(1),int(2)],[ [int(2)], [int(1)] ], _))).
:- assert_must_fail((kernel_objects:not_partition_wf([int(1),int(2)],[ [int(2)], [int(1)], [] ], _))).
:- assert_must_succeed((kernel_objects:not_partition_wf([int(1),int(2)],[ [int(2)], [int(1),int(2)] ], _))).
:- assert_must_succeed((kernel_objects:not_partition_wf([int(1),int(3)],[ [int(1)], [int(2)] ], _))).
:- assert_must_succeed((kernel_objects:not_partition_wf([int(1),int(2),int(3)],[ [int(1)], [int(2)] ], _))).

:- block not_partition_wf(?,-,?).
not_partition_wf(FullSet,ListOfSets,WF) :-
  expand_custom_set_to_list(ListOfSets,EListOfSets,_,not_partition_wf), % necessary ? ListOfSets is actually list
  not_partition_wf2(EListOfSets,[],FullSet,WF).

:- block not_partition_wf2(-,?,?,?).
%not_partition_wf2(Sets,SoFar,_) :- print(not_part2(Sets,SoFar)),nl,fail.
not_partition_wf2([],ElementsSoFar,FullSet,WF) :- not_equal_object_wf(ElementsSoFar,FullSet,WF).
not_partition_wf2([Set1|Rest],ElementsSoFar,FullSet,WF) :-
      expand_custom_set_to_list(Set1,ESet1,_,not_partition_wf2), not_partition_wf3(ESet1,ElementsSoFar,Rest,FullSet,WF).

:- block not_partition_wf3(-,?,?,?,?).
not_partition_wf3([],S,OtherSets,FullSet,WF) :-
   not_partition_wf2(OtherSets,S,FullSet,WF). % finished treating this set
not_partition_wf3([H|T],ElementsSoFar,OtherSets,FullSet,WF) :-
   membership_test_wf(ElementsSoFar,H,MemRes,WF),
   not_partition_wf4(MemRes,H,T,ElementsSoFar,OtherSets,FullSet,WF).

:- block not_partition_wf4(-,?,?,?,?,?,?).
not_partition_wf4(pred_true,_,_,_,_,_,_). % Not disjoint
not_partition_wf4(pred_false,H,T,ElementsSoFar,OtherSets,FullSet,WF) :-
   add_element_wf(H,ElementsSoFar,ElementsSoFar2,WF), % we could also already check whether H in FullSet or not
   not_partition_wf3(T,ElementsSoFar2,OtherSets,FullSet,WF).

:- assert_must_succeed(exhaustive_kernel_succeed_check(check_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(check_subset_of([int(1),int(2),int(5)],[int(2),int(5),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(check_subset_of([int(1),int(3),int(5)],[int(2),int(5),int(1)]))).
:- assert_must_succeed((kernel_objects:power_set(global_set('Name'),PS),kernel_objects:check_subset_of(X,PS),
                        kernel_objects:equal_object(X,[[fd(2,'Name'),fd(1,'Name')]]))).
:- assert_must_succeed(findall(X,kernel_objects:check_subset_of(X,[[int(1),int(2)],[]]),[_1,_2,_3,_4])).
:- assert_must_succeed((kernel_objects:check_subset_of(X,[[int(1),int(2)],[]]),
                        nonvar(X),
                        kernel_objects:equal_object(X,[[int(2),int(1)]]))).
:- assert_must_succeed((kernel_objects:check_subset_of_wf(Y,X,_WF), Y = [fd(1,'Name')],
           nonvar(X),X=[H|T], var(T), H==fd(1,'Name'), X=Y)).
:- assert_must_succeed((kernel_objects:check_subset_of(Y,X), Y = [fd(1,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:check_subset_of(Y,X), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:check_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
:- assert_must_succeed((kernel_objects:sample_closure(C),kernel_objects:check_subset_of(C,global_set('NAT')))).
:- assert_must_succeed((kernel_objects:check_subset_of(global_set('NAT'),global_set('NAT')))).
:- assert_must_succeed((kernel_objects:check_subset_of(global_set('NAT'),global_set('NATURAL')))).
:- assert_must_fail((kernel_objects:check_subset_of(global_set('NAT'),global_set('NATURAL1')))).
:- assert_must_fail((kernel_objects:check_subset_of(global_set('NAT'),global_set('NAT1')))).
:- assert_must_fail((kernel_objects:check_subset_of(X,Y), Y = [fd(1,'Name')], kernel_objects:equal_object(X,global_set('Name')))).
  /* TO DO: add special treatment for closures and type checks !! */

check_subset_of(Set1,Set2) :- init_wait_flags(WF),
  check_subset_of_wf(Set1,Set2,WF),
  ground_wait_flags(WF).

check_finite_subset_of_wf(Set1,Set2,WF) :-
   check_subset_of_wf(Set1,Set2,WF),
   is_finite_set_wf(Set1,WF).

:- block check_subset_of_wf(-,-,?).
check_subset_of_wf(Set1,Set2,WF) :-     % print_term_summary(check_subset_of(Set1,Set2)),nl,
   (both_global_sets(Set1,Set2,G1,G2)
     -> ( %print(check_subset_of_global_sets(G1,G2)),nl,
         check_subset_of_global_sets(G1,G2))
      ;  check_subset_of0(Set1,Set2,WF)
   ).

both_global_sets(S1,S2,G1,G2) :- nonvar(S1),nonvar(S2),
  is_global_set(S1,G1), is_global_set(S2,G2).

% check if we have a global set or interval
% is_global_set([],R) :- !, R=interval(0,-1). % useful ???
is_global_set(global_set(G1),R) :- !,
   (custom_explicit_sets:get_integer_set_interval(G1,Low,Up) -> R=interval(Low,Up) ; R=G1).
is_global_set(Closure,R) :-
    custom_explicit_sets:is_interval_closure_or_integerset(Closure,Low,Up),!,
    R=interval(Low,Up).


:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(0,0),interval(minus_inf,inf))).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(-200,1000),interval(minus_inf,inf))).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(10,1000),interval(0,inf))).
:- assert_must_fail(kernel_objects:check_subset_of_global_sets(interval(-10,1000),interval(0,inf))).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(0,inf),interval(0,inf))).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(0,inf),interval(minus_inf,inf))).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(1,inf),interval(0,inf))).
:- assert_must_succeed(kernel_objects:check_subset_of_global_sets(interval(1,inf),interval(minus_inf,inf))).

% to do: also extend to allow intervals with inf/minus_inf
check_subset_of_global_sets(X,Y) :- (var(X) ; var(Y)),
   add_internal_error('Illegal call: ',check_subset_of_global_sets(X,Y)),fail.
check_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :- !,
   interval_subset(Low1,Up1,Low2,Up2).
check_subset_of_global_sets(X,X) :- !. % both args must be atomic and ground (global set names)
% BUT WE COULD HAVE {x|x>0} <: NATURAL1  ?  interval(0,inf) <: NATURAL1
check_subset_of_global_sets(X,Y) :- check_strict_subset_of_global_sets(X,Y).

% To do: perform some treatment of inf, minus_inf values here <----
interval_subset(Low1,Up1,Low2,Up2) :-  % print(interval_subset(Low1,Up1,Low2,Up2)),nl,
   (var(Low1) ; var(Up1)), % otherwise we can use code below
   finite_interval(Low1,Up1), finite_interval(Low2,Up2), % inf can appear as term; but only directly not later
   !,
   %print(posting),nl,
   % Maybe to do: try to avoid CLPFD overflows if possible
   clpfd_in_interval(Low1,Up1,Low2,Up2).
interval_subset(Low1,Up1,Low2,Up2) :-
   interval_subset_aux(Low1,Up1,Low2,Up2).

% check if we have a finite interval (fails for inf/minus_inf terms)
finite_interval(Low1,Up1) :- (var(Low1) -> true ; number(Low1)), (var(Up1) -> true ; number(Up1)).


% assert Low1..Up1 <: Low2..Up2
clpfd_in_interval(Low1,Up1,Low2,Up2) :-
  (preferences:preference(use_chr_solver,true)
   -> chr_integer_inequality:chr_in_interval(Low1,Up1,Low2,Up2) ; true),
   % TO DO: improve detection of Low1 #=< Up1; maybe outside of CHR ?; we could also add a choice point here
   clpfd_interface:post_constraint2((Low1 #=< Up1) #=> ((Low2 #=< Low1) #/\ (Up1 #=< Up2)),Posted),
   (Posted==true -> true ;  interval_subset_aux(Low1,Up1,Low2,Up2)).

:- block interval_subset_aux(-,?,?,?), interval_subset_aux(?,-,?,?).
interval_subset_aux(Low1,Up1,_,_) :- safe_less_than_with_inf(Up1,Low1). %Set 1 is empty.
interval_subset_aux(Low1,Up1,Low2,Up2) :-
    safe_less_than_equal_with_inf(Low1,Up1), % Set 1 is not empty
    safe_less_than_equal_with_inf_clpfd(Low2,Low1), safe_less_than_equal_with_inf_clpfd(Up1,Up2). % may call CLPFD

% a version of safe_less_than which allows minus_inf and inf, but only if those terms appear straightaway at the first call
% assumes any variable will only be bound to a number
safe_less_than_with_inf(X,Y) :- (X==Y ; X==inf ; Y==minus_inf), !,fail.
safe_less_than_with_inf(X,Y) :- (X==minus_inf ; Y==inf), !.
safe_less_than_with_inf(X,Y) :- safe_less_than(X,Y).

safe_less_than_with_inf_clpfd(X,Y) :- (X==Y ; X==inf ; Y==minus_inf), !,fail.
safe_less_than_with_inf_clpfd(X,Y) :- (X==minus_inf ; Y==inf), !.
safe_less_than_with_inf_clpfd(X,Y) :- less_than_direct(X,Y). % this can also call CLPFD

% a version of safe_less_than_equal which allows minus_inf and inf, but only if those terms appear straightaway at the first call
safe_less_than_equal_with_inf(X,Y) :- X==Y,!.
safe_less_than_equal_with_inf(X,Y) :- (X==inf ; Y==minus_inf), !,fail.
safe_less_than_equal_with_inf(X,Y) :- (X==minus_inf ; Y==inf), !.
safe_less_than_equal_with_inf(X,Y) :- safe_less_than_equal(X,Y).

safe_less_than_equal_with_inf_clpfd(X,Y) :- X==Y,!.
safe_less_than_equal_with_inf_clpfd(X,Y) :- (X==inf ; Y==minus_inf), !,fail.
safe_less_than_equal_with_inf_clpfd(X,Y) :- (X==minus_inf ; Y==inf), !.
safe_less_than_equal_with_inf_clpfd(X,Y) :- less_than_equal_direct(X,Y). % this can also call CLPFD

:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,2),interval(1,3))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,1),interval(1,2))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,1),interval(0,1))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(2,1),interval(33,34))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(3,1),interval(4,2))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(3,1),interval(2,1))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(1,2),interval(1,2))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(1,2),interval(2,3))).
:- assert_must_fail(kernel_objects:check_strict_subset_of_global_sets(interval(2,3),interval(1,2))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(0,1000),interval(0,inf))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(1,1000),interval(1,inf))).
:- assert_must_succeed(kernel_objects:check_strict_subset_of_global_sets(interval(-200,1000),interval(minus_inf,inf))).
% for any other term we have global enumerated or deferred sets: they cannot be a strict subset of each other
check_strict_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :-
	check_strict_subset_intervals(Low1,Up1,Low2,Up2).

check_strict_subset_intervals(Low1,Up1,Low2,Up2) :-
    safe_less_than_equal_with_inf_clpfd(Low2,Up2), % Low2..Up2 not empty
    check_strict_subset_intervals1(Low1,Up1,Low2,Up2).

check_strict_subset_intervals1(Low1,Up1,Low2,Up2) :- % we cannot have inf as term (yet) here
   %preferences:preference(use_clpfd_solver,true),
   (var(Low1) ; var(Up1)),
   finite_interval(Low1,Up1), finite_interval(Low2,Up2),
   %print(posting(Low1,Up1,Low2,Up2)),nl,
   !,
   clpfd_interface:post_constraint2((Low1 #=< Up1) #=> ((Low2 #=< Low1) #/\ (Up1 #=< Up2) #/\ (Low1 #\= Low2 #\/ Up1 #\= Up2)),Posted),
   (Posted==true -> true ; check_strict_subset_intervals2(Low1,Up1,Low2,Up2)).
check_strict_subset_intervals1(Low1,Up1,Low2,Up2) :- check_strict_subset_intervals2(Low1,Up1,Low2,Up2).

:- block check_strict_subset_intervals2(-,?,?,?),check_strict_subset_intervals2(?,-,?,?),
   check_strict_subset_intervals2(?,?,-,?).
check_strict_subset_intervals2(Low1,Up1,_,_) :- safe_less_than_with_inf(Up1,Low1). % interval 1 empty
check_strict_subset_intervals2(Low1,Up1,Low2,Up2) :-
    safe_less_than_equal_with_inf(Low1,Up1), % interval 1 not empty
    ( safe_less_than_with_inf(Low2,Low1), safe_less_than_equal_with_inf_clpfd(Up1,Up2)
      ;
      Low1=Low2,safe_less_than_with_inf_clpfd(Up1,Up2)
    ).

:- use_module(custom_explicit_sets,[is_definitely_maximal_set/1,singleton_set/2]).
:- use_module(kernel_tools,[ground_value_check/2]).

check_subset_of0(Set1,_Set2,_WF) :- Set1==[],!.
check_subset_of0(Set1,Set2,WF) :- Set2==[],
  %nonvar(Set2),Set2=[], %var(Set1),
  !,% print(checking_empty_set(Set1)),nl,
  empty_set_wf(Set1,WF). %  ,print(empty),nl.
check_subset_of0(_Set1,Set2,_WF) :-
  nonvar(Set2),is_definitely_maximal_set(Set2),!. %  , print(subset_maximal(_Set1,Set2)),nl.
  %singleton
check_subset_of0(Set1,Set2,_) :- Set1==Set2,!.
check_subset_of0(Set1,Set2,WF) :-  custom_explicit_sets:singleton_set(Set1,El),!,
   check_element_of_wf(El,Set2,WF).
check_subset_of0(Set1,Set2,WF) :-  % Note: two intervals are treated in check_subset_of_global_sets
    subset_of_explicit_set(Set1,Set2,Code,WF),!,
    % print(subset_explicit(Set1,Set2,Code)),nl,
    call(Code).
check_subset_of0(Set1,Set2,WF) :- nonvar(Set1),!,
    get_cardinality_powset_wait_flag(Set2,check_subset_of0,WF,_,LWF),
    expand_custom_set_to_list(Set1,ESet1,_,check_subset_of1),
    try_expand_and_convert_to_avl_unless_large(Set2,ESet2),
    % print(check_subset(ESet1,Set2,WF,LWF)),nl,
    % b_interpreter_components:observe_instantiation(ESet1,'ESet1',ESet1),
    check_subset_of2(ESet1,[],ESet2,WF,LWF,none).
check_subset_of0(Set1,Set2,WF) :-
    is_wait_flag_info(WF,wfx_no_enumeration),!,
    %print(no_enum(Set1)),nl,
    when(nonvar(Set1),check_subset_of0_lwf(Set1,Set2,WF,_LWF)).
check_subset_of0(Set1,Set2,WF) :-
    % DO we need LWF if Set1=avl_set(_) ??
    get_cardinality_powset_wait_flag(Set2,check_subset_of0,WF,_Card,LWF),
    % print(check_subset_of0(Set1,Set2,WF)),nl,
    ground_value_check(Set2,GS2),
    % TO DO: pass GS2 and use below + use block ?
    %print(subset_card(Card,Set2,WF)),nl, (Card==inf -> trace ; true),
    when((nonvar(Set1);(nonvar(LWF),nonvar(GS2))),check_subset_of0_lwf(Set1,Set2,WF,LWF)).

:- use_module(custom_explicit_sets,[is_infinite_or_very_large_explicit_set/2]).

check_subset_of0_lwf(Set1,_Set2,_WF,_LWF) :- Set1==[],!.
%check_subset_of0_lwf(Set1,Set2,WF,_LWF) :- Set2==[],!, % can never trigger as Set2 was already nonvar
%   empty_set_wf(Set1,WF).
check_subset_of0_lwf(Set1,Set2,WF,_LWF) :-  custom_explicit_sets:singleton_set(Set1,El),!,
   check_element_of_wf(El,Set2,WF).
check_subset_of0_lwf(Set1,Set2,WF,_LWF) :-  % Note: two intervals are treated in check_subset_of_global_sets
    nonvar(Set1), % otherwise we have already checked this code above
    subset_of_explicit_set(Set1,Set2,Code,WF),!,
    % print(subset_explicit(Set1,Set2,Code)),nl,
    call(Code).
check_subset_of0_lwf(Set1,Set2,WF,LWF) :-
    (nonvar(Set1) ; nonvar(Set2),dont_expand_this_explicit_set(Set2)),
    !,
    expand_custom_set_to_list(Set1,ESet1,_,check_subset_of1),
    try_expand_and_convert_to_avl_unless_large(Set2,ESet2),
    % print(check_subset(ESet1,Set2,WF,LWF)),nl,
    % b_interpreter_components:observe_instantiation(ESet1,'ESet1',ESet1),
    check_subset_of2(ESet1,[],ESet2,WF,LWF,none).
check_subset_of0_lwf(Set1,Set2,_WF,_LWF) :-
     expand_custom_set_to_list(Set2,ESet2,_,check_subset_of0_lwf), % Set2 is ground
     % THIS WILL ENUMERATE, for something like dom(f) <: SET this is problematic, as information cannot be used
     % hence we use wfx_no_enumeration above
     non_free(Set1), % as we now enumerate Set1 in a specific order ESet2
     gen_subsets(Set1,ESet2).

:- block check_subset_of2(-,?,?,?,-, ?).
check_subset_of2([],_SoFar,_Set2,_WF,_LWF,_Last).
% :- print_tabs, print('SOLUTION '),  hashing:my_term_hash((_SoFar,_Set2),Hash), print(Hash),nl, (Hash = 30707821301826039 -> trace ; true).
check_subset_of2(HT,SoFar,Set2,WF,LWF,Last) :-  %print(chk(H,Set2,Last)),nl,
      (var(HT),Set2 = avl_set(AVL)
        -> % the value is chosen by the enumerator
           %trace_point(check_subset_of2(HT,SoFar,Set2,WF,LWF,Last)),
           %print_tabs,print(' --> enum '), translate:print_bvalue(SoFar),nl,
           custom_explicit_sets:safe_avl_member(H,AVL),
           % this forces H to be ground; if Last /= none then it will be ground
           (Last==none -> true ; Last @< H),
           % TO DO: we could write a safe_avl_member_greater_than(H,Last,AVL)
           %print_tabs,translate:print_bvalue(H),nl, %(H==avl_set(node([],true,0,empty,empty)) -> trace ; true),
           not_element_of_wf(H,SoFar,WF),
           NewLast=H,
           HT = [H|T]
        ;  % the value may have been chosen by somebody else or will not be enumerated in order below
           %trace_point(normal_checksubset(HT,SoFar,Set2,WF,LWF,Last)),
           HT = [H|T],
           not_element_of_wf(H,SoFar,WF),
           check_element_of_wf_lwf(H,Set2,WF,LWF),
           %check_element_of_wf(H,Set2,WF),

           NewLast = Last
      ),
      %print(check(H,Set2,WF,LWF,Last)),nl,
      check_subset_of3(H,T,SoFar,Set2,WF,LWF,NewLast).

% TO DO: write specific subsets code for avl_set(Set2) + try expand when becomes ground; merge with enumerate_tight_set ,...
% TO DO: ensure that it also works with global_set(T) instead of avl_set(_) or with interval closures


:- block check_subset_of3(?,-,-,?,?,-,?), check_subset_of3(?,-,?,-,?,-,?), check_subset_of3(?,-,-,-,?,?,?).
check_subset_of3(_,T,_,_Set2,_WF,_LWF,_) :- T==[],!.
check_subset_of3(H,T,SoFar,Set2,WF,LWF,Last) :- var(T),!,
    % Sofar, Set2 and LWF must be set
    when((nonvar(T);(ground(Set2),ground(H),ground(SoFar))),
                      (T==[] -> true
                       ; add_new_element_wf(H,SoFar,SoFar2,WF),  %SoFar2 = [H|SoFar],
                         check_subset_of2(T,SoFar2,Set2,WF,LWF,Last))).
check_subset_of3(H,T,SoFar,Set2,WF,LWF,Last) :-
    % T must be set and not equal to []
    T = [H2|T2],
	add_new_element_wf(H,SoFar,SoFar2,WF),  %SoFar2 = [H|SoFar],
    %check_subset_of2(T,SoFar2,Set2,WF,LWF))),
    check_element_of_wf(H2,Set2,WF),
    not_element_of_wf(H2,SoFar2,WF),
    check_subset_of3(H2,T2,SoFar2,Set2,WF,LWF,Last).


:- block gen_subsets(?,-).
gen_subsets([],_).
gen_subsets([H|T],Set) :- %print(gen_subsets(H,T,Set)),nl,
   ordered_delete(DH,Set,NewSet),
   equal_object(DH,H,gen_subsets), % it is important that [H|T] remains uninstantiated, hence we should mark arg1 as non_free before calling gen_subsets
   gen_subsets(T,NewSet).

ordered_delete(H,[H|T],T).
ordered_delete(H,[_|T],R) :- ordered_delete(H,T,R).

gen_strict_subsets(T,[_H2|T2]) :- gen_subsets(T,T2).
gen_strict_subsets([H|T],[H2|T2]) :-
    equal_object(H,H2,gen_strict_subsets),
    gen_strict_subsets(T,T2).

:- assert_must_succeed(exhaustive_kernel_check_wf(check_finite_non_empty_subset_of_wf([int(1),int(5)], [int(2),int(5),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(check_finite_non_empty_subset_of_wf([int(1),int(5)], [int(5),int(1)],WF),WF)).
check_finite_non_empty_subset_of_wf(Set1,Set2,WF) :-
   check_non_empty_subset_of_wf(Set1,Set2,WF),
   is_finite_set_wf(Set1,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(check_non_empty_subset_of_wf([int(1),int(5)], [int(2),int(5),int(1),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(check_non_empty_subset_of_wf([int(2)], [int(5),int(1)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wfdet(check_non_empty_subset_of_wf([], [int(1)],WF),WF)).

check_non_empty_subset_of_wf(S1,S2,WF) :- not_empty_set_wf(S1,WF),
   check_subset_of_wf(S1,S2,WF).

:- assert_must_succeed(exhaustive_kernel_succeed_check(not_subset_of([int(1),int(2),int(5)], [int(2),int(4),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1),int(3)]))).
:- assert_must_succeed((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name')], X=[fd(2,'Name')])).
:- assert_must_succeed((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name')], X=[fd(1,'Name'),fd(2,'Name')])).
:- assert_must_fail((kernel_objects:not_subset_of(Y,X), Y = [fd(1,'Name'),fd(3,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_subset_of(Y,X), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_subset_of(global_set('NAT'),global_set('NAT')))).
:- assert_must_succeed((kernel_objects:not_subset_of(global_set('NAT'),global_set('NAT1')))).


not_subset_of(Set1,Set2) :- init_wait_flags(WF),
  not_subset_of_wf(Set1,Set2,WF),
  ground_wait_flags(WF).

:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf([int(1),int(2),int(5)], [int(2),int(4),int(1),int(3)],_WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf(global_set('NATURAL'), global_set('INTEGER'),_WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf(global_set('INTEGER'), global_set('INTEGER'),_WF))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_finite_subset_of_wf([int(1)], [],_WF))).

:- block not_finite_subset_of_wf(-,?,?).
not_finite_subset_of_wf(Set1,Set2,WF) :- test_finite_set_wf(Set1,Finite,WF),
   not_finite_subset_of_wf_aux(Finite,Set1,Set2,WF).
:- block not_finite_subset_of_wf_aux(-,?,?,?).
not_finite_subset_of_wf_aux(pred_false,_Set1,_Set2,_WF).
not_finite_subset_of_wf_aux(pred_true,Set1,Set2,WF) :- not_subset_of_wf(Set1,Set2,WF).

:- block not_subset_of_wf(-,?,?).
not_subset_of_wf(Set1,Set2,WF) :-  % print_quoted(not_subset_of_wf(Set1,Set2)),nl,trace,
   (both_global_sets(Set1,Set2,G1,G2) % also catches intervals
     ->  check_not_subset_of_global_sets(G1,G2)
      ;  not_subset_of_wf1(Set1,Set2,WF)
   ).
not_subset_of_wf1(_Set1,Set2,_WF) :-
   nonvar(Set2), is_definitely_maximal_set(Set2),!,fail.
not_subset_of_wf1(Set1,Set2,WF) :- Set2==[],!, not_empty_set_wf(Set1,WF).
not_subset_of_wf1(Set1,Set2,_WF) :- Set1==Set2,!, fail.
not_subset_of_wf1(Set1,Set2,WF) :-  custom_explicit_sets:singleton_set(Set1,El),!,
   not_element_of_wf(El,Set2,WF).
not_subset_of_wf1(Set1,Set2,WF) :- not_subset_of_explicit_set(Set1,Set2,Code,WF),!,
   call(Code).
not_subset_of_wf1(Set1,Set2,WF) :-
  expand_custom_set_to_list(Set1,ESet1,_,not_subset_of_wf1),
  % print_quoted(not_subset_of2(ESet1,Set2,WF)),nl,
  not_subset_of2(ESet1,Set2,WF).


:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(0,2),interval(1,3))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(1,2),interval(0,-1))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(1,2),interval(4,3))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(2,4),interval(1,3))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(1,9000),interval(2,9999))).
:- assert_must_succeed((kernel_objects:check_not_subset_of_global_sets(interval(X2,X4),interval(1,3)),
 X2=2, X4=4)).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(2,4),interval(1,4))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(2,4),interval(2,4))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(2,4),interval(0,10))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(0,2),interval(1,inf))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(-1,2),interval(0,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(1,2),interval(1,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(0,2),interval(0,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(-1,2),interval(minus_inf,inf))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(0,inf),interval(1,inf))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(minus_inf,inf),interval(1,inf))).
:- assert_must_succeed(kernel_objects:check_not_subset_of_global_sets(interval(minus_inf,inf),interval(0,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(1,inf),interval(minus_inf,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(1,inf),interval(1,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(1,inf),interval(0,inf))).
:- assert_must_fail(kernel_objects:check_not_subset_of_global_sets(interval(0,inf),interval(0,inf))).

:- block check_not_subset_of_global_sets(-,?), check_not_subset_of_global_sets(?,-).
check_not_subset_of_global_sets(interval(Low1,Up1),G2) :- !,
   safe_less_than_equal_with_inf_clpfd(Low1,Up1), % Set 1 is not empty; otherwise it will always be a subset
   not_subset_interval_gs_aux(G2,Low1,Up1).
check_not_subset_of_global_sets(G1,G2) :-
    \+ check_subset_of_global_sets(G1,G2).

:- use_module(library(clpfd)).
not_subset_interval_gs_aux(interval(Low2,Up2),Low1,Up1) :-
   finite_interval(Low1,Up1), finite_interval(Low2,Up2),
   !,
   % post_constraint2((Low1 #<Low2 #\/ Up1 #> Up2 #\/ Up2 #< Low1),Posted),  %%  X #<100 #\/ X#<0. does not constraint X ! but X #<max(100,0) does
   post_constraint2((Low1 #<Low2 #\/ Up2 #< max(Up1,Low1)),Posted),
   (Posted==true -> true ;  not_interval_subset(Low1,Up1,Low2,Up2)).
not_subset_interval_gs_aux(interval(Low2,Up2),Low1,Up1) :- !, not_interval_subset(Low1,Up1,Low2,Up2).
not_subset_interval_gs_aux(GS2,Low1,Up1) :-
   when((nonvar(Low1),nonvar(Up1)), \+ check_subset_of_global_sets(interval(Low1,Up1),GS2)).

not_interval_subset(Val1,Up1,Low2,Up2) :- var(Val1), Val1==Up1,  %print(not_in(Val1,Up1,Low2,Up2)),nl,
    !, % better propagation for singleton set
    (Up2==inf -> Low2\==minus_inf, less_than_direct(Val1,Low2)
      ; Low2=minus_inf -> less_than_direct(Low2,Val1)
      ;  not_in_nat_range(int(Val1),int(Low2),int(Up2))).
not_interval_subset(Low1,Up1,Low2,Up2) :- not_interval_subset_block(Low1,Up1,Low2,Up2).
:- block not_interval_subset_block(-,?,?,?), not_interval_subset_block(?,-,?,?),
         not_interval_subset_block(?,?,-,?), not_interval_subset_block(?,?,?,-).
not_interval_subset_block(Low1,Up1,Low2,Up2) :- % this could be decided earlier, e.g. 1..n /<: 1..inf is false
         \+ interval_subset(Low1,Up1,Low2,Up2).


:- block not_subset_of2(-,?,?).
%not_subset_of2([],_,_WF) :- fail.
not_subset_of2([H|T],Set2,WF) :-
   membership_test_wf(Set2,H,MemRes,WF),
   not_subset_of3(MemRes,T,Set2,WF).

:- block not_subset_of3(-,?,?,?).
not_subset_of3(pred_false,_T,_Set2,_WF).
not_subset_of3(pred_true,T,Set2,WF)   :- not_subset_of2(T,Set2,WF).



:- assert_must_succeed(exhaustive_kernel_check_wf(not_both_subset_of([int(1),int(2),int(5)], []
,[int(2),int(4),int(1),int(3)],[],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(not_both_subset_of([int(1),int(2),int(5)], [int(3)],
    [int(2),int(5),int(1),int(3)],[int(1),int(4)],WF),WF)).

not_both_subset_of(Set1A,Set1B, Set2A,Set2B, WF) :-
  kernel_equality:subset_test(Set1A,Set2A,Result,WF), % not yet implemented ! % TODO ! -> sub_set,equal,super_set
  not_both_subset_of_aux(Result,Set1B,Set2B,WF).

:- block not_both_subset_of_aux(-,?,?,?).
not_both_subset_of_aux(pred_false,_Set1B,_Set2B,_WF).
not_both_subset_of_aux(pred_true,Set1B,Set2B,WF) :-
    not_subset_of_wf(Set1B,Set2B,WF).

/***********************************/
/* not_strict_subset_of(Set1,Set2) */
/*         Set1 /<<: Set2         */
/**********************************/


:- assert_must_succeed(exhaustive_kernel_succeed_check(not_strict_subset_of([int(1),int(2),int(5)], [int(2),int(4),int(1),int(3)]))).
:- assert_must_succeed(exhaustive_kernel_succeed_check(not_strict_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1)]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_strict_subset_of([int(1),int(2),int(5)], [int(2),int(5),int(1),int(3)]))).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [], X=[int(2),int(1)])).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [[int(1),int(2)]], X=[[int(2)],[int(2),int(1)]])).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [fd(1,'Name')], X=global_set('Name'))).
:- assert_must_fail((kernel_objects:not_strict_subset_of(Y,X), Y = [fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name'),fd(2,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [fd(1,'Name'),fd(3,'Name')], X=global_set('Name'))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(1)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(1),int(2)], X=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(2)], X=[int(2)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [int(2)], X=[int(1)])).
:- assert_must_succeed((kernel_objects:not_strict_subset_of(X,Y), Y = [], X=[int(1)])).

not_strict_subset_of(Set1,Set2) :-
  chr_not_subset_strict(Set1,Set2),
  init_wait_flags(WF),
  not_strict_subset_of_wf(Set1,Set2,WF),
  ground_wait_flags(WF).

:- block not_strict_subset_of_wf(-,?,?),not_strict_subset_of_wf(?,-,?).
not_strict_subset_of_wf(Set1,Set2,WF) :- %print_quoted(check_subset_of(Set1,Set2)),nl,
   (both_global_sets(Set1,Set2,G1,G2)
     -> %print(check_not_strict_subset_of_global_sets(G1,G2)),nl,
         not_strict_subset_of_global_sets(G1,G2)
      ;  not_strict_subset_of_wf1(Set1,Set2,WF)
   ).
not_strict_subset_of_wf1(Set1,Set2,WF) :- not_subset_of_explicit_set(Set1,Set2,Code,WF),!,
   equality_objects_wf(Set1,Set2,EqRes,WF),
   not_strict_eq_check(EqRes,Code).
not_strict_subset_of_wf1(Set1,Set2,WF) :-
   % OLD VERSION: not_subset_of(Set1,Set2) ; check_equal_object(Set1,Set2).
    expand_custom_set_to_list(Set1,ESet1,_,not_strict_subset_of_wf1),
    (nonvar(Set2),is_infinite_explicit_set(Set2) -> Inf=infinite ; Inf=unknown),
    not_strict_subset_of2(ESet1,Set2,Inf,WF).

:- block not_strict_eq_check(-,?).
not_strict_eq_check(pred_true,_). % if equal then not strict subset is true
not_strict_eq_check(pred_false,Code) :- call(Code). % check if not subset

:- block not_strict_subset_of2(-,?,?,?).
not_strict_subset_of2([],R,_,WF) :- empty_set_wf(R,WF).
not_strict_subset_of2([H|T],Set2,Inf,WF) :-
    membership_test_wf(Set2,H,MemRes,WF),
    not_strict_subset_of3(MemRes,H,T,Set2,Inf,WF).

:- block not_strict_subset_of3(-,?,?,?,?,?).
not_strict_subset_of3(pred_false,_H,_T,_Set2,_,_WF).
not_strict_subset_of3(pred_true,H,T,Set2,Inf,WF) :-
      (Inf=infinite
       -> RS2=Set2 % Set1 is finite; we just have to check that all elements are in Set2 and we have a strict subset
       ; remove_element_wf(H,Set2,RS2,WF)),
      not_strict_subset_of2(T,RS2,Inf,WF).


:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(0,2),interval(1,3))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(1,2),interval(0,-1))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(1,2),interval(4,3))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(2,4),interval(1,3))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(1,9000),interval(2,9999))).
:- assert_must_succeed((kernel_objects:not_strict_subset_of_global_sets(interval(X2,X4),interval(1,3)),
 X2=2, X4=4)).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(2,4),interval(1,4))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(2,4),interval(2,4))).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(2,4),interval(0,10))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(0,2),interval(1,inf))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(-1,2),interval(0,inf))).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(1,2),interval(1,inf))).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(0,2),interval(0,inf))).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(-1,2),interval(minus_inf,inf))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(0,inf),interval(1,inf))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(minus_inf,inf),interval(1,inf))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(minus_inf,inf),interval(0,inf))).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(1,inf),interval(minus_inf,inf))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(1,inf),interval(1,inf))).
:- assert_must_fail(kernel_objects:not_strict_subset_of_global_sets(interval(1,inf),interval(0,inf))).
:- assert_must_succeed(kernel_objects:not_strict_subset_of_global_sets(interval(0,inf),interval(0,inf))).

:- block not_strict_subset_of_global_sets(-,?), not_strict_subset_of_global_sets(?,-).
not_strict_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :- !,
   % Note: if Low2>Up2 then nothing is a strict subset of the empty set, i.e., everything is not a strict subset
   (finite_interval(Low1,Up1), finite_interval(Low2,Up2)
    -> clpfd_interface:post_constraint2(((Low2 #=< Up2) #=> (Low1 #=< Up1 #/\ ((Low2 #> Low1) #\/ (Up1 #> Up2) #\/ ((Low1 #= Low2 #/\ Up1 #= Up2))))),Posted)
     ; Posted=false),
   (Posted==true -> true ;  not_strict_subset_intervals(Low1,Up1,Low2,Up2)).
not_strict_subset_of_global_sets(G1,G2) :-
   when((ground(G1),ground(G2)), \+check_strict_subset_of_global_sets(G1,G2)).

:- block not_strict_subset_intervals(?,?,-,?), not_strict_subset_intervals(?,?,?,-).
% Instead of blocking on Low2,Up2 we could post bigger constraint (Low2 <= Up2  =>  (Low1 <= Up1 /\ ....
not_strict_subset_intervals(_Low1,_Up1,Low2,Up2) :- safe_less_than_with_inf(Up2,Low2),!.
not_strict_subset_intervals(Low1,Up1,Low2,Up2) :-
   safe_less_than_equal_with_inf_clpfd(Low1,Up1), % if Low1..Up1 is empty then it would be a strict subset
   not_check_strict_subset_intervals2(Low1,Up1,Low2,Up2).
:- block not_check_strict_subset_intervals2(-,?,?,?),not_check_strict_subset_intervals2(?,-,?,?),
   not_check_strict_subset_intervals2(?,?,-,?).
not_check_strict_subset_intervals2(Low1,Up1,Low2,Up2) :- \+ check_strict_subset_intervals2(Low1,Up1,Low2,Up2).


/* Set1 /: FIN1(Set2) */
:- assert_must_succeed((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X=[int(1)], Y=[int(1),int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[])).
:- assert_must_fail((kernel_objects:not_non_empty_finite_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(1)])).

:- block not_non_empty_finite_subset_of_wf(-,?,?).
not_non_empty_finite_subset_of_wf(Set1,Set2,WF) :- test_finite_set_wf(Set1,Finite,WF),
   not_non_empty_finite_subset_of_aux(Finite,Set1,Set2,WF).
:- block not_non_empty_finite_subset_of_aux(-,?,?,?).
not_non_empty_finite_subset_of_aux(pred_false,_Set1,_Set2,_WF).
not_non_empty_finite_subset_of_aux(pred_true,Set1,Set2,WF) :- not_non_empty_subset_of_wf(Set1,Set2,WF).

/* Set1 /: POW1(Set2) */
:- assert_must_succeed(exhaustive_kernel_check_wf(not_non_empty_subset_of_wf([int(1)], [int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(not_non_empty_subset_of_wf([int(2)], [int(2),int(3)],WF),WF)).
:- assert_must_succeed((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X=[int(1)], Y=[int(1),int(2)])).
:- assert_must_succeed((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[])).
:- assert_must_fail((kernel_objects:not_non_empty_subset_of_wf(Y,X,_WF), X = [int(1)], Y=[int(1)])).

:- block not_non_empty_subset_of_wf(-,-,?).
not_non_empty_subset_of_wf(Set1,Set2,WF) :- % Maybe introduce binary choice point ?
   empty_set_wf(Set1,WF) ;
   not_subset_of_wf(Set1,Set2,WF).


/* min, max */

:- assert_must_succeed(exhaustive_kernel_check(minimum_of_set([int(1)],int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(minimum_of_set([int(2),int(3),int(1)],int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(minimum_of_set([int(2),int(3),int(1)],int(2),unknown,_WF))).
:- assert_must_succeed((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(1)])).
:- assert_must_succeed((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_fail((kernel_objects:minimum_of_set(Y,X,unknown,_WF), X = int(2), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_abort_wf(kernel_objects:minimum_of_set([],_R,unknown,WF),WF).
%:- must_succeed(kernel_waitflags:assert_must_abort2_wf(kernel_objects:minimum_of_set([],_R,WF),WF)).

:- block minimum_of_set_extension_list(-,?,?,?).
minimum_of_set_extension_list(ListOfValues,int(Min),Span,WF) :-
   minimum_of_set2(ListOfValues,Min,Span,WF).

:- block minimum_of_set(-,?,?,?).
minimum_of_set(Set1,Res,_Span,WF) :- is_custom_explicit_set(Set1,minimum_of_set),
    min_of_explicit_set(Set1,Min), !,
 %%print_term_summary(explicit_set_min(Set1,Min)), %%
    equal_object_wf(Min,Res,minimum_of_set,WF).
minimum_of_set(Set1,int(Min),Span,WF) :- expand_custom_set_to_list(Set1,ESet1,_,minimum_of_set),
   minimum_of_set2(ESet1,Min,Span,WF).
:- block minimum_of_set2(-,?,?,?).
minimum_of_set2([],Res,Span,WF) :-
   add_wd_error_set_result('min applied to empty set','',Res,int(0),Span,WF).
minimum_of_set2([int(N)|T],Min,_,_) :- clpfd_geq2(N,Min,_),minimum_of_set3(T,N,Min,[N]).

:- if((current_prolog_flag(version_data,sicstus(4,X,Y,_,_)),(X>2;X=2,Y>0))). % we are 4.2.1 or later
:- block minimum_of_set3(-,?,?,?). % with CLPFD: makes sense to also unfold if Min Variable; hence no longer block on : minimum_of_set3(?,-,-).
minimum_of_set3([],MinSoFar,MinSoFar,ListOfValues) :-
   (var(MinSoFar) -> clpfd_minimum(MinSoFar,ListOfValues) ; true). % THIS CRASHES SICSTUS 4.2.0
:- else.
:- block minimum_of_set3(-,?,?,?).
minimum_of_set3([],MinSoFar,MinSoFar,_ListOfValues).
:- endif.
minimum_of_set3([int(M)|T],MinSoFar,Min,ListOfValues) :- clpfd_geq2(M,Min,_),
   minimum(M,MinSoFar,NewMinSoFar),
   minimum_of_set3(T,NewMinSoFar,Min,[M|ListOfValues]).


:- block minimum(-,?,?), minimum(?,-,?).
minimum(M1,M2,Min) :- M1<M2 -> Min=M1 ; Min=M2.

:- assert_must_succeed(exhaustive_kernel_check(maximum_of_set([int(1)],int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(maximum_of_set([int(2),int(3),int(1)],int(3),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(maximum_of_set([int(2),int(3),int(1)],int(2),unknown,_WF))).
:- assert_must_succeed((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(1), Y=[int(1)])).
:- assert_must_succeed((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(2), Y=[int(2),int(1)])).
:- assert_must_succeed((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(3), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_fail((kernel_objects:maximum_of_set(Y,X,unknown,_WF), X = int(2), Y=[int(1),int(2),int(1),int(3)])).
:- assert_must_fail((preferences:preference(use_clpfd_solver,true),
      kernel_objects:maximum_of_set([int(X),int(_Y)],int(3),unknown,_WF), X = 4)). % in CLPFD modus
:- assert_must_fail((preferences:preference(use_clpfd_solver,true),
      kernel_objects:maximum_of_set([int(_),int(X)],int(3),unknown,_WF), X = 4)).% in CLPFD modus
:- assert_must_abort_wf(kernel_objects:maximum_of_set([],_R,unknown,WF),WF).

:- block maximum_of_set_extension_list(-,?,?,?).
maximum_of_set_extension_list(ListOfValues,int(Max),Span,WF) :-
   maximum_of_set2(ListOfValues,Max,Span,WF).

:- block maximum_of_set(-,?,?,?).
maximum_of_set(Set1,Res,_Span,WF) :- %print(max(Set1,Res)),nl,
    is_custom_explicit_set(Set1,maximum_of_set),
    max_of_explicit_set(Set1,Max), !,
% print_term_summary(explicit_set_max(Set1,Max)), %%
    equal_object_wf(Max,Res,maximum_of_set,WF).
maximum_of_set(Set1,int(Max),Span,WF) :-
   expand_custom_set_to_list(Set1,ESet1,_,maximum_of_set),
   maximum_of_set2(ESet1,Max,Span,WF).
:- block maximum_of_set2(-,?,?,?).
maximum_of_set2([],Res,Span,WF) :-
  add_wd_error_set_result('max applied to empty set','',Res,int(0),Span,WF). %preferences:get_preference(maxint,R))). %R=abort(maximum_of_empty_set))).
maximum_of_set2([int(N)|T],Max,_Span,_) :- clpfd_geq2(Max,N,_),
   maximum_of_set3(T,N,Max,[N]).

:- if((current_prolog_flag(version_data,sicstus(4,X,Y,_,_)),(X>2;X=2,Y>0))). % we are 4.2.1 or later
:- block maximum_of_set3(-,?,?,?). % with CLPFD: makes sense to also unfold if Max Variable; hence no longer block on : maximum_of_set3(?,-,-).
maximum_of_set3([],MaxSoFar,MaxSoFar,ListOfValues) :-
   (var(MaxSoFar) -> clpfd_maximum(MaxSoFar,ListOfValues) ; true). % THIS CRASHES SICSTUS 4.2.0
:- else.
:- block maximum_of_set3(-,?,?,?).
maximum_of_set3([],MaxSoFar,MaxSoFar,_ListOfValues).
:- endif.
maximum_of_set3([int(M)|T],MaxSoFar,Max,ListOfValues) :- clpfd_geq2(Max,M,_),
   maximum(M,MaxSoFar,NewMaxSoFar),
   maximum_of_set3(T,NewMaxSoFar,Max,[M|ListOfValues]).

:- block maximum(-,?,?), maximum(?,-,?).
maximum(M1,M2,Max) :- M1>M2 -> Max=M1 ; Max=M2.


:- assert_must_succeed((cardinality_of_set_extension_list([fd(1,'Name')],R,_WF), R = int(1))).
:- assert_must_succeed((cardinality_of_set_extension_list([int(X),int(Y)],int(1),_WF), X=22, Y==22)).

cardinality_of_set_extension_list(List,int(Card),WF) :-
  length(List,MaxCard), less_than_equal_direct(Card,MaxCard),
  cardinality_of_set_extension_list2(List,[],0,MaxCard,Card,WF).

:- block cardinality_of_set_extension_list2(-,?,?,?,?,?).
cardinality_of_set_extension_list2([],_,AccSz,_MaxCard,Res,_WF) :- Res=AccSz.
cardinality_of_set_extension_list2([H|T],Acc,AccSz,MaxCard,Res,WF) :-
   % print(card_set_ext([H|T],Acc,AccSz,MaxCard,Res)),nl,
   membership_test_wf(Acc,H,MemRes,WF),
   (MaxCard==Res -> /* only solution is for H to be not in Acc */ MemRes=pred_false
    ; AccSz==Res -> /* only solution is for H to be in Acc */ MemRes=pred_true
    ; (var(Res),var(MemRes)) -> kernel_equality:equality_int(MaxCard,Res,EqMaxC),prop_if_pred_true(EqMaxC,MemRes,pred_false),
                                kernel_equality:equality_int(AccSz,Res,EqAccSz),prop_if_pred_true(EqAccSz,MemRes,pred_true)
    ; true),
   cardinality_of_set_extension_list3(MemRes,H,T,Acc,AccSz,MaxCard,Res,WF).

:- block prop_if_pred_true(-,?,?).
prop_if_pred_true(pred_true,X,X).
prop_if_pred_true(pred_false,_,_).

:- block cardinality_of_set_extension_list3(-,?,?,?,?,?,?,?).
cardinality_of_set_extension_list3(pred_true,_,T,Acc,AccSz,MaxCard,Res,WF) :-
   % H is a member of Acc, do not increase Acc nor AccSz; however MaxCard now decreases
   less_than_direct(Res,MaxCard), M1 is MaxCard-1,
   cardinality_of_set_extension_list2(T,Acc,AccSz,M1,Res,WF).
cardinality_of_set_extension_list3(pred_false,H,T,Acc,AccSz,MaxCard,Res,WF) :-
   A1 is AccSz+1, less_than_equal_direct(A1,Res),
   cardinality_of_set_extension_list2(T,[H|Acc],A1,MaxCard,Res,WF).

:- assert_must_succeed(exhaustive_kernel_check(is_finite_set_wf([fd(1,'Name'),fd(2,'Name')],_WF))).
:- assert_must_succeed((is_finite_set_wf(Y,_WF), Y = [])).
:- assert_must_succeed((is_finite_set_wf(Y,_WF), Y = [int(1),int(2)])).
:- use_module(typing_tools,[contains_infinite_type/1]).
:- use_module(custom_explicit_sets,[card_for_specific_custom_set/3]).

is_finite_set_wf(Set,WF) :-  test_finite_set_wf(Set,pred_true,WF).

:- assert_must_succeed(exhaustive_kernel_fail_check(is_infinite_set_wf([fd(1,'Name'),fd(2,'Name')],_WF))).
:- assert_must_fail((is_infinite_set_wf(Y,_WF), Y = [int(1),int(2)])).

is_infinite_set_wf(Set,WF) :- test_finite_set_wf(Set,pred_false,WF).

test_finite_set(Set,PREDRES) :- test_finite_set_wf(Set,PREDRES,_WF). % _WF not used below.
:- block test_finite_set_wf(-,?,?).
%test_finite_set_wf(A,B,C) :- print(test_finite_set_wf(A,B,C)),nl,fail.
test_finite_set_wf([],X,_WF) :- !, X=pred_true.
test_finite_set_wf([_|T],X,WF) :- !, test_finite_set_wf(T,X,WF).  % what if Tail contains closure ??
test_finite_set_wf(avl_set(_),X,_WF) :- !, X=pred_true.
test_finite_set_wf(closure(_P,T,_B),X,_WF) :- \+ contains_infinite_type(T), !, X=pred_true.
test_finite_set_wf(closure(P,T,B),X,WF) :- !, test_finite_closure(P,T,B,X,WF).
test_finite_set_wf(Set,X,WF) :-  /* also deals with global_set(_) */
   /* explicit_set_cardinality may trigger an enum warning */
   explicit_set_cardinality_wf(Set,Card,WF), set_finite_result(Card,Set,X).

:- use_module(bsyntaxtree,[is_a_disjunct/3]).
% we already check that contains_infinite_type above
test_finite_closure(P,T,B,X,WF) :- is_a_disjunct(B,D1,D2),!,
   test_finite_closure(P,T,D1,X1,WF),
   (X1=pred_true -> test_finite_closure(P,T,D2,X,WF)
     ; X = pred_false).
% TO DO: add is_closure1_value_closure
test_finite_closure(P,T,B,X,WF) :- when(ground(B), test_finite_closure_ground(P,T,B,X,WF)).

% first: we need to check all constructors such as POW, FIN, ... which card_for_specific_custom_set supports
% problem: if card becomes very large it is replaced by inf, which may give wrong results here (for card(.) we just get a spurious WD warning, here we may get wrong results)
test_finite_closure_ground(P,T,B,X,WF) :-
    is_powerset_closure(closure(P,T,B),_Type,Subset),
    % note: whether Type is fin, fin1, pow, or pow1 does not matter
    !,
    test_finite_set_wf(Subset,X,WF).
test_finite_closure_ground(P,T,B,X,WF) :-
    custom_explicit_sets:is_lambda_value_domain_closure(P,T,B, Subset,_Expr), !,
    %print(lambda(Subset)),nl,
    test_finite_set_wf(Subset,X,WF).
test_finite_closure_ground(P,T,B,X,WF) :-
    custom_explicit_sets:is_cartesian_product_closure(closure(P,T,B), A1,B2), !,
    test_finite_set_wf(A1,AX,WF),
    test_finite_set_wf(B2,BX,WF),
    %print(result_cart(AX,BX)),nl,
    test_finite_cartesian_product(AX,BX,A1,B2,X).
test_finite_closure_ground(Par,Typ,Body, X,_WF) :- %print(try(Par,Typ,Body)),nl,
    custom_explicit_sets:is_geq_leq_interval_closure(Par,Typ,Body,Low,Up), !,
    % print(geq_leq(Par,Low,Up,Body)),nl,
    custom_explicit_sets:card_of_interval_inf(Low,Up,Card),  %print(Card),nl,
    set_finite_result_no_warn(Card,X).
test_finite_closure_ground(P,T,B,X,WF) :-
     custom_explicit_sets:is_member_closure(P,T,B,_,SET), nonvar(SET),
     unary_member_closure_for_finite(SET,Check,SET1),
     !,
     (Check==finite -> test_finite_set_wf(SET1,X,WF) ; kernel_equality:eq_empty_set(SET1,X)).
% TO DO: catch other special cases :  relations, struct,...
test_finite_closure_ground(P,T,B,X,_WF) :-
    card_for_specific_custom_set(closure(P,T,B),Card,Code),!,
    call(Code), % TO DO: catch if we convert large integer due to overflow to inf !
    % maybe we can set / transmit a flag for is_overflowcheck ? overflow_float_pown ? factorial ?
    set_finite_result(Card,closure(P,T,B),X).
test_finite_closure_ground(P,T,B,X,WF) :- %print(try_expand2(P)),nl,
   on_enumeration_warning(expand_only_custom_closure_global(closure(P,T,B),Result,check),fail),
   !,
   %print(expanded2(P,Result)),nl,
   test_finite_set_wf(Result,X,WF).
test_finite_closure_ground(P,T,B,X,WF) :- X==pred_true, !,
        get_enumeration_finished_wait_flag(WF,AWF), % only add warning if indeed we find a solution
        finite_warning(AWF,P,T,B,is_finite_set_closure(P)).
test_finite_closure_ground(P,T,B,_X,_WF) :- !,
   finite_warning(now,P,T,B,test_finite_closure(P)),
   fail. % now we fail; used to be X=pred_true. % we assume set to be finite, but print a warning
      % we could set up the closure and do a deterministic phase: if it fails or all variables become bounded, then it is finite

unary_member_closure_for_finite(seq(b(value(SET1),_,_)),empty,SET1). % finite if SET1 is empty
unary_member_closure_for_finite(seq1(b(value(SET1),_,_)),empty,SET1).
unary_member_closure_for_finite(perm(b(value(SET1),_,_)),finite,SET1). % finite if SET1 is finite
unary_member_closure_for_finite(iseq(b(value(SET1),_,_)),finite,SET1).
unary_member_closure_for_finite(iseq1(b(value(SET1),_,_)),finite,SET1).
unary_member_closure_for_finite(identity(b(value(SET1),_,_)),finite,SET1).
% we could deal with POW/POW1... here

:- block test_finite_cartesian_product(-,?,?,?,?), test_finite_cartesian_product(?,-,?,?,?).
test_finite_cartesian_product(pred_true, pred_true, _,_,X) :- !, X=pred_true. % both finite
test_finite_cartesian_product(pred_false,pred_false,_,_,X) :- !, X=pred_false. % both infinite
test_finite_cartesian_product(pred_false,pred_true, _,B,X) :- !,
     kernel_equality:eq_empty_set(B,X). % only finite if B empty
test_finite_cartesian_product(pred_true, pred_false,A,_,X) :- !,
     kernel_equality:eq_empty_set(A,X). % only finite if B empty


:- block set_finite_result_no_warn(-,?).
set_finite_result_no_warn(inf,X) :- !,  X=pred_false.
set_finite_result_no_warn(_,pred_true).

:- block set_finite_result(-,?,?).
set_finite_result(inf,Set,X) :- !,
   (Set=closure(P,T,B) %,preferences:preference(disprover_mode,_true))
     -> finite_warning(now,P,T,B,test_finite_closure(P)) % we sometimes return inf for very large sets % TO DO: fix
      ; true),
    X=pred_false.
set_finite_result(_,_,pred_true).


:- assert_must_succeed((finite_cardinality_as_int(Y,int(X),_WF), Y = [fd(1,'Name'),fd(2,'Name')],X==2)).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int([int(2),int(4),int(1)],int(3)))).
:- assert_must_succeed((cardinality_as_int(Y,int(2)), Y = [fd(1,'Name'),fd(2,'Name')])).
:- assert_must_succeed((cardinality_as_int(Y,int(2)),
  nonvar(Y), Y = [H1|YY], nonvar(YY), YY=[H2], H1=int(0), H2=int(3) )).
:- assert_must_succeed((cardinality_as_int([A|Y],int(3)),
  nonvar(Y), Y = [B|YY], nonvar(YY), YY=[C], A=int(1),B=int(3),C=int(2) )).
:- assert_must_succeed((cardinality_as_int(Y,int(1)), Y = [fd(1,'Name')])).
:- assert_must_succeed((cardinality_as_int(Y,int(0)), Y = [])).
:- assert_must_succeed((cardinality_as_int(X,int(3)), equal_object(X,global_set('Name')))).
:- assert_must_fail((cardinality_as_int(Y,int(X)), Y = [fd(1,'Name'),fd(2,'Name')],dif(X,2))).
:- assert_must_succeed_any((preferences:preference(use_clpfd_solver,false) ;
    cardinality_as_int(S,int(C)), clpfd_interface:post_constraint('#>='(C,2)), nonvar(S),S=[_|T],nonvar(T))).
:- assert_must_succeed((cardinality_as_int([int(1)|avl_set(node(int(3),true,0,empty,empty))],int(2)))).
:- assert_must_succeed((cardinality_as_int([int(1)|avl_set(node(int(3),true,0,empty,empty))],X),X==int(2))).
% check that we deal with repeated elements, in case no other predicate sets up a list !
:- assert_must_fail((cardinality_as_int([int(1),int(1)],int(2)))).
:- assert_must_fail((cardinality_as_int([int(1),int(1)],_))).
:- assert_must_fail((cardinality_as_int(X,int(2)),X=[int(1),int(1)])).
:- assert_must_fail((cardinality_as_int([int(3)|avl_set(node(int(3),true,0,empty,empty))],_))).
:- assert_must_fail((cardinality_as_int([X|avl_set(node(int(3),true,0,empty,empty))],int(2)),X=int(3))).

% :- use_module(probsrc(kernel_cardinality)).
%finite_cardinality_as_int(Set,C,WF) :- preference(use_clpfd_solver,true), !, kernel_cardinality:finite_int_cardinality(Set,C,WF). % see also test 34 where this helps
finite_cardinality_as_int(Set,int(Card),WF) :-   % print(card(Set)),nl,trace,
  % if Card is already known we could give it to cardinality_as_int1 straightaway; but we would not detect certain WD-problems
    %(number(Card) -> CardValue=Card ; true),
    cardinality_as_int1(Set,Card,CardValue,WF), % print(card_value(CardValue,Card)),nl,
    % clpfd_domain(CardValue,Low,Up), print(card(CardValue,Low,Up,Set)),nl,trace,
    (clpfd_max_bounded(CardValue)
     -> Card=CardValue % we cannot have an infinite return value for CardValue
     ; (nonvar(Set),is_interval_closure(Set,_,_)) -> % we must have a finite interval; no need to guard against inf
        Card=CardValue  % TO DO: let cardinality_as_int1 return a flag whether set can be infinite
        % example i=2..x & card(i):10..2122110 & x > 2121000; see test 1625
     ; check_finite_card(CardValue,Card,Set,WF) % check that we have obtained a finite value; only propagate then
       % TO DO: maybe detect a few more cases where infinite return values are not possible; e.g., based on type
       % indeed: check_finite_card prevents propagation of CLPFD information
    ).

:- block check_finite_card(-,-,?,?).
% check before assigning that the result is not "inf" (otherwise we may trigger arithmetic co-routines before generating the error)
check_finite_card(CardValue,C,Set,WF) :- % print(check_finite_card(CardValue,C)),nl,
  (CardValue==inf -> add_wd_error('card applied to infinite (or very large) set: ',Set,WF)
    % if CardValue not yet instantiated we may want to delay the unification below for stronger wd-checking:
    % i.e., if find_abort_values preference is true
   ; C=CardValue).


cardinality_as_int(S,I) :- cardinality_as_int_wf(S,I,no_wf_available). % TO DO: remove this predicate ?
:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_data_validation_mode,true)).
:- block cardinality_as_int_wf(-,?,?). % avoid instantiating list skeletons; cause backtracking in unifications,...
:- else.
:- block cardinality_as_int_wf(-,-,?).
:- endif.
% can return inf !
cardinality_as_int_wf(Set,int(Card),WF) :-
   cardinality_as_int1(Set,Card,Card,WF).

cardinality_as_int1(Set,Card,ResCard,WF) :-
  %print(card1(Set,Card,ResCard)), tools_printing:print_var_integer(Card),
  (number(Card) -> cardinality_as_int1b(Set,Card,ResCard,WF) ;
   cardinality_as_int1b(Set,Card,ResCard,WF),
   (var(Set) ->
     (clpfd_domain(Card,Low,_Up), number(Low), Low>1,unbound_variable_for_cons(Set)
       % TO DO: also use this optimization later in cardinality_as_int2
       -> setup_ordered_list_skeleton(Low,Skel,open,WF), %print(open_skel_low(Low,Up,Skel)),nl,
          Skel=Set
       ; get_wait_flag(1,force_non_empty(Set,Card),WF,LWF),
         force_non_empty(Set,0,Card,LWF) % we could consider using guarding this with a waitflag with priority 1.1
      )
     ; true)
   ).
% tests 1418, 1419, 1628, 1776 require that cardinality_as_int1b be triggered quickly
:- block cardinality_as_int1b(-,-,?,?). % with this the self-check with post_constraint('#>='(C,2) fails
% cardinality_as_int1(Set, CardValue, ComputedCardValue) : CardValue should be unified with ComputedCardValue afterwards
cardinality_as_int1b(Set,Card,ResCard,WF) :-  % print(card1b(Card)),nl,
   %portray_waitflags(WF),nl,
   number(Card), unbound_variable_for_cons(Set),
   !, % we know the cardinality and the set is not yet bound; this improvement is tested in tests 1417, 1418
   %frozen(Set,F), print(frozen(Set,F)),nl,
   setup_ordered_list_skeleton(Card,Skel,closed,WF),
    %% print(skel(Card,Skel)),nl, %%
   (Card,Set) = (ResCard,Skel). % bypass equal_object: assign variable in one-go
cardinality_as_int1b(Set,Card,ResCard,WF) :- var(Card),!,
   cardinality_as_int2(Set,0,Card,ResCard,[],WF).
cardinality_as_int1b(Set,Card,ResCard,WF) :-  %print(card2_prio(Card,ResCard,WF)),nl,
    % Set is a variable but not unbound_variable_for_cons
    % Unifications can be very expensive when we set up long lists
    % Idea: multiply Card by a factor and delay instantiating; maybe we get a avl_set; see test 456
    Prio is Card*100,
    get_wait_flag(Prio,cardinality_as_int1(Set,Card),WF,LWF2),
    when((nonvar(Set) ; nonvar(LWF2)),
        cardinality_as_int2(Set,0,Card,ResCard,[],WF)).
   %force_non_empty(Set,0,Card,1). % we could consider using guarding this with a waitflag with priority 1.1

:- if(environ(prob_data_validation_mode,true)).
:- block cardinality_as_int2(-,?,?,?,?,?). % avoid instantiating list skeletons; cause backtracking in unifications,...
:- else.
:- block cardinality_as_int2(-,?,-,?,?,?).
:- endif.
cardinality_as_int2(X,C,Res,ResultValue,_,WF) :-   % print(card2(X,C,Res,ResultValue)),nl,
   C==Res,!,empty_set_wf(X,WF),ResultValue=Res. % avoid choice point below
cardinality_as_int2(X,C,Res,ResultValue,SoFar,WF) :- nonvar(X),  X \= [], X\= [_|_],!,
   (is_custom_explicit_set(X)
    ->  %print(explicit_set_card(X,C,ResultValue)),nl,
        explicit_set_cardinality_wf(X,ESC,WF), add_card(C,ESC,ResultValue),
        disjoint_sets(X,SoFar,WF)
     ;  add_error_fail(cardinality_as_int2,'First argument not set: ',cardinality_as_int2(X,C,Res))
   ).
cardinality_as_int2([],C,Res,ResultValue,_,_WF) :-  C=ResultValue, Res=ResultValue.
cardinality_as_int2([H|T],C,Res,ResultValue,SoFar,WF) :-
   C1 is C+1,
  % print(card(_H,T,C,Res)),nl,
  not_element_of_wf(H,SoFar,WF), % do we always need to check this ? relevant for test 1828
  add_new_element_wf(H,SoFar,SoFar2,WF),
  (ground(Res) -> safe_less_than_equal(cardinality_as_int2,C1,Res)
                  /* check consistency so far if cardinality provided */
                ; clpfd_geq(Res,C1,_) %,(ground(Res)->print(forced(Res)),nl;true)
  ),
  force_non_empty(T,C1,Res,1), % Use WF ?
  cardinality_as_int2(T,C1,Res,ResultValue,[H|SoFar2],WF).

% setup an list skeleton with ordering constraints to avoid duplicate solutions
setup_ordered_list_skeleton(0,R,Closed,_WF) :- !, (Closed=closed -> R=[] ; true).
setup_ordered_list_skeleton(N,[H|T],Closed,WF) :-
   all_different_wf([H|T],WF),
   N1 is N-1, setup_list_skel_aux(N1,H,T,Closed).


:- use_module(kernel_ordering,[ordered_value/2]).
%setup_list_skel_aux(0,_,R,Closed) :- !, (Closed=closed -> R=[] ; true). % if open: TO DO: impose ordering on rest using lazy_ordered_value ? done in next clause below
setup_list_skel_aux(0,Prev,R,Closed) :- !, (Closed=closed -> R=[] ; lazy_ordered_value(R,Prev)).
setup_list_skel_aux(N,Prev,[H|T],Closed) :-  ordered_value(Prev,H),
   N>0, N1 is N-1, setup_list_skel_aux(N1,H,T,Closed).

:- block lazy_ordered_value(-,?).
lazy_ordered_value([H|T],Prev) :- !, ordered_value(Prev,H), lazy_ordered_value(T,H).
lazy_ordered_value(_,_).


% TO DO: use clpfd all_different for integers !?
%  get_integer_list(Set,IntList), clpfd_alldifferent(IntList).
% ensure we have all different constraint in case ordered_value does not succeed in enforcing order!
all_different_wf(ListOfValues,WF) :-
   all_different2(ListOfValues,[],WF).
:- block all_different2(-,?,?).
all_different2([],_,_) :- !.
all_different2([H|T],SoFar,WF) :- !, all_different3(SoFar,H,WF), all_different2(T,[H|SoFar],WF).
all_different2(CS,SoFar,WF) :- is_custom_explicit_set(CS),
    disjoint_sets(CS,SoFar,WF). % already done above by cardinality_as_int2 ?
all_different3([],_,_).
all_different3([H|T],X,WF) :- not_equal_object_wf(H,X,WF), all_different3(T,X,WF).



force_non_empty(Set,CSoFar,TotalCard,LWF) :-
   var(Set), var(TotalCard),
   preference(data_validation_mode,false),!,
  % print(force_non_empty(CSoFar)),nl,
   clpfd_interface:post_constraint( (TotalCard#=CSoFar) #<=> EmptyR01),
   prop_non_empty(EmptyR01,Set,LWF).
force_non_empty(_,_,_,_).
:- block prop_non_empty(-,-,?).
prop_non_empty(_,X,_) :- nonvar(X),!. % do nothing; cardinality_as_int2 will be called anyway
prop_non_empty(0,X,LWF) :- /* X is var; first arg nonvar */ !, not_empty_set_lwf(X,LWF).
%prop_non_empty(1,X,_). % empty_set not really required: TotalCard is now instantiated; cardinality_as_int2 will get called
prop_non_empty(_,_,_).



:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int_for_wf(global_set('NATURAL'),inf))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_as_int_for_wf([],0))).
:- assert_must_succeed(exhaustive_kernel_check_opt(cardinality_as_int_for_wf([int(2)],1),
                        preferences:get_preference(convert_comprehension_sets_into_closures,false))). % in this case inf returned for closures
:- assert_must_succeed(exhaustive_kernel_check_opt(cardinality_as_int_for_wf([int(3),int(1),int(-1),int(100)],4),
                        preferences:get_preference(convert_comprehension_sets_into_closures,false))).
:- assert_must_succeed(exhaustive_kernel_fail_check_opt(cardinality_as_int_for_wf([int(3),int(1),int(-1),int(100)],1000),
                        preferences:get_preference(convert_comprehension_sets_into_closures,false))).
:- assert_must_succeed(exhaustive_kernel_fail_check_opt(cardinality_as_int_for_wf(global_set('NATURAL'),1000),
                        preferences:get_preference(convert_comprehension_sets_into_closures,false))).
% a simpler version without propagation to result; for waitflag priority computation or similar
% it may return inf for closures marked as symbolic !
cardinality_as_int_for_wf(Set,Card) :- cardinality_as_int_for_wf0(Set,0,Card).
:- block cardinality_as_int_for_wf0(-,?,-).
cardinality_as_int_for_wf0(X,C,Res) :-
  (nonvar(X) -> cardinality_as_int_for_wf1(X,C,Res)
    ; Res==inf -> cardinality_as_int_for_inf(X,C)
    ; cardinality_as_int_for_wf2(X,C,Res)).

:- block cardinality_as_int_for_inf(-,?).
cardinality_as_int_for_inf(X,C) :- cardinality_as_int_for_wf1(X,C,inf).

cardinality_as_int_for_wf1([],C,Res) :-  !,C=Res.
cardinality_as_int_for_wf1([_H|T],C,Res) :- !,C1 is C+1,
    cardinality_as_int_for_wf0(T,C1,Res).
cardinality_as_int_for_wf1(X,C,Res) :- is_custom_explicit_set(X),!,
    explicit_set_cardinality_for_wf(X,ESC), add_card(C,ESC,Res).
cardinality_as_int_for_wf1(X,C,Res) :-
   add_internal_error('First arg not set: ',cardinality_as_int_for_wf1(X,C,Res)),fail.

% first argument was var, third argument not inf hence third arg must be set
%cardinality_as_int_for_wf2([],C,C).
cardinality_as_int_for_wf2([],C,Res) :- (C==Res -> ! ; C=Res).
cardinality_as_int_for_wf2([_H|T],C,Res) :- C<Res, C1 is C+1,
    (var(T) -> cardinality_as_int_for_wf2(T,C1,Res) ; cardinality_as_int_for_wf1(T,C1,Res)).

% add two cardinalities together: can be number or inf
:- block add_card(-,-,?).
add_card(X,Y,R) :- X==0,!,R=Y.
add_card(X,Y,R) :- Y==0,!,R=X.
add_card(X,_,R) :- X==inf,!,R=inf.
add_card(_,Y,R) :- Y==inf,!,R=inf.
add_card(X,Y,R) :- add_card2(X,Y,R).

:- block add_card2(-,?,?), add_card2(?,-,?).
add_card2(inf,_,R) :- !,R=inf.
add_card2(_,inf,R) :- !,R=inf.
add_card2(X,Y,R) :- R is X+Y.

:- assert_must_succeed(exhaustive_kernel_check(cardinality_peano_wf([],0,no_wf_available))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_peano_wf([int(11)],s(0),no_wf_available))).
:- assert_must_succeed(exhaustive_kernel_check(cardinality_peano_wf([int(11),int(22)],s(s(0)),no_wf_available))).
% cardinality as peano number
:- block cardinality_peano_wf(-,-,?).
cardinality_peano_wf(Set,PCard,WF) :-
 (nonvar(Set),is_custom_explicit_set(Set,cardinality)
   ->  explicit_set_cardinality_wf(Set,Card,WF), % print(explicit_cardinality(Set,Card)),nl,
       card_convert_int_to_peano(Card,PCard)
   ;   cardinality3(Set,PCard,WF)
 ).

:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(3,s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(2,S),S==s(s(0)))).
:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(X,s(s(s(0)))),X==3)).
:- assert_must_succeed((kernel_objects:card_convert_int_to_peano(X,s(s(s(Y)))),X=4,Y==s(0))).
:- assert_must_fail((kernel_objects:card_convert_int_to_peano(X,s(s(s(_Y)))),X=2)).

:- block card_convert_int_to_peano(-,-).
card_convert_int_to_peano(X,S0) :- var(X), !,
   peel_s(S0,SX,RemS),  %% print(peel_s(S0,SX,RemS)),nl, %%
   (RemS==0 -> X=SX
    ; int_plus(int(X1),int(SX),int(X)),
      greater_than_equal(int(X1),int(0)),
      card_convert_int_to_peano(X1,RemS)).
card_convert_int_to_peano(inf,X) :- !,
    add_message(cardinality,'*** WARNING: Large or infinite Cardinality.'),nl,
    infinite_peano(X).
    %convert_int_to_peano(100,X). % used to limit to 100
card_convert_int_to_peano(X,P) :- convert_int_to_peano(X,P).

:- block infinite_peano(-).
infinite_peano(0) :- fail.
infinite_peano(s(X)) :- infinite_peano(X).

peel_s(0,0,0).
peel_s(s(X),Res,SX) :- (var(X) -> Res=1, SX=X ; peel_s(X,RX,SX), Res is RX+1).

:- block cardinality3(-,?,?). % avoids instantiating set; to do: use kernel_cardinality instead
% relevant, e.g., for "BK-ANT-N-2013" for SlotSolver_v7; but makes 'axm2/WD' fail for test 1448; TO DO: hopefully fixed with kernel_cardinality
% :- block cardinality3(-,-,?).
cardinality3(Set,SC,WF) :- var(Set),!,
  (SC=0 -> Set=[] ; SC=s(C),Set=[_|T],cardinality3(T,C,WF)).
cardinality3([],0,_).
cardinality3([_|T],s(C),WF) :- cardinality3(T,C,WF).
cardinality3(avl_set(AVL),Res,WF) :- cardinality_peano_wf(avl_set(AVL),Res,WF).
cardinality3(closure(P,T,B),Res,WF) :- cardinality_peano_wf(closure(P,T,B),Res,WF).






:- assert_must_succeed(exhaustive_kernel_check(card_geq([int(2),int(4),int(1)],s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_geq(global_set('Name'),s(s(s(0)))))).
:- assert_must_succeed((kernel_objects:card_geq([int(1),int(2)],s(s(0))))).
:- assert_must_succeed((kernel_objects:card_geq([int(1),int(2)],s(0)))).
:- assert_must_fail((kernel_objects:card_geq(global_set('Name'),s(s(s(s(0))))))).
:- assert_must_fail((kernel_objects:card_geq([int(1),int(2)],s(s(s(0)))))).
:- block card_geq(-,-).
card_geq(Set,Card) :-
 (nonvar(Set),is_custom_explicit_set(Set,card_geq)
   -> explicit_set_cardinality(Set,CCard), geq_int_peano(CCard,Card)
   ;  card_geq2(Set,Card) ).
   % should we call setup_ordered_list_skeleton(Card,Set,open)
:- block card_geq2(?,-).
card_geq2(_,C) :- C==0,!.
card_geq2(S,C) :- S==[],!,C=0.
card_geq2(S,s(C)) :- var(S),!,S=[_|T],card_geq2(T,C).
card_geq2([_|T],s(C)) :- card_geq2(T,C).
card_geq2(avl_set(A),s(C)) :- card_geq(avl_set(A),s(C)).
card_geq2(closure(P,T,B),s(C)) :- card_geq(closure(P,T,B),s(C)).
card_geq2(global_set(G),s(C)) :- card_geq(global_set(G),s(C)).

:- block geq_int_peano(-,-).
geq_int_peano(_,0).
geq_int_peano(X,s(C)) :- geq_int_peano1(X,C).
:- block geq_int_peano1(-,?).
geq_int_peano1(inf,_) :- !.
geq_int_peano1(X,C) :- X>0, X1 is X-1, geq_int_peano(X1,C).

:- block convert_int_to_peano(-,?).
convert_int_to_peano(X,Y) :- convert_int_to_peano2(X,Y).
convert_int_to_peano2(inf,_).
convert_int_to_peano2(X,R) :- number(X),
 (X>100000
   -> print('*** Warning: converting large integer to peano: '),print(X),nl,
      (X>1000000000 -> print('*** treat like inf'),nl  % no hope of ever finishing, do not instantiate just like inf
        ; convert_int_to_peano3(X,R))
    ; convert_int_to_peano3(X,R)
 ).
convert_int_to_peano3(0,R) :- !, R=0.
convert_int_to_peano3(X,s(P)) :-
  (X>0 -> X1 is X-1, convert_int_to_peano3(X1,P)
   ; X<0 -> add_error_and_fail(convert_int_to_peano,'Negative nr cannot be converted to peano: ',X)
  ).

% not used:
%:- block convert_peano_to_int(-,?).
%convert_peano_to_int(0,0).
%convert_peano_to_int(s(P),X) :- convert_peano_to_int(P,X1), X is X1+1.

:- assert_must_succeed((kernel_objects:cardinality_greater_equal(Set,set(integer),int(X),integer,_WF), X=3,
 nonvar(Set),Set=[_|S2],nonvar(S2),S2=[_|S3],nonvar(S3),S3=[_|S4],var(S4), Set=[int(1),int(2),int(3)] )).
:- assert_must_succeed((kernel_objects:cardinality_greater(Set,set(integer),int(X),integer,_WF), X=2,
 nonvar(Set),Set=[_|S2],nonvar(S2),S2=[_|S3],nonvar(S3),S3=[_|S4],var(S4), Set=[int(1),int(2),int(3)] )).
   /* special predicates called for e.g. card(Set)>X */
cardinality_greater(Set,TypeSet,int(X),_,WF) :-
   kernel_objects:max_cardinality(TypeSet,MaxCard),
   %print(check_less(X,MaxCard)),nl,
   (number(MaxCard) -> less_than(int(X),int(MaxCard)) ; true),
   card_greater2(Set,X,WF).
:- block card_greater2(?,-,?).
card_greater2(Set,X,WF) :- X1 is X+1, card_greater_equal2(Set,X1,WF).

cardinality_greater_equal(Set,TypeSet,int(X),_,WF) :-
   kernel_objects:max_cardinality(TypeSet,MaxCard),
   %print(check_less_equal(X,MaxCard)),nl,
   (number(MaxCard) -> less_than_equal(int(X),int(MaxCard)) ; true),
   card_greater_equal2(Set,X,WF).
:- block card_greater_equal2(?,-,?).
card_greater_equal2(Set,X,WF) :-
    (X<1 -> true % potential WD issue, hence this predicates should only be called when no wd issue
     ; X=1 -> not_empty_set(Set) % ditto: Set could be infinite
     ; var(Set) -> setup_ordered_list_skeleton(X,Set,open,WF)
     ; convert_int_to_peano(X,Peano),
       card_geq(Set,Peano)).



%is_cartesian_pair_or_times(P,X,Y) :- is_cartesian_pair(P,X,Y).
%is_cartesian_pair_or_times(int(Z),int(X),int(Y)) :- times(int(X),int(Y),int(Z)).

is_cartesian_pair_wf((X,Y),XType,YType,WF) :-
   check_element_of_wf(X,XType,WF), check_element_of_wf(Y,YType,WF).

:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_objects:not_is_cartesian_pair((int(1),int(1)),
           [int(1),int(2)],[int(2),int(3)],WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_objects:not_is_cartesian_pair((int(3),int(2)),
           [int(1),int(2)],[int(2),int(3)],WF),WF)).
:- assert_must_succeed((kernel_objects:not_is_cartesian_pair((int(1),int(1)),
           [int(1),int(2)],[int(2),int(3)],_WF))).
:- assert_must_succeed((kernel_objects:not_is_cartesian_pair((int(3),int(1)),
           [int(1),int(2)],[int(2),int(3)],_WF))).
:- assert_must_fail((kernel_objects:not_is_cartesian_pair((int(1),int(3)),
           [int(1),int(2)],[int(2),int(3)],_WF))).
:- assert_must_succeed((kernel_objects:not_is_cartesian_pair((X,int(3)),
           [int(1),int(2)],[int(2),int(3)],_WF),X=int(4))).


not_is_cartesian_pair((X,Y),XType,YType,WF) :-
   not_is_cartesian_pair0(X,Y,XType,YType,WF).

:- block not_is_cartesian_pair0(-,-,?,?,?).
not_is_cartesian_pair0(X,Y,XType,YType,WF) :-
   (nonvar(X) -> not_is_cartesian_pair1(X,Y,XType,YType,WF)
    ; not_is_cartesian_pair1(Y,X,YType,XType,WF)).

not_is_cartesian_pair1(X,Y,XType,YType,WF) :-
  %print(member_ship_test(XType,X,MemResX)),nl,
  membership_test_wf(XType,X,MemResX,WF),
  (var(MemResX) -> membership_test_wf(YType,Y,MemResY,WF) ; true),
  not_is_cartesian_pair3(MemResX,X,XType,MemResY,Y,YType,WF).

:- block not_is_cartesian_pair3(-,?,?, -,?,?, ?).
not_is_cartesian_pair3(MemResX,X,XType, MemResY,Y,YType, WF) :-
  (MemResX==pred_false -> true
   ; MemResY==pred_false -> true
   ; MemResX==pred_true -> not_element_of_wf(Y,YType,WF)
   ; not_element_of_wf(X,XType,WF)
  ).



/***************************/
/*  power_set(Set,TypeSet) */
/*    Set : POW(TypeSet)   */
/***************************/

:- assert_must_succeed(exhaustive_kernel_check(power_set([int(2),int(4)],[[int(2)],
 [int(4)],[],[int(4),int(2)]]))).
:- assert_must_succeed(power_set([int(1)],[[int(1)],[]])).
:- assert_must_succeed((power_set([int(1),int(2)],R),
           equal_object(R,[[],[int(1)],[int(2)],[int(1),int(2)]]))).
:- assert_must_succeed(power_set([],[[]])).

:- block power_set(-,?).
power_set(S,Res) :-   %print_message(power_set(S,PowerS)),
    cardinality_peano_wf(S,Card,no_wf_available),
    when(ground(Card),  /* when all elements are known */
          (try_expand_custom_set(S,SE),
           findall(Subset, generate_subsets(SE,Subset), PowerS), %print(powerS(PowerS)),nl,
           equal_object_optimized(PowerS,Res,power_set)
          )).


generate_subsets([],[]).
generate_subsets([H|T],R) :- (R=[H|GT]; R=GT), generate_subsets(T,GT).


:- assert_must_succeed(exhaustive_kernel_check(non_empty_power_set([int(2),int(4)],[[int(2)],
 [int(4)],[int(4),int(2)]]))).
:- assert_must_succeed(non_empty_power_set([int(1)],[[int(1)]])).
:- assert_must_succeed((non_empty_power_set([int(1),int(2)],R),
           equal_object(R,[[int(1)],[int(2)],[int(1),int(2)]]))).
:- assert_must_succeed(non_empty_power_set([],[])).

:- block non_empty_power_set(-,?).
non_empty_power_set(S,Res) :-
    cardinality_peano_wf(S,Card,no_wf_available),
    when(ground(Card),  /* when all elements are known */
          (try_expand_custom_set(S,SE),findall(Subset, (generate_subsets(SE,Subset),not_empty_set(Subset)), PowerS),
           equal_object_optimized(PowerS,Res,non_empty_power_set)) ).



/* ------- */
/* BOOLEAN */
/* ------- */

% following predicates are not used:
%is_boolean(pred_true /* bool_true */).
%is_boolean(pred_false /* bool_false */).
%is_not_boolean(X) :- dif(X,pred_true /* bool_true */), dif(X,pred_false /* bool_false */).

/* ------- */
/* NUMBERS */
/* ------- */


is_integer(int(X),_WF) :- when(ground(X),integer(X)).
:- block is_not_integer(-).
is_not_integer(X) :- X \= int(_), % will be called for x /: INTEGER; should always fail.
   add_internal_error('Wrong type argument: ',is_not_integer(X)),fail.

is_natural(int(X),_WF) :- clpfd_geq2(X,0,Posted), (Posted==true -> true  ; number_geq(X,0)).
is_natural1(int(X),_WF) :- clpfd_geq2(X,1,Posted), (Posted==true -> true ; number_geq(X,1)).
:- block number_geq(-,?).
number_geq(X,N) :- X>=N.
:- block number_leq(-,?).
number_leq(X,N) :- X=<N.

:- assert_must_succeed(is_implementable_int(int(0),_WF)).
:- assert_must_fail(is_not_implementable_int(int(0))).


is_implementable_int(int(X),WF) :-  element_of_global_integer_set_wf('INT',X,WF).
is_implementable_nat(int(X),WF) :- element_of_global_integer_set_wf('NAT',X,WF).
is_implementable_nat1(int(X),WF) :- element_of_global_integer_set_wf('NAT1',X,WF).
is_not_implementable_int(X) :- not_element_of_global_set(X,'INT').
is_not_implementable_nat(X) :- not_element_of_global_set(X,'NAT').
is_not_implementable_nat1(X) :- not_element_of_global_set(X,'NAT1').

is_not_natural(int(X)) :- clpfd_geq2(-1,X,Posted), (Posted=true -> true ; number_leq(X,-1)).
is_not_natural1(int(X)) :- clpfd_geq2(0,X,Posted), (Posted==true -> true ; number_leq(X,0)).

:- assert_must_succeed(exhaustive_kernel_check(less_than(int(2),int(3)))).
:- assert_must_succeed(( safe_less_than(A,B),A=3,B=5 )).
:- assert_must_succeed(( safe_less_than(A,B),B=5,A=3 )).
:- assert_must_fail(( safe_less_than(A,B),A=5,B=3 )).
:- assert_must_fail(( safe_less_than(A,B),B=3,A=5 )).
:- assert_must_fail(( safe_less_than(A,B),A=5,B=5 )).
:- assert_must_fail(( safe_less_than(A,B),B=5,A=5 )).

less_than(int(X),int(Y)) :-
    (number(X),number(Y) -> X < Y
      ;  clpfd_lt(X,Y,Posted),
        (Posted=true -> true  ; safe_less_than(X,Y))).
less_than_direct(X,Y) :-
    (number(X),number(Y) -> X < Y
      ;  clpfd_lt(X,Y,Posted),
        (Posted=true -> true  ; safe_less_than(X,Y))).
:- block safe_less_than(-,?), safe_less_than(?,-).
safe_less_than(X,Y) :-
 (number(X),number(Y) -> X<Y
    ; add_internal_error('Arguments not numbers: ',safe_less_than(X,Y))).

:- assert_must_succeed(exhaustive_kernel_check(less_than_equal(int(33),int(33)))).
less_than_equal(int(X),int(Y)) :-
    (number(X),number(Y) -> X =< Y
      ; clpfd_leq(X,Y,Posted),
        (Posted=true -> true  ; safe_less_than_equal(less_than_equal,X,Y))).
less_than_equal_direct(X,Y) :-
    (number(X),number(Y) -> X =< Y
      ; clpfd_leq(X,Y,Posted),
        (Posted=true -> true  ; safe_less_than_equal(less_than_equal_direct,X,Y))).

safe_less_than_equal(X,Y) :-
   safe_less_than_equal(safe_less_than_equal,X,Y).
:- block safe_less_than_equal(?,-,?), safe_less_than_equal(?,?,-).
safe_less_than_equal(PP,X,Y) :- %print(slte(PP,X,Y)),nl,
 (number(X),number(Y) -> X=<Y
    ; add_internal_error('Arguments not numbers: ',safe_less_than_equal(PP,X,Y))).

:- assert_must_succeed(exhaustive_kernel_check(greater_than(int(2),int(1)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(greater_than(int(2),int(2)))).
greater_than(int(X),int(Y)) :- less_than_direct(Y,X).
:- assert_must_succeed(exhaustive_kernel_check(greater_than(int(2),int(1)))).
:- assert_must_succeed(exhaustive_kernel_check(greater_than_equal(int(2),int(2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(greater_than_equal(int(1),int(2)))).
greater_than_equal(int(X),int(Y)) :- less_than_equal_direct(Y,X).





:- assert_must_succeed(exhaustive_kernel_check([commutative],int_plus(int(2),int(3),int(5)))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],int_plus(int(2),int(3),int(6)))).

:- assert_must_succeed(int_plus(int(1),int(2),int(3))).
:- assert_must_succeed(( int_plus2(A,B,C),A=3,B=2,C==5 )).
:- assert_must_succeed(( int_plus2(A,B,C),A=3,C=5,B==2 )).
:- assert_must_succeed(( int_plus2(A,B,C),B=2,A=3,C==5 )).
:- assert_must_succeed(( int_plus2(A,B,C),B=2,C=5,A==3 )).
:- assert_must_succeed(( int_plus2(A,B,C),C=5,A=3,B==2 )).
:- assert_must_succeed(( int_plus2(A,B,C),C=5,B=2,A==3 )).
:- assert_must_succeed(( int_plus2(A,B,C),A=0,B==C )).
:- assert_must_succeed(( int_plus2(A,B,C),B=0,A==C )).

int_plus(int(X),int(Y),int(Plus)) :- %print(int_plus(X,Y,Plus)),nl,
    (two_vars_or_more(X,Y,Plus)
      -> clpfd_eq(Plus,X+Y) % can have performance problems
       ; true % otherwise we can compute the value directly below; we could skip the block declaration
    ),
    int_plus2(X,Y,Plus).
two_vars_or_more(X,Y,Z) :- var(X),!, (var(Y) ; var(Z)).
two_vars_or_more(_X,Y,Z) :- var(Y) , var(Z).

:- block int_plus2(-,-,-).
int_plus2(X,Y,Plus) :-  %print(int_plus2(X,Y,Plus)),nl,
    ( ground(X) -> int_plus3(X,Y,Plus)
    ; ground(Y) -> int_plus3(Y,X,Plus)
    ;              int_minus3(Plus,X,Y)).

% int_plus3/3: the first argument must be ground when called
int_plus3(0,Y,Plus) :- !, Y=Plus. % not inferred by CLP(FD): Z #= Y+X, X=0. does not infer Y==Z
int_plus3(X,Y,Plus) :- % print(dif(Y,Plus)),nl, integer_dif(Y,Plus), % this generates overflows for test 1353, 1014
                       int_plus4(X,Y,Plus).

% int_plus4/3: the first argument must be ground when called
:- block int_plus4(?,-,-).
int_plus4(X,Y,Plus) :-
    ( var(Plus) -> Plus is X+Y
    ;              Y is Plus-X).

:- assert_must_succeed(exhaustive_kernel_check(int_minus(int(2),int(3),int(-1)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(int_minus(int(2),int(3),int(1)))).
:- assert_must_succeed(int_minus(int(3),int(1),int(2))).
:- assert_must_succeed(( int_minus2(A,B,C),A=3,B=2,C==1 )).
:- assert_must_succeed(( int_minus2(A,B,C),A=3,C=1,B==2 )).
:- assert_must_succeed(( int_minus2(A,B,C),B=2,A=3,C==1 )).
:- assert_must_succeed(( int_minus2(A,B,C),B=2,C=1,A==3 )).
:- assert_must_succeed(( int_minus2(A,B,C),C=1,A=3,B==2 )).
:- assert_must_succeed(( int_minus2(A,B,C),C=1,B=2,A==3 )).
:- assert_must_succeed(( int_minus2(A,B,C),B=0,A==C )).
:- assert_must_succeed(( int_minus2(A,B,C),B=0,C=5,A==5 )).
:- assert_must_succeed(( int_minus2(A,B,5),B=0,A==5 )).

int_minus(int(X),int(Y),int(Minus)) :-  %print(int_minus(X,Y,Minus)),nl,
   int_minus2(X,Y,Minus),
   (two_vars_or_more(X,Y,Minus) -> clpfd_eq(Minus,X-Y) % can have performance problems.
   % we could also set Minus to 0 if X==Y; this is done in CHR (chr_integer_inequality)
     ; true). % we can compute the value directly anyway
:- block int_minus2(-,-,-).
int_minus2(X,Y,Minus) :- %print(int_minus2(X,Y,Minus)),nl,
    ( ground(Y) ->
	   ( Y=0       -> X=Minus
	    ; Y2 is -Y, int_plus3(Y2,X,Minus))
    ; ground(X) ->
	   int_minus3(X,Y,Minus)
    ;  int_plus3(Minus,Y,X) % will infer that Y=X if Minus=0
    ).

% int_minus3/3: the first argument must be ground when called
:- block int_minus3(?,-,-).
int_minus3(X,Y,Minus) :-
    ( var(Minus) -> Minus is X-Y
    ;               Y is X-Minus).

:- assert_must_succeed(exhaustive_kernel_check(division(int(2),int(3),int(0),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(7),int(2),int(3),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(8),int(2),int(4),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(9),int(2),int(4),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(2),int(-1),int(-2),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(9),int(-2),int(-4),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(-9),int(-3),int(3),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(division(int(-1),int(4),int(0),unknown,_WF))).
:- assert_must_succeed((platform_is_64_bit
		       -> exhaustive_kernel_check(division(int(4294967296),int(2),int(2147483648),unknown,_WF))
		       ;  exhaustive_kernel_check(division(int(134217728),int(2),int(67108864),unknown,_WF)))).
:- assert_must_succeed((platform_is_64_bit
		       -> exhaustive_kernel_check(division(int(4294967296),int(2147483648),int(2),unknown,_WF))
		       ;  exhaustive_kernel_check(division(int(134217728),int(67108864),int(2),unknown,_WF)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(division(int(2),int(3),int(1),unknown,_WF))).
:- assert_must_succeed(( division3(A,B,C,unknown,_),A=15,B=4,C==3 )).
:- assert_must_succeed(( division3(A,B,C,unknown,_),B=4,A=15,C==3 )).

division(int(X),int(Y),int(XDY),Span,WF) :- var(Y), (var(X) ; var(XDY)),
    preferences:preference(use_clpfd_solver,true),!,
    (preferences:preference(disprover_mode,true)
      -> clpfd_eq_div(XDY,X,Y) /* we can assume well-definedness */
      ;  clpfd_eq_guarded_div(XDY,X,Y),
          % TO DO: we could set up a choice point just before enumeration of infinite types for Y=0 & Y/=0;
          %        same for modulo
         check_nonzero(X,Y,XDY,Span,WF)
    ).
division(int(X),int(Y),int(XDY),Span,WF) :-
    %% clpfd_eq_expr(XDY,X/Y), % can have performance problems; could hide division by 0 !
    division3(X,Y,XDY,Span,WF).

:- block check_nonzero(?,-,?,?,?).
check_nonzero(X,Y,XDY,Span,WF) :-
  (Y=0 -> add_wd_error_set_result('division by zero','/'(X,Y),XDY,0,Span,WF)
   ; true).

:- block division3(?,-,?,?,?).
division3(X,Y,XDY,Span,WF) :-
    ( Y==0 -> add_wd_error_set_result('division by zero','/'(X,Y),XDY,0,Span,WF)
     ; nonvar(X) -> XDY is X // Y
     ; Y == 1 -> X=XDY
     ; (Y == -1,nonvar(XDY)) -> X is -XDY
     ; clpfd_eq_div(XDY,X,Y)). % we could setup constraint before Y is known; could hide division by 0 ?



:- assert_must_succeed(exhaustive_kernel_check(floored_division(int(2),int(3),int(0),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(floored_division(int(7),int(2),int(3),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(floored_division(int(-1),int(4),int(-1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(floored_division(int(-9),int(-3),int(3),unknown,_WF))).
floored_division(int(X),int(Y),int(XDY),Span,WF) :- var(Y), (var(X) ; var(XDY)),
    preferences:preference(use_clpfd_solver,true),!,
    (preferences:preference(disprover_mode,true)
      -> clpfd_eq_fdiv(XDY,X,Y) /* we can assume well-definedness */
      ;  clpfd_eq_guarded_fdiv(XDY,X,Y),
         check_nonzero(X,Y,XDY,Span,WF)
    ).
floored_division(int(X),int(Y),int(XDY),Span,WF) :-
    %% clpfd_eq_expr(XDY,X/Y), % can have performance problems; could hide division by 0 !
    floored_division3(X,Y,XDY,Span,WF).
:- block floored_division3(?,-,?,?,?).
floored_division3(X,Y,XDY,Span,WF) :-
    ( Y==0 -> add_wd_error_set_result('division by zero','/'(X,Y),XDY,0,Span,WF)
     ; nonvar(X) ->  XDY is X div Y
     ; Y == 1 -> X=XDY
     ; (Y == -1,nonvar(XDY)) -> X is -XDY
     ; clpfd_eq_guarded_fdiv(XDY,X,Y)). % we could setup constraint before Y is known; could hide division by 0 ?

:- assert_must_succeed(exhaustive_kernel_check(modulo(int(2),int(3),int(2),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(modulo(int(7),int(2),int(1),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(modulo(int(8),int(2),int(0),unknown,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(modulo(int(9),int(2),int(1),unknown,_WF))).
:- assert_must_succeed((platform_is_64_bit
		       -> exhaustive_kernel_check(modulo(int(4294967296),int(2147483648),int(0),unknown,_WF))
		       ;  exhaustive_kernel_check(modulo(int(134217728),int(67108864),int(0),unknown,_WF)))).
:- assert_must_succeed((platform_is_64_bit
		       -> exhaustive_kernel_check(modulo(int(4294967299),int(2147483648),int(3),unknown,_WF))
		       ;  exhaustive_kernel_check(modulo(int(134217731),int(67108864),int(3),unknown,_WF)))).
:- assert_must_succeed(( modulo2(A,B,C,unknown,_),A=7,B=5,C==2 )).
:- assert_must_fail(( modulo2(A,B,C,unknown,_),A=7,B=5,C==3 )).

modulo(int(X),int(Y),int(Modulo),Span,WF) :-
    %% clpfd_eq(Modulo,X mod Y), % can have performance problems; could hide division by 0 !
    %clpfd_modulo(X,Y,Modulo,WF), % maybe only call in non-CLPFD mode ?
    modulo2(X,Y,Modulo,Span,WF),
    % assert that Modulo<Y, Modulo>=0
    (nonvar(X),nonvar(Y) -> true % we already have computed Modulo using modulo2
      ; nonvar(Modulo), Modulo < 0 -> true % we will generate well-definedness error; see comment next line
      ; number(Y),Y =< 0 -> true % in this case we will generate a well-definedness error; it would be more efficient from a constraint solving perspective to assume that there are no well-definedness errors and remove this case !!
      ; clpfd_modulo_prop(X,Y,Modulo,WF)
    ).
:- use_module(specfile,[z_or_tla_minor_mode/0]).
:- block modulo2(-,?,?,?,?), modulo2(?,-,?,?,?).
modulo2(X,Y,Modulo,Span,WF) :-
    ( Y>0 -> (X<0 -> (z_or_tla_minor_mode -> Modulo is X mod Y
					   ; add_wd_error_set_result('mod not defined for negative numbers in B:',mod(X,Y),Modulo,0,Span,WF))
                   ; Modulo is X mod Y)
    ; Y==0  -> add_wd_error_set_result('mod by zero:',mod(X,Y),Modulo,0,Span,WF)
    ; Y<0 -> add_wd_error_set_result('mod not defined for negative numbers:',mod(X,Y),Modulo,0,Span,WF)). % there seems to be a definition in Z ? at least for Z Live ?

% propagate information about Modulo result if part of the information known
clpfd_modulo_prop(X,Y,Modulo,WF) :- %preferences:preference(use_clpfd_solver,true),!,
    % in CLP(FD) this is sufficient; for non-CLPFD mode it is better to call in_nat_range to restrict enumeration
    less_than_direct(Modulo,Y),
    less_than_equal_direct(0,Modulo),  % 0 <= Modulo < Y  -> by transitivity this forces Y>0  and we no longer detect wd-errors
    %less_than_equal_direct(Modulo,X). %  by transitivity this imposes X >= 0 and we will never find WD problems with negative X
    clpfd_modulo_prop2(X,Y,Modulo,WF).


clpfd_modulo_prop2(X,Y,Modulo,_WF) :-
    number(Modulo), % this test is required for test 1009, 417 : TO DO : investigate cause
    var(X), % or should this be var(X) ; var(Y) ??
    preferences:preference(use_clpfd_solver,true),
    clpfd:fd_min(Y,MinY), number(MinY), MinY>0,
    clpfd:fd_min(X,MinX), number(MinX), MinX>=0,
    !,
    %print(modulo_prop(X,MinX,Y,MinY,Modulo)),nl,
    clpfd_interface:post_constraint(X#>=Modulo),
    clpfd_interface:post_constraint(Modulo #= X mod Y).
clpfd_modulo_prop2(X,_Y,Modulo,_WF) :-
    clpfd_interface:post_constraint(X#>=0 #=> X#>=Modulo). % this would be faster (e.g., {y|y:100000..200000 & y mod 2 = 0}), but would not catch some WD errors: clpfd_interface:post_constraint(X#>=Modulo).
    % we could reify: Y>0 => Modulo <Y ?  Is it worth it ?
    % we could also use the CLP(FD) modulo operator X in 3..100, 1 #= X mod 20 infers X in 21..81
    % post_constraint((X#>=0 #/\ Y#>0) #=> Modulo #= X mod Y)
    % what is still missing is that if Y < Modulo => X=Y  (CLP(FD) does this X in 0..100 , Y in 2..20 ,  X #= Y mod 30.)
/* clpfd_modulo_prop(X,Y,Modulo,WF) :- clpfd_modulo_noclp(X,Y,Modulo,WF).
:- block clpfd_modulo_noclp(-,-,-,?).
clpfd_modulo_noclp(X,Y,Modulo,WF) :-  print(mod(X,Y,Modulo,WF)),nl,
    var(X),var(Modulo),number(Y),!,
    Y1 is Y-1,
    in_nat_range_wf(int(Modulo),int(0),int(Y1),WF).  % problem: could enumerate lambda return variables !!
clpfd_modulo_noclp(_X,_Y,_Modulo,_WF).
*/


:- assert_must_succeed(exhaustive_kernel_check(unary_minus_wf(int(2),int(-2),_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(unary_minus_wf(int(2),int(2),_WF))).
:- assert_must_succeed(( unary_minus2(A,B),A=7,B== -7 )).
:- assert_must_succeed(( unary_minus2(A,B),A= -7,B==7 )).
:- assert_must_succeed(( unary_minus2(B,A),A=7,B== -7 )).
:- assert_must_succeed(( unary_minus2(B,A),A= -7,B==7 )).
:- assert_must_fail(( unary_minus2(B,A),A= -7,B=6 )).
:- assert_must_fail(( unary_minus2(A,B),A= -7,B=6 )).

unary_minus_wf(int(X),int(MX),_WF) :-
    unary_minus2(X,MX),
    (var(X),var(MX) -> clpfd_eq(MX,0 - X) % can have performance problems
       ; true % we can compute the value without CLPFD
    ).
:- block unary_minus2(-,-).
unary_minus2(X,MX) :-
    ( ground(X) -> MX is -X
    ;              X is -MX).

:- assert_must_succeed(first_of_pair((int(1),int(2)),int(1))).
:- assert_must_succeed(second_of_pair((int(1),int(2)),int(2))).

first_of_pair((A,_B),R) :- equal_object(R,A,first_of_pair).
second_of_pair((_A,B),R) :- equal_object(R,B,second_of_pair).


:- assert_must_succeed(exhaustive_kernel_check(cartesian_product([int(2),int(4)],[int(3),int(1)],
 [(int(2),int(1)),(int(2),int(3)),(int(4),int(3)),(int(4),int(1))]))).
:- assert_must_succeed(exhaustive_kernel_check(cartesian_product([],[int(3),int(1)],[]))).
:- assert_must_succeed(exhaustive_kernel_check(cartesian_product([int(3)],[],[]))).
:- assert_must_succeed(exhaustive_kernel_fail_check(cartesian_product([int(3)],[int(2)],[]))).
:- assert_must_succeed((cartesian_product(global_set('NAT'),[int(2)],_Res))).
:- assert_must_succeed((cartesian_product([int(1)],[int(2)],Res),
         equal_object(Res,[(int(1),int(2))]))).
:- assert_must_succeed((cartesian_product([int(1)],[int(2)],[(int(1),int(2))]))).
:- assert_must_succeed((cartesian_product([],[int(1),int(2)],Res),
         equal_object(Res,[]))).
:- assert_must_succeed((cartesian_product([int(1),int(2)],[],Res),
         equal_object(Res,[]))).
:- assert_must_succeed((cartesian_product([int(1),int(2)],[int(2),int(3)],Res),
 equal_object(Res,[(int(1),int(2)),(int(1),int(3)),(int(2),int(2)),(int(2),int(3))]))).
:- assert_must_succeed((cartesian_product([int(1)|T],[int(2)|T2],Res),
             T = [int(2)], T2 = [int(3)],
 equal_object(Res,[(int(1),int(2)),(int(1),int(3)),(int(2),int(2)),(int(2),int(3))]))).
:- assert_must_fail((cartesian_product([int(1)],[int(2),int(3)],Res),(Res=[_];
         equal_object(Res,[_,_,_|_])))).


:- block cartesian_product(-,?,?), cartesian_product(?,-,?).
cartesian_product(Set1,Set2,Res) :-
    expand_custom_set_to_list(Set1,ESet1,_,cartesian_product1),
    (ESet1==[] -> empty_set(Res)
     ; expand_custom_set_to_list(Set2,ESet2,_,cartesian_product2),
       (var(Res) -> cartesian_product2(ESet1,ESet2,CRes), kernel_objects:equal_object_optimized(CRes,Res,cart_product)
        ; cartesian_product2(ESet1,ESet2,Res))
     ).

:- block cartesian_product2(-,?,?).
cartesian_product2([],_,Res) :- empty_set(Res).
cartesian_product2([H|T],Set2,Res) :-
    cartesian_el_product(Set2,H,Res,InnerRes),
    cartesian_product2(T,Set2,InnerRes).

:- block cartesian_el_product(-,?,?,?).
cartesian_el_product([],_El,Res,InnerRes) :- equal_object_optimized(Res,InnerRes,cartesian_el_product_1).
cartesian_el_product([H|T],El,ResSoFar,InnerRes) :-
    equal_object(ResSoFar,[(El,H)|NewResSoFar],cartesian_el_product_2),
    cartesian_el_product(T,El,NewResSoFar,InnerRes).



:- assert_must_succeed(exhaustive_kernel_check(in_nat_range(int(2),int(2),int(3)))).
:- assert_must_succeed(exhaustive_kernel_check(in_nat_range_wf(int(2),int(2),int(3),_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(in_nat_range_wf(int(2),int(3),int(2),_WF))).
:- assert_must_succeed((in_nat_range_wf(X,int(11),int(12),WF),
   kernel_waitflags:ground_wait_flags(WF), X==int(12) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(12),_WF), X=int(10) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(12),_WF),  X=int(13) )).
:- assert_must_succeed((in_nat_range_wf(X,int(11),int(12),_WF), X=int(11) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(10),_WF), X=int(11) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(10),_WF), X=int(10) )).
:- assert_must_fail((in_nat_range_wf(X,int(11),int(10),_WF), X=int(12) )).

in_nat_range(int(X),int(Y),int(Z)) :- % does not enumerate, in contrast to in_nat_range_wf
  clpfd_inrange(X,Y,Z,Posted),  % better to call inrange rather than leq twice, avoids unecessary propagation
  (Posted==true -> true
   ;   safe_less_than_equal(in_nat_range,Y,X),
       safe_less_than_equal(in_nat_range,X,Z)
   ).
in_nat_range_wf(int(X),int(Y),int(Z),WF) :- %print(in_nat_range_wf(X,Y,Z)),nl,
  clpfd_inrange(X,Y,Z,Posted),  % better to call inrange rather than leq twice, avoids unecessary propagation
  (Posted==true ->
     /* if the constraint was posted: we do not need to add safe_less_than_equal,...: if overflow happes whole computation will fail anyway */
       add_nat_range_fd_variable_for_labeling(X,Y,Z,WF)  % do we really need to do this ? maybe add just before enumeration finished ?
       %,print(delay_adding(X,Y,Z)),nl %, portray_waitflags(WF),nl,nl
   ;   safe_less_than_equal(in_nat_range_wf,Y,X),
       safe_less_than_equal(in_nat_range_wf,X,Z),
       (ground(X) -> true
       ; get_int_domain(X,Y,Z,RL,RU),get_nat_range_prio(X,RL,RU,WF,LWF),
         % print(register_enumerate_int(X,RL,RU,WF,LWF)),nl, %% portray_waitflags(WF),nl, %%
         call_enumerate_int(X,RL,RU,LWF))
  ).
 % when((ground(X);nonvar(LWF)),(ground(X) -> true ; enumerate_int(X,RL,RU))).

add_nat_range_fd_variable_for_labeling(X,_Low,_Up,_WF) :- nonvar(X),!.
% TO DO: avoid adding useless choice points; not adding makes test 328 fail
%add_nat_range_fd_variable_for_labeling(X,Low,Up,WF) :- !,Size is 100*(Up+1-Low),
%    get_wait_flag(Size,add_nat_range_fd(X,Low,Up),WF,LWF), when(nonvar(LWF),add_fd_variable_for_labeling(X,WF)).
add_nat_range_fd_variable_for_labeling(X,_Low,_Up,WF) :- !,add_fd_variable_for_labeling(X,WF).


:- block get_nat_range_prio(?,-,?,?,?), get_nat_range_prio(?,?,-,?,?).
get_nat_range_prio(_Variable,Y,Z,WF,LWF) :- Size is Z+1-Y, %print(get_natrange_prio(Size,Y,Z,WF)),nl,
   (Size>1 ->
     % we do not use add_fd_variable_for_labeling(Variable,Size,WF,LWF) % will use CLP(FD) labeling
     % either clpfd is off or we had a time-out or overflow; so labeling may generate instantiation error
     get_wait_flag(Size,get_nat_range_prio(Y,Z),WF,LWF)
    ; LWF=Size  /* Size=0 or 1 -> we can either fail or determine variable */).

:- assert_must_succeed((kernel_objects:call_enumerate_int(X,1,2,g), X==2)).
:- block call_enumerate_int(-,?,?,-).
call_enumerate_int(X,RL,RU,_LWF) :-  % print(enum(X,RL,RU,_LWF)),nl,%
   (ground(X) -> true
     ;   % print(enum_int(X,RL,RU)),nl, %%
       % get_int_domain(X,RL,RU,RLL,RUU) : if clp(fd) active then CLP(FD) labeling is used anyway
       enumerate_int(X,RL,RU)).




:- assert_must_succeed(exhaustive_kernel_check(not_in_nat_range(int(2),int(3),int(2)))).
:- assert_must_succeed(exhaustive_kernel_fail_check(not_in_nat_range(int(2),int(2),int(3)))).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(12)), X=int(10) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(12)), X=int(13) )).
:- assert_must_fail((not_in_nat_range(X,int(11),int(12)), X=int(11) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(10)), X=int(11) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(10)), X=int(10) )).
:- assert_must_succeed((not_in_nat_range(X,int(11),int(10)), X=int(12) )).

not_in_nat_range_wf(X,Y,Z,_WF) :- not_in_nat_range(X,Y,Z).
not_in_nat_range(int(X),int(Y),int(Z)) :-
  (number(Y),number(Z)
    -> (Z>=Y -> clpfd_not_in_non_empty_range(X,Y,Z) ; true /* interval empty */)
    ;  clpfd_not_inrange(X,Y,Z)
  ).


:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(1),int(0),int(10),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(10),int(10),int(10),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(1),int(1),int(10),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(10),int(0),int(10),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(11),int(10),int(9),pred_false,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(11),int(13),int(12),pred_false,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(test_in_nat_range_wf(int(11),int(13),int(15),pred_false,WF),WF)).

% reified version
:- block test_in_nat_range_wf(-,-,?,-,?), test_in_nat_range_wf(-,?,-,-,?), test_in_nat_range_wf(?,-,-,-,?).
test_in_nat_range_wf(X,Y,Z,PredRes,WF) :- PredRes==pred_true,!,
   in_nat_range_wf(X,Y,Z,WF).
test_in_nat_range_wf(X,Y,Z,PredRes,WF) :- PredRes==pred_false,!,
   not_in_nat_range_wf(X,Y,Z,WF).
test_in_nat_range_wf(int(X),int(Low),int(Up),PredRes,WF) :-
   %print(post(X,Low,Up)),nl,
   clpfd_interface:post_constraint2(C1 #<=> (X #>= Low #/\ X #=< Up #/\ Low #=< Up),Posted1),
   %print(p(Posted1,C1)),nl,
   (Posted1 == true -> prop_01(C1,PredRes) ; test_in_nat_range_no_clpfd(X,Low,Up,PredRes,WF)).

% Note: A #<=> (X #>= Low #/\ X#=< Up #/\ Low #=< Up),  Low in 11..15, Up in 7..8. -> CLPFD infers A=0
% without the redundant Low #=< Up it does not infer it !
:- block prop_01(-,-).
prop_01(0,pred_false).
prop_01(1,pred_true).

:- block test_in_nat_range_no_clpfd(-,?,?,-,?), test_in_nat_range_no_clpfd(?,-,?,-,?),
         test_in_nat_range_no_clpfd(?,?,-,-,?).
test_in_nat_range_no_clpfd(X,Y,Z,PredRes,WF) :- PredRes==pred_true,!,
   in_nat_range_wf(int(X),int(Y),int(Z),WF).
test_in_nat_range_no_clpfd(X,Y,Z,PredRes,WF) :- PredRes==pred_false,!,
   not_in_nat_range_wf(int(X),int(Y),int(Z),WF).
test_in_nat_range_no_clpfd(X,Y,Z,PredRes,_WF) :- % X,Y,Z must be ground integers
   (X >= Y, X =< Z, Y =< Z -> PredRes=pred_true ; PredRes=pred_false).

:- assert_must_succeed(exhaustive_kernel_check_wf(square(int(3),int(9),WF),WF)).
% is now only called when CLPFD is FALSE
square(int(X),int(Sqr),WF) :- % print(sqr(X,Sqr)),nl,
   int_square(X,Sqr,WF),
   (var(X) -> clpfd_eq(Sqr,X * X)
     ; true). % we can compute the value directly

:- block int_square(-,-,?).
int_square(X,Sqr,_) :- ground(X),!, Sqr is X*X.
int_square(X,Sqr,WF) :- get_binary_choice_wait_flag(int_square,WF,WF2), int_square2(X,Sqr,WF2).
:- block int_square2(-,?,-).
int_square2(X,Sqr,_) :- ground(X),!, Sqr is X*X.
int_square2(X,Sqr,_WF2) :-
  integer_square_root(Sqr,X).

:- assert_must_succeed(( kernel_objects:integer_square_root(0,X),X==0 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(1,X),X==1 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(4,X),X==2 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(49,X),X==7 )).
:- assert_must_succeed(( kernel_objects:integer_square_root(49,X),X==(-7) )).
:- assert_must_fail(( kernel_objects:integer_square_root(5,_) )).
:- assert_must_succeed(( X= 123456789, Y is X*X, kernel_objects:integer_square_root(Y,Z),Z==X)).
:- assert_must_fail(( X= 123456789, Y is 1+X*X, kernel_objects:integer_square_root(Y,_Z))).
:- assert_must_succeed(( X= 12345678900, Y is X*X, kernel_objects:integer_square_root(Y,Z),Z==X)).

integer_square_root(0,Root) :- !, Root = 0.
integer_square_root(Sqr,PMRoot) :- %print(sqrt(Sqr)),nl,
   Sqr>0, Root is truncate(sqrt(Sqr)), Sqr is Root*Root, %print(root(Root)),nl,
  (PMRoot = Root ; PMRoot is -(Root)).

times(int(X),int(Y),int(Times)) :-
    int_times2(X,Y,Times),
    (two_vars_or_more(X,Y,Times) -> clpfd_eq(Times,X * Y) % can have performance problems.
      ; true). % we can compute the value directly

:- assert_must_succeed(exhaustive_kernel_check([commutative],times(int(2),int(3),int(6)))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],times(int(2),int(1),int(2)))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],times(int(2),int(0),int(0)))).
:- assert_must_succeed(exhaustive_kernel_check(times(int(0),int(1),int(0)))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],times(int(2),int(3),int(5)))).
:- assert_must_succeed(exhaustive_kernel_fail_check([commutative],times(int(1),int(3),int(2)))).
:- assert_must_succeed(( int_times2(A,B,C),A=3,B=2,C==6 )).
:- assert_must_succeed(( int_times2(A,B,C),A=3,C=6,B==2 )).
:- assert_must_succeed(( int_times2(A,B,C),B=2,A=3,C==6 )).
:- assert_must_succeed(( int_times2(A,B,C),B=2,C=6,A==3 )).
:- assert_must_succeed(( int_times2(A,B,C),C=6,A=3,B==2 )).
:- assert_must_succeed(( int_times2(A,B,C),C=6,B=2,A==3 )).
:- assert_must_succeed(( int_times2(A,_,C),A=0,C==0 )).
:- assert_must_succeed(( int_times2(_,B,C),B=0,C==0 )).
:- assert_must_succeed(( int_times2(A,B,C),A=1,B==C )).
:- assert_must_succeed(( int_times2(A,B,C),B=1,A==C )).
:- assert_must_succeed(( int_times2(A,1,C),A=2,C==2 )).
:- assert_must_succeed(( int_times2(_A,0,C),C==0 )).
:- assert_must_succeed(( int_times2(A,_,C),C=0,A=0 )).
:- assert_must_succeed(( int_times2(_,B,C),C=0,B=0 )).
:- assert_must_succeed(( int_times2(A,B,0),A=0,B=2 )).
:- assert_must_succeed(( int_times2(A,B,0),B=2,A=0 )).
:- assert_must_succeed(( int_times2(B,A,0),A=0,B=2 )).
:- assert_must_succeed(( int_times2(B,A,0),B=2,A=0 )).
:- assert_must_fail(( int_times2(A,_,C),A=3,C=7 )).
:- assert_must_fail(( int_times2(A,_,C),C=7,A=3 )).
:- assert_must_fail(( int_times2(_,B,C),B=2,C=7 )).
:- assert_must_fail(( int_times2(_,B,C),C=7,B=2 )).
:- assert_must_fail(( int_times2(A,_,C),C=7,A=0 )).
:- assert_must_fail(( int_times2(_,B,C),C=7,B=0 )).
:- assert_must_fail(( int_times2(B,A,0),B=2,A=1 )).

:- block int_times2(-,-,-).
int_times2(X,Y,Times) :-
    ( ground(X) ->
	( X==1      -> Y=Times
	; X==0      -> Times=0
	;              int_times3(X,Y,Times))
    ; ground(Y) ->
	( Y==1      -> X=Times
	; Y==0      -> Times=0
	;              int_times3(Y,X,Times))
    ; int_times4(X,Y,Times)).
% int_times3/3: First argument must be ground when called and non-zero
:- block int_times3(?,-,-).
int_times3(X,Y,Times) :-
    ( ground(Y) -> Times is X*Y
    ;              Y is Times // X, Times is X*Y).
% int_times4/3: Third argument must be ground when called
:- block int_times4(-,-,?).
int_times4(X,Y,Times) :- %print(int_times4(X,Y,Times)),nl,
    ( Times==0 ->
	  ( ground(X) -> (X==0 -> true; Y=0 )
	  ; /* ground(Y) -> */ (Y==0 -> true; X=0 ))
    ;  /* Times /== 0 */
	  ( ground(X) -> X\==0, Y is Times // X, Times is X*Y
	  ; /* ground(Y) -> */ Y\==0, X is Times // Y, Times is X*Y)).


:- assert_must_succeed(exhaustive_kernel_check(int_power(int(2),int(3),int(8),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(2),int(1),int(2),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(3),int(0),int(1),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(1),int(3),int(1),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_check(int_power(int(0),int(3),int(0),unknown,_))).
:- assert_must_succeed(exhaustive_kernel_fail_check(int_power(int(2),int(3),int(6),unknown,_))).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=2,B=5,C==32 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A= -2,B=5,C== -32 )).
%:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=1,B= -5,C==1 )). % now aborts !
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=1,C=1, B= -5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=1,C= 1,B = -5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=2,C=32,B==5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=10,C=1000,B==3 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A= -2,C= -32,B==5 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A= -2,C= 16,B==4 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=2,C=1,B==0 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,B=2,C==0 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,C=0,B=2 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,B=0,C==1 )).
:- assert_must_succeed(( int_power2(A,B,C,unknown,_),A=0,C=1,B==0 )).
:- assert_must_succeed(( int_power2(17,13,C,unknown,_),C==9904578032905937 )).
:- assert_must_succeed((platform_is_64_bit
		       -> int_power2(A,13,C,unknown,_),C=9904578032905937,A=17
		       ;  int_power2(A,9,C,unknown,_),C=134217728,A=8 )).
:- assert_must_fail((platform_is_64_bit
		    -> int_power2(A,13,C,unknown,_),C=9904578032905936,A=17
		    ; int_power2(A,9,C,unknown,_),C=134217727,A=8 )).
:- assert_must_succeed((platform_is_64_bit
		       -> int_power2(A,10,C,unknown,_),C=576650390625,A=15
		       ;  true)).
:- assert_must_fail((platform_is_64_bit
		    -> int_power2(A,10,C,unknown,_),C=576650390626,A=15
		    ;  false)).
:- assert_must_succeed(( int_power2(A,100,C,unknown,_),A=2,C==1267650600228229401496703205376 )).
:- assert_must_fail(( int_power2(A,100,C,unknown,_),C=1267650600228229401496703205375,A=2 )).
:- assert_must_fail(( int_power2(A,100,C,unknown,_),C=1267650600228229401496703205377,A=2 )).

:- assert_must_fail(( int_power2(A,B,C,unknown,_),A=2,B=5,C=33 )).
:- assert_must_abort_wf(( int_power2(A,B,_,unknown,WF),A=2,B= -5 ),WF).
:- assert_must_fail(( int_power2(A,_,C,unknown,_),A= -2,C=32 )).
:- assert_must_fail(( int_power2(A,_,C,unknown,_),A= -2,C= -16 )).

% TODO: calculate X from Y und Pow (i.e., Yth root of Pow); in CLPFD mode this is more or less done
int_power(int(X),int(Y),int(Pow),Span,WF) :-
  ( preferences:preference(use_clpfd_solver,true)
   -> int_power2(X,Y,Pow,Span,WF), int_power_propagation(X,Y,Pow)
    ; int_power1(X,Y,Pow,Span,WF)).
% TO DO ?: if all are variables we can still infer some knowledge
% e.g. if X is positive then Pow must be positive; but it is probably quite rare that we have models with unknown exponent ?
:- block int_power1(-,?,?,?,?). % ensure that Base X is known
int_power1(X,Y,Pow,Span,WF) :- int_power2(X,Y,Pow,Span,WF).
:- block int_power2(-,-,?,?,?), int_power2(?,-,-,?,?).
int_power2(X,Y,Pow,Span,WF) :-
    ( ground(Y) ->
	  ( Y>=0        -> safe_int_power0(X,Y,Pow)
	    ; otherwise -> add_wd_error_set_result('power with negative exponent','**'(X,Y),Pow,1,Span,WF))
    ; otherwise ->  /* X & POW are ground */
		( X==1 -> Pow==1  /* 1**Y = 1 */
		; X==0, Pow==1 -> Y=0
		; X==0 -> Pow==0
		; X>0, Pow>0 ->
			checked_log(X,Y,Pow)
		; X<0, Pow<0 ->
			PosPow is -(Pow),
			NegX is -(X),
			checked_log(NegX,Y,PosPow),
			odd(Y)
		; X<0, Pow>0 ->
			NegX is -(X),
			checked_log(NegX,Y,Pow),
			even(Y))).
checked_log(1,Exp,Pow) :- !,  % the SICStus Prolog log function does not work for Base=1
    Pow=1, less_than_equal_direct(0,Exp).
checked_log(Base,Exp,Pow) :-
    Try is integer(log(Base,Pow)),
    % the calculation might have a rounding error
    ( safe_int_power(Base,Try,Pow) -> Exp=Try
    ; otherwise -> Exp is Try+1, safe_int_power(Base,Exp,Pow)).

:- block even(-).
even(X) :- 0 is X mod 2.
:- block odd(-).
odd(X) :- 1 is X mod 2.

% propagation rules if only one of the args known
:- block int_power_propagation(-,-,-).
int_power_propagation(Base,Exp,Pow) :- Exp==0, var(Base),var(Pow),!,  % B**0 = 1
    Pow = 1.
int_power_propagation(Base,Exp,Pow) :- Exp==1, var(Base),var(Pow),!,  % B**1 = B
    Pow = Base.
int_power_propagation(Base,Exp,Pow) :- Base==1, var(Exp),var(Pow),!,  % 1**E = 1
    Pow = Base.
%int_power_propagation(Base,Exp,Pow) :- number(Base), Base>0,var(Exp),var(Pow),!,
%    clpfd_leq(1,Pow,_). % causes problem with test 305
int_power_propagation(_,_,_). % TO DO: stronger propagation; maybe implement custom CLPFD propagators

% safe exponentiation using the squaring algorithm (CLPFD does not support exponentiation yet)
% Note: in TLA mode 0^0 is undefined according to TLC; for B/Rodin it is 1
safe_int_power0(Base,Exp,Result) :- var(Base),
    Exp>30,!, % Exp>59 % 2**59 no overflow; but everything above that is guaranteed to generate an overflow unless Base is 0 or 1 or -1
    % 3**38 generates overflow; 4**30 generates overflow on 64-bit systems
    % To do: examine whether we should already delay with a smaller or larger exponent
    when(nonvar(Base),safe_int_power(Base,Exp,Result)). % wait until Base is known to avoid CLPFD overflow
safe_int_power0(Base,Exp,Result) :- safe_int_power(Base,Exp,Result).

safe_int_power(_Base,0,Result) :- !, Result = 1.
safe_int_power(Base,Exp,Result) :- ground(Base),!,
   Result is Base^Exp. % new integer exponentiation operator in SICStus 4.3
safe_int_power(Base,Exp,Result) :-
	Msb is msb(Exp),   % most significant bit
	ExpMask is 1<<Msb,
	safe_int_power_clpfd2(ExpMask,Exp,Base,1,Result).

:- use_module(clpfd_interface,[clpfd_eq_expr/2]).
safe_int_power_clpfd2(0,_,_,Prev,Result) :- !, Prev=Result.
safe_int_power_clpfd2(Mask,Exp,Base,Prev,Result) :-
	P is Exp /\ Mask,   % P is Exp's highest bit
	Mask2 is Mask>>1,
	clpfd_eq_expr(Quad,Prev*Prev),
	( P==0      -> Next = Quad
	; otherwise -> clpfd_eq_expr(Next,Quad*Base) ),
	safe_int_power_clpfd2(Mask2,Exp,Base,Next,Result).
 %% -------------------------------------------------------

:- assert_must_succeed(( singleton_set_element([int(1)],E,_WF), E==int(1) )).
:- assert_must_fail(singleton_set_element([int(1)],int(2),_WF) ).
:- assert_must_abort_wf(kernel_objects:singleton_set_element([int(1),int(2)],_E,WF),WF).
% This predicate computes the effect of the MU operator.
% Set should be a singleton set and Elem its only element.
% In case Set is empty  or has more than one element, an error
% message is generated.
% The waitflag store WFStore is needed to obtain a waitflag for the
% decision when to force whether Set is a singleton or not and
% to save the error if necessary.
singleton_set_element(Set,Elem,WFStore) :-
    cardinality_as_int_wf(Set,Card,WFStore),
    get_enumeration_finished_wait_flag(WFStore,LWF),
    equality_objects_lwf(Card,int(1),IsSingleton,LWF),
    singleton_set_element2(IsSingleton,Set,Elem,Card,WFStore).
:- block singleton_set_element2(-,?,?,?,?).
singleton_set_element2(pred_true,Set,Elem,_Card,_WFStore) :-
    exact_element_of(Elem,Set).
singleton_set_element2(pred_false,_Set,_Elem,Card,WFStore) :-
    ( Card=int(C) -> true ;  C = 'unknown'),
    add_wd_error('argument of MU expression must have cardinality 1, but has ', C, WFStore).


%:- print(finished_loading_kernel_objects),nl.
